/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "HistPlot.h"
#include <QDebug>

#include <qwt_plot_curve.h>
#include "dsplib/dsplib.h"
#include "dsplib/fitter.h"

const double eps = 1e-15;

const unsigned int max_display_bins = 500;
const unsigned int curve_segments = 200;

HistPlot::HistPlot(QWidget *parent) : HistPlotBase(parent)
{

}

void HistPlot::setHist(const th1 &r)
{
    h = r;

//    int now_modseq = h.get_modseq();
//    if (modseq == now_modseq) return;

    setWindowTitle(h.getname());

    const double xmin = h.get_xmin ();
    const double xmax = h.get_xmax ();

    const unsigned int nbins = h.get_nbins ();

    if ((nbins == 0) || abs(xmin - xmax) < eps)
        return;

    unsigned int bfirst = 0;
    unsigned int blast = h.get_nbins();
    h.find_autorange_bins(&bfirst, &blast);
    if (bfirst > 0)
        bfirst--;
    if (blast < h.get_nbins())
        blast++;

    const auto &h_px = h.get_px();
    const auto &h_py = h.get_py();
    // Reset histogram arrays
    px.clear();
    py.clear();
    unsigned int bins = blast - bfirst;
    unsigned int binstep = 1;
    if (bins > max_display_bins) {
        binstep = bins / max_display_bins;
    }
    for (unsigned int i = bfirst; i < blast; i+=binstep) {
        double v = 0;
        for (unsigned int j=0; j<binstep; j++)
            v += h_py[i+j];
        px.push_back(h_px[i]);
        py.push_back(v);
    }
    set_x_range(h.binleft(bfirst), h.binright(blast-1));

    // Fit
    const bool fit_enable = h.fit_enabled();
    const bool need_fit = fit_enable && (h.get_nent () > 10);
    fit_ok = false;
    value=0;mean=0;sigma=0;norm=0;
    if (need_fit && !h.get_py().empty()) {
//        int ret = h.fitgaus (&value, &mean, &sigma, &norm);
//        if (ret==0 && norm < 2 && abs(sigma) > eps) fit_ok = true;
        if (!fit_ok) {
            const auto &stats = h.simplestats();
            mean = stats.mean();
            sigma = stats.stddev();
        }
    }

    // Reset fit arrays
    std::vector<double> fpx;
    std::vector<double> fpy;
    if (need_fit) {
        const double pxf_m = 1.0 * (blast-bfirst) / curve_segments;
        for (double x = h.binleft(bfirst); x < h.binright(blast); x += h.binwidth() * pxf_m) {
            fpx.push_back(x);
            fpy.push_back(value*gauss(x-mean, sigma));
        }
    }
    // Reset graphs in HistPlotBase
    setpoints (px, py);
    set_fit_points (fpx, fpy);
    getFitCurve()->setPen( fit_ok ? QPen(Qt::black, 2, Qt::SolidLine) : QPen(Qt::darkGray, 1, Qt::DashLine));

    modseq = h.get_modseq();
    replot ();
}
