/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include "WaveformController.h"

#include <QDebug>
#include <QSettings>
#include <QFileDialog>

#include "waveformmodeladc.h"
#include "qxw/WaveformAverager.h"
#include "qxw/WaveformViewAdvanced.h"
#include "adcmlib/adcm_ev.h"
#include "dsplib/dsplib.h"
#include "dsplib/Convolver.h"

WaveformController::WaveformController(WaveformModel_ADC *m, QObject *parent)
    : QObject(parent),
    model(m),
    averager(new WaveformAverager)
{
}

WaveformController::~WaveformController()
{
    delete averager;
}

void WaveformController::setView(WaveformViewAdvanced *v)
{
    if (view) {
        disconnect(view);
    }
    view = v;
    connect(model, &WaveformModel_ADC::dataChanged,
            this, &WaveformController::modelChanged);

    connect(view, &WaveformViewAdvanced::FFTChanged,
            this, &WaveformController::setFFT);

    connect(view, &WaveformViewAdvanced::filteringChanged,
            this, &WaveformController::setFiltering);

    connect(view, &WaveformViewAdvanced::averagingChanged,
            this, &WaveformController::setAveraging);

    connect(view, &WaveformViewAdvanced::offsetCompChanged,
            this, &WaveformController::setOffsetComp);

    connect(view, &WaveformViewAdvanced::actionClear,
            this, &WaveformController::actionClear);

    connect(view, &WaveformViewAdvanced::actionExport,
            this, &WaveformController::actionExport);
}

void WaveformController::modelChanged()
{
    const auto &data = getPlotData();
    view->setWaveformData(data);
}

void WaveformController::updateView()
{
    modelChanged();
}

void WaveformController::updateViewControls()
{
    view->updateControls();
}

void WaveformController::setOffsetComp(bool value)
{
    offsetComp = value;
    averager->clear();
    updateView();
}

void WaveformController::setFiltering(bool value)
{
    selectFilter = value;
    averager->clear();
    updateView();
}

void WaveformController::setAveraging(bool value)
{
    enableAverager = value;
    averager->clear();
    updateView();
}

void WaveformController::setFFT(bool value)
{
    enableFFT = value;
    averager->clear();
    updateViewControls();
    updateView();
}

void WaveformController::actionClear()
{
    averager->clear();
    view->clear();
}

void WaveformController::actionExport()
{
    const QString settingsKey = "currentWaveformExportDirName";
    QSettings settings;
    const QString &dirNameSettings = settings.value(settingsKey).toString();
    const QString &dirName =
        QFileDialog::getExistingDirectory(view, tr("Export to directory..."), dirNameSettings);
    if (dirName.isNull())
        return;
    settings.setValue(settingsKey, dirName);

    const WaveformDataAll &wda = getPlotData();
    for (auto index : wda) {
        const WaveformData &wd = index.second;
        QString fileName = QString("waveform_chanel_%1.txt").arg(wd.c);
        QFile f(dirName + "/" + fileName);
        if (!f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
            qWarning() << QString("Can't write file '%1'").arg(fileName);
            return;
        }
        QTextStream ts(&f);
        ts << QString("# channel %1").arg(wd.c) << "\n";
        for (size_t i=0; i<wd.p.size(); i++)
            ts << wd.p[i].x << " " << wd.p[i].y << "\n";
        f.close();
    }
}

void WaveformController::setFilterData(int channel, const fir_t<decoder_fir_coef_t> &fir )
{
    //    if (d->filter.size() < size_t(channel+1)) d->filter.resize(channel+1);
    filter[channel] = fir;
}

void WaveformController::setFullScale( int channel, double value )
{
    fullScale[channel] = value;
}

WaveformDataAll WaveformController::makeWaveform(const WaveformDataAll &wda_in) const
{
    WaveformDataAll wda(wda_in);
    if (offsetComp) {
        for (auto & it : wda) {
            WaveformData &wd = it.second;
            for(size_t i=0; i<wd.p.size(); i++)
                wd.p[i].y -= wd.baseline;
        }
    }
//    if (selectFilter) {
//        for (auto &it : wda) {
//            WaveformData &wd = it.second;
//            size_t len = wd.p.size();
//            if (len<1) continue;
//            int c = wd.c;
//            if (c >= filter.size()) continue;
//            const auto &fir = filter[c];
//            double step = wd.p[1].x - wd.p[0].x;
//            double normfactor = 1. / (1 << fir.bits);
//            std::vector<double> v(len);
//            for(size_t i=0; i<len; i++)
//                v[i] = wd.p[i].y;
//            const auto &firCoef = fir.coef;
//            std::vector<double> u;
//            if (firCoef.empty())
//                u.assign(v.begin(), v.end());
//            else {
//                normfactor = 1. / (1 << fir.bits);
//                u = Convolver::convolve(v, firCoef);
//            }
//            for(size_t i=0; i<len; i++) {
//                wd.p[i].x -= fir.offs * step;
//                wd.p[i].y = u[i] * normfactor;
//            }
//            //  		wd[n].p[i].y -= wd[n].baseline;
//        }
//    }
    if (enableFFT) {
        wda = waveformFilterFFT.filterAll(wda);
    }
    return wda;
}

WaveformDataAll WaveformController::linear_to_dBFS(const WaveformDataAll &input) const
{
    WaveformDataAll wda(input);
    for (WaveformDataAll::iterator it = wda.begin(); it != wda.end(); ++it) {
        WaveformData &wd = (*it).second;
        const double fullScaleSigned = fullScale[wd.c];
        const double fullScaleDb = val2db(fullScaleSigned);
        for(size_t i=0; i< wd.p.size(); i++) {
            wd.p[i].y = val2db(wd.p[i].y) - fullScaleDb;
        }
    }
    return wda;
}

WaveformDataAll WaveformController::getPlotData() const
{
    auto modelData = selectFilter ? model->get_conv() : model->get_raw();
    WaveformDataAll wda = makeWaveform(modelData);
    if (enableAverager)
        averager->add(wda);
    else
        averager->clear();

    if (enableAverager) {
        wda = averager->get();
    }
    if (enableFFT) {
        wda = linear_to_dBFS(wda);
    }
    return wda;
}
