/*
**    Copyright 2010-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <cstring>
#include <cmath>

#include <QTimer>
#include <QAbstractTableModel>
#include <QMap>
#include <QMessageBox>

#include "config/adcconfig.h"
#include "config/calavgconfig.h"
#include "calavgdialog.h"

namespace {
enum {
    TABLE_COL_CHANNEL = 0,
    TABLE_COL_COUNT,
    TABLE_COL_OFFSET,
    TABLE_COL_PREINT,
    TABLE_COL_INT,
    TABLE_COL_POSTINT,
    TABLE_N_COLS
};
}

class CalAvgModel: public QAbstractTableModel
{
public:
    explicit CalAvgModel(QObject *parent = nullptr)
        : QAbstractTableModel(parent) {}
    virtual int rowCount(const QModelIndex &parent = QModelIndex()) const;
    virtual int columnCount(const QModelIndex &parent = QModelIndex()) const;
    virtual QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;
    virtual Qt::ItemFlags flags(const QModelIndex &index) const;
    virtual QVariant headerData(int section, Qt::Orientation orientation,
                                int role = Qt::DisplayRole) const;
    void setCustomData(const AdcConfig &adcConfig, const CalAvgConfig &calAvgConfig);
private:
    QMap<int, QMap<int, QString> > d;
    QMap<int, QMap<int, bool> > d_enable;
};

int CalAvgModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return d[TABLE_COL_COUNT].size();
}

int CalAvgModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return TABLE_N_COLS;
}

QVariant CalAvgModel::data(const QModelIndex &index, int role) const
{
    if (role == Qt::DisplayRole) {
        return d[index.column()][index.row()];
    }
    return QVariant();
}

Qt::ItemFlags CalAvgModel::flags(const QModelIndex &index) const
{
    return d_enable[index.column()][index.row()] ? Qt::ItemIsEnabled : Qt::NoItemFlags;
}

QVariant CalAvgModel::headerData(int section, Qt::Orientation orientation,
                                 int role) const
{
    if (role != Qt::DisplayRole)
        return QVariant();
    if (orientation == Qt::Horizontal) {
        switch(section) {
        case TABLE_COL_COUNT : return "Count";
        case TABLE_COL_CHANNEL : return "Channel";
        case TABLE_COL_OFFSET : return "Offset";
        case TABLE_COL_PREINT : return "Pre-Int";
        case TABLE_COL_INT : return "Integral";
        case TABLE_COL_POSTINT : return "Post-Int";
        }
    }
    if (orientation == Qt::Vertical) {
        return QString::number(section);
    }
    return QVariant();
}

void CalAvgModel::setCustomData(const AdcConfig &adcConfig, const CalAvgConfig &calAvgConfig)
{
    QMap<int, QMap<int, QString> > newData;
    QMap<int, QMap<int, bool> > newEnable;
    int row = 0;
    for (auto i : calAvgConfig.channels.keys()) {
        const bool enable = adcConfig.ch_enabled_shape_calibration(i);
        const int canent = calAvgConfig.channels[i].get_nent();
//        const double base    = cfg.calavg_baseline(i);
//        const double inte    = cfg.calavg_integral(i);
//        const double preint  = cfg.calavg_preintegral(i);
//        const double postint = cfg.calavg_postintegral(i);
        for (int i=0; i<columnCount(); i++) {
            newEnable[i][row] = enable;
        }

        newData[TABLE_COL_CHANNEL][row] = QString("%1").arg(i);
        newData[TABLE_COL_COUNT][row] = QString("%1").arg(canent);
//        newMap[table_col_offset][row] = QString("%1").arg(base, 0, 'f', 1);
//        newMap[table_col_int][row] = QString("%1").arg(inte, 0, 'f', 1);
//        newMap[table_col_preint][row] = QString("%1").arg(preint, 0, 'f', 1);
//        newMap[table_col_postint][row] = QString("%1").arg(postint, 0, 'f', 1);
        row++;
    }
    if (d != newData || d_enable != newEnable) {
        beginResetModel();
        d = newData;
        d_enable = newEnable;
        endResetModel();
//        reset();
    }
}

CalAvgDialog::CalAvgDialog (QWidget * parent)
    : CalibrationDialog(parent),
      d_model(new CalAvgModel(this))
{
    setupUi(this);
    tableView->setModel(d_model);
    tableView->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
}

void CalAvgDialog::update ()
{
    reinterpret_cast<CalAvgModel*>(d_model)->setCustomData(*adcConfig, *calAvg);
}

void CalAvgDialog::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Reset:
        clearEnabled();
        break;
    case QDialogButtonBox::Apply:
        emit calAvgConfigChanged(*calAvg);
        break;
    default:;
    }
}

void CalAvgDialog::clearEnabled()
{
    QStringList list;
    for (auto c : adcConfig->channels.keys()) {
        if (adcConfig->ch_enabled_shape_calibration (c) && !calAvg->channels[c].isEmpty())
            list.push_back(QString("%1").arg(c));
    }
    if (list.isEmpty())
        return;
    switch (QMessageBox::warning (this, "Reset Shape Calibration?",
                                 "Do you want to reset average waveform shape calibrations for channels:\n\n"
                                     + list.join(", "),
                                 QMessageBox::Yes,
                                 QMessageBox::No)) {
    case
        QMessageBox::Yes:
        break;
    default:
        return;
    }
    calAvg->clear_enabled(*adcConfig);
    emit calAvgConfigChanged(*calAvg);
}
