/*
**    Copyright 2010-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <cmath>

#include <QDebug>
#include <QTimer>
#include <QLabel>
#include <QPixmap>
#include <QTableWidget>
#include <QPushButton>
#include <QSettings>
#include <QFileDialog>
#include <QFile>
#include <QTextStream>
#include <QRegExp>
#include <QMessageBox>

#include "adcmraw/adcm_mem.h"
#include "config/adcconfig.h"
#include "config/calzeroconfig.h"

#include "calzerodialog.h"

namespace {
enum {
    TABLE_COL_CHANNEL = 0,
    TABLE_COL_OFFSET,
    TABLE_COL_PERCENT,
    TABLE_COL_RMS,
    TABLE_COL_RMS_DB,
    TABLE_N_COLS
};
}

class CalZeroModel: public QAbstractTableModel
{
public:
    explicit CalZeroModel(QObject *parent = nullptr)
        : QAbstractTableModel(parent) {}
    int rowCount(const QModelIndex &parent = QModelIndex()) const override;
    int columnCount(const QModelIndex &parent = QModelIndex()) const override;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const override;
    Qt::ItemFlags flags(const QModelIndex &index) const override;
    QVariant headerData(int section, Qt::Orientation orientation,
                        int role = Qt::DisplayRole) const override;
    void setCustomData(const conf_mem_t &conf_mem, const AdcConfig &adcConfig, const CalZeroConfig &calZeroConfig);
    static bool adc_baseline_good(const conf_mem_t &conf_mem, const CalZeroConfig &calZero, int ch);
    static bool adc_noiselevel_good(const conf_mem_t &conf_mem, const CalZeroConfig &calZero, int ch);
private:
    QMap<int, QMap<int, QString> > d;
    QMap<int, QMap<int, bool> > d_enable;
    QMap<int, QMap<int, QVariant> > d_check;
    QMap<int, QMap<int, QBrush> > d_bg;
};

int CalZeroModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return d[TABLE_COL_CHANNEL].size();
}

int CalZeroModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return TABLE_N_COLS;
}

QVariant CalZeroModel::data(const QModelIndex &index, int role) const
{
    switch(role) {
    case Qt::DisplayRole:
        return d[index.column()][index.row()];
    case Qt::BackgroundRole:
        return d_bg[index.column()][index.row()];
    case Qt::CheckStateRole:
        return d_check[index.column()][index.row()];
    }
    return QVariant();
}

Qt::ItemFlags CalZeroModel::flags(const QModelIndex &index) const
{
    return d_enable[index.column()][index.row()] ? Qt::ItemIsEnabled : Qt::NoItemFlags;
}

QVariant CalZeroModel::headerData(int section, Qt::Orientation orientation,
                                 int role) const
{
    if (role != Qt::DisplayRole)
        return QVariant();
    if (orientation == Qt::Horizontal) {
        switch(section) {
        case TABLE_COL_CHANNEL : return "Channel";
        case TABLE_COL_OFFSET : return "Offset";
        case TABLE_COL_PERCENT : return "Offset, %";
        case TABLE_COL_RMS : return "RMS noise";
        case TABLE_COL_RMS_DB : return "noise, dBFS";
        }
    }
    if (orientation == Qt::Vertical) {
        return QString::number(section);
    }
    return QVariant();
}

CalZeroDialog::CalZeroDialog (QWidget * parent)
    : CalibrationDialog(parent),
      d_model(new CalZeroModel(this))
{
    setupUi(this);
    tableView->setModel(d_model);
    tableView->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
    tableView->verticalHeader()->hide();
    QPushButton *b = buttonBox->addButton("Export...", QDialogButtonBox::ActionRole);
    connect(b, &QPushButton::clicked, this, &CalZeroDialog::exportAction);
    connect(buttonBox, &QDialogButtonBox::clicked, this, &CalZeroDialog::buttonBox_clicked);
}

bool CalZeroModel::adc_baseline_good(const conf_mem_t &conf_mem, const CalZeroConfig &calZero, int ch)
{
    double val = calZero.get(ch);
    int halfRange = conf_mem.getAdcHalfRange(ch);
    double mid = conf_mem.getAdcMidValue(ch);
    double tolerance = 0.05; // relative to half-scale
    double max = mid + halfRange * tolerance;
    double min = mid - halfRange * tolerance;
    //  printf("bl_good: %f < %f < %f\n", min, val, max);
    return (min < val && val < max);
}

bool CalZeroModel::adc_noiselevel_good(const conf_mem_t &conf_mem, const CalZeroConfig &calZero, int ch)
{
    double val = sqrt(calZero.getcalnoiserms(ch)) * 1000. / conf_mem.getAdcHalfRange(ch);
    //  printf("adc_noiselevel_good=%f\n", val);
    if (abs(val) < 1e-12) return false; // failure
    return (0.05 < val && val < 3.0);
}

void CalZeroModel::setCustomData(const conf_mem_t &conf_mem, const AdcConfig &adcConfig, const CalZeroConfig &calZeroConfig)
{
    const int nch = conf_mem.nch; // calZero->channels.size();
    QList<int> channels;
    for (int i=0; i<nch; i++) {
        channels.push_back(i);
    }
    for (auto i : adcConfig.channels.keys()) {
        if (!channels.contains(i))
            channels.push_back(i);
    }
    for (auto i : calZeroConfig.channels.keys()) {
        if (!channels.contains(i))
            channels.push_back(i);
    }

    QMap<int, QMap<int, QString> > new_data;
    QMap<int, QMap<int, bool> > new_enable;
    QMap<int, QMap<int, QVariant> > new_check;
    QMap<int, QMap<int, QBrush> > new_bg;
    int row = 0;
    for (auto i : channels) {
        bool hasHwChannel = (i < conf_mem.nch);
        int adcmid = conf_mem.getAdcMidValue(i);
        int adcHalfRange = conf_mem.getAdcHalfRange(i);
        if (adcHalfRange <= 0)
            adcHalfRange = 1;
        bool enable = hasHwChannel && adcConfig.ch_enabled_offset_calibration(i);
        const CalZeroChannelConfig &channel = calZeroConfig.channels[i];
        bool empty = channel.isEmpty();
        if (empty)
            continue;
        double base = channel.get_base() - adcmid;
        double base_pcnt = base/(adcHalfRange)*100.;
        double noise = sqrt(channel.get_rms());
        double max_signal = adcHalfRange / sqrt(2);
        double noise_dbfs = 20 * log10(noise / max_signal);
        bool offset_ok = !hasHwChannel || empty || adc_baseline_good(conf_mem, calZeroConfig, i);
        bool noise_ok = !hasHwChannel || empty || adc_noiselevel_good(conf_mem, calZeroConfig, i);

        new_data[TABLE_COL_CHANNEL][row] = QString::number(i);
        new_data[TABLE_COL_OFFSET][row] = QString("%1").arg(base, 0, 'f', 1);
        if (hasHwChannel)
            new_data[TABLE_COL_PERCENT][row] = QString("%1").arg(base_pcnt, 0, 'f', 2);
        else
            new_data[TABLE_COL_PERCENT][row] = QString();
        new_data[TABLE_COL_RMS][row] = QString("%1").arg(noise, 0, 'f', 2);
        new_data[TABLE_COL_RMS_DB][row] = QString("%1").arg(noise_dbfs, 0, 'f', 2);

        for (int i=0; i<columnCount(); i++) {
            new_enable[i][row] = enable;
        }
        new_bg[TABLE_COL_OFFSET][row] = offset_ok ? QBrush() : QBrush(Qt::red);
        new_bg[TABLE_COL_PERCENT][row] = new_bg[TABLE_COL_OFFSET][row];
        new_bg[TABLE_COL_RMS][row] = noise_ok ? QBrush() : QBrush(Qt::red);
        new_bg[TABLE_COL_RMS_DB][row] = new_bg[TABLE_COL_RMS][row];

//        new_check[TABLE_COL_CHANNEL][row] = enable ? Qt::Checked : Qt::Unchecked;

        row++;
    }
    if (d != new_data || d_enable != new_enable || d_bg != new_bg || d_check != new_check) {
        beginResetModel();
        d = new_data;
        d_enable = new_enable;
        d_bg = new_bg;
        d_check = new_check;
        endResetModel();
    }
}

void CalZeroDialog::update ()
{
    reinterpret_cast<CalZeroModel*>(d_model)->setCustomData(*conf_mem, *adcConfig, *calZero);
}

void CalZeroDialog::exportAction()
{
    QSettings settings;
    QString fileName = settings.value("currentBaselineCalibrationFileName").toString();
    if (fileName.isEmpty()) {
        QString str = conf_mem->getMasterTypeStr();
        QRegExp re("[^A-Za-z0-9 -]");
        str = str.replace(re, "_");
        fileName = QString("%1_%2_adc_baseline.csv")
                .arg(str)
                .arg(conf_mem->getMasterSerial());
    }
    fileName = QFileDialog::getSaveFileName(this, "Export as...", fileName,
                                                       "CSV files (*.csv);;Text files (*.txt);;All files (*)");

    if (fileName.isNull())
        return;

    QFile f(fileName);
    if (!f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
        qWarning() << QString("Can't write file '%1'").arg(fileName);
        return;
    }
    QTextStream ts(&f);
    ts << "Channel,Baseline bits,Baseline %,Noise RMS bits,Noise dBFS" << "\r\n";
    for (auto i : calZero->channels.keys()) {
        const CalZeroChannelConfig &channel = calZero->channels[i];
        int adcmid = conf_mem->getAdcMidValue(i);
        int adcHalfRange = conf_mem->getAdcHalfRange(i);
        if (adcHalfRange <= 0)
            adcHalfRange = 1;
        if (!channel.isEmpty()) {
            double base = channel.get_base() - adcmid;
            double base_pcnt = base/(adcHalfRange)*100.;
            double noise = sqrt(channel.get_rms());
            double max_signal = adcHalfRange / sqrt(2);
            double noise_dbfs = 20 * log10(noise / max_signal);
            ts << i << "," << base << "," << base_pcnt << "," << noise << "," << noise_dbfs << "\r\n";
        }
    }
    f.close();
}

void CalZeroDialog::buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Reset :
        clearEnabled();
        break;
    case QDialogButtonBox::Apply :
        emit calZeroConfigChanged(*calZero);
        break;
    default:;
    }
}

bool CalZeroDialog::enabled_ch_has_calzero() const
{
    bool ok = false;

    for (auto c : calZero->channels.keys()) {
        if (adcConfig->ch_enabled_offset_calibration (c) && !calZero->channels[c].isEmpty())
            ok = true;
    }
    return ok;
}

void CalZeroDialog::clearEnabled()
{
    QStringList list;
    for (auto c : adcConfig->channels.keys()) {
        if (enabled_ch_has_calzero() && !calZero->channels[c].isEmpty())
            list.push_back(QString("%1").arg(c));
    }
    if (list.isEmpty())
        return;
    switch (QMessageBox::warning (this, "Reset Baseline Calibration?",
                                 "Do you want to reset baseline calibrations for channels:\n\n"
                                     + list.join(", "),
                                 QMessageBox::Yes,
                                 QMessageBox::No)) {
    case
        QMessageBox::Yes:
        break;
    default:
        return;
    }
    calZero->clear_enabled(*adcConfig);
    emit calZeroConfigChanged(*calZero);
}
