//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QPushButton>
#include <QSpinBox>
#include <QCheckBox>
#include <QComboBox>
#include <QDebug>

#include "fittersettingsdialog.h"

#include "adcmlib/bitop.h"
#include "dsplib/fitter.h"
#include "config/decoderconfig.h"
#include "config/allconfig.h"

class FitterSettingsDialogPrivate
{
public:
    FitterSettingsDialogPrivate();
    bool modified() const;
    FitProfile currentData;
    FitProfile newData;
    QVector<FitProfileEnum_t> prof_values;
    QMap<int, QString> prof_name;
};

FitterSettingsDialogPrivate::FitterSettingsDialogPrivate()
{
    // prof_values index matches QComboBox index
    prof_values.push_back(FIT_PROFILE_PRECISION);
    prof_values.push_back(FIT_PROFILE_OPTIMAL);
    prof_values.push_back(FIT_PROFILE_CUSTOM);
    prof_values.push_back(FIT_PROFILE_FAST);
    prof_values.push_back(FIT_PROFILE_SUPERFAST);
    prof_name[FIT_PROFILE_PRECISION] = "Accurate";
    prof_name[FIT_PROFILE_OPTIMAL] = "Optimal";
    prof_name[FIT_PROFILE_FAST] = "Fast";
    prof_name[FIT_PROFILE_SUPERFAST] = "Superfast";
    prof_name[FIT_PROFILE_CUSTOM] = "Custom";
}

FitterSettingsDialog::FitterSettingsDialog(QWidget * parent) :
    QDialog(parent),
    d_ptr(new FitterSettingsDialogPrivate())
{
    Q_D(FitterSettingsDialog);
    setupUi(this);
    setbuttons();
    sb_ni->setMinimum(1);
    sb_ni->setMaximum(100);
    sb_np->setMinimum(0);
    sb_np->setMaximum((MAX_FITGAUS_POINTS-1)/2);

    for(int i=0; i<d->prof_values.size(); i++) {
        const int prof_value = d->prof_values[i];
        comboBox_prof->insertItem(i, d->prof_name[prof_value], prof_value);
    }
    Restore();
}

bool FitterSettingsDialogPrivate::modified() const
{
    return currentData != newData;
}

void FitterSettingsDialog::setData(const FitProfile &data)
{
    Q_D(FitterSettingsDialog);
    d->newData = data;
    Restore();
}

//FitProfile FitterSettingsDialog::getData() const
//{
//    Q_D(const FitterSettingsDialog);
//    return d->newData;
//}

void FitterSettingsDialog::Restore()
{
    Q_D(const FitterSettingsDialog);
    if (restoring) return;
    restoring = true;

    for(int index=0; index < d->prof_values.size(); index++) {
        const int prof_value = d->prof_values[index];
        if (prof_value == d->newData.fprof) {
            comboBox_prof->setCurrentIndex(index);
        }
    }
    RestoreDetails();
    restoring = false;
    setbuttons();
}

bool FitterSettingsDialog::custom_profile() const
{
    Q_D(const FitterSettingsDialog);
    return d->newData.fprof == FIT_PROFILE_CUSTOM;
}

void FitterSettingsDialog::update_sb_np()
{
    Q_D(const FitterSettingsDialog);
    if (custom_profile()) {
        sb_np->setValue(d->newData.fitside);
////        sb_np->setValue(cfg.get_dec_fitside(0));
//        sb_np->setValue(AllConfig::get_dec_fitside(d->adcConfig.step, d->newData, 0));
    } else {
//        sb_np->setValue(AllConfig::get_dec_fitside(d->adcConfig.step, d->newData, 0));
        sb_np->setValue(0); // special text
    }
}

void FitterSettingsDialog::RestoreDetails()
{
    Q_D(const FitterSettingsDialog);
    switch(d->newData.get_dec_fitalgo()) {
    case FIT_ALGO_ACCURATE:
        rb_slow->setChecked(true);
        break;
    case FIT_ALGO_COARSE:
        rb_fast->setChecked(true);
        break;
    }

    sb_ni->setValue(d->newData.get_dec_fititer());
    update_sb_np();
    cb_fit_a->setChecked(!(d->newData.get_dec_fitfixpar()&1));
    //  cb_fit_t->setChecked(!(cfg.getdec_fitfixpar()&2));
    cb_fit_w->setChecked(!(d->newData.get_dec_fitfixpar()&4));
}

void FitterSettingsDialog::setFitAlgorithm( int v)
{
    Q_D(FitterSettingsDialog);
    if (!custom_profile()) return;
    d->newData.fitalgo = v;
    setbuttons();
}


void FitterSettingsDialog::on_sb_ni_valueChanged( int v)
{
    Q_D(FitterSettingsDialog);
    if (custom_profile())
        d->newData.fititer = v;
    else
        d->newData.fititer = d->currentData.fititer;
    setbuttons();
}


void FitterSettingsDialog::on_sb_np_valueChanged( int v)
{
    Q_D(FitterSettingsDialog);
    if (custom_profile())
        d->newData.fitside = (v > 0) ? v : d->newData.fitside;
    else
        d->newData.fitside = d->currentData.fitside;
    setbuttons();
}


void FitterSettingsDialog::on_cb_fit_a_toggled( bool st)
{
    Q_D(FitterSettingsDialog);
    if (custom_profile())
        set_bit_value(&d->newData.fitfixpar, 1u, !st);
    else
        set_bit_value(&d->newData.fitfixpar, 1u, d->currentData.fitfixpar);
    setbuttons();
}

void FitterSettingsDialog::on_cb_fit_w_toggled( bool st)
{
    Q_D(FitterSettingsDialog);
    if (custom_profile())
        set_bit_value(&d->newData.fitfixpar, 4u, !st);
    else
        set_bit_value(&d->newData.fitfixpar, 4u, d->currentData.fitfixpar);
    setbuttons();
}

void FitterSettingsDialog::accept()
{
    Q_D(FitterSettingsDialog);
    emit modified(d->newData);
    d->currentData = d->newData;
    QDialog::accept();
}

void FitterSettingsDialog::apply()
{
    Q_D(FitterSettingsDialog);
    emit modified(d->newData);
    d->currentData = d->newData;
    setbuttons();
}

void FitterSettingsDialog::setbuttons ()
{
    Q_D(const FitterSettingsDialog);
    buttonBox->button(QDialogButtonBox::Apply)->setEnabled(d->modified());
    buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(d->newData != FitProfile());
    frame_custom->setEnabled(d->newData.fprof == FIT_PROFILE_CUSTOM);
}


void FitterSettingsDialog::restoreDefaults()
{
    Q_D(FitterSettingsDialog);
    d->newData = FitProfile();
    Restore();
}

void FitterSettingsDialog::on_comboBox_prof_activated( int index )
{
    Q_D(FitterSettingsDialog);
    d->newData.fprof = d->prof_values[index];
    RestoreDetails();
    setbuttons();
}

void FitterSettingsDialog::on_rb_fast_clicked()
{
    setFitAlgorithm(FIT_ALGO_COARSE);
}

void FitterSettingsDialog::on_rb_slow_clicked()
{
    setFitAlgorithm(FIT_ALGO_ACCURATE);
}

void FitterSettingsDialog::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Apply :
        apply();
        break;
    case QDialogButtonBox::RestoreDefaults :
        restoreDefaults();
        break;
    default:;
    }
}
