//
//    Copyright 2010-2019 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <vector>

#include <QObject>
#include <QMutex>
#include <QDateTime>

#include "ui_mainwindowbase.h"
#include "sysmon.h"
#include "runcontroladcm.h"
#include "adcmlib/adcm_dec.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmlib/runstats.h"
#include "adcmraw/adcm_mem.h"
#include "adcm_df.h"
#include "config/allconfig.h"

class AdcConfig;
class AdcInstance;
class AdcmApp;
class AdcmDevice;
class AllConfig;
class CalAvgDialog;
class CalZeroDialog;
class EvStorageConfigDialog;
class RawStorageConfigDialog;
class RunConfigDialog;
class ConfigManager;
class DecodeManager;
class DecoderConfig;
class EventListFiller;
class FitterSettingsDialog;
class ProgramArguments;
class QLabel;
class QSettings;
class QSpinBox;
class QTimer;
class RawDataWriter;
class RemoteControlServer;
class RunConfig;
class StorageConfig;
class AdcConfigDialog;
class WaveformModel_ADC;
class WaveformController;
class conf_mem_t;
class DeconvConfigDialog;
struct ReadoutStats;

class MainWindow : public QMainWindow, public Ui::MainWindowBase
{
    Q_OBJECT

public:
    Q_DISABLE_COPY(MainWindow)

    explicit MainWindow(AdcmApp *app, const ProgramArguments &pa, QWidget *parent = nullptr);
    void connectSignals();
    void status (const QString &str);
    void initialize();
    void initialize_gui();

private:
    QString getSaveFileName();
    QSharedPointer<QSettings> readSettingsFile(const QString &fileName = QString());
    void doSaveWork(const QString &fileName);
//    QString currentFileName() const;
//    void loadAutoSave();
    void RestoreAll();
    void DecPermChange();
    void TrigRestore();
    void StorRestore();
    void DecRestore(const DecoderConfig &dec);
    void display_refresh_trig_control(const AdcConfig &adc);
//    void append_decoder_statistics();
//    void display_refresh_run_statistics();
    void clearStats();
    //    double bursttime();
    QString test_report_str(); //Unused
    bool changeStateSynchronous(daq_state_t); //Unused
    bool checkStopCondition();

    void reset_threads();
    taskret_t task_off();
    taskret_t task_reset();
    taskret_t task_run();
    taskret_t task_pause();
    taskret_t task_error();
    taskret_t task_failure();
    void clear_wf_view();
    bool histograms_tab_shown() const;
    bool oscilloscope_tab_shown() const;
    bool live_statistics_tab_shown() const;
    void update_info();
    void updateWaveformViewFilters(const conf_mem_t &conf_mem, const DecoderConfig &dec);
    void readGuiSettings(const QSettings *settings);
    void writeGuiSettings(QSettings *settings);
    void updateWindowTitle(const QString &);
    void setDecoderMode(decoder_mode_t mode );

Q_SIGNALS:
    void stopReadout();
    void startReadout();

    void startRemoteControl(bool);
    void updateRemoteState(QString curState);

    void performanceInfoModified(QString);

    void setEvStorageConfig(const StorageConfig &);
    void setRawStorageConfig(const StorageConfig &);

///////////////////////// Start
public Q_SLOTS:
    void powerOff();
    void transitionTask( daq_state_t );
    void adcMatchWindowChanged(const AdcConfig &adcConfig);
    void setAdcConfMem(const conf_mem_t &);
    void setAdcConfig(const AdcConfig &adcConfig);
    void setDecoderConfig(const DecoderConfig &decoderConfig);
    void setRunConfig(const RunConfig &runConfig);
    void updateEventWriterTimer(const StorageConfig &stor);
    void updateRawWriterTimer(const StorageConfig &stor);
    void setConfigFileName(const QString &s);

protected:
    void changeEvent(QEvent *e) override;
    void closeEvent(QCloseEvent *event) override;

private Q_SLOTS:
    // Autoconnected Menu Actions
    void on_fileNewAction_triggered();
    void on_fileOpenAction_triggered();
    void on_fileSaveAction_triggered();
    void on_fileSaveAsAction_triggered();
    void on_filePrintAction_triggered() {} //no invocation
    void on_fileExitAction_triggered();

    void on_setupFitter_setupAction_triggered();
    void on_setupDecoder_setupAction_triggered();
    void on_fileRawFileOutput_triggered();
    void on_fileDecEvFileOutput_triggered();
    void on_actionAdcConfiguration_triggered();
    void on_actionRunConfiguration_triggered();

    void on_statisticsBaselineOffsetAction_triggered();
    void on_statisticsPulseShapeAction_triggered();

    void on_optionsOptionsAction_triggered();

    void on_helpIndexAction_triggered() {}  //no invocation
    void on_helpContentsAction_triggered() {}  //no invocation
    void on_helpAboutAction_triggered();
    void on_helpAboutQtAction_triggered();

    // Autoconnected slots
    void on_rbDecRec_clicked(); //Full
    void on_rbDecAvg_clicked(); //Shape
    void on_rbDecOffset_clicked(); //Offset
    void on_rbDecOff_clicked(); //Off
    void on_checkBoxHistogramFilling_toggled(bool checked);
    void on_groupBox_dt_cut_toggled(bool checked);
    void on_awl_valueChanged( int value ); //dT Cut left
    void on_awr_valueChanged( int value ); //dT Cut right
    void on_sbPW_valueChanged(double value); //pre-threshold //disable_cfg_slots
    void on_sbMW_valueChanged(double value); //Match Window //disable_cfg_slots
    void on_sbLat_valueChanged(int value); // Data latency //disable_cfg_slots
    void on_spinBoxSamplingLength_valueChanged(int value); // TODO
    void on_tabWidget2_currentChanged( int );
    void on_pushButtonClearDecStat_clicked();
///////////////////////// End
    // Internal Timer slots
//    void autoSave();
    void updateBaseStats();
    void updateCPUStats();
    void updateRawStorageStats();
    void updateEvStorageStats();
    void update_display();
//    void update_scope_event();
    void simulate_event();
    void update_mainhist_view();
    void readoutStatsChanged(const ReadoutStats &);

    // Other Internal slots
    void displayDecodedEvent(const AdcmEvent &ev);
    void decoderProgress_changed(double percent);
    void eventsProceed_changed(quint64 evNumbers);
    void updateRuntime(double);
    void updateRunBytes(quint64);
    void updateTemperature(QVector<double> v);
    void updateStatusLabel();
    void onDeviceConnected();
    void onDeviceDisconnected();
    void onDeviceStatusTextChanged(const QString &str);
    void onDeviceReadoutStateChanged(bool state);

    // decoder
    void onDecoderStatisticsUpdated(const adcm_decstat_t &);
    void onHistListModified(const HistList &hl);

    void rcStartRequest();
    void rcStopRequest();

private:
    AdcmApp * const adcmApp;
    // shortcuts to adcmApp fields
    ConfigManager * const configManager;
    AdcInstance * const adcInstance;
    DecodeManager * const decodeManager;
    EventListFiller * const filler;
    RawDataWriter * const writeInstance;
    RemoteControlServer * const remoteControlServer;

    // new GUI elements
    QLabel * const idLabel;
    QLabel * const statusLabelDeviceState;
    QLabel * const temperatureLabel;

    int temperatureOverflow = 0;
//    QString lastSaveFileName; // Path to last file saved via File->Save/Save as...
//    QString lastFileOpenFileName; // Path to last file opened via File->Open; if valid, writes to lastSaveFileName
    SystemMonitor sysmon;
    const QString defaultConfigFileName;
    const QString defaultConfigDir;
    QString initialConfigFileName; // Uses only on start-up stage
    RunControlAdcm runControl;
    WaveformModel_ADC * const waveformModel_ADC;
    WaveformController * const waveformController;
    runstats_t rst;
    adcm_decstat_t sumdecst;
    adcm_decstat_t rateDecSt, rateDecStLast; //filling rate stat, last/completed rate stat
    QTime rateDecStTime; // start time of filling rateDecSt
    int rateDecStLastMs = 0; // exect time in Ms for wich rateDecStLast has been filled
//    QTimer * const timerAutoSave;
    QTimer * const timerBaseStatistics;
    QTimer * const timerRawWriter;
    QTimer * const timerEventWriter;
    QTimer * const timer_sim;
    QTimer * const timerUpdateDisplay;
    QTimer * const timerUpdateHistograms; //TODO remove it

    // dialogs
    DeconvConfigDialog * const deconvConfigDialog;
    AdcConfigDialog * const adcConfigDialog;
    CalAvgDialog * const calAvgDialog;
    CalZeroDialog * const calZeroDialog;
    FitterSettingsDialog * const fitProfileDialog;
    EvStorageConfigDialog * const evStorageConfigDialog;
    RawStorageConfigDialog * const rawStorageConfigDialog;
    RunConfigDialog * const runConfigDialog;


    quint64 eventsProceed = 0;
    double runSeconds = 0;
    quint64 runBytes = 0;
    double prev_mb = 0;
    QTime prev_tv;
//    bool autoSaveLoaded = false;

    // copies
    conf_mem_t cacheAdcConfMem;
    AdcConfig cacheAdcConfig;
    DecoderConfig cacheDecoderConfig;
    StorageConfig cache_stor_dev;
    StorageConfig cache_stor_raw;
    RunConfig cacheRunConfig;

    QString cacheDeviceStatusText;
    QString cacheConfigFileName; // cache
    bool cacheConnectedState = false; // cache
    bool cacheReadoutState = false; // cache
};

#endif // MAINWINDOW_H
