#ifndef DECODEMANAGER_H
#define DECODEMANAGER_H

#include <QObject>
#include <QMap>
#include <QMutex>
#include <QTimer>
#include <QVector>
#include <QList>

#include "adcmlib/adcm_el.h"

class QByteArray;
class QTime;

//class FrameList; // adcm_raw.h
//class AdcmDataBlockDecoder; // adcm_raw.h
class ConfigManager;
class AdcConfig;
class AdcmEvent;
class CalAvgConfig;
class CalZeroConfig;
struct adcm_counters_t;
class adcm_decstat_t;
class adcm_el_t;
class conf_mem_t;
class DecoderConfig;
class DecodeWorker;
class HistManager;
class HistList;
class QSemaphore;
class storage;
class StorageConfig;
class EventListFiller;
struct hist_index_t;

struct WorkerJob
{
    uint workerIndex;
    EventListIt startPos;
    EventListIt endPos;
    int size;
    bool completed;
    WorkerJob(int index, EventListIt start,EventListIt end, int s):
        workerIndex(index),
        startPos(start),
        endPos(end),
        size(s),
        completed(false)
    {}
};

class DecodeManager : public QObject
{
    Q_OBJECT
public:
    DecodeManager(ConfigManager *configManager, QObject *parent);
    ~DecodeManager();
    EventListFiller * createFiller();
    void pause();
    void resume();
    void reset(); // Calls from filler
    void clear(); // Clears hist & stats (worker+mngr), storage & counters(mngr)
public Q_SLOTS:
    // Slots connected from MainWindow
    void setAdcConfMem(const conf_mem_t &value); // (Re-)Initialize
    void setAdcConfig(const AdcConfig &r); // Any changing
    void setCalAvgConfig(const CalAvgConfig &r);
    void setCalZeroConfig(const CalZeroConfig &r);
    void setDecoderConfig(const DecoderConfig &r);
    void setStorageConfig(const StorageConfig &r);
    void appendChannelCounters(const adcm_counters_t &cnt);
    bool endBurst();
    void setHistEnabled(bool);
    // Slot for internal timer
    // Slot for DecodeWorker feedback
    void addWorkerStatistics(const adcm_decstat_t &, int index=-1);
Q_SIGNALS:
    void decodedCalAvg(const CalAvgConfig &);
    void decodedCalZero(const CalZeroConfig &);
    void monitorEvent(const AdcmEvent &); // we should share monitorEventTime and include mutex
    void decoderConfigModified(const DecoderConfig &);
    void adcConfigModified(const AdcConfig &r);
    void adcConfMemChanged(const conf_mem_t &);

    void summedStatisticsUpdated(const adcm_decstat_t &);
    void histListModified(const HistList &);
    void histogramsChanged(const HistList &);
    void decoderConfigUpdated();
    void startWorkers(int);
    void storageConfigUpdated(bool);
//    void clearAllHists();
//    void clearHist(const hist_index_t &);

public:
    static const int numberWorkers;

private Q_SLOTS:
    void loop();
    void on_histListModified(const HistList &);
    void on_histogramsChanged(const HistList &);
    void on_monitorEvent(const AdcmEvent &); // we should share monitorEventTime and include mutex

private:
    void clearCounters();
    bool startJob();
    void clenUpCompletedJobs(int index=-1);
    void info();
    void checkEvListSize();
    DecodeWorker* getIdleWorker();
    void writeChMap();

    bool paused = false;
    AdcConfig * const adcConfig; // Uses in writeChMap to write triggers
    StorageConfig * const storageConfig;
    DecoderConfig * const decoderConfig; // Unique, created at DecodeManager (can be replaced with decoderConfig->mode)
    HistManager * const hmgr; // Unique created at DecodeManager
    bool histEnabled = true;
    adcm_counters_t * const counters;
    storage * const stor;

    QVector<QThread*> workerThreads;
    QVector<DecodeWorker*> workers;
    QList<WorkerJob> currentJobs;

    QSemaphore * const idleWorkers; // Worker releases (job done); Manager acquires (job assigned)
    QSemaphore * const freeEvents; // Manager releases (clean up job); Filler via eventList acquires (job creating)
    QSemaphore * const managerEvents; // Filler releases (job created); Manager acquires (job assigned)
    QSemaphore * const pauseSem; // Released when loop finaly stoped;

    uint lastEventListSize; // Debug value
    adcm_el_t * const eventList;

    QMap<int, adcm_decstat_t> workersDecStat;
    adcm_decstat_t * const summedDecStat;

    QTime * const lastHistSumTime;
    QTime * const workStatTimer;
    QTimer * const jobTimer;
    int nch = 0; // number of channels (used in channel map)
};

#endif // DECODEMANAGER_H
