#include "DecodeWorker.h"

#include <exception>

#include <QDebug>
#include <QTime>
#include <QTimer>
#include <QSemaphore>

#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"

#include "storage.h"
#include "adcm_df.h"
#include "adcmlib/adcm_el.h"
#include "adcmlib/adcm_ev.h"
#include "adcmlib/adcm_dec.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmraw/adcm_mem.h"
#include "adcmlib/timer.h"
#include "hist_manager/histmanager.h"
#include "config/decoderconfig.h"
#include "decoder/waveformdecoderpeaks.h"
#include "decoder/waveformdecodersine.h"
#include "decoder/decoder_profiling_macros.h"
#include "decoder/AdcmEventDecoder.h"

#ifndef DECODER_PROFILING
static const int DECODER_PROFILING = 0;
#endif

#define PREFIX "DecodeWorker: "

namespace {
const int monitorEventTimeMs = 100;
const int STOR_PH_SIZE = sizeof(stor_packet_hdr_t);
const int MAX_STOR_EV_SIZE = STOR_PH_SIZE + sizeof(stor_ev_hdr_t) + sizeof(stor_puls_t) * ADC_NCH;
}

//______________________________________________________________________________

DecodeWorker::DecodeWorker (int _workerIndex,
                  QSemaphore * _idleWorkers,
                  storage * _stro,
                  DecoderConfig * decConf,
                  adcm_el_t * _eventList,
                  bool storEnabled) :
    QObject (nullptr),
    workerIndex(_workerIndex),
    idleWorkers(_idleWorkers),
    stateMutex(new QMutex),
    stor(_stro),
    hmgr(new HistManager(this)),
    eventList(_eventList),
    decst(new adcm_decstat_t),
#ifdef FIT_SINE
    decoder(new WaveformDecoderSine),
#else
    decoder(new WaveformDecoderPeaks),
#endif /* FIT_SINE */
    monitorEventTime(new QTime()),
    statisticsTimer(new QTimer(this)),
    calResultTimer(new QTimer(this)),
    decodeMode(decConf->mode),
    storageEnabled(storEnabled),
    newCalAvgConfig(new CalAvgConfig),
    newCalZeroConfig(new CalZeroConfig)
{
    setObjectName(QString("DecodeWorker[%1]").arg(workerIndex));
    hmgr->setObjectName(QString("HistManager[%1]").arg(workerIndex));

    connect(this, SIGNAL(decoderConfigModified(DecoderConfig)), decoder, SLOT(setDecoderConfig(DecoderConfig)));
//    connect(this, SIGNAL(decoderConfigModified(DecoderConfig)), hmgr, SLOT(setDecoderConfig(DecoderConfig)));

    connect(this, SIGNAL(adcConfigModified(AdcConfig)), decoder, SLOT(setAdcConfig(AdcConfig)));
//    connect(this, &DecodeWorker::adcConfigModified, hmgr, &HistManager::setAdcConfig);

    connect(this, &DecodeWorker::adcConfMemChanged, hmgr, &HistManager::setAdcConfMem);
    connect(hmgr, &HistManager::histogramsChanged, this, &DecodeWorker::histogramsChanged);

//    connect(this, &DecodeWorker::clearAllHists, hmgr, &HistManager::clearAllHists);
//    connect(this, &DecodeWorker::clearHist, hmgr, &HistManager::clearHist);

    connect(statisticsTimer, SIGNAL(timeout()), SLOT(sendStatistics()));
    statisticsTimer->setInterval(1000);
    statisticsTimer->start();

    connect(calResultTimer,SIGNAL(timeout()), SLOT(sendCalResult()));
    calResultTimer->setInterval(300);
    calResultTimer->setSingleShot(false);
    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
        calResultTimer->start();
    monitorEventTime->start();
}

DecodeWorker::~DecodeWorker ()
{
    delete newCalZeroConfig;
    delete newCalAvgConfig;
    delete calResultTimer;
    delete statisticsTimer;
    delete monitorEventTime;
    delete decoder;
    delete decst;
    delete hmgr;
    delete stateMutex;
}

int DecodeWorker::processEvent (AdcmEvent * ev)
{
    Q_ASSERT(ev);

    //    ev->print_event();

    decoder->decode(ev, decst);
    if (ev->getFlag(AdcmEvent::ADCM_EV_DT_CUT)) {
        return 0; // drop event
    }

    switch (decodeMode) {
    case DM_ZERO:
        *newCalZeroConfig += AdcmEventDecoder::makeCalZero(*ev);
        break;
    case DM_AVG:
        *newCalAvgConfig += AdcmEventDecoder::makeCalAvg(*ev);
        break;

    case DM_REC: {
        ftimer_t t;
        if (DECODER_PROFILING) {
            t.start();
        }
        if(histEnabled){
            hmgr->fill (*ev);
            if(prevEvTs)
                hmgr->fillEvDt(*ev, prevEvTs);
            prevEvTs = ev->getEventTimestamp();
        }

        if (DECODER_PROFILING) {
            decst->prof.ta[decst_n_hist] += t.read_clear();
        }
        break;
    }
    default:
        ;
    }
    if (DECODER_PROFILING) hmgr->fill_status (*decst);

//    if (ev->getEventNumber()<10) {
//        printf("-----------------------------------\n");
//        ev->print_event();
//        ev->print_flags();
//    }

    if (storageEnabled)
        writeEvent (*ev);

    return (1);
}

void DecodeWorker::sendStatistics()
{
//    requestForStat=false;
    emit statisticsUpdated(*decst, workerIndex);
    decst->clear();
}


void DecodeWorker::decodeEventList (EventListIt start, EventListIt end)
{
    EventListIt eventIt=start;
    AdcmEvent* ev;

    // Decode events from Event List
    while(eventIt != end){
        ev = eventIt->second;
        processEvent (ev);
        if(workerIndex==0){
            if(histEnabled && monitorEventTime->elapsed() > monitorEventTimeMs) {
                monitorEventTime->restart();
                emit monitorEvent(*ev);
            }
        }
        eventList->next(eventIt);
//        if(requestForStat) sendStatistics();
    }

//    sendStatistics();
    unlockWorker();
    idleWorkers->release();
}


void DecodeWorker::clear ()
{
    WaveformDecoder::adcm_decoder_clear_stat (decst);
    hmgr->clearAllHists();
}

void DecodeWorker::reset()
{
//    statisticsTimer->start();
}


void DecodeWorker::sendCalResult()
{
    if(decodeMode == DM_ZERO){
        if (!newCalZeroConfig->isEmpty()) {
            emit decodedCalZero(*newCalZeroConfig);
            newCalZeroConfig->clear();
        }
    } else if(decodeMode == DM_AVG){
        if (!newCalAvgConfig->channels.isEmpty())
            emit decodedCalAvg(*newCalAvgConfig);
        newCalAvgConfig->clear();
    } else
        qWarning("Callibration timer works in decoder thread, but it should be stopped");
}


bool DecodeWorker::isIdle()
{
    bool ret=false;
    if(stateMutex->tryLock()){
        ret=true;
        stateMutex->unlock();
    }
    return ret;
}


void DecodeWorker::lockWorker()
{
    stateMutex->lock();
}

void DecodeWorker::unlockWorker()
{
    stateMutex->unlock();
}

void DecodeWorker::setHistEnabled(bool val)
{
    histEnabled = val;
}

//void DecodeWorker::pause()
//{
//    statisticsTimer->stop();
//    calResultTimer->stop();
//}

//void DecodeWorker::resume()
//{
//    statisticsTimer->start();
//    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
//        calResultTimer->start();
//    else
//        calResultTimer->stop();
//}

//void DecodeWorker::requestStats()
//{
//    requestForStat=true;
//    if(stateMutex->tryLock()){
//        sendStatistics();
//        stateMutex->unlock();
//    }
//}


bool DecodeWorker::writeEvent (const AdcmEvent &ev)
{
    unsigned char buf[MAX_STOR_EV_SIZE]; // FIXME
    auto *ph = reinterpret_cast<stor_packet_hdr_t *>(&buf[0]);
    auto *eh = reinterpret_cast<stor_ev_hdr_t *>(&buf[STOR_PH_SIZE]);

    ph->id = STOR_ID_EVNT;
    ph->size = STOR_PH_SIZE;

    eh->ts = ev.getEventTimestamp(); // FIXME: 64 to 32 truncation
    eh->reserved1 = 0;
    eh->reserved2 = 0;
    eh->np = 0;
    ph->size += sizeof(stor_ev_hdr_t);

    for (auto c : ev.channel.keys()) {
        const adcm_chbuf_t &cb = ev.channel[c];
        if (cb.conv.peaks.empty())
            continue;
        {
            auto *p = reinterpret_cast<stor_puls_t *>(& buf[ph->size]);
            p->ch = c;
            p->a = cb.i - cb.b; // Mean value relatively baseline
            p->t = cb.conv.peaks[0].pt*ev.format.step * 1e9; // Peak time, ns
            p->w = cb.conv.peaks[0].pw*ev.format.step * 1e9; // Peak width, ns
            ph->size += sizeof (stor_puls_t);
            eh->np++;
        }
    }
    return stor->write (buf, ph->size);
}

void DecodeWorker::setAdcConfig(const AdcConfig &adcConfig)
{
    emit adcConfigModified(adcConfig);
}

void DecodeWorker::setAdcConfMem(const conf_mem_t &value)
{
    emit adcConfMemChanged(value);
}

void DecodeWorker::setDecoderConfig(const DecoderConfig &decConf)
{
    decodeMode = decConf.mode;
    emit decoderConfigModified(decConf);
    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
        calResultTimer->start();
    else
        calResultTimer->stop();
}


void DecodeWorker::setStorageConfig(bool enabled)
{
    storageEnabled = enabled;
}
