#include "RemoteControlServer.h"

#include <QHostInfo>
#include <QTcpServer>
#include <QTcpSocket>

namespace {
const char * RC_CMD_RUN = "run";
const char * RC_CMD_PAUSE = "pause";
const char * RC_CMD_RESET = "reset";
//const char * RC_CMD_ERROR = "error";
const char * RC_CMD_STATE = "state";
const char * RC_CMD_SUCCESS_PREFIX = "success";
const char * RC_CMD_FAILED_PREFIX = "failed";
const char * RC_CMD_END = "\r\n";
}

static const quint16 REMOTE_CONTROL_TCP_PORT = 33305;

RemoteControlServer::RemoteControlServer(QObject *parent) :
    QObject(parent),
    listening(false),
    tcpPort(REMOTE_CONTROL_TCP_PORT),
    portInUse(false),
    qTcpServer(new QTcpServer(this)),
    tcpSocket(nullptr),
    lastRunState("unknown")
{
}

RemoteControlServer::~RemoteControlServer()
{
    blockSignals(true);
    startListen(false);
    blockSignals(false);
}

void RemoteControlServer::startListen(bool state)
{
    listening = state;
    if(listening){
        init();
    } else {
        qTcpServer->close();
        if(tcpSocket){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        qDebug() << QString("TcpServer closed on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    }
}

void RemoteControlServer::runStateUpdated(const QString &runState)
{
    lastRunState = runState;
    if(!listening)
        return;
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_STATE)
            .arg(runState)
            .arg(RC_CMD_END);
    sendString(mess);
}

void RemoteControlServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        if(tcpSocket==nullptr){
            tcpSocket = pendingConnection;
            portInUse=true;
            connect(tcpSocket, SIGNAL(readyRead()), SLOT(readCommandRequest()));
            connect(tcpSocket, SIGNAL(disconnected()), SLOT(onDisconnected()));
            QHostInfo info = QHostInfo::fromName(tcpSocket->peerAddress().toString());
            QString msg = QString("RemoteControlServer accepted connection from %1:%2")
                    .arg(info.hostName())
                    .arg(tcpSocket->peerPort());
            qInfo() << msg;
            updateStatus();
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(tcpSocket==nullptr){
        qDebug()<<"There are no connections.";
    }
}

void RemoteControlServer::readCommandRequest()
{
    while(tcpSocket->bytesAvailable()){
        QByteArray data = tcpSocket->readAll();
        QString mess = QString::fromUtf8(data.data(), data.size());
        foreach (QString command, mess.split(RC_CMD_END)) {
            command = command.trimmed();
            qDebug() << QString("Remote command:").append(command);
            QStringList commandParts = command.split(" ", QString::SkipEmptyParts);
            if(commandParts.isEmpty())
                continue;
            if(commandParts.first() == RC_CMD_RUN) {
                qDebug() << QString("Received command:").append(RC_CMD_RUN);
                if(commandParts.size()==1)
                    emit startRequest();
                else
                    qWarning() << QString("Run command can't have any arguments:").append(command);
            } else if(commandParts.first() == RC_CMD_PAUSE) {
                qDebug() << QString("Received command:").append(RC_CMD_PAUSE);
                if(commandParts.size()==1)
                    emit stopRequest();
                else
                    qWarning() << QString("Pause command can't have any arguments:").append(command);
            } else if(commandParts.first() == RC_CMD_STATE) {
                qDebug() << QString("Received command:").append(RC_CMD_STATE);
                updateStatus();
            }
        }
    }
}

void RemoteControlServer::onDisconnected()
{
    portInUse=false;
    qInfo() << "RemoteControlServer disconnected";
}

void RemoteControlServer::init()
{
    while(!qTcpServer->listen(QHostAddress::Any, tcpPort)) {
        QString s = tr("Unable to listen on TCP port %1: %2").arg(tcpPort).arg(qTcpServer->errorString());
        qCritical() << s;
        return;
    }
    tcpPort = qTcpServer->serverPort();

    connect(qTcpServer, SIGNAL(newConnection()), this, SLOT(updateConnections()));
    qInfo() << QString("RemoteControlServer listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

bool RemoteControlServer::isRunControlConnected() const
{
    return tcpSocket != nullptr && tcpSocket->isValid() && (tcpSocket->state() == QTcpSocket::ConnectedState);
}

void RemoteControlServer::sendString(const QString &mess)
{
    QByteArray messData = mess.toUtf8();
    if (!isRunControlConnected()) {
        return;
    }
    int written = tcpSocket->write(messData);
    if(written != messData.size())
        qWarning()<<"Failed to send info string to run control client";
    else
        qDebug()<<"sent to run control client:" << mess.trimmed();
}


void RemoteControlServer::updateStatus()
{
    runStateUpdated(lastRunState);
}



