
#include <chrono>
#include <cerrno>
#include <string>

#include <QDebug>
#include <QThread>
#include <QTimer>
#include <QByteArray>
#include <QVector>

#include "adcinstance.h"
#include "adcmdevice/adcmdevice.h"
#include "adcmdevice/adchwcounters.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmraw/adcm_ids.h"
#include "adcm_df.h"
#include "mlink/mlink.h"
#include "mlink/regiomlink.h"
#include "regio/regiopacket.h"
#include "mlink/regiomlink.h"
//#include "regioemu.h"
#include "adcmlib/stopwatchtimer.h"
#include "FrameList.h"

//const int hardwarePollIntervalMsec = 30;
constexpr int hardwareReconnectIntervalMsec = 100;
const int hardwarePollDataIntervalMsec = 1;
const int hardwarePollStatusIntervalMsec = 500;
const int hitcountsIntervalMsec = 1000;

const uint maxQueueEvents = 4;

#ifndef DEBUG_PRINT_DATA_RATE
static const int DEBUG_PRINT_DATA_RATE = 0;
#endif

using namespace std::chrono;

AdcInstance::AdcInstance(QObject *parent)
    : QObject(parent),
    reconnectTimer(new QTimer(this)),
    pollDataTimer(new QTimer(this)),
    pollStatusTimer(new QTimer(this)),
    hitcountsTimer(new QTimer(this)),
    runTimer(new StopWatchTimer(this)),
    prevAdcHwCounters(new AdcHwCounters),
    adcmLink(QSharedPointer<mlink::RegIOMLink>(new mlink::RegIOMLink()))
{
    setObjectName("AdcInstance thread");
    if (hasGainControl()) {
        min_gain_db = adc_gain_val_to_db(0);
        max_gain_db = adc_gain_val_to_db(15);
    }
    qRegisterMetaType<conf_mem_t>();
    qRegisterMetaType<QVector<double> >("QVector<double>");

    connect(this, SIGNAL(connected()), SLOT(on_connected()));
    connect(this, SIGNAL(disconnected()), SLOT(on_disconnected()));
    connect(this, SIGNAL(readoutStateChanged(bool)), SLOT(on_readoutStateChanged(bool)));

    connect(pollDataTimer, SIGNAL(timeout()), SLOT(pollData()));
    pollDataTimer->setInterval(hardwarePollDataIntervalMsec);
    pollDataTimer->start();

    connect(pollStatusTimer, SIGNAL(timeout()), SLOT(pollStatus()));
    pollStatusTimer->setInterval(hardwarePollStatusIntervalMsec);
    pollStatusTimer->start();

    connect(hitcountsTimer, SIGNAL(timeout()), this, SLOT(readHitCounts()));
    hitcountsTimer->setInterval(hitcountsIntervalMsec);
    hitcountsTimer->setSingleShot(true);
    // hitcountsTimer->start in connected()

    auto statisticsTimer(new QTimer(this));
    connect(statisticsTimer, SIGNAL(timeout()), SLOT(updateStatistics()));
    statisticsTimer->start(100);

    connect(reconnectTimer, &QTimer::timeout, this, &AdcInstance::connectDevice);
    reconnectTimer->start(hardwareReconnectIntervalMsec);
    QTimer::singleShot(0, this, SLOT(connectDevice()));
}

AdcInstance::~AdcInstance()
{
    delete prevAdcHwCounters;
    delete adcmDevice;
}

void AdcInstance::setErrorString(const QString &str)
{
    p_errorString = str;
    emit statusTextChanged(str);
}

void AdcInstance::fatalError(const QString &str)
{
    if (p_errorString != str) {
        setErrorString(str);
        qCritical() << __PRETTY_FUNCTION__ << str;
    }
    disconnectDevice();
}

float AdcInstance::adc_gain_val_to_db(int value) const
{
    return AdcmDevice::adc_gain_val_to_db(value);
}
/*
int AdcInstance::adc_gain_val_to_raw(float db_value) const
{
  return (db_value+9) / 3;
}
*/
float AdcInstance::getMinGain() const
{
    return min_gain_db;
}

float AdcInstance::getMaxGain() const
{
    return max_gain_db;
}

float AdcInstance::getGainStep() const
{
    return gain_step_db;
}

bool AdcInstance::hasGainControl() const
{
    return adcmDevice ? adcmDevice->getConfMem().hasGainControl() : false;
}

bool AdcInstance::hasTriggerControl() const
{
    return adcmDevice ? adcmDevice->getConfMem().hasTriggerControl() : false;
}

quint64 AdcInstance::queueSize() const
{
    return lastSentSeq - lastReceivedSeq;
}

bool AdcInstance::isQueueFull() const
{
    return queueSize() >= maxQueueEvents;
}

void AdcInstance::pollData()
{
    if (!adcmDevice) return;
    if (isQueueFull()) {
        pollDataTimer->start(hardwarePollDataIntervalMsec);
//        qDebug().nospace() << "AdcInstance::poll thread " << lastSentSeq <<" | "<< lastReceivedSeq
//                           << "; queue " << queueSize() << "/" << maxQueueEvents;
        return;
    }
    auto start = steady_clock::now();
    try {
        adcmDevice->poll();
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in poll(): %1").arg(e.what()));
    }
    auto timestampPoll = steady_clock::now();

    FrameList frameList(adcmDevice->popDataPackets());
    frameList.created = QTime::currentTime();

    auto timestampFrameList = steady_clock::now();
    const int numFrames = frameList.size();
    if (numFrames > 0) {
        //        qDebug() << "Received" << numFrames << "frames";
            lastSentSeq++;
            frameList.seq = lastSentSeq;
            emit frameReceived(frameList);
    }

    quint64 numBytes = 0;
    for (auto & buf : frameList) {
        numBytes += buf.size();
    }
    numBytesReceived += numBytes;
    auto timestampEmitData = steady_clock::now();

    emit timeElapsed(runTimer->elapsed());
    emit byteCountChanged(numBytesReceived);

    readoutStats.statsNumBytesReceived = numBytes;
    readoutStats.readout_time_poll_ms = duration_cast<milliseconds>(timestampPoll - start).count();
    readoutStats.readout_time_framelist_ms = duration_cast<milliseconds>(timestampFrameList - timestampPoll).count();
    readoutStats.readout_time_emit_data_ms = duration_cast<milliseconds>(timestampEmitData - timestampFrameList).count();
    readoutStats.throughput_Bps = 1.0 * numBytes / (duration_cast<seconds>(timestampPoll - start).count());

//    int nextPollIntervalMs = (numFrames && isQueueFull()) ? 0 : hardwarePollIntervalMsec;
    int nextPollIntervalMs = hardwarePollDataIntervalMsec;
    pollDataTimer->start(nextPollIntervalMs);
}

void AdcInstance::pollStatus()
{
    if (!adcmDevice) return;
    try {
        temperatures = adcmDevice->readTemperatures();
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in pollStatus(): %1").arg(e.what()));
    }
    Q_EMIT tempChanged(temperatures);
}

void AdcInstance::updateStatistics()
{
    emit readoutStatsChanged(readoutStats);
}

adcm_counters_t AdcInstance::makeAdcmCounters(const AdcHwCounters &adcHwCounters) const
{
    adcm_counters_t cnt;
    const int n = std::min((int)ADC_NCH, adcHwCounters.hits.size());
    cnt.n = n;
    for (int i=0; i<n; i++)
        cnt.rawhits[i] = adcHwCounters.hits[i];
    for (int i=n; i<ADC_NCH; i++)
        cnt.rawhits[i] = 0;
    cnt.time = adcHwCounters.current_ts;
    return cnt;
}

void AdcInstance::readHitCounts()
{
    try {
        if (adcmDevice) {
            AdcHwCounters cnt = adcmDevice->adc_get_counters();
            emit hitCountsChanged(makeAdcmCounters(cnt - *prevAdcHwCounters));
            *prevAdcHwCounters = cnt;
        }
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in adc_get_counters(): %1").arg(e.what()));
    }
    hitcountsTimer->start();
}

void AdcInstance::disconnectDevice()
{
    if (!adcmDevice)
        return;
    qInfo() << __PRETTY_FUNCTION__ << " begin " << adcmDevice;
    delete adcmDevice;
    adcmDevice = nullptr;
    emit disconnected();
}

void AdcInstance::connectDevice()
{
    if (adcmDevice)
        return;
    try {
        if (!adcmLink)
            return;
        if (!adcmLink->isOpen()) {
            if (int ret = adcmLink->open()) {
                if (ret == -ENOENT)
                    fatalError("No device driver");
                else
                    fatalError(strerror(-ret));
                return;
            }
        }
        if (!adcmLink->isConnected()) {
            fatalError("Link down");
            return;
        }
        adcmDevice = new AdcmDevice(adcmLink);
        adcmDevice->adc_reset();
    }
    catch(std::exception &e) {
        fatalError(QString("Exception in connectToDevice(): %1").arg(e.what()));
        return;
    }
    setErrorString(QString());
    if (adcmDevice) {
        emit connected();
    }
}

void AdcInstance::on_connected()
{
    qInfo() << __PRETTY_FUNCTION__;
    reconnectTimer->stop();
    emit confMemChanged(adcmDevice->getConfMem());
    pollStatus();
    if (desiredReadoutState)
        startReadout();
    else
        stopReadout();
}

void AdcInstance::on_disconnected()
{
    qInfo() << __PRETTY_FUNCTION__;
    prevAdcHwCounters->clear();
    reconnectTimer->start(hardwareReconnectIntervalMsec);
}

void AdcInstance::on_readoutStateChanged(bool state)
{
    if (state) {
        hitcountsTimer->start();
    } else {
        hitcountsTimer->stop();
    }
}

void AdcInstance::setCalZeroConfig(const CalZeroConfig &r)
{
    if (r == calZero)
        return;
    calZero = r;
    setup();
}

void AdcInstance::setAdcConfig(const AdcConfig &r)
{
    if (r == adcConfig)
        return;
    adcConfig = r;
    setup();
}

void AdcInstance::setup()
{
    if (!adcmDevice)
        return;
    try {
        adcmDevice->setup(calZero, adcConfig);
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in setup(): %1").arg(e.what()));
    }
}

void AdcInstance::startReadout()
{
    desiredReadoutState = true;
    if (!adcmDevice) return;
    if (dma_running)
        return;
    try {
        if (!adcmDevice->adc_start()) {
            fatalError("startReadout failed");
            return;
        }
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in adc_start(): %1").arg(e.what()));
        return;
    }
    runTimer->start();
    adcm_counters.clear();
    dma_running = true;
    emit readoutStateChanged(true);
}

void AdcInstance::stopReadout()
{
    desiredReadoutState = false;
    if (!adcmDevice) return;
    try {
        adcmDevice->adc_stop();
    }
    catch (std::exception &e) {
        fatalError(QString("Exception in adc_stop(): %1").arg(e.what()));
    }
    runTimer->stop();
    dma_running = false;
    emit readoutStateChanged(false);
}

double AdcInstance::runtime () const
{
    return runTimer->elapsed();
}

void AdcInstance::reset()
{
    numBytesReceived = 0;
    adcm_counters.clear();
    if (dma_running)
        runTimer->restart();
    else
        runTimer->reset();
}

void AdcInstance::dataProgress(quint64 value)
{
    lastReceivedSeq = value;
//    qDebug() << QString("lastReceivedSeq=%1").arg(lastReceivedSeq);
}

QStringList AdcInstance::getDeviceIdStrList() const
{
    if (!adcmDevice) return QStringList();
    return adcmDevice->getConfMem().getDeviceIdStrList();
}
