    //
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "adcconfig.h"

#include <cmath>
#include <cstdio>
#include <QDebug>
#include <QTextStream>

#include "adcm16.h"
#include "adcmraw/adcm_mem.h"
#include "ConfigStatus.h"

const double eps = 1e-15;

AdcChannelConfig::AdcChannelConfig()
{
    flags.all = 0;
    flags.bit.en = 1;
    flags.bit.test = 1;
    flags.bit.zero = 1;
}

bool AdcChannelConfig::operator==(const AdcChannelConfig &other) const
{
    return true
            && (abs(gain - other.gain) < eps)
            && (abs(thr - other.thr) < eps)
            && (flags.all == other.flags.all)
//            && (recwin == other.recwin)
            ;
}

AdcChannelConfigMap AdcConfig::defaultChannels()
{
    AdcChannelConfigMap tmp;
    for (int i=0; i<ADCM_N_CH; i++)
        tmp[i] = AdcChannelConfig();
    return tmp;
}

void AdcConfig::resize(int n)
{
    for (int i=0; i<n; i++)
        if (!channels.contains(i))
            channels.insert(i, AdcChannelConfig());
    for (auto i : channels.keys())
        if (i < 0 || i >= n)
            channels.remove(i);
}

ConfigStatus AdcConfig::isValid() const
{
    ConfigStatus status("adc");
    if ((bits < 0) || (bits > 32)) {
        status << QString("bad adc.bits: %1, set default %2").arg(bits).arg(AdcConfig().bits);
    }
    if ((step <= 0) || (step > 1e9)) {
        status << QString("bad adc.step: %1, set default %2").arg(step).arg(AdcConfig().step);
    }
    return status;
}

bool AdcConfig::operator==(const AdcConfig &other) const
{
    return true
            && bits == other.bits
            && abs(step -other.step) < eps
            && channels == other.channels
            && prewin == other.prewin
            && matchwin == other.matchwin
            && latency == other.latency
            && runmode == other.runmode
            && trigbits.all == other.trigbits.all
            && mult_min == other.mult_min
            && mult_max == other.mult_max
            && fragm_len == other.fragm_len
            && fragm_count_m == other.fragm_count_m
            && fragm_count_s == other.fragm_count_s
            && timestamp_is_utc == other.timestamp_is_utc
            ;
}

bool AdcConfig::isBaselineCalibrationMode() const
{
    return runmode == ADCM_RUNMODE_BASECAL;
}

//void AdcConfig::adjust_adc_recwin()
//{
//    const unsigned int recwin_max = adc_cfg_get_recwin_max();
//    for (int i : channels.keys()) {
//        if (channels[i].recwin > recwin_max)
//            channels[i].recwin = recwin_max;
//    }
//}


QTextStream &operator<<(QTextStream &s, const adcm_channel_flags_t &flags)
{
    s << "[";
    if (flags.bit.en)
        s << " en";
    if (flags.bit.test)
        s << " test";
    if (flags.bit.zero)
        s << " z/s";
    if (flags.bit.inv)
        s << " inv";
    if (flags.bit.m)
        s << " m";
    if (flags.bit.s)
        s << " s";
    if (flags.bit.unused)
        s << " unused=" << flags.bit.unused;
    s << "]";
    return  s;
}

QTextStream &operator<<(QTextStream &s, const AdcChannelConfig &a)
{
    s << a.gain
      << " " << a.thr
      << " " << a.flags;
    return s;
}

QTextStream &operator<<(QTextStream &s, const AdcConfig &a)
{
    s << "# --- AdcConfig ---" << endl;
    s << "adc_bits: " << a.bits << endl;
    s << "adc_step: " << a.step << endl;
    s << "adc_prewin: " << a.prewin << endl;
    s << "adc_matchwin: " << a.matchwin << endl;
    s << "adc_latency: " << a.latency << endl;
    s << "adc_runmode: " << a.runmode << endl;
    s << "adc_trigbits:"
      << ((a.trigbits.all == 0) ? " none" : "")
      << (a.trigbits.bit.int_simple ? " internal_simple" : "")
      << (a.trigbits.bit.coinc_logic ? " internal" : "")
      << (a.trigbits.bit.ext1 ? " ext1" : "")
      << (a.trigbits.bit.ext2 ? " ext2" : "")
      << (a.trigbits.bit.ext3 ? " ext3" : "")
      << endl;
    s << "adc_mult_min: " << a.mult_min << endl;
    s << "adc_mult_max: " << a.mult_max << endl;
    s << "adc_fragm_len: " << a.fragm_len << endl;
    s << "adc_fragm_count_m: " << a.fragm_count_m << endl;
    s << "adc_fragm_count_s: " << a.fragm_count_s << endl;
    s << "adc_cfg->timestamp_is_utc: " << a.timestamp_is_utc << endl;
    s << "# Number of channels: " << a.channels.size() << endl;
    s << "# channel, gain, thr, trig, recwin" << endl;
    for (int i : a.channels.keys()) {
        s << "  " << i
          << ": " << a.channels[i]
          << endl;
    }
    return s;
}

// Threshold conversion: mV (-0.5..0.5) and value (0..1023)

const double thr_min_mv = -512;
const double thr_max_mv = 511;

//double AdcConfig::mv_to_value_scaler ()
//{
//    /*
//  double range_mv = thr_max_mv - thr_min_mv;
//  double range_val = ADC_MAX_THR - ADC_MIN_THR;
//  if (range_mv == 0) return(1);
//  return(range_val/range_mv);
//*/
//    return (1);
//}

int AdcConfig::adc_rawdata_maxval(int ch) const
{
//    if (ch<0 || ch>=size()) return 1;
    int n = (1<<(bits-1))-1;
    return n;
}

int AdcConfig::adc_rawdata_minval(int ch) const
{
//    if (ch<0 || ch>=size()) return (-1);
    int n = -(1<<(bits-1));
    return n;
}

int AdcConfig::adc_rawdata_midval(int ch) const
{
    return 0;
}

double AdcConfig::thr_mv_min () const
{
    return thr_min_mv;
}

double AdcConfig::thr_mv_max () const
{
    return thr_max_mv;
}

void AdcConfig::setThr(int ch, double value)
{
    if (value < adc_rawdata_minval(ch))
        value = adc_rawdata_minval(ch);
    if (value > adc_rawdata_maxval(ch))
        value = adc_rawdata_maxval(ch);
    channels[ch].thr = value;
}

void AdcConfig::setGain(int ch, double value)
{
    // FIXME: check gain range
    if (value >= 0 && value <= ADC_MAX_GAIN)
        channels[ch].gain = value;
}

//unsigned int AdcConfig::adc_cfg_get_recwin_max() const
//{
//    int num_enabled = 0; // number of enabled channels
//    for (int ch : channels.keys()) {
//        int enabled = (channels[ch].flags.bit.test) ||
//                      ( (channels[ch].flags.bit.en) && (channels[ch].flags.bit.m || channels[ch].flags.bit.s) );
//        if (enabled)
//            num_enabled++;
//    }

//    unsigned int max_samples_in_event = fragm_len_max * fragm_count_max; // event fifo size
//    unsigned int mult = mult_max;
//    //  if (mult > num_enabled) mult = num_enabled;
//    if (isBaselineCalibrationMode()) { // baseline calibration mode
//        mult = num_enabled;
//    }
//    if (mult<1) mult=1;
//    unsigned int maxsize = max_samples_in_event / mult;
//    if (mult > 1) { // DPM works pass-through when onle 1 channel enabled
//        unsigned int max_total_size = ch_dpm_ks<<10;
//        unsigned int max_ch_size = max_total_size/mult;
//        if (maxsize > max_ch_size)
//            maxsize = max_ch_size;
//    }

//    if (maxsize < ADCM_MIN_N_SAMP)
//        maxsize = ADCM_MIN_N_SAMP;

//    // limit set by decoder software
//    if (maxsize > ADCM_MAX_N_SAMP)
//        maxsize = ADCM_MAX_N_SAMP;
//    maxsize &= ~1;

//    //  printf("num_enabled=%u, max_samples_in_event=%u, mult=%u, maxsize=%u\n", num_enabled, max_samples_in_event, mult, maxsize);

//    return maxsize;
//}

bool AdcConfig::is_testmode() const
{
    return isBaselineCalibrationMode();
}

bool AdcConfig::is_co_trigger() const
{
    return trigbits.bit.coinc_logic;
}

bool AdcConfig::is_ext_trigger() const
{
    return trigbits.bit.ext1;
}

bool AdcConfig::is_simple_trigger() const
{
    return trigbits.bit.int_simple;
}

//void AdcConfig::updateTriggerBits()
//{
//    for (int c = 0; c < trig.size(); c++) {
//        if (is_testmode()) { //baseline calibration
//            //            trig |= ADCM_CCTRL_BIT_M;
//            //            trig |= ADCM_CCTRL_BIT_S;
//            trig[c] |= ADCM_CCTRL_BIT_TEST;
//        } else {
//            // normal run, clear TEST bit
//            trig[c] &= ~ADCM_CCTRL_BIT_TEST;
//            if (is_simple_trigger()) {
//                // sync by any channel independently
//                trig[c] |= ADCM_CCTRL_BIT_M;
//                trig[c] |= ADCM_CCTRL_BIT_S;
//            } else if (is_ext_trigger()) {
//                // external trigger mode, master is special channel N+1
//                trig[c] &= ~ADCM_CCTRL_BIT_M;
//                //      trig &= ~ADCM_CCTRL_BIT_S;
//                trig[c] |= ADCM_CCTRL_BIT_S;
//            } else {
//                // in coincidence mode, leave settings as it is
//            }
//        }
//    }
//}


bool AdcConfig::ch_enabled_offset_calibration (int ch) const
{
//  return (trig[ch] & ADCM_CCTRL_BIT_TEST);
    return channels[ch].flags.bit.en || !channels[ch].flags.bit.zero;
}

bool AdcConfig::ch_enabled_shape_calibration (int ch) const
{
    return channels[ch].flags.bit.en && channels[ch].flags.bit.zero && channels[ch].flags.bit.m && channels[ch].flags.bit.s;
}

bool AdcConfig::ch_enabled_reconstruction (int ch) const
{
    if (! channels.contains(ch))
        return false;
    if (!(channels[ch].flags.bit.en))
        return false;
    return channels[ch].flags.bit.m || channels[ch].flags.bit.s;
}

bool AdcConfig::cfg_channel_may_have_data(int ch) const
{
    //  bool en = false;
    const adcm_channel_flags_t &trig = channels[ch].flags;
    //  int trigmode = cfg.adc.trigbits;
    if (trigbits.bit.int_simple || trigbits.bit.ext1)
        return trig.bit.en;

    bool en_any_m = false, en_any_s = false;
    for (auto i : channels.keys()) {
        Q_UNUSED(i)
        if (trig.bit.m) en_any_m = true;
        if (trig.bit.s) en_any_s = true;
    }
    if (en_any_m && en_any_s)
        return trig.bit.m || trig.bit.s;
    return false;
}

int AdcConfig::fragm_count(int ch) const
{
    int count = 1;
    if (channels[ch].flags.bit.m)
        count = fragm_count_m;
    else
        if (channels[ch].flags.bit.s)
        count = fragm_count_s;
    return count;
}

int AdcConfig::recwin(int ch) const
{
    return fragm_count(ch) * fragm_len;
}
