//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "allconfig.h"

#include <cmath>
#include <QDir>
#include <QDebug>

#include "adcm16.h"
#include "dsplib/fir.h"
#include "dsplib/fitter.h"
#include "adcmraw/adcm_mem.h"
#include "ConfigStatus.h"

const int CFG_VERSION = 0x00000086;

template<class T>
bool INRANGE(const T &x, const T &min, const T &max)
{
    return !(x < min || x > max);
}

int clear_cfgcalzero (AllConfig * cfg)
{
    if (!cfg)
        return (-1);
    for (int c = 0; c < ADC_NCH; c++)
        if (cfg->adc.channels[c].flags.bit.en) {
            cfg->dec.calzero.channels[c].clear();
        }
    return (0);
}

AllConfig::AllConfig()
{
    version = CFG_VERSION;

    stor_raw.fileName = "/data/adcm_raw.dat";

    stor_dev.enable = true;
    stor_dev.fileName = "/data/adcm.dat";

    // Resolve cross-dependencies
// FIXME: implement in decoder
//    for (int c = 0; c < std::min(dec.int2.size(), adc.size()); c++) {
//        dec.int2[c] = adc.channels[c].recwin - adc.latency - (adc.matchwin - adc.prewin);
//    }
}

bool AllConfig::operator==(const AllConfig &other) const
{
    return true
            && version == other.version
            && adc == other.adc
            && stor_raw == other.stor_raw
            && stor_dev == other.stor_dev
            && dec == other.dec
            && hist == other.hist
            && run == other.run
            ;
}

ConfigStatus AllConfig::isValid() const
{
    ConfigStatus status("all");
    //  double ts;

    if (version != CFG_VERSION) {
        status << QString("bad version %1, expected %2").arg(version).arg(CFG_VERSION);
    }
    status &= (adc.isValid() & dec.isValid());
    return status;
}

//bool AllConfig::enabled_ch_has_calzero() const
//{
//    bool ok = false;

//    for (auto c : dec.calzero.channels.keys()) {
//        if (adc.ch_enabled_offset_calibration (c) && !dec.calzero.channels[c].isEmpty())
//            ok = true;
//    }
//    return ok;
//}

bool AllConfig::enabled_ch_has_calavg() const
{
    bool ok = false;

    for (auto c : dec.calavg.channels.keys()) {
        if (adc.ch_enabled_shape_calibration (c) && !dec.calavg.channels[c].isEmpty())
            ok = true;
    }
    return ok;
}

bool AllConfig::enabled_ch_has_fir () const
{
    bool ok = true;

    for (auto c : dec.channels.keys()) {
        if (adc.ch_enabled_reconstruction (c))
            if (! dec.channels[c].deconv.fir.validate() || dec.channels[c].deconv.fir.isEmpty())
                ok = false;
    }
    return ok;
}

double AllConfig::calavg_integrate (int ch, int start, int end) const
{

    int i1, i2;
    double val;

    if (dec.calavg.channels[ch].isEmpty())
        return 0;

    //  st = 1;
    val = 0;
    i1 = start;
    i2 = end;
    if (i1 < 0) {
        //    i1 = 0;
    }
    if (i1 >= (int)ADCM_MAX_N_SAMP) {
        i1 = (int)ADCM_MAX_N_SAMP - 1;
        //    st = 0;
    }
    if (i2 < 0) {
        i1 = 0;
        //    st = 0;
    }
    if (i2 >= (int)ADCM_MAX_N_SAMP) {
        i2 = (int)ADCM_MAX_N_SAMP - 1;
        //    st = 0;
    }
    const auto &data = dec.calavg.channels[ch].get();
    for (int i = i1; i < i2; i++)
        val += data[i];
    if (i2 > i1)
        val /= (double) (i2 - i1);
    return val;
}

double AllConfig::calavg_baseline (unsigned int ch) const
{
    return calavg_integrate (ch, adc.prewin + dec.channels[ch].blr1, adc.prewin + dec.channels[ch].blr2);
}

double AllConfig::calavg_integral (unsigned int ch) const
{
    return (calavg_integrate (ch, adc.prewin + dec.channels[ch].int1, adc.prewin + dec.channels[ch].int2));
}

double AllConfig::calavg_preintegral (unsigned int ch) const
{
    return (calavg_integrate (ch, 0, adc.prewin + dec.channels[ch].int1));
}

double AllConfig::calavg_postintegral (unsigned int ch) const
{
    return (calavg_integrate (ch, adc.prewin + dec.channels[ch].int2, adc.recwin(ch)));
}

//double AllConfig::getcalzero (unsigned int ch) const
//{
//    double cz;

//    if (ch >= ADC_NCH)
//        return 0;
//    if (dec.calzero.nent[ch] > 0)
//        cz = (double) dec.calzero.base[ch] / dec.calzero.nent[ch];
//    else
//        //    cz = adc.offs[ch];
//        cz = 1<<(adc.bits[ch]-1);
//    //	cz=0;
//    return cz;
//}

//int AllConfig::getcalzero_all (unsigned int nch, float *result) const
//{
//    if (nch > ADC_NCH)
//        return (-1);

//    for(unsigned int ch=0; ch<nch; ch++) {
//        if (dec.calzero.nent[ch] > 0)
//            result[ch] = dec.calzero.base[ch] / dec.calzero.nent[ch];
//        else
//            //      result[ch] = adc.offs[ch];
//            result[ch] = 1<<(adc.bits[ch]-1);
//        //	result[ch]=0;
//    }
//    return 0;
//}

int AllConfig::thr_mv_to_value (const CalZeroConfig &calZero, const AdcConfig &adc, unsigned int ch, double mv)
{
//    double val = mv * AdcConfig::mv_to_value_scaler () + calZero.get(ch);
    double val = mv + calZero.get(ch);

    if (val < adc.adc_rawdata_minval(ch))
        val = adc.adc_rawdata_minval(ch);
    if (val > adc.adc_rawdata_maxval(ch))
        val = adc.adc_rawdata_maxval(ch);
    //  fprintf (stderr, "mv_to_value: %f, %u\n", mv, (unsigned int) val);
    return val;
}

//int AllConfig::thr_mv_to_value (unsigned int ch, double mv) const
//{
//    return thr_mv_to_value(dec.calzero, adc, ch, mv);
//}

//double thr_value_to_mv (const CalZeroConfig &calZero, unsigned int ch, unsigned int val)
//{
//    double mv = val / AdcConfig::mv_to_value_scaler () - calZero.get(ch);
//    //  fprintf (stderr, "value_to_mv: %d, %f\n", val, mv);
//    return (int) mv;
//}

//double AllConfig::thr_value_to_mv (unsigned int ch, unsigned int val) const
//{
//    return ::thr_value_to_mv(dec.calzero, ch, val);
//}

int AllConfig::get_dec_fitside(double step_ns, const DecoderConfig &dec, int ch)
{
    return dec.fitprof.get_fit_half_np(dec.channels[ch].deconv.pwid / step_ns);
/*
    int default_value = 6;

//    double step_ns = conf_mem.getMasterAdcStepNs();

    int fside = 0;
    double k = 1.0, b = 0;
    if (dec.fprof == FIT_PROFILE_SUPERFAST) {k = 0.7; }
    if (dec.fprof == FIT_PROFILE_FAST) {k = 1.0; }
    if (dec.fprof == FIT_PROFILE_OPTIMAL) { k=1.2; b = 1; }
    if (dec.fprof == FIT_PROFILE_PRECISION) {k = 1.6; b = 1; }
    if (dec.fprof == FIT_PROFILE_CUSTOM) {fside = dec.fitside;}

    if (fside <= 0) // auto
        fside = static_cast<int>(trunc(dec.channels[ch].pwid / step_ns * k + b));
    if (fside<=0 || fside > 63)
        fside = default_value;
    if (fside < 2)
        fside = 2;
    // safe check
    int maxfside = (MAX_FITGAUS_POINTS-1)/2;
    if (fside > maxfside)
        fside = maxfside;

    //  printf("ch %2d: pwid=%f, k=%.1f, b=%.0f  fside=%d\n", ch, dec.pwid[ch], k, b, fside);
    return fside;
    */
}

//int AllConfig::dt_cut_good(double dt) const
//{
//    if (!dec.dt_cut_en)
//        return 1;
//    return (dec.dt_cut_l <= dt && dt <= dec.dt_cut_r);
//}
