/*
**    Copyright 2013 Ivan Filippov
**    Copyright 2013-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "combined_ini_io.h"

#include <QString>
#include <QSettings>
#include <QDebug>

#include "inibased.h"
#include "allconfig.h"

template<>
void AdcChannelConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "gain", data->gain);
    writeValue(settings, "thr", data->thr);
    writeValue(settings, "trig", static_cast<int>(data->flags.all));
}

template<>
void AdcChannelConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "gain", &data->gain);
    readValue(settings, "thr", &data->thr);
    int tmp = data->flags.all;
    readValue(settings, "trig", &tmp);
    data->flags.all = tmp;
}

template<>
void AdcConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "bits", data->bits);
    writeValue(settings, "step", data->step);
    writeTree(settings, "channels", data->channels);
//    AdcChannelConfigIni::staticSaveMap(data->channels, settings, path() + "channels");
//    for (int i : data->channels.keys()) {
//        const AdcChannelConfig &channels = data->channels[i];
//        AdcChannelConfigIni::staticSave(channels, settings, path() + QString("channels/%1").arg(i));
//    }
    writeValue(settings, "prewin", data->prewin);
    writeValue(settings, "matchwin", data->matchwin);
    writeValue(settings, "latency", data->latency);
    writeValue(settings, "runmode", static_cast<int>(data->runmode));
    writeValue(settings, "trigbits", data->trigbits.all);
    writeValue(settings, "mult_min", data->mult_min);
    writeValue(settings, "mult_max", data->mult_max);
    writeValue(settings, "fragm_len", data->fragm_len);
    writeValue(settings, "fragm_count_m", data->fragm_count_m);
    writeValue(settings, "fragm_count_s", data->fragm_count_s);
    writeValue(settings, "timestamp_is_utc", data->timestamp_is_utc);
}

template<>
void AdcConfigIni::readFromIni(const QSettings &settings)
{
    const AdcConfig orig = *data;
    Q_ASSERT(data->isValid());

    readValue(settings, "bits", &data->bits);
//    Q_ASSERT(isValid());
    if (data->bits < 2 || data->bits > 32)
        data->bits = orig.bits;
    readValue(settings, "step", &data->step);
//    Q_ASSERT(isValid());
    if (data->step < 1e-12)
        data->step = 1;
    readValue(settings, "prewin", &data->prewin);
    readValue(settings, "matchwin", &data->matchwin);
    readValue(settings, "latency", &data->latency);
    int tmp = data->runmode;
    readValue(settings, "runmode", &tmp);
    data->runmode = static_cast<adcm_runmode_t>(tmp);
    int tmp_trigbits(data->trigbits.all);
    readValue(settings, "trigbits", &tmp_trigbits);
    data->trigbits.all = tmp_trigbits;
    readValue(settings, "mult_min", &data->mult_min);
    readValue(settings, "mult_max", &data->mult_max);
    readValue(settings, "fragm_len", &data->fragm_len);
    readValue(settings, "fragm_count_m", &data->fragm_count_m);
    readValue(settings, "fragm_count_s", &data->fragm_count_s);
    readValue(settings, "timestamp_is_utc", &data->timestamp_is_utc);

    readTree(settings, "channels", &data->channels);

//    const QString &arrayName = "channels";
//    const QList<int> &keys = childKeysNumeric(settings, arrayName);
//    for (auto & i : keys) {
//        AdcChannelConfigIni::staticLoad(settings, &data->channels[i], path() + QString("%1/%2").arg(arrayName).arg(i));
//    }
}

template<>
void StorageConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "enable", data->enable);
    writeValue(settings, "mode", data->mode);
    writeValue(settings, "period", data->period);
    writeValue(settings, "filename", data->fileName);
    writeValue(settings, "szlimit", data->szlimit);
}

template<>
void StorageConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "enable", &data->enable);
    readValue(settings, "mode", &data->mode);
    readValue(settings, "period", &data->period);
    readValue(settings, "filename", &data->fileName);
    readValue(settings, "szlimit", &data->szlimit);
}

template<>
void CalZeroChannelConfigIni::writeToIni(QSettings *settings) const
{
    if (data->nent) {
        writeValue(settings, "base", data->get_base());
        writeValue(settings, "rms", data->get_rms());
    }
}

template<>
void CalZeroChannelConfigIni::readFromIni(const QSettings &settings)
{
    if (settings.contains(path() + "base")) {
        data->nent = 1;
        readValue(settings, "base", &data->base);
        readValue(settings, "rms", &data->rms);
    }
}

template<>
void CalZeroConfigIni::writeToIni(QSettings *settings) const
{
    writeTree(settings, "channels", data->channels);
//    for (auto ch : data->channels.keys()) {
//        const CalZeroChannelConfig &channels = data->channels[ch];
//        CalZeroChannelConfigIni::staticSave(channels, settings, path() + QString("channel/%1").arg(ch));
//    }
}

template<>
void CalZeroConfigIni::readFromIni(const QSettings &settings)
{
    readTree(settings, "channels", &data->channels);

//    QStringList groups = settings.childGroups();

//    for (const QString &key : settings.allKeys()) {
//        if (! key.startsWith(path()))
//            continue;
//        QString s = key.section(path(), 1);
//        bool ok = false;
//        int ch = s.section("/", 0).toInt(&ok);
//        CalZeroChannelConfigIni::staticLoad(settings, &data->channels[ch], QString::number(ch));
//    }
//    for (auto & group : groups) {
//        int ch = group.toInt();
//        CalZeroChannelConfig tmp;
//        tmp.readFromIni(settings, group);
//        channels[ch] = tmp;
//    }
}

template<>
void BaseConfigIni<CalAvgChannelConfig>::writeToIni(QSettings *settings) const
{
    const CalAvdChannelData &tmp = data->get();
    writeValue(settings, "step", data->step);
    writeVector(settings, "data", tmp);
}

template<>
void BaseConfigIni<CalAvgChannelConfig>::readFromIni(const QSettings &settings)
{
    const CalAvgChannelConfig defVal;
    CalAvdChannelData d;
    double step = defVal.step;
    readValue(settings, "step", &step);
    readVector(settings, "data", &d);
    *data = CalAvgChannelConfig(d, step);
}

template<>
void BaseConfigIni<CalAvgConfig>::writeToIni(QSettings *settings) const
{
    writeTree(settings, "channels", data->channels);
//    for (auto ch : data->channels.keys()) {
//        BaseConfigIni<CalAvgChannelConfig>::staticSave(data->channels[ch], settings, path() + QString("channel/%1").arg(ch));
//    }
}

template<>
void BaseConfigIni<CalAvgConfig>::readFromIni(const QSettings &settings)
{
    readTree(settings, "channels", &data->channels);
#if 0
    //    CalAvgConfig newConf;
    settings->beginGroup(path());
    QStringList groups = settings->childGroups();
    for (auto & group : groups) {
        int ch = group.toInt();
        CalAvgChannelConfig tmp = channels[ch];
        tmp.readFromIni(settings, group);
        if (!tmp.isValid()) {
            qWarning() << QString("Invalid CalAvgChannelConfig for channel %1 in %2").arg(ch).arg(settings->fileName());
        }
        channels[ch] = tmp;
    }
//    readQVectorFromIni(settings, "nent", newConf.nent);
//    readQVectorFromIni(settings, "nf", newConf.nf);

//    int size = settings->beginReadArray("data");
//    for(int i=0; i<size; ++i){
//        settings->setArrayIndex(i);
//        int ch = settings->value("ch").toInt();
//        readQVectorFromIni(settings, "val", newConf.data[ch]);
//    }
//    settings->endArray();

//    readQVectorFromIni(settings, "step", newConf.step);

    settings->endGroup();
//    if(newConf.isValid())
//        *this = newConf;
//    else
//        qWarning() << "CalAvgConfig in ini-file is invalid";
#endif
}

//void writeQVectorToIni(QSettings settings, QString name, QVector<int> values)
//QVector<int> readQVectorFromIni(QSettings &settings, const QString &name, const int type)

template<>
void DecoderChannelConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "fwinf", static_cast<int>(data->deconv.fwinf));
    writeValue(settings, "ppos", data->deconv.ppos);
    writeValue(settings, "pwid", data->deconv.pwid);
    writeValue(settings, "snf", data->deconv.snf);
    writeValue(settings, "pnf", data->deconv.pnf);
    writeValue(settings, "tadj", data->deconv.tadj);
    writeValue(settings, "expfall", data->deconv.expfall);
//    writeValue(settings, "fir", data->deconv.fir); // TODO
    FirConfigIni::staticSave(data->deconv.fir, settings, path() + "fir");
    writeValue(settings, "blr1", data->blr1);
    writeValue(settings, "blr2", data->blr2);
    writeValue(settings, "int1", data->int1);
    writeValue(settings, "int2", data->int2);
    writeValue(settings, "puthr", data->puthr);
    writeValue(settings, "pumindist", data->pumindist);
    writeValue(settings, "baserc", data->baserc);
}

template<>
void DecoderChannelConfigIni::readFromIni(const QSettings &settings)
{
    int tmp = data->deconv.fwinf;
    readValue(settings, "fwinf", &tmp);
    data->deconv.fwinf = static_cast<fwinf_t>(tmp);
    readValue(settings, "ppos", &data->deconv.ppos);
    readValue(settings, "pwid", &data->deconv.pwid);
    readValue(settings, "snf", &data->deconv.snf);
    readValue(settings, "pnf", &data->deconv.pnf);
    readValue(settings, "tadj", &data->deconv.tadj);
    readValue(settings, "expfall", &data->deconv.expfall);
    FirConfigIni::staticLoad(settings, &data->deconv.fir, path() + "fir");
    readValue(settings, "blr1", &data->blr1);
    readValue(settings, "blr2", &data->blr2);
    readValue(settings, "int1", &data->int1);
    readValue(settings, "int2", &data->int2);
    readValue(settings, "puthr", &data->puthr);
    readValue(settings, "pumindist", &data->pumindist);
    readValue(settings, "baserc", &data->baserc);
}

template<>
void DecoderConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "mode", data->mode);

    CalZeroConfigIni::staticSave(data->calzero, settings, path() + "calzero");
    CalAvgConfigIni::staticSave(data->calavg, settings, path() + "calavg");

//    for (auto ch : data->channels.keys()) {
//        DecoderChannelConfigIni::staticSave(data->channels[ch], settings, path() + QString("channel/%1").arg(ch));
//    }
    writeTree(settings, "channels", data->channels);

    writeValue(settings, "fprof", static_cast<int>(data->fitprof.fprof));
    writeValue(settings, "fititer", data->fitprof.fititer);
    writeValue(settings, "fitside", data->fitprof.fitside);
    writeValue(settings, "fitfixpar", data->fitprof.fitfixpar);
    writeValue(settings, "fitalgo", data->fitprof.fitalgo);


    writeValue(settings, "zero_thr", data->zero_thr);
    writeValue(settings, "dt_cut_en", data->dt_cut_en);
    writeValue(settings, "dt_cut_l", data->dt_cut_l);
    writeValue(settings, "dt_cut_r", data->dt_cut_r);
}

template<>
void DecoderConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "mode", &data->mode);

    CalAvgConfigIni::staticLoad(settings, &data->calavg, path() + "calavg");
    CalZeroConfigIni::staticLoad(settings, &data->calzero, path() + "calzero");

#if 0
    QStringList groups = settings->childGroups();
    for (auto & group : groups) {
        int ch = group.toInt();
        DecoderChannelConfigIni tmp(this, QString("channel/%1").arg(ch), DecoderChannelConfig());
        tmp.readFromIni(settings, group);
        if (!tmp.data->isValid()) {
            qWarning() << QString("Invalid DecoderChannelConfig for channel %1 in %2").arg(ch).arg(settings->fileName());
        }
        data->channels[ch] = tmp.data;
    }
#endif
    readTree(settings, "channels", &data->channels);

    readValue(settings, "fprof", &data->fitprof.fprof);
    readValue(settings, "fititer", &data->fitprof.fititer);
    readValue(settings, "fitside", &data->fitprof.fitside);
    readValue(settings, "fitfixpar", &data->fitprof.fitfixpar);
    readValue(settings, "fitalgo", &data->fitprof.fitalgo);


    readValue(settings, "zero_thr", &data->zero_thr);
    readValue(settings, "dt_cut_en", &data->dt_cut_en);
    readValue(settings, "dt_cut_l", &data->dt_cut_l);
    readValue(settings, "dt_cut_r", &data->dt_cut_r);
}

template<>
void HistConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "enable", data->enable);
}

template<>
void HistConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "enable", &data->enable);
}

template<>
void RunConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "stopCondition", data->stopCondition);
    writeValue(settings, "stopByTimer", data->stopByTimer);
    writeValue(settings, "stopByEventCount", data->stopByEventCount);
    writeValue(settings, "stopTimerValue", data->stopTimerValue);
    writeValue(settings, "stopEventCountValue", data->stopEventCountValue);
}

template<>
void RunConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "stopCondition", &data->stopCondition);
    readValue(settings, "stopByTimer", &data->stopByTimer);
    readValue(settings, "stopByEventCount", &data->stopByEventCount);
    readValue(settings, "stopTimerValue", &data->stopTimerValue);
    readValue(settings, "stopEventCountValue", &data->stopEventCountValue);
}

template<>
void AllConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "version", data->version);

    AdcConfigIni::staticSave(data->adc, settings, path() + "adc");
    StorageConfigIni::staticSave(data->stor_raw, settings, path() + "StorageConfig_Raw");
    StorageConfigIni::staticSave(data->stor_dev, settings, path() + "StorageConfig_Dev");
    DecoderConfigIni::staticSave(data->dec, settings, path() + "dec");
    HistConfigIni::staticSave(data->hist, settings, path() + "hist");
    RunConfigIni::staticSave(data->run, settings, path() + "run");
}

template<>
void AllConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "version", &data->version);

    AdcConfigIni::staticLoad(settings, &data->adc, "adc");
    StorageConfigIni::staticLoad(settings, &data->stor_raw, "StorageConfig_Raw");
    StorageConfigIni::staticLoad(settings, &data->stor_dev, "StorageConfig_Dev");
    DecoderConfigIni::staticLoad(settings, &data->dec, "dec");
    HistConfigIni::staticLoad(settings, &data->hist, "hist");
    RunConfigIni::staticLoad(settings, &data->run, "run");
}

template<>
void FirConfigIni::writeToIni(QSettings *settings) const
{
    writeValue(settings, "offs", data->offs);
    writeValue(settings, "bits", data->bits);
    writeVector(settings, "coef", QVector<decoder_fir_coef_t>::fromStdVector(data->coef));
}

template<>
void FirConfigIni::readFromIni(const QSettings &settings)
{
    readValue(settings, "offs", &data->offs);
    readValue(settings, "bits", &data->bits);
    auto tmp(QVector<decoder_fir_coef_t>::fromStdVector(data->coef));
    readVector(settings, "coef", &tmp);
    data->coef = tmp.toStdVector();
//    QVector<double> tmp = QVector<double>::fromStdVector(data->coef);
//    readValue(settings, "coef", &tmp);
//    data->coef = tmp.toStdVector();
}
