//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "decoderconfig.h"

#include <cstdio>
#include <cmath>
#include <QDebug>
#include <QTextStream>

#include "adcm16.h"
#include "dsplib/fir.h"
#include "dsplib/fitter.h"
#include "dsplib/dsplib.h"
#include "util/TextFormatter.h"

const double eps = 1e-15;

template<class T>
bool INRANGE(const T &x, const T &min, const T &max)
{
    return !(x < min || x > max);
}

DecoderChannelConfig::DecoderChannelConfig()
{
    puthr = 75;     // %
    pumindist = 100; // ns
    blr1 = -10;
    blr2 = -3;
    int1 = -2;
    int2 = 10;//adc.recwin-adc.latency-(adc.matchwin-adc.prewin);
    baserc = 5;
    //    search_start = PREWIN-12;
    //    search_end = PREWIN+32;
    /*
// a bypass filter
  fir.klen=1;
  fir.offs=0;
  fir.bits=1;
  fir.coef[0]=1;
*/
//    // a useful filter
//    fir.offs = 5;
//    fir.bits = 10;
//    fir.coef.clear();
//    fir.coef.push_back(11);
//    fir.coef.push_back(37);
//    fir.coef.push_back(183);
//    fir.coef.push_back(449);
//    fir.coef.push_back(511);
//    fir.coef.push_back(133);
//    fir.coef.push_back(-349);
//    fir.coef.push_back(-455);
//    fir.coef.push_back(-235);
//    fir.coef.push_back(-57);
//    fir.coef.push_back(-20);
//    fir.coef.push_back(-6);
}

bool DeconvConfig::operator==(const DeconvConfig &other) const
{
    return true
           && fwinf == other.fwinf
           && abs(ppos - other.ppos) < eps
           && abs(pwid - other.pwid) < eps
           && abs(snf - other.snf) < eps
           && abs(pnf - other.pnf) < eps
           && abs(tadj - other.tadj) < eps
           && fir == other.fir
           && expfall == other.expfall
        ;
}

bool DecoderChannelConfig::operator==(const DecoderChannelConfig &other) const
{
    return true
            && deconv == other.deconv
            && blr1 == other.blr1
            && blr2 == other.blr2
            && int1 == other.int1
            && int2 == other.int2
            && puthr == other.puthr
            && pumindist == other.pumindist
            && abs(baserc - other.baserc) < eps
            ;
}

QTextStream &operator<<(QTextStream &s, const DeconvConfig &channel)
{
    s << "# --- Deconvolution Config ---" << endl;
    s << "fwinf: " << channel.fwinf << endl;
    s << "ppos: " << channel.ppos << endl;
    s << "pwid: " << channel.pwid << endl;
    s << "snf: " << channel.snf << endl;
    s << "pnf: " << channel.pnf << endl;
    s << "tadj: " << channel.tadj << endl;
    s << "expfall: " << channel.expfall << endl;
    s << "fir: offs=" << channel.fir.offs << ", coef=" << channel.fir.coef << endl;
    return s;
}

QTextStream &operator<<(QTextStream &s, const DecoderChannelConfig &channel)
{
    s << "# --- Decoder Channel Config ---" << endl;
    s << "blr: " << channel.blr1 << " - " << channel.blr2 << endl;
    s << "int: " << channel.int1 << " - " << channel.int2 << endl;
    s << "puthr: " << channel.puthr << endl;
    s << "pumindist: " << channel.pumindist << endl;
    s << "baserc: " << channel.baserc << endl;
    s << channel.deconv << endl;
    return s;
}

ConfigStatus DeconvConfig::isValid() const
{
    ConfigStatus status("deconv");
    if ((tadj < -200) || (tadj > 200)) {
        status << QString("bad tadj: %1, set default %2").arg(tadj).arg(DeconvConfig().tadj);
    }
    if ((snf < -100) || (snf > 100)) {
        status << QString("bad snf: %1, set default %2").arg(snf).arg(DeconvConfig().snf);
    }
    if ((pnf < -100) || (pnf > 100)) {
        status << QString("bad pnf: %1, set default %2").arg(pnf).arg(DeconvConfig().pnf);
    }
    status &= fir.validate("fir");
    return status;
}
ConfigStatus DecoderChannelConfig::isValid() const
{
    return validate();
}

ConfigStatus DecoderChannelConfig::validate(const QString &title) const
{
    ConfigStatus status(title);
    if ((baserc < 0) || (baserc > 100)) {
        status << QString("bad baserc: %1, set default %2").arg(baserc).arg(DecoderChannelConfig().baserc);
    }
    if (!INRANGE (blr1, -1*(int)ADCM_MAX_N_SAMP, (int)ADCM_MAX_N_SAMP)) {
        status << QString("bad blr1: %1, set default %2").arg(blr1).arg(DecoderChannelConfig().blr1);
    }
    if (!INRANGE (blr2, blr1, (int)ADCM_MAX_N_SAMP)) {
        status << QString("bad blr2: %1, set default %2").arg(blr2).arg(DecoderChannelConfig().blr2);
    }
    if (!INRANGE (int1, -1*(int)ADCM_MAX_N_SAMP, (int)ADCM_MAX_N_SAMP)) {
        status << QString("bad int1: %1, set default %2").arg(int1).arg(DecoderChannelConfig().int1);
    }
    if (!INRANGE (int2, int1, (int)ADCM_MAX_N_SAMP)) {
        status << QString("bad int2: %1, set default %2").arg(int2).arg(DecoderChannelConfig().int2);
    }
    status &= deconv.isValid();
    return status;
}

static const int default_fitside = 6;
static const unsigned int  default_fititer = 2; // 1 is enough. 2 for safe.
static const int default_fitalgo = FIT_ALGO_ACCURATE;
static const int default_fitfixpar = 0;       // bit 0: amplitude, bit1: time, bit2: width. Default: no fixed par

FitProfile::FitProfile()
{
    fprof = FIT_PROFILE_DEFAULT;
    fititer = default_fititer;
    fitside = 0; // 0 - auto
    fitfixpar = default_fitfixpar;
    fitalgo = default_fitalgo;
}

bool FitProfile::operator==(const FitProfile &other) const
{
    return true
            && fprof == other.fprof
            && fititer == other.fititer
            && fitside == other.fitside
            && fitfixpar == other.fitfixpar
            && fitalgo == other.fitalgo
            ;
}

ConfigStatus FitProfile::isValid() const
{
    ConfigStatus status("fitprofile");
    if ((fititer < 0) || (fititer > 10000)) {
        status << QString("bad fititer %1, set %2").arg(fititer).arg(FitProfile().fititer);
    }
    if ((fitside < 0) || (fitside > 63)) {
        status << QString("bad fitside %1, set %2").arg(fitside).arg(FitProfile().fitside);
    }
    return status;
}

int FitProfile::get_dec_fititer() const
{
    int fiter;
    switch (fprof) {
    case FIT_PROFILE_SUPERFAST:
        fiter = 1;
        break;
    case FIT_PROFILE_FAST:
        fiter = 1;
        break;
    case FIT_PROFILE_OPTIMAL:
        fiter = 2;
        break;
    case FIT_PROFILE_PRECISION:
        fiter = 26;
        break;
    case FIT_PROFILE_CUSTOM:
        fiter = fititer;
        break;
    default:
        fiter = default_fititer;
        break;
    }

    if (fiter > 1000)
        fiter = default_fititer;
    return fiter;
}

/*fit_algo_t*/
int FitProfile::get_dec_fitalgo() const
{
    int algo = default_fitalgo;
    if (fprof == FIT_PROFILE_CUSTOM) algo = fitalgo;

    return algo;
}

int FitProfile::get_dec_fitfixpar() const
{
    int fixpar = default_fitfixpar;
    if (fprof == FIT_PROFILE_CUSTOM) fixpar = fitfixpar;

    return fixpar;
}

int FitProfile::get_fit_half_np(double pulse_width_steps) const
{
    int fside = 0;
    double k = 1.0, b = 0;
    if (fprof == FIT_PROFILE_SUPERFAST) {k = 0.7; }
    if (fprof == FIT_PROFILE_FAST) {k = 1.0; }
    if (fprof == FIT_PROFILE_OPTIMAL) { k=1.2; b = 1; }
    if (fprof == FIT_PROFILE_PRECISION) {k = 1.6; b = 1; }
    if (fprof == FIT_PROFILE_CUSTOM) {fside = fitside;}

    if (fside <= 0) // auto
        fside = static_cast<int>(trunc(pulse_width_steps * k + b));
    if (fside<=0 || fside > 63)
        fside = default_fitside;
    if (fside < 2)
        fside = 2;
    // safe check
    int maxfside = (MAX_FITGAUS_POINTS-1)/2;
    if (fside > maxfside)
        fside = maxfside;

    //  printf("ch %2d: pwid=%f, k=%.1f, b=%.0f  fside=%d\n", ch, dec.pwid[ch], k, b, fside);
    return fside;
}


DecoderConfig::DecoderConfig(int n)
    : initialized_nch(n)
{
//    channels.resize(nch());

    mode = DM_OFF;
    zero_thr = 10;       // mV
    dt_cut_en = 0;
    dt_cut_l = -300;
    dt_cut_r = 300;
}

ConfigStatus DecoderConfig::isValid() const
{
    ConfigStatus status("dec");
    if (magic != expected_magic) {
        status << QString("bad dec.magic: %1, expected %2").arg(magic).arg(expected_magic);
    }
    status &= fitprof.isValid();
    for (auto c : channels.keys()) {
        //    ts = adc.step[c];
        //    if ((dec.search_start[c]<0) || (dec.search_start[c]>RECWIN)) {printf("bad dec.search_start[%d]: %d\n", c, dec.search_start[c]); ok=0;}
        //    if ((dec.search_end[c]<0) || (dec.search_end[c]>RECWIN)) {printf("bad dec.search_end[%d]: %d\n", c, dec.search_end[c]); ok=0;}
        status &= channels[c].validate(QString("channel[%1]").arg(c));
    }
    return status;
}

bool DecoderConfig::operator==(const DecoderConfig &other) const
{
    return true
            && mode == other.mode
            && calzero == other.calzero
            && calavg == other.calavg
            && channels == other.channels
            && fitprof == other.fitprof
            && abs(zero_thr - other.zero_thr) < eps
            && dt_cut_en == other.dt_cut_en
            && abs(dt_cut_l - other.dt_cut_l) < eps
            && abs(dt_cut_r - other.dt_cut_r) < eps
            ;
}

int DecoderConfig::nch() const
{
    return initialized_nch;
}

QTextStream &operator<<(QTextStream &s, const FitProfile &fitprof)
{
    s << "# --- FitProfile ---" << endl;
    s << "fprof: " << fitprof.fprof << endl;
    s << "fititer: " << fitprof.fititer << endl;
    s << "fitside: " << fitprof.fitside << endl;
    s << "fitfixpar: " << fitprof.fitfixpar << endl;
    s << "fitalgo: " << fitprof.fitalgo << endl;
    //             (fitfixpar & 1) ? "a " : "", (fitfixpar & 2) ? "t " : "", (fitfixpar & 4) ? "w " : "");
    return s;
}

QTextStream &operator<<(QTextStream &s, const DecoderConfig &dec)
{
    s << "# --- Decoder Config ---" << endl;
    s << "mode: " << dec.mode << endl;
    s << "zero_thr: " << dec.zero_thr << endl;
    s << "dT cut: " << (dec.dt_cut_en ? "enabled" : "disabled") << ", " << dec.dt_cut_l << " - " << dec.dt_cut_r << " ns" << endl;
    s << "# channels" << endl;
    if (dec.channels.isEmpty()) {
        s << "# empty" << endl;
    } else {
        for (auto i : dec.channels.keys())
            s << i << dec.channels[i] << endl;
    }
    s << dec.calzero;
    s << dec.calavg;
    s << dec.fitprof;
    return s;
}
