/*
**    Copyright 2013 Ivan Filippov
**    Copyright 2013-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef INIBASED_H
#define INIBASED_H

#include <QDebug>
#include <QString>
#include <QVector>
#include <QMap>
#include <QVariant>
#include <QSettings>

#include "TreeElement.h"

template <typename T>
class BaseConfigIni;

class IniBased : public TreeElement
{
public:
    IniBased(const IniBased &) = delete;
    IniBased(const IniBased &&) = delete;
    IniBased &operator=(const IniBased &) = delete;
    IniBased &operator=(const IniBased &&) = delete;

    IniBased(IniBased *parent = nullptr, const QString &section = QString());
    ~IniBased() override = default;

    void save(QSettings *settings) const;
    void load(const QSettings &settings);
    virtual QString path() const;

    void writeValue(QSettings *settings, const QString &key, const QVariant &data) const;

    template<typename T>
    void writeVector(QSettings *settings, const QString &key, const QVector<T> &data) const
    {
        QStringList list;
        for (auto & item : data) {
            const QVariant &var = QVariant::fromValue(item);
            Q_ASSERT(var.isValid());
            list.push_back(var.value<QString>());
        }
        if (data.isEmpty())
            writeValue(settings, key, "");
        else
            writeValue(settings, key, list);
    }

    template<typename K, typename V>
    void writeValue(QSettings *settings, const QString &key, const QMap<K, V> &data) const
    {
        QStringList list;
        for (auto & k : data.keys()) {
            const V &v = data[k];
            const QVariant &var = QVariant::fromValue(v);
            QString strVal = var.value<QString>();
            list.push_back(QString("{%1, %2}").arg(k).arg(strVal));
//            pv.push_back(QPair<K, V> (k, data[k]));
        }
        if (list.isEmpty())
            writeValue(settings, key, "");
        else
            writeValue(settings, key, list.join("; "));
//        QVariantList list;
//        for (auto & item : data) {
//            const QVariant &var = QVariant::fromValue(item);
//            Q_ASSERT(var.isValid());
//            list.push_back(var);
//        }
//        writeValue(settings, key, list);
    }

    template<typename T>
    void writeTree(QSettings *settings, const QString &key, const QVector<T> &data) const
    {
        for (int i=0; i<data.size(); i++)
            BaseConfigIni<T>::staticSave(data[i], settings, path() + key + QString("/%1").arg(i));
    }

    template<typename K, typename V>
    void writeTree(QSettings *settings, const QString &key, const QMap<K, V> &data) const
    {
        const QString &fullKey = path() + key;
        for (auto & i : data.keys()) {
            BaseConfigIni<V>::staticSave(data[i], settings, fullKey + QString("/%1").arg(i));
//            writeValue(settings, QString("%1/%2").arg(key).arg(i), QVariant::fromValue(data[i]));
        }
    }

//    void writeValue(QSettings *settings, const QString &key, const QVector<int> &data) const;
//    void writeValue(QSettings *settings, const QString &key, const QVector<unsigned int> &data) const;
//    void writeValue(QSettings *settings, const QString &key, const QVector<double> &data) const;

    template<typename T>
    T readValue(const QSettings &settings, const QString &key, T *data = nullptr)
    {
        const QString &fullKey = path() + key;
        T tmp;
        if (data) tmp = *data;
        if (!settings.contains(fullKey))
            return tmp;
        tmp = unpack<T>(settings.value(fullKey, tmp));
        if (data)
            *data = tmp;
        return tmp;
    }

    template<typename T>
    QVector<T> readVector(const QSettings &settings, const QString &key, QVector<T> *data = nullptr)
    {
        const QString &fullKey = path() + key;
        QVector<T> tmp;
        if (data) tmp = *data;
        if (!settings.contains(fullKey))
            return tmp;
        QVariant var = settings.value(fullKey);
        if (var.isNull() || !var.isValid() || !var.canConvert<QVariantList>())
            return tmp;
//        var.convert(QMetaType::QStringList);
        QVariantList list = var.toList();
        tmp.resize(0);
        for (auto & s : list) {
            QVariant var(s);
            const T &value = var.value<T>();
            tmp.push_back(value);
        }
        if (data)
            *data = tmp;
        return tmp;
    }

    template<typename K, typename V>
    QMap<K, V> readMap(const QSettings &settings, const QString &key, QMap<K, V> *data = nullptr)
    {
        const QString &fullKey = path() + key;
        QMap<K, V> tmp;
        if (data) tmp = *data;
        if (!settings.contains(fullKey))
            return tmp;
        QString valStr = settings.value(fullKey).toString();
        QMap<QString, QString> m(parseMap(valStr));
        for (auto &k : m.keys()) {
            const K &kVar = QVariant::fromValue(k).value<K>();
            QTextStream stream(&m[k]);
            const V &vVar = QVariant::fromValue(m[k]).value<V>();
            tmp[kVar] = vVar;
        }
        if (data)
            *data = tmp;
        return tmp;
    }

    template<typename T>
    QVector<T> readTree(const QSettings &settings, const QString &key, QVector<T> *data = nullptr)
    {
        const QList<int> &keys = childKeysNumeric(settings, key);
        QVector<T> tmp;
        if (data) tmp = *data;
        for (auto & i : keys) {
            if (tmp.size() <= i)
                tmp.resize(i+1);
            BaseConfigIni<T>::staticLoad(settings, &tmp[i], path() + key + QString("/%1").arg(i));
        }
        if (data)
            *data = tmp;
        return tmp;
    }

    template<typename K, typename V>
    QMap<K, V> readTree(const QSettings &settings, const QString &key, QMap<K, V> *data = nullptr)
    {
        const QString &fullKey = path() + key;
        QMap<K, V> tmp;
        if (data) tmp = *data;

        const QStringList &keys = childKeys(settings, key);
        for (auto & i : keys) {
            QVariant ivar; ivar = i;
            const K &k = ivar.value<K>();
            BaseConfigIni<V>::staticLoad(settings, &tmp[k], path() + key + QString("/%1").arg(i));
//            QVariant var = settings.value(fullKey + "/" + i);
//            if (var.isNull() || !var.isValid() || !var.canConvert<V>())
//                continue;
//            const V &v = var.value<V>();
//            tmp[k] = v;
        }
        if (data)
            *data = tmp;
        return tmp;
    }

protected:
    virtual void writeToIni(QSettings *settings) const = 0;
    virtual void readFromIni(const QSettings &settings) = 0;
    QStringList childKeys(const QSettings &settings, const QString &key) const;
    QList<int> childKeysNumeric(const QSettings &settings, const QString &key) const;

private:
    template<typename T>
    T unpack(const QVariant &var, const T &defVal = T())
    {
        if (var.isValid() && var.canConvert<T>()) {
            return var.value<T>();
        }
        return defVal;
    }
    static QMap<QString, QString> parseMap(const QString &str);
    QString sectionName;
};

enum {
    NO_MERGE_CONFIG = false,
    MERGE_CONFIG = true,
};

template <typename T>
class BaseConfigIni : public IniBased
{
public:
    BaseConfigIni() = delete;
    BaseConfigIni(const BaseConfigIni &) = delete;
    BaseConfigIni(const BaseConfigIni &&) = delete;
    BaseConfigIni &operator=(const BaseConfigIni &) = delete;
    BaseConfigIni &operator=(const BaseConfigIni &&) = delete;

    BaseConfigIni(IniBased *parent = nullptr, const QString &section = QString())
        : IniBased(parent, section),
        data(new T), created(true) {}

    BaseConfigIni(T *r, IniBased *parent = nullptr, const QString &section = QString())
        : IniBased(parent, section),
        data(r) {}
    ~BaseConfigIni() override {
        if (created)
            delete data;
    }

    static void staticSaveMap(const QMap<int, T> &m, QSettings *settings, const QString &fullPath = QString())
    {
        for (int i : m.keys()) {
            const T &channels = m[i];
            BaseConfigIni<T>::staticSave(channels, settings, fullPath + QString("/%1").arg(i));
        }
    }

    static void staticSave(const T &r, QSettings *settings, const QString &fullPath = QString())
    {
        BaseConfigIni ini(nullptr, fullPath);
        ini.setData(r);
        ini.save(settings);
    }

    static T staticLoad(const QSettings &settings, const QString &key = QString())
    {
        T r;
        BaseConfigIni ini(&r, nullptr, key);
        ini.load(settings);
        return r;
    }
    static T staticLoad(const QSettings &settings, T *r, const QString &fullPath = QString(), bool merge = MERGE_CONFIG)
    {
        if (merge == NO_MERGE_CONFIG)
            *r = T();
        BaseConfigIni ini(r, nullptr, fullPath);
        ini.load(settings);
        return *r;
    }

protected:
    void setData(const T &r) { *data = r; }
    T getData() const  { return *data; }
    void writeToIni(QSettings *settings) const override;
    void readFromIni(const QSettings &settings) override;

    T *data;

private:
    bool created = false;
};

template <typename T>
class AdvConfigIni : public BaseConfigIni<T>
{
public:
    AdvConfigIni(IniBased *parent = nullptr, const QString &section = QString())
        : BaseConfigIni<T>(parent, section) {}

};

#endif // INIBASED_H
