//
//    Copyright 2010-2019 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "DeconvConfigDialog.h"

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <algorithm>

#include <QDebug>
#include <QFileDialog>
#include <QMessageBox>
#include <QComboBox>
#include <QLabel>
#include <QSpinBox>
#include <QSettings>
#include <qwt_scale_engine.h>

//#include "qxw/WaveformView.h"
#include "qxw/WaveformPlot.h"
#include "qxw/WaveformData.h"

#include "config/inibased.h"
#include "config/allconfig.h"
//#include "adcmraw/adchardw.h"
#include "dsplib/fir.h"
#include "config/firconfig.h"
#include "dsplib/fir_ini.h"
#include "dsplib/dsplib.h"
#include "dsplib/fitter.h"
#include "dsplib/fftengine.h"
#include "dsplib/Convolver.h"
#include "adcmlib/ch_label.h"
#include "qlog/qlog.h"
//#include "plot.h"

#include "dsplib/fftengine.h"
#include "util/TextFormatter.h"

#ifdef ALL_DEBUG
# undef DCCONF_DEBUG
# define DCCONF_DEBUG ALL_DEBUG
#endif

#ifndef DCCONF_DEBUG
# define DCCONF_DEBUG 0         /* dummy */
#endif

#define PREFIX "dcconf: "

static double INPFFT_MIN = 1e-1;
static double INPFFT_MAX = 1e3;

static const int calibrationColorIndex = 4;
static const int residualsColorIndex = 0;
static const int referenceColorIndex = 1;
static const int filterColorIndex = 6;
static const int testColorIndex = 4;

DeconvConfigDialog::DeconvConfigDialog (QWidget* parent)
    : QDialog(parent),
      ui(this),
      fft(new RFFTEngine)
{
    setupUi(this);
    //    extcfg = incfg;
    disable_control_slots = 0;
    restoring = 0;
    ch = 0;

    comboBoxWindowingFunction->addItem ("Rectangle", WF_RECT);
    comboBoxWindowingFunction->addItem ("Blackman", WF_BLACKMAN);
    comboBoxWindowingFunction->addItem ("Hamming", WF_HAMMING);
    comboBoxWindowingFunction->addItem ("Hanning", WF_HANN);
    comboBoxWindowingFunction->addItem ("Barlett", WF_BARLETT);
    comboBoxWindowingFunction->addItem ("Tukey 0.75", WF_TUKEY_075);
    comboBoxWindowingFunction->addItem ("Tukey 0.5", WF_TUKEY_050);
    comboBoxWindowingFunction->addItem ("Tukey 0.25", WF_TUKEY_025);
    comboBoxWindowingFunction->addItem ("Gauss 0.25", WF_GAUSS_025);
    comboBoxWindowingFunction->addItem ("Gauss 0.5", WF_GAUSS_050);
    comboBoxWindowingFunction->addItem ("Gauss 0.75", WF_GAUSS_075);
//    win_type = WF_BLACKMAN;             // default
//    comboBoxWindowingFunction->setCurrentIndex(win_id_by_type((fwinf_t)newDecoderConfig.fwinf[0]));

    sbfirnp->setMaximum(FIRMAXLEN);
    //  sbfirnp->setValue(FIRMAXLEN);

    Restore();
}

DeconvConfigDialog::~DeconvConfigDialog ()
{
    delete fft;
}

void DeconvConfigDialog::setAdcConfig(const AdcConfig &r)
{
    if (r == adcConfig)
        return;
    adcConfig = r;
}

void DeconvConfigDialog::setDecoderConfig(const DecoderConfig &r)
{
    if (r == newDecoderConfig)
        return;
    newDecoderConfig = r;
    Restore();
}

fwinf_t DeconvConfigDialog::win_type() const
{
    return newDecoderConfig.channels[ch].deconv.fwinf;
}

QVector<double> DeconvConfigDialog::getCalAvgY(int channel) const
{
    QVector<double> calAvg;
    const auto & data = newDecoderConfig.calavg.channels[channel].get();
    calAvg.reserve(data.size());
    for (auto i : data) {
        calAvg.push_back(i);
    }
    return calAvg;
}

double DeconvConfigDialog::getCalAvgAmplitude(int channel) const
{
    const auto &calAvg = getCalAvgY(channel);
    if (calAvg.empty())
        return 0;
    return std::max(1.0, *std::max_element(calAvg.begin(), calAvg.end()));
}

size_t DeconvConfigDialog::getRefSize() const
{
    return newDecoderConfig.calavg.channels[ch].get().size();
}

std::vector<double> DeconvConfigDialog::getRefY_n(int channel, double amplitudeScaler, int _n) const
{
    const int lat = data_latency ();
    const size_t n = _n;
    std::vector<double> ref;
    ref.reserve(n);
    for (size_t i = 0; i < n; i++) {
        ref.push_back(amplitudeScaler * reqfunscaled (int(i) - lat));
    }
    return ref;
}

//std::vector<double> DeconvConfigDialog::getRefY(int channel, double amplitudeScaler) const
//{
//    return getRefY_n(channel, amplitudeScaler, getRefSize());
//}

std::vector<double> DeconvConfigDialog::getRefY(int channel) const
{
    return getRefY_n(channel, getCalAvgAmplitude(channel), getRefSize());
}

std::vector<double> DeconvConfigDialog::getRefXInNs(int channel) const
{
    const int lat = data_latency ();
    const size_t n = adcConfig.recwin(channel);
    std::vector<double> x;
    x.reserve(n);
    for (size_t i = 0; i < n; i++) {
        x.push_back((int(i) - lat) * getTimeStepInNs());
    }
    return x;
}

std::vector<double> DeconvConfigDialog::getCalAvgXInNs(int channel) const
{
    return getRefXInNs(channel);
}

void DeconvConfigDialog::replotINP ()
{
    WaveformPlot *wp = ui->plotINP; // Input signal

    std::vector<double> vx = getCalAvgXInNs(ch);
    std::vector<double> vy0 = getCalAvgY(ch).toStdVector();
    std::vector<double> vy1 = getRefY(ch);
    size_t n = std::min(vx.size(), std::min(vy0.size(), vy1.size()));
    if (n == 0) {
        wp->clear();
        wp->replot();
        return;
    }
    vx.resize(n);
    vy0.resize(n);
    vy1.resize(n);
    WaveformDataAll wda;
    wda[0] = WaveformData(vx, vy0, tr("Calibration data"), calibrationColorIndex);
    wda[1] = WaveformData(vx, vy1, tr("Reference data"), referenceColorIndex);
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time, ns"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();
}

void DeconvConfigDialog::replotINPFFT ()
{
    WaveformPlot *wp = ui->plotINPFFT; // Input signal FFT


    const std::vector<double> &inp = getCalAvgY(ch).toStdVector();
    const std::vector<double> &req = getRefY(ch);
    const size_t N = std::min(inp.size(), req.size());
    // Find power spectra and peaks
    std::vector<std::complex<double> > inpDFT = fft->DFT(inp);
    inpDFT.resize(N/2+1);
    std::vector<std::complex<double> > reqDFT = fft->DFT(req);
    reqDFT.resize(N/2+1);
    const double dftScaler = 1. / sqrt(N);
    const double fullScaleSigned = 1 << (adcConfig.bits-1);
    const double fullScaleDb = val2db(fullScaleSigned);

    const size_t inpDftSize = inpDFT.size();
    std::vector<double> inpDftPy;
    for (size_t i = 0; i < inpDftSize; i++) {
        inpDftPy.push_back(val2db(std::abs(inpDFT[i]) * dftScaler) - fullScaleDb); // normalize DFT
    }
    const size_t reqDftSize = reqDFT.size();
    std::vector<double> dftPx;
    dftPx.reserve(reqDftSize);
    for (size_t i = 0; i < reqDftSize; i++) {
        dftPx.push_back(double(i) / N / getTimeStepInNs() * 1000.0);
    }

    std::vector<double> reqDftPy;
    reqDftPy.reserve(reqDftSize);
    for (size_t i = 0; i < reqDftSize; i++) {
        reqDftPy.push_back(val2db(std::abs(reqDFT[i]) * dftScaler) - fullScaleDb); // normalize DFT
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(dftPx, inpDftPy, "Source FFT", calibrationColorIndex);
    wda[1] = WaveformData(dftPx, reqDftPy, "Target FFT", referenceColorIndex);

    // horizontal markers
//    const CalAvgConfig &calavg = newDecoderConfig.calavg;
//    const CalAvgChannelConfig &cal = calavg.data[ch];
    if (true) {
        const double snfdb = newDecoderConfig.channels[ch].deconv.snf - fullScaleDb;
        const double rnfdb = newDecoderConfig.channels[ch].deconv.pnf - fullScaleDb;
        //  const double snf = db2val (snfdb);
        //  const double rnf = db2val (rnfdb);

        std::vector<double> nfpx(2), snfpy(2, snfdb), rnfpy(2, rnfdb);
        if (reqDftPy.size() >= 2) {
            nfpx[0] = 0;
            //      nfpx[1] = reqDftPy.size() - 1;
            nfpx[1] = dftPx.back();
        } else {
            nfpx[0] = 0;
            nfpx[1] = 1;
        }

        wda[2] = WaveformData(nfpx, snfpy, "Signal Noise Floor", calibrationColorIndex+8);
        wda[3] = WaveformData(nfpx, rnfpy, "Target Noise Floor", referenceColorIndex+8);
    }
    wp->setWaveformData(wda);

    //  wp->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
    //  wp->axisScaleEngine(QwtPlot::yLeft)->setAttribute(QwtScaleEngine::Floating);
    //  wp->setAxisOptions (QwtPlot::yLeft, QwtAutoScale::Floating | QwtAutoScale::Logarithmic);
    //  wp->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
    //  wp->setxrange (0, H);

    //  printf("%edB  %e   %edB  %e\n", snf, snfdb, rnf, rnfdb);
    //  printf("%e:%e - %e:%e\n", px[0], p1[0], px[1], p1[1]);
    //  printf("%e:%e - %e:%e\n", px[0], p2[0], px[1], p2[1]);
    wp->enableLegend(false);

    QwtText xTitle(tr("Frequency, MHz"));
    QwtText yTitle(tr("dBFS"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}

void DeconvConfigDialog::replotFIR ()
{
    WaveformPlot *wp = ui->plotFIR; // FIR impulse

    const auto &fir = newDecoderConfig.channels[ch].deconv.fir;
    if (fir.isEmpty()) {
        wp->clear();
        ui->tlklen->clear();
        ui->tloffs->clear();
        ui->tlsum->clear();
        return;
    }

    const size_t klen = fir.coef.size();
    const int firdelay = fir.offs;
    std::vector<double> px, FIRpy, WFpy;
    for (size_t i = 0; i < fir.coef.size(); i++) {
        px.push_back(i * getTimeStepInNs());
        FIRpy.push_back(fir.coef[i]);
        WFpy.push_back(windowingFunction (win_type(), i, klen));
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(px, FIRpy, "Filter Impulse Response", filterColorIndex);
    wda[1] = WaveformData(px, WFpy, "Windowing", filterColorIndex+2);
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time, ns"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();

    ui->tlklen->setText (QString("Size: %1").arg(klen));
    ui->tloffs->setText (QString("Offs: %1").arg(firdelay));
    double sum = 0;
    for (size_t  i = 0; i < klen; i++)
        sum += fir.coef[i];
    ui->tlsum->setText (QString("Sum: %1").arg(sum));
}

void DeconvConfigDialog::replotFIRFFT ()
{
    WaveformPlot *wp = ui->plotFIRFFT; // FIR impulse FFT
    const auto &fir = newDecoderConfig.channels[ch].deconv.fir;

    if (fir.isEmpty()) {
        wp->clear();
        return;
    }
    std::vector<double> firCoef;
    for (auto value : fir.coef)
        firCoef.push_back(value);


    const size_t N = firCoef.size();
    std::vector<std::complex<double> > firDft = fft->DFT(firCoef);
    firDft.resize(N/2+1); // result is symmetric
    const double dftScaler = 1. / sqrt(N);
    std::vector<double> px, FirDftMagnitude, FirDftPhase;
    for (size_t i = 0; i < firDft.size(); i++) {
        px.push_back(double(i) / N / getTimeStepInNs() * 1000.0); // ns, MHz
        FirDftMagnitude.push_back(val2db(std::abs(firDft[i]) * dftScaler));
        FirDftPhase.push_back(std::arg(firDft[i]));
    }
    std::vector<double> FirDftPhaseUnwrap;
    FirDftPhaseUnwrap.push_back(FirDftPhase[0]);
    for (size_t i=1; i < FirDftPhase.size(); i++) {
        double delta = FirDftPhase[i] - FirDftPhase[i-1];
        if (delta > M_PI)
            delta -= 2*M_PI;
        if (delta < -M_PI)
            delta += 2*M_PI;
        FirDftPhaseUnwrap.push_back(FirDftPhaseUnwrap[i-1] + delta);
    }
    std::vector<double> FirDftGroupDelay;
    FirDftGroupDelay.push_back(0);
    for (size_t i=1; i < FirDftPhaseUnwrap.size(); i++) {
        FirDftGroupDelay.push_back(FirDftPhaseUnwrap[i] - FirDftPhaseUnwrap[i-1]);
    }
    WaveformDataAll wda;
    wda[0] = WaveformData(px, FirDftMagnitude, "Filter FFT", filterColorIndex);
//    wda[1] = WaveformData(px, FirDftGroupDelay, "Group delay", filterColorIndex+2);
//    wda[2] = WaveformData(px, FirDftPhaseUnwrap, "Filter phase", filterColorIndex+4);
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Frequency, MHz"));
    QwtText yTitle(tr("Amplitude, dB"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}

struct gauss_param_t {
    double a = 0;
    double t = 0;
    double w = 1;
    double aerr = 1;
    double terr = 1;
    double werr = 1;
    double error = 1;
};

gauss_param_t DeconvConfigDialog::find_hit(const std::vector<double> &testResult) const
{
    gauss_param_t p;
    FitterGaus fitterGaus(testResult.size());

    int peakIndex = 0;
    double peakValue = testResult[0];
    for(size_t i=0; i<testResult.size(); i++) {
        if (testResult[i] > peakValue) {
            peakIndex = i;
            peakValue = testResult[peakIndex];
        }
    }
    //double err[3] = { 0, 0, 0 };
    const CalAvgChannelConfig &calAvgChannel = newDecoderConfig.calavg.channels[ch];
    int fs = AllConfig::get_dec_fitside(calAvgChannel.step, newDecoderConfig, ch);
    const size_t N = testResult.size();
    int xl = peakIndex - fs;
    int xr = peakIndex + fs;
    if (xl < 0) xl = 0;
    if (xr >= int(N)) xr = int(N) - 1;
    qDebug() << QString("peakIndex=%1, fit range %2:%3").arg(peakIndex).arg(xl).arg(xr);
    //    fitstatus_t fitstatus;
    //    fitstatus.niter = 100;
    //    FindGausParam (0, TSTpy, rlen, par, &fitstatus);
    std::vector<double> fitData;
    for (int i=xl; i<xr; i++)
        fitData.push_back(testResult[i]);

    //    GausPar par = GaussianApproximator::FindGausApproxShifted (fitData);
    //    p.a = par.a;
    //    p.t = xl + par.x0;
    //    p.w = par.sigma;

    {
        double par[3]={0,0,0};
        fitstatus_t fitstatus;
        fitstatus.niter = newDecoderConfig.fitprof.get_dec_fititer();
        fitterGaus.fit (newDecoderConfig.fitprof.get_dec_fitfixpar(), &fitData[0], fitData.size(), par, &fitstatus);
        p.a = par[0];
        p.t = par[1];
        p.w = par[2];
        double sum2 = 0;
        double sumerr2 = 0;
        for (int i=0; i<fitData.size(); i++) {
            double y = fitData[i];
            double yfit = p.a * gauss(i - p.t, p.w);
            sum2 += pow(y, 2);
            sumerr2 += pow(y - yfit, 2);
        }
        sum2 /= fitData.size();
        sumerr2 /= fitData.size();
        p.error = sqrt(sumerr2 / sum2);
        qDebug() << sum2 << sumerr2 << p.error;
    }
    p.t += xl;

    //    QuickFitGaus (&TSTpy[xl], xr-xl, par, &fitstatus);
    //    printf("(2) par = %f, %f, %f; xp=%d, fs=%d, newrange=%d-%d\n", par[0], par[1]+xl, par[2], xp, fs, xl, xr);
    //    printf("(3) par err = %f, %f, %f\n", (apar[0]-par[0]), (apar[1]-(par[1]+xl)), (apar[2]-par[2]));

    //	printf ("Expected latency: %f\n", (double)lat);
    //	printf ("Measured latency: %f\n", par[1]+xl);
    return p;
}

void DeconvConfigDialog::replotTEST ()
{
    WaveformPlot *wp = ui->plotTEST; // Test impulse

    const auto &fir = newDecoderConfig.channels[ch].deconv.fir;
    if (fir.isEmpty()) {
        wp->clear();
        return;
    }

    const int firdelay = fir.offs;

    const CalAvgChannelConfig &calAvgChannel = newDecoderConfig.calavg.channels[ch];

    const std::vector<double> &inp = getCalAvgY(ch).toStdVector();

    const std::vector<double> &testResult = Convolver::convolve (fir.coef, inp);

    const int lat = data_latency () + firdelay;
    const double timeStepInNs = getTimeStepInNs();     // nanoseconds

    //    QwtText yTitle(tr("Amplitude, mV"));
    //    yTitle.setFont(font());
    //    wp->setAxisTitle(QwtPlot::yLeft, yTitle);

    //printf("replot ok\n");

    if (calAvgChannel.isEmpty() || testResult.empty()) {
        ui->textLabelFitStatus->clear ();
        wp->clear();
        wp->replot();
        return;
    }
    // find peak value
    const gauss_param_t &par = find_hit(testResult);
    const QString plusMinus = QString::fromUtf8("\u00B1");
    QString text_a = QString("Ampl: %1 %2 %3")
                       .arg(par.a, 0, 'f', 1)
                       .arg(plusMinus)
                       .arg(par.aerr, 0, 'f', 1);
    QString text_t = QString("Pos: %1 %2 %3")
                         .arg((par.t/* + xl*/ - lat) * timeStepInNs, 0, 'f', 1)
                         .arg(plusMinus)
                         .arg(par.terr * timeStepInNs, 0, 'f', 2);
    QString text_w = QString("Width: %1 %2 %3")
                         .arg(GausFWHM * par.w * timeStepInNs, 0, 'f', 2)
                         .arg(plusMinus)
                         .arg(par.werr * timeStepInNs, 0, 'f', 2);
    QString text_err = QString("Rel. error: %1")
                         .arg(par.error, 0, 'e', 2);
    ui->textLabelFitStatus->setText (text_a + "<br>" +  text_t + "<br>" + text_w + "<br>" + text_err);

    std::vector<double> px;
    std::vector<double> py;
    std::vector<double> perr;
    for (size_t i = 0; i < testResult.size(); i++) {
        px.push_back((int(i) - lat) * timeStepInNs);
        double y = testResult[i];
        py.push_back(y);
        double yfit = par.a * gauss(i - par.t, par.w);
        perr.push_back((y - yfit) * 100);
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(px, py, "Result waveform", testColorIndex);
    wda[1] = WaveformData(px, perr, "Residuals waveform * 100", residualsColorIndex);
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time")+QString::fromUtf8(", ns"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();
}

void DeconvConfigDialog::Restore ()
{
    //printf("Restore:%d\n", restoring);

    if (restoring)
        return;

    //  printf("Restore (ch=%d)\n", ch);
    restoring = 1;
    disable_control_slots = 1;

    //const char *chname[ADC_NCH] = {"A", "B", "C", "D", "E", "F", "G", "H"};
    QString s;
    s = "[";
    if (adcConfig.channels[ch].flags.bit.m)
        s += ustr_ch_label_M;
    else
        s += " ";
    if (adcConfig.channels[ch].flags.bit.s)
        s += ustr_ch_label_S;
    else
        s += " ";
    s += "]";
    tlchannel->setText (s);
    tlchannel->setEnabled (adcConfig.channels[ch].flags.bit.en);
    //  tlchannel->setPaletteForegroundColor(QColor(rgbcmap[ch][0],rgbcmap[ch][1],rgbcmap[ch][2]));

    sbpos->setValue (newDecoderConfig.channels[ch].deconv.ppos);
    sbwid->setValue (newDecoderConfig.channels[ch].deconv.pwid);
    sbexp->setValue (newDecoderConfig.channels[ch].deconv.expfall);
    sbblrmd->setValue ((100 * newDecoderConfig.channels[ch].baserc));
    /*
  CreateFIR();
  calcRES();

*/
    if (newDecoderConfig.nch() > 0) {
        spinBoxChannel->setMaximum(newDecoderConfig.nch());
        spinBoxChannel->setEnabled(true);
//        spinBoxChannel->setMaxValue (newDecoderConfig.nch() - 1);
  } else {
        spinBoxChannel->setMaximum(0);
        spinBoxChannel->setEnabled(false);
    }

    const double noiseRmsAbsMin = 1. / sqrt (12); // quantization noise
//    double noiseRms = std::max(noiseRmsAbsMin, cfg->getcalnoiserms(ch));
    const double fftGain = sqrt(adcConfig.recwin(ch));
    INPFFT_MIN = noiseRmsAbsMin / fftGain;  // set to noise level
    INPFFT_MAX = adcConfig.adc_rawdata_maxval(ch) / fftGain;    // set to maximum signal level
    //    ui->plotINPFFT->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
    //    QwtScaleEngine *scaleEngine = ui->plotINPFFT->axisScaleEngine(QwtPlot::yLeft);
    //    scaleEngine->setReference(val2db(INPFFT_MIN));
    int snf_db_min = val2db(INPFFT_MIN);
    int snf_db_max = val2db(INPFFT_MAX);
    sbsnf->setRange (snf_db_min, snf_db_max);
    sbrnf->setRange (snf_db_min, snf_db_max);
    sbsnf->setValue (newDecoderConfig.channels[ch].deconv.snf);
    sbrnf->setValue (newDecoderConfig.channels[ch].deconv.pnf);
    //    const double defaultNoiseFloorDb = val2db(noiseRms / fftGain);
    //    sbsnf->setValue ((defaultNoiseFloorDb));
    //    sbrnf->setValue ((defaultNoiseFloorDb));

    char str[40];
    double nrms = 0;
    bool empty = newDecoderConfig.calzero.channels[ch].isEmpty();
    if (empty)
        sprintf (str, "Baseline: uncal.");
    else
        sprintf (str, "Baseline: %.1f", newDecoderConfig.calzero.channels[ch].get_base());
    tlbl->setText (str);
    if (empty) {
        sprintf (str, "RMS Noise: uncal.");
        nrms = 1000;
    } else {
        nrms = sqrt (newDecoderConfig.calzero.channels[ch].get_rms());
        sprintf (str, "RMS Noise: %.1f", nrms);
    }
    tlnrms->setText (str);

    if (empty) {
        tlbl->setStyleSheet("QLabel { color : red; }");
        tlnrms->setStyleSheet("QLabel { color : red; }");
    } else {
        tlbl->setPalette(QApplication::palette(tlbl));
        if (nrms < 10.0)
            tlnrms->setPalette(QApplication::palette(tlnrms));
        else
            tlnrms->setStyleSheet("QLabel { color : red; }");
    }


    sbput->setValue ((int) newDecoderConfig.channels[ch].puthr);
    sbspumd->setValue ((int) newDecoderConfig.channels[ch].pumindist);

    sbblr1->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbblr2->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbblr1->setValue (newDecoderConfig.channels[ch].blr1);
    sbblr2->setValue (newDecoderConfig.channels[ch].blr2);
    sbi1->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbi2->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbi1->setValue (newDecoderConfig.channels[ch].int1);
    sbi2->setValue (newDecoderConfig.channels[ch].int2);

    const auto &fir = newDecoderConfig.channels[ch].deconv.fir;
//    if (fir.offs > fir.coef.size())
//        fir.offs = fir.coef.size() / 2;
    sbfirnp->setValue (fir.coef.size());

    for (int i=0; i<comboBoxWindowingFunction->count(); i++) {
        if (comboBoxWindowingFunction->itemData(i) == win_type())
        comboBoxWindowingFunction->setCurrentIndex(i);
    }

    disable_control_slots = 0;

    on_pushButtonGenerateFir_clicked ();
    Redraw ();
    restoring = 0;
    //  printf("Restore... Ok\n");
}

void DeconvConfigDialog::Redraw ()
{
    //printf("Redraw\n");

    replotINP ();
    replotINPFFT ();
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

void DeconvConfigDialog::on_pushButtonClearFir_clicked()
{
    newDecoderConfig.channels[ch].deconv.fir.bits = 1;
    newDecoderConfig.channels[ch].deconv.fir.offs = 0;
    newDecoderConfig.channels[ch].deconv.fir.coef.clear();
//    newDecoderConfig.channels[ch].deconv.fir.coef.push_back(1.0);
    replotFIR();
    replotFIRFFT();
    replotTEST();
}

void DeconvConfigDialog::on_pushButtonGenerateFir_clicked ()
{
    if (restoring)
        return;
    //printf("CreateFIR\n");

    if (!newDecoderConfig.calavg.channels.contains(ch)
        || newDecoderConfig.calavg.channels[ch].isEmpty()) {
        //    log_notice ("ch %d, No average waveform, please calibrate average first", ch);
        return;
    }
    const std::vector<double> &calAvg = getCalAvgY(ch).toStdVector();
    const size_t n = calAvg.size();
    std::vector<double> req = getRefY_n(ch, getCalAvgAmplitude(ch), n);
    if (req.size() > calAvg.size())
        req.resize(calAvg.size());
    Q_ASSERT(calAvg.size() == req.size());

    // double orig_thr = db2val (newDecoderConfig.channels[ch].deconv.snf);
    // double req_thr = db2val (newDecoderConfig.channels[ch].deconv.pnf);

    FirConfig df;

    create_deconvolution_dblFIR(&df, calAvg, req);

    // FIR truncation. FIR will have fixed number of samples (np)
    const int olen = df.coef.size();

    int np = sbfirnp->value ();
    if (np > olen)
        np = olen;
    int i1 = (olen - np) / 2;
    int i2 = (olen + np) / 2;
    if (i1 < 0)
        i1 = 0;
    if (i2 > olen)
        i2 = olen;

    for (int i = i1; i < i2; i++) {
        df.coef[i - i1] = df.coef[i];
    }

    df.coef.resize(i2 - i1); // np
    df.offs -= i1;

    // end of FIR truncation

    // Windowing
    for (unsigned int i = 0; i < df.coef.size(); i++) {
        df.coef[i] *= windowingFunction (win_type(), i, df.coef.size() - 1);
    }

    // After windowing, first and last samples may be near zero
    // TODO

    // Copy new fir to cfg
//    intfir_t newFIR;
//    newFIR.coef.resize(df.klen);
//    newFIR.offs = df.offs;
//    newFIR.bits = FIR_COEFF_BITS;
//    Make_Discrete (&df.coef[0], &newFIR.coef[0], df.size(), FIR_COEFF_BITS);
    newDecoderConfig.channels[ch].deconv.fir = df; // newFIR;

    /*
  if (!restoring)
    Restore ();
*/
    //  Restore(); -- is called from Restore() --
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

double DeconvConfigDialog::reqf (int i) const
{
    double ampl = 1; // 0.5 * BITS2MAXUINT (adcConfig.bits);
    return ampl * reqfunscaled(i);
}

double DeconvConfigDialog::reqfunscaled (int i) const
{
    // i in timebins, not ns!

    const int nl=10;

    double step = getTimeStepInNs();
    if (step <= 1e-12)
        step = 1;
    double pos = newDecoderConfig.channels[ch].deconv.ppos / step;
    double wid = newDecoderConfig.channels[ch].deconv.pwid / step;
    double value=0;
    for(int l=0;l<nl;l++)
        value += 1. * gauss (i+(1.*l/nl - .5) - pos, wid / GausFWHM);
    value /= nl;
    return value;
}

//double DeconvConfigDialog::ca (unsigned int ch, int i) const
//{
//    int na = newDecoderConfig.calavg.channels[ch].nent;
//    double z = newDecoderConfig.calzero.channels[ch].get_base();
//    if (na <= 0)
//        na = 1;
////    if (nz <= 0)
////        nz = 1;

//    //  return (1. * (1. * cfg->dec.calavg.data[ch][i] / na));
//    double expfall_param = newDecoderConfig.channels[ch].deconv.expfall;
//    double expfall = exp(-10.*i/expfall_param)/exp(-10.*20/expfall_param);
//    double value = newDecoderConfig.calavg.channels[ch].data[i] / na;
//    value -= z;
//    return (1. * value * expfall);

//    //  return (1. *(1.*cfg->dec.calavg.data[ch][i] / na -
//    //  1.*cfg->dec.calzero.base[ch] / nz )*expfall/* * 10.*exp(-.1*i)*(1+erf(.3*(i-5))*/);

//}

double DeconvConfigDialog::getTimeStepInNs() const
{
    if (newDecoderConfig.calavg.channels.contains(ch))
        return newDecoderConfig.calavg.channels[ch].step * 1e9;
    return adcConfig.step;
}

//--------- Slot actions -----------------------------------------------------------------

void DeconvConfigDialog::on_comboBoxWindowingFunction_activated (int item)
{
    fwinf_t new_win_type = static_cast<fwinf_t>(comboBoxWindowingFunction->itemData(item).toInt());
    if (new_win_type == win_type()) return;
    newDecoderConfig.channels[ch].deconv.fwinf = new_win_type;
    on_pushButtonGenerateFir_clicked ();
}

void DeconvConfigDialog::on_sbfirnp_valueChanged (int)
{
    //  if (restoring) return;
    on_pushButtonGenerateFir_clicked ();
}


void DeconvConfigDialog::on_sbsnf_valueChanged (double v)
{
    if (restoring)
        return;
    newDecoderConfig.channels[ch].deconv.snf = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}

void DeconvConfigDialog::on_sbrnf_valueChanged (double v)
{
    if (restoring)
        return;
    newDecoderConfig.channels[ch].deconv.pnf = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}



void DeconvConfigDialog::on_sbput_valueChanged (int v)
{
    newDecoderConfig.channels[ch].puthr = v;
    setButtons();
}

void DeconvConfigDialog::on_sbspumd_valueChanged (int v)
{
    newDecoderConfig.channels[ch].pumindist = v;
    setButtons();
}

void DeconvConfigDialog::on_sbi1_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi1->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.channels[ch].int1 = v;
    if (sbi2->value () < v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi2->setValue (v);
        newDecoderConfig.channels[ch].int2 = v;
        disable_control_slots = tmp;
    }
    setButtons();
}

void DeconvConfigDialog::on_sbi2_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi2->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.channels[ch].int2 = v;
    if (sbi1->value () > v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi1->setValue (v);
        newDecoderConfig.channels[ch].int1 = v;
        disable_control_slots = tmp;
    }
    setButtons();
}


void DeconvConfigDialog::on_sbblr1_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr1->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.channels[ch].blr1 = v;
    if (sbblr2->value () < v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr2->setValue (v);
        newDecoderConfig.channels[ch].blr2 = v;
        disable_control_slots = tmp;
    }
    setButtons();
}

void DeconvConfigDialog::on_sbblr2_valueChanged (int v)
{
    if (disable_control_slots)
        return;

    int v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        int tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr2->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.channels[ch].blr2 = v;
    if (sbblr1->value () > v) {
        int tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr1->setValue (v);
        newDecoderConfig.channels[ch].blr1 = v;
        disable_control_slots = tmp;
    }
    setButtons();
}



void DeconvConfigDialog::on_sbpos_valueChanged (double v)
{
    newDecoderConfig.channels[ch].deconv.ppos = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void DeconvConfigDialog::on_sbwid_valueChanged (double v)
{
    newDecoderConfig.channels[ch].deconv.pwid = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void DeconvConfigDialog::on_spinBoxChannel_valueChanged (int inch)
{
    if (inch >= int(currentDecoderConfig.nch())) {
        spinBoxChannel->setValue(newDecoderConfig.nch());
        return;
    }
    ch = inch;
    Restore ();
}

void DeconvConfigDialog::on_sbexp_valueChanged (double v)
{
    newDecoderConfig.channels[ch].deconv.expfall = v;

    on_pushButtonGenerateFir_clicked ();
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    // replotTEST ();
}


void DeconvConfigDialog::on_sbblrmd_valueChanged (int v)
{
    newDecoderConfig.channels[ch].baserc = 0.01 * v;
}


int DeconvConfigDialog::data_latency () const
{
    return adcConfig.latency;
}

template<class T>
QStringList vectorToStringList(const QVector<T> &values)
{
    QStringList sl;
    foreach (int value, values) {
        sl += QString("%1").arg(value);
    }
    return sl;
}

//--------------- Button actions --------------------------------------------------------

void DeconvConfigDialog::on_pushButtonSaveFir_clicked()
{
    QSettings settings;
    QString fileName = settings.value("currentFirFileName").toString();
    fileName = QFileDialog::getSaveFileName (this, "Saving FIR file", fileName,
                                                       "FIR files (*.fir);;All files (*)");
    if (fileName.isNull())
        return;
    if (!fileName.endsWith(".fir"))
        fileName += ".fir";
    QSettings s(fileName, QSettings::IniFormat);
    BaseConfigIni<FirConfig>::staticSave(newDecoderConfig.channels[ch].deconv.fir, &s);
//    if (ret)
//        QMessageBox::warning (this, "Saving FIR filter", "The file cannot be written", QMessageBox::Cancel, Qt::NoButton);

    settings.setValue("currentFirFileName", fileName);
}

void DeconvConfigDialog::on_pushButtonLoadFir_clicked ()
{
    QString title("Load FIR filter");
    QSettings settings;
    QString fileName = settings.value("currentFirFileName").toString();
    fileName = QFileDialog::getOpenFileName (this, title, fileName,
                                                       "FIR files (*.fir);;All files (*)");

    if (fileName.isNull())
        return;
    settings.setValue("currentFirFileName", fileName);

    QSettings s(fileName, QSettings::IniFormat);
    const auto &tmp = BaseConfigIni<FirConfig>::staticLoad(s);
    if (!tmp.coef.empty()) {
        newDecoderConfig.channels[ch].deconv.fir = tmp;
        Restore ();
        QString str;
        QTextStream stream(&str);
        stream << tmp.coef;
        QString text= QString("Channel %1\n\nRead file\n'%2':\n\n%3")
                .arg(ch)
                .arg(fileName)
                .arg(str);
        QMessageBox::information(this, title, text, QMessageBox::Ok | QMessageBox::Default, Qt::NoButton);
    } else {
        QMessageBox::warning (this, title, "No FIR coefficients found in file", QMessageBox::Cancel, Qt::NoButton);
        return;
    }
}

void DeconvConfigDialog::on_pushButtonLoadShape_clicked ()
{
    QString title("Load average shape");
    QSettings settings;
    QString fileName = settings.value("currentShapeFileName").toString();
    fileName = QFileDialog::getOpenFileName (this, title, fileName,
                                                         "Shape files (*.shp);;All files (*)");

    if (fileName.isNull())
        return;
    if (!fileName.endsWith(".shp"))
        fileName += ".shp";
    settings.setValue("currentShapeFileName", fileName);

    QSettings s(fileName, QSettings::IniFormat);
    const auto &tmp = BaseConfigIni<CalAvgChannelConfig>::staticLoad(s);
    if (!tmp.isEmpty()) {
        newDecoderConfig.calavg.channels[ch] = tmp;
        Restore ();
        QString text= QString("Channel %1\n\nRead file\n'%2':\n\n%3")
                .arg(ch)
                .arg(fileName)
                .arg(vectorToStringList(tmp.get()).join(", "));
        QMessageBox::information(this, title, text, QMessageBox::Ok | QMessageBox::Default, Qt::NoButton);
    } else {
        QMessageBox::warning (this, title, "No shape calibration waveform found in file", QMessageBox::Cancel, Qt::NoButton);
        return;
    }
}

void DeconvConfigDialog::on_pushButtonSaveShape_clicked ()
{
    QSettings settings;
    QString currentShapeFileName = settings.value("currentShapeFileName").toString();
    currentShapeFileName = QFileDialog::getSaveFileName (this, "Saving average shape", currentShapeFileName,
                                                         "Shape files (*.shp);;All files (*)");

    if (currentShapeFileName.isNull())
        return;

    if (!currentShapeFileName.endsWith(".shp"))
        currentShapeFileName += ".shp";

    // TODO: write current channel (ch)
    QSettings s(currentShapeFileName, QSettings::IniFormat);
    BaseConfigIni<CalAvgChannelConfig>::staticSave(newDecoderConfig.calavg.channels[ch], &s);
    settings.setValue("currentShapeFileName", currentShapeFileName);
}

bool DeconvConfigDialog::modified() const
{
    return currentDecoderConfig != newDecoderConfig;
}

void DeconvConfigDialog::apply()
{
    emit decoderConfigChanged(newDecoderConfig);
    currentDecoderConfig = newDecoderConfig;
    setButtons();
}

void DeconvConfigDialog::setButtons ()
{
    buttonBox->button(QDialogButtonBox::Apply)->setEnabled(modified());
    buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(newDecoderConfig != DecoderConfig());
}

void DeconvConfigDialog::restoreDefaults()
{
    newDecoderConfig = AllConfig().dec;
    setButtons();
    Restore();
}

void DeconvConfigDialog::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Ok :
    case QDialogButtonBox::Apply :
        apply();
        break;
    case QDialogButtonBox::RestoreDefaults :
        restoreDefaults();
        break;
    default:;
    }
}
