//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#define _USE_MATH_DEFINES
#include <cmath>
#include "dsplib.h"
#include "WindowFunction.h"

WindowFunction::WindowFunction()
{
}

// Windowing functions
// i = 0, 1, ..., m

double wf_barthann (double i, int m)
{
    double arg = i / m - .5;
    double parg = arg;
    if (parg < 0)
        parg = -parg;
    return (0.62 - 0.48 * parg + 0.38 * cos (2 * M_PI * arg));
}

double wf_barlett (double i, int m)
{
    if (m <= 0)
        return (1);
    if ((i < 0) || (i > m))
        return (0);
    if (i < .5 * m)
        return (2. * i / m);
    else
        return (2. - 2. * i / m);
}

double wf_blackman (double i, int m)
{
    return (0.42 - 0.5 * cos (2 * M_PI * i / m) + 0.08 * cos (4 * M_PI * i / m));
}

double wf_blackmanharris (double i, int m)
{
    // returns an (m+1)-point, minimum , 4-term Blackman-harris window.
    // The window is minimum in the sense that its maximum sidelobes are minimized.
    double arg = 2 * M_PI * i / m;
    return (0.35875 - 0.48829 * cos (arg) + 0.14128 * cos (2 * arg) - 0.01168 * cos (3 * arg));
}

double wf_bohman (double i, int m)
{
    double arg = 2 * i / m - 1;
    double parg = arg;

    if ((i < 0) || (i > m))
        return (0);
    if (parg < 0)
        parg = -parg;
    return ((1 - parg) * cos (M_PI * parg) + 1 / M_PI * sin (M_PI * parg));
}

double wf_flattop (double i, int m)
{
    double arg = 2 * M_PI * i / m;
    return (1 - 1.93 * cos (arg) + 1.29 * cos (2 * arg) - 0.388 * cos (3 * arg) + 0.0322 * cos (4 * arg));
}

double wf_gauss (double i, int m, double r)
{
    double x = i / m - .5;
    return (gauss (x, r));
}
double wf_gauss_075 (double i, int m)
{
    return (wf_gauss (i, m, .75));
}
double wf_gauss_050 (double i, int m)
{
    return (wf_gauss (i, m, .50));
}
double wf_gauss_025 (double i, int m)
{
    return (wf_gauss (i, m, .25));
}

double wf_hamming (double i, int m)
{
    return (0.53836 - 0.46164 * cos (2 * M_PI * i / m));
}

double wf_hann (double i, int m)
{
    return (0.5 * (1 - cos (2 * M_PI * i / m)));
}

double wf_nuttall (double i, int m)
{
    // Similar to BlackmanHarris, but produce slightly lower sidelobes.
    double arg = 2 * M_PI * i / m;
    return (0.3635819 - 0.4891775 * cos (arg) + 0.1365995 * cos (2 * arg) - 0.0106411 * cos (3 * arg));
}

double wf_rect (double, int)
{
    return (1);
}

double wf_triang (double i, int m)
{
    return (wf_barlett (i + 1, m + 2));
}

double wf_tukey (double i, int m, double r)
{
    double j = i - .5 * m;
    double pj = j;

    if (r <= 0)
        return (1);
    if (r >= 1)
        r = 1;

    if (pj < 0)
        pj = 0. - pj;
    if (pj < (r * .5 * m))
        return (1);
    return (cos (M_PI * (pj - r * .5 * m) / ((1 - r) * m)));
}
double wf_tukey_075 (double i, int m)
{
    return (wf_tukey (i, m, .75));
}
double wf_tukey_050 (double i, int m)
{
    return (wf_tukey (i, m, .50));
}
double wf_tukey_025 (double i, int m)
{
    return (wf_tukey (i, m, .25));
}

double windowingFunction (fwinf_t win_type, double i, int m)
{
    if (i<0) i=0;
    if (i>m) i=m;
    switch (win_type) {
    case WF_BARTHANN: return wf_barthann(i, m);
    case WF_BARLETT: return wf_barlett(i, m);
    case WF_BLACKMAN: return wf_blackman(i, m);
    case WF_BLACKMANHARRIS: return wf_blackmanharris(i, m);
    case WF_BOHMAN: return wf_bohman(i, m);
    case WF_FLATTOP: return wf_flattop(i, m);
    case WF_GAUSS_075: return wf_gauss(i, m, 0.75);
    case WF_GAUSS_050: return wf_gauss(i, m, 0.50);
    case WF_GAUSS_025: return wf_gauss(i, m, 0.25);
    case WF_HAMMING: return wf_hamming(i, m);
    case WF_HANN: return wf_hann(i, m);
    case WF_NUTTALL: return wf_nuttall(i, m);
    case WF_RECT: return 1;
    case WF_TRIANG: return wf_triang(i, m);
    case WF_TUKEY_075: return wf_tukey(i, m, 0.75);
    case WF_TUKEY_050: return wf_tukey(i, m, 0.50);
    case WF_TUKEY_025: return wf_tukey(i, m, 0.25);
    default: return 1;
    }
}
