
#define _USE_MATH_DEFINES
#include <cmath>
#include <cstring> // FIXME: memset
#include <cassert>
#include <iostream>
#include <algorithm>

// Compile this source with gcc -fast-math option

#include <complex>
#ifdef USE_FFTW
#include <fftw3.h> // NOLINT
#endif
#include <vector>

#include "dsplib.h"

inline double sqr(double x) { return pow(x, 2); }

inline double hypot2(double x, double y) { return x*x+y*y; }

const long double pi = M_PI;
//const long double pi = acosl((long double) -1);

/*
#ifdef M_PI
# define pi    M_PI
#else
# define pi    (double)3.14159265358979312e+00
#endif
*/
//#define e     (double)2.71828182845904509e+00
//#define pi    (double)3.14159265358979312e+00
//#define twopi (double)6.28318530717958623e+00

//void LinearFit(double x[], double y[], int n, double *a, double *b)
/// Linear Least Squares fit
void LinearFit (const double *x, const double *y, int n, double *a, double *b, double *c)
{
    // Find a,b for array of (x,y) of length n
    // y = ax + b

    int i;
    double xa = 0, ya = 0, stt = 0, xy = 0, aa, bb, chi2 = 0;

    if (n < 1)
        return;

    for (i = 0; i < n; i++) {
        xa += x[i];
        ya += y[i];
        //    printf("%d: %f, %f\n", i, x[i], y[i]);
    }
    xa /= n;
    ya /= n;
    for (i = 0; i < n; i++)
        stt += sqr (x[i] - xa);
    for (i = 0; i < n; i++)
        xy += (x[i] - xa) * y[i];
    if (stt != 0)
        bb = xy / stt;
    else
        bb = 0;
    aa = ya - xa * bb;
    for (i = 0; i < n; i++)
        chi2 += sqr (y[i] - aa - bb * x[i]);
    //  printf("Line fit: a=%f, b=%f, chi2=%f\n", aa, bb, chi2);
    if (a)
        *a = bb;
    if (b)
        *b = aa;
    if (c)
        *c = chi2;
}

void LinearFitInt (const int *y, int n, double *a, double *b, double *c)
{
    // Find a,b for array of (x,y) of length n
    // y = ax + b

    int i;
    double ya = 0, stt = 0, xy = 0, aa, bb, chi2 = 0;

    if (n < 1)
        return;

    for (i = 0; i < n; i++) {
        ya += y[i];
    }
    ya /= n;
    for (i = 0; i < n; i++)
        stt += sqr (i - 1);
    for (i = 0; i < n; i++)
        xy += (i - 1) * y[i];
    if (stt != 0)
        bb = xy / stt;
    else
        bb = 0;
    aa = ya - bb;
    for (i = 0; i < n; i++)
        chi2 += sqr (y[i] - aa - bb * i);
    printf ("Line fit int: a=%f, b=%f, chi2=%f\n", aa, bb, chi2);
    if (a)
        *a = bb;
    if (b)
        *b = aa;
    if (c)
        *c = chi2;
}

void deriv2 (const int *x, int len, int *out)
{
    assert(x);
    assert(out);
    assert(len>=2);
    for (int i = 1; i < len - 1; i++)
        out[i] = 2 * x[i] - x[i - 1] - x[i + 1];
}

/// @brief An unnormalized gaussian function with mean=0 and @a sigma
///
/// Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed
double gauss (double x, double sigma)
{
    double arg, res;
    // Calculate a gaussian function with mean and sigma.
    if (sigma == 0) {
        if (x==0) return 1;
        else return 0;
    }
    arg = x / sigma;
    res = exp (-0.5 * arg * arg);
    return res;
    //   return res/(2.50662827463100024*sigma); //sqrt(2*Pi)=2.50662827463100024
}

double fft_noisefloor (double rmsnoise, int rank)
{
    return (20. * log (rmsnoise / sqrt (1.*rank)) / log (10.));
}

/// Rectangular to halfcomplex DFT format
std::vector<double> rect2hc(const std::vector<double> &S, const std::vector<double> &C)
{
    const size_t N = S.size() + C.size() - 1;
    std::vector<double> x(N);
    for(size_t i=0; i<N/2+1; i++) x[i] = C[i]; // Re
    for(size_t i=1; i<(N+1)/2; i++) x[N-i] = S[i]; // Im
    return x;
}

/// Halfcomplex to rectangular DFT format
void hc2rect(const std::vector<double> &hc, std::vector<double> *S, std::vector<double> *C)
{
    const size_t N = hc.size();
    for(size_t i=0; i<N/2+1; i++) (*C)[i] = hc[i]; // Re
    for(size_t i=1; i<(N+1)/2; i++) (*S)[i] = hc[N-i]; // Im
}


// Discrete Fourier Transform, order N, Real form
// symmetry: X[k] = X*[N-k]
void BruteFT::DFT (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C)
{
    assert(S);
    assert(C);
    const size_t n = x.size();
    const size_t out_size = n/2+1;
    S->resize(out_size);
    C->resize(out_size);
    for (size_t k = 0; k < out_size; k++) {
        (*S)[k] = 0;
        (*C)[k] = 0;
        long double s=0, c=0;
        for (size_t i = 0; i < n; i++) {
            const double arg = 2 * pi * k * i / n;
            const double is = -sin(arg);
            const double ic = cos(arg);
            s += is * x[i];
            c += ic * x[i];
        }
        (*S)[k] = s;
        (*C)[k] = c;
        //        (*S)[k] /= sqrt (n / 2);
        //        (*C)[k] /= sqrt (n / 2);
    }
    //  if (n%2 == 0) /* N is even */ C[n/2] *= .5; /* Nyquist freq. */
}

// Inverse Discrete Fourier Transform, order n, Real form
std::vector<double> BruteFT::IDFT (const std::vector<double> &S, const std::vector<double> &C)
{
    // BUG: the result is not consistent with fftw
    //    const size_t N = (C.size()-1)*2;
    //    std::vector<double> out(N);
    //    fftw_plan p = fftw_plan_r2r_1d(N, &x[0], &out[0], FFTW_HC2R, FFTW_ESTIMATE);
    //    Q_ASSERT(p);
    //    fftw_execute(p);
    if (S.size() < 1 || C.size() < 1)
        return std::vector<double> ();
    assert(S.size() <= C.size());
    const size_t n = C.size() + S.size() - 1;
    std::cerr << "IDFT: n = " << n << std::endl;
    std::vector<double> out(n);

    for (size_t k = 0; k < n; k++) {
        long double x = 0;
        for(size_t i=1; i<S.size(); i++) {
            const long double a = (long double)(2. * i * k) / n;
            const long double arg = pi*a;
            x += sinl(arg)*S[i];
        }
        for(size_t i=0; i<C.size(); i++) {
            const long double a = (long double)(2. * i * k) / n;
            const long double arg = pi*a;
            long double m = 2;
            if ((i==0) || i+1==C.size()) m=1;
            //            if (i==0) m=1;
            //            if (((i+1)==C.size())) m=2;
            x += cosl(arg)*C[i]*m;
        }
        /*
        for (size_t i = 0; i < (n / 2 + 1); i++) {
            c = C[i];
//            if ((i == 0) || (i == n / 2))
//                c *= 0.5;
            arg = 2 * pi * i * k / n;
            is = sin(arg);
            ic = cos(arg);
            x += (is * S[i] + ic * c);
            //      x[k] += (sin(2*pi*i*k/n)*S[i] + cos(2*pi*i*k/n)*c);
        }
*/
        //        x[k] /= sqrt (n / 2);
        out[k] = x;
    }
    return out;
}

// Calculate power from rectangular FFT
std::vector<double> BruteFT::RECT2POW (const std::vector<double> &S, const std::vector<double> &C)
{
    std::vector<double> out;
    RECT2POW(S, C, &out);
    return out;
}

void BruteFT::RECT2POW (const std::vector<double> &S, const std::vector<double> &C, std::vector<double> *out)
{
    const size_t n = S.size();
    assert(n == C.size());
    if (out->size() != n)
        out->resize(n);
    for (size_t i = 0; i < n; i++)
        (*out)[i] = hypot (S[i], C[i]);
}

// Calculate phase from rectangular FFT
std::vector<double> BruteFT::RECT2PHA (const std::vector<double> &S, const std::vector<double> &C)
{
    assert(S.size() == C.size());
    std::vector<double> out(S.size());
    for (size_t i = 0; i < S.size(); i++)
        out[i] = atan2 (S[i], C[i]);
    return out;
}

/// @brief Complex Divide operation in rectangular form:
/// (e+if)=(a+ib)/(c+id)
void Compdiv (double a, double b, double c, double d, double *e, double *f)
{
    double mod2;
    if (!e || !f)
        return;
    //  mod2 = c*c+d*d;
    mod2 = hypot2 (c, d);
    if (mod2 == 0) {
        *e = 1.;
        *f = 1.;
        return;
    }
    *e = (a * c + b * d) / mod2;
    *f = (b * c - a * d) / mod2;
}

double detpulse (double a, double t, double rt, double ft)
{
    // Simulate detector pulse
    //

    double v, r, f, e;

    //  if (rt > ft) m = ft; else m = rt;

    //  e = TMath::Log(ADC_RES*10); // set error to 0.1 resolution
    e = log (1e6);

    r = t / rt;
    f = t / ft;
    if (f > e)
        return (0);
    if (r < -4)
        return (0);

#ifndef _MSC_VER
    v = a * (erf (r) + 1) / 2 * exp (-f);
#else
    if (f>0)
        v = a * exp (-f);
    else
        v = 0;
#endif
    // if (v<0) v=0;
    return v;
}

/// @brief Normalize to integral=1
void Normalize (std::vector<double> *x)
{
    double sum;
    sum = 0;
    for (size_t i = 0; i < x->size(); i++)
        sum += (*x)[i];
    if (sum != 0)
        for (size_t i = 0; i < x->size(); i++)
            (*x)[i] /= sum;
}

/// @brief Scale to fit in <bits>-bit integer
std::vector<int> Make_Discrete (const std::vector<double> &x, int bits)
{
    const size_t n = x.size();
    double scalepos, scaleneg, scale;

    if (bits <= 1)
        return std::vector<int> ();
    int imin = BITS2MININT (bits); /* - (1<<(bits-1)) */
    int imax = BITS2MAXINT (bits); /* ((1<<(bits-1)) - 1) */
    double min = 0;
    double max = 0;

    for (size_t i = 0; i < n; i++) {
        if (x[i] > max)
            max = x[i];
        if (x[i] < min)
            min = x[i];
    }
    if (max > 0)
        scalepos = imax / max;
    else
        scalepos = 1e38;
    if (min < 0)
        scaleneg = imin / min;
    else
        scaleneg = 1e38;
    if (scalepos < scaleneg)
        scale = scalepos;
    else
        scale = scaleneg;
    // printf("Make_Discrete(): min=%f, max=%f, scaling by %f (neg %f, pos %f)\n", min, max, scale, scaleneg, scalepos);
    std::vector<int> out(n);
    if (scale != 0)
        for (size_t i = 0; i < n; i++)
            out[i] = (int) (x[i] * scale);
    return out;
}

int intsum (const std::vector<int> &x)
{
    int s = 0;

    for (size_t i = 0; i < x.size(); i++)
        s += x[i];
    return s;
}

/// @brief Max element of vector of doubles
double maxdbl (const std::vector<double> &x)
{
    return *std::max_element(x.begin(), x.end());
}

/// @brief Min element of vector of doubles
double mindbl (const std::vector<double> &x)
{
    return *std::min_element(x.begin(), x.end());
}

/// @brief DC value
double dcdbl (const std::vector<double> &x)
{
    if (x.empty())
        return 0;
    double v = 0;

    for (size_t i = 0; i < x.size(); i++)
        v += x[i];
    v /= x.size();
    return v;
}

/// @brief RMS value
double rmsdbl (const std::vector<double> &x, double dc)
{
    if (x.empty())
        return 0;
    double v = 0;
    for (size_t i = 0; i < x.size(); i++)
        v += sqr (x[i] - dc);
    v /= x.size();
    return v;
}

/// @brief Decibels to amplitude ratio conversion
double db2val (double db)
{
    double value;

    value = exp (db / 20. * log (10.));
    return (value);
}

/// @brief Amplitude ratio to decibels conversion
double val2db (double val)
{
    double db;

    db = 20. * (log(val) / log (10.));
    return (db);
}

double rndm ()
{
    // System buil-in
    // Random number in range [0,1) with period approx. 16*((2**31)-1)
    double value;
    value = (double) rand () / RAND_MAX;
    return (value);
}

unsigned int f1Seed;

///  @brief Machine independent random number generator
///
///  Produces uniformly-distributed floating points between 0 and 1.
///  Identical sequence on all machines of >= 32 bits.
///  Periodicity = 10**8
///  Universal version (Fred James 1985).
///  generates a number in ]0,1]
double rndm1 ()
{
    const double kCONS = 4.6566128730774E-10;
    const int kMASK24 = 2147483392;
    unsigned int jy;

    f1Seed *= 69069;
    jy = (f1Seed & kMASK24);      // Set lower 8 bits to zero to assure exact float
    if (jy)
        return kCONS * jy;
    return rndm ();
}



int fSeed1 = 9876;
int fSeed2 = 54321;

///  @brief Machine independent random number generator
///
///  Produces uniformly-distributed floating points in [0,1]
///  Identical sequence on all machines of >= 32 bits.
///  Periodicity > 10**14
double rndm2 ()
{
    double iz, r;
    int k = (int) (fSeed1 / 53668);
    fSeed1 = 40014 * (fSeed1 - k * 53668) - k * 12211;
    if (fSeed1 < 0)
        fSeed1 += 2147483563;
    k = (int) (fSeed2 / 52774);
    fSeed2 = 40692 * (fSeed2 - k * 52774) - k * 3791;
    if (fSeed2 < 0)
        fSeed2 += 2147483399;
    iz = fSeed1 - fSeed2;
    if (iz <= 0)
        iz += 2147483562;
    r = iz * 4.6566128e-10;
    return r;
}

/// @brief An uniform deviate on the interval ]x1, x2].
double ranuni (double x1, double x2)
{
    double ans = rndm ();
    return x1 + (x2 - x1) * ans;
}

/// @brief A number distributed following a gaussian with @a sigma
double rangaus (double sigma)
{
    double x, y, z, result;

    y = rndm ();
    z = rndm ();
    x = z * 6.28318530717958623;
    result = sigma * sin (x) * sqrt (-2 * log (y));
    return result;
}


// Compatibility
void BruteFT::DFT_tmp (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C)
{
    assert(S);
    assert(C);
    const size_t n = x.size();
    const size_t out_size = n/2+1;
    S->resize(out_size);
    C->resize(out_size);
    for (size_t k = 0; k < out_size; k++) {
        (*S)[k] = 0;
        (*C)[k] = 0;
        for (size_t i = 0; i < n; i++) {
            double arg = 2 * pi * k * i / n;
            double is = sin(arg);
            double ic = cos(arg);
            (*S)[k] += is * x[i];
            (*C)[k] += ic * x[i];
        }
        (*S)[k] /= sqrt (n / 2 + .5);
        (*C)[k] /= sqrt (n / 2 + .5);
    }
    //  if (n%2 == 0) /* N is even */ C[n/2] *= .5; /* Nyquist freq. */
}

void BruteFT::IDFT (std::vector<double> *x, const std::vector<double> &S, const std::vector<double> &C)
{

    // Inverse Discrete Fourier Transform, order n
    // Real form

    size_t n = (std::min(S.size(), C.size()) - 1) * 2;
    if (x->size() != n)
        x->resize(n);

    for (size_t k = 0; k < n; k++) {
        (*x)[k] = 0;
        for (size_t i = 0; i < (n / 2 + 1); i++) {
            double c = C[i];
            if ((i == 0) || (i == n / 2))
                c *= 0.5;
            double arg = 2 * pi * i * k / n;
            double is = sin(arg);
            double ic = cos(arg);
            (*x)[k] += (is * S[i] + ic * c);
            //      x[k] += (sin(2*pi*i*k/n)*S[i] + cos(2*pi*i*k/n)*c);
        }
        (*x)[k] /= sqrt (n / 2 + .5);
    }
}

//void RECT2PHA (const double *S, const double *C, int n, double *phase)
//{

//  // Calculate phase from rectangular FFT

//  int i;
//  for (i = 0; i < n; i++)
//    phase[i] = atan2 (S[i], C[i]);
//}

void Make_Discrete (const double *x, int *out, int n, int bits)
{

    // Scale to fit in <bits>-bit integer

    int i, imin, imax;
    double min, max, scalepos, scaleneg, scale;

    if (bits <= 1)
        return;
    imin = BITS2MININT (bits) /* - (1<<(bits-1)) */ ;
    imax = BITS2MAXINT (bits) /* ((1<<(bits-1)) - 1) */ ;
    min = 0;
    max = 0;

    for (i = 0; i < n; i++) {
        if (x[i] > max)
            max = x[i];
        if (x[i] < min)
            min = x[i];
    }
    if (max > 0)
        scalepos = imax / max;
    else
        scalepos = 1e38;
    if (min < 0)
        scaleneg = imin / min;
    else
        scaleneg = 1e38;
    if (scalepos < scaleneg)
        scale = scalepos;
    else
        scale = scaleneg;
    // printf("Make_Discrete(): min=%f, max=%f, scaling by %f (neg %f, pos %f)\n", min, max, scale, scaleneg, scalepos);
    if (scale != 0)
        for (i = 0; i < n; i++)
            out[i] = (int) (x[i] * scale);
}

void convolve_i_s_i (const int *x1, const int l1, const short int *x2, const int l2, int *out)
{
    if (l1==0 || l2==0) return;

    // Important! Zero buffer. Buffer is durty on enter.
    // memset is fastest
    memset(out, 0, sizeof(int)*(l1 + l2 - 1));
    //  int *optr = &out[0];
    //  for (i = 0; i < l1 + l2 - 1; i++) *optr++ = 0;

    /*
// This is trivial implementation. Demo.

  {
  int i, j;
  for (i = 0; i < l1; i++)
    for (j = 0; j < l2; j++)
    {
      out[i + j] += x1[i] * x2[j];
    }
  }
*/
    /*
  memset(test, 0, sizeof(int)*(l1+l2-1));
// Good code for continuous filtering with head overlap, assuming missing samples = 0
  {
  const int *coef = x1;
  int nk = l1;
  const short int *data = x2;
  int nd = l2;
  int k, j;
  assert(nk<=nd);

// Compute head overlap
  for (k = 0; k < nk-1; k++) {
  {
    int tmp=0;
    for(j=0;j<=k;j++) {
      tmp += coef[k-j] * data[j];
    }
    test[k] = tmp;
  }
  }
// Compute main part
  for (j = 0; j < nd; j++) {
    int tmp = 0;
    for (k = 0; k < nk; k++)
    {
      tmp += coef[nk-1-k] * data[j+k];
    }
    test[nk-1+j] = tmp;

  }

  }
*/

    // Welcome to real life. This algo is two times faster.
    // Avoid index operations. Use pointers.
    {
        int i;
        for (i = 0; i < l1; i++)
        {
            int *optr = &out[i];
            const int x1i = x1[i];
            const short int *x2ptr = x2;
            int j;
            for (j = 0; j < l2; j++)
            {
                *optr++ += x1i * *x2ptr++;
            }
        }
    }

    /*
// Another solution. Result is short, cannot use it.
  {
    int y_ptr, c_ptr, x_ptr;
    int n, k;
    register int temp;
    y_ptr = 0;
    for (n = 0; n < l2; n++) {
      c_ptr = 0;
      x_ptr = n;
      temp = 0;
      for (k = 0; k < l1; k++)
            temp += x1[l1-1-c_ptr++] * x2[x_ptr++];
      test[l1-1+y_ptr++] = temp;
    }
  }
*/
    /*
  {
    int i;
    printf("l1=%2d, l2=%2d, result %d\n", l1, l2, l1+l2);
    for (i = 0; i < l1 + l2; i++)
//    if (out[i] != test[i])
      printf("[%2d]: int %8d out %8d test %8d\n", i, x2[i], out[i], test[i]);
  }
*/
    /*
// code above optimized
  {
    int *y_ptr;
    const int *c_ptr;
    const short int *x_ptr;
    register int temp;
    int n, k;
    y_ptr = &out[0];
    for (n = 0; n < l2; n++) {
      c_ptr = &x1[0];
      x_ptr = &x2[n];
      temp = *c_ptr++ * *x_ptr++;
      for (k = 1; k < l1; k++)
            temp = temp + *c_ptr++ * *x_ptr++;
      *y_ptr++ = temp;
    }
  }
*/
}
