
#define _USE_MATH_DEFINES
#include <cmath>
#include <complex>
#include <cstdio>
#include <cstring>

#include "dsplib.h"
#include "fitter.h"
#include "lmmin.h"

#ifdef USE_FFT
#ifdef FIT_SINE
#include "rfftw.h"
#endif
#endif

const double M_1_SQRT2PI = 1./sqrt(2 * M_PI);

inline double sqr(double x) { return pow(x, 2); }

double FindGausDevNorm (const int *x, int len, double *par)
{
    double rms = 0;
    if (par[2]==0) {
        for (int i = 0; i < len; i++)
            rms += sqr(1.*x[i]);
        return sqrt(rms);
    }
    for (int i = 0; i < len; i++) {
        //	double expectval = par[0] * gauss (i - par[1], par[2]);
        double arg = (i - par[1]) / par[2];
        double expectval = par[0] * exp(-.5*arg*arg);
        rms += sqr (x[i] - expectval);// / sqr(expectval);
    }
    return sqrt(rms);
}

double FindGausDevNormD (const double *x, int len, double *par)
{
    int i;
    double rms = 0;
    for (i = 0; i < len; i++) {
        double expectval = par[0] * gauss (i - par[1], par[2]);
        rms += sqr (x[i] - expectval);// / sqr(expectval);
    }
    return sqrt(rms);
}

void FindGausApproxD (const double *x, int len, double *par)
{

    int i;
    double tt, val, mean, rms;
    double allcha = 0, sumx = 0, sumx2 = 0;

    if (!x || !par)
        return;

    par[0] = 0;
    par[1] = 0;
    par[2] = 0;

    for (i = 0; i < len; i++) {
        tt = i + .5;
        val = x[i];
        if (val < 0)
            val = 0;                  // FIXME
        sumx += val * tt;
        sumx2 += val * tt * tt;
        allcha += val;
        //    printf("[%d]: x %g, sumx %g, sumx2 %g, allcha %g\n", i, val, sumx, sumx2, allcha);
    }
    if (allcha > 0) {
        mean = sumx / allcha;
        rms = sqrt (sumx2 / allcha - mean * mean);
        //    printf("mean: %g, rms %g\n", mean, rms);
        //   par[0] = 0.5*allcha/(sqrt(2*pi)*rms);
        par[0] = allcha / (sqrt (2 * M_PI) * rms);
        par[1] = mean - 0.5;
        par[2] = rms;
    }
}

void FindGausApprox (const int *x, int len, double *par)
{

    int i;
    double allcha = 0, sumx = 0, sumx2 = 0;

    if (!x || !par)
        return;

    for (i = 0; i < len; i++) {
        int val = x[i];
        int tt = i;
        if (val < 0)
            val = 0;                  // FIXME
        sumx += val * tt;
        sumx2 += val * tt * tt;
        allcha += val;
        //    printf("[%d]: x %g, sumx %g, sumx2 %g, allcha %g\n", i, val, sumx, sumx2, allcha);
    }
    {
        double mean, rms;
        if (allcha==0) {
            mean = len/2;
            rms = len/2;
        } else {
            mean = sumx / allcha;
            rms = sqrt (sumx2 / allcha - mean * mean);
        }
        //    printf("mean: %g, rms %g\n", mean, rms);
        //   par[0] = 0.5*allcha/(sqrt(2*pi)*rms);
        par[0] = allcha / (sqrt (2 * M_PI) * rms);
        par[1] = mean;
        par[2] = rms;
    }
}


typedef struct {
    const double *y;
    double fp[3];
    int fix;
} fitdata_gaus_t;

void lm_evaluate_gaus (double *par, int m_dat, double *fvec, double *fjac, void *data, int *info)
/*
 *	par is an input array. At the end of the minimization, it contains
 *        the approximate solution vector.
 *
 *	m_dat is a positive integer input variable set to the number
 *	  of functions.
 *
 *	fvec is an output array of length m_dat which contains the function
 *        values the square sum of which ought to be minimized.
 *
 *	data is a read-only pointer to fitdata_gaus_t
 *
 *      info is an integer output variable. If set to a negative value, the
 *        minimization procedure will stop.
 */
{
    fitdata_gaus_t *mydata = (fitdata_gaus_t *) data;
    double m1, arg, ex, fac, p0, p1, p2, p0ex;

    if (mydata->fix) {
        int nmp=0;
        if (!(mydata->fix&1)) {p0 = par[nmp]; nmp++;} else p0 = mydata->fp[nmp];
        if (!(mydata->fix&2)) {p1 = par[nmp]; nmp++;} else p1 = mydata->fp[nmp];
        if (!(mydata->fix&4)) {p2 = par[nmp]; nmp++;} else p2 = mydata->fp[nmp];
    } else {
        p0 = par[0];
        p1 = par[1];
        p2 = par[2];
    }

    m1 = p2 ? 1. / p2 : 1;

    // duplicated code: optimized for speed
    if (fjac) {
        for (int i = 0; i < m_dat; i++) {
            arg = (i - p1) * m1;
            ex = exp(-0.5*arg*arg);
            p0ex = -p0*ex;
            fvec[i] = mydata->y[i] +p0ex;
            fac = p0ex*arg;
            fjac[0*m_dat+i] = -ex;
            fjac[1*m_dat+i] = fac*m1;
            fjac[2*m_dat+i] = fac*m1*arg;
        }
    } else {
        for (int i = 0; i < m_dat; i++) {
            arg = (i - p1) * m1;
            ex = exp(-0.5*arg*arg);
            p0ex = -p0*ex;
            fvec[i] = mydata->y[i] +p0ex;
        }
    }
    /*
*** Original gaussian() code for reference (not optimized)
  double arg, ex, fac, y;
  arg = (x - a[1]) / a[2];
  ex = exp (-arg * arg);
  fac = 2.0 * a[0] * ex * arg;
  y = a[0] * ex;
  jac[0] = ex;
  jac[1] = fac / a[2];
  jac[2] = jac[1] * arg;
  return y;
*/
    *info = *info;                /* to prevent a 'unused variable' warning */
    /* if <parameters drifted away> { *info = -1; } */
}

void lm_print_default( int n_par, const double* par, int m_dat, const double* fvec,
                      const void *data, int iflag, int iter, int nfev )
/*
 *       data  : for soft control of printout behaviour, add control
 *                 variables to the data struct
 *       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
 *       iter  : outer loop counter
 *       nfev  : number of calls to *evaluate
 */
{
    //    double f, y, t;
    int i;
    //    const fitdata_t *mydata = (fitdata_t*)data;

    if (iflag==2) {
        printf ("trying step in gradient direction\n");
    } else if (iflag==1) {
        printf ("determining gradient (iteration %d)\n", iter);
    } else if (iflag==0) {
        printf ("starting minimization\n");
    } else if (iflag==-1) {
        printf ("terminated after %d evaluations\n", nfev);
    }

    printf( "  par: " );
    for( i=0; i<n_par; ++i )
        printf( " %12g", par[i] );
    printf ( " => norm: %12g\n", lm_enorm( m_dat, fvec ) );
    /*
    if ( iflag == -1 ) {
        printf( "  fitting data as follows:\n" );
        for( i=0; i<m_dat; ++i ) {
//            t = (mydata->t)[i];
            y = (mydata->y)[i];
//            f = mydata->user_func( t, par );
            printf( "    t[%2d]=%12g y=%12g fit=%12g residue=%12g\n",
                    i, t, y, f, y-f );
        }
    }
*/
}
void lm_print_none (int n_par, const double *par, int m_dat, const double *fvec, const void *data, int iflag, int iter, int nfev)
{
}


FitterBase::FitterBase(unsigned int fit_points)
{
    ws = 0;
}

FitterBase::~FitterBase()
{
    struct lm_ws_t *p = ws;
    ws = 0;
    if (p)
        lm_free_ws(p);
}

unsigned int FitterBase::getMaxFitPoints() const
{
    if (!ws) return 0;
    return ws->dim_m;
}


FitterGaus::FitterGaus(unsigned int fit_points)
    : FitterBase(fit_points)
{
    ws = lm_alloc_ws (fit_points, 3);
}

int FitterGaus::fit(int fixpar, const double *x, int len, double *par, fitstatus_t *fs)   // approximation
{
    int i, nmp;
    double mp[3];
    lm_control_type control;
    fitdata_gaus_t fitdata;

    if (!x || !par || (len<=0))
        return (-1);

    if (!ws) return(-1);

    if ((unsigned int)len > getMaxFitPoints()) len = getMaxFitPoints();

    for (i = 0; i < 3; i++) {
        par[i] = 0;
        //    err[i] = 0;
    }

    lm_initialize_control (&control);
    fitdata.y = x;
    fitdata.fix = fixpar;
    if (fs) control.maxcall = fs->niter;
    FindGausApproxD (x, len, par);
    nmp=0;
    fitdata.fix=fixpar;
    if (!(fixpar&1)) {mp[nmp] = par[0]; nmp++;} else {fitdata.fp[0] = par[0];}
    if (!(fixpar&2)) {mp[nmp] = par[1]; nmp++;} else {fitdata.fp[1] = par[1];}
    if (!(fixpar&4)) {mp[nmp] = par[2]; nmp++;} else {fitdata.fp[2] = par[2];}
    lm_minimize (ws, len, nmp, mp, lm_evaluate_gaus, lm_print_none, &fitdata, &control);
    //  lm_minimize (ws, len, nmp, mp, lm_evaluate_gaus, lm_print_default, &fitdata, &control);
    nmp=0;
    if (!(fixpar&1)) {par[0] = mp[nmp]; nmp++;}
    if (!(fixpar&2)) {par[1] = mp[nmp]; nmp++;}
    if (!(fixpar&4)) {par[2] = mp[nmp]; nmp++;}
    if (fs) {
        fs->fnorm = control.fnorm;
        fs->niter= control.nfev;
    }
    //  printf ("status: %s after %d evaluations, fnorm=%e\n", lm_shortmsg[control.info], fs->niter, fs->fnorm);
    if (control.info <= 0) return(2);
    //  lm_free_ws(ws);
    return (0);
}

int QuickFitGaus (const double *x, int len, double *par, fitstatus_t *fs) {

    lm_control_type lmc;
    fitdata_gaus_t lmfd;
    struct lm_ws_t *lmws;

    if (len<=0) return(-1);

    lmws = lm_alloc_ws (len, 3);
    if (!lmws) return(-1);

    lm_initialize_control(&lmc);
    lmfd.y=x;
    lmfd.fix=0;

    FindGausApproxD(x,len,par);
    lm_minimize (lmws, len, 3, par, lm_evaluate_gaus, lm_print_none, &lmfd, &lmc);
    lm_free_ws(lmws);
    if (fs) {
        fs->fnorm = lmc.fnorm;
        fs->niter= lmc.nfev;
    }

    return(0);
}

//______________________________________________________________________________
//
// f(x) = p0 + p1 * sin(p2 + p3*x)
//

#ifdef FIT_SINE

#if ! NUMERICAL_JACOBIAN
#warning *** for sine function fit, it is required to set NUMERICAL_JACOBIAN=1
#endif

typedef struct {
    const double *y;
    double fp[4];
    int fix;
} fitdata_sine_t;

struct lm_ws_t *g_ws_sine = 0;

FitterSine::FitterSine(unsigned int fit_points)
    : FitterBase(fit_points)
{
    ws = lm_alloc_ws (fit_points, 4);
#ifdef USE_FFT
    fftw_p = 0;
    fftwr_p = 0;
    fftw_p_size = 0;
#endif
}

double sine_func(double *p, double i)
{
    return p[0] + p[1] * sin(p[2] + p[3]*i);
}

void lm_evaluate_sine (double *par, int m_dat, double *fvec, double *fjac, void *data, int *info)
/*
 *	par is an input array. At the end of the minimization, it contains
 *        the approximate solution vector.
 *
 *	m_dat is a positive integer input variable set to the number
 *	  of functions.
 *
 *	fvec is an output array of length m_dat which contains the function
 *        values the square sum of which ought to be minimized.
 *
 *	data is a read-only pointer to fitdata_sine_t
 *
 *      info is an integer output variable. If set to a negative value, the
 *        minimization procedure will stop.
 */
{
    int i;
    fitdata_sine_t *mydata = (fitdata_sine_t *) data;

    // f(x) = p0 + p1 * sin(p2 + p3*x)
    for (i = 0; i < m_dat; i++) {
        //	fvec[i] = mydata->y[i]  -  (p0 + p1 * sin(p2 + p3*i));
        fvec[i] = mydata->y[i]  -  sine_func(par, i);
    }
    /*
// f(x) = p0 + p1 * sin(p2 + p3*x)
  if (fjac) {
//  printf("BEFORE\n");
//  for (i = 0; i < 7; i++) {
//	printf("fjac[%d] = %e %e %e %e\n", i, fjac[0*m_dat+i], fjac[1*m_dat+i], fjac[2*m_dat+i], fjac[3*m_dat+i]);
//  }

  double f1[4*m_dat];
  double npar[4];
  for(int j=0; j<4; j++)
    npar[j] = par[j];

#define LM_MACHEP .555e-16 // resolution of arithmetic
#define MAX(a,b) (((a)>=(b)) ? (a) : (b))
  double temp = MAX(1.e-14,LM_MACHEP);
  double eps = sqrt(temp); // used in calculating the Jacobian by forward differences
  for(int j=0; j<4; j++) {
        double step = eps*fabs(par[j]);
        if (step == 0.) step = eps;
        npar[j] = par[j] + step;

    for (i = 0; i < m_dat; i++) {
          fjac[j*m_dat+i] = -(sine_func(npar, i)-sine_func(par, i) )/step;
        }
        npar[j] = par[j];

//	f1[0*m_dat+i] = 1;
//	f1[1*m_dat+i] = sin(p2+p3*i);
//	f1[2*m_dat+i] = p1*cos(p2+p3*i);
//	f1[3*m_dat+i] = p1*( cos(p2+p3*i) + sin (p2+p3*i)* i);
  }

//  printf("AFTER\n");
//  for (i = 0; i < 7; i++) {
//	printf("f1[%d] = %e %e %e %e\n", i, f1[0*m_dat+i], f1[1*m_dat+i], f1[2*m_dat+i], f1[3*m_dat+i]);
//  }
//  printf("-----------------------------------------------------------------------\n");
  }
*/
    //  printf("lm_evaluate_sine: f(x) = p0 + p1 * sin(p2 + p3*x), p0=%e, p1=%e, p2=%e, p3=%e\n", p0, p1, p2, p3);
    //  for (i = 0; i < m_dat; i++) {
    //	printf("fvec[%3d]: % 5.3f\n", i, fvec[i]);
    //  }


    *info = *info;                /* to prevent a 'unused variable' warning */
    /* if <parameters drifted away> { *info = -1; } */
}

// Frequency Interpolation
double tau(double x)
{
    return 1./4 * log(3*x*x + 6*x + 1) - sqrt(6)/24 * log((x + 1 - sqrt(2./3))  /  (x + 1 + sqrt(2./3)));
}

double QuinnEstimator(double xr[3], double xi[3])
{
    // Quinn's Second Estimator
    //    * k =  index of the max (possibly local) magnitude of an DFT.
    //    * X[i]  =  bin "i" of an DFT |X[i]| =  magnitude of DFT at bin "i".
    //    * k'  =  the interpolated bin location.

    double ap = (xr[2] * xr[1] + xi[2] * xi[1])  /  (xr[1] * xr[1] + xi[1] * xi[1]);
    double dp = -ap / (1 - ap);
    double am = (xr[0] * xr[1] + xi[0] * xi[1])  /  (xr[1] * xr[1] + xi[1] * xi[1]);
    double dm = am / (1 - am);
    double d = (dp + dm) / 2 + tau(dp * dp) - tau(dm * dm);
    return d;
}

double GausEstimator(double x[3])
{
    double gp[3]={0,0,0};
    FindGausApproxD(x, 3, gp);
    return gp[1]-1;
}
double QuadraticEstimator(double x[3])
{
    return (x[2] - x[0]) / (2 * (2 * x[1] - x[0] - x[2]));
}
double BarycentricEstimator(double x[3])
{
    return (x[2] - x[0]) / (x[0] + x[1] - x[2]);
}
/*
double Quinn1Estimator(complex x[3])
{
  double  alpha1 = creal(x[0]/x[1]);
  double  alpha2 = creal(x[2]/x[1]);

  double delta1 =  alpha1/(1-alpha1);
  double delta2 = -alpha2/(1-alpha2);

  if ((delta1>0) & (delta2>0))
        return delta2;
  else
    return delta1;
}
*/
int FitterSine::fit (int fixpar, const double *x, int len, double *par0, double *par, fitstatus_t *fs)   // approximation
{
    int i, nmp;
    double mp[4];
    lm_control_type control;
    fitdata_sine_t fitdata;
    int fft_used = 0;

    //  printf("============= x=%p, par=%p, len=%u\n", x, par, len);
    if (!x || !par || (len<=0))
        return (-1);

    if (!ws) return(-1);
    if (len>getMaxFitPoints()) len = getMaxFitPoints();
    for (i = 0; i < 4; i++) {
        par[i] = 0;
        //    err[i] = 0;
    }

    lm_initialize_control (&control);
    fitdata.y = x;
    fitdata.fix = fixpar;
    //  if (fs) control.maxcall = fs->niter; // warning
    // preset f(x) = p0 + p1 * sin(p2 + p3*x)

    // Find initial parameter values - important!
    double xw[MAX_FITSINE_POINTS];
    for(i=0; i<len; i++) {
        double wf = windowingFunction(WF_HAMMING, i, len);
        xw[i] = x[i] * wf;
        //	printf("xw[%2d]: %f %f\n", i, xw[i], wf);
    }

    double imax=0;
    //  double phimax=0;

#ifdef USE_FFT
    {
        if (fftw_p_size != len) { // check and create plan
            if (fftwr_p) rfftw_destroy_plan(fftwr_p);
            fftwr_p = rfftw_create_plan(len, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
            fftw_p_size = len;
        }
        int N = len;
        fftw_real in[N], out[N], power_spectrum[N/2+1];
        for(i=0;i<N;i++) {
            in[i] = xw[i];
        }
        rfftw_one(fftwr_p, in, out);
        power_spectrum[0] = out[0]*out[0];  // DC component
        for (int k = 1; k < (N+1)/2; ++k)  // (k < N/2 rounded up)
            power_spectrum[k] = out[k]*out[k] + out[N-k]*out[N-k];
        if (N % 2 == 0) // N is even
            power_spectrum[N/2] = out[N/2]*out[N/2];  // Nyquist freq.

#if 0
        for(int k = 1; k < N/2+1; ++k) {
            double complex c = out[k] + I*out[N-k];
            double mag = cabs(c); double pwr = mag*mag;
            double phi = carg(c);
            printf("%3d  %e  %e  %e\n", k, pwr, power_spectrum[k], phi/M_PI);
        }
        exit(0);
#endif

        int imaxbin = 0;
        double vmax= 0;
        for(i=1; i<(N+1)/2; i++) // start from 1st harmonic
        {
            //	printf("[%2d]: %f\n", i, P[i]);
            if (power_spectrum[i] > vmax) {
                vmax = power_spectrum[i];
                imaxbin = i;
            }
        }
        imax = imaxbin;
        fft_used = 1;

        if ((imaxbin >= 1) && (imaxbin+1 <(N+1)/2)) {
            double complex xc[3];
            double xpow[3];
            double xre[3];
            double xim[3];
            for(i=0; i<3; i++) {
                int k = imaxbin-1+i;
                xc[i] = out[k] + I*out[N-k]; // halfcomplex FFT format
                xre[i] = creal(xc[i]);
                xim[i] = cimag(xc[i]);
                xpow[i] = cabs(xc[i]) * cabs(xc[i]);
            }
            phimax = carg(xc[1]);
            //	double imax1 = imaxbin + Quinn1Estimator(xc);
            //	imax = imaxbin + QuinnEstimator(xc); // not working
            //	imax = imaxbin + GausEstimator(xre2);
            //	imax = imaxbin + QuadraticEstimator(&power_spectrum[imaxbin-1]); // best
            imax = imaxbin + QuadraticEstimator(xpow); // best
            //	imax = imaxbin + BarycentricEstimator(xre);
            //	printf("%e %e    %f\n", imax1, imax, imax1-imax);
            //	imax = imax1;
            //	printf("%f pi\n", phimax/M_PI);
        }

    }
#endif

    /*
#ifdef USE_FFT
  {
        if (fftw_p_size != len) { // check and create plan
          if (fftw_p) fftw_destroy_plan(fftw_p);
          fftw_p = fftw_create_plan(len, FFTW_FORWARD, FFTW_ESTIMATE);
          fftw_p_size = len;
        }
        int N = len;
    fftw_complex fftw_in[N], fftw_out[N];
        fftw_real power_spectrum[N];
        for(i=0;i<N;i++) {
          fftw_in[i].re = xw[i];
          fftw_in[i].im = 0;
        }
        fftw_one(fftw_p, fftw_in, fftw_out);
        double complex out[N];
        for(i=0;i<N;i++) {
          out[i] = fftw_out[i].re + I*fftw_out[i].im;
        }
    for (int k = 0; k < N; ++k)
      power_spectrum[k] = cabs(out[k]) * cabs(out[k]);

#if 0
        for(int k = 0; k < N; ++k) {
          double pwr = cabs(out[k])*cabs(out[k]);
          double phi = carg(out[k]);
          printf("%3d  %e  %e  %e\n", k, pwr, power_spectrum[k], phi/M_PI);
        }
        exit(0);
#endif

        int imaxbin = 0;
    double vmax= 0;
    for(i=1; i<(N+1)/2; i++) // start from 1st harmonic
    {
//	printf("[%2d]: %f\n", i, P[i]);
          if (power_spectrum[i] > vmax) {
                vmax = power_spectrum[i];
                imaxbin = i;
          }
        }
        imax = imaxbin;
        fft_used = 1;

  if ((imaxbin >= 1) && (imaxbin+1 <len)) {
        double mxre[3], mxim[3];
        double pwr[3];
        for(i=0;i<3;i++) {
          int k = imaxbin-1+i;
          pwr[i] = cabs(out[k])*cabs(out[k]);
        }
//	mx[0] = power_spectrum[imaxbin-1];
//	mx[1] = power_spectrum[imaxbin];
//	mx[2] = power_spectrum[imaxbin+1];
//	mxre[0] = out[0].re; mxim[0] = out[0].im;
//	mxre[1] = out[1].re; mxim[1] = out[1].im;
//	mxre[2] = out[2].re; mxim[2] = out[2].im;
        mxre[0] = cabs(out[0]); mxim[0] = carg(out[0]);
        mxre[1] = cabs(out[1]); mxim[1] = carg(out[1]);
        mxre[2] = cabs(out[2]); mxim[2] = carg(out[2]);
//	imax = imaxbin + Quinn1Estimator(&out[imaxbin-1]);
        imax = imaxbin + QuadraticEstimator(pwr); // best
        phimax = carg(out[imaxbin]);
  }

  }
#endif
*/
    if (!fft_used) {
        double S[MAX_FITSINE_POINTS/2+1], C[MAX_FITSINE_POINTS/2+1], P[MAX_FITSINE_POINTS/2+1];
        DFT (xw, len, S, C);
        RECT2POW (S, C, len, P);
        double vmax = 0;
        for(i=1; i<len/2+1; i++) // start from 1st harmonic
        {
            //	printf("[%2d]: %f\n", i, P[i]);
            if (P[i] > vmax) {
                vmax = P[i];
                imax = i;
            }
        }
    }

    double fpeak = 2*M_PI*imax/len;

    double ymin, ymax;
    ymin=ymax=x[0];
    for(i=0; i<len; i++) {
        if (x[i] < ymin) ymin = x[i];
        if (x[i] > ymax) ymax = x[i];
    }
    double yrange = ymax - ymin;
    if (yrange < 0) yrange = 1;

    //// find first zero-crossing
    //  int zc=0;
    //  for(i=1; i<len; i++)
    //	if ((x[i-1] <= 0) && (x[i] > 0)) {zc=i; break;}

    par[0] = 0;
    par[1] = yrange/2;
    par[3] = fpeak;
    par[2] = 0;

    for(i=0;i<4;i++) par0[i] = par[i];
    //  par[2] = par[3]?(-zc/par[3]):0;
    nmp=0;
    fitdata.fix=fixpar;
    if (!(fixpar&1)) {mp[nmp] = par[0]; nmp++;} else {fitdata.fp[0] = par[0];}
    if (!(fixpar&2)) {mp[nmp] = par[1]; nmp++;} else {fitdata.fp[1] = par[1];}
    if (!(fixpar&4)) {mp[nmp] = par[2]; nmp++;} else {fitdata.fp[2] = par[2];}
    if (!(fixpar&8)) {mp[nmp] = par[3]; nmp++;} else {fitdata.fp[3] = par[3];}
    lm_minimize (ws, len, nmp, mp, lm_evaluate_sine, lm_print_none, &fitdata, &control);
    //  lm_minimize (ws, len, nmp, mp, lm_evaluate_sine, lm_print_default, &fitdata, &control);
    nmp=0;
    if (!(fixpar&1)) {par[0] = mp[nmp]; nmp++;}
    if (!(fixpar&2)) {par[1] = mp[nmp]; nmp++;}
    if (!(fixpar&4)) {par[2] = mp[nmp]; nmp++;}
    if (!(fixpar&8)) {par[3] = mp[nmp]; nmp++;}
    if (fs) {
        fs->fnorm = control.fnorm;
        fs->niter= control.nfev;
    }
    //  printf ("status: %s after %d evaluations, fnorm=%e\n", lm_shortmsg[control.info], fs->niter, fs->fnorm);
    if (control.info <= 0) return(2);
    //  lm_free_ws(g_ws_sine);

    //  printf("fpeak error: %e %e %e\n", fpeak, par[3], fpeak-par[3]);
    //  printf("fpeak error: %8.f ppm, ipeak=%f\n", fabs(fpeak-par[3])/(fpeak+par[3])*2*1e6, imax);
    //  par[3] = fpeak;

    //  printf("phimax = %f, par2 = %f\n", phimax/M_PI, par[2]/M_PI);
    return (0);
}

#endif /* FIT_SINE */

// New interface

double GausFcn(const GausPar &par, double x)
{
    return par.a * gauss(x-par.x0, par.sigma);
}

double GaussianApproximator::FindGausDevNorm (const std::vector<int> &x, const GausPar &par)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausDevNorm(xd, par);
}

double GaussianApproximator::FindGausDevNorm (const std::vector<double> &x, const GausPar &par)
{
    const size_t len = x.size();
    double rms = 0;
    for (size_t i = 0; i < len; i++) {
        rms += pow(x[i] - GausFcn(par, i), 2);
    }
    return sqrt(rms);
}

double GaussianApproximator::FindGausDevNormShifted (const std::vector<int> &x, const GausPar &par)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausDevNormShifted(xd, par);
}

double GaussianApproximator::FindGausDevNormShifted (const std::vector<double> &x, const GausPar &par)
{
    const size_t len = x.size();
    double rms = 0;
    for (size_t i = 0; i < len; i++) {
        rms += pow(x[i] - GausFcn(par, i+0.5), 2);
    }
    return sqrt(rms);
}

/** @brief Gaussian function approximation, x shifted by 0.5
 *
 *  @par[in] x data values defined at 0.5, 1.5, 2.5, ...
 */
GausPar GaussianApproximator::FindGausApproxShifted (const std::vector<double> &x)
{
    GausPar par;
    const size_t len = x.size();
    double allcha = 0, sumx = 0, sumx2 = 0;

    for (size_t i = 0; i < len; i++) {
        const double binCenter = i + 0.5;
        const double &val = x[i];
        sumx += val * binCenter;
        sumx2 += val * binCenter * binCenter;
        allcha += val;
    }
    if (allcha > 0) {
        const double mean = sumx / allcha;
        double rms = sqrt (sumx2 / allcha - mean * mean);
        if (rms < M_1_SQRT2PI) rms = M_1_SQRT2PI; // force lower limit
        par.a = M_1_SQRT2PI * allcha / rms;
        par.x0 = mean;
        par.sigma = rms;
    } else {
        par.a = 0;
        par.x0 = len/2 + 0.5;
        par.sigma = len/2;
    }
    return par;
}

GausPar GaussianApproximator::FindGausApproxShifted (const std::vector<int> &x)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausApproxShifted(xd);
}

/** @brief Gaussian function approximation
 *
 *  @par[in] x data values defined at 0, 1, 2, ...
 */
GausPar GaussianApproximator::FindGausApprox (const std::vector<int> &x)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausApprox(xd);
}

GausPar GaussianApproximator::FindGausApprox (const std::vector<double> &x)
{
    GausPar par;
    const size_t len = x.size();
    double allcha = 0, sumx = 0, sumx2 = 0;
    for (size_t i = 0; i < len; i++) {
        const double val = x[i];
        const double binCenter = i;
        sumx += val * binCenter;
        sumx2 += val * pow(binCenter, 2);
        allcha += val;
    }
    {
        if (allcha > 0) {
            const double mean = sumx / allcha;
            double rms = sqrt (sumx2 / allcha - mean * mean);
            if (rms < M_1_SQRT2PI) rms = M_1_SQRT2PI; // force lower limit
            par.a = M_1_SQRT2PI * allcha / rms;
            par.x0 = mean;
            par.sigma = rms;
        } else {
            par.a = 0;
            par.x0 = len/2;
            par.sigma = len/2;
        }
    }
    return par;
}

