
#include <array>
#include <cmath>
#include <vector>
//#include <algorithm>

#include "../dsplib/fitter.h"
#include "th1.h"

const double eps = 1e-100;

th1::th1 (const QString &_title, const QString &_xLabel, int fnbins, double fmin, double fmax, int foptions):
    title(_title), xLabel(_xLabel)
{
    options = foptions;
    if (0==(options & NO_AUTORANGE))
        options |= AUTORANGE;
    resize(fnbins, fmin, fmax);
    clear ();
}

th1::th1(const QString &_title, int fnbins, double fmin, double fmax, int foptions)
    : th1(_title, QString(), fnbins, fmin, fmax, foptions)
{
}

bool th1::isEmpty() const
{
    return nent == 0 && over == 0 && under == 0;
}


bool th1::isEqualRange(const th1 &r) const
{
    return isEqualRange(r.nbins, r.xmin, r.xmax);
}

bool th1::isEqualRange(size_t _nbins, double _xmin, double _xmax) const
{
    return _nbins == nbins
            && abs(_xmin - xmin) < eps
            && abs(_xmax - xmax) < eps;
}

void th1::resize(const th1 &r)
{
    resize(r.nbins, r.xmin, r.xmax);
}

void th1::resize(size_t _nbins, double _xmin, double _xmax)
{
    if (_nbins <= 0)
        _nbins = 1;
    if (isEqualRange(_nbins, _xmin, _xmax))  // no change
        return;
    nbins = _nbins;
    clear();
    py.resize(nbins);
    setrange(_xmin, _xmax);
    update_cache();
    mod++;
}

void th1::clear ()
{
    under = 0;
    over = 0;
    nent = 0;
    py.clear();
    py.resize(nbins);
    //    std::fill(py.begin(), py.end(), 0.);
    statistics.clear();
    mod++;
}

th1 &th1::operator+=(const th1 &r)
{
    if (!isEqualRange(r)) {
        resize(r);
    }
    Q_ASSERT(py.size() == r.py.size());
    if (py != r.py) {
        for (size_t i=0; i<nbins; i++)
            py[i] += r.py[i];
        mod++;
    }
    if (r.over) {
        over += r.over;
        mod++;
    }
    if (r.under) {
        under += r.under;
        mod++;
    }
    if (r.nent) {
        nent += r.nent;
        mod++;
    }
    statistics += r.statistics;
    return *this;
}

//void th1::add(const th1 *summand)
//{
//    Q_ASSERT(summand->get_nbins()==get_nbins());
////    std::transform(py.begin(), py.end(), summand->get_all_bins().begin(),
////                   py.begin(), std::plus<double>());
//    std::vector<double> all_bins = summand->get_py();
//    std::vector<double>::iterator itAdd = all_bins.begin();
//    for(std::vector<double>::iterator itSum=py.begin(); itSum!=py.end();++itSum){
//        double add = *itAdd;
//        double sum = *itSum;
//        sum +=add;
//        *itSum = sum;
////        *itSum += *itAdd;
//        ++itAdd;
//    }

//    bool isModified=false;
//    if(summand->get_under()){
//        isModified=true;
//        under += summand->get_under();
//    }
//    if(summand->get_over()){
//        isModified=true;
//        over += summand->get_over();
//    }
//    if(summand->get_nent()){
//        isModified=true;
//        nent += summand->get_nent();
//    }

//    if(isModified)
//        mod++;
//}


//void th1::copyEntry(const th1 *copy)
//{
//    Q_ASSERT(copy->get_nbins()==get_nbins());
//    if(get_nbins())
//        memcpy(&py[0], &(copy->get_all_bins()[0]), sizeof(double)*get_nbins());
//    under=copy->get_under();
//    over=copy->get_over();
//    nent=copy->get_nent();
//}

bool th1::autorange_enabled() const
{
    return ((options & AUTORANGE) == AUTORANGE);
}

bool th1::fit_enabled() const
{
    bool state = ((options & FIT) == FIT);
    return state;
}

bool th1::fit_resol_enabled() const
{
    bool state = ((options & FIT_RESOL) == FIT_RESOL);
    return state;
}

void th1::set_autorange_enable(bool state) const
{
    if (state == autorange_enabled()) return;
    if (state)
        options |= AUTORANGE;
    else
        options &= ~AUTORANGE;
    update_autorange();
}

void th1::set_fit_enable(bool state) const
{
    if (state)
        options |= FIT;
    else
        options &= ~FIT;
}

double th1::binwidth() const
{
    return cache_binwidth;
}

double th1::binleft (int n) const
{
    return xmin + n * cache_binwidth;
}

double th1::binright (int n) const
{
    return xmin + (1 + n) * cache_binwidth;
}

double th1::bincenter (int n) const
{
    //    double binwidth = (xmax - xmin) / nbins;
    return xmin + (0.5 + n) * cache_binwidth;
}

int th1::findbin (double x) const
{
//    double hw = xmax - xmin;
//    if (hw == 0)
//        hw = 1;

    return static_cast<int>(floor((x - xmin) / cache_binwidth));
}

void th1::update_cache()
{
    cache_xrange = xmax - xmin;
    if (abs(cache_xrange) < eps)
        cache_xrange = 1;
    cache_binwidth = cache_xrange / nbins;
}

void th1::setrange (double inmin, double inmax)
{
    if (inmin > inmax)  // reversed
        std::swap(inmin, inmax);
    if (inmax - inmin < eps)  // too narrow
        inmax = inmin + 1.0;
    if (abs(inmin - xmin) < eps && abs(inmax - xmax) < eps)  // not changed
        return;
    xmin = inmin;
    xmax = inmax;
    update_cache();
    ar_min = xmin;
    ar_max = xmax;
    mod++;
}

void th1::getrange (double *outmin, double *outmax) const
{
    double min = ar_min;
    double max = ar_max;
    if (!autorange_enabled() || (ar_min>=ar_max)) { // no autorange
        min = xmin;
        max = xmax;
    }

    if (outmin)
        *outmin = min;
    if (outmax)
        *outmax = max;
}

void th1::fill (double x)
{
    filln (x, 1);
}

void th1::filln (double x, double count)
{
    mod++;
    int bin = findbin (x);
    if (bin < 0) {
        //    under+=(int)count;
        under++;
        return;
    }
    if ((unsigned int) bin >= nbins) {
        //    over+=(int)count;
        over++;
        return;
    }
    //  nent += (int)count;
    nent ++;
    py[bin] += count;
    statistics.fill(x);
}

GenericStatistics th1::simplestats() const
{
    return  statistics;
    GenericStatistics s;
    for (auto d : py)
        s.fill(d);
    return s;
//    std::array<double, 3> par = {0};
//    FindGausApproxD(&py[0], py.size(), &par[0]);
//    if (value)
//        *value = par[0];
//    if (mean)
//        *mean = xmin + (0.5 + par[1]) * cache_binwidth;
//    if (sigma)
//        *sigma = par[2] * cache_binwidth;
}

int th1::fitgaus (double *fvalue, double *fmean, double *fsigma, double *fnorm) const
{
    double value, mean, sigma, norm;

    unsigned int fitstart = findbin(ar_min);
    unsigned int fitend = findbin(ar_max);
    if (fitstart >= fitend || fitend > nbins) { // bad range
        fitstart=0;
        fitend=nbins;
    }
    unsigned int fitnbins = fitend-fitstart;

    //  int *inp = new int[fitnbins];
    double par[3];
    //  double err[3], chi2;
    //  for (i = 0; i < fitnbins; i++)
    //    inp[i] = (int)py[fitstart+i];
    // select fit method: simple or with MIGRAD

    //  if (nbins > 100)
    //	FindGausApprox(inp, nbins, par);
    //  else

    fitstatus_t fs;
    fs.niter = 20;
    QuickFitGaus(&py[fitstart], fitnbins, par, &fs);

    //  delete (inp);
    value = par[0];


    double binwidth = (xmax - xmin) / nbins;
    mean = xmin + (fitstart+0.5 + par[1]) * binwidth;
    sigma = par[2] * binwidth;
    norm = (value>0)?(fs.fnorm/value):1000;

    if (fvalue)
        *fvalue = value;
    if (fmean)
        *fmean = mean;
    if (fsigma)
        *fsigma = sigma;
    if (fnorm)
        *fnorm = norm;
    return 0;
}

GenericStatistics th1::getStatistics() const
{
    return statistics;
}

QString th1::getname() const
{
    return title;
}

unsigned int th1::get_nbins () const
{
    return nbins;
}

double th1::get_bin (unsigned int index) const
{
    if (index >= nbins)
        return 0;
    return py.at(index);
}

unsigned int th1::get_nent () const
{
    return nent;
}

unsigned int th1::get_over () const
{
    return over;
}

unsigned int th1::get_under () const
{
    return under;
}

double th1::get_sum () const
{
    double s=0;
    for(unsigned int i=0;i<py.size();i++)
        s += py[i];
    return s;
}

int th1::get_modseq () const
{
    return mod;
}

double th1::get_xmin () const
{
    return xmin;
}

double th1::get_xmax () const
{
    return xmax;
}

bool th1::find_autorange_bins(unsigned int *bfirst, unsigned int *blast) const
{
    if (py.empty()) {
        if (bfirst)
            *bfirst = 0;
        if (blast)
            *blast = nbins;
        return false;
    }

    const double maxvalue = *std::max_element(py.begin(), py.end());
    double threshold = maxvalue * autorange_ratio;
    if (threshold < 2) threshold = autorange_threshold;

    size_t bmin = 0;
    size_t bmax = nbins;
    bool found = false;
    for (size_t i = 0; i < nbins; i++)
        if (py[i] > threshold) {
            bmin = i;
            found = true;
            break;
        }
    for (size_t i = nbins-1; i > 1; i--)
        if (py[i-1] > threshold) {
            bmax = i;
            found = true;
            break;
        }
    double pxmin = bincenter(bmin) - cache_binwidth / 2;
    double pxmax = bincenter(bmax-1) + cache_binwidth / 2;
    ar_min = pxmin;
    ar_max = pxmax;
    if (bfirst)
        *bfirst = bmin;
    if (blast)
        *blast = bmax;
    return found;
}

void th1::update_autorange() const
{
    if (!autorange_enabled()) {
        ar_min = xmin;
        ar_max = xmax;
        return;
    }
    if (py.empty())
        return;

//    double maxvalue = 0;
//    for (unsigned int i = 0; i < nbins; i++) {
//        if (py[i] > maxvalue) maxvalue = py[i];
//    }
    const double maxvalue = *std::max_element(py.begin(), py.end());

    double threshold = maxvalue * 0.025;
    if (threshold < 2) threshold = 2;

    double pxmin = bincenter(0), pxmax = bincenter(nbins-1);
    int i;
    for (i = 1; i < (int)nbins; i++)
        if (py[i] > threshold) {
            pxmin = bincenter(i-1);
            break;
        }
    for (int j = (int)nbins - 3; j >= i; --j)
        if (py[j] > threshold) {
            pxmax = bincenter(j+2);
            break;
        }
    if (pxmin<pxmax) {
        ar_min = pxmin;
        ar_max = pxmax;
    }
}
