//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QtGlobal>
#ifdef Q_WS_X11
#include <cstdlib>
#endif
#include <QMutex>
#include <QDateTime>
#include <QDebug>
#include <QMap>
#include "qlog.h"
#include "qlog_p.h"
#include "ansi_col.h"

namespace
{
    QMutex instanceMutex; // FIXME: non-POD static
}

QLog* QLog::_instance = nullptr;

QLog::Cleanup::~Cleanup()
{
    // If nothing else (and it may not be clear exactly how a multi-threaded
    // program shuts down, or how locking still works at this time),
    // this provides proper synchronisation with thread-initiated initialisation,
    // although that could also be assured by making _instance an std::atomic.
      QMutexLocker guard(&instanceMutex);

    if (_instance)
    {
        delete QLog::_instance;
        QLog::_instance=nullptr;
    }
}

QLog &QLog::getInstance()
{
#if 0 // "broken" optimisation
    // You might execute the protected code only "if (!_instance)",
    // and thus see a potentially substantial performance gain, but this
    // "double-checked locking" technique has been proven to be unreliable,
    // and in specific circumstances, depending on processor, compiler, level
    // of optimisation, and program timing, such a program may misbehave.
    // Then again, you might get away with it most of the time, and maybe
    // your application allows to trade off reliability for performance.
    // A portable way to make double-checked locking work reliably
    // is to make _instance a std::atomic type (or it will be when this
    // becomes part of the standard, and your programming environment
    // may already provide advance support for it).
    // Non-portable ways to achieve the same effect may be the use of "volatile"
    // with Visual C++ 2005 or later (implying std::atomic-like behaviour
    // like indivisible access and associated memory fences, which is not
    // standard now and very unlikely to become so), or any other special
    // provisions that sufficiently resemble std::atomic.
    if (!_instance)
#else // safe but suboptimal
    // To amortise the cost of obtaining the lock, the client code may
    // eliminate unnecessary calls by locally storing a reference or pointer
    // to the single instance it previously obtained through getInstance().
    if (true)
#endif
    {
        // Assuming that we have some UDT that locks the mutex in its constructor
        // and unlocks in the destructor.
            QMutexLocker guard(&instanceMutex);

        if (!_instance)
        {
            _instance = new QLog();
        }
    }

    return *_instance;
}


QLog::QLog()
    : d(new QLogPrivate)
{
}

QLog::~QLog()
{
    delete d;
}

void QLog::setDebugLevel(int debugLevel)
{
    d->debugLevel = debugLevel;
}

int QLog::getDebugLevel() const
{
    return d->debugLevel;
}

void QLog::err(const char *str) const
{
    log_pri(QLog::Err, 0, str);
}

void QLog::err(const QString &str) const
{
    log_pri(QLog::Err, 0, str);
}

void QLog::warn(const char *str) const
{
    log_pri(QLog::Warning, 0, str);
}

void QLog::warn(const QString &str) const
{
    log_pri(QLog::Warning, 0, str);
}

void QLog::notice(const char *str) const
{
    log_pri(QLog::Notice, 0, str);
}

void QLog::notice(const QString &str) const
{
    log_pri(QLog::Notice, 0, str);
}

void QLog::info(const char *str) const
{
    log_pri(QLog::Info, 0, str);
}

void QLog::info(const QString &str) const
{
    log_pri(QLog::Info, 0, str);
}

#ifdef QT_DEBUG
void QLog::debug(int level, const char *str) const
{
    log_pri(QLog::Debug, level, str);
}

void QLog::debug(int level, const QString &str) const
{
    log_pri(QLog::Debug, level, str);
}
#endif

void QLog::log_pri(int priority, int debugLevel, const char *str) const
{
    log_pri(priority, debugLevel, QString::fromUtf8(str));
}

void QLog::log_pri(int priority, int debugLevel, const QString &str) const
{
    if ((priority == QLog::Debug) && (getDebugLevel() < debugLevel))
        return;

    QString prioTag = "default";
    switch (priority) {
    case QLog::Alert:
        prioTag = "alert";
        break;
    case QLog::Crit:
        prioTag = "critical";
        break;
    case QLog::Err:
        prioTag = "error";
        break;
    case QLog::Warning:
        prioTag = "warning";
        break;
    case QLog::Notice:
        prioTag = "notice";
        break;
    case QLog::Info:
        prioTag = "info";
        break;
    case QLog::Debug:
        prioTag = QString("debug:%1").arg(debugLevel);
        break;
    }
#ifdef Q_WS_X11
    bool ansiCapable = false;
    QString term(getenv("TERM"));
    if (term == "ansi" || term == "linux" || term == "xterm") ansiCapable = true;
    if (ansiCapable) {
        QMap<int, QString> prioColor;
        prioColor[QLog::Alert] = ANSI_MAGENTA;
        prioColor[QLog::Crit] = ANSI_MAGENTA;
        prioColor[QLog::Err] = ANSI_RED;
        prioColor[QLog::Warning] = ANSI_YELLOW;
        prioColor[QLog::Notice] = ANSI_BLUE;
        prioColor[QLog::Info] = ANSI_NORM;
        prioColor[QLog::Debug] = ANSI_NORM;
        prioTag = prioColor[priority]+prioTag + ANSI_NORM;
    }
#endif
    QString timestamp = QDateTime::currentDateTime().toString("MMM dd hh:mm:ss.zzz ");
    qDebug() << timestamp << QString("["+prioTag+"] " + str+"\n");
    //    d->qout.device()->flush();
}

QLog &logger = QLog::getInstance();
