//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>

// Compile this source with gcc -fast-math option

#include "qxw_randgen.h"

#define BITS2MAXINT(bits)  ((1<<((bits)-1)) - 1)
#define BITS2MININT(bits)  (-(1<<((bits)-1)))

namespace qxw {

double gauss (double x, double sigma)
{

    // An unnormalized gaussian function with mean=0 and sigma
    // Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed

    double arg, res;
    // Calculate a gaussian function with mean and sigma.
    if (sigma == 0)
        return 1.e30;
    arg = x / sigma;
    res = exp (-0.5 * arg * arg);
    return res;
    //   return res/(2.50662827463100024*sigma); //sqrt(2*Pi)=2.50662827463100024
}

// Windowing functions

double detpulse (double a, double t, double rt, double ft)
{

    // Simulate detector pulse
    //

    double v, r, f, e;

    //  if (rt > ft) m = ft; else m = rt;

    //  e = TMath::Log(ADC_RES*10); // set error to 0.1 resolution
    e = log (1e6);

    r = t / rt;
    f = t / ft;
    if (f > e)
        return (0);
    if (r < -4)
        return (0);
#ifndef _MSC_VER
    v = a * (erf (r) + 1) / 2 * exp (-f);
#else
    if (f > 0)
        v = a * exp(-f);
    else
        v = 0;
#endif
    // if (v<0) v=0;
    return v;
}

void Normalize (double *x, int n)
{

    // Normalize to integral=1

    int i;
    double sum;
    sum = 0;
    for (i = 0; i < n; i++)
        sum += x[i];
    if (sum != 0)
        for (i = 0; i < n; i++)
            x[i] /= sum;
}

void Make_Discrete (const double *x, int *out, int n, int bits)
{

    // Scale to fit in <bits>-bit integer

    int i, imin, imax;
    double min, max, scalepos, scaleneg, scale;

    if (bits <= 1)
        return;
    imin = BITS2MININT (bits) /* - (1<<(bits-1)) */ ;
    imax = BITS2MAXINT (bits) /* ((1<<(bits-1)) - 1) */ ;
    min = 0;
    max = 0;

    for (i = 0; i < n; i++) {
        if (x[i] > max)
            max = x[i];
        if (x[i] < min)
            min = x[i];
    }
    if (max > 0)
        scalepos = imax / max;
    else
        scalepos = 1e38;
    if (min < 0)
        scaleneg = imin / min;
    else
        scaleneg = 1e38;
    if (scalepos < scaleneg)
        scale = scalepos;
    else
        scale = scaleneg;
    // printf("Make_Discrete(): min=%f, max=%f, scaling by %f (neg %f, pos %f)\n", min, max, scale, scaleneg, scalepos);
    if (scale != 0)
        for (i = 0; i < n; i++)
            out[i] = (int) (x[i] * scale);
}

int intsum (const int *x, int len)
{
    int i, s = 0;

    for (i = 0; i < len; i++)
        s += x[i];
    return (s);
}

double rndm ()
{
    // System buil-in
    // Random number in range [0,1) with period approx. 16*((2**31)-1)
    double value;
    value = (double) rand () / RAND_MAX;
    return (value);
}

unsigned int f1Seed;

double rndm1 ()
{
    //  Machine independent random number generator.
    //  Produces uniformly-distributed floating points between 0 and 1.
    //  Identical sequence on all machines of >= 32 bits.
    //  Periodicity = 10**8
    //  Universal version (Fred James 1985).
    //  generates a number in ]0,1]

    const double kCONS = 4.6566128730774E-10;
    const int kMASK24 = 2147483392;
    unsigned int jy;

    f1Seed *= 69069;
    jy = (f1Seed & kMASK24);      // Set lower 8 bits to zero to assure exact float
    if (jy)
        return kCONS * jy;
    return rndm ();
}



int fSeed1 = 9876;
int fSeed2 = 54321;

double rndm2 ()
{
    //  Machine independent random number generator.
    //  Produces uniformly-distributed floating points in [0,1]
    //  Identical sequence on all machines of >= 32 bits.
    //  Periodicity > 10**14

    double iz, r;
    int k = (int) (fSeed1 / 53668);
    fSeed1 = 40014 * (fSeed1 - k * 53668) - k * 12211;
    if (fSeed1 < 0)
        fSeed1 += 2147483563;
    k = (int) (fSeed2 / 52774);
    fSeed2 = 40692 * (fSeed2 - k * 52774) - k * 3791;
    if (fSeed2 < 0)
        fSeed2 += 2147483399;
    iz = fSeed1 - fSeed2;
    if (iz <= 0)
        iz += 2147483562;
    r = iz * 4.6566128e-10;
    return r;
}

// returns a uniform deviate on the interval ]x1, x2].
double ranuni (double x1, double x2)
{
    double ans = rndm ();
    return x1 + (x2 - x1) * ans;
}

double rangaus (double sigma)
{
    //      Return a number distributed following a gaussian with sigma

    double x, y, z, result;

    y = rndm ();
    z = rndm ();
    x = z * 6.28318530717958623;
    result = sigma * sin (x) * sqrt (-2 * log (y));
    return result;
}
}
