#include "EthFlashDev.h"

#include <QHostAddress>

#include "flash-prog/FlashDevTypes.h"
#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegWrite;

namespace {
const uint MAX_PART_SIZE = 256;
constexpr fw_version_t REV_FLASH_MIN = fw_version_t(1, 0, 24140);
constexpr fw_version_t REV_FLASH_ID = fw_version_t(1, 0, 27846);
}

EthFlashDev::EthFlashDev(quint16 id, quint64 serial) :
    QObject(),
    dev(new QtMregDevice(id, serial, this)),
    hw_maj_ver(0),
    writeInProgress(false)
{
    connect(this, SIGNAL(deviceUpdated(DeviceDescription)), dev, SLOT(deviceUpdated(DeviceDescription)));
    connect(this, SIGNAL(deviceUpdated(DeviceDescription)), this, SLOT(updatedHwVer(DeviceDescription)));

    connect(dev, SIGNAL(deviceConnected()), SLOT(on_devConnected()));
    connect(dev, SIGNAL(deviceDisconnected()), SIGNAL(deviceDisconnected()));
    connect(dev, SIGNAL(setStatus(QString)), SIGNAL(setStatus(QString)));
    connect(dev, SIGNAL(TempChanged(double)), SIGNAL(TempChanged(double)));
    connect(dev, SIGNAL(liveMagicFailed(DeviceIndex)), SIGNAL(liveMagicFailed(DeviceIndex)));
}

void EthFlashDev::updatedHwVer(const DeviceDescription &dd)
{
    if(dd.getIndex() != DeviceIndex(dev->getDeviceId(), dev->getDeviceSerialNumber()))
            return;
    bool hwMajVerOk=false;
    quint16 ver = dd.hw_str.left(2).toUShort(&hwMajVerOk);
    if(hwMajVerOk)
        hw_maj_ver = ver;
}

void EthFlashDev::setDeviceEnabled(const DeviceIndex &index, bool en)
{
    if(dev->getDeviceIndex() != index)
        return;

    enabled = en;
    if(writeInProgress && !enabled)
        writeOk = false;
}

void EthFlashDev::startDevWrite(uint id, const QByteArray &fileData, int headerSize, bool isGoldenImage, bool verify)
{
    writeInProgress = true;
    bool ok = startWrite(id, fileData.constData(), static_cast<uint32_t>(fileData.size()), headerSize, isGoldenImage, verify);
    writeInProgress = false;
    emit taskFinished(dev->getDeviceIndex(), id, true, ok, isGoldenImage);
}

void EthFlashDev::startDevVerify(uint id, const QByteArray &fileData, int headerSize, bool isGoldenImage)
{
    fileHeaderSize = headerSize;
    fullSize = static_cast<uint32_t>(fileData.size());
    writeInProgress = true;
    bool ok = verifyFlash(isGoldenImage, fullSize, fileData.constData());
    writeInProgress = false;
    emit taskFinished(dev->getDeviceIndex(), id, false, ok, isGoldenImage);
}

void EthFlashDev::startDevErase(uint id, bool isGoldenImage)
{
    writeInProgress = true;
    bool ok = sectorErase(isGoldenImage, getImageOffset(1));
    writeInProgress = false;
    emit taskFinished(dev->getDeviceIndex(), id, false, ok, isGoldenImage);
}

void EthFlashDev::on_devConnected()
{
    configureFlash();

    emit devConnected(dev->getDeviceIndex(),
                      dev->getFirmwareVersionStr(),
                      dev->getDeviceFwVersion());
}

void EthFlashDev::disconnectFromHost(const DeviceIndex &index)
{
    if(index.isNull() || (index==dev->getDeviceIndex()) ) {
        QHostAddress host;
        dev->setHostAddress(host);
    }
}

bool EthFlashDev::sectorErase(bool isGoldenImage, uint32_t sizeToErase)
{
    emit progressFormatChanged(dev->getDeviceIndex(), "Erased %p%");
    return FlashDev::sectorErase(isGoldenImage, sizeToErase);
}

bool EthFlashDev::writeFlash(const char *fileData, bool isGoldenImage)
{
    emit progressFormatChanged(dev->getDeviceIndex(), "Wrote %p%");
    return FlashDev::writeFlash(fileData, isGoldenImage);
}

bool EthFlashDev::verifyFlash(bool isGoldenImage, uint sizeToRead, const char *data)
{
    emit progressFormatChanged(dev->getDeviceIndex(), "Verify %p%");
    return FlashDev::verifyFlash(isGoldenImage, sizeToRead, data);
}

bool EthFlashDev::isConnected() const
{
    return dev->isConnected();
}

quint16 EthFlashDev::devReadReg(int regAddr)
{
    return dev->regRead(regAddr);
}

void EthFlashDev::devWriteReg(int regAddr, quint16 data)
{
    dev->regWrite(regAddr, data);
}

void EthFlashDev::devWriteReg32(int regAddr, quint32 data)
{
    dev->regWrite32(regAddr, data);
}

bool EthFlashDev::devBlkWriteReg(uint count, const uint *addr, uint16_t *data)
{
    RegOpVector opVec;
    uint written = 0;
    while(written != count) {
        uint sizeToWrite = qMin(count-written, MAX_PART_SIZE);
        for(uint w=0; w<sizeToWrite; ++w)
            opVec << RegWrite(addr[written+w], data[written+w]);

        RegOpVector opRes = dev->regOpExec(opVec);

        written += sizeToWrite;
        if(opRes.size() != opVec.size())
            return false;
    }
    return true;
}

bool EthFlashDev::devBlkReadReg(uint count, const uint *addr, uint16_t *data)
{
    RegOpVector opVec;
    uint read = 0;
    while(read != count) {
        uint sizeToRead = qMin(count-read, MAX_PART_SIZE);
        for(uint w=0; w<sizeToRead; ++w)
            opVec << RegRead(addr[read+w]);

        RegOpVector opRes = dev->regOpExec(opVec);

        if(opRes.size() != opVec.size())
            return false;

        for(uint w=0; w<sizeToRead; ++w)
            data[read+w] = opRes[w].data;
        read += sizeToRead;
    }
    return true;
}

void EthFlashDev::sendProgress(int p)
{
    emit progressUpdated(dev->getDeviceIndex(), p);
}

std::string EthFlashDev::getPrefix() const
{
    return (dev->getDeviceIdStr() +" "+ dev->getSerialIdStr()).toStdString();
}

uint EthFlashDev::getRegBase() const
{
    return 0x7000;
}

bool EthFlashDev::canReadFlashId() const
{
    return dev->getDeviceFwVersion() < REV_FLASH_MIN
        || dev->getDeviceFwVersion() >= REV_FLASH_ID;
}

bool EthFlashDev::is_ADC64VE_module() const
{
    return dev->getDeviceId() == DEVICE_ID_ADC64VE
            || dev->getDeviceId() == DEVICE_ID_ADC64VE_XGE;
}
