/*
**    Copyright 2017-2020 Ivan Filippov, Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*/

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <math.h>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QFileDialog>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QStatusBar>
#include <QSettings>
#include <QThread>
#include <QTimer>
#include <QProgressBar>

#include "EthFlashDev.h"
#include "flash-prog/DataBase.h"
#include "device-discover/DiscoverDialog.h"
#include "mregdev/device_id.h"
#include "base/AboutBox.h"

namespace {
enum {COL_ID, COL_SERIAL, COL_NAME, COL_EN, COL_STATE, COL_FIRMWARE, COL_TEMP, COL_PROGRESS, COL_NUM};
enum TASK_ID {TASK_ERASE, TASK_WRITE, TASK_VERIFY};
constexpr fw_version_t REV_ETHERNET_WRITE = fw_version_t(1, 0, 24200);
}

Q_DECLARE_METATYPE(fw_version_t);

MainWindow::MainWindow(QWidget *parent):
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    connectedState(false),
    discoverDialog(new DiscoverDialog(this)),
    testValue(0),
    tempLabel(new QLabel(this)),
    fwLabel(new QLabel(this)),
    lastStatusLabel(new QLabel(this)),
    dataBase(new DataBase(this)),
    fileHeaderSize(-1),
    fileSize(-1),
    cmdPerformed(0),
    taskId(100)
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<Config>();

    ui->setupUi(this);
    ui->mainToolBar->hide();
    connect(ui->actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));
    setWindowTitle(QCoreApplication::applicationName());

    initTable();
    readConfig();

    QVector<DeviceDescription> deviceList;
    for(DeviceIndex index : config.devices) {
        DeviceDescription dd(index.getDevId(), index.getSerial());
        dd.enabled = true;
        deviceList.push_back(dd);

        addDevice(dd);
    }
    if(!deviceList.isEmpty())
        discoverDialog->setSelected(deviceList);

    if(!config.filePath.isEmpty()) {
        ui->labelFile->setText(config.filePath);
        readFile();
    }
    ui->actionVerify_after_write->blockSignals(true);
    ui->actionVerify_after_write->setChecked(config.verifyAfterWrite);
    ui->actionVerify_after_write->blockSignals(false);
//    write65k();
    statusBar()->addPermanentWidget(lastStatusLabel);
    statusBar()->addPermanentWidget(tempLabel);
    statusBar()->addPermanentWidget(fwLabel);
}

MainWindow::~MainWindow()
{
    saveConfig();

    disconnectFromAll();
    delete ui;
}

void MainWindow::on_actionWrite_Golden_Image_triggered()
{
    QMessageBox msgBox;
    msgBox.setText("Do you want to write \"Golden Image\"?");
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if(ret == QMessageBox::Yes)
        startWrite(true);
}

void MainWindow::on_actionGet_record_logs_triggered()
{
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    QSet<int> rows;
    for(QTableWidgetSelectionRange range : ranges) {
        for(int row = range.bottomRow(); row <= range.bottomRow(); ++row) {
            rows << row;
        }
    }
    QList<DeviceIndex> devIndexes;
    for(int r : rows) {
        devIndexes.append(getIndexByRow(r));
    }
    if(devIndexes.isEmpty())
        devIndexes = devices.keys();
    for(DeviceIndex i : devIndexes) {
        qInfo().noquote() << QString("\t=== Log for %1 ===").arg(i.getSerialStr());
        dataBase->readInfo(i.getSerial());
    }
}

void MainWindow::on_actionVerify_after_write_toggled(bool verify)
{
    config.verifyAfterWrite = verify;
    config.saveSettings();
}

void MainWindow::on_actionSend_reboot_signal_triggered()
{
    QMessageBox msgBox;
    msgBox.setText("Will be implemented soon.");
    msgBox.setStandardButtons(QMessageBox::Ok);
    msgBox.exec();
}

void MainWindow::on_actionVerify_Golden_image_triggered()
{
    startVerify(true);
}

void MainWindow::on_actionVerify_regular_image_triggered()
{
    startVerify(false);
}

void MainWindow::on_actionErase_Golden_part_triggered()
{
    startErase(true);
}

void MainWindow::on_actionErase_regular_part_triggered()
{
    startErase(false);
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::on_pushButtonChange_clicked()
{
    if(discoverDialog->exec() == QDialog::Accepted) {
        if(!devices.isEmpty()) {
            disconnectFromAll();
        }
        QVector<DeviceDescription> ddList= discoverDialog->getSelectedDevices();
        config.devices.clear();
        for(DeviceDescription dd : ddList) {
            if(dd.enabled) {
                config.devices.push_back(dd.getIndex());
                addDevice(dd);
            }
        }

        config.saveSettings();
    }
}

void MainWindow::on_pushButtonFile_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this,
        tr("Open Bit file"), config.filePath, tr("Bit Files (*.bit)"));
    if(!fileName.isEmpty()) {
        fileHeaderSize = -1;
        fileSize = -1;
        config.filePath = fileName;
        ui->labelFile->setText(fileName);
        saveConfig();
        readFile();
    }
    updateWriteButtonState();
}

void MainWindow::on_pushButtonWrite_clicked()
{
    startWrite();
}

void MainWindow::deviceConnected(const DeviceIndex &index,
                                 QString fwStr, const fw_version_t &fwVersion)
{
    const bool supported = fwVersion >= REV_ETHERNET_WRITE;
    const int row = getRowByIndex(index);
    if(row != -1) {
        ui->tableWidget->item(row, COL_NAME)->setTextColor(Qt::darkGreen);
        ui->tableWidget->item(row, COL_FIRMWARE)->setText(fwStr);
        ui->tableWidget->item(row, COL_FIRMWARE)->setTextColor(supported ? Qt::black : Qt::red);
    }
    devOk[index] = true; // supported;
    updateWriteButtonState();
    updateTableSize();
}

void MainWindow::deviceDisconnected()
{
    DeviceIndex index = getIndexBySender(sender());
    if(index.isNull()) {
        qWarning() << "Unknown device disconnect";
        return;
    }
    const int row = getRowByIndex(index);
    if(row != -1) {
        ui->tableWidget->item(row, COL_NAME)->setTextColor(Qt::red);
    }
    devOk[index] = false;
    updateWriteButtonState();
    updateTableSize();
}

void MainWindow::deviceTempChanged(double temp)
{
    const int row = getRowBySender(sender());
    if(row != -1) {
        QString tempStr = QString("%1 %2C").arg(temp, 0, 'f', 1).arg(QChar(0x02da));
        ui->tableWidget->item(row, COL_TEMP)->setText(tempStr);
    }
    updateTableSize();
}

void MainWindow::deviceStatusUpdated(QString s)
{
    DeviceIndex index = getIndexBySender(sender());
    const int row = getRowByIndex(index);
    if(row != -1 && !devLiveMagicErr.contains(index)) {
        ui->tableWidget->item(row, COL_STATE)->setText(s);
    }
    updateTableSize();
}

void MainWindow::progressFormatChanged(DeviceIndex index, QString format)
{
    const int row = getRowByIndex(index);
    if(row != -1) {
        QProgressBar *bar = qobject_cast<QProgressBar*>(ui->tableWidget->cellWidget(row, COL_PROGRESS));
        if(bar) {
            bar->setFormat(format);
            updateTableSize();
        }
    }
}

void MainWindow::progressUpdated(DeviceIndex index, int p)
{
    const int row = getRowByIndex(index);
    if(row != -1) {
        QProgressBar *bar = qobject_cast<QProgressBar*>(ui->tableWidget->cellWidget(row, COL_PROGRESS));
        if(bar && bar->value() != p) {
            bar->setValue(p);
            updateTableSize();
        }
    }
}

void MainWindow::taskFinished(DeviceIndex index, uint id, bool isWrite, bool ok, bool isGoldenImage)
{
    qInfo() << QString("%1 at %2 taskId:%3 status:%4")
               .arg(isWrite ? "dataWrote" : "dataVerify/secorErase")
               .arg(index.getIdent()).arg(id).arg(ok?"success":"fail");
    if(ok) {
        devTaskId[index] = id;
        if(isWrite)
            dataBase->writeInfo(index.getSerial(), isGoldenImage, config.filePath);
    } else {
        devTaskId[index] = 0;
    }

    updateTableSize();
}

void MainWindow::liveMagicFailed(DeviceIndex index)
{
    int row = getRowByIndex(index);
    if(row == -1)
        return;
    QWidget *w = ui->tableWidget->cellWidget(row, COL_EN);
    QCheckBox *cb = qobject_cast<QCheckBox*>(w);
    if(cb == NULL)
        return;
    ui->tableWidget->item(row, COL_STATE)->setText("LiveMagic error");
    updateTableSize();
    devLiveMagicErr.insert(index);
    cb->setChecked(false);
    devEn[index] = false;
    emit setDeviceEnabled(index, devEn[index]);

    qWarning() << QString("Disable device %1 due to liveMagic error").arg(index.getIdent());
}

void MainWindow::deviceEnabled(bool st)
{
    QTableWidget *t = ui->tableWidget;
    QObject *cb = sender();
    int row=t->rowCount()-1;
    for(; row>=0; --row){
        if(t->cellWidget(row, COL_EN) == cb)
            break;
    }

    if(row==-1)
        return;

    const DeviceIndex index = getIndexByRow(row);
    EthFlashDev *dev = devices[index];
    if(index.isNull())
        return;
    devEn[index] = st;
    emit setDeviceEnabled(index, devEn[index]);
    if(st) {
        devLiveMagicErr.remove(index);
        ui->tableWidget->item(row, COL_STATE)->setText("Connecting...");
        updateTableSize();
        connect(discoverDialog, SIGNAL(deviceDiscovered(DeviceDescription)),
                dev, SIGNAL(deviceUpdated(DeviceDescription)));
    } else {
        disconnect(discoverDialog, SIGNAL(deviceDiscovered(DeviceDescription)),
                   dev, SIGNAL(deviceUpdated(DeviceDescription)));
        emit disconnectFromHost(index);
    }
}

void MainWindow::startWrite(bool isGoldenImage)
{
    startTask(TASK_WRITE, isGoldenImage);
}

void MainWindow::startVerify(bool isGoldenImage)
{
    startTask(TASK_VERIFY, isGoldenImage);
}

void MainWindow::startErase(bool isGoldenImage)
{
    QMessageBox msgBox;
    msgBox.setText(QString("Do you want to erase \"%1 Image\"?")
                   .arg(isGoldenImage ? "Golden" : "Regular"));
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if(ret == QMessageBox::Yes)
        startTask(TASK_ERASE, isGoldenImage);
}

void MainWindow::startTask(int task, bool isGoldenImage)
{
    setLabelStatus("in progress...", "gray");

    QString taskStr;
    qint64 taskTimeout;
    switch (task) {
    case TASK_ERASE:
        taskStr = "erase";
        taskTimeout = 5*60*1000;
        break;
    case TASK_WRITE:
        taskStr = "write";
        taskTimeout = 10*60*1000;
        break;
    case TASK_VERIFY:
        taskStr = "verify";
        taskTimeout = 3*60*1000;
        break;
    default:
        qFatal("Unexpected TASK_ID: %d",task);
    }
    qInfo().noquote() << QString("\n=== Start %1 ===").arg(taskStr);
    qInfo().noquote() << QString("%1 %2 image %3").arg(QDateTime::currentDateTime().toString())
               .arg(isGoldenImage?"golden": "second").arg(taskStr);
    setPushButtonEnabled(false);

    // Read data from file
    if(!readFile())
        return;
    cmdPerformed = 0;
    bool allOk=true;

    ++taskId;
    if(taskId==0)
        taskId=100;
    devTaskId.clear();
    QSet<DeviceIndex> deviceForTask;
    for (DeviceIndex i : devices.keys())
        if(devEn[i])
            deviceForTask << i;

    switch (task) {
    case TASK_ERASE:
        emit startDevErase(taskId, isGoldenImage);
        break;
    case TASK_WRITE:
        emit startDevWrite(taskId, fileData, fileHeaderSize, isGoldenImage, config.verifyAfterWrite);
        break;
    case TASK_VERIFY:
        emit startDevVerify(taskId, fileData, fileHeaderSize, isGoldenImage);
        break;
    }

    QElapsedTimer taskTimer;
    taskTimer.start();
    bool finished = false;
    bool someFailed = false;
    while(!finished) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 500);
        if(taskTimer.elapsed() > taskTimeout) {
            qWarning() << QString("%1 timeouted").arg(taskStr);
            break;
        }
        finished = true;
        for (DeviceIndex index : deviceForTask) {
            if(!devTaskId.contains(index) || (devTaskId[index] != taskId && devTaskId[index] != 0)){
                finished = false;
                break;
            }
            if(devTaskId[index] == 0)
                someFailed = true;
        }
    }
    allOk &= finished && !someFailed;

    qInfo().noquote() << "==="<<(allOk?"end":"fail")<<"===";
    setLabelStatus(QString("%1 (%2)")
                   .arg(allOk?QString("ok in %1 sec").arg(taskTimer.elapsed()/1000):"fail")
                   .arg(QTime::currentTime().toString()),
                   allOk?"green":"red");
    updateTableSize();
    setPushButtonEnabled(true);
}

void MainWindow::initTable()
{
    QTableWidget *t = ui->tableWidget;
    t->verticalHeader()->hide();
    t->setColumnCount(COL_NUM);
    t->setHorizontalHeaderItem(COL_ID, new QTableWidgetItem("Id"));
    t->setHorizontalHeaderItem(COL_SERIAL, new QTableWidgetItem("Serial"));
    t->setHorizontalHeaderItem(COL_NAME, new QTableWidgetItem("Name"));
    t->setHorizontalHeaderItem(COL_EN, new QTableWidgetItem("En"));
    t->setHorizontalHeaderItem(COL_STATE, new QTableWidgetItem("State"));
    t->setHorizontalHeaderItem(COL_FIRMWARE, new QTableWidgetItem("f/w"));
    t->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    t->setHorizontalHeaderItem(COL_PROGRESS, new QTableWidgetItem("Progress"));
    t->hideColumn(COL_ID);
    t->hideColumn(COL_SERIAL);

    t->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    t->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize();
}

void MainWindow::updateTableSize()
{
    QTableWidget *table = ui->tableWidget;
    table->resizeRowsToContents();
     table->resizeColumnsToContents();

     int width = table->verticalHeader()->width()+5;
     for(int col=0; col<table->columnCount(); ++col) {
         if(table->isColumnHidden(col))
             continue;
         width += table->columnWidth(col);
     }
     if(width<20)
         width = 20;

     int height = table->horizontalHeader()->height()+5;
     for(int row=0; row<table->rowCount(); ++row) {
         if(table->isRowHidden(row))
             continue;
         height += table->rowHeight(row);
     }
     if(height<20)
         height = 20;
     table->setFixedSize(width, height);
     table->updateGeometry();
}

DeviceIndex MainWindow::getIndexBySender(QObject *sender) const
{
    return devices.key(reinterpret_cast<EthFlashDev *>(sender), DeviceIndex());
}

DeviceIndex MainWindow::getIndexByRow(int row) const
{
    QTableWidget *t = ui->tableWidget;
    bool ok;
    quint16 id = t->item(row, COL_ID)->text().toUShort(&ok, 16);
    quint32 ser;
    if(ok) {
        ser = t->item(row, COL_SERIAL)->text().toUInt(&ok, 16);
    }
    if(ok)
        return DeviceIndex(id, ser);
    else
        return DeviceIndex();
}

int MainWindow::getRowByIndex(DeviceIndex index) const
{
    QTableWidget *t = ui->tableWidget;

    for(int row=0; row<t->rowCount(); ++row){
        if(getIndexByRow(row) == index)
            return row;
    }
    return -1;
}

int MainWindow::getRowBySender(QObject *sender) const
{
    DeviceIndex index = getIndexBySender(sender);
    if(index.isNull())
        return -1;
    else
        return getRowByIndex(index);
}

void MainWindow::setPushButtonEnabled(bool enabled) const
{
    ui->pushButtonFile->setEnabled(enabled);
    ui->pushButtonChange->setEnabled(enabled);
    ui->actionVerify_after_write->setEnabled(enabled);
    ui->actionVerify_Golden_image->setEnabled(enabled);
    ui->actionVerify_regular_image->setEnabled(enabled);
    if(enabled) {
        updateWriteButtonState();
    } else {
        ui->pushButtonWrite->setEnabled(enabled);
        ui->actionWrite_Golden_Image->setEnabled(enabled);
    }
}

void MainWindow::setLabelStatus(QString msg, QString color) const
{
    lastStatusLabel->setText(msg);
    lastStatusLabel->setStyleSheet(QString("QLabel { color : %1; }").arg(color));
}

void MainWindow::deviceUpdated(const DeviceDescription &dd)
{
    const DeviceIndex &index = dd.getIndex();
    if(!dd.enabled) {
        return;
    }

    QTableWidget *table = ui->tableWidget;
    int row = getRowByIndex(index);
    if(row==-1) {
        row = table->rowCount();
        table->setRowCount(row+1);
        table->setItem(row, COL_ID, new QTableWidgetItem(QString::number(index.getDevId(), 16)));
        table->setItem(row, COL_SERIAL, new QTableWidgetItem(QString::number(index.getSerial(), 16)));
        table->setItem(row, COL_NAME, new QTableWidgetItem(QString("%1 %2")
                                                           .arg(getDeviceTypeName(dd.device_id),
                                                                dd.getSerialIdStr())));
        QCheckBox *cb = new QCheckBox(table);
        cb->setChecked(true);
        devEn[index] = true;
        table->setCellWidget(row, COL_EN, cb);
        connect(cb, SIGNAL(toggled(bool)), this, SLOT(deviceEnabled(bool)));
        table->setItem(row, COL_FIRMWARE, new QTableWidgetItem());
        table->setItem(row, COL_STATE, new QTableWidgetItem());
        table->setItem(row, COL_TEMP, new QTableWidgetItem());
        QProgressBar *bar = new QProgressBar(table);
        bar->setMinimumWidth(111);
        table->setCellWidget(row, COL_PROGRESS, bar);
        bar->setValue(0);
    }
}

void MainWindow::addDevice(const DeviceDescription &dd)
{
    deviceUpdated(dd);
    EthFlashDev *dev = new EthFlashDev(dd.device_id, dd.serial_id);
    devices[dd.getIndex()] = dev;
    QThread *thread = new QThread(this);
    deviceThreads[dd.getIndex()] = thread;
    dev->moveToThread(thread);
    thread->start();
    thread->setObjectName(QString(dd.getIndex().getSerialStr()));

    connect(discoverDialog, SIGNAL(deviceDiscovered(DeviceDescription)),
            dev, SIGNAL(deviceUpdated(DeviceDescription)));

    qRegisterMetaType<fw_version_t>();
    connect(dev, &EthFlashDev::devConnected,
            this, &MainWindow::deviceConnected);
    connect(dev, SIGNAL(deviceDisconnected()), SLOT(deviceDisconnected()));
    connect(dev, SIGNAL(setStatus(QString)), SLOT(deviceStatusUpdated(QString)));
    connect(dev, SIGNAL(TempChanged(double)), SLOT(deviceTempChanged(double)));
    connect(this, SIGNAL(closeDevices()), dev, SLOT(deleteLater()));
    connect(dev, SIGNAL(destroyed(QObject*)), thread, SLOT(quit()));
    connect(dev, SIGNAL(liveMagicFailed(DeviceIndex)),
            SLOT(liveMagicFailed(DeviceIndex)));
    connect(this, SIGNAL(disconnectFromHost(DeviceIndex)),
            dev, SLOT(disconnectFromHost(DeviceIndex)));

    connect(this, SIGNAL(startDevWrite(uint,QByteArray,int,bool,bool)),
            dev, SLOT(startDevWrite(uint,QByteArray,int,bool,bool)));
    connect(this, SIGNAL(startDevVerify(uint,QByteArray,int,bool)),
            dev, SLOT(startDevVerify(uint,QByteArray,int,bool)));
    connect(this, SIGNAL(startDevErase(uint,bool)),
            dev, SLOT(startDevErase(uint,bool)));
    connect(dev, SIGNAL(progressFormatChanged(DeviceIndex, QString)),
            SLOT(progressFormatChanged(DeviceIndex, QString)));
    connect(dev, SIGNAL(progressUpdated(DeviceIndex, int)),
            SLOT(progressUpdated(DeviceIndex, int)));
    connect(dev, SIGNAL(taskFinished(DeviceIndex,uint,bool,bool,bool)),
            SLOT(taskFinished(DeviceIndex,uint,bool,bool,bool)));
    connect(this, SIGNAL(setDeviceEnabled(DeviceIndex,bool)),
            dev, SLOT(setDeviceEnabled(DeviceIndex,bool)));

    QHostAddress host(dd.ip_addr);
    qInfo() << QString("Added device %1 at %2")
               .arg(dd.getSerialIdStr())
               .arg(host.toString());
    updateTableSize();
}

void MainWindow::disconnectFromAll()
{
    emit closeDevices();

    QElapsedTimer timer;
    timer.start();
    while(true) {
        if(timer.elapsed()>2*60000){
            qWarning() << "Timeout in disconnectFromAll";
            break;
        }
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);

        bool ok=true;
        for(DeviceIndex index : devices.keys()) {
            QThread *thread = deviceThreads[index];
            if(thread->isRunning()){
                ok = false;
                break;
            }
        }
        if(ok)
            break;
    }

    devices.clear();
    for(DeviceIndex index : devices.keys()) {
        QThread *thread = deviceThreads[index];
        delete thread;
    }
    deviceThreads.clear();
    ui->tableWidget->setRowCount(0);
}

void MainWindow::readConfig()
{
    config.loadSettings();

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.endGroup();
}

void MainWindow::saveConfig()
{
    config.saveSettings();

    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();
}

void MainWindow::updateWriteButtonState() const
{
    bool enable = true;
    bool hasEnabledDevice = false;
    for(DeviceIndex index : devices.keys()) {
        if(!devEn[index])
            continue;
        if(!devOk.contains(index) || !devOk[index]) {
            enable = false;
            break;
        }
        hasEnabledDevice = true;
    }
    enable &= hasEnabledDevice;

    if(enable) {
        QFile file(config.filePath);
        enable = file.open(QIODevice::ReadOnly);
        file.close();
    }
    if(enable) {
        enable = fileHeaderSize!=-1;
    }
    ui->pushButtonWrite->setEnabled(enable);
    ui->actionWrite_Golden_Image->setEnabled(enable);
}

void MainWindow::write65k()
{
    QFile file("file.dat");
    file.open(QIODevice::WriteOnly);
    QDataStream out(&file);
    quint16 val=0;
    quint32 cnt=0;
    while(true){
        out << val;
        val++;
        cnt++;
        if(cnt>=65*1024/2)
            break;
    }

    file.close();
    exit(1);
}

bool MainWindow::readFile()
{
    ui->labelFile->setStyleSheet("QLabel { color : black; }");
    QFile file(config.filePath);
    QFileInfo fileInfo(file);
    if(fileInfo.lastModified() == fileModified && fileInfo.size() == fileSize
            && fileData.size() == fileSize
            && fileHeaderSize>0
            && fileSize>fileHeaderSize) {
        qInfo() << "File wasn't modified. Use cached file data.";
        return true;
    }
    if(!fileInfo.exists()
            && fileData.size() == fileSize
            && fileHeaderSize>0
            && fileSize>fileHeaderSize
            ) {
        qWarning() << "File is no longer available, but we have a cache.";
        return true;
    }

    fileModified = fileInfo.lastModified();
    fileSize = fileInfo.size();
    fileHeaderSize = -1;

    if(!file.open(QIODevice::ReadOnly)){
            qWarning() << "Can't open bit file";
            return false;
    }
    fileData = file.readAll();
    file.close();
    if(fileData.isEmpty()) {
        qWarning() << "Read file is empty";
        return false;
    }
    if(fileData.size() != fileSize) {
        qWarning() << "data.size() != fileSize";
        return false;
    }

    qInfo() << QString("Read bit file '%1' (%2 bytes)").arg(config.filePath).arg(fileData.size());

    fileHeaderSize = FlashDev::findBitFileHeader(fileData.data(), fileData.size());
    ui->labelFile->setStyleSheet(QString("QLabel { color : %1; }")
                                 .arg((fileHeaderSize!=-1)?"green":"red"));
    if(fileHeaderSize == -1) {
        qWarning() << "Can't find header size of bit-file";
        return false;
    }

    qInfo() << "File header takes"<<fileHeaderSize<<"bytes of bit-file";
    return true;
}
