#include "ClientManager.h"
#include "params.h"
#include "Analizer.h"
#include "RingBuffer.h"
#include "TcpClient.h"
#include "base/threadQuit.h"
#include "base/RcRunId.h"

#include <QCoreApplication>
#include <QElapsedTimer>
#include <QThread>

ClientManager::ClientManager(int _index, int _evNumIndex, bool readout, int preferedBufSize, QObject *parent) :
    QObject(parent),
    index(_index),
    evNumIndex(_evNumIndex),
    dataRingBuff( new RingBuffer<char>(preferedBufSize)),
    evRingBuff(new RingBuffer<EventMetaInfo>(EVENT_RING_SIZE)),
    tcpClientThread(new QThread(this)),
    analizerThread(new QThread(this)),
    tcpClient(new TcpClient(index, dataRingBuff)),
    analizer(new Analizer(index, dataRingBuff, evRingBuff)),
    isAnalizerRunning(false),
    isTcpClientRunning(false),
    isAnalizerResetting(false),
    isTcpClientResetting(false),
    readoutEnabled(readout),
    gotError(false),
    curStateIndex(0)
{
    qRegisterMetaType<ProgramInterface> ();
    tcpClientThread->setObjectName(QString("tcpClient #%1").arg(index));
    analizerThread->setObjectName(QString("analizer #%1").arg(index));
    tcpClient->moveToThread(tcpClientThread);
    tcpClientThread->start();

    analizerThread->start();
    analizer->moveToThread(analizerThread);

    connect(this, &ClientManager::stopReadOut, tcpClient, &TcpClient::stopReadOut);
    connect(this, &ClientManager::startReadOut, tcpClient, &TcpClient::startReadOut);

    connect(this, &ClientManager::starting, analizer, &Analizer::startAnalize);
    connect(this, &ClientManager::starting, tcpClient, &TcpClient::startRun);

    connect(this, &ClientManager::stoping, analizer, &Analizer::stopAnalize);
    connect(this, &ClientManager::stoping, tcpClient, &TcpClient::stopRun);

    connect(this, &ClientManager::resetting, analizer, &Analizer::resetAnalize);
    connect(this, &ClientManager::resetting, tcpClient, &TcpClient::resetReadOut);

    connect(analizer, &Analizer::updateRunState, this, &ClientManager::updateRunState);
    connect(tcpClient, &TcpClient::updateRunState, this, &ClientManager::updateRunState);

    connect(analizer, &Analizer::resetted, this, &ClientManager::onResetted);
    connect(tcpClient, &TcpClient::resetted, this, &ClientManager::onResetted);

    connect(tcpClient, &TcpClient::sendStatus, this, &ClientManager::sendStatus);
    connect(tcpClient, &TcpClient::sendStatistic, this, &ClientManager::sendStatistic);

    connect(this, &ClientManager::destroyingClient, analizer, &Analizer::deleteLater);
    connect(this, &ClientManager::destroyingClient, tcpClient, &TcpClient::deleteLater);

    connect(analizer, &Analizer::destroyed, analizerThread, &QThread::quit);
    connect(tcpClient, &TcpClient::destroyed, tcpClientThread, &QThread::quit);

    connect(this, QOverload<const ProgramDescription &, const ProgramInterface &>::of(&ClientManager::setTcpAddress),
            tcpClient, &TcpClient::setAddress);

    connect(this, &ClientManager::clientDescrChanged, analizer, &Analizer::clientDescrChanged);
    connect(this, &ClientManager::clientDescrChanged, tcpClient, &TcpClient::clientDescrChanged);

    connect(analizer, &Analizer::statUpdated, this, &ClientManager::statUpdated);
    connect(analizer, &Analizer::statUpdated, tcpClient, &TcpClient::analyzerStatUpdated);

    connect(tcpClient, &TcpClient::inputBufferSizeUpdated, this, &ClientManager::inputBufferSizeUpdated);
}

ClientManager::~ClientManager()
{
    emit destroyingClient();
    QElapsedTimer timer;
    timer.start();

    if(!threadQuiting(analizerThread, 1000, &timer))
        qWarning()<<"Analizer's thread isn't stopped for client #"<<index;

    if(!threadQuiting(tcpClientThread, 1000, &timer))
        qWarning()<<"TcpClient's thread isn't stopped for client #"<<index;

    delete dataRingBuff;
    delete evRingBuff;
}

void ClientManager::setProgDescription(const ProgramDescription &descr)
{
    if(progDescription.getName() != descr.getName())
        emit clientDescrChanged(descr.getName());

    progDescription = descr;
    ProgramInterface newInterface;
    for(const ProgramInterface &interface : descr.interfaces) {
        if(interface.type == ProgramInterface::getTypeName(ProgramInterfaceDataFlow)
                && interface.id == evNumIndex){
            newInterface = interface;
            break;
        }
    }
//    qDebug() << QString("[ClientMgr #%1] connect to interface '%2' (id:%3)").arg(index)
//                .arg(newInterface.type).arg(newInterface.id);
    if(!newInterface.host.isNull() && newInterface.port)
        emit setTcpAddress(descr, newInterface);
    tcpInterface = newInterface;
}

void ClientManager::startClient(uint stateIndex)
{
    curStateIndex = stateIndex;
    gotError=false;
    if(!readoutEnabled){
        emit started(index, true, curStateIndex);
        return;
    }
    emit starting(curStateIndex);
    QElapsedTimer timer;
    timer.start();
    while(!timer.hasExpired(2000)) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
        if(isAnalizerRunning && isTcpClientRunning)
            break;
    }
    const QString suffix = QString(" is still stopped, clIndex=%1").arg(index);
    if(!isAnalizerRunning)
        qWarning() << ("Analizer"+suffix);
    if(!isTcpClientRunning)
        qWarning() << ("TcpClient"+suffix);
}

void ClientManager::stopClient(uint stateIndex)
{
    curStateIndex = stateIndex;
    if(!readoutEnabled){
        emit stopped(index,curStateIndex);
        return;
    }
    emit stoping(curStateIndex);
    QElapsedTimer timer;
    timer.start();
    while(!timer.hasExpired(2000)){
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
        if(!isAnalizerRunning && !isTcpClientRunning)
            break;
    }
    const QString suffix = QString(" is still running, clIndex=%1").arg(index);
    if(isAnalizerRunning)
        qWarning() << ("Analizer"+suffix);
    if(isTcpClientRunning)
        qWarning() << ("TcpClient"+suffix);
}

void ClientManager::resetClient(const RcRunId &runId, uint stateIndex)
{
    curStateIndex = stateIndex;
    if(!readoutEnabled){
        emit started(index, true, curStateIndex);
        return;
    }
    qDebug() << QString("ClientManager index #%1 start resetting (run %2)")
                .arg(index).arg(QString(runId));
    isAnalizerResetting = true;
    isTcpClientResetting = true;
    emit resetting(curStateIndex);
}

void ClientManager::destroyClient(int i)
{
    if(i!=0 && index!=i)
        return;
    deleteLater();
}

void ClientManager::updateClientReadout(int i, bool readout)
{
    if(i!=0 && index!=i)
        return;
    readoutEnabled = readout;
    if(readoutEnabled){
        emit startReadOut();
    } else
        emit stopReadOut();
}

void ClientManager::updateRunState(bool isAnalizer, bool isInRunMode, uint stateIndex)
{
    if(curStateIndex != stateIndex){
        qWarning()<<QString("%1 (id:%2) send wrong stateIndex(%3; curIndex=%4) on state update")
                    .arg(progDescription.getName()).arg(index)
                    .arg(stateIndex).arg(curStateIndex);
        return;
    }
    qDebug()<<(isAnalizer?"Analizer":"TcpClient")<<index<<(isInRunMode?"started":"stopped");
    if(isAnalizer){
        isAnalizerRunning = isInRunMode;
    } else {
        isTcpClientRunning = isInRunMode;
    }
    if(gotError)
        return;

    if(isTcpClientRunning == isAnalizerRunning) {
        if(isInRunMode)
            emit started(index, true, curStateIndex);
        else
            emit stopped(index, curStateIndex);
    }
}

void ClientManager::onResetted(bool isAnalizer,uint stateIndex)
{
    if(curStateIndex != stateIndex){
        qWarning()<<QString("%1 (id:%2) send wrong stateIndex(%3; curIndex=%4) on reset")
                    .arg(progDescription.getName()).arg(index)
                    .arg(stateIndex).arg(curStateIndex);
        return;
    }
    if(isAnalizer)
        isAnalizerResetting = false;
    else
        isTcpClientResetting = false;
    if(!isTcpClientResetting && !isAnalizerResetting){
        qDebug()<<"ClientManager index #"<<index<<" resetting completed";
        emit resetted(index, curStateIndex);
    }
}

void ClientManager::onTcpError(QString err)
{
    if(!gotError){
        gotError=true;
        emit started(index, false, curStateIndex);
    }
}
