#ifndef COMBINER_H
#define COMBINER_H

#include <EventMetaInfo.h>
#include "params.h"
#include "EvBConfig.h"
#include "FileWriter.h"
#include "OutputTcpServer.h"
#include "lib-common/MpdRawTypes.h"
#include "base/RcRunId.h"

#include <QMap>
#include <QObject>
#include <QSet>
#include <QVector>

class Analizer;
class FileWriter;
template<class T> class RingBuffer;

class QSemaphore;
class QTimer;
class OutputTcpServer;

typedef QPair<quint64, quint64> EventsRange;
typedef QMap<quint64, EventsRange> ClientTrashStat;

struct CombinerStats {
    CombinerStats() { clear(); newPeriod(); }
    void clear() {
        eob = 0;
        regularEvents = 0;
        fullEvents = 0;
        partedEvents = 0;
        skipped = 0;
        monitorEvCount = 0;
        missedFileSize = 0;
        lastEventClients.clear();
//        lastEventNum.clear();
        for (int key : lastEventNum.keys())
            lastEventNum[key] = 0;
        for (int key : flushedSize.keys())
            flushedSize[key] = 0;
        for (int key : dataEvents.keys())
            dataEvents[key] = 0;
        for (int key : trashedEvents.keys())
            trashedEvents[key] = 0;
        clientNumDistr.clear();
        timeDistr.clear();
        timeDistrCnt = 0;
        fwDataPart = 0;
        fwFullSize = 0;
        hasPendingEvent = false;
//        tcpPorts.clear();
//        tcpIsFree.clear();
    }

    bool isNextPeriod(const QElapsedTimer &curT) const {
        return periodCnt>EV_SIZE_PERIOD_MIN_CNT &&
                ( !timeMark.isValid() || (timeMark.msecsTo(curT)>EV_SIZE_PERIOD_MS));
    }

    void newPeriod() {
        periodSize = 0;
        periodCnt = 0;
        timeMark.start();
    }

    void addEvent(quint32 size) {
        periodSize += size;
        ++periodCnt;
    }

    quint64 eob;
    quint64 regularEvents;
    quint64 fullEvents;
    quint64 partedEvents;
    quint64 skipped;
    quint64 expectedEvNum;
    quint64 monitorEvCount;
    quint64 missedFileSize;
    QSet<int> lastEventClients; // set of clientIds in last flushed event
    QMap<int, quint64> lastEventNum; // last EvNum gained by Combiner
    QMap<int, quint64> flushedSize; // summed size of all events gained by Combiner
    QMap<int, quint64> dataEvents; // total events count gained by Combiner
    QMap<int, quint64> trashedEvents; //  number of trashed evnets
    QMap<int, quint64> clientNumDistr; //  [clientNumInEv] = count
                                       // (for 1st level EvB: clientNumInEv=devNumInEv)
    QMap<int, qint64> timeDistr; // [index] time from erliest ev in ms
    QMap<int, qint64> timeDistrFull; // [index] time from decoding till flush
    quint32 timeDistrCnt;
    QMap<int, int> freeEvPart; //  % of free events
    QMap<int, int> preparedEvPart; //  % of prepared events
    QMap<int, int> flushedEvPart; //  % of flushed events
    QMap<int, OutputTcpStatus> tcpStatus;
    double fwDataPart;
    int fwFullSize;
    int fwPrefSize;
    bool hasPendingEvent;

    quint64 periodSize;
    quint64 periodCnt;
    QElapsedTimer timeMark;
    QElapsedTimer timeStamp; // updated on send stat
};

class Combiner : public QObject
{
    Q_OBJECT
public:
    explicit Combiner();
    ~Combiner();
    void initOutputBuffers(int size, bool outToTcp, int outTcpNum);

signals:
    void status(const QString &);
    void statusRaw(const CombinerStats &);
    void started(int index, bool ok, uint stateIndex);
    void stopped(int index, uint stateIndex);
    void resetted(int index, uint stateIndex);
    void error(int index, QString);
    void startTcpServer(int index, bool state) const;
    void deleteTcp(int) const;
    void clearTcp() const;
    void startingFile(const RcRunId &runId, int writerIndex);
    void startingTcpWork();
    void closeFile();
    void closeTcpWork(uint);
    void storageChanged(QString dir);
    void deleteFileWriter();
    void fileWriterStatsUpdated(const FileWriterStats &s);
    void outputBufferSizeUpdated(quint64 size) const;
    void setMonitorPortNum(quint16 port) const;

public slots:
    void start(uint stateIndex);
    void stop(uint stateIndex);
    void reset(const RcRunId &newRunId, uint stateIndex);
    void destroyClient(int clientId);
    void setOutput(COMBINER_OUT combinerOut, int tcpNum);
    void updateAnalizerList(QVector<Analizer*> analizers, QMap<int, QString> clNames);
    void enableTimeDistr(bool en);
    void setStartEvNumber(quint32 startEvNum, quint32 step);

private slots:
    void init();
    void workLoop();
    void sendStatus();
    void tcpStatusUpdated(const OutputTcpStatus &s);
    void tcpFinished(int tcpId, uint fId);
    void writerFileOpened(const RcRunId &runId);
    void outputError(int index, QString errStr);

private:
    void processInit();
    void flushEvent(int singleIndex=-1);
    void flushSingleEvent(int index);
    void sendData(char *src, int size, bool addToOutEvent=false);
    void writeData(RingBuffer<char> *buffer, char *src, int size);
    bool updateClientsEventInfo(int index, int timeout=0);
    int addTcpServer();
    void writeStartStopRecrd(bool isStart);
    void addTrashStat(int index, quint32 evNum);
    QString getClientNameByIndex(int index) const { return clientNames[clientIds[index]]; }
    void resizeBuffer();

    QVector<int> clientIds; // client id by index of enabled client
    QMap<int, QString> clientNames; // by client id
    QVector<EventMetaInfo*> events; // Events pending to be summed
    QVector<RingBuffer<EventMetaInfo> *> eventRingBuffers;
    QVector<int> eventRingBuffersDataSpace;
    bool initFlag;
    bool enabled;
    bool eventStarted;
    int clientNumbers;
    int readoutClientNumbers;
    quint32 startEvNumber;
    quint32 evNumberStep;
    quint32 expectedEvNum;
    CombinerStats stats;
    QMap<int, ClientTrashStat> trashStat;

    RcRunId runId;

    MpdEventHdr mpdDataHdr;
    QTimer * const workLoopTimer;
    QTimer * const statusTimer;
    QString predefStorage;
    QMap<int, OutputTcpServer *> outputTcpServers;
    QMap<int, QThread *> outputTcpServerThreads;
    OutputTcpServer * monitorTcpServer;
    int outputBufferSize;
    RingBuffer<char> * const writerRingBuf;
    FileWriter * const fileWriter;
    QThread * const fileWriterThread;
    QByteArray outputEvent;
    bool outToFile;
    bool outToTcpServer;
    bool noFreeSpaceError; // for single error string
    uint eventsWithDebug;
    uint finishId;
    QMap<int, uint> clientsFinishSt; // <tcpIndex, finishId>
    RcRunId writerRunId; // runId of opened file
//    bool prevEventIncomplet;
    int prevEventIncomplet;
    QElapsedTimer updateClientsTimer;
    uint curStateIndex;
    QMap<int, quint32> lastMissingEvFromClient;
    int suppressLogLagging;
    bool timeDistrEn;

    bool waitingResize;
    int preferedBufSize;
};

#endif // COMBINER_H
