#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "Combiner.h"
#include "RunControl.h"
#include "pnp-server/PNPServer.h"
#include "mongo/StartupOptions.h"
#include "configuration-manager/ConfigurationManager.h"
#include "base/RcRunId.h"

#include <QMainWindow>
#include <QMap>

namespace Ui {
class MainWindow;
}

class EvBStartupOptions;
class Analizer;
class ClientManager;
class MongoConfig;
class PNPServer;
class RemoteControlServer;
class RemoteControlServerStatus;
class RcRunId;
class StatisticDialog;
class TcpStatistic;


class QElapsedTimer;
class QSemaphore;
class QThread;
class QLabel;

struct LogHistorEntry{
    int time;
    QString mess;
    int repeatCnt;
};

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(StartupOptions &start_opt, QWidget * parent = nullptr);
    void init();
    ~MainWindow();

public slots:
    void addLog(QtMsgType severity, QString log, QString thrName);

private slots:
    void on_actionAbout_triggered();
    void on_pushButtonAdd_clicked();
    void on_pushButtonRemove_clicked();
    void on_tableWidget_itemSelectionChanged();
    void updateClientStatus(int index, bool isConnected, QString status);
    void updateTcpStatistic(int index, const TcpStatistic &statistic);
    void runControlStateUpdated(const RC_STATE &state, uint stateIndex);
    void clientStarted(int index, bool ok, uint stateIndex); // from combiner and clients
    void clientStopped(int index, uint stateIndex); // from combiner and clients
    void deviceResetted(int index, uint stateIndex); // from combiner and clients
    void deviceError(int index, QString errStr); // from combiner and clients
    void updateStatus(const QString&);
    void updateCombinerStats(const CombinerStats&);
    void fileWriterStatsUpdated(const FileWriterStats &fwStat);
    void inputBufferSizeUpdated(int id, quint64 size);
    void outputBufferSizeUpdated(quint64 size);

    void on_pushButtonReset_clicked();
    void remoteStartRequest(const RcRunId &runId, quint64 startEvNumber);
    void remoteStopRequest(const RcRunId &runId, quint64 stopEvNumber);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat);
    void addLog(QString log, Qt::GlobalColor color=Qt::black);

    void on_actionChange_storage_dir_triggered();
    void on_actionReset_to_defaults_triggered();
    void on_actionAdvanced_counters_toggled(bool showColumns);
    void on_actionAllow_Remote_Control_toggled(bool checked);
    void data_out_changed(QAction *);
    void on_actionSet_number_of_outputTcpServers_triggered();
    void on_actionSet_EvB_index_triggered();
    void on_actionShow_statistics_triggered();
    void on_actionEnable_client_time_distribution_toggled(bool en);
    void on_actionSet_monitor_port_number_triggered();

    void setClientReadOut(bool readout);
    void gotProgramDescription(const ProgramDescription &newDescr);
    void gotClosedProgram(const ProgramDescription &newDescr);
    void updateTablePalette();

    void switch_to_config(QString program_index, QString configuration_name);
//    bool clientsDestroying(int client_id, int timeout, QElapsedTimer *timer);
    void on_actionConfiguration_manager_triggered();

signals:
    void starting(uint stateIndex);
    void stoping(uint stateIndex);
    void reset(const RcRunId &runId, uint stateIndex);
    void destroyClient(int id);
    void destroyCombiner(int id);
    void destroyAll();
    void remoteCompletedStart(bool success, const RcRunId &runId);
    void remoteCompletedStop(bool success, const RcRunId &runId);
    void sendRemoteState(QString state, bool read4Run);
    void sendRunError(const RcRunId &runId, QString errStr);
    void sendRunWarning(const RcRunId &runId, QString warnStr);
    void sendBackPreasure(const RcRunId &runId, bool backPreasure);
    void storageChanged(QString storage);
    void updateClientReadout(int id, bool readout);
    void analizerListUpdated(QVector<Analizer *> list, QMap<int, QString> clientNames);
    void setOutput(COMBINER_OUT combinerOut, int tcpNum);
    void setEvNumberStep(quint32 startEvNum, quint32 step);
    void allowRemoteControl(bool);
    void changeProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void enableTimeDistr(bool en) const;
    void resizeOutputBuffer(int outputBufSize) const;
    void setMonitorPortNum(quint16 portNum);
    void config_transfer_to_rc(QByteArray data);

private:
    void initCombiner();
    void update_combiner_to_config();
    void setupTable();
    void setup_clients();
    int addClient(const ClientConfig &cc);
    void updatePushButtonState();
    void updateTableSize();
    void updateAnalizerList();
    void updateRemoteStatus();
    void initProgramDescription();
    void updateProgramDescription();
    void updateTcpPortLabels();
    void setOutTcpServerNumber(int tcpNum);
    QVector<ProgramDescription> getFreeProgs() const;
    bool isFreeProg(const ProgramDescription &pd) const;
    int getTableRowByClientIndex(int index) const;
    QString getClientNameByIndex(int index) const;
    int getSelectedRow() const;
    int getClientIdByRow(int row) const;
    void updateTableTextColor(int row, int clientId=-1) const;
    void combinerOutUpdated();
    QLabel *getTcpLabel(int id) const;
    QString humanReadableSize(double size);
    bool tryToStop();

    Ui::MainWindow *ui;
    ConfigurationManager *configuration_manager;
    EvBConfig config;
    QString default_program_index;

    QMap<int, ClientManager*> clients; // stored by client's id
    int clientLastId;
    Combiner *combiner;
    QThread *combinerThread;
    QMap<int, bool> recivedStates;
    QMap<int, bool> lastConnactionState; // by clientId
    bool allOk;
    RemoteControlServer * const remoteControl;
    PNPServer * const pnpServer;
    ProgramDescription progDesc;
    bool run;
    RcRunId loc_runId;
    quint32 lastEvNum;
    QList<LogHistorEntry> logHistory;
    QLabel * const labelFileErrSize;
    QVector<ProgramDescription> progDescrs;
    CombinerStats lastCombinerStats;
    QTime * const clientDistrTime;
    uint enabledClientNum;
    uint curStateIndex;
    StatisticDialog * const statisticDialog;
    bool backPreasure;
};

#endif // MAINWINDOW_H
