#include "RunControl.h"
#include "ui_RunControl.h"

#include <QDebug>
#include <QElapsedTimer>
#include <QTimer>

QString getRunControlStateStr(RC_STATE st){
    switch (st) {
    case RC_RESETTING:
        return "resetting";
    case RC_STOPPED:
        return "stopped";
    case RC_STARTING:
        return "starting";
    case RC_STARTED:
        return "started";
    case RC_STOPPING:
        return "stopping";
    case RC_ERROR:
        return "error";
    default:
        return "undefined";
    }
}

RunControl::RunControl(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::RunControl),
    curState(RC_STOPPED),
    curStateIndex(1),
    transitionStateTimer(new QTimer(this))
{
    ui->setupUi(this);
    connect(transitionStateTimer, &QTimer::timeout, this, &RunControl::stateTimedOut);
    transitionStateTimer->setSingleShot(true);
    transitionStateTimer->setInterval(2000);
}

RunControl::~RunControl()
{
    delete ui;
}

bool RunControl::setState(RC_STATE state)
{
    switch (state) {
    case RC_ERROR:
        qWarning() << "RunControl error state";
        break;
    case RC_RESETTING:
        if(curState!=RC_STOPPED && curState!=RC_ERROR){
            qCritical() << "Trying to go to Resetting state from state:"
                        << getRunControlStateStr(curState);
            return false;
        }
        qInfo() << "RunControl resetting";
        transitionStateTimer->start();
        break;
    case RC_STOPPED:
        if(curState!=RC_STOPPING && curState!=RC_RESETTING){
            qCritical() << "Trying to go to Stoped state from state:"
                        << getRunControlStateStr(curState);
            return false;
        }
        qInfo() << "RunControl stopped";
        transitionStateTimer->stop();
        break;
    case RC_STARTING:
        if(curState!=RC_STOPPED){
            qCritical() << "Trying to go to Starting state from state:"
                        << getRunControlStateStr(curState);
            return false;
        }
        qInfo() << "RunControl is starting";
        transitionStateTimer->start();
        break;
    case RC_STARTED:
        if(curState!=RC_STARTING){
            qCritical() << "Trying to go to Started state from state:"
                        << getRunControlStateStr(curState);
            return false;
        }
        transitionStateTimer->stop();
        qInfo() << "RunControl started";
        break;
    case RC_STOPPING:
        if(curState!=RC_STARTED && curState!=RC_ERROR){
            qCritical() << "Trying to go to Stopping state from state:"
                        << getRunControlStateStr(curState);
            return false;
        }
        qInfo() << "RunControl is stopping";
        transitionStateTimer->start();
        break;
    default:
        qCritical("Unknown RunControl state");
        break;
    }
    ++curStateIndex;
    if(curStateIndex==0)
        curStateIndex=1;
    curState=state;
    updateStateButtons();
//    qInfo() << curStateIndex << getRunControlStateStr(curState);
    emit stateChanged(curState, curStateIndex);
    return true;
}

bool RunControl::waitState(RC_STATE st, int timeoutMs)
{
    QElapsedTimer timer;
    timer.start();
    while(curState != st && !timer.hasExpired(timeoutMs)){
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 50);
    }
    if(curState != st) {
        qCritical() << QString("failed on state=%1 timeoutMs=%2")
                       .arg(getRunControlStateStr(st)).arg(timeoutMs);
    }
    return curState == st;
}

void RunControl::setButtonsEnable(bool enabled)
{
    ui->pushButtonStart->setEnabled(enabled);
    ui->pushButtonStop->setEnabled(enabled);
}

void RunControl::on_pushButtonStart_clicked()
{
    if(!setState(RC_STARTING))
        qWarning() << "Starting prohibited";
}

void RunControl::on_pushButtonStop_clicked()
{
    if(!setState(RC_STOPPING))
        qWarning() << "Stopping prohibited";
}

void RunControl::stateTimedOut()
{
    qCritical() << QString("transitionState timed out; curState is: %1;"
                           "go to resetting state...")
                   .arg(getRunControlStateStr(curState));
    const bool needStop = curState == RC_STARTING;
    setState(RC_ERROR);
    if(needStop)
        setState(RC_STOPPING);
    setState(RC_RESETTING);
}

void RunControl::updateStateButtons()
{
    ui->pushButtonStarted->setChecked(curState == RC_STARTED);
    ui->pushButtonStarting->setChecked(curState == RC_STARTING);
    ui->pushButtonStopped->setChecked(curState == RC_STOPPED);
    ui->pushButtonStopping->setChecked(curState == RC_STOPPING);
    ui->pushButtonResetting->setChecked(curState == RC_RESETTING);
    ui->pushButtonStart->setEnabled(curState == RC_STOPPED);
    ui->pushButtonStop->setEnabled(curState == RC_STARTED);
    ui->pushButtonError->setChecked(curState == RC_ERROR);
}
