#include "TcpClient.h"

#include "params.h"
#include "Analizer.h"
#include "RingBuffer.h"
#include "base/DnsLookup.h"
#include "pnp-server/PNPMessage.h"
#include "metric-sender/MetricSender.h"

#include <QDebug>
#include <QElapsedTimer>
#include <QHostInfo>
#include <QSemaphore>
#include <QTimer>
#include <QtNetwork>

#ifdef Q_OS_UNIX
#include <arpa/inet.h>
#endif
#include <new>

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

namespace {
//const int PACK_SIZE_LIMIT = 20*1024*1024;
//const int QUEUE_SIZE_LIMIT = 100;
//const int RING_SIZE = 100*1024*1024;
const int RING_SIZE = 50*1000*1000;
const int RECONNECT_INTERVAL = 200; //ms
const double RB_HIGH_WATERMARK = 0.9;
const double RB_LOW_WATERMARK = 0.5;
const int LOW_SPACE_INTERVAL = 1000; //ms
const qint64 SOCKET_READ_BUFFER_SIZE=1e8; // 100 MB
}

TcpClient::TcpClient(int _index, RingBuffer<char> *_byteRB) :
    QObject(nullptr),
    index(_index),
    readoutEnabled(false),
    idleReadout(true),
    tcpSocket(new QTcpSocket(this)),
    dnsLookup(new DnsLookup(this)),
    reconnectTimer(new QTimer(this)),
    delayReadTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    sendMetricsTimer(new QTimer(this)),
    resizeTimer(new QTimer(this)),
    low_space_timer(new QTimer(this)),
    connectingAttempts(0),
    packageCounter(0),

    byteRB(_byteRB),
    noFreeSpace(false),
    loopRecivedSize(0),
    curStateIndex(0),
    srcBusy(false),

    waitingResize(false),
    preferedBufSize(0)
{
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<QAbstractSocket::SocketError>();
    qRegisterMetaType<TcpStatistic>();
//    Q_ASSERT(index>=0 && index<10);
    Q_ASSERT(index>=0);
    statistic.index = index;

    tcpSocket->setReadBufferSize(SOCKET_READ_BUFFER_SIZE);
    connect(tcpSocket, &QTcpSocket::readyRead, this, &TcpClient::readData, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::connected, this, &TcpClient::connected, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::disconnected, this, &TcpClient::disconnected, Qt::QueuedConnection);
    connect(tcpSocket, QOverload<QAbstractSocket::SocketError>::of(&QTcpSocket::error),
        this, &TcpClient::displayError, Qt::QueuedConnection);
    connect(dnsLookup, &DnsLookup::reverseLookupComplete,
            this, &TcpClient::gotHostInfo);
    connect(reconnectTimer, &QTimer::timeout, this, &TcpClient::connectToHost, Qt::QueuedConnection);
    reconnectTimer->setSingleShot(true);

    delayReadTimer->setSingleShot(true);
    delayReadTimer->setInterval(1000);
    connect(delayReadTimer, &QTimer::timeout, this, &TcpClient::readData, Qt::QueuedConnection);


    connect(statusTimer, &QTimer::timeout, this, &TcpClient::updateStatus);
    statusTimer->start(500);

    connect(sendMetricsTimer, &QTimer::timeout, this, &TcpClient::doSendMetrics);
    sendMetricsTimer->start(1000);

    connect(resizeTimer, &QTimer::timeout, this, &TcpClient::resizeBuffer);
    resizeTimer->setInterval(100);
    resizeTimer->setSingleShot(100);

    connect(low_space_timer, &QTimer::timeout, this, &TcpClient::low_space_timer_slot);
    low_space_timer->setSingleShot(true);
    low_space_timer->setInterval(LOW_SPACE_INTERVAL);
}


TcpClient::~TcpClient()
{
    readoutEnabled = false;
    disconnectFromHost();
}

void TcpClient::connectToHost()
{
    try{
        if(!readoutEnabled || isConnectionAlive()){
            return;
        }
        tcpSocket->abort();
        if(hostAddr.isNull() || hostPort==0){
            qWarning() << getName().append(" can't connect; client address is unknown.");
            updateStatus();
        } else {
            tcpSocket->connectToHost(hostAddr, hostPort);
            ++connectingAttempts;
            if(tcpSocket->waitForConnected(1000)){
                qInfo() << getName().append(" connected.");
                updateStatus();
            } else {
                QString mess = getName().append(" can't connect.Err:%1")
                        .arg(tcpSocket->errorString());
                qWarning() << mess;
                //        if(connectingAttempts<=5
                //                && tcpSocket->error() == QAbstractSocket::ConnectionRefusedError) {
                if(connectingAttempts<=10) {
                    reconnectTimer->start(RECONNECT_INTERVAL);
                } else {
                    // disable this tcp client
                    emit gotError(mess);
                }
            }
        }
    } catch(...) {
        qDebug()<<"Exception in TcpClient::requestNewData()";
    }
}

void TcpClient::startRun(uint stateIndex)
{
    curStateIndex = stateIndex;
    idleReadout = false;
    if(isConnectionAlive())
        emit updateRunState(false, true, curStateIndex);
}

void TcpClient::stopRun(uint stateIndex)
{
    curStateIndex = stateIndex;
    idleReadout = true;
    emit updateRunState(false, false, curStateIndex);
}

void TcpClient::startReadOut()
{
    if(readoutEnabled){
        qWarning()<<"Trying to start started tcp-client";
        return;
    }
    readoutEnabled = true;
    connectToHost();
}


void TcpClient::stopReadOut()
{
    if(!readoutEnabled){
        qWarning()<<"Trying to stop stopped tcp-client";
        return;
    }
    readoutEnabled = false;

    disconnectFromHost();
    updateStatus();
}

void TcpClient::disconnectFromHost()
{
    if(tcpSocket->state() != QTcpSocket::UnconnectedState){
        tcpSocket->disconnectFromHost();
        qInfo() << getName().append(" disconnected.");
    }
    reconnectTimer->stop();
}


void TcpClient::readData()
{
    if(waitingResize)
        resizeBuffer();

    try{
        qint64 avail64 = tcpSocket->bytesAvailable();
        if(!avail64)
            return;
        if(avail64 > LONG_MAX)
            avail64 = LONG_MAX;
        int avail = static_cast<int>(avail64);

        if(idleReadout){
            loopRecivedSize += tcpSocket->readAll().size();
            return;
        }

        while(avail) {
            int freeSize = byteRB->getFreeSize();
            if(freeSize < (1-RB_HIGH_WATERMARK) * byteRB->getAllSize()){
                if(!noFreeSpace){
                    if(low_space_timer->isActive()) {
                        low_space_timer->stop();
                    } else {
                        qWarning() << QString("%1: free space low").arg(getName());
                    }
                }
                noFreeSpace=true;
                delayReadTimer->start(100);
                return;
            }
            avail = qMin(avail,freeSize);
            if(noFreeSpace && freeSize < RB_LOW_WATERMARK * byteRB->getAllSize()) {
                noFreeSpace=false;
                low_space_timer->start();
            }

            ++packageCounter;
            int free1 = byteRB->getContinuosFreeSize();

            if(free1<0)
                qWarning() << "free1" << free1;
            if(avail<0)
                qWarning() << "avail" << avail;
            if(avail>free1){
                int read1 = tcpSocket->read(byteRB->getWritePointer(), free1);
                if(read1 > 0)
                    loopRecivedSize += read1;
                Q_ASSERT(read1==free1);//todo
                avail -= read1;
                byteRB->writeData(read1);
                //            qDebug()<<"TcpClient::readData() ringBuff->writeData: "<<read1;
                Q_ASSERT(byteRB->getWriterPos()==0);
                //            qDebug()<<"TcpClient read "<<read1;
                avail = qMin(avail, byteRB->getContinuosFreeSize());
            }
            int read2 = tcpSocket->read(byteRB->getWritePointer(), avail);
            if(read2 > 0)
                loopRecivedSize += read2;
            Q_ASSERT(read2==avail);//todo
            byteRB->writeData(read2);
            //        qDebug()<<"TcpClient::readData() ringBuff->writeData: "<<read2;

            avail = tcpSocket->bytesAvailable();
            //        if(avail){
            //            qDebug()<<"double recive";
            //        }
            //        qDebug()<<"TcpClient read "<<read2;
        }
    }catch(...){
        qDebug()<<"Exception in TcpClient::readData()";
    }
}

void TcpClient::low_space_timer_slot()
{
    qInfo() << QString("%1: free space available").arg(getName());
}

void TcpClient::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qDebug() << "Remote hoste closed connection.";
//        emit sendStatus(index, "Remote hoste closed connection");
        updateStatus();
        break;
    case QAbstractSocket::HostNotFoundError:
//        qDebug() << "The host was not found. Please check the "
//                    "host name and port settings.";
        qDebug() << "The host was not found. Please check the "
                    "host name and port settings.";
//        emit sendStatus(index, "host was not found");
        updateStatus();
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
//            qDebug() << tr("The connection was refused by the peer. "
//                           "Make sure the fortune server is running, "
//                           "and check that the host name and port "
//                           "settings are correct.");
            qDebug() << "The connection was refused by the peer. "
                        "Make sure the fortune server is running, "
                        "and check that the host name and port "
                        "settings are correct.";
//            emit sendStatus(index, "The connection was refused by the peer");
            updateStatus();
        }
        break;
    default:
//        qDebug() << tr("The following error occurred: %1.")
//                    .arg(tcpSocket->errorString());
        qDebug() << tr("The following error occurred: %1.")
                    .arg(tcpSocket->errorString());
    }
}


void TcpClient::setAddress(const ProgramDescription &descr, const ProgramInterface &interface)
{
    statistic.name = descr.getName();
    srcBusy=false;
    programIndex = descr.index;
    if(hostAddr == interface.host && hostPort == interface.port) {
        if(!isConnectionAlive())
            connectToHost();
        return;
    }
    if(readoutEnabled && isConnectionAlive()) {
        qWarning() << getName().append("Skip source programm with same name/index");
        return;
    }

    if(readoutEnabled && !interface.isFree) {
        qWarning() << QString("[%1] Can't connect to busy interface '%2' (id:%3)")
                      .arg(descr.getName())
                      .arg(interface.type).arg(interface.id);
        srcBusy = true;
        return;
    }

    connectingAttempts=0;
    disconnectFromHost();

    if(hostAddr != interface.host) {
        hostName.clear();
        dnsLookup->reverseLookup(interface.host.toString());
    }
    hostAddr = interface.host;
    hostPort = interface.port;
//    qInfo()<<getName().append("; address=%2:%3; bufferSize=%4")
//              .arg(hostAddr.toString()).arg(hostPort).arg(RING_SIZE);
    if(readoutEnabled)
        connectToHost();
}


void TcpClient::resetReadOut(uint stateIndex)
{
    curStateIndex = stateIndex;
    qInfo() << getName().append(" start resetting");
    Q_ASSERT(readoutEnabled);
//    Q_ASSERT(tcpSocket->state() == QTcpSocket::UnconnectedState);
//    Q_ASSERT(reconnectTimer->isActive()==false);
//    delayReadTimer;
//    statusTimer;
    connectingAttempts=0;
    packageCounter=0;
    noFreeSpace=false;

#ifdef RESET_ROUGH
    byteRB->clear();
    emit resetted(false, curStateIndex);
#else
    int curFreeSpace = byteRB->getFreeSize();
    int needToFreeSize = byteRB->getAllSize()-curFreeSpace;
    sendLog(QString("curFreeSpace=%1; needToFreeSize=%2")
            .arg(curFreeSpace).arg(needToFreeSize));
    QElapsedTimer timer;
    timer.start();
    while(byteRB->getFreeSize() != byteRB->getAllSize() && !timer.hasExpired(2000)){
        if(byteRB->waitTrashSpace(needToFreeSize, 50))
            byteRB->freeData(needToFreeSize);
    }

//    newDataWC;
    if(timer.hasExpired(2000)){
        sendLog(QString("free=%1 data=%2 trash=%3")
                .arg(byteRB->getFreeSize())
                .arg(byteRB->getDataSize())
                .arg(byteRB->getTrashSize()), Qt::red );
        sendLog(QString("TcpClient index #%1 resetting failed").arg(index));
    } else {
        sendLog(QString("TcpClient index #%1 resetting completed").arg(index));
        emit resetted(false);
    }
#endif
}


void TcpClient::connected()
{
    if(readoutEnabled && !idleReadout)
        emit updateRunState(false, true, curStateIndex);
    connectingAttempts=0;
    updateStatus();
}


void TcpClient::disconnected()
{
    if(readoutEnabled && !idleReadout)
        emit updateRunState(false, false, curStateIndex);
    if(readoutEnabled)
        reconnectTimer->start(RECONNECT_INTERVAL);
    updateStatus();
}

void TcpClient::doSendMetrics()
{
    QString metricPrefix = QString("in.%1.").arg(programIndex);
    MetricSender::send(metricPrefix+"buffer.free",  (double) byteRB->getFreeSize()  / byteRB->getAllSize());
    MetricSender::send(metricPrefix+"buffer.data",  (double) byteRB->getDataSize()  / byteRB->getAllSize());
    MetricSender::send(metricPrefix+"buffer.trash", (double) byteRB->getTrashSize() / byteRB->getAllSize());
}

void TcpClient::updateStatus()
{
    statistic.freePart  = 100.*byteRB->getFreeSize() /byteRB->getAllSize();
    statistic.dataPart  = 100.*byteRB->getDataSize() /byteRB->getAllSize();
    statistic.trashPart = 100.*byteRB->getTrashSize()/byteRB->getAllSize();
#ifdef Q_OS_UNIX
    statistic.rcvBufSize = 0;
    if(tcpSocket && tcpSocket->state()==QAbstractSocket::ConnectedState) {
        bool ok;
        int val = tcpSocket->socketOption(QAbstractSocket::ReceiveBufferSizeSocketOption).toInt(&ok);
        if(ok)
            statistic.rcvBufSize = val;
    }
#endif
    statistic.readBufferSize = tcpSocket->readBufferSize();
    statistic.bytesAvailable = tcpSocket->bytesAvailable();

    emit sendStatistic(index, statistic);

    if(!readoutEnabled) {
        emit sendStatus(index, false, QString("Disabled"));
    } else if(hostAddr.isNull() || hostPort==0) {
        emit sendStatus(index, false, srcBusy ? "Source is busy" : "Address is unknown");
    } else {
        switch (tcpSocket->state()) {
        case QTcpSocket::ConnectingState:
            emit sendStatus(index, false, "Connecting...");
            break;
        case QTcpSocket::ConnectedState:
        {
            if(!idleReadout) {
                int writtenSize = byteRB->getDataSize();
                float buffLoad = writtenSize*100.0/byteRB->getAllSize();
                emit sendStatus(index, true, QString("%1 packages recived; load=%2%%3")
                                .arg(packageCounter).arg(buffLoad,0,'f',1)
                                .arg((buffLoad>80)?"; lack of space":""));
            } else {
                emit sendStatus(index, true, QString("Idle readout (%1)")
                                .arg(loopRecivedSize));
            }
            break;
        }
        case QTcpSocket::UnconnectedState:
            emit sendStatus(index, false, QString("Reconnecting...(%1 attempt) (%2)")
                            .arg(connectingAttempts).arg(tcpSocket->errorString()));
            break;
        case QTcpSocket::ClosingState:
            emit sendStatus(index, false, "Stopping...");
            break;
        default:
            emit sendStatus(index, false, "Unknown state");
            break;
        }
    }
    loopRecivedSize = 0;
}

void TcpClient::clientDescrChanged(const QString &descr)
{
    clientDescr = descr;
}

void TcpClient::gotHostInfo(const QHostInfo &hostInfo)
{
// Not working: always "Unknown error"
//    if(h.error() != QHostInfo::NoError)
//        return;
    if(!QHostAddress(hostInfo.hostName()).isNull())
        return;

    const QString &tail = "." + QHostInfo::localDomainName();
    for(auto &addr : hostInfo.addresses()) {
        if(addr != hostAddr) continue;
        hostName = hostInfo.hostName();
        int i = hostName.lastIndexOf(tail);
        if (i > 0)
            hostName.remove(i, hostName.length()-i);
    }
}

void TcpClient::analyzerStatUpdated(const AnalyzerStat &aStat)
{
    if(aStat.periodCnt==0)
        return;

    int newEvSize = aStat.periodSize/aStat.periodCnt;
    newEvSize = qMin(newEvSize, EVENT_SIZE_MAX);
    newEvSize = qMax(newEvSize, EVENT_SIZE_MIN);
    int newBufSize = newEvSize*EVENT_RING_SIZE;

    if(waitingResize) {
        preferedBufSize = newBufSize;
    } else {
        const int curBufSize = byteRB->getAllSize();
        double factor = 1.*newBufSize/curBufSize;
        if(factor<0.5 || factor>1.1) {
            waitingResize = true;
            preferedBufSize = newBufSize;
            qInfo() << QString("WaitingResize up to %1MB")
                       .arg(1e-6*preferedBufSize,0,'f',3);
        }
    }
    if(waitingResize) {
        emit inputBufferSizeUpdated(index, preferedBufSize);
        resizeBuffer();
    }
}

void TcpClient::resizeBuffer()
{
    Q_ASSERT(preferedBufSize>0);
    if(byteRB->getFreeSize()==byteRB->getAllSize()) {
        qInfo() << QString("Perform buffer resize from %1MB to %2MB")
                   .arg(1e-6*byteRB->getAllSize(),0,'f',3)
                   .arg(1e-6*preferedBufSize,0,'f',3);
        bool ok = byteRB->resize(preferedBufSize);
        if(ok) {
            waitingResize = false;
            return;
        }
    }
    resizeTimer->start();
}

bool TcpClient::isConnectionAlive() const
{
    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return true;
    return false;
}

QString TcpClient::getName() const
{
    return QString("TcpClient #%1 '%2'").arg(index).arg(clientDescr);
}
