#include "DeviceWidget.h"
#include "ui_DeviceWidget.h"

#include "DeviceConfig.h"
#include "mregdev/device_id.h"
#include "base/threadQuit.h"
#include "regio/regop.h"

#include "QCheckBox"
#include "QFileDialog"
#include "QMessageBox"
#include "QThread"
#include "QLineEdit"
#include "QTableWidgetItem"
#include "QScrollBar"

using mlink::RegOpVector;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;

namespace {
const int CHANNEL_COUNT = 16;
enum {COL_THR, COL_HITS, COL_CNT};
}

DeviceWidget::DeviceWidget(DeviceConfig device_config, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::DeviceWidget),
    device_thread(new QThread(this)),
    device(new MscDevice(device_config)),
    device_config(device_config)
{
    init_gui();

    device_thread->setObjectName(QString("dev %1").arg(device_config.device_index.getSerialStr()));

    connect(device, SIGNAL(destroyed()), device_thread, SLOT(quit()));
    connect(this, SIGNAL(destroyAll()), device, SLOT(deleteLater()));
    connect(device, SIGNAL(update_device_config(DeviceConfig)), this, SIGNAL(update_device_config(DeviceConfig)));
    connect(this, SIGNAL(update_device_config(DeviceConfig)), device, SLOT(write_config_to_device(DeviceConfig)));

    connect(this, SIGNAL(soft_clear()), device, SLOT(soft_clear()));
    connect(this, SIGNAL(find_base_line()), device, SLOT(find_base_line()));
    connect(this, SIGNAL(write_loot(QByteArray)), device, SLOT(write_loot(QByteArray)));

    connect(this, SIGNAL(device_discovered(DeviceDescription)), device, SLOT(deviceUpdated(DeviceDescription)));
    connect(device, SIGNAL(update_status(DeviceStatus)), this, SIGNAL(update_status(DeviceStatus)));
    connect(device, SIGNAL(update_status(DeviceStatus)), this, SLOT(update_status_device(DeviceStatus)));

    device->moveToThread(device_thread);
    device_thread->start();
}

DeviceWidget::~DeviceWidget()
{
    emit destroyAll();

    if(!threadQuiting(device_thread, 1000)) {
        qWarning() << QString("device %1 thread isn't stopped").arg(device_config.device_index.getSerialStr());
    }

    delete ui;
}

void DeviceWidget::init_gui()
{
    ui->setupUi(this);
    ui->label_serial->setText(device_config.device_index.getSerialStr());
    ui->label_lut_path->setText(device_config.lut_path);
    ui->doubleSpinBox->blockSignals(true);
    ui->doubleSpinBox->setValue(device_config.channel_thr_map[0]);
    ui->doubleSpinBox->blockSignals(false);

    QTableWidget *table = ui->tableWidget;
    table->setRowCount(CHANNEL_COUNT);
    for(int i=0; i<CHANNEL_COUNT; ++i) {
        //--- double spin box
        QDoubleSpinBox *spin_box_channel = new QDoubleSpinBox(table);
        spin_box_channel->setValue(device_config.channel_thr_map[i]);
        spin_box_channel->setObjectName(QString::number(i));
        spin_box_channel->setMinimum(ui->doubleSpinBox->minimum());
        spin_box_channel->setMaximum(ui->doubleSpinBox->maximum());
        spin_box_channel->setDecimals(ui->doubleSpinBox->decimals());
        spin_box_channel->setSingleStep(ui->doubleSpinBox->singleStep());
        connect(spin_box_channel, SIGNAL(valueChanged(double)), this, SLOT(spin_box_channel_value_changed(double)));
        table->setCellWidget (i, COL_THR, spin_box_channel);

        //--- line edit
        auto hit_item = new QTableWidgetItem();
        hit_item->setFlags(hit_item->flags() & ~Qt::ItemIsEditable);
        table->setItem(i, COL_HITS, hit_item);
    }
    table->resizeRowsToContents();
    ui->tableWidget->resizeColumnToContents(0);
}

void DeviceWidget::on_pushButton_clicked()
{
    QFileDialog file_dialog;
    file_dialog.setDirectory(device_config.lut_path);

    QString fileName=file_dialog.getOpenFileName(this, tr("OpenFile"), "", tr("Bin Files (*.bin);;"));
    if (fileName !=""){
        QFile file(fileName);
        if (!file.open(QIODevice::ReadOnly)){
            QMessageBox::critical(this,tr("Error"),tr("Could not open file"));
            return;
        }
        device_config.lut_path = fileName;
        ui->label_lut_path->setText(device_config.lut_path);


        QByteArray lut_data;
        lut_data = file.readAll();
        file.close();

        emit write_loot(lut_data);
    }

    emit update_device_config(device_config);
}

void DeviceWidget::on_pushButton_upload_clicked()
{
    QFile file(device_config.lut_path);
    if (!file.open(QIODevice::ReadOnly)){
        QMessageBox::critical(this,tr("Error"),tr("Could not open file"));
        return;
    }
    ui->label_lut_path->setText(device_config.lut_path);

    QByteArray lut_data = file.readAll();
    file.close();
    emit write_loot(lut_data);
}

void DeviceWidget::spin_box_channel_value_changed(double value)
{
    device_config.channel_thr_map[QObject::sender()->objectName().toInt()] = value;
    emit update_device_config(device_config);
}


void DeviceWidget::addLog(const QString &log)
{
    qInfo()<<QString("[%1] [%2] %3")
             .arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss"))
             .arg(device_config.device_index.getSerialStr())
             .arg(log);
}

void DeviceWidget::on_doubleSpinBox_valueChanged(double thr)
{
    for(int ch : device_config.channel_thr_map.keys()) {
        device_config.channel_thr_map[ch] = thr;
    }

    for(int i=0; i<CHANNEL_COUNT; ++i){
        QDoubleSpinBox *spin_box_channel = qobject_cast<QDoubleSpinBox*>(ui->tableWidget->cellWidget(i, COL_THR));
        bool oldState = spin_box_channel->blockSignals(true);
        spin_box_channel->setValue(thr);
        spin_box_channel->blockSignals(oldState);
    }
    emit update_device_config(device_config);
}

void DeviceWidget::update_status_device(DeviceStatus device_status)
{
    QColor serial_color = device_status.connected ? Qt::black : Qt::red;
    if(ui->label_serial->palette().color(ui->label_serial->foregroundRole()) != serial_color) {
        auto palette = ui->label_serial->palette();
        palette.setColor(ui->label_serial->foregroundRole(), serial_color);
        ui->label_serial->setPalette(palette);
    }
    if(device_status.ch_cnt.size() != CHANNEL_COUNT) {
        return;
    }

    ui->tableWidget->setUpdatesEnabled(false);
    for(int ch=0; ch<CHANNEL_COUNT; ++ch) {
        ui->tableWidget->item(ch, COL_HITS)->setText(QString::number(device_status.ch_cnt[ch]));
    }
    ui->tableWidget->setUpdatesEnabled(true);
    ui->tableWidget->resizeColumnToContents(0);
}
