#include "MscDevice.h"
#include "DeviceConfig.h"

#include "QDebug"
#include "regio/regop.h"
#include "MscRegs.h"

const double DacScale = 1.25;
const quint16 DacRange = 2048;
const auto MSC_N_CH = DeviceConfig::N_CH;
double dacToVolt(quint16 daq) {
    return DacScale*daq/DacRange-DacScale;
}
quint16 voltToDac(double v) {
    return (v/DacScale+1)*DacRange;
}

using mlink::RegOpVector;
using mlink::RegWrite;

MscDevice::MscDevice(DeviceConfig device_config) :
    QtMregDevice(device_config.device_index.getDevId(), device_config.device_index.getSerial(), nullptr),
    last_config(device_config),
    configValid(false)
{
    device_status.index = device_config.device_index;
    connect(this, SIGNAL(setStatus(const QString &)), this, SLOT(set_status(const QString &)));
}

bool MscDevice::checkLiveMagic()
{
    bool ok = QtMregDevice::checkLiveMagic();
    configValid &= ok;
    if(!configValid)
        write_config_to_device(last_config);
    return ok;
}

void MscDevice::write_config_to_device(DeviceConfig device_config)
{
    mlink::RegOpVector regOp;

    if(!configValid || last_config.channel_thr_map!=device_config.channel_thr_map || last_config.channel_thr_map!=device_config.channel_thr_map){
        QMap<int, int> chThr;
//        const QVector<double> &bl = last_config.baseline;
        for(int ch=0; ch<MSC_N_CH; ++ch)
            chThr[ch] = voltToDac(device_config.channel_thr_map[ch]+device_config.baseLine[ch]);
        regOp << setChThreshold(chThr);
    }
//    memWrite(0, device_config.test_val);

    RegOpVector resReg = regOpExec(regOp);
    bool writeOk = resReg.size() == regOp.size();
    writeLiveMagic();

//    runFlag = dc.runEn;
    last_config = device_config;
    configValid = writeOk && isConnected();
}

void MscDevice::set_status(QString status)
{
    device_status.status = status;
    emit update_status(device_status);
}

void MscDevice::statusPoll()
{
    QtMregDevice::statusPoll();
    device_status.connected = isConnected();
    if(isConnected()) {
        device_status.temp = getTemperature();
        device_status.ch_cnt.clear();
        for(int i=0; i<MSC_N_CH; ++i){
            device_status.ch_cnt.append(regRead32(REG_CYCLE_CNT_BASE + 2*i));
        }
    }

    emit update_status(device_status);
}

void MscDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
//    emit deviceConnected(getDeviceSerialNumber(), getFirmwareVersionStr());
    reset_dac();

    device_status.fwStr = getFirmwareVersionStr();
    device_status.fwRev = getDeviceFwVersion();
    if(!configValid)
        write_config_to_device(last_config);
}

void MscDevice::on_deviceDisconnected()
{
    QtMregDevice::on_deviceDisconnected();
    configValid = false;
}

void MscDevice::reset_dac()
{
    RegOpVector regOp;
    for(int reg=REG_DAC_SPI_1; reg<=REG_DAC_SPI_2;++reg) {
        regOp << RegWrite(reg, DAC_CMD_RESET); // DAC Data and control reset
        regOp << RegWrite(reg, 0x8000); // gain x1, unbuff, Vref
        regOp << RegWrite(reg, 0xA001); // LDAC High
    }
    regOpExec(regOp);
}


mlink::RegOpVector MscDevice::setChThreshold(QMap<int, int> chThr)
{
    mlink::RegOpVector regOp;

    for (int ch : chThr.keys()) {
        const int regAddr = ch/8 ? REG_DAC_SPI_2 : REG_DAC_SPI_1;
        quint16 dacData = chThr[ch] & 0xFFF;
        dacData |= (7-(ch%8))<<12;

        regOp << RegWrite(regAddr, dacData);
//        qInfo() << hex << regAddr << dacData << dec << ch << (chThr[ch] & 0xFFF);
    }
    regOp << RegWrite(REG_DAC_SPI_1, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC_SPI_2, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC_SPI_1, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC_SPI_2, DAC_CMD_SINGLE_UPDATE);
    return regOp;
}

void MscDevice::find_base_line()
{
    QVector<double> baseLine;
    baseLine.resize(MSC_N_CH);
    const int numSamples = 20;

    RegOpVector getCmpSt;
    for(int i=0; i<numSamples; ++i)
        getCmpSt << mlink::RegRead(REG_COMPSTATE);

    QMap<int, QMap<quint16, int> > chCompSt;
    QMap<int, int> dacMap;
    const auto minDac = DacRange*0.8;
    const auto maxDac = DacRange*1.2;
    for(quint16 dacVal=minDac; dacVal<maxDac; ++dacVal) {
        for(int ch=0; ch<MSC_N_CH; ++ch)
            dacMap[ch] = dacVal;
        regOpExec(setChThreshold(dacMap));

        for(int ch=0; ch<MSC_N_CH; ++ch)
            chCompSt[ch][dacVal] = 0;

        RegOpVector cmpStRegs = regOpExec(getCmpSt);
        for(int i=0; i<numSamples; ++i) {
            const quint16 &compSt = cmpStRegs[i].data;
            for(int ch=0; ch<MSC_N_CH; ++ch) {
                if(compSt & (1<<ch))
                    ++chCompSt[ch][dacVal];
            }
        }
    }

    QStringList info;
    info << QString("Baseline for device %1%2:").arg(getDeviceInfoStr()).arg(QChar(0x02da));
    for (int ch : chCompSt.keys()) {
        bool missing=true;
        const auto &compSt = chCompSt[ch];
        for (auto dacVal : compSt.keys()) {
            if(!compSt.contains(dacVal-1)) continue;
            const int &val1 = compSt[dacVal-1];
            const int &val2 = compSt[dacVal];
            if((val1 >= (numSamples/2)) && (val2 < (numSamples/2))) {
                const double v = dacToVolt(dacVal);
                baseLine[ch] = v;
                const double prevV = last_config.baseLine.value(ch,0);
                info << QString("baseline %1: %2(%3mV) shift:%4(%5mV)")
                           .arg(ch, 2).arg(dacVal, 4).arg(v*1e3,5,'f',1)
                           .arg((qint32)dacVal-voltToDac(prevV), 3)
                           .arg((v-prevV)*1e3,5,'f',1);
                missing=false;
                break;
            }
        }
        if(missing) {
            info << QString("Failed to detect baseline for ch: %1").arg(ch);
        }
    }
    qInfo().noquote() << info.join('\n');

    configValid = false;
    last_config.baseLine = baseLine;

    emit update_device_config(last_config);
}

void MscDevice::write_loot(QByteArray lut_data)
{
    std::vector<quint32> src;
    src.resize(lut_data.size()/sizeof (quint32));
    memcpy(reinterpret_cast<void *>(&src[0]), lut_data.data(), lut_data.size());
    int size = src.size();
    for(int i=0; i<size; ++i) {
        memWrite(i, src[i]);
    }
    return;

//    memWriteBlk(0, src);
    if(isConnected())
        qInfo() << "test" << regRead(REG_CTRL);
    else
        qInfo() << "test dead";
    qInfo() << "isConnected():" << isConnected();
    for(int i=1; i<size; ++i) {
        if(!isConnected()) break;
        std::vector<quint32> src;
        src.resize(i);
        qInfo() << i;
        memWriteBlk(0, src);
        if(isConnected())
            qInfo() << "test" << regRead(REG_CTRL);
        else
            qInfo() << "test dead";
        qInfo() << "isConnected():" << isConnected();
    }
}

void MscDevice::soft_clear()
{
    if(!isConnected()) return;

    RegOpVector regOp;
    quint16 soft_clear = regRead(REG_CTRL);

    soft_clear |= (quint16)1 << 0;
    regOp << RegWrite(REG_CTRL, soft_clear);

    soft_clear &= (quint16)0xFFFE << 0;
    regOp << RegWrite(REG_CTRL, soft_clear);

    regOpExec(regOp);
}

