//
//    Copyright 2011 Ilja Slepnev & Alexey Shutov
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <exception>
#include <iostream>
#include <signal.h>
#include <QCoreApplication>
#include <QThread>

#include "mstream-lib/MStreamDump.h"
#include "mstream-lib/MStreamDumpParameters.h"
#include "base/SyslogMessageSender.h"
#include "metric-sender/MetricSender.h"

#include "pnp-server/PNPServer.h"
#include "mldiscover/MldpListener.h"

static std::function<void()> callbackMStreamTerm;
void mstreamTerm(int signum) {
    qInfo() << "Interrupt signal (" << signum << ") received.\n";

    if(callbackMStreamTerm) {
        callbackMStreamTerm();
    }
    signal(SIGTERM, nullptr);
    raise(SIGTERM);
}

int main(int argc, char *argv[])
{
    int ret;

    QCoreApplication::setOrganizationName("AFI Electronics");
    QCoreApplication::setOrganizationDomain("local");
    QCoreApplication::setApplicationName("MStream");

#ifdef QT_DEBUG
    qInfo() << "Debug version";
#endif
    qInfo() << "Starting up";
    SyslogMessageSender::instance().install();
    MetricSender::instance().install();

    try {
        QCoreApplication app(argc, argv);
        MStreamDumpParameters msParams;
        msParams.parseCommandLine();

        MStreamDump *mStreamDumper = new MStreamDump(msParams);
        callbackMStreamTerm = std::bind(&MStreamDump::terminate, mStreamDumper);
        signal(SIGTERM, mstreamTerm);
        QObject::connect(mStreamDumper, &QObject::destroyed, &app, &QCoreApplication::quit);

        QThread *pnp_thread = new QThread();
        PNPServer *pnp_server = new PNPServer();
        pnp_server->moveToThread(pnp_thread);
        pnp_thread->setObjectName("PNP Server");
        pnp_thread->start();
        QObject::connect(mStreamDumper, &MStreamDump::closing, pnp_server, &QObject::deleteLater);
        QObject::connect(pnp_server, &QObject::destroyed, pnp_thread, &QThread::quit);
        QObject::connect(mStreamDumper, &MStreamDump::updateProgramDescription,
            pnp_server, &PNPServer::updateProgramDescription);

        QThread *mldp_listener_thread= new QThread();
        MldpListener *mldp_listener = new MldpListener();
        mldp_listener_thread->setObjectName("MLDP Listener");
        mldp_listener->moveToThread(mldp_listener_thread);
        mldp_listener_thread->start();
        QObject::connect(mldp_listener, &MldpListener::received, mStreamDumper, &MStreamDump::receivedMldpDescr);
        QObject::connect(mStreamDumper, &MStreamDump::closing, mldp_listener, &MldpListener::deleteLater);
        QObject::connect(mldp_listener, &QObject::destroyed, mldp_listener_thread, &QThread::quit);

        ret = app.exec();
    }
    catch (std::exception &exc) {
        std::cerr << exc.what() << std::endl;
        qCritical() << exc.what();
        exit(1);
    }
    SyslogMessageSender::instance().remove();

    qInfo() << "Shutdown complete";
    return ret;
}
