//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "version.h"
#include "device-discover/DiscoverDialog.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/RemoteControlServer.h"
#include "metric-sender/MetricSender.h"
#include "configuration-manager/ConfigurationManager.h"
#include "mongo/RootConfig.h"
#include "TrigDevice.h"
#include "remote-control-server/EvNumChecker.h"
#include "SlowControlServer.h"
#include "base/SyslogMessageSender.h"
#include "run-control-client-manager/ClientManagerWidget.h"
#include "run-control-client-manager/RunStates.h"
#include "base/AboutBox.h"
#include "base/ThreadController.h"


//#include "mldiscover/DeviceDescription.h"
#include "mongo/StartupOptions.h"
//#include "mongo/RootConfig.h"
//#include "configuration-manager/ConfigurationManager.h"
//#include "TrigDevice.h"

//#include <QtGui/QApplication>
#include <QDesktopWidget>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostAddress>
#include <QLabel>
#include <QStatusBar>
#include <QStyle>
#include <QSettings>
#include <QShortcut>
#include <QTabBar>
#include <QThread>
#include <QTimer>
#include <QMessageBox>
#include <QFileDialog>

Q_DECLARE_METATYPE(Qt::GlobalColor)
Q_DECLARE_METATYPE(TrigStatus)
Q_DECLARE_METATYPE(RcTrigConfig)

const int CONFIG_TIMEOUT = 4000;

MainWindow::MainWindow(StartupOptions &startup_options,  QWidget *parent):
    QMainWindow(parent)
    , ui(new Ui::MainWindow)
    , config(ConfigSelector(startup_options.program_type, startup_options.program_index, startup_options.configuration_name))
    , labelState(new QLabel("disconnected", this))
    , labelSlowControl(new QLabel("SC: disabled", this))
    , labelDevice(new QLabel(this))
    , labelFirmware(new QLabel("f/w:unknown", this))
    , labelTemp(new QLabel(this))
    , discoverDialog(new DiscoverDialog(this))
    , pnpServer(new PNPServer(this))
    , trigDevice(nullptr)
    , slowControl(new SlowControlServer(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , remoteConnected(false)
    , evNumChecker(new EvNumChecker(this))
    , hasRunError(false)
    , config_timer(new QTimer(this))
{
    QThread::currentThread()->setObjectName("GUI thread");

    ui->setupUi(this);
    connect(ui->actionQuit, &QAction::triggered, qApp, &QCoreApplication::quit);
    ui->tabWidget->setCurrentIndex(0); // always start with main controls view

    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DeviceDescription>();
    qRegisterMetaType<TrigStatus>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<RcTrigConfig>();
    qRegisterMetaType<RootConfig>();

    setWindowTitle(config.cs.toString());

    megamaster_flg = false;

    QTimer::singleShot(0, this, &MainWindow::init);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    discoverDialog->close();

    emit closeProgram(progDesc);
    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stateRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();

    if(trigDevice){
        emit deleteHardware(DeviceIndex(trigDevice->getDeviceId(), trigDevice->getDeviceSerialNumber()));
    }

    emit destroyAll();

    QObject::disconnect(&SyslogMessageSender::instance(), &SyslogMessageSender::newMsg,
                     this, &MainWindow::addLog);
    event->accept();
}

MainWindow::~MainWindow()
{
    delete ui->trigMainForm;
    delete ui->clientManagerWidget;
    delete ui->vmeTrigsWidget;
    delete ui;
}

void MainWindow::initDiscoverDialog()
{
    discoverDialog->setSingleMode(true);
    discoverDialog->addDeviceIdFilter(DEVICE_ID_U40VE_RC);
    discoverDialog->addDeviceIdFilter(DEVICE_ID_UT24VE);
    connect(discoverDialog, &DiscoverDialog::deviceDiscoveredAll,
            ui->vmeTrigsWidget, &VmeTrigsWidget::deviceDiscovered);
    connect(discoverDialog, &DiscoverDialog::deviceDiscoveredAll,
            ui->vmeTrigsWidget, &VmeTrigsWidget::insertAvailableDevice);
}

void MainWindow::initRemoteControl()
{
    connect(remoteControl, &RemoteControlServer::startRequest,
            this, &MainWindow::onRemoteStartRequest);
    connect(remoteControl, &RemoteControlServer::stopRequest,
            this, &MainWindow::onRemoteStopRequest);
    connect(this, &MainWindow::remoteStartCompleted,
            remoteControl, &RemoteControlServer::completedStart);
    connect(this, &MainWindow::remoteStopCompleted,
            remoteControl, &RemoteControlServer::completedStop);
    connect(this, &MainWindow::sendRemoteState,
            remoteControl, &RemoteControlServer::sendState);
    connect(this, &MainWindow::sendVmeFlag,
            remoteControl, &RemoteControlServer::onVmeFlag);
    connect(this, &MainWindow::sendRunError,
            remoteControl, &RemoteControlServer::onRunError);
    connect(this, &MainWindow::sendRunWarn,
            remoteControl, &RemoteControlServer::onRunWarning);
    connect(this, &MainWindow::sendBackPressure,
            remoteControl, &RemoteControlServer::onBackPressure);

    connect(remoteControl, &RemoteControlServer::connectionStatusUpdated,
            this, &MainWindow::onRemoteStatusUpdated);
    connect(this, &MainWindow::allowRemoteControl,
            remoteControl, &RemoteControlServer::startListen);
    connect(this, &MainWindow::changeProgramDescription,
            remoteControl, &RemoteControlServer::changeProgramDescription);

    connect(remoteControl, &RemoteControlServer::evNumCheckRequest,
            evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
    connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
            remoteControl, &RemoteControlServer::evNumCheckResponse);

    connect(this, &MainWindow::destroyAll,
            remoteControl, &RemoteControlServer::deleteLater);

    // thread stuff
    new ThreadController(remoteControl, this);
}

void MainWindow::init()
{
    QObject::connect(&SyslogMessageSender::instance(), &SyslogMessageSender::newMsg,
                     this, &MainWindow::addLog);

    qInfo() << QString("Program index is: '%1'").arg(config.cs.getIndex());

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.rc." + shortHostName + "." + config.cs.getIndex());

    ui->groupBoxWarn->hide();
    ui->labelWarTrigDev->hide();
    ui->labelWarClients->hide();
    ui->labelWarTmwr->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnClients->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTrigDev->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnClients->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnTmwr->setStyleSheet("QLabel { color : orange;}");

    ui->mainToolBar->hide();
    statusBar()->addWidget(labelState);
    statusBar()->addPermanentWidget(labelSlowControl);
    statusBar()->addPermanentWidget(labelDevice);
    statusBar()->addPermanentWidget(labelTemp);
    statusBar()->addPermanentWidget(labelFirmware);

    QString styleLargeFont = "font-size: 18pt";
//    ui->groupBoxRunControl->setStyleSheet();
    ui->pushButtonRestartAll->setStyleSheet(styleLargeFont);
    ui->pushButtonStopAll->setStyleSheet(styleLargeFont);
    ui->labelRunState->setStyleSheet(styleLargeFont);
    ui->pushButtonRestartAll->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonStopAll->setIcon(style()->standardIcon(QStyle::SP_MediaStop));
//  labelState->setStyleSheet("QLabel {color: red}");

    initDiscoverDialog();

    connect(ui->actionCheck_LVDS_parity, &QAction::toggled,
            ui->trigMainForm, &TrigMainForm::checkLvdsParityToggled);

    connect(ui->clientManagerWidget, &ClientManagerWidget::runWarn,
            this, &MainWindow::onRunWarning);
    connect(ui->clientManagerWidget, &ClientManagerWidget::runError,
            this, &MainWindow::onRunError);
    connect(ui->clientManagerWidget, &ClientManagerWidget::updateTabState,
            this, &MainWindow::setClientTabState);
    connect(ui->clientManagerWidget, &ClientManagerWidget::client_config_changed,
            this, &MainWindow::client_config_receiver);

    connect(ui->vmeTrigsWidget, &VmeTrigsWidget::updateTabState,
            this, &MainWindow::setTabState);
    connect(ui->vmeTrigsWidget, &VmeTrigsWidget::runError,
            this, &MainWindow::onRunError);

    connect(this, &MainWindow::runStart,
            ui->clientManagerWidget, &ClientManagerWidget::onRestartAll);
    connect(this, &MainWindow::runStart,
            ui->vmeTrigsWidget, &VmeTrigsWidget::onRunStart);
    connect(this, &MainWindow::runStart,
            ui->trigMainForm, &TrigMainForm::runStart);
    connect(this, &MainWindow::runStart,
            slowControl, &SlowControlServer::onRunStart);
    connect(this, &MainWindow::sendControlEnable,
            ui->clientManagerWidget, &ClientManagerWidget::setControlEnable);
    connect(this, &MainWindow::sendControlEnable,
            ui->vmeTrigsWidget, &VmeTrigsWidget::setControlEnable);
    connect(this, &MainWindow::sendControlEnable,
            ui->trigMainForm, &TrigMainForm::setControlEnable);

    connect(pnpServer, &PNPServer::programDiscovered, ui->clientManagerWidget, &ClientManagerWidget::programDiscovered);
    connect(pnpServer, &PNPServer::programClosed, ui->clientManagerWidget, &ClientManagerWidget::programClosed);
    connect(pnpServer, &PNPServer::programDiscovered, ui->vmeTrigsWidget, &VmeTrigsWidget::gotProgramDescription);
    connect(pnpServer, &PNPServer::programClosed, ui->vmeTrigsWidget, &VmeTrigsWidget::gotClosedProgram);
    connect(this, &MainWindow::changeProgramDescription, pnpServer, &PNPServer::changeProgramDescription);
    connect(this, &MainWindow::closeProgram, pnpServer, &PNPServer::closeProgram);
    connect(this, &MainWindow::deleteAll, pnpServer, &QObject::deleteLater);

    //--- configurations transfer
    connect(ui->clientManagerWidget, &ClientManagerWidget::config_transfer,
            this, &MainWindow::config_transfer_from_clients);
    connect(this, &MainWindow::config_transfer_to_rc,
            remoteControl, &RemoteControlServer::config_transfer);
    connect(config_timer, &QTimer::timeout,
            this, &MainWindow::config_timeout);
    config_timer->setSingleShot(true);
    config_timer->setInterval(CONFIG_TIMEOUT);

    initRemoteControl();

    connect(this, &MainWindow::destroyAll,
            slowControl, &SlowControlServer::deleteLater);
    connect(this, &MainWindow::allowSlowControl,
            slowControl, &SlowControlServer::startListen);
    connect(slowControl, &SlowControlServer::connectionStatusUpdated,
            this, &MainWindow::onSlowCntStatusUpdated);
    connect(this, &MainWindow::slowControlStateUpdated,
            slowControl, &SlowControlServer::slowControlStateUpdated);

    new ThreadController(slowControl, this);
    initProgramDescription();

    QStringList searchTargs;
    searchTargs << DaqConfig::getTypeName(DaqConfigAdc64System);
    searchTargs << DaqConfig::getTypeName(DaqConfigEventBuilder);
    searchTargs << DaqConfig::getTypeName(DaqConfigHrbSystem);
    searchTargs << DaqConfig::getTypeName(DaqConfigRunControl);
    searchTargs << DaqConfig::getTypeName(DaqConfigVmeDaq);
    searchTargs << DaqConfig::getTypeName(DaqConfigLand);
    searchTargs << DaqConfig::getTypeName(DaqConfigTtvxs);
    searchTargs << DaqConfig::getTypeName(DaqConfigGenerator);
    searchTargs << DaqConfig::getTypeName(DaqConfigTdcControl);
    pnpServer->startSearch(searchTargs);

    connect(this, &MainWindow::runStop,
            ui->clientManagerWidget, &ClientManagerWidget::onStopAll);
    connect(this, &MainWindow::runStop,
            ui->vmeTrigsWidget, &VmeTrigsWidget::onRunStop);

    loadSettings();

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.endGroup();

    configuration_manager = new ConfigurationManager(this, config.cs.program_type, config.default_root);
    connect(static_cast<ConfigurationManager *>(configuration_manager), &ConfigurationManager::switch_to_config,
            this, &MainWindow::switch_to);

    connect(evNumChecker, &EvNumChecker::evNumRequest,
            ui->clientManagerWidget, &ClientManagerWidget::evNumRequest);
    connect(evNumChecker, &EvNumChecker::evNumRequest,
            ui->vmeTrigsWidget, &VmeTrigsWidget::evNumRequest);

    connect(ui->clientManagerWidget, &ClientManagerWidget::evNumResponse,
            evNumChecker, &EvNumChecker::gotClientEvNum);
    connect(ui->vmeTrigsWidget, &VmeTrigsWidget::evNumResponse,
            evNumChecker, &EvNumChecker::gotVmeTrigEvNum);

    connect(ui->clientManagerWidget, &ClientManagerWidget::clientMapUpdated,
            evNumChecker, &EvNumChecker::updateClientMap);
    connect(ui->vmeTrigsWidget, &VmeTrigsWidget::vmeMapUpdated,
            evNumChecker, &EvNumChecker::updateVmeMap);

    connect(ui->clientManagerWidget, &ClientManagerWidget::backPressure,
            this, &MainWindow::onBackPressure);
    connect(ui->clientManagerWidget, &ClientManagerWidget::vmeFlagUpdated,
            this, &MainWindow::onVmeFlagUpdated);

}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.cs.program_index, config.cs.configuration_name);
}

void MainWindow::switch_to(QString program_index, QString configuration_name)
{
    configuration_manager->hide();
    config.cs.program_index = program_index;
    config.cs.configuration_name = configuration_name;

    loadSettings();
    setWindowTitle(config.cs.toString());
}

void MainWindow::config_transfer_from_clients(int id, RootConfig root)
{
    for(int id_tmp : config.clients.keys()){
        if(config.clients[id_tmp].id == id){
            clients_config[id] = root;
        }
    }

    bool all_clients_config = true;
    for(int id_tmp : config.clients.keys()) {
        if(!clients_config.contains(id_tmp)) {
            all_clients_config = false;
        }
    }

    if(all_clients_config) {
        run_root_prepare();
        config_timer->stop();
    }
}

void MainWindow::run_root_prepare()
{
    RootConfig my_config = config.root;
    my_config.item_name = config.cs.program_type + "_" +
            config.cs.program_index + "_" +
            QString::number(config.pc.runId.runNumber);

    for(int id_tmp : clients_config.keys()){
        RootConfig root_client = clients_config[id_tmp];
        QString program_type = root_client.get_data<QString>("program_type");
        QString program_index = root_client.get_data<QString>("program_index");
        QString run_number = QString::number(root_client.get_data<int>("run_number"));
        QString item_name = program_type + "_" + program_index + "_" + run_number;
        root_client.item_name = item_name;

        RootConfig my_child;
        QString child_name;
        for(int i=1; i<=clients_config.size(); ++i){
            child_name = QString("Client_%1").arg(i);
            my_child = my_config.child(child_name);
            QString program_type_tmp = my_child.get_data<QString>("program_type");
            QString program_index_tmp = my_child.get_data<QString>("program_index");
            if((program_type == program_type_tmp)&&(program_index == program_index_tmp)){
                my_child.append_child(root_client);
                my_config.remove_child(child_name);
                my_config.append_child(my_child);
                break;
            }
        }
    }

    if(megamaster_flg){
        RootConfig run_root = RootConfig();
        run_root.set_data("program_type", config.cs.program_type);
        run_root.set_data("run_index", config.pc.runId.runIndex);
        run_root.set_data("run_number", config.pc.runId.runNumber);
        QDateTime current_dt = QDateTime::currentDateTime();
        run_root.set_data("date_of_creation", current_dt);

        run_root.append_child(my_config);
        config.write_history_config(run_root);
    }
    else{
        emit config_transfer_to_rc(my_config.get_json().toBinaryData());
    }
}

void MainWindow::clearRunWarning()
{
    ui->labelWarClients->hide();
    ui->labelWarnClients->clear();
    ui->labelWarnClients->hide();

    ui->labelWarTrigDev->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnTrigDev->clear();

    ui->labelWarTmwr->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTmwr->clear();
    updateGroupBoxWarn();
}

void MainWindow::config_timeout()
{
    run_root_prepare();
}

void MainWindow::client_config_receiver(QMap<int, ClientConfig> client_config_)
{
    config.clients = client_config_;
    config.write_config();
}

void MainWindow::on_actionQuit_triggered()
{
    close();
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::findDevice()
{
    if(config.rcTrigConfig.rcSerial){
        DeviceDescription dd(config.rcTrigConfig.rcDeviceId, config.rcTrigConfig.rcSerial);
        dd.enabled = true;
        discoverDialog->addSelected(dd);
    }

    if(discoverDialog->exec() == QDialog::Accepted){
        bool found = false;
        for(const DeviceDescription &dd: discoverDialog->getSelectedDevices()) {
            if(!dd.enabled)
                continue;
            found = true;
            QHostAddress host(dd.ip_addr);
            config.rcTrigConfig.rcSerial = dd.serial_id;
            config.rcTrigConfig.rcDeviceId = dd.device_id;
            config.rcTrigConfig.hostStr = dd.ip_addr ? host.toString() : "";
            config.rcTrigConfig.hostTs = QDateTime::currentDateTime();
            qInfo() << QString("Selected Run Control device %1(%2) at %3")
                       .arg(dd.getSerialIdStr()).arg(dd.getModelStr())
                       .arg(host.toString());
            updateDeviceLabel();
            ui->trigMainForm->updateExtTrigGroup();
            ui->trigMainForm->updateRemoteBlocking(remoteConnected);
            connectToTrig(dd);
        }
        if(!found) {
            qInfo() << "No Run Control device selected";
            labelDevice->clear();
            config.rcTrigConfig.rcSerial = 0;
            config.rcTrigConfig.rcDeviceId = 0;
            config.rcTrigConfig.hostStr.clear();
            config.rcTrigConfig.hostTs = QDateTime();
            updateDeviceLabel();
            ui->trigMainForm->updateExtTrigGroup();
            ui->trigMainForm->updateRemoteBlocking(remoteConnected);
            config.write_config();

            connectToTrig(DeviceDescription());
        }
    }
}

void MainWindow::onRemoteStartRequest(const RcRunId &runId, quint64 startEvNumber)
{
    Q_UNUSED(startEvNumber)
    megamaster_flg = false;
    clients_config.clear();
    bool success = perfomeStart(runId);
    emit remoteStartCompleted(success, runId);
    if(success && config.clients.isEmpty()){
        emit config_transfer_to_rc(config.root.get_json().toBinaryData());
    }

    updateProgramDescription();
}

void MainWindow::onRemoteStopRequest(const RcRunId &runId, quint64 stopEvNumber)
{
    Q_UNUSED(stopEvNumber)
    bool success = perfomeStop(false);

    emit remoteStopCompleted(success, runId);
    updateProgramDescription();
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl) == 0) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        progDesc.interfaces.push_back(interface);
    }

    ProgramInterface &i = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !status.portInUse;

    if(status.portInUse){
       i.setPeer(status.peerHost, status.peerPort);
    }
    else
        i.peers.clear();

    updateProgramDescription();
    bool doStop=false;
    if(config.pc.runEnable && remoteConnected && !status.portInUse)
        doStop = true;
    remoteConnected = status.portInUse;
    updateRemoteBlocking();
    if(doStop) {
        qWarning() << "Stop run due to remoteControl disconnect";
        perfomeStop(true);
    }
    updateStatStopButton();
}

void MainWindow::onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse)
{
    QString txt = "SlowControl: ";

    if(listening)
        if(portInUse)
            txt.append("connected");
        else
            txt.append("disconnected");
    else
        txt.append("disabled");

    labelSlowControl->setStyleSheet(QString("QLabel { color : %1; }")
                                    .arg(listening ? (portInUse?"green":"red") : "gray"));

    if(tcpPort)
        labelSlowControl->setToolTip(QString("Listening on %1 port").arg(tcpPort));
    labelSlowControl->setText(txt);
}

void MainWindow::setClientTabState(bool isOk)
{
    setTabState(isOk, TAB_CLIENTS);
}

void MainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    QColor c = bar->tabTextColor(tabIndex);
    if(c.isValid() == isOk) {
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
    }
    bool prevOk = isReadForRun();
    tabOk[tabIndex] = isOk;
    bool newOk = isReadForRun();
    if(!config.pc.runEnable)
        emit slowControlStateUpdated(newOk ? SC_STATE_STANDBY: SC_STATE_FAULT,
                                     newOk ? "Ready for run" : "Not ready for run");
    if(newOk != prevOk)
        updateProgramDescription();
}

void MainWindow::blockControl(bool block)
{
    const bool enabled = !block;
    emit sendControlEnable(enabled);
    ui->menuOptions->setEnabled(enabled);
}

bool MainWindow::perfomeStart(const RcRunId &runId)
{
    blockControl(true);
    ui->pushButtonRestartAll->setEnabled(false);
    bool isOk = isReadForRun();
    QElapsedTimer timer;
    timer.start();
    if(isOk) {
        setTabState(true, TAB_LOGS);
    } else {
        QString msg = "Can't start run. There are some errors. Check ";
        if(!tabOk.value(TAB_TRIG_DEV, false))
            msg.append("Runcontrol device;");
        if(!tabOk.value(TAB_CLIENTS, false))
            msg.append("Clients;");
        if(!tabOk.value(TAB_VME_TRIGS, false))
            msg.append("TMWR modules;");
        qCritical() << msg;
    }

    // Wait till end of spill for VME system
    if(megamaster_flg && config.hasVme()) {
        ui->trigMainForm->closeTrig();
        qInfo() << "Wait till end of spill for VME system";
        const int SPILL_TIMEOUT = 15000;
        bool spillEndTimeout=false;
        while(lastTrigStatus.spillState.reg.bit.phase != 4 ) {
            QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 10);
            if(timer.elapsed() > SPILL_TIMEOUT) {
                spillEndTimeout=true;
                break;
            }
        }
        while(lastTrigStatus.spillState.reg.bit.phase == 4 ) {
            QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 10);
            if(timer.elapsed() > SPILL_TIMEOUT) {
                spillEndTimeout=true;
                break;
            }
        }
        if(spillEndTimeout) {
            qWarning() << "Can't detect end of spill for correct start of Run";
        } else {
            QThread::currentThread()->msleep(100);
            timer.start();
        }
    }

    hasRunError = !isOk;
    if(isOk) {
        const int START_TIMEOUT = 3000;
        emit slowControlStateUpdated(SC_STATE_INIT, "Starting");

        ui->pushButtonRestartAll->setEnabled(false);
        config.pc.runId.set(runId);
        config.pc.runEnable = true;
        config.write_config();
        ui->clientManagerWidget->set_run_enable(config.pc.runEnable);
        ui->labelRunState->setText(QString("Starting Run %1...")
                                   .arg(QString(config.pc.runId)));
        ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
        ui->plainTextEditLog->appendPlainText("");
        qInfo() << QString("!!! Preparing Run %1... !!! ")
                   .arg(QString(config.pc.runId));
        ui->trigMainForm->closeTrig();
        ui->trigMainForm->clearStats();
        emit runStart(config.pc.runId);
        // Get started state from Vme Modules
        while(!timer.hasExpired(START_TIMEOUT)){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
            if(ui->vmeTrigsWidget->get_runId()==config.pc.runId &&
                    ui->vmeTrigsWidget->getRunState() != RS_STARTING )
                // Break on error or started
                break;
        }
        if(ui->vmeTrigsWidget->getRunState() != RS_STARTED){
            qCritical() << "Some vme trigger failed to reset EvNumber";
            isOk = false;
        }

        // Get started state from Tcp Clients (sub-programms)
        while(!timer.hasExpired(START_TIMEOUT)){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
            if(ui->clientManagerWidget->get_run_id()==config.pc.runId &&
                    ui->clientManagerWidget->getRunState() != RS_STARTING)
                // Break on error or started
                break;
        }
        auto sttt = ui->clientManagerWidget->getRunState();
        if(sttt != RS_STARTED){
            qDebug() << "Some client failed to start";
            isOk = false;
        }
    }

    if(isOk) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 2000);
        ui->trigMainForm->openTrig();
        QString txt = QString("Run started: %1")
                          .arg(QString(config.pc.runId));
        ui->labelRunState->setText(txt);
        qInfo() << "!!! Run started !!!";
        ui->pushButtonStopAll->setEnabled(true);
        QVector<quint64> devSerials;
        if(!megamaster_flg)
            devSerials << config.rcTrigConfig.rcSerial;
        evNumChecker->set_dev_serials(devSerials);
        if(megamaster_flg)
            evNumChecker->startCheck();
        emit slowControlStateUpdated(SC_STATE_RUNNING, txt);
    } else {
        hasRunError = true;
        qCritical() << "!!! Run failed !!!";
        QString txt = QString("Run failed: %1")
                          .arg(QString(config.pc.runId));
        ui->labelRunState->setText(txt);
        if(config.pc.runEnable)
            perfomeStop(true);
        emit slowControlStateUpdated(SC_STATE_FAULT, txt);
    }
    ui->labelRunState->setStyleSheet(isOk ? "QLabel { color : green;}" : "QLabel { color : red;}");
    updateStatStopButton();
    blockControl(false);

    if(isOk){
        config_timer->start();
    }
    return isOk;
}

bool MainWindow::perfomeStop(bool isError)
{
    blockControl(true);
    bool isOk=true;
    ui->pushButtonStopAll->setEnabled(false);
    ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
    if(!hasRunError)
        ui->labelRunState->setText(QString("Stopping Run %1...")
                                   .arg(QString(config.pc.runId)));
    QString msg = QString("!!! Stopping Run %1... !!! ")
                      .arg(QString(config.pc.runId));
    update();
    if(isError)
        qWarning() << msg;
    else
        qInfo() << msg;
    config.pc.runEnable = false;
    config.write_config();
    ui->clientManagerWidget->set_run_enable(config.pc.runEnable);
    emit slowControlStateUpdated(SC_STATE_INIT, "Stoping...");

    evNumChecker->stopCheck();
    ui->trigMainForm->closeTrig();
    // TODO: Smart wait
    if(remoteConnected) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 1000);
    } else {
        QTime t;
        t.start();
        int elapsedTime;
        while((elapsedTime=t.elapsed())<1000){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 1000-elapsedTime);
            QThread::msleep(100);
        }
    }
    emit runStop();

    QElapsedTimer timer;
    timer.start();
    // Get stoped state from Tcp Clients (sub-programms)
    while(!timer.hasExpired(10000)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if(ui->clientManagerWidget->get_run_id().runNumber==config.pc.runId.runNumber &&
            // FIXME: compare runId, not number
            ui->clientManagerWidget->getRunState() != RS_STOPPING )
            break;
    }
    if(ui->clientManagerWidget->getRunState() != RS_STOPPED){
        qDebug() << "Some client failed to stop";
        isOk = false;
    }

    if(isOk) {
        if(!hasRunError){
            ui->labelRunState->setText(QString("Run %1 %2")
                                       .arg(QString(config.pc.runId))
                                       .arg(isError ? "crashed" : "stopped"));
        }
        if(isError)
            qWarning() << "!!! Run stoped !!!";
        else
            qInfo() << "!!! Run stoped !!!";
    } else {
        if(!hasRunError){
            hasRunError = true;
            ui->labelRunState->setText(QString("Run stop failed: %1")
                                       .arg(QString(config.pc.runId)));
        }
        qCritical() << "!!! Run stop failed !!!";
        emit runStop();
    }
    emit slowControlStateUpdated(hasRunError ? SC_STATE_FAULT : SC_STATE_STANDBY, "Run stopped");
    ui->labelRunState->setStyleSheet((isOk && !isError) ? "QLabel { color : green;}" : "QLabel { color : red;}");
    updateStatStopButton();

    blockControl(false);
    return isOk;
}

void MainWindow::updateDeviceLabel()
{
    QString devLabelStr;
    if(config.rcTrigConfig.rcSerial) {
        const QString serialStr = QtMregDevice::getSerialIdStr(config.rcTrigConfig.rcSerial);
        devLabelStr = QString("%1 %2")
            .arg(getDeviceTypeName(config.rcTrigConfig.rcDeviceId))
            .arg(serialStr);
        if(config.rcTrigConfig.hostStr.isEmpty())
            devLabelStr.append(QString(", IP %3").arg(config.rcTrigConfig.hostStr));
        else
            devLabelStr.append(", no IP");
    } else {
        devLabelStr = "No selected dev.";
    }
    labelDevice->setText(devLabelStr);
}

void MainWindow::trigStatusUpdated(const TrigStatus &status)
{
    labelTemp->setText(QString::number(status.temp, 'f', 1).append(QChar(0x02da)).append('C'));
    labelFirmware->setText(QString("f/w:%1").arg(status.fwStr));
    if(ui->actionCheck_LVDS_parity->isEnabled() != status.supportLvdsParity)
        ui->actionCheck_LVDS_parity->setEnabled(status.supportLvdsParity);
    if(!status.connected && config.pc.runEnable) {
        onRunError(config.pc.runId, QString(status.ident + " disconnected"));
    }

    if(status.connectStChange) {
        if(labelDevice->styleSheet().isEmpty() != status.connected)
            labelDevice->setStyleSheet(status.connected ? "" : "QLabel { color : red;}");
        config.rcTrigConfig.hostStr = status.hostStr;
        config.rcTrigConfig.hostTs = QDateTime::currentDateTime();
        config.write_config();
        updateDeviceLabel();
    }
    lastTrigStatus = status;
}

void MainWindow::setTrigStatus(QString status)
{
    labelState->setText(status);
}

void MainWindow::onRunError(const RcRunId &runId, QString errStr)
{
    if(config.pc.runEnable && config.pc.runId == runId) {
        if(remoteConnected){
            QString mess = QString("Retransmit runError to the remoteControl in Run %1")
                    .arg(QString(config.pc.runId));
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning() << mess;
            emit sendRunError(runId, errStr);
        } else {
            QString mess = QString("Errors in Run %1")
                    .arg(QString(config.pc.runId));
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning()<< mess;

            perfomeStop(true);
        }
//    } else {
//        qWarning() << QString("Got RunError from %1 run").arg(config.pc.runEnable?"previous":"stopped");
    }
}

void MainWindow::onRunWarning(const RcRunId &runId, QString warnStr)
{
    if(!warnStr.isEmpty() &&
            (!config.pc.runEnable || config.pc.runId != runId) )
        return;

    QLabel *lab=nullptr;
    QLabel *label=nullptr;
    if(sender() == ui->clientManagerWidget) {
        lab = ui->labelWarClients;
        label = ui->labelWarnClients;
    } else if(sender() == ui->trigMainForm) {
        lab = ui->labelWarTrigDev;
        label = ui->labelWarnTrigDev;
    } else if(sender() == ui->vmeTrigsWidget) {
        lab = ui->labelWarTmwr;
        label = ui->labelWarnTmwr;
    }
    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    label->setText(warnStr);
    label->setHidden(warnStr.isEmpty());
    lab->setHidden(label->isHidden());

    updateGroupBoxWarn();
}

void MainWindow::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(config.pc.runEnable && config.pc.runId == runId) {
        if(remoteConnected) {
            QString mess = QString("Retransmit backPressure(%2) to the remoteControl in Run %1")
                    .arg(QString(config.pc.runId))
                    .arg(backPressure);
            if(backPressure)
                qWarning() << mess;
            else
                qInfo() << mess;
            emit sendBackPressure(runId, backPressure);
        } else {
            if(backPressure) {
                ui->trigMainForm->closeTrig();
                qWarning()<< "BackPressure enabled";
            } else {
                ui->trigMainForm->openTrig();
                qInfo()<< "BackPressure disabled";

            }
        }
    }
}

void MainWindow::onVmeFlagUpdated()
{
    emit sendVmeFlag(config.hasVme());
}

void MainWindow::addLog(QtMsgType severity, const QString &log, const QString &thrName)
{
    if (!ui)
        return;
    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<<"";
    }

    if((ui->tabWidget->tabBar()->currentIndex() != TAB_LOGS)
            && (severity==QtCriticalMsg || severity==QtWarningMsg)) {
        setTabState(false, TAB_LOGS);
    }

    QTextCharFormat tf = ui->plainTextEditLog->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEditLog->setCurrentCharFormat(tf);
    QString str = log;
    if(!thrName.isEmpty())
        str.prepend(QString("[%1] ").arg(thrName));
    str.prepend(QString("[%1] ").arg(QTime::currentTime().toString()));
    ui->plainTextEditLog->appendPlainText(str);
//    ui->plainTextEditLog->appendPlainText(QString("[%1] [%2] %3")
//                                          .arg(QTime::currentTime().toString(), thrName, log));
}

void MainWindow::loadSettings()
{
    config.read_config();
    setTabState(false, TAB_TRIG_DEV);
    updateDeviceLabel();
    if(config.rcTrigConfig.rcSerial) {
        quint32 ipAddr = 0;
        if(!config.rcTrigConfig.hostStr.isEmpty()){
            QHostAddress host(config.rcTrigConfig.hostStr);
            if(!host.isNull())
                ipAddr = host.toIPv4Address();
        }
        DeviceDescription dd(config.rcTrigConfig.rcDeviceId, config.rcTrigConfig.rcSerial, ipAddr);
        connectToTrig(dd);
    }

    ui->clientManagerWidget->init(config.clients);

    ui->trigMainForm->init(&config);
    ui->vmeTrigsWidget->init(&config);

    ui->labelRunState->setText(QString("Previous Run %1")
                               .arg(QString(config.pc.runId)));
    bool oldState1 = ui->actionAllow_Remote_Control->blockSignals(true);
    ui->actionAllow_Remote_Control->setChecked(config.pc.allowRemoteControl);
    ui->actionAllow_Remote_Control->blockSignals(oldState1);
    ui->actionEnableStatusDump->setChecked(config.pc.enableStatusDump);
    ui->actionCheck_LVDS_parity->setChecked(config.rcTrigConfig.checkLvdsParity);
    emit setStatusDumpEnable(config.pc.enableStatusDump);
    emit setStatusDumpDirName(config.pc.statusDumpDirName);
    emit allowRemoteControl(config.pc.allowRemoteControl);

    if(config.cs.getIndex() == "Glob") {
        emit allowSlowControl(true);
    } else {
#ifdef QT_DEBUG
//        emit allowSlowControl(true);
#endif
    }
    updateProgramDescription();
}

void MainWindow::connectTrigDevice(TrigDevice *trigDevice)
{
    connect(this, &MainWindow::deleteHardware,
            trigDevice, &TrigDevice::deleteHardware);

    connect(discoverDialog, &DiscoverDialog::deviceDiscovered,
            trigDevice, &TrigDevice::deviceUpdated);
    connect(this, &MainWindow::deviceDiscovered,
            trigDevice, &TrigDevice::deviceUpdated);

    connect(trigDevice, &TrigDevice::chNumberUpdated,
            ui->trigMainForm, &TrigMainForm::setChNumber);

    connect(trigDevice, &TrigDevice::trigSrcCntUpdated,
            ui->trigMainForm, &TrigMainForm::trigSrcCntUpdated);
    connect(trigDevice, &TrigDevice::setStatus,
            this, &MainWindow::setTrigStatus);
    connect(trigDevice, &TrigDevice::updateStatus,
            this, &MainWindow::trigStatusUpdated);
    connect(trigDevice, &TrigDevice::updateStatus,
            ui->trigMainForm, &TrigMainForm::trigStatusUpdated);
    connect(trigDevice, &TrigDevice::updateStatus,
            slowControl, &SlowControlServer::trigStatusUpdated);

    connect(trigDevice, &TrigDevice::spillPlotUpdated,
            ui->trigMainForm, &TrigMainForm::spillPlotUpdated);
    connect(trigDevice, &TrigDevice::spillPlotUpdated,
            slowControl, &SlowControlServer::spillPlotUpdated);

    connect(trigDevice, &TrigDevice::triggerHistogramUpdated,
            ui->trigMainForm, &TrigMainForm::triggerHistogramUpdated);
    connect(trigDevice, &TrigDevice::triggerHistogramUpdated,
            slowControl, &SlowControlServer::triggerHistogramUpdated);
    connect(this, &MainWindow::setStatusDumpEnable,
            trigDevice, &TrigDevice::setStatusDumpEnable);
    connect(this, &MainWindow::setStatusDumpDirName,
            trigDevice, &TrigDevice::setStatusDumpDirName);

    connect(ui->trigMainForm, &TrigMainForm::trigSetupUpdated,
            trigDevice, &TrigDevice::writeSetup);
//    connect(trigDevice, &TrigDevice::setupWritten,
//            this, &MainWindow::writeSetup);
    connect(trigDevice, &TrigDevice::updateTabState,
            this, &MainWindow::setTabState);
    connect(ui->trigMainForm, &TrigMainForm::runStart,
            trigDevice, &TrigDevice::onRunStart);
    connect(ui->trigMainForm, &TrigMainForm::clearTrigStats,
            trigDevice, &TrigDevice::clearStats);
    connect(ui->trigMainForm, &TrigMainForm::resetAll,
            trigDevice, &TrigDevice::resetAll);

    connect(evNumChecker, &EvNumChecker::getTrigEvNum,
            trigDevice, &TrigDevice::evNumRequest);
    connect(trigDevice, &TrigDevice::trigEvNumResponse,
            evNumChecker, &EvNumChecker::gotTopLevelEvNum);
    connect(trigDevice, &TrigDevice::devEvNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum);
}

void MainWindow::connectToTrig(const DeviceDescription &dd)
{
    if(trigDevice){
        emit deleteHardware(trigDevice->getDeviceIndex());
        setTabState(false, TAB_TRIG_DEV);
        setTrigStatus("disconnected");
        labelFirmware->setText("f/w:unknown");
        labelTemp->clear();
    }
    if(config.rcTrigConfig.rcDeviceId == 0 || config.rcTrigConfig.rcSerial == 0)
        return;

    trigDevice = new TrigDevice(config.rcTrigConfig);
    connectTrigDevice(trigDevice);

    auto tc = new ThreadController(trigDevice, this);
    connect(static_cast<QObject *>(trigDevice), &QObject::destroyed,
            tc, &QObject::deleteLater);

    // Connecting
    if(!dd.isNull())
        emit deviceDiscovered(dd);
}

void MainWindow::initProgramDescription()
{
    progDesc.type = config.cs.getTypeName();
    progDesc.index = config.cs.getIndex();
    progDesc.name = progDesc.getName();
}

void MainWindow::updateProgramDescription()
{
    QString remoteState;

    progDesc.options.clear();
    progDesc.options["runEnable"] = QVariant(config.pc.runEnable).toString();
    progDesc.options["run_index"] = config.pc.runId.runIndex;
    progDesc.options["run_number"] = QVariant(config.pc.runId.runNumber).toString();

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    bool readyForRun = isReadForRun();
    for(const ClientConfig &c: config.clients.values()) {
        remoteState += QString("%1 #%2; ").arg(c.program_type).arg(c.program_index);
        /*
        progDesc.options[QString("client_%1_progType").arg(c.id)] = c.progType;
        progDesc.options[QString("adc_%1_progIndex").arg(c.id)] = c.progIndex;
        progDesc.options[QString("adc_%1_readout").arg(c.id)] = QVariant(c.readout).toString();
        */
        progDesc.options[QString("Client-%1").arg(c.id)] = c.program_type +" "+ c.program_index +" , "+ QVariant(c.readout).toString();
    }

    int n=0;
    for(const VmeTrigConfig &vme: config.vmeTrigs.values()) {
        if(vme.readout)
            remoteState += QString("tmwr:%1; ").arg(QtMregDevice::getSerialIdStr(vme.serial));
        progDesc.options[QString("vmeTrigModule_%1_serial").arg(n)] = QtMregDevice::getSerialIdStr(vme.serial);
        progDesc.options[QString("vmeTrigModule_%1_readout").arg(n)] = QVariant(vme.readout).toString();
        progDesc.options[QString("vmeTrigModule_%1_checkWr").arg(n)] = QVariant(vme.checkWr).toString();
    }

    emit changeProgramDescription(progDesc);
    emit sendRemoteState(remoteState, readyForRun);
    emit sendVmeFlag(config.hasVme());
}

void MainWindow::updateStatStopButton()
{
    ui->pushButtonRestartAll->setEnabled(!config.pc.runEnable && !remoteConnected);
    ui->pushButtonStopAll->setEnabled(config.pc.runEnable && !remoteConnected);
}

bool MainWindow::isReadForRun() const
{
    bool allOk=true;
    auto tabIt = tabOk.begin();
    while(tabIt != tabOk.end()) {
        if(tabIt.key() != TAB_LOGS)
            allOk &= tabIt.value();
        ++tabIt;
    }
    return allOk;
}

void MainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!ui->labelWarnTrigDev->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTrigDev->text();
    }
    if(!ui->labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnClients->text();
    }
    if(!ui->labelWarnTmwr->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTmwr->text();
    }

    ui->groupBoxWarn->setHidden(hidden);
    if(!hidden)
        emit sendRunWarn(config.pc.runId, warnStrList.join("; "));
}

void MainWindow::updateRemoteBlocking()
{
    ui->trigMainForm->updateRemoteBlocking(remoteConnected);
}

void MainWindow::on_pushButtonRestartAll_clicked()
{
    megamaster_flg = true;
    clients_config.clear();

    const RcRunId newRunId(config.cs.getIndex(), ++config.pc.localRunNumber);
    bool success = perfomeStart(newRunId);

    if(success && config.clients.isEmpty()){
        RootConfig run_root = RootConfig();
        run_root.set_data("program_type", config.cs.program_type);
        run_root.set_data("run_index", config.pc.runId.runIndex);
        run_root.set_data("run_number", config.pc.runId.runNumber);
        QDateTime current_dt = QDateTime::currentDateTime();
        run_root.set_data("date_of_creation", current_dt);

        RootConfig my_root = config.root;
        my_root.item_name = config.cs.program_type + "_" +
                config.cs.program_index + "_" +
                QString::number(config.pc.runId.runNumber); // FIXME: why not runId?

        run_root.append_child(my_root);
        config.write_history_config(run_root);
    }
}

void MainWindow::on_pushButtonStopAll_clicked()
{
    perfomeStop(false);
}

void MainWindow::on_actionFindDevice_triggered()
{
    findDevice();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.pc.allowRemoteControl = val;
    config.write_config();
    emit allowRemoteControl(config.pc.allowRemoteControl);
    updateProgramDescription();
}

void MainWindow::on_actionEnableStatusDump_toggled(bool val)
{
    config.pc.enableStatusDump = val;
    config.write_config();
    emit setStatusDumpEnable(config.pc.enableStatusDump);
}

void MainWindow::on_actionSetStatusDumpDirectory_triggered()
{
    QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"),
                                                    config.pc.statusDumpDirName,
                                                    QFileDialog::ShowDirsOnly
                                                    | QFileDialog::DontResolveSymlinks);
    if (dir.isNull())
        return;
    config.pc.statusDumpDirName = dir;
    config.write_config();
    emit setStatusDumpDirName(dir);
}

void MainWindow::on_tabWidget_currentChanged(int index)
{
    if(index == TAB_LOGS && !tabOk.value(TAB_LOGS, true))
        setTabState(true, TAB_LOGS);
}
