#include "SlowControlServer.h"

#include <QTcpServer>
#include <QTimer>
#include <QtNetwork>

#include "SlowControlServer_p.h"
#include "TrigStatus.h"

namespace {
const quint16 SLOW_CONTROL_PORT = 33306;
}

SlowControlServer::SlowControlServer(QObject *parent) :
    QObject(parent),
    d_ptr(new SlowControlServerPrivate()),
    qTcpServer(new QTcpServer(this)),
    tcpPort(SLOW_CONTROL_PORT)
{
    responce();

#ifdef QT_DEBUG
    QTimer *periodicResponceTimer = new QTimer(this);
    connect(periodicResponceTimer, &QTimer::timeout,
            this, &SlowControlServer::responce);
    periodicResponceTimer->start(1000);
#endif
}

SlowControlServer::~SlowControlServer()
{
    blockSignals(true);
    startListen(false);
    delete qTcpServer;
    blockSignals(false);
}

void SlowControlServer::startListen(bool state)
{
    listening = state;
    if(listening){
        init();
    } else {
        qTcpServer->close();
        if(tcpSocket){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        qDebug() << QString("TcpServer closed on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    }
    updateStatus();
}

void SlowControlServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        if(!tcpSocket){
            tcpSocket = pendingConnection;
            portInUse=true;
            connect(tcpSocket, &QAbstractSocket::readyRead,
                    this, &SlowControlServer::readCommandRequest);
            connect(tcpSocket, &QAbstractSocket::disconnected,
                    this, &SlowControlServer::onDisconnected);
            QHostInfo info = QHostInfo::fromName(tcpSocket->peerAddress().toString());
            QString msg = QString("Accepted connection from %1:%2")
                    .arg(info.hostName())
                    .arg(tcpSocket->peerPort());
            qInfo() << msg;
            updateStatus();
#ifdef QT_DEBUG
responce();
#endif
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(!tcpSocket){
        qDebug()<<"There are no connections.";
    }
}

void SlowControlServer::readCommandRequest()
{
    if(!tcpSocket)
        return;
    while(tcpSocket->bytesAvailable()){
        QByteArray lineData = tcpSocket->readLine();
        QString line = QString::fromLatin1(lineData);
        if(line.startsWith("a")){
            responce();
        }
    }
}

void SlowControlServer::onDisconnected()
{
    portInUse=false;
    updateStatus();
    qInfo() << "Disconnected";
}

void SlowControlServer::slowControlStateUpdated(QString _state, QString _status)
{
    Q_D(SlowControlServer);
    d->daqState = _state;
    d->daqStatus = _status;
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::trigStatusUpdated(const TrigStatus &status)
{
    Q_D(SlowControlServer);
    d->rcState = (status.configValid && status.liveMagikOk) ? SC_STATE_ON : SC_STATE_FAULT;
    d->rcDescr = status.ident;
    d->spillCount = status.spillCount;
    d->evNumber = status.eventNumber();
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::spillPlotUpdated(const SpillPlotData &spillPlot)
{
    Q_D(SlowControlServer);
    d->lastSpillPlot = spillPlot;
}

void SlowControlServer::triggerHistogramUpdated(const TriggerHistogramData &trigData)
{
    Q_D(SlowControlServer);
    d->lastTrigData = trigData.histograms.isEmpty() ?
                TriggerHistogramChannelData() :
                trigData.histograms.last(); // fbp.multihist ? [15] : [0]
}

void SlowControlServer::onRunStart(const RcRunId &runId)
{
    Q_D(SlowControlServer);
    d->runId.set(runId);
}

void SlowControlServer::init()
{
    if(!qTcpServer->isListening()) {
        if(!qTcpServer->listen(QHostAddress::Any, tcpPort)) {
            QString s = tr("Unable to start on port %1: %2").arg(tcpPort).arg(qTcpServer->errorString());
            qCritical() << s;
            throw std::runtime_error(s.toStdString());
        }
    }
    tcpPort = qTcpServer->serverPort();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &SlowControlServer::updateConnections);
    qInfo() << QString("Listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void SlowControlServer::updateStatus() const
{
    emit connectionStatusUpdated(tcpPort, listening, portInUse);
}

void SlowControlServer::responce() const
{
    Q_D(const SlowControlServer);

    if(!tcpSocket)
        return;

    tcpSocket->write(d->constructResponse());
}
