//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TrigMainForm.h"
#include "ui_TrigMainForm.h"
#include "RcConfig.h"
#include "TrigDevice.h"
#include "RunControlRegs.h"
#include "TrigChannelConfig.h"
#include "TrigChannelState.h"
#include "metric-sender/MetricSender.h"
#include "rc_gui_utils.h"

#include <QComboBox>
#include <QHostAddress>
#include <QMessageBox>
#include <QThread>

namespace {
enum TRIG_STAT_COLS {
//    TRIG_STAT_COL_CH_NUM,
//    TRIG_STAT_COL_TRIG_IN,
    TRIG_STAT_COL_TIME_STBY,
    TRIG_STAT_COL_TIME_RUN,
    TRIG_STAT_COL_TIME_READOUT,
    TRIG_STAT_COL_TIME_XOFF,
//    TRIG_STAT_COL_STATE,
    TRIG_STAT_COL_TRIG,
    TRIG_STAT_COL_TRIG_RUN,
    TRIG_STAT_COL_TRIG_SPILL,
    TRIG_STAT_COL_TRIG_RUN_SPILL,
    TRIG_STAT_COL_TRIG_ACCEPTED,
    TRIG_STAT_COL_PARITY_ERR,
    TRIG_STAT_COL_L0_CNT,
    TRIG_STAT_COL_BEFORE_ERR,
    TRIG_STAT_COL_AFTER_ERR,
    TRIG_STAT_COL_TOTAL
};
inline bool trigInMoreThatOne(RcConfig * config) {
    return config->rcTrigConfig.rcDeviceId == DEVICE_ID_UT24VE;
}
//const int BEFORE_PROT_BASE=32;
//const int AFTER_PROT_BASE=64;
const int BEFORE_PROT_BASE=0;
const int AFTER_PROT_BASE=0;

} // namespace

TrigMainForm::TrigMainForm(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::TrigMainForm),
    cfgSeqNumber(0),
    prevTrigStatus(new TrigStatus),
    controlEnabled(true)
{
    ui->setupUi(this);
    trigChannelConfig = new TrigChannelConfig(ui->tableWidgetChCfg);
    qRegisterMetaType<TrigChannelState>();
    qRegisterMetaType<SpillPlotData>();
    qRegisterMetaType<TriggerHistogramChannelData>();
    qRegisterMetaType<TriggerHistogramData>();

//    connect(ui->trigConfigForm, &TrigConfigForm::configUpdated,
//            this, &TrigMainForm::configUpdated);
    connect(trigChannelConfig, &TrigChannelConfig::configUpdated,
            this, &TrigMainForm::configUpdated);
//    connect(this, &TrigMainForm::updateLeds,
//            ui->trigLedForm, &TrigLedForm::updateLeds);
    connect(this, &TrigMainForm::updateChState,
            trigChannelConfig, &TrigChannelConfig::updateChState);

    if (ui->tableWidgetStats->rowCount() != TRIG_STAT_COL_TOTAL)
        ui->tableWidgetStats->setRowCount(TRIG_STAT_COL_TOTAL);
    for(int row=0;row<TRIG_STAT_COL_TOTAL; ++row)
        ui->tableWidgetStats->setItem(row, 0, new QTableWidgetItem());
#if QT_VERSION >= 0x050000
    ui->tableWidgetStats->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui->tableWidgetStats->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    ui->tableWidgetStats->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    ui->tableWidgetStats->verticalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
    ui->radioButtonSpillSourceWr->hide();
    ui->lineEditPeriodWr->hide();
    ui->spinBoxBeforeWidth->setMinimum(BEFORE_PROT_BASE);
    ui->spinBoxAfterWidth->setMinimum(AFTER_PROT_BASE);
    // hide phase 1, 2
    ui->labelAcceleration->hide();
    ui->spinBoxSpillDurPhase1->hide();
    ui->labelPrepare->hide();
    ui->spinBoxSpillDurPhase2->hide();

    connect(this, &TrigMainForm::trigSrcCntUpdated,
            ui->trigHistView, &TrigHistView::trigSrcCntUpdated);
    connect(this, &TrigMainForm::spillPlotUpdated,
            ui->trigHistView, &TrigHistView::spillPlotUpdated);
    connect(this, &TrigMainForm::triggerHistogramUpdated,
            ui->trigHistView, &TrigHistView::triggerHistogramUpdated);
    connect(ui->trigHistView, &TrigHistView::rcConfigUpdated,
            this, &TrigMainForm::writeConfig);
}

TrigMainForm::~TrigMainForm()
{
    delete prevTrigStatus;
    delete ui;
}

void TrigMainForm::init(RcConfig *conf)
{
    config = conf;
    trigChannelConfig->init(config);
    ui->trigHistView->init(config);
    restoreGuiFromConfig();
}

void TrigMainForm::updateExtTrigGroup()
{
    bool moreThanOne = trigInMoreThatOne(config);
    ui->checkBoxTrigIn->setText(QString("External TTL")
                                .append(moreThanOne ? " 1" : ""));
    ui->checkBoxTrigIn2->setVisible(moreThanOne);
    ui->checkBoxTrigIn3->setVisible(moreThanOne);
    ui->checkBoxTrigIn4->setVisible(moreThanOne);
}

double TrigMainForm::timeFromClocks(quint64 t)
{
    return (double)t/TrigDevice::getClockFreqHz();
}

void TrigMainForm::updateRemoteBlocking(bool remoteConnected)
{
    if(remoteConnected) {
        config->rcTrigConfig.spill.source = SpillConfig::SPILL_GATE;
        config->rcTrigConfig.trig.trigSrc = trigInMoreThatOne(config) ? 0x8 : 1;
        restoreGuiFromConfig();
    }
    ui->groupBoxTrigSrc->setEnabled(!remoteConnected);
    ui->groupBoxTrigState->setEnabled(!remoteConnected);
    ui->groupBoxSoftSpill->setEnabled(!remoteConnected);
}

QString TrigMainForm::timeStr(quint64 t)
{
    quint64 msec = 1000.0 * timeFromClocks(t);
    quint64 sec = msec/1000;
    quint64 min = sec/60;
    quint64 h = min/60;
    QString str = QString("%1:%2:%3.%4")
            .arg(h)
            .arg(min % 60, 2, 10, QChar('0'))
            .arg(sec % 60, 2, 10, QChar('0'))
            .arg(msec % 1000, 3, 10, QChar('0'));
    return str;
}

void TrigMainForm::sendTrigStatusMetrics(const TrigStatus &prevTrigStatus, const TrigStatus &status)
{
    // Gauges - Summary
    MetricSender::send("spillNumber", status.spillCount);
    MetricSender::send("eventNumber", status.eventNumber());

    MetricSender::send("errors.lvdsParity", status.parityErrorCnt - prevTrigStatus.parityErrorCnt);
    // Gauges - Spill Timers
    MetricSender::send("spill.period.current",   status.spillCurrentTimeMs);
    MetricSender::send("spill.period.TTL", status.spillTtlLastPeriodMs);
// TODO: MetricSender::send("spill.period.WR", status.spillWrLastPeriodMs);
    // WR
    MetricSender::send("errors.wr.other", status.wrOtherErrorCount - prevTrigStatus.wrOtherErrorCount);
    MetricSender::send("errors.wr.code",  status.wrCodeErrorCount  - prevTrigStatus.wrCodeErrorCount);
    MetricSender::send("errors.wr.linkDown",   status.wrLinkLostCount   - prevTrigStatus.wrLinkLostCount);
    MetricSender::send("errors.wr.syncLoss",   status.wrSyncLostCount   - prevTrigStatus.wrSyncLostCount);
    // other
    MetricSender::send("temp", status.temp);
    // per-channel counters
    MetricSender::send("triggers.All",      status.trigAll()      - prevTrigStatus.trigAll());
//        MetricSender::send("triggers.Spill",    status.trigOnSpill()  - prevTrigStatus.trigOnSpill());
    MetricSender::send("triggers.Accepted", status.trigAccepted() - prevTrigStatus.trigAccepted());
    MetricSender::send("triggers.Missed",   status.trigMissed()   - prevTrigStatus.trigMissed());
    if (status.isL0Module) {
        MetricSender::send("triggers.L0",   status.l0RegCnt[1] - prevTrigStatus.l0RegCnt[1]);
        MetricSender::send("triggers.BeforeProtection",   status.l0RegCnt[2] - prevTrigStatus.l0RegCnt[2]);
        MetricSender::send("triggers.AfterProtection",   status.l0RegCnt[3] - prevTrigStatus.l0RegCnt[3]);
    }

    MetricSender::send("time.Clock",   timeFromClocks(status.timeClock()   - prevTrigStatus.timeClock()));
    MetricSender::send("time.Run",     timeFromClocks(status.timeRun()     - prevTrigStatus.timeRun()));
    MetricSender::send("time.Spill",   timeFromClocks(status.timeSpill()   - prevTrigStatus.timeSpill()));
    MetricSender::send("time.Readout", timeFromClocks(status.timeReadout() - prevTrigStatus.timeReadout()));
    MetricSender::send("time.Standby", timeFromClocks(status.timeStandby() - prevTrigStatus.timeStandby()));
    MetricSender::send("time.Xoff",    timeFromClocks(status.timeXoff()    - prevTrigStatus.timeXoff()));

    for(int ch=0; ch<status.nch; ch++) {
        bool xOff = config->rcTrigConfig.trig.xOffMask&(1<<ch);
        if (!xOff)
            continue;
        qint64 events = status.xOffEventCnt[ch] - prevTrigStatus.xOffEventCnt[ch];
        qint64 clocks = status.cnt[CNT_INDEX_TIME_RUN_SPILL][ch] - prevTrigStatus.cnt[CNT_INDEX_TIME_RUN_SPILL][ch];
        double time = timeFromClocks(clocks);
        if (events == 0 && clocks == 0) continue; // unused channel
        double eventTime = (events > 0) ? time / events : 0.;
        QString descr = config->channelDescr[ch];
        if (descr.isEmpty()) descr = QString("%1").arg(ch);
        MetricSender::send(QString("channel.%1.%2.time_xoff").arg(ch).arg(config->channelDescr[ch]), eventTime);
    }
}

void TrigMainForm::trigStatusUpdated(const TrigStatus &status)
{
    QTableWidget *tableStats = ui->tableWidgetStats;
    if(status.isL0Module != prevTrigStatus->isL0Module) {
        tableStats->setRowHidden(TRIG_STAT_COL_L0_CNT, !status.isL0Module);
        tableStats->setRowHidden(TRIG_STAT_COL_BEFORE_ERR, !status.isL0Module);
        tableStats->setRowHidden(TRIG_STAT_COL_AFTER_ERR, !status.isL0Module);
        ui->checkBoxTrigLvds->setVisible(status.isL0Module);
        ui->groupBoxTrigWord->setVisible(status.isL0Module);
        const bool trigLvds = config->rcTrigConfig.trig.trigSrc & TrigConfig::TrigMaskLvds;
        ui->spinBoxLvdsMask->setVisible(status.isL0Module && trigLvds);
        ui->labelLvdsMask->setVisible(status.isL0Module && trigLvds);
    }
    if (status.spillCount < prevTrigStatus->spillCount) {
        prevTrigStatus->clear();
    }
    if (status.spillCount > prevTrigStatus->spillCount) {
        sendTrigStatusMetrics(*prevTrigStatus, status);
        *prevTrigStatus = status;
    }
    QString digitStyle = "font-weight: bold; font-family: courier; font-size: 20pt";
    {
        QString text = QString("Run    %1")
                .arg(QString(config->pc.runId));
        QString curText = ui->labelRunNumber->text();
        if (text != curText) {
            ui->labelRunNumber->setStyleSheet(digitStyle);
            ui->labelRunNumber->setText(text);
        }
    }

    ui->labelSpillNumber->setText(QString("Spill  %1").arg(status.spillCount));
    ui->labelSpillNumber->setStyleSheet(digitStyle);

    ui->labelEventNumber->setText(QString("Event  %L1").arg(status.eventNumber()));
    ui->labelEventNumber->setStyleSheet(digitStyle);

    ui->labelEfficiency->setText(QString("Missed %1 %").arg(status.missRatio()*100.0, 0, 'f', 2));
    ui->labelEfficiency->setStyleSheet(digitStyle);

    TrigChannelState chState;
    chState.globalEventCount = status.eventNumber();
    chState.xOff = status.curXOffState;
    chState.timeout = status.curXOffTimeoutState;
    chState.timeoutXoff = status.xOffHiTimeoutCnt;
    chState.timeoutRead = status.xOffLowTimeoutCnt;
    chState.globalXoffTime = 1.0 * status.timeXoff() / TrigDevice::getClockFreqHz();

    for(int cntType : status.cnt.keys()) {
//        if (cntType % 2) continue; // not a time counter
        if (cntType != CNT_INDEX_TIME_RUN_SPILL) continue; // make only run spill time
        for(int ch : status.cnt[cntType].keys()) {
            chState.eventCount[ch] = status.trigAccepted();
            chState.xOffTime[ch] = 1.0 * status.cnt[cntType][ch] / TrigDevice::getClockFreqHz();
            chState.xOffTimeStr[ch] = timeStr(chState.xOffTime[ch]);
        }
    }
    emit updateChState(chState);
//    emit updateChState(status.curXOffState, status.curXOffTimeoutState,
//                       xOffTimeMap, xOffTimeStrMap,
//                       status.xOffHiTimeoutCnt, status.xOffLowTimeoutCnt);

    //    tableStats->item(TRIG_STAT_COL_TRIG_IN, 0)->setText(QString("%1").arg(status.trigSrc,16,2,QChar('0')));
    qint64 timeXoff = status.timeXoff();
    qint64 timeReadout = status.timeReadout();
    qint64 timeStandby = status.timeStandby();
    qint64 timeRun = status.timeRun();
    tableStats->item(TRIG_STAT_COL_TIME_XOFF,    0)->setText(timeStr(timeXoff));
    tableStats->item(TRIG_STAT_COL_TIME_READOUT, 0)->setText(timeStr(timeReadout));
    tableStats->item(TRIG_STAT_COL_TIME_STBY,    0)->setText(timeStr(timeStandby));
    tableStats->item(TRIG_STAT_COL_TIME_RUN,     0)->setText(timeStr(timeRun));
//    tableStats->item(TRIG_STAT_COL_STATE, 0)->setText(stateStr);
    tableStats->item(TRIG_STAT_COL_TRIG,            0)->setText(QString("%L1").arg(status.trigAll()));
    tableStats->item(TRIG_STAT_COL_TRIG_RUN,        0)->setText(QString("%L1").arg(status.trigOnRun()));
    tableStats->item(TRIG_STAT_COL_TRIG_SPILL,      0)->setText(QString("%L1").arg(status.trigOnSpill()));
    tableStats->item(TRIG_STAT_COL_TRIG_RUN_SPILL,  0)->setText(QString("%L1").arg(status.trigOnRunOnSpill()));
    tableStats->item(TRIG_STAT_COL_TRIG_ACCEPTED,   0)->setText(QString("%L1").arg(status.trigAccepted()));
    tableStats->item(TRIG_STAT_COL_PARITY_ERR,      0)->setText(QString("%L1").arg(status.parityErrorCnt));


    if(status.isL0Module) {
        tableStats->item(TRIG_STAT_COL_L0_CNT,      0)->setText(QString("%L1").arg(status.l0RegCnt[1]));
        tableStats->item(TRIG_STAT_COL_BEFORE_ERR,  0)->setText(QString("%L1").arg(status.l0RegCnt[2]));
        tableStats->item(TRIG_STAT_COL_AFTER_ERR,   0)->setText(QString("%L1").arg(status.l0RegCnt[3]));
    }

    const bool hasBapBlockControl = (status.fwRev >= fw_version_t(1, 0, 25502));
    ui->radioButtonBapCheck->setVisible(hasBapBlockControl);
    ui->radioButtonBapBlock->setVisible(hasBapBlockControl);

    const bool hasDeadTimeControl = (status.fwRev >= fw_version_t(1, 0, 25367));
    ui->labelTrigSourceDeadTime->setVisible(hasDeadTimeControl);
    ui->doubleSpinBoxTrigRandomDeadTime->setVisible(hasDeadTimeControl);

    ui->labelDeadTime->setText(QString("Dead Time %1 %").arg(status.deadTimeRatio()*100.0, 0, 'f', 2));
    ui->labelDeadTime->setStyleSheet(digitStyle);

    int spillPhase;
    switch (status.spillState.reg.bit.phase) {
    case 1: spillPhase = 1; break;
    case 2: spillPhase = 2; break;
    case 4: spillPhase = 3; break;
    case 8: spillPhase = 4; break;
    case 16: spillPhase = 5; break;
    default: spillPhase = 0;
    }
    QString activeStyle = "background-color: green; color:white";
    QString passiveStyle;
    ui->labelSpillPhase0->setStyleSheet(spillPhase == 0 ? activeStyle : passiveStyle);
    ui->labelSpillPhase1->setStyleSheet(spillPhase == 1 ? activeStyle : passiveStyle);
    ui->labelSpillPhase2->setStyleSheet(spillPhase == 2 ? activeStyle : passiveStyle);
    ui->labelSpillPhase3->setStyleSheet(spillPhase == 3 ? activeStyle : passiveStyle);
    ui->labelSpillPhase4->setStyleSheet(spillPhase == 4 ? activeStyle : passiveStyle);
    ui->labelSpillPhase5->setStyleSheet(spillPhase == 5 ? activeStyle : passiveStyle);
    ui->labelSpillPhase0->setEnabled(spillPhase == 0);
    ui->labelSpillPhase1->setEnabled(spillPhase == 1);
    ui->labelSpillPhase2->setEnabled(spillPhase == 2);
    ui->labelSpillPhase3->setEnabled(spillPhase == 3);
    ui->labelSpillPhase4->setEnabled(spillPhase == 4);
    ui->labelSpillPhase5->setEnabled(spillPhase == 5);

    ui->labelSpillPhase1->setVisible(!config->rcTrigConfig.isSpillGate());
    ui->labelSpillPhase2->setVisible(!config->rcTrigConfig.isSpillGate());
//    ui->labelSpillPhase4->setVisible(!config->rcTrigConfig.isSpillGate());
    ui->labelSpillPhase5->setVisible(!config->rcTrigConfig.isSpillGate());

    ui->labelSpillPhase4->setVisible(false);

    displaySpillSource(status);
    if(!status.liveMagikOk){
        qInfo() << QString("%1 Device restart detected - writing config").arg(status.ident);
        emit resetAll();
    }
}

void TrigMainForm::displaySpillSource(const TrigStatus &status)
{
    if (status.spillTtlLastPeriodMs > 0) {
        ui->lineEditPeriodTtl->setText(QString("%1 ms").arg(status.spillTtlLastPeriodMs));
    } else {
        ui->lineEditPeriodTtl->clear();
    }
//    ui->lineEditPeriodTtl->setEnabled(status.spillState.reg.bit.ttl_ok);

    if (status.spillWrLastPeriodMs > 0) {
        ui->lineEditPeriodWr ->setText(QString("%1 ms").arg(status.spillWrLastPeriodMs));
    } else {
        ui->lineEditPeriodWr->clear();
    }
//    ui->lineEditPeriodWr->setEnabled(status.spillState.reg.bit.wr_ok);
    QString spillSourceStr;
    switch (status.spillState.reg.bit.sel) {
    case SpillConfig::SPILL_OFF:   spillSourceStr = "OFF"; break;
    case SpillConfig::SPILL_TIMER: spillSourceStr = "Timer"; break;
    case SpillConfig::SPILL_TTL:   spillSourceStr = "TTL"; break;
    case SpillConfig::SPILL_WR:    spillSourceStr = "WR"; break;
    case SpillConfig::SPILL_GATE:   spillSourceStr = "Gate"; break;
    default: spillSourceStr = "AUTO"; break;
    }
    ui->labelSpillSourceSelected->setEnabled(config->rcTrigConfig.spill.source == SpillConfig::SPILL_AUTO);
    ui->labelSpillSourceSelected->setText(spillSourceStr);
    bool spillSourceOk = true;
    if (status.spillState.reg.bit.sel == SpillConfig::SPILL_TTL && !status.spillState.reg.bit.ttl_ok) spillSourceOk = false;
    if (status.spillState.reg.bit.sel == SpillConfig::SPILL_WR  && !status.spillState.reg.bit.wr_ok) spillSourceOk = false;
    ui->labelSpillSourceSelected->setStyleSheet(spillSourceOk ? "background-color: green; color: white" : "background-color: red; color: white");
}

void TrigMainForm::openTrig()
{
    ui->radioButtonTrigOpen->setChecked(true);
}


void TrigMainForm::closeTrig()
{
    ui->radioButtonTrigClose->setChecked(true);
}

void TrigMainForm::clearStats()
{
    prevTrigStatus->clear();
    emit clearTrigStats(false);
}

void TrigMainForm::setChNumber(int chNum)
{
    config->rcTrigConfig.chNum = chNum;
    writeConfig();

    trigChannelConfig->updateConfig();
}

void TrigMainForm::checkLvdsParityToggled(bool checked)
{
    config->rcTrigConfig.checkLvdsParity = checked;
    writeConfig();

    ui->tableWidgetStats->setRowHidden(TRIG_STAT_COL_PARITY_ERR,
                                       !config->rcTrigConfig.checkLvdsParity);
}

void TrigMainForm::spillPlotStepUpdated()
{
    writeConfig();
}

void TrigMainForm::setControlEnable(bool enabled)
{
    controlEnabled = enabled;
    setEnabled(controlEnabled);
    trigChannelConfig->runUpdated();
}

void TrigMainForm::setTrigSrcMask(quint16 mask, bool checked)
{
    if(checked)
        config->rcTrigConfig.trig.trigSrc |= mask;
    else
        config->rcTrigConfig.trig.trigSrc &= ~mask;
    writeConfig();
}

void TrigMainForm::writeConfig()
{
    ++cfgSeqNumber;
    config->write_config();
    emit trigSetupUpdated(config->rcTrigConfig, cfgSeqNumber);
}

void TrigMainForm::restoreGuiFromConfig()
{

    trigChannelConfig->updateConfig();

    bool oldState1 = ui->spinBoxSpillDurPhase4->blockSignals(true);
    bool oldState2 = ui->spinBoxPeriodTimer->blockSignals(true);
    bool oldState3 = ui->radioButtonSpillSourceAuto->blockSignals(true);
    bool oldState4 = ui->spinBoxBeforeWidth->blockSignals(true);
    bool oldState5 = ui->spinBoxAfterWidth->blockSignals(true);
    bool oldState6 = ui->checkBoxTrigIn->blockSignals(true);
    bool oldState7 = ui->checkBoxTrigPeriod->blockSignals(true);
    bool oldState8 = ui->checkBoxTrigRandom->blockSignals(true);
    bool oldState9 = ui->doubleSpinBoxTrigTimerFreq->blockSignals(true);
    bool oldState10 = ui->doubleSpinBoxTrigRandomFreq->blockSignals(true);
    bool oldState11 = ui->doubleSpinBoxTrigRandomDeadTime->blockSignals(true);


    ui->radioButtonTrigOpen->blockSignals(true);
    if(config->rcTrigConfig.trig.trigOpen)
        ui->radioButtonTrigOpen->setChecked(true);
    else
        ui->radioButtonTrigClose->setChecked(true);
    ui->radioButtonTrigOpen->blockSignals(false);

    bool trigIn = config->rcTrigConfig.trig.trigSrc & 0x1;
    bool trigPeriodical = config->rcTrigConfig.trig.trigSrc & TrigConfig::TrigMaskTimer;
    bool trigRandom = config->rcTrigConfig.trig.trigSrc & TrigConfig::TrigMaskRandom;
    bool trigLvds = config->rcTrigConfig.trig.trigSrc & TrigConfig::TrigMaskLvds;
    ui->checkBoxTrigIn->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x2;
    ui->checkBoxTrigIn2->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x4;
    ui->checkBoxTrigIn3->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x8;
    ui->checkBoxTrigIn4->setChecked(trigIn);
    ui->checkBoxTrigPeriod->setChecked(trigPeriodical);
    ui->checkBoxTrigRandom->setChecked(trigRandom);
    ui->checkBoxTrigLvds->setChecked(trigLvds);
    ui->doubleSpinBoxTrigTimerFreq->setMinimum(TrigDevice::getTrigTimerMinFreqHz());
    ui->doubleSpinBoxTrigTimerFreq->setMaximum(std::min(10e6, TrigDevice::getTrigTimerMaxFreqHz()));
    ui->doubleSpinBoxTrigTimerFreq->setValue(config->rcTrigConfig.trig.trigTimerFreq);
    ui->doubleSpinBoxTrigRandomFreq->setMinimum(TrigDevice::getTrigTimerMinFreqHz());
    ui->doubleSpinBoxTrigRandomFreq->setMaximum(std::min(10e6, TrigDevice::getTrigTimerMaxFreqHz()));
    ui->doubleSpinBoxTrigRandomFreq->setValue(config->rcTrigConfig.trig.trigRandomFreq);

    ui->doubleSpinBoxTrigRandomFreq->setMinimum(0);
    ui->doubleSpinBoxTrigRandomFreq->setMaximum(1e6);
    ui->doubleSpinBoxTrigRandomDeadTime->setValue(config->rcTrigConfig.trig.trigRandomDeadTimeUs);

    ui->spinBoxSpillDurPhase1->setMaximum(4000000);
    ui->spinBoxSpillDurPhase2->setMaximum(4000000);
    ui->spinBoxSpillDurPhase3->setMaximum(4000000);
    ui->spinBoxSpillDurPhase4->setMaximum(4000000);
    ui->spinBoxSpillDurPhase5->setMaximum(4000000);
    ui->spinBoxSpillDurPhase1->setValue(config->rcTrigConfig.spill.durPhase1);
    ui->spinBoxSpillDurPhase2->setValue(config->rcTrigConfig.spill.durPhase2);
    ui->spinBoxSpillDurPhase3->setValue(config->rcTrigConfig.spill.durPhase3);
    ui->spinBoxSpillDurPhase4->setValue(config->rcTrigConfig.spill.durPhase4);
    ui->spinBoxSpillDurPhase5->setValue(config->rcTrigConfig.spill.durPhase5);
    ui->labelPostReadout->hide(); ui->spinBoxSpillDurPhase4->hide();

    ui->groupBoxSoftSpill->setChecked(config->rcTrigConfig.spill.softSpill);
    ui->spinBoxPeriodTimer->setValue(config->rcTrigConfig.spill.timerPeriod);

    ui->radioButtonSpillSourceAuto->setChecked(config->rcTrigConfig.spill.source == SpillConfig::SPILL_AUTO);
    ui->radioButtonSpillSourceTimer->setChecked(config->rcTrigConfig.spill.source == SpillConfig::SPILL_TIMER);
    ui->radioButtonSpillSourceTtl->setChecked(config->rcTrigConfig.spill.source == SpillConfig::SPILL_TTL);
    ui->radioButtonSpillSourceWr->setChecked(config->rcTrigConfig.spill.source == SpillConfig::SPILL_WR);
    ui->radioButtonSpillSourceGate->setChecked(config->rcTrigConfig.spill.source == SpillConfig::SPILL_GATE);

    ui->tableWidgetStats->setRowHidden(TRIG_STAT_COL_PARITY_ERR,
                                       !config->rcTrigConfig.checkLvdsParity);
    ui->tableWidgetStats->setRowHidden(TRIG_STAT_COL_L0_CNT,
                                       !prevTrigStatus->isL0Module);
    ui->tableWidgetStats->setRowHidden(TRIG_STAT_COL_BEFORE_ERR,
                                       !prevTrigStatus->isL0Module);
    ui->tableWidgetStats->setRowHidden(TRIG_STAT_COL_AFTER_ERR,
                                       !prevTrigStatus->isL0Module);
    ui->spinBoxLvdsMask->setVisible(prevTrigStatus->isL0Module && trigLvds);
    ui->labelLvdsMask->setVisible(prevTrigStatus->isL0Module && trigLvds);
    ui->checkBoxTrigLvds->setVisible(prevTrigStatus->isL0Module);
    ui->groupBoxTrigWord->setVisible(prevTrigStatus->isL0Module);
    ui->spinBoxLvdsMask->setValue(config->rcTrigConfig.lvdsMask);
    ui->spinBoxTrigWord->setValue(config->rcTrigConfig.trigWordMask);
    ui->checkBoxBeforeProtect->setChecked(config->rcTrigConfig.checkBeforeProtection);
    ui->checkBoxAfterProtect->setChecked(config->rcTrigConfig.checkAfterProtection);
    ui->radioButtonBapCheck->setChecked(!config->rcTrigConfig.bapBlock);
    ui->radioButtonBapBlock->setChecked(config->rcTrigConfig.bapBlock);
    ui->spinBoxTrigRejectDelay->setValue(config->rcTrigConfig.trigRejectDelay*16);
    ui->spinBoxBeforeWidth->setValue(config->rcTrigConfig.beforeProtWidth*16+BEFORE_PROT_BASE);
    ui->spinBoxAfterWidth->setValue(config->rcTrigConfig.afterProtWidth*16+AFTER_PROT_BASE);
    ui->spinBoxBeforeWidth->setEnabled(config->rcTrigConfig.checkBeforeProtection);
    ui->spinBoxAfterWidth->setEnabled(config->rcTrigConfig.checkAfterProtection);


    updateTableSize();
    updateExtTrigGroup();

    ui->spinBoxSpillDurPhase4->blockSignals(oldState1);
    ui->spinBoxPeriodTimer->blockSignals(oldState2);
    ui->radioButtonSpillSourceAuto->blockSignals(oldState3);
    ui->spinBoxBeforeWidth->blockSignals(oldState4);
    ui->spinBoxAfterWidth->blockSignals(oldState5);
    ui->checkBoxTrigIn->blockSignals(oldState6);
    ui->checkBoxTrigPeriod->blockSignals(oldState7);
    ui->checkBoxTrigRandom->blockSignals(oldState8);
    ui->doubleSpinBoxTrigTimerFreq->blockSignals(oldState9);
    ui->doubleSpinBoxTrigRandomFreq->blockSignals(oldState10);
    ui->doubleSpinBoxTrigRandomDeadTime->blockSignals(oldState11);
}

void TrigMainForm::on_pushButtonClearStat_clicked()
{
    clearStats();
}

void TrigMainForm::on_pushButtonResetAll_clicked()
{
    int ret = QMessageBox::question(this, "Confirm ResetAll", "Are you sure to ResetAll settings in Run Control device."
//                                    ,QMessageBox::Yes
//                                    ,QMessageBox::No
                                    );
    if(ret == QMessageBox::Yes) {
        emit resetAll();
    }
}

void TrigMainForm::configUpdated(quint16 xOff)
{
    TrigConfig newConfig(config->rcTrigConfig.trig);
    newConfig.xOffMask = xOff;
    if(config->rcTrigConfig.trig == newConfig)
        return;
    config->rcTrigConfig.trig = newConfig;
    writeConfig();
}

void TrigMainForm::on_radioButtonTrigOpen_toggled(bool open)
{
    config->rcTrigConfig.trig.trigOpen = open;
    writeConfig();
}

void TrigMainForm::on_doubleSpinBoxTrigTimerFreq_valueChanged(double arg1)
{
    config->rcTrigConfig.trig.trigTimerFreq = arg1;
    writeConfig();
}

void TrigMainForm::on_doubleSpinBoxTrigRandomFreq_valueChanged(double arg1)
{
    config->rcTrigConfig.trig.trigRandomFreq = arg1;
    writeConfig();
}

void TrigMainForm::on_doubleSpinBoxTrigRandomDeadTime_valueChanged(double arg1)
{
    config->rcTrigConfig.trig.trigRandomDeadTimeUs = arg1;
    writeConfig();
}

void TrigMainForm::on_checkBoxTrigIn_toggled(bool checked)
{
    setTrigSrcMask(0x1, checked);
}

void TrigMainForm::on_checkBoxTrigIn2_toggled(bool checked)
{
    setTrigSrcMask(0x2, checked);
}

void TrigMainForm::on_checkBoxTrigIn3_toggled(bool checked)
{
    setTrigSrcMask(0x4, checked);
}

void TrigMainForm::on_checkBoxTrigIn4_toggled(bool checked)
{
    setTrigSrcMask(0x8, checked);
}

void TrigMainForm::on_checkBoxTrigLvds_toggled(bool checked)
{
    ui->spinBoxLvdsMask->setVisible(prevTrigStatus->isL0Module && checked);
    ui->labelLvdsMask->setVisible(prevTrigStatus->isL0Module && checked);
    setTrigSrcMask(TrigConfig::TrigMaskLvds, checked);
}

void TrigMainForm::on_checkBoxTrigPeriod_toggled(bool checked)
{
    ui->doubleSpinBoxTrigTimerFreq->setEnabled(checked);
    setTrigSrcMask(TrigConfig::TrigMaskTimer, checked);
}

void TrigMainForm::on_checkBoxTrigRandom_toggled(bool checked)
{
    ui->doubleSpinBoxTrigRandomFreq->setEnabled(checked);
    setTrigSrcMask(TrigConfig::TrigMaskRandom, checked);
}

void TrigMainForm::on_spinBoxSpillDurPhase1_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.durPhase1 = arg1;
    writeConfig();
}

void TrigMainForm::on_spinBoxSpillDurPhase2_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.durPhase2 = arg1;
    writeConfig();
}

void TrigMainForm::on_spinBoxSpillDurPhase3_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.durPhase3 = arg1;
    writeConfig();
}

void TrigMainForm::on_spinBoxSpillDurPhase4_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.durPhase4 = arg1;
    writeConfig();
}

void TrigMainForm::on_spinBoxSpillDurPhase5_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.durPhase5 = arg1;
    writeConfig();
}

void TrigMainForm::updateTableSize()
{
    int height=ui->tableWidgetStats->horizontalHeader()->height();
    for(int i=0;i<ui->tableWidgetStats->rowCount();++i){
        height += ui->tableWidgetStats->rowHeight(i);
    }
    height += 5;
    ui->tableWidgetStats->setMinimumHeight(height);
    QRect geo = ui->tableWidgetStats->geometry();
    geo.setHeight(height);
    ui->tableWidgetStats->setGeometry(geo);
//    qDebug()<<ui->tableWidgetStats <<" height="<<height;
}

void TrigMainForm::on_groupBoxSoftSpill_toggled(bool arg1)
{
    config->rcTrigConfig.spill.softSpill = arg1;
    writeConfig();
}

void TrigMainForm::on_spinBoxPeriodTimer_valueChanged(int arg1)
{
    config->rcTrigConfig.spill.timerPeriod = arg1;
    writeConfig();
}

void TrigMainForm::setSpillMode(const SpillConfig::spill_sel_t &sel)
{
    config->rcTrigConfig.spill.source = sel;
    ui->groupBoxSpillPhases->setVisible(!config->rcTrigConfig.isSpillGate());
    writeConfig();
}

void TrigMainForm::on_radioButtonSpillSourceAuto_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_AUTO);
}

void TrigMainForm::on_radioButtonSpillSourceTtl_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_TTL);
}

void TrigMainForm::on_radioButtonSpillSourceWr_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_WR);
}

void TrigMainForm::on_radioButtonSpillSourceTimer_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_TIMER);
}

void TrigMainForm::on_radioButtonSpillSourceGate_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_GATE);
}

void TrigMainForm::on_spinBoxLvdsMask_valueChanged(int val)
{
    config->rcTrigConfig.lvdsMask = val;
    writeConfig();
}

void TrigMainForm::on_spinBoxTrigWord_valueChanged(int val)
{
    config->rcTrigConfig.trigWordMask = val;
    writeConfig();
}

void TrigMainForm::on_spinBoxTrigRejectDelay_valueChanged(int val)
{
    if(val%16) {
        ui->spinBoxTrigRejectDelay->setValue(val/16*16);
        return;
    }
    config->rcTrigConfig.trigRejectDelay = val/16;
    writeConfig();
}

void TrigMainForm::on_checkBoxBeforeProtect_toggled(bool checked)
{
    config->rcTrigConfig.checkBeforeProtection = checked;
    ui->spinBoxBeforeWidth->setEnabled(checked);
    writeConfig();
}

void TrigMainForm::on_checkBoxAfterProtect_toggled(bool checked)
{
    config->rcTrigConfig.checkAfterProtection = checked;
    ui->spinBoxAfterWidth->setEnabled(checked);
    writeConfig();
}

void TrigMainForm::on_radioButtonBapBlock_toggled(bool checked)
{
    config->rcTrigConfig.bapBlock = checked;
    writeConfig();
}

void TrigMainForm::on_spinBoxBeforeWidth_valueChanged(int val)
{
    if(val%16) {
        ui->spinBoxBeforeWidth->setValue(val/16*16);
        return;
    }
    config->rcTrigConfig.beforeProtWidth = (val-BEFORE_PROT_BASE)/16;
    writeConfig();
}

void TrigMainForm::on_spinBoxAfterWidth_valueChanged(int val)
{
    if(val%16) {
        ui->spinBoxAfterWidth->setValue(val/16*16);
        return;
    }
    config->rcTrigConfig.afterProtWidth = (val-AFTER_PROT_BASE)/16;
    writeConfig();
}
