/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef TRIGSTATUS_H
#define TRIGSTATUS_H

#include <QtGlobal>
#include <QString>
#include <QMap>

#include "mregdev/fw_version.h"
#include "WrEpState.h"
#include "RunControlRegs.h"

struct SpillStateBits
{
    SpillStateBits(quint16 init = 0) {reg.all = init;}
    QString toString() const;
    reg_spill_status_t reg;
};

class TrigStatus
{
public:
    TrigStatus();
    void clear();
    double acceptRatio() const;
    double missRatio() const;
    double deadTimeRatio() const;

    /// First counter index: [0,1] - Total, [2,3] - Run, [4,5] - Spill, [6,7] - Run&Spill
    /// Even - time, odd - edges
    /// cntLevel[*][15] - Elapsed Time
    /// cntLevel[*][14] - XOFF Time

    qint64 trigAll() const {return cnt[1][14];}
    qint64 trigOnRun() const {return cnt[3][14];}
    qint64 trigOnSpill() const {return cnt[5][14];}
    qint64 trigOnRunOnSpill() const {return cnt[7][14];}
    qint64 trigAccepted() const {return cnt[7][15];}
    qint64 trigMissed() const {return trigOnRunOnSpill() - trigAccepted();}
    qint64 timeXoff() const { return cnt[6][14];}

    // Elapsed Time
    qint64 timeClock() const { return cnt[0][15];}
    qint64 timeRun() const { return cnt[2][15];}
    qint64 timeSpill() const { return cnt[4][15];}
    qint64 timeReadout() const { return cnt[6][15];}
    qint64 timeStandby() const {qint64 v = timeClock() - timeRun(); return (v>0) ? v : 0;}
    quint64 eventNumber() const;

    // trig
    quint16 curTrigState;
    quint16 curXOffState;
    quint16 curXOffTimeoutState;
    quint64 parityErrorCnt;
    // spill
    quint32 spillPlotAddr;
    SpillStateBits spillState;
    quint32 spillCount;
    quint32 spillCurrentTimeMs;
    quint32 spillTtlLastPeriodMs;
    quint32 spillWrLastPeriodMs;
    // WR
    WrStatusBits wrStatus;
    WrTaiCode ncuTai;
    quint32 wrUptime;
    quint32 wrOtherErrorCount;
    quint32 wrCodeErrorCount;
    quint32 wrLinkLostCount;
    quint32 wrSyncLostCount;
    QMap<int, QMap<int, qint64> > cnt; // <cntType, <ch, cntVal> >
    QMap<int, quint64> xOffHiTimeoutCnt; // <ch, cntVal>
    QMap<int, quint64> xOffLowTimeoutCnt; // <ch, cntVal>
    QMap<int, quint64> xOffEventCnt; // <ch, cntVal>
    QMap<int, quint64> l0RegCnt; // <ch, cntVal>
    // other
    fw_version_t fwRev;
    int nch;
    bool connected;
    bool connectStChange;
    bool configValid;
    bool supportLvdsParity;
    bool isL0Module;
    bool liveMagikOk;
    QString ident;
    double temp;
    QString fwStr;
    QString hostStr;
    quint16 trigSrc;
    QString hostAddress;

    QMap<int, quint64> xOffEventCnt_base;
    quint64 globEvNum; // since f/w 28815
    quint64 globEvNum_cash; // no need since f/w 28815
};


#endif // TRIGSTATUS_H
