//
//    Copyright 2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TriggerHistogram.h"
#include "ui_TriggerHistogram.h"

#include <QLayout>
#include <QTimer>

#include "TrigDevice.h"
#include "qxw/WaveformPlot.h"
#include "qwt_scale_engine.h"
#include "qwt_plot_zoomer.h"

static const int updatePeriodMs = 100;

TriggerHistogram::TriggerHistogram(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::TriggerHistogram)
{
    ui->setupUi(this);
    plot = new WaveformPlot();
    QLayout *layout = new QVBoxLayout(this);
    layout->addWidget(plot);
//    plot->axisScaleEngine(QwtPlot::yLeft)->setAttributes(QwtScaleEngine::IncludeReference);
    plot->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine());
    plot->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine());
    plot->setAxisTitle(QwtPlot::xBottom, "Busy time and Trigger intervals, μs");
    plot->enableLegend(false);
    QTimer::singleShot(0, this, &TriggerHistogram::checkForReplot);
}

TriggerHistogram::~TriggerHistogram()
{
    delete ui;
}

WaveformData truncateHistData(const TriggerHistogramChannelData &thd)
{
    WaveformData wd0;
    if (thd.data.size() && !thd.isEmpty()) {
        int iMin = thd.getFirstNotemptyBin();
        int iMax = thd.getLastNotemptyBin();
        Q_ASSERT(thd.px.size() == thd.data.size());
        if (iMin > 0) iMin--;
        if (iMax < thd.data.size()-1) iMax++;
        for (int i=iMin; i<=iMax; i++) {
            double x = thd.px[i];
            double y = thd.data[i];
            if (y < 1) y = 0.1; // for logscale
            wd0.p.push_back(point_t(x, y));
        }
    }
    wd0.c = 0;
    return wd0;
}

static const double default_xMin = 0.1;
static const double default_xMax = 10;
static const double default_yMin = 0.1;
static const double default_yMax = 1;

QRectF getBoundingBox(const TriggerHistogramChannelData &thd)
{
    if (thd.px.isEmpty() || thd.data.isEmpty() || thd.isEmpty())
        return QRectF();
    Q_ASSERT(thd.px.size() == thd.data.size());
    double yMin = default_yMin;
    double yMax = default_yMax;
    for (int i=0; i<thd.data.size(); i++) {
        if (thd.data[i] > yMax)
            yMax = thd.data[i];
    }
    double xMin = default_xMin;
    double xMax = default_xMax;
    if (!thd.isEmpty()) {
        xMin = thd.px[thd.getFirstNotemptyBin()];
        xMax = thd.px[thd.getLastNotemptyBin()];
    }
    QPointF topLeft = QPointF(xMax, yMax);
    QPointF bottomRight = QPointF(xMin, yMin);
    return QRectF(topLeft, bottomRight);
}

QRectF getBoundingBox(const TriggerHistogramData &thd)
{
    QRectF box;
    Q_FOREACH(const TriggerHistogramChannelData &thcd, thd.histograms) {
        box = box.united(getBoundingBox(thcd));
    }
    return box;
}

void TriggerHistogram::scheduleReplot()
{
    replot = true;
}

void TriggerHistogram::checkForReplot()
{
    if (replot)
        performReplot();
    QTimer::singleShot(updatePeriodMs, this, &TriggerHistogram::checkForReplot);
}

void TriggerHistogram::performReplot()
{
    QRectF box, box1;
    WaveformDataAll wda;
    // Z-order: first draw previous line, then new line
    if (history.size() > 1) {
        const TriggerHistogramData &previous = history.at(history.size()-2);
        box1 = getBoundingBox(previous);
        //        WaveformData wd1 = truncateHistData(previous.histograms[channel]);
        //        wd1.c = 1;
        //        wda.push_back(wd1);
    }
    if (! history.isEmpty()) {
        const TriggerHistogramData &thd = history.last();
        Q_FOREACH(const TriggerHistogramChannelData &thcd, thd.histograms)
        {
            WaveformData wd0 = truncateHistData(thcd);
            wd0.c = thcd.channel;
            wda.push_back(wd0);
        }
        box = getBoundingBox(thd);
    }

    bool autoReplot = plot->autoReplot();
    plot->setAutoReplot(false);
    plot->setAllData(wda);
    plot->setCurvePen(0, QPen(Qt::blue, 2, Qt::SolidLine));
    plot->setCurvePen(1, QPen(Qt::gray, 1, Qt::SolidLine));
    //    plot->setCurveStyle(0, QwtPlotCurve::Sticks);
    //    plot->setCurveStyle(1, QwtPlotCurve::Steps);
    if (! box1.isNull())
        box = box.united(box1);
    if (box.isNull()) {
        box.setRect(default_xMin, default_yMin, default_xMax-default_xMin, default_yMax-default_yMin);
    }
    double xMin = std::min(box.left(), box.right());
    double xMax = std::max(box.left(), box.right());
    plot->setAxisScale(QwtPlot::xBottom, xMin * 0.8, xMax * 1.2);
    double yMin = std::min(box.top(), box.bottom());
    double yMax = std::max(box.top(), box.bottom());
    plot->setAxisScale(QwtPlot::yLeft, yMin * 0.8, yMax * 1.2);
    plot->setAutoReplot(autoReplot);
}

void TriggerHistogram::updateHistogram(const TriggerHistogramData &thd)
{
    if (history.isEmpty()) {
        history.push_back(TriggerHistogramData()); // initialize
    }
    if (history.last().spillCount != thd.spillCount) {
        history.push_back(thd);
    }
    history.last() = thd;
    while (history.size() > 2) {
        history.pop_front();
    }
    scheduleReplot();
}
