//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "VmeTrigDevice.h"

#include <QDebug>
#include "regio/regop.h"
#include "mregdev/mregdevice.h"
#include "regio/regop.h"
#include "mregdev/fw_version.h"
#include "base/RcRunId.h"

using mlink::RegOp;
using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

namespace {
enum {
    REG_FVME2TMWR_STATUS = 0x201, // 16-bit RO, bit[0] - tm_valid
    REG_FVME2TMWR_RUN_EVENT_NUMBER = 0x204, // 64-bit RO
    REG_FVME2TMWR_WR_LINK_ERROR_COUNTER = 0x100, // Reg 0x100, 32-bit RO counter of PHY code error
    REG_FVME2TMWR_WR_SYNC_LOST_COUNTER  = 0x102, // Reg 0x102, 32-bit RO counter of white rabbit sync lost
    REG64_FVME2TMWR_BUSY_TIMER = 0x300,
    REG64_FVME2TMWR_IDLE_TIMER = 0x310
};

constexpr fw_version_t REV_HOLD_COUNTERS(1, 0, 22708);
}

VmeTrigDevice::VmeTrigDevice(quint64 serId, QObject *parent) :
    QtMregDevice(DEVICE_ID_VME_TRIG, serId, parent),
    serial(serId),
    readout(true)
{
}

void VmeTrigDevice::deleteTriger(quint64 serial)
{
    if(serial==0 || serial==this->serial){
        qDebug()<<"Deleteing vme trigger:"<<hex<<this->serial;
        deleteLater();
    }
}

void VmeTrigDevice::statusPoll()
{
    QtMregDevice::statusPoll();
    status.ident = getIdent();
    status.connected = isConnected();
    status.liveMagikOk = checkLiveMagic();
    if(status.connected){
        status.clear();
        status.temp = getTemperature();
        RegOpVector v;

        if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 2);
        v << RegRead64(REG_FVME2TMWR_RUN_EVENT_NUMBER);
        v << RegRead32(REG_FVME2TMWR_WR_SYNC_LOST_COUNTER);
        v << RegRead32(REG_FVME2TMWR_WR_LINK_ERROR_COUNTER);
        v << RegRead(REG_FVME2TMWR_STATUS);
        v << RegRead64(REG64_FVME2TMWR_BUSY_TIMER);
        v << RegRead64(REG64_FVME2TMWR_IDLE_TIMER);
        if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0);

        RegOpVector result = regOpExec(v);

        foreach (mlink::RegOp regOp, result) {
            switch (regOp.addr) {
            case REG_FVME2TMWR_RUN_EVENT_NUMBER:
            case REG_FVME2TMWR_RUN_EVENT_NUMBER+1:
            case REG_FVME2TMWR_RUN_EVENT_NUMBER+2:
            case REG_FVME2TMWR_RUN_EVENT_NUMBER+3: {
                int byte = regOp.addr % 4;
                status.evNum |= (quint64)regOp.data << 16*byte;
                break;
            }
            case REG_FVME2TMWR_WR_SYNC_LOST_COUNTER:
            case REG_FVME2TMWR_WR_SYNC_LOST_COUNTER+1: {
                int byte = regOp.addr % 4;
                status.wrSyncFail |= (quint64)regOp.data << 16*byte;
                break;
            }
            case REG_FVME2TMWR_WR_LINK_ERROR_COUNTER:
            case REG_FVME2TMWR_WR_LINK_ERROR_COUNTER+1: {
                int byte = regOp.addr % 4;
                status.wrLinkError |= (quint64)regOp.data << 16*byte;
                break;
            }
            case REG_FVME2TMWR_STATUS:
                status.deviceStatus = regOp.data;
                break;
            case REG64_FVME2TMWR_BUSY_TIMER:
            case REG64_FVME2TMWR_BUSY_TIMER+1:
            case REG64_FVME2TMWR_BUSY_TIMER+2:
            case REG64_FVME2TMWR_BUSY_TIMER+3: {
                int byte = regOp.addr % 4;
                status.busyTimer |= (quint64)regOp.data << 16*byte;
                break;
            }
            case REG64_FVME2TMWR_IDLE_TIMER:
            case REG64_FVME2TMWR_IDLE_TIMER+1:
            case REG64_FVME2TMWR_IDLE_TIMER+2:
            case REG64_FVME2TMWR_IDLE_TIMER+3: {
                int byte = regOp.addr % 4;
                status.idleTimer |= (quint64)regOp.data << 16*byte;
                break;
            }
            }
        }
    }

    emit updateStatus(status);
}

void VmeTrigDevice::resetEvNum(const RcRunId &runId)
{
    if(!readout){
        emit runStateUpdated(serial, RS_STARTED , runId);
        return;
    }

    bool ok=isConnected();
    if(ok && false){
        try{
            RegOpVector v;
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 1);
            v << RegRead(mlink::MregDeviceAdcm::ADCM16_CTRL_REG);
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0);
            v << RegRead(mlink::MregDeviceAdcm::ADCM16_CTRL_REG);

            RegOpVector result = regOpExec(v);
            ok = (result.size()==4);
            if(ok) {
                const RegOp &val1 = result[1];
                const RegOp &val3 = result[3];
                ok = (!val1.write) && val1.data==1 &&
                        (!val3.write) && val3.data==0;
            }
        } catch(...) {
        }
    }
    if(ok)
        qDebug()<<QString("Vme Trig Module Event Number reseted; serial=0x%1")
                  .arg(serial, 8, 16, QChar('0'));
    else
        qCritical() << QString("Failed to reset Vme Trig Module Event Number; serial=0x%1")
                       .arg(getSerialIdStr());
    emit runStateUpdated(serial, ok ? RS_STARTED : RS_ERROR, runId);
}

void VmeTrigDevice::vmeReadoutUpdated(quint64 serial, bool readout)
{
    if(this->serial==serial)
        this->readout = readout;
}

void VmeTrigDevice::evNumRequest(int checkIndex)
{
    if(!readout) return;
    if(!isConnected()) return;

    try{
        RegOpVector v;
        if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0x2);
        v << RegRead64(REG_FVME2TMWR_RUN_EVENT_NUMBER);
        if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
            v << RegWrite(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0);

        RegOpVector result = regOpExec(v);

        quint64 evNum = 0;

        foreach (const RegOp &regOp, result) {
            if(regOp.addr >= REG_FVME2TMWR_RUN_EVENT_NUMBER && (regOp.addr < REG_FVME2TMWR_RUN_EVENT_NUMBER+4)){
                int byte = regOp.addr % 4;
                evNum |= (quint64)regOp.data << 16*byte;
            }
        }
        if(result.size() == 6)
            emit evNumResponse(checkIndex, serial, evNum);
    } catch(...) {
    }
}
