#include "Config.h"
#include "mongo/BaseConfig.h"
#include "mregdev/device_id.h"

#include <QJsonDocument>
#include <QDirIterator>
#include <QSettings>

Config::Config(TDC2ViewerStartupOptions &startup_options):
    program_type(startup_options.program_type),
    program_version(startup_options.program_version),
    program_index(startup_options.program_index),
    configuration_name(startup_options.configuration_name),
    debug_mode(startup_options.debug_mode),
    rawDataEveryEv(false),
    rawDataLimit(1),
    rawDataPeriodic(false),
    rawDataPeriod(5),
    source(SOURCE_IDLE)
{
    update_root();
    default_root = root;
    default_root.set_data("configuration_name", QString("default"));
}

Config::~Config()
{
}

bool Config::read_config()
{
    root = BaseConfig::read_config(program_type, program_index, configuration_name);
    if(root.isEmpty()){
        write_config();
    }
    program_index = root.get_data<QString>("program_index");
    configuration_name = root.get_data<QString>("configuration_name");

    sourse_program_type = root.get_data<QString>("sourse_program_type");
    sourse_program_index = root.get_data<QString>("sourse_program_index");
    rawDataEveryEv = root.get_data<bool>("rawDataEveryEv");
    rawDataLimit = root.get_data<int>("rawDataLimit");
    rawDataPeriodic = root.get_data<bool>("rawDataPeriodic");
    rawDataPeriod = root.get_data<int>("rawDataPeriod");
    progInterfaceType = root.get_data<QString>("progInterfaceType");
    useSeparateInl = root.get_data<bool>("useSeparateInl");
    separateInlFolder = root.get_data<QString>("separateInlFolder");

    QString str = root.get_data<QString>("SOURCE_DATA");
    if(str == "SOURCE_IDLE"){
        source = SOURCE_IDLE;
    }
    if(str == "SOURCE_STREAM"){
        source = SOURCE_STREAM;
    }
    if(str == "SOURCE_FILE"){
        source = SOURCE_FILE;
    }

    for(QString child_name : root.children().keys()){
        if(child_name.split("_").at(0) == "device"){
            RootConfig root_device = root.child(child_name);
            DeviceIndex device_index;
            device_index.first = root_device.get_data<quint16>("device_id");
            device_index.second = root_device.get_data<quint64>("device_serial");
            for(QString cgrandson_name : root_device.children().keys()){
                if(cgrandson_name == "channels"){
                    RootConfig root_channels = root_device.child(cgrandson_name);
                    QMap<int, bool> map;
                    for(QString str : root_channels.data().keys()){
                        map[str.toInt()] = root_device.get_data<bool>(str);
                    }
                    invertMap[device_index] = map;
                }
            }
            continue;
        }

        if(child_name == "file_reader_config"){
            RootConfig root_file_reader = root.child(child_name);
            file_reader_config.lastFile = root_file_reader.get_data<QString>("lastFile");
            file_reader_config.playbackInterval = root_file_reader.get_data<int>("playbackInterval");
            file_reader_config.readBy = root_file_reader.get_data<int>("readBy");
            continue;
        }

        if(child_name == "tdc_ttvxs_map") {
            RootConfig root_map = root.child(child_name);

            for(auto pair : root_map.children().values()){
                DeviceIndex tdcIndex;
                tdcIndex.first = pair.get_data<quint16>("tdc_id");
                tdcIndex.second = pair.get_data<quint64>("tdc_ser");
                DeviceIndex ttvxsIndex;
                ttvxsIndex.first = pair.get_data<quint16>("ttvxs_id");
                ttvxsIndex.second = pair.get_data<quint64>("ttvxs_ser");
                if(!tdcIndex.isNull() && !ttvxsIndex.isNull())
                    tdc_ttvxs_map[tdcIndex] = ttvxsIndex;
            }
        }
    }
    reloadCalibration();

    return true;
}

bool Config::write_config()
{
    update_root();
    return BaseConfig::write_config(root);
}

void Config::update_root()
{
    root.remove_all_data();
    root.set_data("program_type", program_type);
    root.set_data("program_version", program_version);
    root.set_data("program_index", program_index);
    root.set_data("configuration_name", configuration_name);

    root.set_data("sourse_program_type", sourse_program_type);
    root.set_data("sourse_program_index", sourse_program_index);
    root.set_data("rawDataEveryEv", rawDataEveryEv);
    root.set_data("rawDataLimit", rawDataLimit);
    root.set_data("rawDataPeriodic", rawDataPeriodic);
    root.set_data("rawDataPeriod", rawDataPeriod);
    root.set_data("progInterfaceType", progInterfaceType);
    root.set_data("useSeparateInl", useSeparateInl);
    root.set_data("separateInlFolder", separateInlFolder);

    if(source == 0){
        root.set_data("SOURCE_DATA", QString("SOURCE_IDLE"));
    }
    if(source == 1){
        root.set_data("SOURCE_DATA", QString("SOURCE_STREAM"));
    }
    if(source == 2){
        root.set_data("SOURCE_DATA", QString("SOURCE_FILE"));
    }

    for(DeviceIndex device_index : invertMap.keys()){
        RootConfig root_device;
        root_device.set_data("device_id", device_index.first);
        root_device.set_data("device_serial", device_index.second);
            RootConfig root_map;
            for(auto ch : invertMap[device_index].keys()){
                root_map.set_data(QString::number(ch), invertMap[device_index][ch]);
            }
            root_map.item_name = "channels";
            root_device.append_child(root_map);
        root_device.item_name = QString("device_%1").arg(device_index.getSerialStr());
        root.append_child(root_device);
    }

    RootConfig root_file_reader;
    root_file_reader.set_data("lastFile", file_reader_config.lastFile);
    root_file_reader.set_data("playbackInterval", file_reader_config.playbackInterval);
    root_file_reader.set_data("readBy", file_reader_config.readBy);
    root_file_reader.item_name = "file_reader_config";
    root.append_child(root_file_reader);

    if(!tdc_ttvxs_map.isEmpty()) {
        RootConfig root_tdc_ttvxs_map("tdc_ttvxs_map");
        for(auto tdcIndex : tdc_ttvxs_map.keys()) {
            const auto &ttvxsIndex = tdc_ttvxs_map[tdcIndex];
            if(ttvxsIndex.isNull()) continue;
            RootConfig pair(QString("%1 -> %2")
                            .arg(tdcIndex.getIdent())
                            .arg(ttvxsIndex.getIdent()));
            pair.set_data("tdc_id", tdcIndex.getDevId());
            pair.set_data("tdc_ser", tdcIndex.getSerial());
            pair.set_data("ttvxs_id", ttvxsIndex.getDevId());
            pair.set_data("ttvxs_ser", ttvxsIndex.getSerial());
            root_tdc_ttvxs_map.append_child(pair);
        }
        if(!root_tdc_ttvxs_map.children_const().isEmpty())
            root.append_child(root_tdc_ttvxs_map);
    }
}

DeviceCalibration &Config::getCalibration(const DeviceIndex &index)
{
    Q_ASSERT(!index.isNull());
    return allCal[index];
}

void Config::reloadCalibration()
{
    CalConfig::reloadCalibration();

    if(useSeparateInl) {
        QStringList nameFilters;
        nameFilters << "*.ini" << "*.conf";
        QDirIterator it(separateInlFolder, nameFilters, QDir::Files | QDir::NoDotDot, QDirIterator::Subdirectories);
        while(it.hasNext()) {
            auto fn = it.next();

            QSettings sSep(fn, QSettings::IniFormat);
            CalConfig::reloadCalibration(sSep);
        }
    }
    QSettings s;
    CalConfig::reloadCalibration(s);
}
