#ifndef DEVICEDATA
#define DEVICEDATA

#include "mldiscover/DeviceIndex.h"
#include "lib-common/MpdRawTypes.h"

#include <map>
#include <vector>

#include <QElapsedTimer>
#include <QVector>

namespace  {
const double HPTDC_TS_STEP_NS = 0.024/1.024;
}

struct  TdcEdge
{
    TdcEdge(quint16 ts, quint32 t, bool l):time(t),tdcTs(ts),rising(l){}
    TdcEdge(){ clear(); }
    void clear() {
        time=0;
        tdcTs=0;
        rising=true;
    }

    quint32 time;
    quint16 tdcTs;
    bool rising;
};

struct  TdcChHit
{
    TdcChHit(){ clear(); }
    void clear() {
        ch=0;
        tdcTs=0;
        leadingTime=0;
        trailingTime=0;
        hasLeadTimeInl=false;
        hasTrailTimeInl=false;
        hasWidthTimeInl=false;
    }
    double getHitTime(const QVector<double> &inlCorr) {
        const quint32 *t = leadingTime ? (&leadingTime) : (&trailingTime);
        double time = HPTDC_TS_STEP_NS*(inlCorr.value((*t)%1024, 0)+*t);
        if(leadingTime) {
            hasLeadTimeInl=true;
            leadTimeInl=time;
        } else {
            hasTrailTimeInl=true;
            trailTimeInl=time;
        }
        return time;
    }
    double getLeadTimeNs(const QVector<double> &inlCorr) {
        hasLeadTimeInl = true;
        return leadTimeInl=HPTDC_TS_STEP_NS*(inlCorr.value(leadingTime%1024, 0)+leadingTime);
    }
    double getLeadTimeNs() const {
        return HPTDC_TS_STEP_NS*leadingTime;
    }
    double getTrailTimeNs() const {
        return HPTDC_TS_STEP_NS*trailingTime;
    }
    double getWidthNs(const QVector<double> &inlCorr) {
        quint32 t = trailingTime-leadingTime;
        hasWidthTimeInl=true;
        return widthTimeInl=HPTDC_TS_STEP_NS*(inlCorr.value(t%1024, 0)+t);
    }

    quint8 ch;
    quint16 tdcTs;
    quint32 leadingTime;
    quint32 trailingTime;
    double leadTimeInl;
    bool hasLeadTimeInl;
    double trailTimeInl;
    bool hasTrailTimeInl;
    double widthTimeInl;
    bool hasWidthTimeInl;
};

struct TdcStat {
    TdcStat() { clear(); }
    void clear() {
        index.clear();
        regIoTimeout = 0;
        regIoError = 0;
        stats.clear();
    }

    DeviceIndex index;
    bool regIoTimeout;
    bool regIoError;
    QMap<quint16, quint16> stats;
};

struct  TAI
{
    TAI(const AdcDataType1 *tsTAI) {
        flags = tsTAI->taiFlags;
        sec = tsTAI->wfTsHi;
        nSec = tsTAI->wfTsLow;
    }
    TAI(){ clear(); }
    void clear() {
        sec=0;
        flags=0;
        nSec=0;
    }

    quint32 sec;
    quint32 flags:2;
    quint32 nSec:30;
};

struct  TdcEvent
{
    TdcEvent(){ clear(); }
    void clear() {
        evNum=0;
        tai.clear();
        tdcHits.clear();
        temp.clear();
        fw_rev.clear();
        bmc_rev.clear();
        errors.clear();
    }

    DeviceIndex index;
    quint32 evNum;
    TAI tai;
    QMap<int, QVector<TdcChHit> > tdcHits; // per channel
    QMap<int, double> temp; // per tempId
    QVariant fw_rev;
    QVariant bmc_rev;
    std::vector<quint32> errors; // full 32-bit hptdc words
};

typedef union {
    unsigned int all;
    struct {
        unsigned int pulser: 1; // no connection
        unsigned int random_pulser: 1; // link data error
        unsigned int external_ttl: 1; // pci bus signaled error
        unsigned int digital_sfp: 1; // error propagation
        unsigned int reserved: 28; // no clock from remote
    } bit;
} ttvxs_trig_src_t;

struct  TtvxsEvent
{
    TtvxsEvent(){ clear(); }
    void clear() {
        evNum=0;
        tai.clear();

        temp.clear();
        fw_rev.clear();
        bmc_rev.clear();
        errors.clear();
    }

    DeviceIndex index;
    quint32 evNum;
    TAI tai;
    ttvxs_trig_src_t trig_src;
    QMap<int, double> temp; // per tempId
    QVariant fw_rev;
    QVariant bmc_rev;
    std::vector<quint32> errors; // full 32-bit hptdc words
};

struct CompleteEvent
{
    void clear() {
        evNum = 0;
        tdcEv.clear();
        ttvxsEv.clear();
        tdcStats.clear();
    }

    quint32 evNum;
    QMap<DeviceIndex, TdcEvent> tdcEv;
    QMap<DeviceIndex, TtvxsEvent> ttvxsEv;
    QMap<DeviceIndex, TdcStat> tdcStats;
};

struct DecodedData
{
    DecodedData () { clear(); }
    void clear() {
        events.clear();
        event_cnt = 0;
    }

    QMap<quint32, CompleteEvent> events;
    quint32 event_cnt;
};
Q_DECLARE_METATYPE(DecodedData)

struct TdcHits
{
    void clear() { hits.clear();}
    QMap<int, quint32> hits; // per channel
};
struct TdcErrors
{
    void clear() { errors.clear();}
    QMap<quint32, QList< QPair<quint32, quint32> > > errors; // per channel
};

struct AnalyzerStats
{
    AnalyzerStats ():
        lastEvNum(0)
    {
        clear();
    }
    void clear() {
        eventDecoded=0;
        sizeDecoded=0;
//        lastEvNum=0;
        decodeTimeNs=0;
        decodeLoop=0;
        formatErrorCnt=0;
        tdcHits.clear();
        tdcErrors.clear();
        ttvxsList.clear();

        timer.invalidate();
        skip=0;
        pass=0;
    }

    qint64 eventDecoded;
    qint64 sizeDecoded;
    quint32 lastEvNum;
    qint64 decodeTimeNs; // decode time 200-300 us per max adcWin
    qint64 decodeLoop;
    int formatErrorCnt;
    QMap<DeviceIndex, TdcHits> tdcHits;
    QMap<DeviceIndex, TdcErrors> tdcErrors;
    QList<DeviceIndex> ttvxsList;

    QElapsedTimer timer;
    qint64 skip;
    qint64 pass;
};
Q_DECLARE_METATYPE(AnalyzerStats)

#endif // DEVICEDATA

