#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QAbstractButton>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QProcess>
#include <QMessageBox>
#include <QMetaType>
#include <QTime>
#include <QMap>

#include "Config.h"
#include "pnp-server/PNPMessage.h"
#include "TDC2ViewerStartupOptions.h"
#include "configuration-manager/ConfigurationManager.h"

namespace Ui {
class MainWindow;
}

class QTableWidget;
class QSemaphore;
class QTreeWidgetItem;

#include "DeviceData.h"

class CalibrationDialog;
class FileReaderDialog;
class LogDialog;
class InlCheckDialog;
class TDCAnalyzer;
class TcpClient;
class PNPServer;
class HistWorker;
class StatusDialog;

Q_DECLARE_METATYPE(QHostAddress)

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(TDC2ViewerStartupOptions &startup_options, QWidget *parent = nullptr);
    ~MainWindow();

signals:
    void deleteDevice(const DeviceIndex &);
    void closeAll();
    void addLog(QString);
    void resetAll();
    void requestRawData(int events);
    void tcpAddressChanged(QHostAddress addr, quint16 port);
    void rawDataEveryEvUpdated(bool dataFromFile);
    void closeFile();
    void invertUpdated(InvertMapType invertMap) const;
    void tdcTtvxMapUpdated(TdcTtvxsMapType ttMap) const;
    void inlConfigUpdated(InlResults) const;
    void requestInlResults(bool allRes) const;
    void set_can_skip(bool skip) const;
    void set_show_skip_info(bool skip) const;

private slots:
    void on_actionConfiguration_manager_triggered();
    void switch_to_config(QString program_index, QString configuration_name);
    void on_actionSet_source_file_triggered();
    void on_actionConnect_to_source_program_triggered();
    void on_actionDisconnect_from_program_triggered();

    void on_pushButtonResetAll_clicked();

    void tcpStatusUpdated(bool connOk, double dataRate);
    void programDiscovered(const ProgramDescription &newDescr);
    void programClosed(const ProgramDescription &newDescr);
    void updateHitStats(const AnalyzerStats &stats);
    void rawDataUpdated(const QStringList &data);

    void on_actionExit_triggered();
    void on_actionCalibrationData_triggered();
    void on_actionReset_all_triggered();
    void inl_corrections_updated(InlResults res, bool allRes);
    void on_actionSave_INL_corrections_triggered();
    void on_actionShow_INL_check_triggered();
    void on_actionSet_INL_separate_location_toggled(bool checked);
    void on_actionHPTDC_vol_check_triggered();
    void on_actionSet_TDC_TTVXS_map_triggered();
    void on_actionShow_skip_info_toggled(bool show);
    void on_actionSatistic_dialog_triggered();
    void on_actionHPTDC_errors_triggered();
    void on_actionAbout_triggered();

    void on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column);
    void on_radioButtonRawEvery_toggled(bool checked);
    void on_pushButtonRequestRaw_clicked();
    void on_spinBoxRawEvents_valueChanged(int limit);
    void on_checkBoxRawPeriodical_toggled(bool checked);
    void on_spinBoxRawTimer_valueChanged(int val);

    void sourceFileNameUpdated(QString fn);
    void tcpConnected(bool connected);
    void rawDataPeriodicRequest();
    void updateStatus();

    void writeSettings();

public slots:
    void addLogFromSender(QtMsgType severity, const QString &log, const QString &thrName);

private:
    void read_config();
    void initTree();
    void updateModelList(const AnalyzerStats &analStat);
    void updateTreeWithStats();
    void updateTreeItemStates();
    void updateInlMarks();
    void updateTtvxsMarks();
    void setTableRowsCount(int nch);
    void updateControlState();
    void sendLog(QString l);
    void updateSourceLabel();
    void changeSourceState(SOURCE_DATA st);
    QVector<ProgramDescription> getFreeProgs() const;
    bool hasFreeProgs() const;
    bool isFreeProg(const ProgramDescription &pd) const;
    ProgramInterface getFreeProgramInterface(const ProgramDescription &pd) const;
    void updateTcpInterface(const ProgramDescription &descr);
    void updateRawDataControl();
    void disconnectFromProg();
    void setSourceColor(QString col);

    ConfigurationManager *configuration_manager;
private:
    Ui::MainWindow *ui;
    QSemaphore * const analyzerSem;// free space in analyzer queue
    QSemaphore * const histSem;// free space in histogram queue
    QSemaphore * const writerSem;

    Config config;
    InvertMapType prev_invert_map;
    FileReaderDialog * const fileReaderDialog;
    PNPServer * const pnpServer;
    ProgramDescription selectedProgram;
    ProgramInterface selectedInterface;


    QMap<DeviceIndex, QTreeWidgetItem *> modTreeItem;
    QLabel * const dataSourceLabel;
    QLabel * const dataRateLabel;
    QLabel * const analyzerStatusLabel;
    QLabel * const semStatusLabel;
    QLabel * const vxsClockLabel;
    QLabel * const analyzerLabel;
    LogDialog *const logDialog;
    CalibrationDialog * const calibrationDialog;
    InlCheckDialog * const inlCheckDialog;

    TcpClient * const tcpClient;
    QThread * const tcpClientThread;
    TDCAnalyzer * const analyzer;
    QThread * const analyzerThread;
    HistWorker * const worker;
    QThread * const worker_thread;

    QTreeWidgetItem * const rootTreeItem;
    QVector<ProgramDescription> progDescrs;
    QMap<QUuid, ProgramDescription> progDescrsMap;
    AnalyzerStats lastAnalStat;
    QTimer * const rawDataPeriodicTimer;
    QTimer * const statusTimer;
    bool rawDataReqDone;
    uint rawEventCnt;
    QElapsedTimer rawEventCntTimer;
    bool tcpConnOk;
    StatusDialog * const statusDialog;
};

#endif // MAINWINDOW_H
