#include "StatusDialog.h"
#include "ui_StatusDialog.h"

#include <QDebug>
#include <QScrollBar>

#include "ConfigParams.h"
#include <QSettings>
#include <QTimer>

namespace {
enum {COL_VAL, COL_DIFF, COL_SPEED, COL_TOTAL};
enum {ROW_SIZE, ROW_EV_NUM, ROW_DECODE_TOTAL, ROW_DECODE_SKIP, ROW_DECODE_PASS, ROW_DECODE_SKIP_RATIO, ROW_DECODE_PASS_RATIO, ROW_FORMAT_ERROR, ROW_TOTAL};
}

StatusDialog::StatusDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::StatusDialog),
    show_skip_info(false),
    loop(0),
    freeze(false)
{
    ui->setupUi(this);
    init();

    QSettings settings;
    settings.beginGroup("StatusDialog");
    restoreGeometry(settings.value("geometry").toByteArray());
    if(settings.value("visible").toBool())
        QTimer::singleShot(100, this, &StatusDialog::show);
    settings.endGroup();
}

StatusDialog::~StatusDialog()
{
    QSettings settings;
    settings.beginGroup("StatusDialog");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("visible", isVisible());
    settings.endGroup();

    delete ui;
}

void StatusDialog::updateHitStats(const AnalyzerStats &stats)
{
    auto table = ui->tableWidget;
    qint64 dt = prevStat.timer.isValid() ? prevStat.timer.msecsTo(stats.timer) : 0;
    if(freeze) {
        pendingStat = stats;
        update_indicator(dt);
        return;
    }

    table->setUpdatesEnabled(false);
    table->item(ROW_SIZE, COL_VAL)->setText(QString("%L1 B").arg(stats.sizeDecoded));
    table->item(ROW_EV_NUM, COL_VAL)->setText(QString("%L1").arg(stats.lastEvNum));
    qint64 total = stats.skip+stats.pass;
    table->item(ROW_DECODE_TOTAL, COL_VAL)->setText(QString("%L1").arg(total));
    QString ratioStr;
    ratioStr = total ? QString(" (%1%)").arg(100*stats.skip/total) : "";
    table->item(ROW_DECODE_SKIP, COL_VAL)->setText(QString("%L1").arg(stats.skip).append(ratioStr));
    ratioStr = total ? QString(" (%1%)").arg(100*stats.pass/total) : "";
    table->item(ROW_DECODE_PASS, COL_VAL)->setText(QString("%L1").arg(stats.pass).append(ratioStr));
    table->item(ROW_DECODE_SKIP_RATIO, COL_VAL)->setText(total ? QString("%1%").arg(100*stats.skip/total) : "-");
    table->item(ROW_DECODE_PASS_RATIO, COL_VAL)->setText(total ? QString("%1%").arg(100*stats.pass/total) : "-");
    if(stats.formatErrorCnt) {
        table->item(ROW_FORMAT_ERROR, COL_VAL)->setText(QString("%L1").arg(stats.formatErrorCnt));
        if(table->isRowHidden(ROW_FORMAT_ERROR))
            table->showRow(ROW_FORMAT_ERROR);
    }
    if(dt) {
        qint64 dVal;
        double speed;
        dVal = stats.sizeDecoded-prevStat.sizeDecoded;
        speed = 1e3*(dVal)/dt;
        table->item(ROW_SIZE, COL_DIFF)->setText(QString("%L1 B").arg(dVal));
        table->item(ROW_SIZE, COL_SPEED)->setText(QString("%1 MB/s").arg(speed/1e6,0,'f',2));
        dVal = stats.lastEvNum-prevStat.lastEvNum;
        speed = 1e3*(dVal)/dt;
        table->item(ROW_EV_NUM, COL_DIFF)->setText(QString("%L1").arg(dVal));
        table->item(ROW_EV_NUM, COL_SPEED)->setText(QString("%L1").arg((qint64)speed));
        dVal = stats.skip+stats.pass-prevStat.skip-prevStat.pass;
        speed = 1e3*(dVal)/dt;
        qint64 dValTotal = dVal;
        table->item(ROW_DECODE_TOTAL, COL_DIFF)->setText(QString("%L1").arg(dVal));
        table->item(ROW_DECODE_TOTAL, COL_SPEED)->setText(QString("%L1").arg((qint64)speed));
        dVal = stats.skip-prevStat.skip;
        speed = 1e3*(dVal)/dt;
        ratioStr = dValTotal ? QString(" (%1%)").arg(100*dVal/dValTotal) : "";
        table->item(ROW_DECODE_SKIP, COL_DIFF)->setText(QString("%L1%2").arg(dVal).arg(ratioStr));
        table->item(ROW_DECODE_SKIP, COL_SPEED)->setText(QString("%L1").arg((qint64)speed));
        table->item(ROW_DECODE_SKIP_RATIO, COL_DIFF)->setText(dValTotal ? QString("%1%").arg(100*dVal/dValTotal) : "-");
        dVal = stats.pass-prevStat.pass;
        speed = 1e3*(dVal)/dt;
        ratioStr = dValTotal ? QString(" (%1%)").arg(100*dVal/dValTotal) : "";
        table->item(ROW_DECODE_PASS, COL_DIFF)->setText(QString("%L1%2").arg(dVal).arg(ratioStr));
        table->item(ROW_DECODE_PASS, COL_SPEED)->setText(QString("%L1").arg((qint64)speed));
        table->item(ROW_DECODE_PASS_RATIO, COL_DIFF)->setText(dValTotal ? QString("%1%").arg(100*dVal/dValTotal) : "-");
        if(stats.formatErrorCnt) {
            dVal = stats.formatErrorCnt-prevStat.formatErrorCnt;
            speed = 1e3*(dVal)/dt;
            table->item(ROW_FORMAT_ERROR, COL_DIFF)->setText(QString("%L1").arg(dVal));
            table->item(ROW_FORMAT_ERROR, COL_SPEED)->setText(QString("%1").arg((qint64)speed));
            if(table->isRowHidden(ROW_FORMAT_ERROR))
                table->showRow(ROW_FORMAT_ERROR);
        }
    }

    log_info(stats);

    table->resizeColumnsToContents();
    update_table_size(table);
    table->setUpdatesEnabled(true);

    if(stats.eventDecoded)
    ui->labelAvrEvSize->setText(QString("%L1 B").arg(stats.sizeDecoded/stats.eventDecoded));
    if(stats.pass)
    ui->labelAvrDecodeTime->setText(QString("%L1 ns").arg(stats.decodeTimeNs/stats.pass));
    if(stats.decodeLoop)
    ui->labelAvrAnalizeLoopSize->setText(QString("%L1 B").arg(stats.sizeDecoded/stats.decodeLoop));
    if(stats.decodeLoop)
    ui->labelAvrAnalizeLoopTime->setText(QString("%L1 ms").arg(1e-6*stats.decodeTimeNs/stats.decodeLoop, 0,'f',1));

    ui->label_config_param->setText(QString("Data queue size: %1\nPack size limit: %L2 B\nDecoded events limit: %L3 B")
                                    .arg(DATA_QUEUE_SIZE).arg(PACK_SIZE_LIMIT).arg(DECODED_EVENTS_LIMIT));
    prevStat = stats;
    update_indicator(dt);
}

void StatusDialog::reset()
{
    updateHitStats(AnalyzerStats());
    ui->tableWidget->hideRow(ROW_FORMAT_ERROR);
}

void StatusDialog::init()
{
    auto table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);
    table->setRowCount(ROW_TOTAL);
    for(int c=0; c<COL_TOTAL; ++c) {
        for(int r=0; r<ROW_TOTAL; ++r){
            auto i = new QTableWidgetItem();
            table->setItem(r, c, i);
            if(r==ROW_FORMAT_ERROR) {
                auto brush = i->foreground();
                brush.setColor(Qt::red);
                i->setForeground(brush);
            }
        }
    }

    table->setHorizontalHeaderItem(COL_VAL, new QTableWidgetItem("Value"));
    table->setHorizontalHeaderItem(COL_DIFF, new QTableWidgetItem("Diff"));
    table->setHorizontalHeaderItem(COL_SPEED, new QTableWidgetItem("Speed (per 1s)"));

    table->setVerticalHeaderItem(ROW_SIZE, new QTableWidgetItem("Recieved size"));
    table->setVerticalHeaderItem(ROW_EV_NUM, new QTableWidgetItem("Cur evNum"));
    table->setVerticalHeaderItem(ROW_DECODE_TOTAL, new QTableWidgetItem("Total events"));
    table->setVerticalHeaderItem(ROW_DECODE_SKIP, new QTableWidgetItem("Skipped events"));
    table->setVerticalHeaderItem(ROW_DECODE_PASS, new QTableWidgetItem("Parsed events"));
    table->setVerticalHeaderItem(ROW_DECODE_SKIP_RATIO, new QTableWidgetItem("Skipped events %"));
    table->setVerticalHeaderItem(ROW_DECODE_PASS_RATIO, new QTableWidgetItem("Parsed events %"));
    table->setVerticalHeaderItem(ROW_FORMAT_ERROR, new QTableWidgetItem("Format error"));
    table->resizeColumnsToContents();

    table->hideRow(ROW_DECODE_SKIP_RATIO);
    table->hideRow(ROW_DECODE_PASS_RATIO);
    table->hideRow(ROW_FORMAT_ERROR);
}

void StatusDialog::log_info(const AnalyzerStats &stats)
{
    if(!show_skip_info) return;
    qint64 dt = 0;
    if(lastLogTimer.isValid() ||
            (dt = lastLogTimer.msecsTo(stats.timer))<4900)
        return;

    lastLogTimer = stats.timer;
    if(stats.skip) {
        quint64 tot = stats.pass+stats.skip;
        quint64 dSkip = stats.skip - prevStat.skip;
        quint64 dPass = stats.pass - prevStat.pass;
        quint64 dTot = dPass+dSkip;
        qInfo() << QString("skip: %1; pass: %2; skip: %3% (%4/%5-%6%); speed: %7")
                   .arg(stats.skip).arg(stats.pass)
                   .arg(100.*stats.skip/tot, 0, 'f', 2)
                   .arg(dSkip).arg(dPass)
                   .arg(100.*dSkip/dTot, 0, 'f', 2)
                   .arg(1e3*dPass/dt);
    }
}

void StatusDialog::update_table_size(QTableWidget * const table, bool widthOnly)
{
    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

void StatusDialog::update_indicator(qint64 dt)
{
    QString str = QString("....").replace(loop,1,QChar('*'));
    if(dt)
        str.append(QString("\n%1 ms").arg(dt));
    ui->labelIndicator->setText(str);
    loop = (loop+1)%4;
}

void StatusDialog::on_checkBoxFreeze_toggled(bool checked)
{
//    ui->checkBoxFreeze->setText(checked ? "Freeze":"Release");
    freeze = checked;
    if(!checked && pendingStat.timer.isValid()) {
        updateHitStats(pendingStat);
        pendingStat.timer.invalidate();
    }
}
