#include "TdcTtvxsMapDialog.h"
#include "ui_TdcTtvxsMapDialog.h"

#include "DeviceData.h"
#include "Config.h"

#include <QComboBox>
#include <QLabel>
#include <QPushButton>

TdcTtvxsMapDialog::TdcTtvxsMapDialog(const AnalyzerStats analyzerStats,
                                     const TdcTtvxsMapType &ttMap,
                                     QWidget *parent) :
    QDialog(parent),
    ui(new Ui::TdcTtvxsMapDialog),
    original_tdc_ttvxs_map(ttMap)
{
    ui->setupUi(this);
    ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setText("Restore");

    absentTdc  = original_tdc_ttvxs_map.keys();
    presentTdc = analyzerStats.tdcHits.keys();
    for(auto tdc : presentTdc) {
        absentTdc.removeAll(tdc);
    }
    presentTtvxs = analyzerStats.ttvxsList;
    for(auto ttvxs : original_tdc_ttvxs_map.values())
        if(!presentTtvxs.contains(ttvxs))
            absentTtvxs << ttvxs;

    initGui();
}

TdcTtvxsMapDialog::~TdcTtvxsMapDialog()
{
    delete ui;
}

void TdcTtvxsMapDialog::currentTextChanged()
{
    QComboBox *cb = qobject_cast<QComboBox*>(sender());
    if(!cb) return;

    DeviceIndex tdc;

    auto tdcLabel = qobject_cast<QLabel*>(ui->formLayout->labelForField(cb));
    QString tdcStr = tdcLabel->text();
    for(auto i : presentTdc) {
        if(tdcStr == i.getSerialStr()) {
            tdc = i;
            break;
        }
    }
    if(tdc.isNull()) {
        for(auto i : absentTdc) {
            if(tdcStr == i.getSerialStr()) {
                tdc = i;
                break;
            }
        }
    }
    Q_ASSERT(!tdc.isNull());
    DeviceIndex ttvxs = getTtvxsIndex(cb);
    if(ttvxs.isNull())
        tdc_ttvxs_map.remove(tdc);
    else
        tdc_ttvxs_map[tdc] = ttvxs;
}

void TdcTtvxsMapDialog::on_buttonBox_clicked(QAbstractButton *button)
{
    auto pb = qobject_cast<QPushButton*>(button);
    if(pb == ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)) {
        initGui();
        return;
    }
    if(pb != ui->buttonBox->button(QDialogButtonBox::Reset))
        return;

    auto layout = ui->formLayout;
    for(int row=0; row<layout->rowCount();) {
        auto label = qobject_cast<QLabel*>(layout->itemAt(row, QFormLayout::LabelRole)->widget());
        if(label) {
            bool absent=false;
            for(auto index : absentTdc)
                if(index.getSerialStr()==label->text()) {
                    absent = true;
                    tdc_ttvxs_map.remove(index);
                    break;
                }
            if(absent) {
                layout->removeRow(row);
                continue;
            }
        }
        auto cb = qobject_cast<QComboBox*>(layout->itemAt(row, QFormLayout::FieldRole)->widget());
        ++row;
        if(!cb) continue;
        cb->setCurrentText("None");
        for(int i=0; i<cb->count();) {
            auto index = getTtvxsIndex(cb, i);
            if(absentTtvxs.contains(index))
                cb->removeItem(i);
            else
                ++i;
        }
    }
    if(presentTdc.isEmpty()) {
        auto label = new QLabel(DeviceIndex().getSerialStr(), this);
        label->setEnabled(false);
        auto cb = createComboBox(DeviceIndex());
        cb->setEnabled(false);
        ui->formLayout->addRow(label, cb);
    }

    Q_ASSERT(tdc_ttvxs_map.isEmpty());
}

void TdcTtvxsMapDialog::initGui()
{
    while (ui->formLayout->rowCount()) {
        ui->formLayout->removeRow(0);
    }
    tdc_ttvxs_map = original_tdc_ttvxs_map;
    ui->formLayout->addRow(new QLabel("TDC", this), new QLabel("TTVXS", this));

    for(auto tdc : presentTdc) {
        auto label = new QLabel(tdc.getSerialStr(), this);
        auto cb = createComboBox(tdc_ttvxs_map.value(tdc, DeviceIndex()));
        ui->formLayout->addRow(label, cb);
    }

    for(auto tdc : absentTdc) {
        auto label = new QLabel(tdc.getSerialStr(), this);
        label->setEnabled(false);
        auto cb = createComboBox(tdc_ttvxs_map.value(tdc, DeviceIndex()));
        ui->formLayout->addRow(label, cb);
    }

    if(presentTdc.isEmpty() && absentTdc.isEmpty()) {
        auto label = new QLabel(DeviceIndex().getSerialStr(), this);
        label->setEnabled(false);
        auto cb = createComboBox(DeviceIndex());
        cb->setEnabled(false);
        ui->formLayout->addRow(label, cb);
    }
}

QComboBox *TdcTtvxsMapDialog::createComboBox(DeviceIndex prefered_ttvxs)
{
    auto cb = new QComboBox(this);
    cb->addItem("None", QVariant());
    if(!presentTtvxs.isEmpty())
        cb->insertSeparator(cb->count());
    for(auto ttvxs : presentTtvxs){
        cb->addItem(ttvxs.getSerialStr(),
                    QList<QVariant>() << ttvxs.getDevId() << ttvxs.getSerial());
        if(prefered_ttvxs == ttvxs)
            cb->setCurrentText(ttvxs.getSerialStr());
    }
    if(!absentTtvxs.isEmpty())
        cb->insertSeparator(cb->count());
    for(auto ttvxs : absentTtvxs){
        QString itemName = ttvxs.getSerialStr()+" (off)";
        cb->addItem(itemName,
                    QList<QVariant>() << ttvxs.getDevId() << ttvxs.getSerial());

        cb->setItemData(cb->findText(itemName), QBrush(Qt::red), Qt::TextColorRole);
        if(prefered_ttvxs == ttvxs)
            cb->setCurrentText(itemName);
    }
    connect(cb, &QComboBox::currentTextChanged,
            this, &TdcTtvxsMapDialog::currentTextChanged);
    return cb;
}

DeviceIndex TdcTtvxsMapDialog::getTtvxsIndex(QComboBox *cb, int id)
{
    QList<QVariant> data;
    if(id==-1)
        data = cb->currentData().toList();
    else
        data = cb->itemData(id).toList();

    DeviceIndex index;
    if(data.size() == 2) {
        index = DeviceIndex(static_cast<quint16>(data.first().toUInt()),
                            data.last().toULongLong());
    }
    return index;
}
