﻿#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <QBitArray>
#include <QBoxLayout>
#include <QClipboard>
#include <QFileDialog>
#include <QJsonDocument>
#include <QHostAddress>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include "QShortcut"
#include <QSpinBox>
#include <QSplitter>
#include <QStandardPaths>
#include <QTableWidgetItem>
#include <QThread>
#include <QTimer>
#include <QTreeWidgetItem>
#ifdef __linux__
#include <signal.h>
#endif

#include "device-discover/DiscoverDialog.h"
#include "mregdev/device_id.h"
#include "daq-config/DaqConfig.h"
#include "base/threadQuit.h"
#include "pnp-server/PNPServer.h"
#include "metric-sender/MetricSender.h"
#include "lib-common/UIStyles.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/RunControlCommands.h"

#include "LogDialog.h"
#include "AdjustDialog.h"
#include "OrderDialog.h"
#include "base/AboutBox.h"
#include "base/ThreadController.h"

namespace {
    enum {COL_DEV_SERIAL, COL_DEV_SLOT, COL_DEV_FW, COL_DEV_MCU_VER, COL_DEV_PLL_UNLOCK, COL_DEV_PLL_STATUS, COL_DEV_TEMP, COL_DEV_POWER, COL_DEV_TIME,
          COL_DEV_RUN_TIME, COL_DEV_CUR_EV_NUM, COL_DEV_UNBLOCK, COL_DEV_BLOCK, COL_DEV_MS_ERR,
          COL_DEV_STATUS, COL_HPTDC_STATUS,
          COL_BUSY_TIME, COL_BUSY_STATUS,COL_TRIG_QUEUE,
          COL_CLOCK_STATUS, COL_DEV_STREAM, COL_ERR_CNT, COL_DEV_TOTAL};
    enum {
        tree_tdc_name, // moduleName or modCh
        tree_tdc_num_cols
    };
    const QString BUSY_STATUS_TOOL_TIP = "Current Busy Status\n"
                                         "1x - TDC busy\n"
                                         "x1 - ReadOut busy";
    const QString BUSY_TIME_TOOL_TIP = "Moment Busy time\n"
                                       "1) TDC busy\n"
                                       "2) ReadOut busy\n"
                                       "3) Total busy";
    const QString DEV_FW_TOOL_TIP = "Doubli-click copies product-status information to clipboard";
    const int REMOTE_CONTROL_PORT = 0;

    void setItemTextColor(QTableWidgetItem* const item, QColor col, bool isVertical=false) {
        if(!item)
            return;
        if(isVertical) {
            item->setTextColor(col);
        } else {
            auto brush = item->foreground();
            brush.setColor(col);
            item->setForeground(brush);
        }
    }
    void setItemBgColor(QTableWidgetItem* const item, QColor col) {
        if(!item)
            return;
        auto brush = item->background();
        brush.setColor(col);
        brush.setStyle(Qt::SolidPattern);
        item->setBackground(brush);
    }
    void setTableItem(QTableWidget *tableDev, int row, int col) {
        bool blockSignals = tableDev->blockSignals(true);
        if(col==COL_CLOCK_STATUS) {
            auto label = new QLabel(tableDev);
            label->setAlignment(Qt::AlignCenter);
            tableDev->setCellWidget(row, col, label);
        } else {
            auto item = new QTableWidgetItem();
                item->setTextAlignment(Qt::AlignCenter);
            if(col == COL_DEV_MS_ERR) {
                setItemTextColor(item, Qt::red);
            }
            tableDev->setItem(row, col, item);
        }
        tableDev->blockSignals(blockSignals);
    }

    class TableSortLocker {
    public:
        TableSortLocker(QTableWidget *t) :
            table(t),
            sort_en(t->isSortingEnabled())
        {
            if(sort_en) table->setSortingEnabled(false);
        }
        ~TableSortLocker() {
            if(sort_en) table->setSortingEnabled(true);

        }
        QTableWidget * const table;
        const bool sort_en;
    };

    class TableUpdateLocker {
    public:
        TableUpdateLocker(QWidget *t) :
            w(t),
            up_en(t->updatesEnabled())
        {
            if(up_en) w->setUpdatesEnabled(false);
        }
        ~TableUpdateLocker() {
            if(up_en) w->setUpdatesEnabled(true);

        }
        QWidget * const w;
        const bool up_en;
    };

    void initTableItemAtRow(QTableWidget *tableDev, int row) {
        auto sortLocker = TableSortLocker(tableDev);
        for(int col=0; col<COL_DEV_TOTAL; ++col) {
            setTableItem(tableDev, row, col);
        }
        tableDev->setVerticalHeaderItem(row, new QTableWidgetItem());
        tableDev->item(row, COL_BUSY_STATUS)->setToolTip(BUSY_STATUS_TOOL_TIP);
        tableDev->item(row, COL_BUSY_TIME)->setToolTip(BUSY_TIME_TOOL_TIP);
        tableDev->item(row, COL_DEV_FW)->setToolTip(DEV_FW_TOOL_TIP);
        if(row==0) {
            tableDev->item(row, COL_DEV_SERIAL)->setText("No device selected");
            tableDev->verticalHeaderItem(row)->setText("No device selected");
        }
    }

    QString getDevMetricPrefix(DeviceIndex i) {
        return QString("%1.").arg(i.getSerialStr());
    }

    void setLabelColor(QLabel * const lab, QColor col) {
        if(!lab) return;
        QPalette p = lab->palette();
        p.setColor(lab->foregroundRole(), col);
        lab->setPalette(p);
    }
}

MainWindow::MainWindow(StartupOptions &start_options, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    config(start_options),
    discoverDialog(new DiscoverDialog(this)),
    remoteControl(new RemoteControlServer(this, REMOTE_CONTROL_PORT)),
    pnpServer(new PNPServer(this)),
    evNumChecker(new EvNumChecker(this)),
    blockControl(0),
    logDialog(new LogDialog(this)),
    adjustDialog(start_options.debug_mode ? new AdjustDialog(this) : nullptr),
    rootTreeItem(new QTreeWidgetItem(QStringList("All"))),
    deviceStatusTimer(new QTimer(this)),
    writeConfigTimer(new QTimer(this)),
    criticalPowerMode(false),
    on_remote(false),
    setup_index(0),
    update_remote_enabled(true)
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<TdcStatus>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<QMap<int,double> >();
    qRegisterMetaType<QMap<int,int> >();
    qRegisterMetaType<Adjust_setup>();
    ui->setupUi(this);
    statusBar()->hide();

    widgets_to_block.append(ui->comboBoxEdges);
    widgets_to_block.append(ui->comboBoxPreset);
    widgets_to_block.append(ui->actionAllow_remote_control);
    widgets_to_block.append(ui->actionon_board_PLL_bypass);
    widgets_to_block.append(ui->actionAdd_HPTDC_event_header);
    widgets_to_block.append(ui->actionAdd_HPTDC_event_tralier);
    widgets_to_block.append(ui->groupBoxStat);
    widgets_to_block.append(ui->spinBoxStatPeriod);

    connect(discoverDialog, &DiscoverDialog::deviceDiscovered,
            this, &MainWindow::deviceUpdated);
    discoverDialog->setSingleMode(false);
    QSet<quint16> supportedDevices;
    supportedDevices << DEVICE_ID_TDC72VXS;
    supportedDevices << DEVICE_ID_TDC64VHLE;
    discoverDialog->setDeviceIdFilter(supportedDevices);

    connect(this, &MainWindow::set_run_state, ui->m_stream_output, &StatisticOutput::set_run_state);

    connect(evNumChecker, &EvNumChecker::completedEvNumCheck, remoteControl, &RemoteControlServer::evNumCheckResponse);
    connect(evNumChecker, &EvNumChecker::checkFail, this, &MainWindow::sendRunWarning);
    connect(remoteControl, &RemoteControlServer::evNumCheckRequest, evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);

    connect(remoteControl, &RemoteControlServer::startRequest, this, &MainWindow::onRemoteStartRequest);
    connect(remoteControl, &RemoteControlServer::stopRequest, this, &MainWindow::onRemoteStopRequest);
    connect(remoteControl, &RemoteControlServer::connectionStatusUpdated, this, &MainWindow::remoteStatusUpdated);
    connect(this, &MainWindow::remoteStartCompleted, remoteControl, &RemoteControlServer::completedStart);
    connect(this, &MainWindow::remoteStopCompleted, remoteControl, &RemoteControlServer::completedStop);
    connect(this, &MainWindow::sendRemoteState, remoteControl, &RemoteControlServer::sendState);
    connect(this, &MainWindow::closeAll, remoteControl, &RemoteControlServer::deleteLater);
    connect(this, &MainWindow::runError, remoteControl,
        QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));
    connect(this, &MainWindow::runWarning, remoteControl, &RemoteControlServer::onRunWarning);
    connect(this, &MainWindow::allowRemoteControl, remoteControl, &RemoteControlServer::startListen);
    connect(this, &MainWindow::config_transfer_to_rc, remoteControl, &RemoteControlServer::config_transfer);
    connect(this, &MainWindow::changeProgramDescription, remoteControl, &RemoteControlServer::changeProgramDescription);

    new ThreadController(remoteControl, this);

    auto hhView = ui->tableWidgetDevice->horizontalHeader();
    connect(hhView, &QHeaderView::sortIndicatorChanged,
            this, &MainWindow::sortIndicatorChanged);

    pnpServer->startSearch(DaqConfig::getTypeName(DaqConfigMStream));
    connect(pnpServer, &PNPServer::programDiscovered, this, &MainWindow::programDiscovered);
    connect(this, &MainWindow::changeProgramDescription, pnpServer, &PNPServer::changeProgramDescription);
    connect(this, &MainWindow::closeProgram, pnpServer, &PNPServer::closeProgram);
    connect(this, &MainWindow::closeAll, pnpServer, &QObject::deleteLater);
    new ThreadController(pnpServer, this);

    configuration_manager = new ConfigurationManager(this, start_options.program_type, config.default_root);
    connect(configuration_manager, &ConfigurationManager::switch_to_config, this, &MainWindow::switch_to_config);

    connect(deviceStatusTimer, &QTimer::timeout, this, &MainWindow::applyDeviceStatusUpdates);
    deviceStatusTimer->setSingleShot(true);
    deviceStatusTimer->start(500);

    connect(writeConfigTimer, &QTimer::timeout, this, &MainWindow::delayed_write_config);
    writeConfigTimer->setSingleShot(true);
    writeConfigTimer->start(500);

    initTable();
    initTree();
    initShortcut();
    read_config();
    init_gui(start_options);
    updateControlState();
}

MainWindow::~MainWindow()
{
    writeSettings();
    emit closeProgram(program_description);

    if(writeConfigTimer->isActive()) {
        writeConfigTimer->stop();
        delayed_write_config();
    }
    for (const DeviceIndex &i: devGlob.keys()) {
        disconnectFromDev(i);
    }

    emit closeAll();
    QElapsedTimer closeTimer;
    closeTimer.start();
    for (const DeviceIndex &i: devGlob.keys()) {
        if(!threadQuiting(devGlob[i]->devThread, 2000, &closeTimer)) {
            qWarning() << "Can't stop device thread";
        }
    }

    delete ui;
}

void MainWindow::init_gui(StartupOptions &start_options)
{
    ui->label_remote_status->hide();

    ui->labelPreset->setVisible(config.debug_mode);
    ui->comboBoxPreset->setVisible(config.debug_mode);
    ui->actionReset_power_status->setVisible(false);

    ui->groupBoxTrigger->setStyleSheet(gbox_fill_blue);
    ui->groupBoxReadout->setStyleSheet(gbox_fill_blue);
    ui->groupBoxStat->setStyleSheet(gbox_fill_blue);

    if(start_options.debug_mode) {
        connect(adjustDialog, &AdjustDialog::adjust_setup_changed, this, &MainWindow::adjust_setup_changed);
    }
    ui->actionAdjust_HPTDC->setVisible(start_options.debug_mode);
    ui->actionTrace_dev_info_changes->setVisible(start_options.debug_mode);
    ui->actionShow_power_current->setVisible(start_options.debug_mode);
    ui->actionon_board_PLL_bypass->setVisible(start_options.debug_mode);

    block_widgets_signal(true);
    ui->comboBoxEdges->addItem("Leading", 1);
    ui->comboBoxEdges->addItem("Trailing", 2);
    ui->comboBoxEdges->addItem("Lead+Trail", 3);

    for(auto preset : StatPresetFabric::getAllPresets()){
        ui->comboBoxPreset->addItem(preset.label, preset.type);
    }

    ui->actionAllow_remote_control->setChecked(config.allow_remote_control);
    emit allowRemoteControl(config.allow_remote_control);

    ui->actionon_board_PLL_bypass->setChecked(config.default_setup.pll_bypass);
    block_widgets_signal(false);

    if(config.sort_column >= 0) {
        ui->tableWidgetDevice->sortByColumn(config.sort_column,
                                            config.sort_asc ? Qt::AscendingOrder : Qt::DescendingOrder);
    }
    ui->actionConfiguration_window->setChecked(config.show_cfg_widget);

    ui->splitter_vertical->setCollapsible(0, false);
    ui->splitter_vertical->setCollapsible(1, false);
    ui->splitter_vertical->setStretchFactor(0,1);
    ui->splitter_vertical->setStretchFactor(1,2);

    if(adjustDialog) {
        adjustDialog->setConfig(config.default_setup.adjust_set);
    }

}

void MainWindow::switch_to_config(QString program_index, QString configuration_name)
{
    configuration_manager->hide();
    config.program_index = program_index;
    config.configuration_name = configuration_name;

    for (const DeviceIndex &i: devGlob.keys()) {
        disconnectFromDev(i);
    }

    read_config();

    if(adjustDialog) {
        adjustDialog->setConfig(config.default_setup.adjust_set);
    }
    updateControlState();

    block_widgets_signal(true);
    ui->actionAllow_remote_control->setChecked(config.allow_remote_control);
    emit allowRemoteControl(config.allow_remote_control);
    block_widgets_signal(false);

}

void MainWindow::on_actionAllow_remote_control_toggled(bool checked)
{
    config.allow_remote_control = checked;
    config.write_config();
    emit allowRemoteControl(config.allow_remote_control);
    updateProgramDescription();
}

void MainWindow::on_actionon_board_PLL_bypass_toggled(bool pll_bypass)
{
    config.default_setup.pll_bypass = pll_bypass;
    param_updated();
}

void MainWindow::updateProgramDescription()
{
    program_description.type = config.program_type;
    program_description.index = config.program_index;
    program_description.name = program_description.getName();

#ifdef GIT_DESCR
    program_description.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    program_description.ver_date = GIT_LAST_COMMIT;
#endif

    program_description.options.clear();
    program_description.options["run"] = config.run ? "true" : "false";
    program_description.options["run_index"] = config.runId.runIndex;
    program_description.options["run_number"] = QString::number(config.runId.runNumber);

    int i=0;
    for(DeviceIndex &index : config.knownSetups.keys()){
        if(!config.knownSetups[index].enabled)
            continue;
        program_description.options[QString("TDC_%1").arg(i)] = QString("0x%1;0x%2")
                .arg(index.first, 2, 16, QChar('0'))
                .arg(index.second, 8, 16, QChar('0'));
        ++i;
    }
    emit changeProgramDescription(program_description);
}

void MainWindow::onRemoteStartRequest(const RcRunId &runId, quint64 start_event_number)
{
    auto label_status = ui->label_remote_status;
    setLabelColor(label_status, QColor("orange"));
    label_status->setText("Preparing for start...");
    qInfo() << QString("Got remote request to start Run %1")
               .arg(QString(runId));
    Q_UNUSED(start_event_number)
    bool success=true;
    if(config.runId.runIndex == runId.runIndex &&
            config.runId.runNumber > runId.runNumber) {
        qWarning() << QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                      .arg(config.runId.runNumber)
                      .arg(runId.runNumber);
    }

    if(success) {
        config.runId.set(runId);
        success &= perform_start();
        config.write_config();
        on_actionReset_MStream_triggered();
        // todo: remove this dummy wait (for mstream connect to device)
        QThread::msleep(2000);
    }

    if(!success) {
        qWarning() << "Something was wrong while executing remote start";
    } else {
        qInfo() << QString("Remote start completed successfully for run %1")
                   .arg(QString(runId));
    }

    emit remoteStartCompleted(success, runId);
    if(success){
        emit config_transfer_to_rc(config.root.get_json().toBinaryData());
        label_status->setText(QString("Running %1...")
                              .arg(QString(runId)));
        setLabelColor(label_status, Qt::darkGreen);
    } else {
        label_status->setText(QString("Failed to start run %1")
                              .arg(QString(runId)));
        setLabelColor(label_status, Qt::red);
    }

    updateProgramDescription();
}

void MainWindow::onRemoteStopRequest(const RcRunId &runId, quint64 start_event_number)
{
    auto label_status = ui->label_remote_status;
    label_status->setText("Stopping run...");
    setLabelColor(label_status, QColor("orange"));
    Q_UNUSED(start_event_number)
    bool success=true;
    if(config.runId.runIndex != runId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(config.runId.runIndex)
                     .arg(runId.runIndex);
        success = false;
    }
    if(success && config.runId.runNumber != runId.runNumber){
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(config.runId.runNumber)
                     .arg(runId.runNumber);
        success = false;
    }
    if(success) {
        success &= perform_stop();
    }

    if(success) {
        qInfo("Remote stop completed successfully");
        ui->label_remote_status->setText(QString("Run finished: %1")
                                         .arg(QString(runId)));
        setLabelColor(label_status, Qt::black);
    } else {
        qWarning("Something was wrong while executing remote stop");
        ui->label_remote_status->setText(QString("Errors on run %1...")
                                         .arg(QString(runId)));
        setLabelColor(label_status, Qt::red);
    }

    emit remoteStopCompleted(success, runId);
        updateProgramDescription();
}

void MainWindow::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(!program_description.containsProgramInterface(ProgramInterfaceRemoteControl, 0)) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        interface.port = 0;
        interface.enabled = config.allow_remote_control;
        interface.isFree = false;
        program_description.interfaces.push_back(interface);
    }
    const bool &port_in_use = status.portInUse;
    on_remote = port_in_use && status.listening;

    ProgramInterface &i = program_description.getProgramInterface(ProgramInterfaceRemoteControl);
    ProgramInterface prevIt = i;
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !port_in_use;

    if(port_in_use) {
        i.setPeer(status.peerHost, status.peerPort);
    } else {
        i.peers.clear();
        if(config.run)
            perform_stop();
    }

    updateRemoteStatus();
    updateControlState();
    enable_gui_controls();
    ui->label_remote_status->setVisible(port_in_use);

    // lock all control widgets
    updateProgramDescription();
}

void MainWindow::updateRemoteStatus()
{
    if(!on_remote || !update_remote_enabled)
        return;
    QString readyStr, notReadyStr;
    bool ready4Run = true;

    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        const auto &dg = dg_it.value();
        const auto &curSt = dg->pendingDevStatus;
        const bool devOk = dg->setup_index == setup_index;
        const auto devStr = QString("0x%1 ").arg(dg_it.key().getSerial(), 8, 16, QChar('0'));
        if(devOk && curSt.connected
//                && dg.msStat == QProcess::Running
                ) {
            readyStr.append(devStr);
        } else {
            notReadyStr.append(devStr);
            ready4Run = false;
        }
        ++dg_it;
    }

    QString stateStr;
    if(!notReadyStr.isEmpty())
        stateStr.append("Not ready:").append(notReadyStr);
    if(!readyStr.isEmpty()) {
        if(!stateStr.isEmpty())
            stateStr.append("| ");
        stateStr.append("Ready:").append(readyStr);
    }
    if(stateStr.isEmpty())
        stateStr = "No devices";
    emit sendRemoteState(stateStr, ready4Run);
}

void MainWindow::enable_gui_controls()
{
    const bool en = !on_remote || config.debug_mode;

    ui->actionReset_MStream->setEnabled(en);
    ui->actionSend_soft_reset->setEnabled(en);
    ui->actionForce_write_setup->setEnabled(en);
    ui->actionAdjust_HPTDC->setEnabled(en);
    ui->actionon_board_PLL_bypass->setEnabled(en);
    ui->menuHPTDC_Event_hdr_trl->setEnabled(en);
    ui->actionConfiguration_manager->setEnabled(en);
    updateControlState();
}

void MainWindow::sendRunError(QString mess)
{
    if(config.isRun() && config.allow_remote_control)
        emit runError(config.runId,
                      QString("TDC #%1 %2").arg(config.program_index).arg(mess));
}

void MainWindow::sendRunWarning(QString mess)
{
    if(config.isRun() && config.allow_remote_control)
        emit runWarning(config.runId,
                        QString("TDC #%1 %2").arg(config.program_index).arg(mess));
}

void MainWindow::temperatureSet(const TdcStatus &status)
{
    const auto devMetricPrefix = getDevMetricPrefix(status.getIndex())+"temp.";
    const QMap<int, double> &temps = status.temp;
    auto table = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(table);
    auto updateLocker = TableUpdateLocker(table);
    int row = getRowByDevice(status.getIndex());
    if(temps.isEmpty() || row==-1) return;
    QList<double> tempList = temps.values();
    double minTemp = *std::min_element(tempList.begin(), tempList.end());
    double maxTemp = *std::max_element(tempList.begin(), tempList.end());

    const bool tempWarn = maxTemp>=70 || minTemp<0;
    bool warn_changed = false;
    auto dg_it = devGlob.find(status.getIndex());
    if(dg_it != devGlob.end()) {
        auto &warn = dg_it.value()->warn_flags[COL_DEV_TEMP];
        if(warn != tempWarn) {
            warn = tempWarn;
            warn_changed = true;
        }
        if(warn) {
            sendRunWarning(QString("%1 temp %2-%3")
                           .arg(dg_it.key().getSerialStr())
                           .arg(minTemp).arg(maxTemp));
        }
    }

    QString tempStr = QString::number(minTemp,'f',1);
    QString tempMaxStr = QString::number(maxTemp,'f',1);
    if(tempStr != tempMaxStr) tempStr += QString("-%2").arg(tempMaxStr);
    tempStr.append(QChar(0x02da));
    QStringList fullInfo;
    for (int tempId: temps.keys()) {
        MetricSender::send(devMetricPrefix + QString::number(tempId), temps[tempId]);
        QString suffixStr;
        if(status.deviceId == DEVICE_ID_TDC64VHLE) {
            if(tempId) {
                char c = 'A';
                c += tempId-1;
                suffixStr = QString("HPTDC Chain %1").arg(c);
            } else
                suffixStr = "Board temp.";
        } else {
            switch (tempId) {
            case 0:
                suffixStr = "Board temp.";
                break;
            case 1:
                suffixStr = "Pll temp.";
                break;
            case 2:
            case 3:
            case 4:
                suffixStr = QString("Ch %1-%2").arg((tempId-2)*24).arg((tempId-1)*24-1);
                break;
            default:
                break;
            }
        }
        if(!suffixStr.isEmpty())
            fullInfo << QString("%1%2 - %3").arg(temps[tempId],0,'f',1).arg(QChar(0x02da)).arg(suffixStr);
    }
    if(warn_changed)
        table_warn_updated(status.getIndex(), COL_DEV_TEMP);
    QTableWidgetItem * item = table->item(row, COL_DEV_TEMP);
    item->setText(tempStr);
    item->setToolTip(fullInfo.join('\n'));

    updateTableSize(table);
}

void MainWindow::setDeviceStatus(const QString &status)
{
    return; // planing to delete this column
    DeviceIndex index;
    for (const DeviceIndex &i: devGlob.keys()) {
        if(devGlob[i]->dev == sender()) {
            index = i;
            break;
        }
    }
    if(index.isNull())
        return;
    int row = getRowByDevice(index);
    if(row == -1) {
        row = getRowByDevice(index);
        Q_ASSERT(row!=-1);
        return;
    }
    ui->tableWidgetDevice->item(row, COL_DEV_STATUS)->setText(status);
}

void MainWindow::deviceStatusUpdated(const TdcStatus &status)
{
    const auto index = status.getIndex();
    auto dg_it = devGlob.find(index);
    if(dg_it == devGlob.end())
        return;

    auto &dg = dg_it.value();
    bool doMStreamReset = (dg->pendingDevStatus.mstreamHwBufSize != status.mstreamHwBufSize)
            || (dg->pendingDevStatus.mstreamMultiAck != status.mstreamMultiAck);
    dg->pendingDevStatus = status;

    if(status.connected) {
        if(dg->mstream) {
            if(doMStreamReset)
                resetMStream(index);
        } else {
            addMStreamDump(index);
        }
    }

    if(status.bmc_status.power.isCriticalCurrent() ||
            status.isMStreamRestartNeeded(dg->lastDevStatus)) {
        applyDeviceStatusUpdates();
    }
}

void MainWindow::applyDeviceStatusUpdates()
{
    QTableWidget *table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);
    auto sortLocker = TableSortLocker(table);

    auto dgIt = devGlob.begin();
    while(dgIt != devGlob.end()) {
        if(!dgIt.value()->pendingDevStatus.processed)
            applySingleDeviceStatusUpdates(dgIt.key());
        ++dgIt;
    }

    if(updateLocker.up_en) {
        updateTableSize(table);
    }

    updateRemoteStatus();
    deviceStatusTimer->start(500);
}

void MainWindow::applySingleDeviceStatusUpdates(DeviceIndex index)
{
    auto dg_it = devGlob.find(index);
    if(dg_it == devGlob.end())
        return;

    auto &dg = dg_it.value();
    TdcStatus &status = dg->pendingDevStatus;
    if(status.processed)
        return;
    status.processed = true;
    TdcStatus &lastDevStatus = dg->lastDevStatus;

    QTableWidget *table = ui->tableWidgetDevice;
    const int row = getRowByDevice(status.getIndex());
    if(row == -1) {
        lastDevStatus = status;
        return;
    }
    const auto &statistic = status.statistic;
    auto updateLocker = TableUpdateLocker(table);
    if(status.connected) {
        const auto devMetricPrefix = getDevMetricPrefix(index);
        table->item(row, COL_DEV_RUN_TIME)->setText(status.statistic.getRunTime());

        temperatureSet(status);
        table->item(row, COL_DEV_FW)->setText(status.fwStr);
        deviceBmcStatusUpdated(row, status);

//        table->item(row, COL_DEV_TIME)->setText(QDateTime::fromTime_t(status.wrTime).toString());
        table->item(row, COL_DEV_CUR_EV_NUM)->setText(QString("%L1").arg(status.curEvNum));
        MetricSender::send(devMetricPrefix + "curEvNum", status.curEvNum);
        table->item(row, COL_DEV_UNBLOCK)->setText(QString("%L1").arg(statistic.evNumAccepted));
        table->item(row, COL_DEV_BLOCK)->setText(QString("%L1").arg(statistic.getBlocked()));

        QString str;
        bool hptdcErr = status.hptdcChipErrorFlags;
        if(status.hptdcChipErrorFlags)
            str += QString("HPTDC Err flags: %1").arg(status.hptdcChipErrorFlags, 0, 2);
        for (int hptdcId: status.hptdcStatus.keys()) {
            const HptdcStatus &hptdcStatus = status.hptdcStatus[hptdcId];
            if(!hptdcStatus.getSummary().isEmpty()) {
                if(!str.isEmpty()) str += "\n";
                str += QString("HPTDC #%1:\n%2").arg(hptdcId).arg(hptdcStatus.getSummary());
            }
            hptdcErr |= hptdcStatus.hasError();
        }
        if(str.isEmpty()) {
            str = "Ok";
        }

        table->item(row, COL_HPTDC_STATUS)->setText(str);

        qobject_cast<QLabel*>(table->cellWidget(row, COL_CLOCK_STATUS))
                ->setText(status.getClockStatus());
        table->item(row, COL_BUSY_STATUS)->setText(QString("%1").arg(status.busyStatus, 2, 2, QChar('0')));
        table->item(row, COL_TRIG_QUEUE)->setText(QString("%1").arg(status.trigQueue));
        if(statistic.evNumAccepted != lastDevStatus.statistic.evNumAccepted){
            table->item(row, COL_BUSY_TIME)->setText(statistic.getBusyStr(config.debug_mode, lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.total", statistic.getBusyTotal(lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.tdc", statistic.getBusyTdc(lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.readout", statistic.getBusyReadout(lastDevStatus.statistic));
        }
        MetricSender::send(devMetricPrefix+"time.uptime", status.statistic.getRunTimeSec());
        QStringList errCntStr;
        auto errCnt = statistic.errCnt;
        auto &powerOnErrCnt = statistic.powerOnErrCnt;
        errCnt -= powerOnErrCnt;

        QString errStr = errCnt.toString();
        bool errCntFlag = !errStr.isEmpty();
        auto errCntItem = table->item(row, COL_ERR_CNT);
        errCntItem->setText(errStr);

        bool warn_changed = false;
        if(dg->warn_flags[COL_HPTDC_STATUS] != hptdcErr) {
            dg->warn_flags[COL_HPTDC_STATUS] = hptdcErr;
            warn_changed = true;
        }
        if(dg->warn_flags[COL_ERR_CNT] != errCntFlag) {
            dg->warn_flags[COL_ERR_CNT] = errCntFlag;
            warn_changed = true;
        }
        if(warn_changed)
            table_warn_updated(index, COL_HPTDC_STATUS);
    } else {
//        table->item(row, COL_DEV_STATUS)->setText("Offline");
//        table->item(row, COL_DEV_STATUS)->setTextColor(Qt::black);
    }
    if(updateLocker.up_en) {
        updateTableSize(table);
    }

    const bool doResetMStream = status.isMStreamRestartNeeded(lastDevStatus);
    lastDevStatus = status;
    if(doResetMStream) {
        resetMStream(index);
    }
}

void MainWindow::deviceBmcStatusUpdated(int row, const TdcStatus &tdcStatus)
{
    const BmcStatus &status = tdcStatus.bmc_status;
    const auto index = tdcStatus.getIndex();
    if(row<0) return;
    QTableWidget *table = ui->tableWidgetDevice;
    Q_ASSERT(!table->isSortingEnabled());
    QColor status_col = status.get_overall_status_col();
    const bool mcu_ver_warn = status.overall_status==0b11;
    auto mcu_ver_item = table->item(row, COL_DEV_MCU_VER);
    mcu_ver_item->setText(status.getVerStr());
    auto bg_brush = mcu_ver_item->background();
    if(bg_brush.color() != status_col) {
        bg_brush.setColor(status_col);
        bg_brush.setStyle(Qt::SolidPattern);
        mcu_ver_item->setBackground(bg_brush);
    }
    mcu_ver_item->setToolTip(QString("overall - %1\npower - %2\nPLL- %3")
                             .arg(status.get_status_str(status.overall_status))
                             .arg(status.get_status_str(status.power_status))
                             .arg(status.get_status_str(status.pll_status)));

    auto mcu_pll_unlock_item = table->item(row, COL_DEV_PLL_UNLOCK);
    mcu_pll_unlock_item->setText(QString::number(status.pllUnlockCnt));
    bool prev_was_set;
    bool refresh_col = (mcu_pll_unlock_item->data(Qt::UserRole).toInt(&prev_was_set)==0) != (status.pllUnlockCnt==0);
    mcu_pll_unlock_item->setData(Qt::UserRole, status.pllUnlockCnt);
    if(refresh_col || !prev_was_set) update_pll_unlock_visible();
//    auto mcu_pll_st_item = table->item(row, COL_DEV_PLL_STATUS);
//    mcu_pll_st_item->setText(status.onboard_pll_ok() ? "Ok" : "Fail");
//    mcu_pll_st_item->setBackgroundColor(status.onboard_pll_ok() ? Qt::green : Qt::red);

    // Power
    const BmcPower &power = status.power;
    const auto powerItem = table->item(row, COL_DEV_POWER);
    const bool powerCritical = power.isCriticalCurrent();
    auto dg_it = devGlob.find(index);
    if(dg_it != devGlob.end()) {
        bool warn_changes = false;
        auto &warn_flags = dg_it.value()->warn_flags;
        if(powerCritical && warn_flags[COL_DEV_POWER]!=powerCritical) {
            warn_flags[COL_DEV_POWER]=powerCritical;
            warn_changes = true;
        }
        if(warn_flags[COL_DEV_MCU_VER]!=mcu_ver_warn) {
            warn_flags[COL_DEV_MCU_VER]=mcu_ver_warn;
            warn_changes = true;
        }
        if(warn_changes)
            table_warn_updated();
        if(powerCritical) {
            sendRunWarning(QString("%1 power crit.").arg(dg_it.key().getSerialStr()));
        }
    }

    if(config.debug_mode || powerCritical || criticalPowerMode) {
        powerItem->setText(power.getSummary(config.show_power_current));
        if(powerCritical) {
            logDevInfo(powerItem, true);
        }
    }
}

void MainWindow::deviceConnected()
{
    deviceConnectStatueChanged(true);
}

void MainWindow::deviceDisconnected()
{
    deviceConnectStatueChanged(false);
}

void MainWindow::table_row_color(QTableWidget *table, int row, QColor color)
{
    for(int i=0; i<COL_DEV_TOTAL; ++i) {
        if(i==COL_CLOCK_STATUS) {
            auto label = qobject_cast<QLabel*>(table->cellWidget(row, i));
            if(label) {
                auto palette = label->palette();
                label->setAutoFillBackground(true);
                palette.setColor(label->backgroundRole(), color);
                label->setPalette(palette);
            }
        } else {
            setItemBgColor(table->item(row, i), color);
        }
    }
}

bool MainWindow::wait_device_setup_written()
{
    QElapsedTimer timer;
    timer.start();
    bool ok=false;
    while (!ok && timer.elapsed()<1000) {
        ok=true;
        QCoreApplication::processEvents(QEventLoop::AllEvents, 10);
        auto dg_it = devGlob.begin();
        while(ok && (dg_it != devGlob.end())) {
            if(dg_it.value()->setup_index != setup_index) {
                ok = false;
                break;
            }
            ++dg_it;
        }
    }
    if(!ok)
        qWarning() << "Some devices failed to write setup";
    return ok;
}

void MainWindow::increase_setup_index()
{
    ++setup_index;
    if(setup_index<100)
        setup_index=100;
}

bool MainWindow::set_dev_warn(DeviceIndex index, int col, bool warn_flag)
{
    bool warn_changes = false;
    auto dg_it = devGlob.find(index);
    if(dg_it != devGlob.end()) {
        auto &warn = dg_it.value()->warn_flags[col];
        if(warn != warn_flag) {
            warn = warn_flag;
            warn_changes = true;
        }
    }
    return warn_changes;
}

void MainWindow::table_warn_updated()
{
    auto table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);

    QSet<DeviceIndex> dev_warn;
    for(int col=0; col<COL_DEV_TOTAL; ++col) {
        bool col_warn_flag=false;
        auto dg_it = devGlob.begin();
        while(dg_it != devGlob.end()) {
            auto &dg = dg_it.value();
            bool item_warn = dg->warn_flags[col];
            if(item_warn)
                dev_warn.insert(dg->index);
            col_warn_flag |= item_warn;
            QColor fg_color = item_warn ? Qt::red : Qt::black;
            QColor bg_color = Qt::transparent;
            if(col==COL_HPTDC_STATUS || col==COL_DEV_POWER || col==COL_ERR_CNT) {
                fg_color = Qt::black;
                bg_color = item_warn ? Qt::red : Qt::green;
//            } else if(col==COL_DEV_MCU_VER) {
//                // specific logic
            }

            const int row = getRowByDevice(dg->index);
            auto item = table->item(row, col);
            if(item && col!=COL_DEV_MCU_VER) {
                setItemTextColor(item, fg_color);
                setItemBgColor(item, bg_color);
            }
            ++dg_it;
        }
        auto col_item = table->horizontalHeaderItem(col);
        QColor col_color = col_warn_flag ? Qt::red : Qt::black;
        if(1) {
            col_item->setTextColor(col_color);
        } else {
            // Setting color to horizontal headers doesn't work
            if(col==COL_DEV_POWER)
                qInfo() << col << col_item->text() << col_item->textColor();
            setItemTextColor(col_item, col_color);
            if(col==COL_DEV_POWER)
                qInfo() << col << col_item->text() << col_item->textColor() << col_color;
        }
        // Additional actions for group (column)
        switch (col) {
        case COL_DEV_POWER:
            criticalPowerMode=col_warn_flag;
            ui->actionReset_power_status->setVisible(criticalPowerMode);
            if(criticalPowerMode && col_warn_flag)
                ui->actionShow_power_current->setChecked(true);
            table->setColumnHidden(col, !config.debug_mode && !col_warn_flag);
            break;
        case COL_ERR_CNT:
            ui->actionReset_error_counters->setVisible(col_warn_flag);
            table->setColumnHidden(col, !col_warn_flag);
            break;
        }
    }
    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        auto &dg = dg_it.value();
        dg->dev_warn = dev_warn.contains(dg->index);
        int row = getRowByDevice(dg->index);
        setItemTextColor(table->verticalHeaderItem(row), dg->getColor(), true);
        ++dg_it;
    }

    updateTableSize(table);
}

void MainWindow::table_warn_updated(DeviceIndex warn_index, int warn_col)
{
    Q_UNUSED(warn_index)
    Q_UNUSED(warn_col)
    table_warn_updated();
}

void MainWindow::deviceConnectStatueChanged(bool connected)
{
    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        if(dg_it.value()->dev == sender())
            break;
        ++dg_it;
    }

    if(dg_it == devGlob.end()) return;
    dg_it.value()->dev_online = connected;
    auto table = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(table);
    const int row = getRowByDevice(dg_it.key());
    if(row == -1) return;

    setItemTextColor(table->item(row, COL_DEV_SERIAL), connected ? Qt::darkGreen : Qt::darkGray);
    setItemTextColor(table->verticalHeaderItem(row), dg_it.value()->getColor(), true);
    table_row_color(table, row, connected ? Qt::white : Qt::gray);
    updateTableSize(table);
    if(config.isRun() && !connected) {
        sendRunError(QString("%1 offline").arg(dg_it.key().getIdent()));
    }
}

void MainWindow::deviceUpdated(const DeviceDescription &dd)
{
    knownDevDescr[dd.getIndex()] = dd;
    auto sortLocker = TableSortLocker(ui->tableWidgetDevice);

    const int row = getRowByDevice(dd.getIndex());
    if(row == -1) {
        return;
    }
    QString txt;
    if(dd.chassis_slot){
        txt = QString("%1").arg(dd.chassis_slot, 2, 10, QChar('0'));
    }
    auto itemSlot = ui->tableWidgetDevice->item(row, COL_DEV_SLOT);
    if(itemSlot && itemSlot->text() != txt) {
        itemSlot->setText(txt);
        updateModelList();
    }
}

void MainWindow::programDiscovered(const ProgramDescription &pd)
{
    if(pd.type == DaqConfig::getTypeName(DaqConfigMStream)){
        bool ok;
        quint32 progSerial = pd.index.toUInt(&ok, 16);
        if(!ok) return;
        for (const DeviceIndex &index: devGlob.keys()) {
            if(index.getSerial() == progSerial) {
                ProgramInterface i = pd.getProgramInterface(ProgramInterfaceDataFlow);
                if(!i.isEmpty()){
//                    qInfo()<<"mstream interface updated:"<<i.host.toString()<<":"<<i.port;
//                    emit setTcpAddress(index, i.host, i.port);
                }
            }
        }
    }
}

void MainWindow::deviceInfoUpdated(DeviceIndex index, int chNum)
{
    if(config.knownSetups[index].getNCh() != chNum) {
        config.knownSetups[index].setNCh(chNum);
        writeSettings();
        updateModelList();
        configUpdated(index);
    }

    ui->spinBoxLatency->setMaximum(2048);
//    ui->spinBoxMatchWin -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
    updateGuiVal();
}

void MainWindow::deviceSetupWritten(DeviceIndex index, quint32 setup_index)
{
    auto dg_it=devGlob.find(index);
    if(dg_it != devGlob.end()) {
        dg_it.value()->setup_index = setup_index;
        updateRemoteStatus();
    }
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.program_index, config.configuration_name);
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::on_actionChange_device_order_triggered()
{
    OrderDialog dialog(&config, knownDevDescr, this);
    if(dialog.exec() == QDialog::Accepted) {
        config.knownSetups = dialog.get_known_setups();
        writeSettings();
        updateDeviceOrder(true);
    }
}

void MainWindow::on_actionReset_MStream_triggered()
{
    for (DeviceIndex i: devGlob.keys()) {
        resetMStream(i);
    }
}

void MainWindow::on_actionSend_soft_reset_triggered()
{
    emit softReset();
}

void MainWindow::on_actionTxt_data_format_triggered()
{
    QString s;
    s += "<h2>TDC2 Txt Data format</h2>";
    s += "<p>File constructs with following lines:</p>";
    s += "<ul>";
    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
    s += "</ul>";
    s += "<p>where:</p>";
    s += "<dl>";
    s += "<dt><b>TAI time</b></dt>";
    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
    s += "<dt><b>ch</b></dt>";
    s += "<dd><i>Channel number in range 0-15</i></dd>";
    s += "<dt><b>tdcLeadTime</b></dt>";
    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>adcTime</b></dt>";
    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>sample</b></dt>";
    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
    s += "</dl>";

    QMessageBox::about (this, "Txt data format", s);
}

void MainWindow::on_actionRaw_data_format_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
//    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d.setLayout(mainLayout);
    d.exec();

//    QMessageBox::about (this, "Raw data format", html);
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::customContextMenuRequested(const QPoint &pos)
{
    QMenu contextMenu(tr("Context menu"), this);

    QAction action1("Enabled", this);
    connect(&action1, &QAction::triggered, this, &MainWindow::customContextMenuAction);
    action1.setData(true);
    contextMenu.addAction(&action1);

    QAction action2("Disabled", this);
    action2.setData(false);
    connect(&action2, &QAction::triggered, this, &MainWindow::customContextMenuAction);
    contextMenu.addAction(&action2);

    contextMenu.exec(ui->treeWidgetChCfg->mapToGlobal(pos));
}

void MainWindow::customContextMenuAction()
{
    auto s = qobject_cast<QAction*>(QObject::sender());

    if(s == nullptr) return;
    bool enable = s->data().toBool();

    auto tree = ui->treeWidgetChCfg;
    auto updateLocker = TableUpdateLocker(tree);
    for(auto item : tree->selectedItems()) {
        item->setCheckState(tree_tdc_name, enable ? Qt::Checked : Qt::Unchecked);
    }
}

void MainWindow::on_pushButtonChangeDevice_clicked()
{
    if(discoverDialog->exec() == QDialog::Accepted) {
        QMap<DeviceIndex, DeviceSetup> oldSetups = config.knownSetups;
        QMap<DeviceIndex, DeviceSetup> newSetups;
        for (const DeviceDescription &dd: discoverDialog->getSelectedDevices()) {
            const DeviceIndex index = dd.getIndex();
            newSetups[index] = config.getConfig(index);
            newSetups[index].enabled = dd.enabled;
        }

        DeviceIndex selected_device;
        if(newSetups.contains(config.curDev) && newSetups[config.curDev].enabled)
            selected_device = config.curDev;

        // clean-up
        for (const DeviceIndex &index: oldSetups.keys()) {
            if(!oldSetups[index].enabled) continue;
            if(!newSetups.contains(index) || !newSetups[index].enabled)
                disconnectFromDev(index);
        }

        config.knownSetups = newSetups;
        // add new
        for (const DeviceIndex &index: newSetups.keys()) {
            auto newSetup = newSetups[index];
            if(!newSetup.enabled) {
                continue;
            }
            if(!oldSetups.contains(index) || !oldSetups[index].enabled){
                newSetup.copy_config(config.default_setup);
                connectToDev(index);
                if(index==selected_device) {
                    int row = getRowByDevice(selected_device);
                    Q_ASSERT(row!=-1);
                    ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setSelected(true);
                }
            }
        }

        writeSettings();

        updateDeviceOrder();
        updateControlState();
    }
}

void MainWindow::on_tableWidgetDevice_itemSelectionChanged()
{
    QTableWidget *table = ui->tableWidgetDevice;
    QList<QTableWidgetSelectionRange> ranges = table->selectedRanges();
    config.curDev = DeviceIndex();
    if(ranges.isEmpty()) {
        updateGuiVal();
        return;
    }
    if(ranges.size()!=1) {
        updateGuiVal();
        return;
    }
    Q_ASSERT(ranges.size()==1);
    QTableWidgetSelectionRange range = ranges.first();
    Q_ASSERT(range.bottomRow() == range.topRow());
    int row = range.bottomRow();

    DeviceIndex index = getDevIndexByRow(row);
    if(!index.isNull() && devGlob.contains(index)
            && config.knownSetups.contains(index)
            && config.knownSetups[index].enabled) {
        config.curDev = index;
        config.default_setup.copy_config(config.getConfig(index));
    }
    updateGuiVal();
}

void MainWindow::on_tableWidgetDevice_itemChanged(QTableWidgetItem *item)
{
    logDevInfo(item);
}

void MainWindow::on_tableWidgetDevice_cellDoubleClicked(int row, int column)
{
    if(column != COL_DEV_FW) return;
    DeviceIndex index = getDevIndexByRow(row);
    if(!devGlob.contains(index)) return;

    const auto &stat = devGlob[index]->lastDevStatus;

    const auto &power = stat.bmc_status.power;
    QStringList powerList;
    for(auto id : power.vol.keys()) {
        auto str = QString("%1-%2").arg(power.getIdStr(id)).arg(1e-3*power.vol[id],0,'f',3);
        if(str.contains("TDC",Qt::CaseInsensitive))
            powerList << str.remove("TDC_");
    }
    QString powerStr = powerList.join(' ').prepend("Long test passed. Voltages ");

    QStringList list;
    list << stat.fwStr.split('.').last();
    list << stat.bmc_status.getVerStr();
    list << QDateTime::currentDateTime().toString("dd.MM.yyyy");
    list << "Robot";
#ifdef GIT_DESCR
    list << GIT_DESCR;
#endif
    list << powerStr;

    QApplication::clipboard()->setText(list.join('\t'));
}

void MainWindow::sortIndicatorChanged(int logicalIndex, Qt::SortOrder order)
{
    config.sort_column = logicalIndex;
    config.sort_asc = order==Qt::AscendingOrder;
}

void MainWindow::on_pushButtonStart_clicked()
{
    perform_start();
    updateControlState();
}

bool MainWindow::perform_start()
{
    increase_setup_index();
    if(!config.has_enabled_device())
        return true;

    update_remote_enabled = false;
    config.run = true;
    config.default_setup.run = true;
    for (const DeviceIndex &index: devGlob.keys()) {
        DeviceSetup &ds  = config.getConfig(index);
        if(ds.enabled) {
            ds.run = true;
            ds.setup_index = setup_index;
        }
    }

    configUpdated();
    writeSettings(true);

    const bool ok = wait_device_setup_written();
    update_remote_enabled = true;
    updateRemoteStatus();
    return ok;
}

void MainWindow::on_pushButtonStop_clicked()
{
    perform_stop();
    updateControlState();
}

bool MainWindow::perform_stop()
{
    increase_setup_index();
    bool noDevice = true;

    update_remote_enabled = false;
    config.run = false;
    config.default_setup.run = false;
    for (const DeviceIndex &index: devGlob.keys()) {
        DeviceSetup &ds  = config.getConfig(index);
        if(ds.enabled) {
            ds.run = false;
            noDevice = false;
            ds.setup_index = setup_index;
        }
    }
    if(noDevice)
        return true;
    configUpdated();

    const bool ok = wait_device_setup_written();
    update_remote_enabled = true;
    updateRemoteStatus();
    return ok;
}

void MainWindow::on_radio_button_global_toggled(bool checked)
{
    if(blockControl) return;
    config.global_mode = checked;
    if(!config.global_mode) return;

    for(DeviceIndex i : config.knownSetups.keys())
        config.getConfig(i).copy_config(config.default_setup);

    configUpdated();
}

void MainWindow::on_checkBoxTrigTimer_toggled(bool checked)
{
    config.default_setup.trigsetup.bit.timer_en = checked ? 1 : 0;
    param_updated();
    ui->spinBoxTimeTrig->setEnabled(checked);
}

void MainWindow::on_spinBoxTimeTrig_valueChanged(int rateHz)
{
    const quint32 period = static_cast<quint32>(125e6/rateHz/3); // period in 24ns
    config.default_setup.trigPeriod = period;
    param_updated();

    updateRealRate();
}

void MainWindow::on_checkBoxTrigTtl_toggled(bool checked)
{
    config.default_setup.trigsetup.bit.ttl_en = checked ? 1 : 0;
    param_updated();
}

void MainWindow::on_checkBoxTrigTtc_toggled(bool checked)
{
    config.default_setup.trigsetup.bit.ttc_en = checked ? 1 : 0;
    param_updated();
}

void MainWindow::on_checkBoxTrigVxs_stateChanged(int arg1)
{
    config.default_setup.trigsetup.bit.vxs_en = (Qt::Checked == arg1) ? 1 : 0;
    param_updated();
}

void MainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    ui->spinBoxMatchWin->setMaximum(lat);

    config.default_setup.latency = static_cast<quint32>(lat);
    param_updated();
}

void MainWindow::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    emit matchWinUpdated(win);

    config.default_setup.matchWin = static_cast<quint32>(win);
    param_updated();
}

void MainWindow::on_comboBoxEdges_currentIndexChanged(int index)
{
    bool ok;
    int v = ui->comboBoxEdges->itemData(index).toInt(&ok);
    Q_ASSERT(ok);

    config.default_setup.leadEdge = v&1;
    config.default_setup.trailEdge = v&2;
    param_updated();
}

void MainWindow::on_groupBoxStat_toggled(bool checked)
{
    config.default_setup.stat_readout = checked;
    param_updated();
}

void MainWindow::on_spinBoxStatPeriod_valueChanged(int period_ms)
{
    config.default_setup.stat_period = static_cast<quint16>(period_ms);
    param_updated();
}

void MainWindow::on_comboBoxPreset_currentIndexChanged(int index)
{
    StatPreset stat_preset;
    int preset_type = ui->comboBoxPreset->itemData(index).toInt();
    stat_preset = StatPresetFabric::getPreset(preset_type);

    config.default_setup.stat_preset = stat_preset;
    // use for all  devices regardless of global_mode
    for(const DeviceIndex &index: config.knownSetups.keys()){
        config.getConfig(index).stat_preset = stat_preset;
    }
    configUpdated();
}

void MainWindow::on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column)
{
    updateTreeSize(ui->treeWidgetChCfg);

    if(blockControl)
        return;
    if(column != tree_tdc_name)
        return;
    Qt::CheckState state = item->checkState(tree_tdc_name);
    if(state == Qt::PartiallyChecked)
        return;
    bool checked = (state == Qt::Checked);
    bool modified = false;
    if(item==rootTreeItem) {
        for (const DeviceIndex &index: modTreeItem.keys()) {
            if(modTreeItem[index]->checkState(tree_tdc_name) != state) {
                modTreeItem[index]->setCheckState(tree_tdc_name, state);
                modified = true;
            }
        }
    } else {
        QTreeWidgetItem *p = item->parent();
        bool found=false;
        for (const DeviceIndex &index: modTreeItem.keys()){
            if(item==modTreeItem[index]){
                // Module item
                found = true;
                DeviceSetup &ds = config.getConfig(index);
                for (int ch: ds.chTdcEn.keys()) {
                    if(ds.chTdcEn[ch] != checked) {
                        ds.chTdcEn[ch] = checked;
                        modified = true;
                    }
                }
                if(modified) {
                    ui->treeWidgetChCfg->blockSignals(true);
                    for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
                        item->child(childIndex)->setCheckState(tree_tdc_name, state);
                    }
                    ui->treeWidgetChCfg->blockSignals(false);
                }
            } else if(p==modTreeItem[index]) {
                // Channel item
                found = true;
                bool ok;
                int modCh = item->text(tree_tdc_name).toInt(&ok);
                DeviceSetup &ds = config.getConfig(index);
                Q_ASSERT(ok);
                Q_ASSERT(modCh<=ds.getNCh());
                bool &chState = ds.chTdcEn[modCh];
                if(chState != checked) {
                    chState = checked;
                    modified = true;
                }
            }
            if(modified)
                updateTreeItemStates();
            if(found)
                break;
        }
        Q_ASSERT(found);
    }

    if(modified) {
        updateGuiVal();
        configUpdated();
    }
}

void MainWindow::read_config()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.beginGroup("splitter_vertical");
    ui->splitter_vertical->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_vertical->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.beginGroup("splitter_bottom");
    ui->splitter_bottom->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_bottom->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.endGroup();

    config.read_config();

    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
        .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    config.curDev = DeviceIndex();
     for(auto cfg : config.knownSetups.values()) {
         if(cfg.enabled) {
             config.curDev = cfg.index;
             break;
         }
     }

    QVector<DeviceDescription> devicesAll;
    QVector<DeviceDescription> devicesEnabled;

    for (const DeviceSetup &ds: config.knownSetups.values()) {
        DeviceDescription dd;
        dd.device_id = ds.index.getDevId();
        dd.serial_id = ds.index.getSerial();
        dd.enabled = ds.enabled;
        devicesAll.append(dd);
        if(dd.enabled) {
            connectToDev(ds.index);
            devicesEnabled.push_back(dd);
        }
    }
    updateDeviceOrder();

    discoverDialog->setSelected(devicesAll);
//    emit deviceListUpdated(devicesEnabled);
    (config.global_mode ? ui->radio_button_global : ui->radio_button_separate)
            ->setChecked(true);
    ui->actionTrace_dev_info_changes->setChecked(config.trace_device_info);
    ui->actionShow_power_current->setChecked(config.show_power_current);

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.tdc2-ctrl." + shortHostName + "." + config.program_index);
}

void MainWindow::writeSettings(bool force)
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.beginGroup("splitter_vertical");
    settings.setValue("geometry", ui->splitter_vertical->saveGeometry());
    settings.setValue("state", ui->splitter_vertical->saveState());
    settings.endGroup();
    settings.beginGroup("splitter_bottom");
    settings.setValue("geometry", ui->splitter_bottom->saveGeometry());
    settings.setValue("state", ui->splitter_bottom->saveState());
    settings.endGroup();
    settings.endGroup();

    if(force) {
        config.write_config();
        writeConfigTimer->stop();
    } else {
        writeConfigTimer->start();
    }
}

void MainWindow::initTable()
{
    ++blockControl;
    QTableWidget * const tableDev = ui->tableWidgetDevice;
    tableDev->setSortingEnabled(true);
    tableDev->setColumnCount(COL_DEV_TOTAL);

    tableDev->setHorizontalHeaderItem(COL_DEV_SERIAL, new QTableWidgetItem("Serial"));
    tableDev->setHorizontalHeaderItem(COL_DEV_SLOT, new QTableWidgetItem("Slot"));
    tableDev->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));

    tableDev->setHorizontalHeaderItem(COL_DEV_PLL_UNLOCK, new QTableWidgetItem("PLL unlock"));
    tableDev->setHorizontalHeaderItem(COL_DEV_PLL_STATUS, new QTableWidgetItem("PLL st."));

    tableDev->setHorizontalHeaderItem(COL_DEV_TEMP, new QTableWidgetItem("temp"));
    tableDev->setHorizontalHeaderItem(COL_DEV_POWER, new QTableWidgetItem("power"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TIME, new QTableWidgetItem("time"));
    tableDev->setHorizontalHeaderItem(COL_DEV_RUN_TIME, new QTableWidgetItem("Run time"));
    tableDev->setHorizontalHeaderItem(COL_DEV_CUR_EV_NUM, new QTableWidgetItem("Cur evNum"));
    tableDev->setHorizontalHeaderItem(COL_DEV_UNBLOCK, new QTableWidgetItem("Accepted"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BLOCK, new QTableWidgetItem("Blocked"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STATUS, new QTableWidgetItem("Dev. St."));
    tableDev->setHorizontalHeaderItem(COL_HPTDC_STATUS, new QTableWidgetItem("HPTDC St."));
    tableDev->setHorizontalHeaderItem(COL_BUSY_TIME, new QTableWidgetItem("Busy time"));
    tableDev->setHorizontalHeaderItem(COL_BUSY_STATUS, new QTableWidgetItem("Busy st."));
    tableDev->setHorizontalHeaderItem(COL_TRIG_QUEUE, new QTableWidgetItem("Trig queue"));
    tableDev->setHorizontalHeaderItem(COL_CLOCK_STATUS, new QTableWidgetItem("Clock"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STREAM, new QTableWidgetItem("Data rate"));
    tableDev->setHorizontalHeaderItem(COL_ERR_CNT, new QTableWidgetItem("Error Cnt"));
    tableDev->setRowCount(1);
    initTableItemAtRow(tableDev, 0);
    tableDev->verticalHeader()->hide();
//    tableDev->hideColumn(COL_DEV_SERIAL);
    tableDev->hideColumn(COL_DEV_TIME);
    tableDev->hideColumn(COL_DEV_MS_ERR);
    tableDev->hideColumn(COL_DEV_PLL_UNLOCK);
    tableDev->hideColumn(COL_DEV_PLL_STATUS);
    tableDev->hideColumn(COL_ERR_CNT);
    tableDev->hideColumn(COL_DEV_STATUS);
    if(!config.debug_mode) {
        tableDev->hideColumn(COL_DEV_POWER);
        tableDev->hideColumn(COL_DEV_CUR_EV_NUM);
        tableDev->hideColumn(COL_BUSY_STATUS);
        tableDev->hideColumn(COL_TRIG_QUEUE);
    }

    tableDev->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    tableDev->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(tableDev);

    connect(tableDev->horizontalHeader(), &QHeaderView::sectionClicked, this, &MainWindow::update_order_index);

    tableDev->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
    tableDev->verticalScrollBar()->setSingleStep(tableDev->rowHeight(0)/3);
    --blockControl;
}

void MainWindow::initTree()
{
    ++blockControl;

    QTreeWidget *tree = ui->treeWidgetChCfg;
    tree->setColumnCount(tree_tdc_num_cols);

    QStringList headerLabels;
    headerLabels << "name"
                 << "ch";
//    tree->hideColumn(tree_tdc_ch);
    tree->setHeaderLabels(headerLabels);

    rootTreeItem->setCheckState(tree_tdc_name, Qt::PartiallyChecked);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    tree->resizeColumnToContents(tree_tdc_name);
    --blockControl;

    tree->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(tree, &QTreeWidget::customContextMenuRequested,
            this, &MainWindow::customContextMenuRequested);
}

void MainWindow::initShortcut()
{
    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &QWidget::show);
    connect(this, &MainWindow::addLog, logDialog, &LogDialog::addLog);

    if(adjustDialog) {
        QShortcut *shortcut_adjust = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_A), this);
        connect(shortcut_adjust, &QShortcut::activated, adjustDialog, &QWidget::show);
        connect(shortcut_adjust, &QShortcut::activated, adjustDialog, &AdjustDialog::activate_dialog);
    }

    QShortcut *shortcut_change_device = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_F), this);
    connect(shortcut_change_device, &QShortcut::activated, this, &MainWindow::on_pushButtonChangeDevice_clicked);
}

void MainWindow::updateModelList()
{
    QTreeWidget *tree = ui->treeWidgetChCfg;

    tree->blockSignals(true);
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        rootTreeItem->removeChild(i);
        delete i;
    }
    modTreeItem.clear();

    int modChOffest = 0;
    int devCnt = 0;
    for (const DeviceIndex &index: devGlob.keys()) {
        QStringList list;
        QTreeWidgetItem * const modItem = new QTreeWidgetItem(rootTreeItem, list);
        modTreeItem[index] = modItem;
        modItem->setText(tree_tdc_name, getDeviceNameForTree(index, devCnt++));
        modItem->setCheckState(tree_tdc_name, Qt::PartiallyChecked);
        modItem->setExpanded(false);

        const DeviceSetup &ds = config.getConfig(index);
        const int modNch = ds.getNCh();
        for(int ch=0; ch<modNch; ++ch){
            QStringList list;
            list << QString::number(ch)
                 << QString::number(modChOffest+ch);
            QTreeWidgetItem * const chItem = new QTreeWidgetItem(modItem, list);
            Qt::CheckState st = ds.chTdcEn[ch] ? Qt::Checked : Qt::Unchecked;
            chItem->setCheckState(tree_tdc_name, st);
        }

        modChOffest += modNch;
    }
    tree->blockSignals(false);
    updateTreeSize(tree);
    updateTreeItemStates();
}

void MainWindow::updateTreeItemStates()
{
    ui->treeWidgetChCfg->blockSignals(true);
    Qt::CheckState allSt = rootTreeItem->checkState(tree_tdc_name);
    Qt::CheckState st = Qt::PartiallyChecked;
    bool firstMod=true;
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        for(int chIndex=0; chIndex<i->childCount(); ++chIndex){
            Qt::CheckState childState = i->child(chIndex)->checkState(tree_tdc_name);
            if(chIndex==0){
                st = childState;
                continue;
            }
            if(st!=childState){
                st = Qt::PartiallyChecked;
                break;
            }
        }
        i->setCheckState(tree_tdc_name, st);
        if(firstMod) {
            allSt = st;
            firstMod = false;
        } else if(allSt != st)
            allSt = Qt::PartiallyChecked;
    }

    rootTreeItem->setCheckState(tree_tdc_name, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void MainWindow::configUpdated(DeviceIndex singleIndex)
{
    if(blockControl) return;
    emit set_run_state(config.run);
    updateRemoteStatus();
    writeSettings();
    for (const DeviceIndex &index: devGlob.keys()) {
        if(singleIndex.isNull() || index == singleIndex){
            emit setupChanged(config.getConfig(index));
        }
    }
    // restartProcess();
}

void MainWindow::updateGuiVal()
{
    ++blockControl;

    DeviceSetup ds;
    if(config.global_mode) {
        if(config.curDev.isNull()) {
            --blockControl;
            return;
        }
        ds = config.getCurrentConfig();
    } else {
        ds = config.getConfig(config.curDev);
    }

    ui->checkBoxTrigTimer->setChecked(ds.trigsetup.bit.timer_en);
    ui->spinBoxTimeTrig->setEnabled(ds.trigsetup.bit.timer_en);
    int rate = static_cast<int>(125e6/config.getCurrentConfig().trigPeriod/3);
    ui->spinBoxTimeTrig->setValue(rate);
    updateRealRate();

    ui->checkBoxTrigTtl->setChecked(ds.trigsetup.bit.ttl_en);
    ui->checkBoxTrigTtc->setChecked(ds.trigsetup.bit.ttc_en);
    ui->checkBoxTrigVxs->setChecked(ds.trigsetup.bit.vxs_en);

    ui->spinBoxLatency->setValue(static_cast<int>(ds.latency));
    ui->spinBoxMatchWin->setMaximum(static_cast<int>(ds.latency));
    ui->spinBoxMatchWin->setValue(static_cast<int>(ds.matchWin));
    int edgeData = 0;
    if(ds.leadEdge) edgeData |= 1;
    if(ds.trailEdge) edgeData |= 2;
    ui->comboBoxEdges->setCurrentIndex(ui->comboBoxEdges->findData(edgeData));

    block_widgets_signal(true);
    ui->actionAdd_HPTDC_event_header->setChecked(ds.hptdcHdr);
    ui->actionAdd_HPTDC_event_tralier->setChecked(ds.hptdcTrl);
    ui->groupBoxStat->setChecked(ds.stat_readout);
    ui->spinBoxStatPeriod->setValue(ds.stat_period);
    block_widgets_signal(false);


    ui->comboBoxPreset->setCurrentText(config.default_setup.stat_preset.label);
    --blockControl;
    updateControlState();
}

void MainWindow::updateControlState()
{
    bool enableControl = (blockControl==0) && (!on_remote || config.debug_mode);
    const bool isRun = config.isRun();
    ui->pushButtonChangeDevice->setEnabled(enableControl && !isRun);

    enableControl &= !config.curDev.isNull();
    ui->frame_configuration->setEnabled(enableControl);
    ui->treeWidgetChCfg->setEnabled(enableControl);
    ui->pushButtonStart->setEnabled(!on_remote && enableControl && !isRun);
    ui->pushButtonStop ->setEnabled(!on_remote && enableControl &&  isRun);
}

void MainWindow::disconnectFromDev(const DeviceIndex &index)
{
    auto dgIt = devGlob.find(index);
    if(dgIt != devGlob.end()) {
        auto &dg = dgIt.value();
        sendLog(dg->dev->getDeviceInfoStr().prepend("Disconnect from:"));
        emit deleteDevice(index);
        if(!threadQuiting(dg->devThread, 2000)){
            qWarning() << "Can't stop dev. h/w thread for "<< index.getIdent();
        }
        if(!threadQuiting(dg->msThread, 2000)){
            qWarning() << "Can't stop MStream thread for "<< index.getIdent();
        }
        ui->m_stream_output->remove_device(index);

        /// TODO: Check for memory leak
        devGlob.erase(dgIt);
    } else {
        qCritical() << "disconnectFromDev: no dev in devGlob map";
    }

    const int row = getRowByDevice(index);
    if(row != -1) {
        QTableWidget *tableDev = ui->tableWidgetDevice;
        if(tableDev->rowCount() == 1) {
            Q_ASSERT(row==0);
            initTableItemAtRow(tableDev, 0);
            config.curDev = DeviceIndex();
        } else {
            tableDev->removeRow(row);
            if(index == config.curDev)
                tableDev->item(0, COL_DEV_SERIAL)->setSelected(true);
        }
        updateTableSize(tableDev);
//        ui->tableWidgetDevice->sortByColumn(0, Qt::AscendingOrder);
    } else {
        qCritical() << "disconnectFromDev: no row in table";
    }

    updateControlState();
    update_pll_unlock_visible();
    updateModelList();
    table_warn_updated();
}

void MainWindow::connectToDev(const DeviceIndex &index)
{
//  qInfo() << "==== Begin of connectToDev:" << index.getSerialStr() << " ====\n";
    Q_ASSERT(config.knownSetups.contains(index));
    auto tableDev = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(tableDev);
    if(devGlob.contains(index))
        disconnectFromDev(index);
    int row = -1;
    if(tableDev->rowCount() == 1 && tableDev->item(0, COL_DEV_SERIAL)->data(Qt::UserRole).isNull()) {
        row = 0;
    }
    if(row == -1) {
        row = tableDev->rowCount();
        tableDev->setRowCount(row+1);
        initTableItemAtRow(tableDev, row);
    }

    devGlob.insert(index, new DeviceGlob(index, config.debug_mode));
    auto &dg = devGlob[index];
    connect(this, &MainWindow::deleteDevice, dg, &DeviceGlob::deleteDevice);
    dg->warn_flags.resize(COL_DEV_TOTAL);
    for(int col=0; col<COL_DEV_TOTAL; ++col)
        dg->warn_flags[col] = false;

    TDCDevice *dev = dg->dev;
    connect(evNumChecker, &EvNumChecker::evNumRequest, dev, &TDCDevice::evNumRequest);
    connect(dev, &TDCDevice::evNumResponse, evNumChecker, &EvNumChecker::gotDevEvNum);

    bool blockSignals = tableDev->blockSignals(true);
    tableDev->verticalHeaderItem(row)->setText(index.getSerialStr());
    tableDev->item(row, COL_DEV_SERIAL)->setText(index.getSerialStr());
    QList<QVariant> index_data;
    index_data << index.getDevId() << index.getSerial();
    tableDev->verticalHeaderItem(row)->setData(Qt::UserRole, index_data);
    tableDev->item(row, COL_DEV_SERIAL)->setData(Qt::UserRole, index_data);
    tableDev->item(row, COL_DEV_STATUS)->setText("connecting...");
    tableDev->item(row, COL_DEV_STREAM)->setText("");
    tableDev->item(row, COL_ERR_CNT)->setText("");
    updateTableSize(tableDev);

    if(config.curDev.isNull()) {
        config.curDev = index;
    } else if(index == config.curDev) {
        tableDev->item(0, COL_DEV_SERIAL)->setSelected(true);
    }

    updateGuiVal();

    try{
        ui->m_stream_output->add_device(index);

        connect(this, &MainWindow::closeAll, dev, &QObject::deleteLater);
        connect(this, &MainWindow::deleteDevice, dev, &QtMregDevice::deleteHardware);

        connect(this, &MainWindow::setupChanged, dev, &TDCDevice::setSetup);
        connect(this, &MainWindow::setupChangedForce, dev, &TDCDevice::writeSetupForce);
        connect(ui->actionForce_write_setup, &QAction::triggered, dev, &TDCDevice::writeSetupForce);
        connect(this, &MainWindow::softReset, dev, &TDCDevice::softReset);
        connect(this, &MainWindow::blockDeviceRun, dev, &TDCDevice::blockDeviceRun);
        connect(this, &MainWindow::reset_error_counters, dev, &TDCDevice::reset_error_counters);
        connect(this, &MainWindow::setDevAddr, dev, &TDCDevice::deviceUpdated);
        connect(discoverDialog, &DiscoverDialog::deviceDiscovered, dev, &TDCDevice::deviceUpdated);
        connect(dev, &QtMregDevice::setStatus, this, &MainWindow::setDeviceStatus);
        connect(dev, &TDCDevice::deviceInfoUpdated, this, &MainWindow::deviceInfoUpdated);
        connect(dev, &TDCDevice::deviceSetupWritten, this, &MainWindow::deviceSetupWritten);
        connect(dev, &TDCDevice::statusUpdated, this, &MainWindow::deviceStatusUpdated);
        connect(dev, &QtMregDevice::deviceConnected, this, &MainWindow::deviceConnected);
        connect(dev, &QtMregDevice::deviceDisconnected, this, &MainWindow::deviceDisconnected);

        dg->startThreads();

        // Connecting
        QHostAddress hostAddr;
        quint32 ip_addr = knownDevDescr.value(index).ip_addr;
        if(!ip_addr)
            emit setDevAddr(knownDevDescr.value(index));
        quint32 slot = knownDevDescr.value(index).chassis_slot;
        if(slot)
            tableDev->item(row, COL_DEV_SLOT)->setText(QString::number(slot));
        emit setupChanged(config.getConfig(index));
    } catch (std::exception &e){
        sendLog(QString("Connect to device %1").arg(e.what()));
    }
    tableDev->blockSignals(blockSignals);

//    qInfo() << "==== End of connectToDev:" << index.getSerialStr() << " ====\n";
    updateModelList();

    QVector<quint64> serials;
    for(auto ds: config.knownSetups.values()){
        if(ds.enabled)
            serials.append(ds.index.getSerial());
    }
    evNumChecker->set_dev_serials(serials);
}

void MainWindow::sendLog(QString l)
{
    emit addLog(l);
}

void MainWindow::logDevInfo(QTableWidgetItem *item, bool force)
{
    if(!config.trace_device_info) return;

    const int col = item->column();
    const int row = item->row();

    if(!force) {
        if(col == COL_DEV_RUN_TIME || col == COL_DEV_CUR_EV_NUM || col == COL_DEV_UNBLOCK || col == COL_DEV_BLOCK ||
           col == COL_DEV_STREAM || col == COL_BUSY_STATUS || col == COL_TRIG_QUEUE ||
           col == COL_BUSY_TIME ||  col == COL_DEV_TEMP || col == COL_DEV_POWER)
            return;
    }
// Columns to log:
//    COL_DEV_SERIAL; COL_DEV_SLOT; COL_DEV_FW; COL_DEV_MCU_VER; COL_DEV_PLL_UNLOCK; COL_DEV_PLL_STATUS; COL_DEV_TIME;
//    COL_DEV_MS_ERR;
//    COL_DEV_STATUS; COL_HPTDC_STATUS;
//    COL_CLOCK_STATUS; COL_ERR_CNT;

    auto newTxt = item->text();
    DeviceIndex index = getDevIndexByRow(row);
    auto devGlobIt = devGlob.find(index);
    if(devGlobIt == devGlob.end())
        return;

    QString &txt = devGlobIt.value()->tbl_info_log[col];
    if(newTxt == txt)
        return;

    QString colStr = ui->tableWidgetDevice->horizontalHeaderItem(col)->text();
    if(colStr.isEmpty()) colStr = QString::number(col);
    QString runTime;
    const auto &devSt = devGlob[index]->pendingDevStatus;
    if(!devSt.getIndex().isNull() && devSt.connected)
        runTime = QString(" [RunTime:%2]").arg(devSt.statistic.getRunTime());
    qInfo().noquote() << QDateTime::currentDateTime().toString("hh:mm:ss")
                      << QString("[%1]%2 Dev. info change col='%3': '%4' -> '%5'")
                         .arg(index.getSerialStr(), runTime, colStr, txt, newTxt);
    txt = newTxt;
}

void MainWindow::updateRealRate()
{
    if(config.curDev.isNull()) {
        ui->labelRealRate->setText("-");
    } else {
        const auto &ds = config.getCurrentConfig();
        int realRateHz = static_cast<int>(125e6/ds.trigPeriod/3);
        ui->labelRealRate->setText(QString("%1 Hz").arg(realRateHz));
        ui->labelRealRate->setToolTip(QString("Trig period: %1*24ns = %2ns")
                                      .arg(ds.trigPeriod).arg(ds.trigPeriod*24));
    }
}

void MainWindow::update_pll_unlock_visible()
{
    const int col = COL_DEV_PLL_UNLOCK;
    auto table = ui->tableWidgetDevice;
    bool hidden = true;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->item(row, col)->data(Qt::UserRole).toInt()) {
            hidden = false;
            break;
        }
    }
    table->setColumnHidden(col, hidden);
}

void MainWindow::update_order_index(int column)
{
    for(DeviceIndex index1 : config.knownSetups.keys()){
        for(int row = 0; row < ui->tableWidgetDevice->rowCount(); ++row){
            DeviceIndex index2 = getDevIndexByRow(row);
            if(index1 == index2){
                config.knownSetups[index1].order_index = row;
            }
        }
    }
    update_tree_order();
}

void MainWindow::updateDeviceOrder(bool from_dialog)
{
    auto table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);
    auto sortLocker = TableSortLocker(table);

    if(!from_dialog)
        update_order_index(1);

    QMap<DeviceIndex, QList<QTableWidgetItem*> > upperItems;
    int row;
    QList<DeviceSetup> devices;

    //deattach
    for(row=0; row<table->rowCount(); ++row) {
        DeviceIndex i = getDevIndexByRow(row);
        if(i.isNull()) break;
        upperItems[i] = OrderDialog::takeRow(table, row);
        Q_ASSERT(config.knownSetups.contains(i));
        devices.append(config.knownSetups[i]);
        rootTreeItem->takeChild(0);
    }

    std::sort(devices.begin(),devices.end(), DeviceSetup::compare);

    //attach
    row=0;
    bool blockSignals=table->blockSignals(true);
    for(DeviceSetup ds : devices) {
        DeviceIndex index = ds.index;
        if(!devGlob.contains(index)) {
            Q_ASSERT(!ds.enabled); continue;
        }
        OrderDialog::setRow(table, row, upperItems[index]);
        Q_ASSERT(modTreeItem.contains(index));
        auto treeItem = modTreeItem[index];
        treeItem->setText(tree_tdc_name, getDeviceNameForTree(index, row));
        rootTreeItem->insertChild(row, treeItem);
        ++row;
    }
    table->blockSignals(blockSignals);
}

void MainWindow::update_tree_order()
{
    if(config.knownSetups.isEmpty())
        return;
    QTreeWidgetItem *item_all = ui->treeWidgetChCfg->topLevelItem(0);

    QMap<int, QTreeWidgetItem *> order_treewitem;
    for(DeviceIndex index : config.knownSetups.keys()){
        const DeviceSetup &setup = config.knownSetups[index];
        if(!setup.enabled) continue;
        auto treeItem = modTreeItem.value(index);
        Q_ASSERT(treeItem);
        int tree_index = item_all->indexOfChild(treeItem);
        Q_ASSERT(tree_index != -1);
        int order = config.knownSetups[index].order_index;
        order_treewitem[order] = treeItem;
    }

    for(int order : order_treewitem.keys()) {
        item_all->removeChild(order_treewitem[order]);
    }

    QList<QTreeWidgetItem *> list_to_append;
    for(int key : order_treewitem.keys()) {
        list_to_append.append(order_treewitem[key]);
    }

    item_all->addChildren(list_to_append);
    ui->treeWidgetChCfg->addTopLevelItem(item_all);

    for(int i=0; i < item_all->childCount(); ++i) {
        QString str = item_all->child(i)->text(0);
        QString index_str = str.split(" ").at(3);
        for(int k : order_treewitem.keys()) {
            if(order_treewitem[k] == item_all->child(i)){
                str.remove(0,1);
                str.prepend(QString::number(k+1));
                item_all->child(i)->setText(0, str);
            }
        }
    }
}

void MainWindow::updateTableSize(QTableWidget * const table, bool widthOnly)
{
    widthOnly = true;
    table->resizeRowsToContents();
    table->resizeColumnsToContents();
//    table->setColumnWidth(COL_RATE, 500);

    int width = 0;
    width += table->verticalScrollBar()->width();
    if(table->verticalHeader()->isVisible())
        width += table->verticalHeader()->width();
    bool test=false;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
        if(test)
            qInfo() << table->horizontalHeaderItem(col)->text() << table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    width = qMin(width, 1000);
    height = qMin(height, 200);
    if(widthOnly) {
        table->setMinimumWidth(width);
    } else {
        table->setMinimumWidth(width);
        table->setMinimumHeight(height);
    }
    table->updateGeometry();
    table->horizontalHeader()->setStretchLastSection(true);
}

void MainWindow::updateTreeSize(QTreeWidget * const tree)
{
    QString colInfo;
    int width = 0;
    colInfo += "fix:5;";
    for(int col=0; col<tree->columnCount(); ++col) {
        if(tree->isColumnHidden(col))            continue;
        tree->resizeColumnToContents(col);
        colInfo += QString("%1:%2;").arg(tree->headerItem()->text(col)).arg(tree->columnWidth(col));
        width += tree->columnWidth(col);
    }

//    if(width<20)
//        width = 20;

//    width += tree->verticalScrollBar()->width();
//    colInfo += QString("verticalScrollBar:%2;").arg(tree->verticalScrollBar()->width());
//    qInfo() << colInfo;
//    tree->setMinimumWidth(width);
//    tree->updateGeometry();
}

DeviceIndex MainWindow::getDevIndexByRow(int row) const
{
    auto index_data = ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->data(Qt::UserRole).toList();
    if(!index_data.isEmpty()) {
        bool ok=false,ok2=false;
        const quint32 id = index_data.first().toUInt(&ok2);
        const quint64 ser = index_data.last().toULongLong(&ok);
        if(ok && ok2) {
            return DeviceIndex(static_cast<quint16>(id), ser);
        }
    }

    return DeviceIndex();
}

int MainWindow::getRowByDevice(const DeviceIndex &i) const
{
    for(int row=ui->tableWidgetDevice->rowCount()-1; row>=0; --row) {
        if(getDevIndexByRow(row) == i) return row;
    }
    return -1;
}

void MainWindow::param_updated()
{
    for (DeviceIndex i: devGlob.keys()) {
        if(config.global_mode || i == config.curDev) {
            config.getConfig(i).copy_config(config.default_setup);
        }
    }
    if(!blockControl) {
        writeSettings();
        configUpdated(config.global_mode ? DeviceIndex() : config.curDev);
    }
}

QString MainWindow::getDeviceNameForTree(DeviceIndex index, int devCnt) const
{
    quint16 slot_num = knownDevDescr.value(index).chassis_slot;
    return QString("%1: %2 (s/n: %3%4)").arg(++devCnt)
            .arg(getDeviceTypeName(index.getDevId()))
            .arg(index.getSerialStr())
            .arg(slot_num ? QString(" slot:%1").arg(slot_num):"");
    auto modItem = modTreeItem.value(index);
    return QString("%1: %2 (s/n: %3%4)").arg(rootTreeItem->indexOfChild(modItem))
            .arg(getDeviceTypeName(index.getDevId()))
            .arg(index.getSerialStr())
            .arg(slot_num ? QString(" slot:%1").arg(slot_num):"");
}

void MainWindow::addMStreamDump(const DeviceIndex &index)
{
    auto dgIt = devGlob.find(index);
    if(dgIt == devGlob.end())
        return;

    auto &dg = dgIt.value();
    const auto &tdcStatus = dg->pendingDevStatus;

    MStreamDumpParameters dump_parameters;
    dump_parameters.deviceIndex = index;
    const auto dd = discoverDialog->getDescriptionByIndex(index);
    if(dd.getIndex() == index && dd.ip_addr){
        dump_parameters.deviceAddress = QHostAddress(dd.ip_addr);
    }
    Q_ASSERT(dd.ip_addr);
    dump_parameters.bigFragId = true;
    dump_parameters.hwBufSize = tdcStatus.mstreamHwBufSize;
    dump_parameters.ackSizeLimit = tdcStatus.mstreamMultiAck ? tdcStatus.mstreamHwBufSize : 1;
    dump_parameters.preRunInit();

    MStreamDump *dump = dg->mstream = new MStreamDump(dump_parameters, false);
    connect(dump, &MStreamDump::closing, pnpServer, &PNPServer::closeProgram);
    connect(dump, &MStreamDump::updateProgramDescription,
            pnpServer, &PNPServer::updateProgramDescription);
    connect(dump, &MStreamDump::stat_updated,
            ui->m_stream_output, &StatisticOutput::stat_updated);
    connect(dump, &MStreamDump::stat_updated, this, &MainWindow::mstream_status_updated);
    connect(discoverDialog, &DiscoverDialog::deviceDiscovered, dump, &MStreamDump::receivedMldpDescr);
    connect(dump, &MStreamDump::stat_from_receiver_updated, ui->m_stream_output, &StatisticOutput::stat_from_receiver_updated);
    connect(this, &MainWindow::msdump_reset, dump, &MStreamDump::msdump_reset);
    connect(this, &MainWindow::mstream_send_connect, dump, &MStreamDump::send_connect_to_dev);

    QThread *thread = dg->msThread = new QThread(this);
    thread->setObjectName("MSD_"+index.getSerialStr());
    connect(dump, &MStreamDump::destroyed, thread, &QThread::quit);
    connect(this, &MainWindow::deleteDevice, dump, &MStreamDump::delete_dump);
    dump->moveToThread(thread);
    thread->start();
}

void MainWindow::resetMStream(const DeviceIndex &index, bool doStatClear)
{
    auto dgIt = devGlob.find(index);
    if(dgIt == devGlob.end())
        return;
    auto &dg = dgIt.value();
    auto &lastStatus = dg->pendingDevStatus;
    if(!lastStatus.connected)
        return;

    emit msdump_reset(index, true, lastStatus.mstreamHwBufSize,
                     lastStatus.mstreamMultiAck ? lastStatus.mstreamHwBufSize : 1, QString(), doStatClear);
}

void MainWindow::adjust_setup_changed(Adjust_setup adjust_set)
{
    config.default_setup.adjust_set = adjust_set;
    // use for all  devices regardless of global_mode
    for(const DeviceIndex &index: config.knownSetups.keys()){
        config.getConfig(index).adjust_set = adjust_set;
    }
    configUpdated();
}

void MainWindow::mstream_status_updated(const MStreamStat &status)
{
//    ui->m_stream_output->stat_updated(status);

    auto table = ui->tableWidgetDevice;
    auto dg_it = devGlob.begin();

    DeviceIndex devIndex;
    while(dg_it != devGlob.end()) {
        if(dg_it.value()->mstream == sender()) {
            break;
        }
        ++dg_it;
    }
    if(dg_it == devGlob.end())
        return;
    auto &prevStatus = dg_it.value()->msStat;
    auto sortLocker = TableSortLocker(table);
    int devRow = getRowByDevice(dg_it.key());
    auto updateLocker = TableUpdateLocker(table);
    if(prevStatus.timestamp.isValid()) {
        qint64 milisecs = 0;
        milisecs = prevStatus.timestamp.msecsTo(status.timestamp);
        double receive_data_rate = 0;
        if(milisecs != 0) {
            receive_data_rate = 1.*(status.recvDataSize - prevStatus.recvDataSize)/milisecs;
        }
        table->item(devRow, COL_DEV_STREAM)->setText(QString("%L1 kB/s").arg(receive_data_rate,0,'f',2));
    }
    prevStatus = status;
    table->item(devRow, COL_DEV_MS_ERR)->setText(QString::number(status.badFormatPackage));
    auto &warn = dg_it.value()->warn_flags[COL_DEV_TEMP];
    if(warn != (status.badFormatPackage != 0)) {
        warn = status.badFormatPackage != 0;
        table_warn_updated(dg_it.key(), COL_DEV_MS_ERR);
    }
//    bool hide_col = true;
//    for(int row=0; row<table->rowCount(); ++row) {
//        if(!table->item(row,COL_DEV_MS_ERR)->text().isEmpty()) {
//            hide_col = false;
//            break;
//        }
//    }

//    if(hide_col != table->isColumnHidden(COL_DEV_MS_ERR))
//        table->setColumnHidden(COL_DEV_MS_ERR, hide_col);
    updateTableSize(table);
}

void MainWindow::on_actionAdjust_HPTDC_triggered()
{
    if(!adjustDialog) return;
    adjustDialog->activate_dialog();
    adjustDialog->show();
}

void MainWindow::on_actionTrace_dev_info_changes_toggled(bool checked)
{
    config.trace_device_info = checked;
    if(!config.trace_device_info) {
        auto devIt = devGlob.begin();
        while(devIt != devGlob.end()) {
            devIt.value()->tbl_info_log.clear();
            ++devIt;
        }
    }
}

void MainWindow::on_actionShow_power_current_toggled(bool checked)
{
    config.show_power_current = checked;
    if(!checked) {
        ui->actionReset_power_status->trigger();
    }
    writeSettings();
}

void MainWindow::on_actionReset_power_status_triggered()
{
    auto dg_it = devGlob.begin();
    while (dg_it != devGlob.end()) {
        dg_it.value()->warn_flags[COL_DEV_POWER] = false;
        ++dg_it;
    }

    table_warn_updated();
}

void MainWindow::on_actionReset_error_counters_triggered()
{
    // todo make slot for column auto-hide
//    ui->actionReset_error_counters->setVisible(false);
    emit reset_error_counters();

    auto dg_it = devGlob.begin();
    while (dg_it != devGlob.end()) {
        dg_it.value()->warn_flags[COL_ERR_CNT] = false;
        ++dg_it;
    }

    table_warn_updated();
}

void MainWindow::on_actionAdd_HPTDC_event_header_toggled(bool checked)
{
    config.default_setup.hptdcHdr = checked;
    param_updated();
}

void MainWindow::on_actionAdd_HPTDC_event_tralier_toggled(bool checked)
{
    config.default_setup.hptdcTrl = checked;
    param_updated();
}

void MainWindow::addLogFromSender(QtMsgType severity, QString log, QString thrName)
{
    Q_UNUSED(severity)
    Q_UNUSED(log)
    Q_UNUSED(thrName)
//    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

//    Qt::GlobalColor color;
//    switch (severity) {
//    case QtCriticalMsg:
//        color = Qt::magenta;
//        break;
//    case QtWarningMsg:
//        color = Qt::red;
//        break;
//    case QtInfoMsg:
//        color = Qt::blue;
//        break;
//    case QtDebugMsg:
//#ifndef QT_DEBUG
//        return;
//#endif
//    default:
//        color = Qt::black;
//    }
}

void MainWindow::delayed_write_config()
{
    config.write_config();
}

void MainWindow::on_actionConfiguration_window_toggled(bool show)
{
    config.show_cfg_widget = show;
    writeSettings();
    ui->widget_bottom->setVisible(show);
}

void MainWindow::block_widgets_signal(bool block)
{
    for(int i=0; i<widgets_to_block.size(); ++i)
        widgets_to_block[i]->blockSignals(block);
}

