#include "Config.h"
#include "mregdev/device_id.h"
#include "mongo/BaseConfig.h"

#include <QSettings>
#include <QDebug>

Config::Config(const TQDC2StartupOptions& startup_options) :
    program_type(startup_options.program_type),
    program_index(startup_options.program_index),
    configuration_name(startup_options.configuration_name),
    separateChEn(false),
    writeEnabled(false),
    dataFileIsRaw(true),
    fileTime(true),
    rawDataLimit(1)
{
    update_root();
    default_root = root;
}

void Config::read_config()
{
    curDev.clear();
    allCal.clear();

    root = BaseConfig::read_config(program_type, program_index, configuration_name);
    if(root.isEmpty()){
        BaseConfig::write_config(default_root);
    }

    separateChEn = root.get_data<bool>("separateChEn");
    writeEnabled = root.get_data<bool>("writeEnabled");
    dataFileIsRaw = root.get_data<bool>("dataFileIsRaw");
    fileDir = root.get_data<QString>("fileDir");
    filePrefix = root.get_data<QString>("filePrefix");
    fileTime = root.get_data<bool>("fileTime");
    fileSuffix= root.get_data<QString>("fileSuffix");
    rawDataLimit = root.get_data<uint>("rawDataLimit", rawDataLimit);
    zmq_config.root_to_variables(root.child("ZMQ"));

    for(auto child : root.children().values()){
        if(child.item_name.startsWith("current_device")){
            curDev = DeviceIndex(child.get_data<quint64>("device_id"),
                                 child.get_data<quint64>("device_serial"));
        }

        if(child.item_name == "known_setups"){
            for(RootConfig device_root : child.children().values()){
                DeviceSetup device_setup;
                QMap<int, int> thrOffset;
                bool ok = device_setup.root_to_variables(device_root, thrOffset);
                if(ok) {
                    knownSetups[device_setup.index] = device_setup;
                    allCal[device_setup.index].thrOffset = thrOffset;
                }
            }
        }
    }

    reloadCalibration();
}

void Config::write_config()
{
    update_root();
    BaseConfig::write_config(root);
}

void Config::update_root()
{
    root.remove_all_data();
    root.set_data("program_type", program_type);
    root.set_data("program_index", program_index);
    root.set_data("configuration_name", configuration_name);

    root.set_data("separateChEn", separateChEn);
    root.set_data("writeEnabled", writeEnabled);
    root.set_data("dataFileIsRaw", dataFileIsRaw);
    root.set_data("fileDir", fileDir);
    root.set_data("filePrefix", filePrefix);
    root.set_data("fileTime", fileTime);
    root.set_data("fileSuffix", fileSuffix);
    root.set_data("rawDataLimit", rawDataLimit);

    if(!curDev.isNull()){
        RootConfig root_cur_dev(QString("current_device_%1").arg(curDev.getSerialStr()));
        root_cur_dev.set_data("device_id", curDev.first);
        root_cur_dev.set_data("device_serial", curDev.second);
        root.append_child(root_cur_dev);
    }

    if(!knownSetups.isEmpty()){
        RootConfig  known_setup_root("known_setups");
        for(DeviceSetup ds : knownSetups.values()) {
            RootConfig root_device;
            if(ds.index.isNull())
                qFatal("DeviceSetup with empty device index detected on write_config");
            root_device = ds.variables_to_root(allCal.value(ds.index));
            known_setup_root.append_child(root_device);
        }

        root.append_child(known_setup_root);
    }

    RootConfig zmq_root = zmq_config.variables_to_root();
    root.append_child(zmq_root);
}

void Config::loadSettings()
{
    QSettings s;
    {
        s.beginGroup("Main");
        DeviceIndex index;
        bool ok;
        index.first = s.value("devId").toUInt(&ok);
        index.second = s.value("devSer").toULongLong(&ok);
        if(!index.isNull())
            curDev = index;
        separateChEn = s.value("separateChEn", separateChEn).toBool();
        writeEnabled = s.value("writeEnabled", writeEnabled).toBool();
        dataFileIsRaw = s.value("dataFileIsRaw", dataFileIsRaw).toBool();
        fileDir = s.value("fileDir").toString();
        filePrefix = s.value("filePrefix").toString();
        fileTime = s.value("fileTime", fileTime).toBool();
        fileSuffix = s.value("fileSuffix").toString();
        rawDataLimit = s.value("rawDataLimit", rawDataLimit).toUInt();
        s.endGroup();

        zmq_config.loadSettings(s);
    }
    for(QString key : s.childGroups()) {
        if(key.startsWith("DevConf_")) {
            s.beginGroup(key);

            DeviceSetup setup;
            if(setup.loadSettings(s))
                knownSetups[setup.index] = setup;

            s.endGroup();
        }
    }
    reloadCalibration();
}

void Config::saveSettings()
{
    QSettings s;
    {
        s.beginGroup("Main");
        s.setValue("devId", curDev.getDevId());
        s.setValue("devSer", curDev.getSerial());
        s.setValue("separateChEn", separateChEn);
        s.setValue("writeEnabled", writeEnabled);
        s.setValue("dataFileIsRaw", dataFileIsRaw);
        s.setValue("fileDir", fileDir);
        s.setValue("filePrefix", filePrefix);
        s.setValue("fileTime", fileTime);
        s.setValue("fileSuffix", fileSuffix);
        s.setValue("rawDataLimit", rawDataLimit);
        s.endGroup();

        zmq_config.saveSettings(s);
    }
    for (QString key : s.childGroups()) {
        if(key.startsWith("DevConf_"))
            s.remove(key);
    }

    int i=0;
    for (const DeviceSetup &setup : knownSetups.values()) {
        s.beginGroup(QString("DevConf_%1").arg(i++));

        setup.saveSettings(s, allCal.value(setup.index));

        s.endGroup();
    }
}


DeviceSetup &Config::getCurrentConfig()
{
    Q_ASSERT(!curDev.isNull());
//    Q_ASSERT(knownSetups.contains(curDev));
    if(!knownSetups.contains(curDev)) {
        DeviceSetup ds;
        ds.index = curDev;
        knownSetups[curDev] = ds;
    }
    return knownSetups[curDev];
}

DeviceCalibration &Config::getCurrentCalibration()
{
    Q_ASSERT(!curDev.isNull());
    return allCal[curDev];
}

bool Config::isRun()
{
    if(curDev.isNull())
        return false;
    const DeviceSetup &ds = getCurrentConfig();
    return ds.run;
}

void Config::updateInl(const DeviceIndex &index, const QMap<int, QVector<double> > &inlCorr, double temp)
{
    QSettings s;
    s.beginGroup(getInlSectionStr(index));
    s.setValue("temp", QString::number(temp, 'f', 1));
    s.setValue("data", QDateTime::currentDateTime().toString());
    for(int ch : inlCorr.keys()) {
        QList<QVariant> v;
        const QVector<double> &corrValues = inlCorr[ch];
        if(corrValues.size() != 1024)
            continue;

        for(int i=0; i<1024; i++)
            v.push_back(QString::number(corrValues[i], 'f', 2));
        s.setValue(QString::number(ch), v);
        allCal[index].inl[ch] = corrValues;
    }
    s.endGroup();
}

void Config::updateBaseCal(const DeviceIndex &index, const QMap<int, double> &baseline, const QMap<int, double> &baseline4, const QMap<int, int> &thr, double temp)
{
    QSettings s;
    s.beginGroup(getCalSectionStr(index));
    s.setValue("temp", QString::number(temp, 'f', 1));
    s.setValue("data", QDateTime::currentDateTime().toString());
    int maxNCh = -1;
    for(int ch : baseline.keys()) {
        s.setValue(getKeyStr_X1(ch), baseline[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    for(int ch : baseline4.keys()) {
        s.setValue(getKeyStr_X4(ch), baseline4[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    for(int ch : thr.keys()) {
        s.setValue(getKeyStr_ThrOffset(ch), thr[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    s.setValue("nch", maxNCh+1);
    s.endGroup();
    allCal[index].x1 = baseline;
    allCal[index].x4 = baseline4;
    allCal[index].thrOffset = thr;
}

void Config::clearCalibration(const DeviceIndex &index)
{
    allCal[index].clearBase();
}

void Config::reloadCalibration()
{
    CalConfig::reloadCalibration();

    QSettings s;
    CalConfig::reloadCalibration(s);
}


