#include "HistMngrWidget.h"
#include "ui_HistMngrWidget.h"

#include "DeviceData.h"

#include "qwt_plot.h"
#include "qwt_interval_data.h"
#include "qwt_plot_curve.h"
#include <qwt_scale_engine.h>

#include <QTimer>
#include <QSettings>
#include <QSemaphore>

HistMngrWidget::HistMngrWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::HistMngrWidget),
    refreshTimer(new QTimer(this))
{
    connect(refreshTimer, &QTimer::timeout, this, &HistMngrWidget::refreshCurrentPlotBypass);
    refreshTimer->setInterval(500);
    ui->setupUi(this);
}

void HistMngrWidget::refreshCurrentPlotBypass()
{
    refreshCurrentPlot();
}

HistMngrWidget::~HistMngrWidget()
{
    delete ui;
}

void HistMngrWidget::dataUpdated(const DeviceData &data, bool)
{
    histSem->release();
    if(!cfg.globalEnabled)
        return;
    if(inlDeviceIndex.isNull() || inlDeviceIndex != data.index) {
        if(inlDeviceIndex!=data.index) {
            // Clear previous INL-corr data
            for(int ch=0; ch<HistId::nch; ++ch) {
                if(HistId::chInl==-1 || !histTRoot.contains(HistId::chInl+ch))
                    continue;
                histTRoot[HistId::chInl+ch].reset();
            }
        }
        inlDeviceIndex = data.index;
    }
    int tdcHits=0,adcHits=0;
    const bool doTdcProcessing = cfg.doTdcProcessing();
    const bool doAdcProcessing = cfg.doAdcProcessing();
    for (const CompleteEvent &ev : data.events) {
        tdcHits=0;
        adcHits=0;
        double tdcPrevTime = -1, adcPrevTime = -1;
        if(doTdcProcessing)
            dataProcessingTdc(ev, tdcHits, tdcPrevTime);
        if(doAdcProcessing)
            dataProcessingAdc(ev, adcHits, adcPrevTime);

        if(histTRoot.contains(HistId::tdcAdcDt) && tdcPrevTime!=-1 && adcPrevTime!=-1 && cfg.histCfg.value(HistId::tdcAdcDt).histGroupEn){
            histTRoot[HistId::tdcAdcDt].fill(tdcPrevTime-adcPrevTime);
        }
    }

    if(histTRoot.contains(HistId::tdcHitEv) && cfg.histCfg.value(HistId::tdcHitEv).histGroupEn){
        histTRoot[HistId::tdcHitEv].fill(tdcHits);
    }
    if(histTRoot.contains(HistId::adcHitEv) && cfg.histCfg.value(HistId::adcHitEv).histGroupEn){
        histTRoot[HistId::adcHitEv].fill(adcHits);
    }

    if(histTRoot.contains(HistId::errors) && cfg.histCfg.value(HistId::errors).histGroupEn){
        std::vector<quint16> errorsBits;
        errorsBits.resize(16, 0);
        for (quint16 err : data.errors) {
            for(int bit=0; err; ++bit){
                if(err&1){
                    ++errorsBits[bit];
                }
                err = (err>>1);
            }
        }
        for(int bit=0; bit<16; ++bit){
            if(errorsBits[bit]){
                histTRoot[HistId::errors].fill(bit, errorsBits[bit]);
            }
        }
    }
    refreshCurrentPlot();
}

void HistMngrWidget::refreshCurrentPlot(bool force)
{
    if(!force && lastRefreshTime.isValid() && lastRefreshTime.elapsed()<500) {
        return;
    }
    if(!histTRoot.contains(cfg.curHistId))
        return;
    const Histogramm &curHist = histTRoot.value(cfg.curHistId);
    const bool isInlHist = HistId::isInlId(cfg.curHistId);
    const int nBins = curHist.nBins;

    QwtArray<double> values(curHist.bins);

    int leftEdge = curHist.nEntries ? curHist.findBin(curHist.min_filled_val) : -1;
    int rightEdge = curHist.nEntries ? curHist.findBin(curHist.max_filled_val) : -1;
    if(isInlHist) {
        QwtArray<double> xValues(nBins);
        for (int bin = 0; bin < nBins; ++bin)
            xValues[bin] = curHist.getBinLower(bin);

        if(values.size()==1024)
            values = makeDNLCorrectionTable(values);

        QwtPlotCurve *inlCurv = inlCurvs[cfg.curHistId];
        Q_ASSERT(inlCurv);
        inlCurv->setData(xValues, values);
        inlCurv->itemChanged();
    } else {
        QwtDoubleInterval interval;
        const double binWidth = curHist.binStep;
        if(leftEdge != -1 || rightEdge != -1){
            if(leftEdge == -1)
                leftEdge = 0;
            if(rightEdge == -1)
                rightEdge = curHist.nBins-1;
//            intervals = intervals.mid(leftEdge, rightEdge-leftEdge+1);
            values = values.mid(leftEdge, rightEdge-leftEdge+1);
        }
        QwtArray<QwtDoubleInterval> intervals(rightEdge-leftEdge+1);
        for (int bin = leftEdge; bin <= rightEdge; ++bin)
        {
            interval.setMinValue(curHist.getBinLower(bin));
            interval.setMaxValue(curHist.getBinLower(bin)+binWidth);
            intervals[bin-leftEdge] = interval;
        }
        HistogramItem *qwtHist = histQwtItems[cfg.curHistId];
        Q_ASSERT(qwtHist);
        qwtHist->setData(QwtIntervalData(intervals, values));
        updateFit(curHist);
        qwtHist->itemChanged();
    }
    ui->qwtPlotHists->setAxisAutoScale(QwtPlot::xBottom);
    ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom)->setAttribute(QwtScaleEngine::Floating, true);
//    ui->qwtPlotHists->setAxisScale(QwtPlot::xBottom, intervals.first().minValue(), intervals.last().maxValue());
    ui->qwtPlotHists->replot();

    QStringList histInfo;
    if(isInlHist) {
        const int ch = cfg.curHistId-HistId::chInl;
        double dt = 1e-3*inlStat.time.restart();
        double cnt = inlStat.count[ch];

        inlStat.count[ch] = curHist.getEntries();
        double dCnt = inlStat.count[ch] - cnt;
        qint32 restTime = (1e7-inlStat.count[ch])/dCnt*dt;
        double rateHz = dCnt/dt;
        QString sufix;
        if(rateHz>1e6){
            rateHz/=1e6;
            sufix = 'M';
        } else if(rateHz>1e3){
            rateHz/=1e3;
            sufix = 'k';
        }
        histInfo << QString("rate: %1 %2Hz").arg(rateHz, 0, 'f', 2).arg(sufix);
        QString timeStr;
        if(restTime>0){
        if(restTime>3600){
            timeStr += QString("%1h ").arg(restTime/3600);
            restTime %= 3600;
        }
        if(restTime>60 || !timeStr.isEmpty()){
            timeStr += QString("%1m ").arg(restTime/60);
            restTime %= 60;
        }

        timeStr += QString("%1s").arg(restTime);
        } else {
            timeStr = "done.";
        }
        histInfo << QString("rest time: %1").arg(timeStr);
    } else {
        histInfo << QString("under:%1; over:%2")
                    .arg(curHist.getUnderFlow())
                    .arg(curHist.getOverFlow());
        if(curHist.getEntries())
            histInfo << QString("range: [%1; %2]")
                        .arg(curHist.min_filled_val)
                        .arg(curHist.max_filled_val);
    }

    histInfo << QString("count: %1").arg((qint64)curHist.getEntries());
    lastRefreshTime.start();
    ui->labelHistInfo->setText(histInfo.join('\n'));
}

void HistMngrWidget::on_comboBoxGroup_currentIndexChanged(int)
{
    onCurHistChanged();
}

void HistMngrWidget::on_checkBoxHistGroupEn_toggled(bool checked)
{
    bool ok;
    int groupId = ui->comboBoxGroup->currentData().toInt(&ok);
    Q_ASSERT(ok && groupId!=-1);
    for(int id=0; id<cfg.histCfg.size(); ++id) {
        HistCfg &c = cfg.histCfg[id];
        if(c.groupId == groupId)
            c.histGroupEn = checked;
    }

    HistCfg hc = getCurHist();
    const int groupIndex = ui->comboBoxGroup->findData(hc.groupId);
    if(groupIndex!=-1)
        ui->comboBoxGroup->setItemText(groupIndex, hc.getGroupTitle());
    cfg.save();
}

void HistMngrWidget::on_spinBoxCh_valueChanged(int)
{
    onCurHistChanged();
}

void HistMngrWidget::on_groupBoxStopCh_toggled(bool val)
{
    cfg.useStopCh = val;
    cfg.save();
}

void HistMngrWidget::on_spinBoxStopCh_valueChanged(int ch)
{
    cfg.stopCh = ch-1;
    cfg.save();
}

void HistMngrWidget::on_radioButtonMin_toggled(bool checked)
{
    cfg.amplHistWithMax = !checked;
    cfg.save();
}

void HistMngrWidget::on_radioButtonMax_toggled(bool checked)
{
    cfg.amplHistWithMax = checked;
    cfg.save();
}

void HistMngrWidget::on_pushButtonClear_clicked()
{
    histTRoot[cfg.curHistId].reset();
    refreshCurrentPlot(true);
}

void HistMngrWidget::on_pushButtonClearAll_clicked()
{
    for(int i : histTRoot.keys()){
        histTRoot[i].reset();
    }
    refreshCurrentPlot(true);
}

void HistMngrWidget::on_checkBoxEnable_toggled(bool checked)
{
    cfg.globalEnabled = checked;
    cfg.save();
    if(cfg.globalEnabled)
        refreshTimer->start();
    else {
        refreshTimer->stop();
        refreshCurrentPlot(true);
    }
}

void HistMngrWidget::on_groupBoxFit_toggled(bool checked)
{
    bool ok;
    const int groupId = ui->comboBoxGroup->currentData().toInt(&ok);
    const int cfgCh = getCurCh();
    Q_ASSERT(ok);
    for(int id=0; id<cfg.histCfg.size(); ++id) {
        HistCfg &c = cfg.histCfg[id];
        if(c.groupId == groupId && (c.ch==-1 || c.ch == cfgCh)) {
            c.showFit = checked;
            break;
        }
    }
    cfg.save();
    refreshCurrentPlot(true);
}

void HistMngrWidget::matchWinUpdated(int matchWin)
{
    if(HistId::tdcDoublePulseTime==-1)
        return;

    for(int i=0; i<cfg.histCfg.size(); ++i) {
        HistCfg &hc = cfg.histCfg[i];
        if(hc.histId != HistId::tdcDoublePulseTime)
            continue;

        const int matchWinNs = matchWin*24;
        hc.min = -matchWinNs;
        hc.max = matchWinNs;
        hc.nBins = 2048*matchWin;

        Q_ASSERT(histTRoot.contains(HistId::tdcDoublePulseTime));
        histTRoot[HistId::tdcDoublePulseTime] =
                Histogramm(hc.title, hc.nBins, hc.min, hc.max);
    }
}

void HistMngrWidget::init(QSemaphore *_histSem, bool isInlEnabled)
{
    histSem = _histSem;

    cfg.init(isInlEnabled);
    cfg.load();
    ui->spinBoxCh->blockSignals(true);
    ui->spinBoxCh->setMinimum(1);
    ui->spinBoxCh->setMaximum(HistId::nch);
    ui->spinBoxCh->blockSignals(false);

    ui->comboBoxGroup->blockSignals(true);
    for(const HistCfg &c : cfg.histCfg) {
        const int &id = c.histId;
        histTRoot[c.histId] =  Histogramm(c.title, c.nBins, c.min, c.max);

        const bool isInlHist = HistId::isInlId(id);
        if(isInlHist) {
            inlCurvs[id] = new QwtPlotCurve(c.getGroupTitle());
        } else {
            histQwtItems[id] = new HistogramItem(c.getGroupTitle());
            histQwtItems[id]->setColor(Qt::darkCyan);
        }

        if(ui->comboBoxGroup->findData(c.groupId) == -1)
            ui->comboBoxGroup->insertItem(c.groupId, c.getGroupTitle(), c.groupId);
        if(cfg.curHistTitle == c.title) {
            if(c.ch != -1) {
                ui->spinBoxCh->blockSignals(true);
                ui->spinBoxCh->setValue(c.ch+1);
                ui->spinBoxCh->blockSignals(false);
            }
            ui->comboBoxGroup->setCurrentIndex(ui->comboBoxGroup->findData(c.groupId));
        }
    }
    ui->comboBoxGroup->blockSignals(false);

    onCurHistChanged();

    ui->checkBoxEnable->setChecked(cfg.globalEnabled);

    ui->qwtPlotHists->replot();
    ui->groupBoxStopCh->setChecked(cfg.useStopCh);
    ui->spinBoxStopCh->setValue(cfg.stopCh+1);
    ui->radioButtonMin->setChecked(!cfg.amplHistWithMax);
    ui->radioButtonMax->setChecked(cfg.amplHistWithMax);

    if(HistId::tdcDoublePulseTime != -1)
        matchWinUpdated(5);
}

QVector<double> HistMngrWidget::makeDNLCorrectionTable(const QVector<double> code_density)
{
    if (code_density.size() != 1024) {
        qWarning() << "makeDNLCorrectionTable: invalid input vector size";
        return QVector<double>();
    }
    double sumv = 0;
    for(int i=0; i<1024; i++) {
        double v = code_density[i];
        sumv += v;
    }
    QVector<double> inl(1024, 0.0);
    double avgv = sumv / 1024.;
    double sumrelv = 0;
    double inlsum = 0;
    for(int i=0; i<1024; i++) {
        double relv = (code_density[i] - avgv) / avgv;
        sumrelv += relv;
        inl[i] = sumrelv;
        inlsum += sumrelv;
    }
//    qInfo() << "avgv = " << avgv; // Ok
//    qInfo() << "sumrelv = " << sumrelv;
//    qInfo() << "inlsum = " << inlsum;
    QVector<double> corr(1024, 0.0);
    for(int i=0; i<1024; i++) {
        corr[i] = inl[i] - inlsum/1024.;
        corr[i] = round(corr[i]*100.0) / 100.0; // quantize
    }
//    qInfo() << "TQDC DNL for channel " << ch << corr;
    return corr;
}

void HistMngrWidget::updateInlCorr(const DeviceIndex &index, const QMap<int, QVector<double> > &corrMap)
{
    Q_UNUSED(index);
    for (int ch : corrMap.keys()) {
        curDevCorrMap[ch] = corrMap[ch];
    }
}

void HistMngrWidget::fillDoublePulse(const TqdcEvent &stopEv, const std::map<int, QVector<TqdcEvent> > &tqdcEvents)
{
    if(!cfg.histCfg.value(HistId::tdcDoublePulseTime).histGroupEn)
        return;
    auto &doublePulseHist = histTRoot[HistId::tdcDoublePulseTime];
    const double stopTimeNs = stopEv.getLeadTimeNs(curDevCorrMap.value(stopEv.ch, QVector<double>()));
    std::map<int, QVector<TqdcEvent>>::const_iterator tdcEvIt = tqdcEvents.begin();
    while(tdcEvIt != tqdcEvents.end()){
        const int &ch = tdcEvIt->first;
        if(!cfg.useStopCh || ch != cfg.stopCh) {
            const QVector<TqdcEvent> &tdcEvents = tdcEvIt->second;
//            if(tdcEvents.size()>1)
//                qInfo() << "fillDoublePulse for not-single event";
            for (const TqdcEvent &tdcEv : tdcEvents) {
                if(stopEv.ch == tdcEv.ch && stopEv.leadingTime == tdcEv.leadingTime)
                    continue;
                const double dt = tdcEv.getLeadTimeNs(curDevCorrMap.value(stopEv.ch, QVector<double>()))
                        -stopTimeNs;
                if(cfg.useStopCh || dt>=0){
                    histTRoot[HistId::tdcDoublePulseTime].fill(dt);
                    doublePulseHist.fill(dt);
                }
            }
        }
        ++tdcEvIt;
    }
}

void HistMngrWidget::updateFit(const Histogramm &curHist)
{
    HistCfg curHistCfg = getCurHist();

    if(curHistCfg.showFit && curHist.getEntries()){
        auto dist = curHist.getDistribution();

        double mean = dist.mean;
        double rms = dist.rms;

        if(qFuzzyIsNull(mean))
            mean = 0;
        if(qFuzzyIsNull(rms))
            rms = 0;

        ui->labelFitInfo->setText(QString("Mean:  %1\nSigma: %2")
                            .arg(mean, 0, 'f', 3).arg(rms, 0, 'f', 3));
    } else {
        ui->labelFitInfo->setText(curHistCfg.showFit?"No fit data":"");
    }
}

int HistMngrWidget::getCurCh() const
{
    return ui->spinBoxCh->value()-1;
}

HistCfg HistMngrWidget::getCurHist()
{
    bool ok;
    const int groupId = ui->comboBoxGroup->currentData().toInt(&ok);
    const int ch = getCurCh();
    HistCfg curHistCfg;
    Q_ASSERT(ok);
    for (const HistCfg &c : cfg.histCfg) {
        if(c.groupId == groupId && (c.ch == ch || c.ch == -1)) {
            curHistCfg = c;
            break;
        }
    }
    return curHistCfg;
}

HistCfg HistMngrWidget::onCurHistChanged()
{
    const HistCfg curHistCfg = getCurHist();
    const int histId=curHistCfg.histId;
    if(histId==-1) {
        qWarning() << "GUI logic of hist change is broken";
        return curHistCfg;
    }

    Q_ASSERT(HistId::isInlId(histId) ? inlCurvs.contains(histId) : histQwtItems.contains(histId));
    if(cfg.curHistId != -1) {
        if(HistId::isInlId(cfg.curHistId)) {
            if(inlCurvs.contains(cfg.curHistId))
                inlCurvs[cfg.curHistId]->detach();
        } else {
            if(histQwtItems.contains(cfg.curHistId))
                if(histQwtItems.contains(cfg.curHistId))
                    histQwtItems[cfg.curHistId]->detach();
        }
    }
    cfg.curHistId = histId;
    cfg.curHistTitle = curHistCfg.title;

    if(HistId::isInlId(cfg.curHistId)) {
        inlCurvs[histId]->attach(ui->qwtPlotHists);
    } else {
        histQwtItems[cfg.curHistId]->attach(ui->qwtPlotHists);
    }

    ui->qwtPlotHists->setTitle(curHistCfg.getGroupTitle());
    // bring fit curv to foreground, curve will be drawn on next updateFit(...)
    refreshCurrentPlot(true);
    cfg.save();

    ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom)->setMargins(1, 1);
//    ui->qwtPlotHists->setAxisAutoScale(QwtPlot::xBottom);
//    qInfo()<<"axisAutoScale:"<<ui->qwtPlotHists->axisAutoScale(QwtPlot::xBottom);

    // Update GUI for this hist
    const int groupIndex = ui->comboBoxGroup->findData(curHistCfg.groupId);
    if(groupIndex!=-1) {
        ui->comboBoxGroup->setItemText(groupIndex, curHistCfg.getGroupTitle());
    }
    QString toolTip = cfg.getGroupToolTip(curHistCfg.groupId);
    ui->comboBoxGroup->setToolTip(toolTip);
//    ui->labelHistDescr->setText(toolTip);
    ui->labelHistDescr->setVisible(false && !toolTip.isEmpty());
    ui->checkBoxHistGroupEn->blockSignals(true);
    ui->checkBoxHistGroupEn->setChecked(curHistCfg.histGroupEn);
    ui->checkBoxHistGroupEn->blockSignals(false);
    ui->groupBoxChSelect->setVisible(curHistCfg.ch!=-1);
    ui->groupBoxStopCh->setVisible(HistId::tdcDoublePulseTime!=-1 && histId==HistId::tdcDoublePulseTime);
    ui->groupBoxAmpMinMax->setVisible(HistId::groupChAmpl!=-1 && curHistCfg.groupId==HistId::groupChAmpl);
    ui->groupBoxFit->setVisible(histId!=-1 && curHistCfg.fittable);
    ui->groupBoxFit->setChecked(curHistCfg.showFit);

    return curHistCfg;
}

void HistMngrWidget::requestInlCorr(const DeviceIndex &index)
{
    if(index != inlDeviceIndex)
        return;
    QMap<int, QVector<double> > inlCorrMap;
    for(int ch=0; ch<HistId::nch; ++ch) {
        if(HistId::chInl==-1 || !histTRoot.contains(HistId::chInl+ch))
            continue;

        Histogramm cutHist = histTRoot[HistId::chInl+ch];
        if(cutHist.getEntries()==0)
            continue;
        int nBins = cutHist.nBins;
        Q_ASSERT(nBins==1024);

        QVector<double> values = makeDNLCorrectionTable(cutHist.bins);
        inlCorrMap[ch] = values;
    }
    if(!inlCorrMap.isEmpty()){
        emit inlCorrUpdated(index, inlCorrMap);
        updateInlCorr(index, inlCorrMap);
    }
}

void HistMngrWidget::dataProcessingTdc(const CompleteEvent &ev, int &tdcHits, double &tdcPrevTime)
{
    std::map<int, QVector<TqdcEvent>>::const_iterator tdcEvIt = ev.tqdcEvents.begin();
    // TDC processing

    Histogramm *tdcChHist=nullptr;
    if(histTRoot.contains(HistId::tdcCh) && cfg.histCfg.value(HistId::tdcCh).histGroupEn)
        tdcChHist = &histTRoot[HistId::tdcCh];
    while(tdcEvIt != ev.tqdcEvents.end()){
        const int &ch = tdcEvIt->first;
        const QVector<TqdcEvent> &tdcEvents = tdcEvIt->second;
        tdcHits += tdcEvents.size();
        for (const TqdcEvent &tdcEv : tdcEvents) {
            const double leadTime = tdcEv.getLeadTimeNs();
            if(tdcChHist){
                tdcChHist->fill(ch+1);
            }

            if(histTRoot.contains(HistId::tdcTime) && cfg.histCfg.value(HistId::tdcTime).histGroupEn){
                histTRoot[HistId::tdcTime].fill(leadTime);
            }
            tdcPrevTime = leadTime;
            if(HistId::chInl != -1 && histTRoot.contains(HistId::chInl+ch) && cfg.histCfg.value(HistId::chInl).histGroupEn){
                histTRoot[HistId::chInl+ch].fill(tdcEv.leadingTime%1024);
            }

            if(histTRoot.contains(HistId::tdcDoublePulseTime) && (!cfg.useStopCh || ch == cfg.stopCh))
                fillDoublePulse(tdcEv, ev.tqdcEvents);
        }
        ++tdcEvIt;
    }
}

void HistMngrWidget::dataProcessingAdc(const CompleteEvent &ev, int &adcHits, double &adcPrevTime)
{
    // ADC processing
    std::map<int, QVector<AdcEvent>>::const_iterator adcEvIt = ev.adcEvents.begin();
    const bool chSamplesEn = HistId::chSamples != -1
            && cfg.histCfg.value(HistId::chSamples).histGroupEn;
    const bool chAmplEn = HistId::chAmpl != -1
            && cfg.histCfg.value(HistId::chAmpl).histGroupEn;
    const bool doWaveformProcessing = chSamplesEn || chAmplEn;

    Histogramm *adcChHist = nullptr;
    if(histTRoot.contains(HistId::adcCh) && cfg.histCfg.value(HistId::adcCh).histGroupEn) {
        adcChHist = &histTRoot[HistId::adcCh];
    }
    Histogramm *adcTimeHist = nullptr;
    if(histTRoot.contains(HistId::adcTime) && cfg.histCfg.value(HistId::adcTime).histGroupEn) {
        adcTimeHist = &histTRoot[HistId::adcTime];
    }

    while(adcEvIt != ev.adcEvents.end()) {
        const int &ch = adcEvIt->first;

        const QVector<AdcEvent> &adcEvents = adcEvIt->second;
        adcHits += adcEvents.size();
        for (const AdcEvent &adcEv : adcEvents) {
            bool gotAnySample = false;
            double minMaxSampleVal = 0;
            if(adcChHist){
                adcChHist->fill(ch+1);
            }

            if(adcTimeHist){
                adcTimeHist->fill(adcEv.adcTs*8);
            }

            adcPrevTime = adcEv.adcTs*8;

            if(!doWaveformProcessing)
                continue;

            Histogramm *chSampleHist = nullptr;
            if(chSamplesEn && histTRoot.contains(HistId::chSamples+ch)) {
                chSampleHist = &histTRoot[HistId::chSamples+ch];
            }
            Histogramm *chAmplHist = nullptr;
            if(chAmplEn && histTRoot.contains(HistId::chAmpl+ch)) {
                chAmplHist = &histTRoot[HistId::chAmpl+ch];
            }

            auto wfIt = adcEv.wf.begin();
            int samplePos = adcEv.adcTs;
            while(wfIt != adcEv.wf.end()) {
                const double &val = *wfIt;

                if(chSampleHist){
                    chSampleHist->fill(val);
                }

                if(gotAnySample ||
                        (cfg.amplHistWithMax ?
                         minMaxSampleVal < val :
                         minMaxSampleVal > val)) {
                    gotAnySample = true;
                    minMaxSampleVal = val;
                }
                ++wfIt;
                ++samplePos;
            }

            if(chAmplHist && gotAnySample){
                chAmplHist->fill(minMaxSampleVal);
            }
        }
        ++adcEvIt;
    }
}
