#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "CalibrationDialog.h"
#include "TqdcFileReader.h"
#include "TQDCAnalyzer.h"
#include "device-discover/DiscoverDialog.h"
#include "mregdev/device_id.h"
#include "daq-config/DaqConfig.h"
#include "base/threadQuit.h"
#include "TcpClient.h"
#include "LogDialog.h"
#include "pnp-server/PNPServer.h"
#include "TQDC2StratupOptions.h"
#include "lib-common/UIStyles.h"
#include "zmq-lib/ZmqDialog.h"
#include "TqdcZmqPublisher.h"
#include "m-stream-output/StatisticOutput.h"
#include "base/AboutBox.h"

#include <QFileDialog>
#include <QHostAddress>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include "QShortcut"
#include <QSpinBox>
#include <QStandardPaths>
#include <QThread>
#include <QTimer>
#include <signal.h>
#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qxw/WaveformPlot.h>
#include "TqdcFileWriter.h"
#include <QPlainTextEdit>
#include <QBoxLayout>

namespace {
const int DATA_QUEUE_SIZE = 50;
const int STOP_EVENT_TIMEOUT_MS = 10e3;
enum {COL_CH, COL_EN, COL_TDC_EN, COL_ADC_EN, COL_TRIG_EN, COL_INVERT, COL_THR, COL_TDC_HIT, COL_ADC_HIT, COL_RATE, COL_TOTAL};
enum {COL_DEV_SERIAL, COL_DEV_FW, COL_DEV_MCU_VER, COL_DEV_TEMP, COL_DEV_TIME, COL_DEV_UNBLOCK, COL_DEV_BLOCK, COL_DEV_ANAL_ERR, COL_DEV_MS_ERR, COL_DEV_STATUS, COL_DEV_STREAM, COL_DEV_TOTAL};

QString getRateStr(double rate) {
    QString mask;
    double downScale;
    if(rate<1e3) {
        mask = "%1 Hz";
        downScale = 1;
    } else if(rate<1e6) {
        mask = "%1 kHz";
        downScale = 1e3;
    } else {
        mask = "%1 MHz";
        downScale = 1e6;
    }
    return mask.arg(rate/downScale, 0, 'f', 2);
}

QTableWidgetItem *createTableCheckBox()
{
    QTableWidgetItem *i = new QTableWidgetItem();
    i->setFlags(Qt::ItemIsSelectable | Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
    i->setCheckState(Qt::Unchecked);
    i->setTextAlignment(Qt::AlignCenter);
    return i;
}
}

MainWindow::MainWindow(const TQDC2StartupOptions &startup_options, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    config(startup_options),
    bakConfig(startup_options),
    analyzerSem(new QSemaphore(DATA_QUEUE_SIZE)),
    writerSem(new QSemaphore(DATA_QUEUE_SIZE)),
    histSem(new QSemaphore(DATA_QUEUE_SIZE)),
    reader(new TqdcFileReader(analyzerSem)),
    analyzer(new TQDCAnalyzer(analyzerSem, histSem, writerSem)),
    analyzerThread(new QThread(this)),
    discoverDialog(new DiscoverDialog(this)),
    pnpServer(new PNPServer(this)),
    tqdc_device(nullptr),
    tqdc_thread(new QThread(this)),
    mstream(nullptr),
    blockControl(0),
    fileWriter(new TqdcFileWriter(writerSem)),
    fileWriterThread(new QThread(this)),
    tcpClient(new TcpClient(analyzerSem, writerSem)),
    tcpClientThread(new QThread(this)),
    devStatusLabel(new QLabel(this)),
    busyStatusLabel(new QLabel(this)),
    analyzerStatusLabel(new QLabel(this)),
    writerStatusLabel(new QLabel(this)),
    semStatusLabel(new QLabel(this)),
    logDialog(new LogDialog(this)),
    calibrationDialog(new CalibrationDialog(this)),
    isDevBlocked(false),
    analIsReadyToStop(false),
    writerIsReadyToStop(false),
    lookForStop(false),
    stopEvNum(0),
    config_timer(new QTimer),
    write_request(false),
    tqdcZmqPublisher(new TqdcZmqPublisher(this)),
    zmqDialog(new ZmqDialog(this))
{
    fileWriterThread->setObjectName("Writer");
    analyzerThread->setObjectName("Analyzer");
    tqdc_thread->setObjectName("TQDC device thread");
    tcpClientThread->setObjectName("TCP cl.");
    tcpClientThread->setObjectName("TcpClient thread");
    bool inl_enabled = startup_options.inl;

    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<TqdcStatus>();
    qRegisterMetaType<DeviceCalibration>();
    qRegisterMetaType<QMap<int,double> >();
    qRegisterMetaType<QMap<int,int> >();
    qRegisterMetaType<DeviceData>();
    qRegisterMetaType<AnalyzerStats>();
    qRegisterMetaType<WriterStatistic>();
    ui->setupUi(this);
    ui->widgetHistMngr->init(histSem, inl_enabled);
    ui->labelAdcPreWin->hide();
    ui->labelAdcPreWinNs->hide();
    ui->spinBoxAdcPreWin->hide();
    if(!inl_enabled) {
        ui->menuHPTDC_INL->menuAction()->setVisible(false);
    }

#ifndef QT_DEBUG
    ui->pushButtonBlockRun->hide();
#endif

    ui->checkBoxRawPeriodical->hide();
    ui->spinBoxRawTimer->hide();

    ui->groupBoxFileWrite->setStyleSheet(gbox_fill_blue);
    ui->groupBoxTrigger->setStyleSheet(gbox_fill_blue);
    ui->groupBoxAdc->setStyleSheet(gbox_fill_blue);

    statusBar()->addWidget(devStatusLabel,1);
    statusBar()->addWidget(busyStatusLabel,1);
    statusBar()->addWidget(analyzerStatusLabel,10);
    statusBar()->addWidget(writerStatusLabel,5);
    statusBar()->addWidget(semStatusLabel,4);
    busyStatusLabel->setToolTip("Current Busy Status\n"
                                "1xx - ADC busy\n"
                                "x1x - TDC busy\n"
                                "xx1 - ReadOut busy");
    analyzerStatusLabel->setToolTip("Analyzer status");
    writerStatusLabel->setToolTip("File Writer status");
    semStatusLabel->setToolTip("Queue's semaphores");

    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &LogDialog::show);
    connect(this, &MainWindow::addLog, logDialog, &LogDialog::addLog);

    connect(reader, &TqdcFileReader::dataAccuired, analyzer, &TQDCAnalyzer::dataAccuired);
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated, this, &MainWindow::showWaveform);
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated, tqdcZmqPublisher, &TqdcZmqPublisher::publish_event);
    connect(analyzer, &TQDCAnalyzer::deviceStatsUpdated, this, &MainWindow::updateHitStats);
    connect(this, &MainWindow::resetAnalyzerStats, analyzer, &TQDCAnalyzer::resetStats);
    connect(this, &MainWindow::requestRawData, analyzer, &TQDCAnalyzer::requestRawData);
    connect(analyzer, &TQDCAnalyzer::rawDataUpdated, this, &MainWindow::rawDataUpdated);

    connect(this, &MainWindow::startWrite, analyzer, &TQDCAnalyzer::startWrite);
    connect(this, &MainWindow::stopWrite, analyzer, &TQDCAnalyzer::stopWrite);
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated, ui->widgetHistMngr, &HistMngrWidget::dataUpdated);
    connect(this, &MainWindow::matchWinUpdated, ui->widgetHistMngr, &HistMngrWidget::matchWinUpdated);
    connect(this, &MainWindow::requestInlCorr, ui->widgetHistMngr, &HistMngrWidget::requestInlCorr);
    connect(ui->widgetHistMngr, &HistMngrWidget::inlCorrUpdated, this, &MainWindow::inlCorrUpdated);

    connect(discoverDialog, &DiscoverDialog::deviceDiscovered, this, &MainWindow::deviceUpdated);
//    connect(pnpServer, &PNPServer::programDiscovered, this, &MainWindow::programDiscovered);
    pnpServer->startSearch(DaqConfig::getTypeName(DaqConfigMStream));

    connect(tcpClient, &TcpClient::dataAccuired, fileWriter, &TqdcFileWriter::writeRawData);
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated, fileWriter, &TqdcFileWriter::writeDecodedData);
    connect(this, &MainWindow::startWrite, fileWriter, &TqdcFileWriter::startWrite);
    connect(this, &MainWindow::stopWrite, fileWriter, &TqdcFileWriter::stopWrite);
    connect(fileWriter, &TqdcFileWriter::fileStatisticUpdated, this, &MainWindow::writerStatisticUpdated);
    connect(this, &MainWindow::prepareForStop, fileWriter, &TqdcFileWriter::prepareForStop);
    connect(this, &MainWindow::prepareForStop, analyzer, &TQDCAnalyzer::prepareForStop);
    connect(fileWriter, &TqdcFileWriter::gotStopEvernt, this, &MainWindow::fileWriterReadyForStop);
    connect(analyzer, &TQDCAnalyzer::gotStopEvernt, this, &MainWindow::analyzerReadyForStop);

    connect(zmqDialog, &ZmqDialog::zmq_cfg_change, tqdcZmqPublisher, &ZmqPublisher::config_changed);
    connect(zmqDialog, &ZmqDialog::zmq_cfg_change, this, &MainWindow::zmq_cfg_change);
    connect(tqdcZmqPublisher, &ZmqPublisher::statusUpdated, zmqDialog, &ZmqDialog::statusUpdated);

    connect(this, &MainWindow::startWrite, tcpClient, &TcpClient::startWrite);
    connect(this, &MainWindow::stopWrite, tcpClient, &TcpClient::stopWrite);
    connect(this, &MainWindow::setTcpAddress, tcpClient, &TcpClient::setAddress);
    connect(tcpClient, &TcpClient::dataAccuired, analyzer, &TQDCAnalyzer::dataAccuired);
    connect(tcpClient, &TcpClient::statusUpdated, this,  &MainWindow::tcpClientStatusUpdated);

    connect(this, &MainWindow::closeAll, analyzer, &TQDCAnalyzer::deleteLater);
    connect(analyzer, &TQDCAnalyzer::destroyed, analyzerThread, &QThread::quit);
    connect(this, &MainWindow::closeAll, fileWriter, &TqdcFileWriter::deleteLater);
    connect(fileWriter, &TqdcFileWriter::destroyed, fileWriterThread, &QThread::quit);
    connect(this, &MainWindow::closeAll, tcpClient, &TcpClient::deleteLater);
    connect(tcpClient, &TcpClient::destroyed, tcpClientThread, &QThread::quit);

    connect(calibrationDialog, &CalibrationDialog::calibrationRequest, this,  &MainWindow::calibrationRequest);
    connect(calibrationDialog, &CalibrationDialog::clearCalibration, this, &MainWindow::clearCalibration);

    fileWriter->moveToThread(fileWriterThread);
    fileWriterThread->start();
//   ui->widgetHistMngr->moveToThread(analyzerThread);
    analyzer->moveToThread(analyzerThread);
    analyzerThread->start();
    tcpClient->moveToThread(tcpClientThread);
    tcpClientThread->start();

    connect(this, &MainWindow::setFileNam, reader, &TqdcFileReader::setFileName);
    //emit setFileNam("/home/noafi/data/016940e2_20160321_113252.data");
    //emit setFileNam("/home/noafi/data/06ea9706_20160316_155654.data");

    connect(config_timer, &QTimer::timeout, this, &MainWindow::write_config_timer);
    config_timer->setSingleShot(true);

    discoverDialog->setSingleMode();
    QSet<quint16> supportedDevices;
    supportedDevices
//            << DEVICE_ID_ADC64
//            << DEVICE_ID_ADC64S2
//            << DEVICE_ID_ADC64WR
//            << DEVICE_ID_ADC64VE
//            << DEVICE_ID_ADC8E
//            << DEVICE_ID_ADCM163
//            << DEVICE_ID_VME_TQDC16
            << DEVICE_ID_TQDC16VS_E;
//    supportedDevices.clear();supportedDevices << DEVICE_ID_AnyDevice;
    discoverDialog->setDeviceIdFilter(supportedDevices);
    initTable();
    read_config();

    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    updateControlState();

    configuration_manager = new ConfigurationManager(this, startup_options.program_type, config.default_root);
    connect(configuration_manager, &ConfigurationManager::switch_to_config, this, &MainWindow::switch_to_config);
}

MainWindow::~MainWindow()
{
    write_config(true);
    disconnectFromDev();

    emit closeAll();
    QElapsedTimer closeTimer;
    closeTimer.start();
    if(!threadQuiting(fileWriterThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop fileWriter thread";
    }
    if(!threadQuiting(tqdc_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop device thread";
    }
    if(!threadQuiting(analyzerThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop analyzer thread";
    }
    if(!threadQuiting(tcpClientThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop Mstream-readout thread";
    }
//    if(!threadQuiting(dumpThread, 2000, &closeTimer)) {
//        qWarning() << "Can't stop MStreamDump thread";
//    }

    delete ui;
}

void MainWindow::on_actionConfiguration_manger_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.program_index, config.configuration_name);
}

void MainWindow::switch_to_config(QString program_index, QString configuration_name)
{
    configuration_manager->hide();
    config.program_index = program_index;
    config.configuration_name = configuration_name;

    disconnectFromDev();

    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    read_config();
    updateControlState();
}

void MainWindow::zmq_cfg_change(const ZmqConfig &cfg)
{
    config.zmq_config = cfg;
    config_timer->start(500);
}

void MainWindow::write_config(bool force)
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.setValue("splitterState", ui->splitter->saveState());
    settings.setValue("tabIndex", ui->tabWidget->currentIndex());
    settings.endGroup();

    if(force) {
        config.write_config();
        config_timer->stop();
    } else {
        config_timer->start(500);
    }
}

void MainWindow::write_config_timer()
{
    config.write_config();
}

void MainWindow::temperatureSet(QMap<int, double> temps)
{
    if(temps.isEmpty()) return;
    QList<double> tempList = temps.values();
    double minTemp = *std::min_element(tempList.begin(), tempList.end());
    double maxTemp = *std::max_element(tempList.begin(), tempList.end());
    QColor col = (maxTemp>=70 || minTemp<0) ? Qt::red : Qt::black;
    QString tempStr = QString("%1-%2%3").arg(minTemp,0,'f',1).arg(maxTemp,0,'f',1)
            .arg(QChar(0x02da));
    QStringList fullInfo;
    for (int tempId : temps.keys()) {
        QString suffixStr;
        switch (tempId) {
        case 0:
            suffixStr = "Board temp.";
            break;
        case 1:
            suffixStr = "Pll temp.";
            break;
        case 2:
            suffixStr = "Analog input temp.";
            break;
        default:
            break;
        }
        if(!suffixStr.isEmpty())
            fullInfo << QString("%1%2 - %3").arg(temps[tempId],0,'f',1).arg(QChar(0x02da)).arg(suffixStr);
    }
    QTableWidgetItem * item = ui->tableWidgetDevice->item(0, COL_DEV_TEMP);
    item->setText(tempStr);
    item->setTextColor(col);
    item->setToolTip(fullInfo.join('\n'));
    updateTableSize(ui->tableWidgetDevice);
}

void MainWindow::setDeviceStatus(const QString &status)
{
    devStatusLabel->setText(status);
}

void MainWindow::deviceStatusUpdated(const TqdcStatus &status)
{
    ui->tableWidgetDevice->setUpdatesEnabled(false);

    const bool doResetMStream =
            tqdc_status.mstreamMultiAck != status.mstreamMultiAck ||
            tqdc_status.mstreamHwBufSize != status.mstreamHwBufSize;
    tqdc_status = status;

    if(status.connected) {
        temperatureSet(status.temp);
        ui->tableWidgetDevice->item(0, COL_DEV_FW)->setText(status.fwStr);
        ui->tableWidgetDevice->item(0, COL_DEV_MCU_VER)->setText(QString("%1.%2").arg(status.mcuVer).arg(status.mcuRev));
        ui->tableWidgetDevice->item(0, COL_DEV_TIME)->setText(QDateTime::fromTime_t(status.wrTime).toString());
        ui->tableWidgetDevice->item(0, COL_DEV_UNBLOCK)->setText(QString::number(status.statistic.evNumAccepted));
        if(status.statistic.evNumCandidateRun<status.statistic.evNumAccepted) {
            qint64 delta = (qint64)status.statistic.evNumCandidateRun-status.statistic.evNumAccepted;
            ui->tableWidgetDevice->item(0, COL_DEV_BLOCK)->setText(QString::number(delta));
            ui->tableWidgetDevice->item(0, COL_DEV_BLOCK)->setTextColor(Qt::red);
        } else {
            ui->tableWidgetDevice->item(0, COL_DEV_BLOCK)->setText(QString::number(status.statistic.evNumCandidateRun-status.statistic.evNumAccepted));
            if(ui->tableWidgetDevice->item(0, COL_DEV_BLOCK)->textColor() == Qt::red)
                ui->tableWidgetDevice->item(0, COL_DEV_BLOCK)->setTextColor(Qt::black);
        }
#ifdef QT_DEBUG
//        qInfo() << QString("All:%1; Candidate:%2; Accepted:%3;")
//                   .arg(status.statistic.evNumAll)
//                   .arg(status.statistic.evNumCandidateRun)
//                   .arg(status.statistic.evNumAccepted);
#endif

        QString str;
        if(!status.hptdcStatus[0].getSummary().isEmpty()) {
            str += QString("HPTDC #1:\n%1").arg(status.hptdcStatus[0].getSummary());
        }
        if(!status.hptdcStatus[1].getSummary().isEmpty()) {
            if(!str.isEmpty()) str += "\n";
            str += QString("HPTDC #2:\n%1").arg(status.hptdcStatus[1].getSummary());
        }
        if(!status.adcSelfTestOk) {
            if(!str.isEmpty()) str += "\n";
            str += QString("DesStatus: %1").arg(status.adcStatus, 8, 2, QChar('0'));
        }
        if(str.isEmpty()) {
            str = "Ok";
        }

        bool err = status.hptdcStatus[0].hasError() || status.hptdcStatus[1].hasError()
                 || !status.adcSelfTestOk;
        ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setText(str);
        ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setTextColor(err?Qt::red : Qt::black);

        ui->tableWidgetChCfg->setUpdatesEnabled(false);
        const QMap<int, double> &hitRate = status.hitRate;
        for(int ch = ui->tableWidgetChCfg->rowCount()-1; ch>=0; --ch){
            QString s;
            if(hitRate.contains(ch)){
                const double &rate = hitRate[ch];
                s = getRateStr(rate);
            }
            ui->tableWidgetChCfg->item(ch, COL_RATE)->setText(s);
        }
        updateTableSize(ui->tableWidgetChCfg, true);
        ui->tableWidgetChCfg->setUpdatesEnabled(true);

        QString busyStatusStr = QString("Busy:%1").arg(status.busyStatus, 3, 2, QChar('0'));
        if(busyStatusLabel->text() != busyStatusStr)
            busyStatusLabel->setText(busyStatusStr);
        if(lookForStop && status.runFinished) {
            if(status.statistic.evNumAccepted) {
                stopEvNum = status.statistic.evNumAccepted-1;
                emit prepareForStop(stopEvNum);
            } else {
                analIsReadyToStop = true;
                writerIsReadyToStop = true;
                stopEvNum = 0;
            }
            lookForStop = false;
            qInfo().noquote() << "[MainWindow]: Stop event number detected:"<< status.statistic.evNumAccepted-1;
        }

        if(!mstream) {
            addMStreamDump(status.index);
        }

    } else {
        ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setText("Offline");
        ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setTextColor(Qt::black);
    }
    updateTableSize(ui->tableWidgetDevice);
    ui->tableWidgetDevice->setUpdatesEnabled(true);

    if(doResetMStream) {
        emit mstream_reset(status.index, true,
                           status.mstreamHwBufSize,
                           status.mstreamMultiAck ? status.mstreamHwBufSize : 1,
                           QString(), true);
    }
}

void MainWindow::tcpClientStatusUpdated(bool connOk, double dataRate)
{
//    qInfo() << QString("Sem status: anal=%1; writer=%2; hist=%3")
//               .arg(analyzerSem->available())
//               .arg(writerSem->available())
//               .arg(histSem->available());
    if(ui->tableWidgetDevice->isColumnHidden(COL_DEV_STREAM))
        return;
    ui->tableWidgetDevice->setUpdatesEnabled(false);
    ui->tableWidgetDevice->item(0, COL_DEV_STREAM)->setText(QString("%1").arg(dataRate/1024, 0, 'f', 2));
    ui->tableWidgetDevice->item(0, COL_DEV_STREAM)->setTextColor(connOk ? Qt::black : Qt::red);

    updateTableSize(ui->tableWidgetDevice);
    ui->tableWidgetDevice->setUpdatesEnabled(true);
    if(tqdc_device && !connOk) {
        pnpServer->startSearch(DaqConfig::getTypeName(DaqConfigMStream));
//        if(tqdc_device && /*mstreamProc->state() == QProcess::NotRunning*/ ){
//            emit mstream_reset(tqdc_status.index, true,
//                               tqdc_status.mstreamHwBufSize,
//                               tqdc_status.mstreamMultiAck ? tqdc_status.mstreamHwBufSize : 1,
//                               QString(), true);
//        }
    }
}

void MainWindow::deviceConnected()
{
    if(sender() != tqdc_device)
        return;
    ui->tableWidgetDevice->item(0, COL_DEV_SERIAL)->setTextColor(Qt::darkGreen);
    updateTableSize(ui->tableWidgetDevice);
}

void MainWindow::deviceDisconnected()
{
    if(sender() != tqdc_device)
        return;
    ui->tableWidgetDevice->item(0, COL_DEV_SERIAL)->setTextColor(Qt::red);
}

void MainWindow::deviceUpdated(const DeviceDescription &dd)
{
    knownDevDescr[dd.getIndex()] = dd;
}

void MainWindow::programDiscovered(const ProgramDescription &pd)
{
    if(pd.type == DaqConfig::getTypeName(DaqConfigMStream)){
        bool ok = true;
        quint32 progSerial = pd.index.toUInt(&ok, 16);
        if(ok && config.curDev.getSerial()==progSerial) {
            auto pd2 = pd;
            ProgramInterface i = pd2.getProgramInterface(ProgramInterfaceDataFlow);
            if(!i.isEmpty()){
                emit setTcpAddress(i.host, i.port);
            }
        }
    }
}

void MainWindow::showWaveform(const DeviceData &data, bool)
{
    if(lastWaveformTime.isNull() || lastWaveformTime.elapsed()>500){
        if(data.events.empty())
            return;

        // find non-empty event
        int evIndex=data.events.size()-1;
        for(; evIndex>=0; --evIndex) {
            const CompleteEvent &event = data.events[evIndex];
            if(!event.adcEvents.empty())
                break;
        }
        if(evIndex<0)
            return;

        const DeviceSetup &ds = config.knownSetups.value(data.index, DeviceSetup());
        const DeviceCalibration &devCal = config.allCal.value(data.index, DeviceCalibration());
        const QMap<int, double> &baselineMap = ds.isGain4() ? devCal.x4 : devCal.x1;

        WaveformDataAll mvv;
        const CompleteEvent &event = data.events[evIndex];
        std::map<int, QVector<AdcEvent> >::const_iterator itAdc=event.adcEvents.begin();
        while(itAdc!=event.adcEvents.end()) {
            const QVector<AdcEvent> &adcEvs = itAdc->second;
            const int &ch = itAdc->first;
            const double baseline = baselineMap.value(ch, 0);
            for (const AdcEvent &adcEv : adcEvs) {
                std::vector<double> vx(adcEv.wf.size());
                std::vector<double> vy(adcEv.wf);
                for(uint i=0; i<vx.size(); ++i) {
                    vx[i] = (adcEv.adcTs+i)*8;
                    vy[i] -= baseline;
                    vy[i] /= 4; //  from 16-bit to 14-bit
                }
                WaveformData wd(vx, vy, QString("ch%1").arg(ch+1).toStdString(), ch);
                mvv.push_back(wd);
//                qInfo() << QString("Adc ev: ts=%1").arg(adcEv.adcTs*8);
            }
            itAdc++;
        }
//        std::map<int, QVector<TqdcEvent> >::const_iterator itTdc=event.tqdcEvents.begin();
//        while(itTdc!=event.tqdcEvents.end()) {
//            const QVector<TqdcEvent> &tdcEvs = itTdc->second;
//            for (const TqdcEvent &tdcEv : tdcEvs) {
//                qInfo() << QString("Tdc ev: ch=%1; lead=%2; trail=%3; width=%4; tdcTs=%5")
//                           .arg(tdcEv.ch).arg(tdcEv.getLeadTimeNs()).arg(tdcEv.getTrailTimeNs())
//                           .arg(tdcEv.getWidthNs())
//                           .arg(tdcEv.tdcTs);
//            }
//            itTdc++;
//        }
        lastWaveformTime.start();
        ui->widgetWaveform->replot();

        ui->widgetWaveform->setAllData(mvv);
    }
}

void MainWindow::updateHitStats(const AnalyzerStats &stats)
{
    QTableWidget *table = ui->tableWidgetChCfg;
    table->setUpdatesEnabled(false);
    for(int ch=0; ch<table->rowCount(); ++ch) {
        table->item(ch, COL_TDC_HIT)->setText(QString::number(stats.tdcHits[ch]));
        table->item(ch, COL_ADC_HIT)->setText(QString::number(stats.adcHits[ch]));
    }
    table->setUpdatesEnabled(true);

    ui->tableWidgetDevice->setUpdatesEnabled(false);
    ui->tableWidgetDevice->item(0, COL_DEV_ANAL_ERR)->setText(QString::number(stats.formatErrorCnt));
    if(stats.formatErrorCnt)
        ui->tableWidgetDevice->showColumn(COL_DEV_ANAL_ERR);
    else
        ui->tableWidgetDevice->hideColumn(COL_DEV_ANAL_ERR);
    ui->tableWidgetDevice->setUpdatesEnabled(true);
    updateTableSize(ui->tableWidgetDevice);
    QString analStr = QString("Ev:%1").arg(stats.eventDecoded);
    if(stats.lastEvNum+1 != stats.eventDecoded) {
        analStr += QString("(lastEv=%2)").arg(stats.lastEvNum);
    }
    if(stats.eventDecoded) {
        analStr += QString("; Time per ev:%1us; Event: %2B; Size per sec: %3KB")
                .arg(1e-3*stats.decodeTimeNs/stats.eventDecoded)
                .arg(stats.sizeDecoded/stats.eventDecoded)
                .arg(1e6*stats.sizeDecoded/stats.decodeTimeNs);
    }

    analyzerStatusLabel->setText(analStr);
}

void MainWindow::rawDataUpdated(const QStringList &data)
{
    ui->plainTextEditRawData->setPlainText(data.join('\n'));
    ui->plainTextEditRawData->setEnabled(true);
    ui->pushButtonRequestRaw->setEnabled(true);
}

void MainWindow::deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCount)
{
    if(config.curDev != index)
        return;
    if(config.knownSetups[index].getNCh() != chNum) {
        config.knownSetups[index].setNCh(chNum);
        write_config();
    }

    ui->spinBoxLatency  -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
//    ui->spinBoxMatchWin -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
    ui->spinBoxAdcWin   -> setMaximum(maxSampleCount ? maxSampleCount   : 625);
    updateGuiVal();
}

void MainWindow::writerStatisticUpdated(const WriterStatistic &s)
{
    qint64 fileSize = s.fileSize;
    QString sizeStr;
    if(fileSize==0)
        sizeStr = "0";
    while (fileSize) {
        int partSize = (fileSize/1000) ? 3 : 0;
        if(!sizeStr.isEmpty()) sizeStr.prepend('.');
        sizeStr.prepend(QString("%1").arg(fileSize%1000, partSize, 10, QChar('0')));
        fileSize /= 1000;
    }
    ui->labelFileInfo->setText(QString("File size: %1 bytes").arg(sizeStr));

    QString statusStr;
    statusStr += QString("Speed:%1 MB/s;")
            .arg(1e3*s.fileSize/s.writeTimeNs, 0, 'f', 2);
    if(!config.dataFileIsRaw) {
        statusStr += QString(" Per event: %2 us")
                .arg(1e-3*s.writeTimeNs/s.evWritten,0,'f',0);
    }
    writerStatusLabel->setText(statusStr);
}

void MainWindow::analyzerReadyForStop(quint32 evNum)
{
    if(stopEvNum == evNum) {
        analIsReadyToStop = true;
    } else {
        qWarning() << QString("Analyzer found wrong stop event number; stopEvNum=%1; writerStopEvNum=%2")
                      .arg(stopEvNum).arg(evNum);
    }
}

void MainWindow::fileWriterReadyForStop(quint32 evNum)
{
    if(stopEvNum == evNum) {
        writerIsReadyToStop = true;
    } else {
        qWarning() << QString("Writer found wrong stop event number; stopEvNum=%1; writerStopEvNum=%2")
                      .arg(stopEvNum).arg(evNum);
    }
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::on_actionCalibrationData_triggered()
{
    calibrationDialog->show();
}

void MainWindow::on_actionSeparate_channel_enable_toggled(bool checked)
{
    config.separateChEn = checked;
    updateGuiVal();
}

void MainWindow::on_actionReset_MStream_triggered()
{
    emit mstream_reset(tqdc_status.index, true,
                       tqdc_status.mstreamHwBufSize,
                       tqdc_status.mstreamMultiAck ? tqdc_status.mstreamHwBufSize : 1,
                       QString(), true);
}

void MainWindow::on_actionSend_soft_reset_triggered()
{
    emit softReset();
}

void MainWindow::on_action0MQ_publisher_triggered()
{
    zmqDialog->show();
}

void MainWindow::on_actionSave_INL_corrections_triggered()
{
    emit requestInlCorr(config.curDev);
}

void MainWindow::on_actionTxt_data_format_triggered()
{
    QString s;
    s += "<h2>TQDC2 Txt Data format</h2>";
    s += "<p>File constructs with following lines:</p>";
    s += "<ul>";
    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
    s += "</ul>";
    s += "<p>where:</p>";
    s += "<dl>";
    s += "<dt><b>TAI time</b></dt>";
    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
    s += "<dt><b>ch</b></dt>";
    s += "<dd><i>Channel number in range 0-15</i></dd>";
    s += "<dt><b>tdcLeadTime</b></dt>";
    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>adcTime</b></dt>";
    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>sample</b></dt>";
    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
    s += "</dl>";

    QMessageBox::about (this, "Txt data format", s);
}

void MainWindow::on_actionRaw_data_format_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
//    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d.setLayout(mainLayout);
    d.exec();

//    QMessageBox::about (this, "Raw data format", html);
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::calibrateFinished(const DeviceIndex &index,
                                   const QMap<int, double> &adcVals, const QMap<int, double> &rms,
                                   const QMap<int, double> &adcVals4, const QMap<int, double> &rms4,
                                   const QMap<int, int> &dacOffset)
{
    --blockControl;
    updateControlState();
    if(config.curDev != index)
        return;
    calibrationDialog->show();
    const double curDevTemp = tqdc_status.temp.value(0,0);
    config.updateBaseCal(index, adcVals, adcVals4, dacOffset, curDevTemp);

    calibrationDialog->updateBaseline(adcVals, adcVals4, rms, rms4);
    calibrationDialog->updateDacOffset(dacOffset);
    setupDevice();
}

void MainWindow::calibrationRequest()
{
    if(blockControl)
        return;
    if(config.curDev.isNull())
        return;

    ++blockControl;
    updateControlState();

    emit startCalibration();
}

void MainWindow::clearCalibration()
{
    config.clearCalibration(config.curDev);
    const DeviceCalibration &devCal = config.getCurrentCalibration();

    calibrationDialog->updateBaseline(devCal.x1, devCal.x4);
    calibrationDialog->updateDacOffset(devCal.thrOffset);

    setupDevice();
}

void MainWindow::inlCorrUpdated(const DeviceIndex &index, const QMap<int, QVector<double> > &inlCorr)
{
    if(index != config.curDev)
        return;
    const double curDevTemp = tqdc_status.temp.value(0,0);
    config.updateInl(index, inlCorr, curDevTemp);
    calibrationDialog->updateInl(inlCorr);
}

void MainWindow::on_pushButtonRequestRaw_clicked()
{
    if(!config.rawDataLimit) {
        qWarning() << "config.rawDataLimit was dropped to 0";
        ui->spinBoxRawEvents->setValue(1);
    }
    ui->pushButtonRequestRaw->setEnabled(false);
    emit requestRawData(config.rawDataLimit);
    ui->plainTextEditRawData->setEnabled(false);
}

void MainWindow::on_spinBoxRawEvents_valueChanged(int limit)
{
    config.rawDataLimit = static_cast<uint>(limit);
    write_config();
}

void MainWindow::on_pushButtonChangeDevice_clicked()
{
    if(discoverDialog->exec() == QDialog::Accepted){
        bool haveEnabledDevice=false;
        for(auto dd : discoverDialog->getSelectedDevices()) {
            if(!dd.enabled)
                continue;
            haveEnabledDevice = true;

            if(config.curDev == dd.getIndex()){
                break;
            }

            connectToDev(dd);
            break;
        }
        if(!haveEnabledDevice){
            config.curDev = DeviceIndex();
            disconnectFromDev();
        }
        write_config();
    }

    updateControlState();
}

void MainWindow::on_pushButtonStart_clicked()
{
    qInfo() << "";
    if(config.curDev.isNull())
        return;
    if(config.writeEnabled) {
        QString fileName = updateFileName(true);
        QFileInfo fileInfo(fileName);
        if(fileInfo.exists()){
            QMessageBox msgBox(QMessageBox::Warning,
                               "File overwrite",
                               QString("File '%1' is already exists.\nDo you want to overwrite it?").arg(fileName),
                               QMessageBox::Yes | QMessageBox::No);
            msgBox.setDefaultButton(QMessageBox::No);
            int ret = msgBox.exec();
            if(ret != QMessageBox::Yes) {
                updateFileName();
                return;
            }
        }
        emit startWrite(fileName, config.dataFileIsRaw);
    }
    DeviceSetup &ds  = config.getCurrentConfig();
    ds.run = true;
    updateControlState();

    setupDevice();
    emit resetAnalyzerStats();
}

void MainWindow::on_pushButtonStop_clicked()
{
    qInfo() << "";
    if(config.curDev.isNull())
        return;
    DeviceSetup &ds  = config.getCurrentConfig();
    ds.run = false;
    setupDevice();

    writerIsReadyToStop = !config.writeEnabled;
    analIsReadyToStop = false;
    lookForStop = true;
    stopEvNum = 0;

    QElapsedTimer t;
    t.start();
    bool ok=false;
    semStatusLabel->show();
    while(t.elapsed() < STOP_EVENT_TIMEOUT_MS) {
        QString semStatusStr = QString("Sem status: anal=%1; writer=%2; hist=%3")
                .arg(analyzerSem->available())
                .arg(writerSem->available())
                .arg(histSem->available());
        semStatusLabel->setText(semStatusStr);

        if(!config.writeEnabled || config.dataFileIsRaw) {
            if(analIsReadyToStop) {
                ok = true;
                break;
            }
        } else {
            if(writerIsReadyToStop) {
                ok = true;
                break;
            }
        }
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
    }
    if(!ok)
        qWarning() << "Correct stop run failed";
    else
        semStatusLabel->hide();
    qInfo().noquote() << "[MainWindow]: Stop run in"<< t.elapsed()<<"ms";

    if(config.writeEnabled)
        emit stopWrite();
    updateControlState();
}

void MainWindow::on_pushButtonBlockRun_clicked()
{
    isDevBlocked = !isDevBlocked;
    ui->pushButtonBlockRun->setText(isDevBlocked?"Unblock run":"Block run");

    emit blockDeviceRun(isDevBlocked);
}

void MainWindow::on_groupBoxFileWrite_toggled(bool en)
{
    config.writeEnabled = en;
    write_config();
}

void MainWindow::on_radioButtonRawFile_toggled(bool checked)
{
    config.dataFileIsRaw = checked;
    write_config();
    updateFileName();
}

void MainWindow::on_pushButtonChangeDir_clicked()
{
    QString fileName = QFileDialog::getExistingDirectory(this,
        tr("Write file"), config.fileDir);
    if(!fileName.isEmpty()) {
        ui->lineEditFilePath->setText(fileName);
        config.fileDir = fileName;
        write_config();
        updateFileName();
    }
}

void MainWindow::on_lineEditFilePrefix_textChanged(const QString &prefix)
{
    config.filePrefix = prefix;
    write_config();
    updateFileName();
}

void MainWindow::on_checkBoxAddTime_toggled(bool checked)
{
    config.fileTime = checked;
    write_config();
    updateFileName();
}

void MainWindow::on_lineEditFileSuffix_textChanged(const QString &suffix)
{
    config.fileSuffix = suffix;
    write_config();
    updateFileName();
}

void MainWindow::on_radioButtonGain1_toggled(bool checked)
{
    if(config.curDev.isNull())
        return;
    DeviceSetup &ds  = config.getCurrentConfig();
    for(int ch=0; ch<ds.getNCh(); ++ch)
        ds.chGain[ch] = !checked;
    setupDevice();
}

void MainWindow::on_radioButtonGain4_toggled(bool checked)
{
    if(config.curDev.isNull())
        return;
    DeviceSetup &ds  = config.getCurrentConfig();
    for(int ch=0; ch<ds.getNCh(); ++ch)
        ds.chGain[ch] = checked;
    setupDevice();
}

void MainWindow::on_checkBoxAdcZs_toggled(bool zsEn)
{
    if(config.curDev.isNull())
        return;
    DeviceSetup &ds  = config.getCurrentConfig();
    ds.zsEn = zsEn;
    ui->spinBoxAdcWin->setEnabled(zsEn);
    ui->spinBoxAdcPreWin->setEnabled(zsEn);
    setupDevice();
}

void MainWindow::on_checkBoxTrigTimer_toggled(bool checked)
{
    if(config.curDev.isNull())
        return;
    config.getCurrentConfig().trigsetup.bit.timer_en = checked ? 1 : 0;
    setupDevice();
    ui->spinBoxTimeTrig->setEnabled(checked);
}

void MainWindow::on_spinBoxTimeTrig_valueChanged(int rateHz)
{
    if(config.curDev.isNull())
        return;
    double period = 125e6/rateHz;
    config.getCurrentConfig().trigPeriod = period;
    setupDevice();
}

void MainWindow::on_checkBoxTrigTtl_toggled(bool checked)
{
    if(config.curDev.isNull())
        return;
    config.getCurrentConfig().trigsetup.bit.ttl_en = checked ? 1 : 0;
    setupDevice();
}

void MainWindow::on_checkBoxTrigTtc_toggled(bool checked)
{
    if(config.curDev.isNull())
        return;
    config.getCurrentConfig().trigsetup.bit.ttc_en = checked ? 1 : 0;
    setupDevice();
}

void MainWindow::on_checkBoxSelfTrig_toggled(bool checked)
{
    ui->spinBoxSelfTrigDelay->setEnabled(checked);
    if(config.curDev.isNull())
        return;
    config.getCurrentConfig().trigsetup.bit.selftrig_en = checked ? 1 : 0;
    setupDevice();
    ui->tableWidgetChCfg->setColumnHidden(COL_TRIG_EN, !checked);
}

void MainWindow::on_spinBoxSelfTrigDelay_valueChanged(int delay)
{
    ui->labelSelfTrigDelay->setText(QString("%1 ns").arg(delay*8));
    if(config.curDev.isNull())
        return;
    config.getCurrentConfig().selfTrigDelay = delay;
    setupDevice();
}

void MainWindow::on_spinBoxThreshold_valueChanged(int thr)
{
    config.getCurrentConfig().setThreshold(thr);
    if(blockControl) return;
    updateGuiVal();
    setupDevice();
}

void MainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    ui->spinBoxMatchWin->setMaximum(lat);
    config.getCurrentConfig().latency = lat;
    if(blockControl) return;
    setupDevice();
}

void MainWindow::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    emit matchWinUpdated(win);
    config.getCurrentConfig().matchWin = win;
    if(blockControl) return;
    setupDevice();
}

void MainWindow::on_spinBoxAdcWin_valueChanged(int val)
{
    ui->labelAdcWinNs->setText(QString("%1 ns").arg(val*8));

    config.getCurrentConfig().readCellNumber = val;
    if(blockControl) return;
    setupDevice();
}

void MainWindow::on_spinBoxAdcPreWin_valueChanged(int preWin)
{
    ui->labelAdcPreWinNs->setText(QString("%1 ns").arg(preWin*8));
    config.getCurrentConfig().adcPreWin = preWin;
    if(blockControl) return;
    setupDevice();
}

void MainWindow::on_tabWidget_currentChanged(int index)
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("tabIndex", index);
    settings.endGroup();
}

void MainWindow::on_tableWidgetChCfg_cellChanged(int row, int column)
{
    if(blockControl) return;
    if(column != COL_EN
            && column != COL_TDC_EN
            && column != COL_ADC_EN
            && column != COL_TRIG_EN
            && column != COL_INVERT)
        return;
    QTableWidgetItem *item = ui->tableWidgetChCfg->item(row, column);
    const bool checked = item->checkState()== Qt::Checked;
    DeviceSetup &setup = config.getCurrentConfig();
    switch (column) {
    case COL_EN:
        setup.chTdcEn[row] = checked;
        setup.chAdcEn[row] = checked;
        break;
    case COL_TDC_EN:
        setup.chTdcEn[row] = checked;
        break;
    case COL_ADC_EN:
        setup.chAdcEn[row] = checked;
        break;
    case COL_TRIG_EN:
        setup.chTrigEn[row] = checked;
        break;
    case COL_INVERT:
        setup.chInv[row] = checked;
        break;
    default:
        return;
    }
    updateTableSize(ui->tableWidgetChCfg, true);
    setupDevice();
}

void MainWindow::onCellThresholdChange(int thr)
{
    if(blockControl) return;
    QTableWidget *table = ui->tableWidgetChCfg;
    int ch=0;
    for(; ch < table->rowCount(); ++ch){
        if(sender() == table->cellWidget(ch, COL_THR))
            break;
    }
    if(ch == table->rowCount()) {
        qWarning()<<"Can't find channal number for threshold spinbox";
        return;
    }

    config.getCurrentConfig().chThr[ch] = thr;
    setupDevice();
}

void MainWindow::read_config()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    ui->splitter->restoreState(settings.value("splitterState").toByteArray());
    int tabCurrentIndex = settings.value("tabIndex").toInt();
    settings.endGroup();

    config.read_config();

    QVector<DeviceDescription> devicesAll;
    QVector<DeviceDescription> devicesEnabled;

    for (const DeviceIndex &index : config.knownSetups.keys()) {
        const DeviceSetup &ds = config.knownSetups[index];

        DeviceDescription dd;
        dd.device_id = ds.index.getDevId();
        dd.serial_id = ds.index.getSerial();
        dd.enabled = (ds.index==config.curDev);
        devicesAll.append(dd);
        if(dd.enabled){
            devicesEnabled.push_back(dd);
        }
    }
    updateTableSize(ui->tableWidgetChCfg, true);
    if(!config.curDev.isNull())
        connectToDev(config.curDev);

    discoverDialog->setSelected(devicesAll);

    (config.dataFileIsRaw ?
                ui->radioButtonRawFile :
                ui->radioButtonDecodedFile)->setChecked(true);
    ui->groupBoxFileWrite->setChecked(config.writeEnabled);
    ui->actionSeparate_channel_enable->setChecked(config.separateChEn);
    ui->lineEditFilePrefix->setText(config.filePrefix);
    ui->checkBoxAddTime->setChecked(config.fileTime);
    ui->lineEditFileSuffix->setText(config.fileSuffix);
    updateFileName();
    if(!config.curDev.isNull())
        emit matchWinUpdated(config.getCurrentConfig().matchWin);
    ui->tabWidget->setCurrentIndex(tabCurrentIndex);

    ui->spinBoxRawEvents->setValue(config.rawDataLimit);

    zmqDialog->setConfig(config.zmq_config);
    tqdcZmqPublisher->config_changed(config.zmq_config);
}

void MainWindow::initTable()
{
    ++blockControl;
    QTableWidget * const table = ui->tableWidgetChCfg;
    table->verticalHeader()->hide();
    table->setColumnCount(COL_TOTAL);
    table->setHorizontalHeaderItem(COL_CH, new QTableWidgetItem("Ch #"));
    table->setHorizontalHeaderItem(COL_EN, new QTableWidgetItem("En"));
    table->setHorizontalHeaderItem(COL_TDC_EN, new QTableWidgetItem("TdcEn"));
    table->setHorizontalHeaderItem(COL_ADC_EN, new QTableWidgetItem("AdcEn"));
    table->setHorizontalHeaderItem(COL_TRIG_EN, new QTableWidgetItem("Trig En"));
    table->setHorizontalHeaderItem(COL_INVERT, new QTableWidgetItem("Invert"));
    table->setHorizontalHeaderItem(COL_THR, new QTableWidgetItem("Threshold"));
    table->setHorizontalHeaderItem(COL_TDC_HIT, new QTableWidgetItem("TDC hits"));
    table->setHorizontalHeaderItem(COL_ADC_HIT, new QTableWidgetItem("ADC hits"));
    table->setHorizontalHeaderItem(COL_RATE, new QTableWidgetItem(" Rate, Hz "));

    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    table->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(table, true);
    setTableRowsCount(16);

    QTableWidget * const tableDev = ui->tableWidgetDevice;
    tableDev->verticalHeader()->hide();
    tableDev->setColumnCount(COL_DEV_TOTAL);
    tableDev->setHorizontalHeaderItem(COL_DEV_SERIAL, new QTableWidgetItem("Serial"));
    tableDev->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TEMP, new QTableWidgetItem("temp"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TIME, new QTableWidgetItem("time"));
    tableDev->setHorizontalHeaderItem(COL_DEV_UNBLOCK, new QTableWidgetItem("Acepted"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BLOCK, new QTableWidgetItem("Blocked"));
    tableDev->setHorizontalHeaderItem(COL_DEV_ANAL_ERR, new QTableWidgetItem("Analyzer err."));
    tableDev->setHorizontalHeaderItem(COL_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STATUS, new QTableWidgetItem("Status"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STREAM, new QTableWidgetItem("Data Stream, kB/s"));
    tableDev->setRowCount(1);
    tableDev->hideColumn(COL_DEV_TIME);
    for(int col=0; col<COL_DEV_TOTAL; ++col) {
        tableDev->setItem(0, col, new QTableWidgetItem());
        if(col == COL_DEV_MS_ERR || col == COL_DEV_ANAL_ERR)
            tableDev->item(0, col)->setTextColor(Qt::red);
    }
    tableDev->hideColumn(COL_DEV_ANAL_ERR);
    tableDev->hideColumn(COL_DEV_MS_ERR);

    tableDev->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    tableDev->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(tableDev);
    --blockControl;
}

void MainWindow::setTableRowsCount(int nch)
{
    QTableWidget *table = ui->tableWidgetChCfg;
    int oldNch = table->rowCount();
    table->setRowCount(nch);
    for(int row=oldNch; row<nch; ++row){
        QTableWidgetItem *i = new QTableWidgetItem(QString::number(row+1));
        i->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
        i->setTextAlignment(Qt::AlignCenter);
        table->setItem(row, COL_CH, i);

        table->setItem(row, COL_EN, createTableCheckBox());
        table->setItem(row, COL_TDC_EN, createTableCheckBox());
        table->setItem(row, COL_ADC_EN, createTableCheckBox());
        table->setItem(row, COL_TRIG_EN, createTableCheckBox());
        table->setItem(row, COL_INVERT, createTableCheckBox());

        QSpinBox *sb = new QSpinBox(this);
        table->setCellWidget(row, COL_THR, sb);
//        sb->setRange(-32000, 32000);
        sb->setRange(ui->spinBoxThreshold->minimum(), ui->spinBoxThreshold->maximum());
        sb->setSingleStep(ui->spinBoxThreshold->singleStep());
        sb->setSuffix(ui->spinBoxThreshold->suffix());
        connect(sb, QOverload<int>::of(&QSpinBox::valueChanged), this, &MainWindow::onCellThresholdChange);

        i = new QTableWidgetItem();
        i->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
        i->setTextAlignment(Qt::AlignRight);
        table->setItem(row, COL_TDC_HIT, i);

        i = new QTableWidgetItem();
        i->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
        i->setTextAlignment(Qt::AlignRight);
        table->setItem(row, COL_ADC_HIT, i);

        i = new QTableWidgetItem();
        i->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
        i->setTextAlignment(Qt::AlignRight);
        table->setItem(row, COL_RATE, i);
    }
}

void MainWindow::setupDevice(bool force)
{
    if(blockControl) return;
    write_config(force);
    emit setupChanged(config.getCurrentConfig(),
                      config.getCurrentCalibration());
    emit mstream_reset(tqdc_status.index, true,
                       tqdc_status.mstreamHwBufSize,
                       tqdc_status.mstreamMultiAck ? tqdc_status.mstreamHwBufSize : 1,
                       QString(), true);
}

void MainWindow::updateGuiVal()
{
    if(config.curDev.isNull())
        return;
    ++blockControl;
    DeviceSetup &ds = config.getCurrentConfig();
    QTableWidget *table = ui->tableWidgetChCfg;

    const bool selfTrig = ds.trigsetup.bit.selftrig_en;
    ui->checkBoxSelfTrig->setChecked(selfTrig);
    ui->spinBoxSelfTrigDelay->setEnabled(selfTrig);
    table->setColumnHidden(COL_TRIG_EN, !selfTrig);

    ui->checkBoxTrigTimer->setChecked(ds.trigsetup.bit.timer_en);
    ui->spinBoxTimeTrig->setEnabled(ds.trigsetup.bit.timer_en);
    double rate = 125e6/config.getCurrentConfig().trigPeriod;
    ui->spinBoxTimeTrig->setValue(rate);
    ui->spinBoxSelfTrigDelay->setValue(ds.selfTrigDelay);
    ui->labelSelfTrigDelay->setText(QString("%1 ns").arg(ds.selfTrigDelay*8));

    ui->checkBoxTrigTtl->setChecked(ds.trigsetup.bit.ttl_en);
    ui->checkBoxTrigTtc->setChecked(ds.trigsetup.bit.ttc_en);

    ui->spinBoxAdcPreWin->setValue(ds.adcPreWin);
    ui->spinBoxAdcWin->setValue(ds.readCellNumber);
    ui->spinBoxLatency->setValue(ds.latency);
    ui->spinBoxMatchWin->setMaximum(ds.latency);
    ui->spinBoxMatchWin->setValue(ds.matchWin);

    table->setColumnHidden(COL_EN, config.separateChEn);
    table->setColumnHidden(COL_TDC_EN, !config.separateChEn);
    table->setColumnHidden(COL_ADC_EN, !config.separateChEn);

    Q_ASSERT(ds.getNCh());
    setTableRowsCount(ds.getNCh());
    int minThr=0;
    bool minThrFound = false;
    for(int ch=0; ch<ds.getNCh(); ++ch){
        table->item(ch, COL_EN)
                ->setCheckState(ds.chTdcEn[ch]==ds.chAdcEn[ch] ?
                                    (ds.chTdcEn[ch] ? Qt::Checked : Qt::Unchecked)
                                  : Qt::PartiallyChecked);
        table->item(ch, COL_TDC_EN)
                ->setCheckState(ds.chTdcEn[ch] ? Qt::Checked : Qt::Unchecked);
        table->item(ch, COL_ADC_EN)
                ->setCheckState(ds.chAdcEn[ch] ? Qt::Checked : Qt::Unchecked);
        table->item(ch, COL_INVERT)
                ->setCheckState(ds.chInv[ch] ? Qt::Checked : Qt::Unchecked);
        table->item(ch, COL_TRIG_EN)
                ->setCheckState(ds.chTrigEn[ch] ? Qt::Checked : Qt::Unchecked);

        QSpinBox *sb = qobject_cast<QSpinBox*>(table->cellWidget(ch, COL_THR));
        Q_ASSERT(sb);
        sb->setValue(ds.chThr[ch]);
        if(ds.chTdcEn[ch] || ds.chAdcEn[ch]){
            if(minThrFound)
                minThr = qMin(ds.chThr[ch], minThr);
            else {
                minThr = ds.chThr[ch];
                minThrFound = true;
            }
        }
    }
    if(minThrFound) {
        ui->spinBoxThreshold->blockSignals(true);
        ui->spinBoxThreshold->setValue(minThr);
        ui->spinBoxThreshold->blockSignals(false);
    }
    ui->checkBoxAdcZs->setChecked(ds.zsEn);
    ui->spinBoxAdcWin->setEnabled(ds.zsEn);
    ui->spinBoxAdcPreWin->setEnabled(ds.zsEn);

    ui->radioButtonGain1->setChecked(!ds.isGain4());
    ui->radioButtonGain4->setChecked(ds.isGain4());
    --blockControl;

    updateControlState();
    updateTableSize(table, true);
}

void MainWindow::updateControlState()
{
    bool enableControl = (blockControl==0) && !config.curDev.isNull();
    ui->groupBoxCtrl->setEnabled(enableControl);
    const bool isRun = config.isRun();
    ui->pushButtonStart->setEnabled(enableControl && !isRun);
    ui->pushButtonStop->setEnabled(enableControl && isRun);
    ui->groupBoxFileWrite->setEnabled(enableControl && !isRun);
    ui->pushButtonChangeDevice->setEnabled((blockControl==0) && !isRun);

    calibrationDialog->enableControl(enableControl && !isRun);
}

void MainWindow::disconnectFromDev()
{
    if(tqdc_device != nullptr) {
        sendLog(tqdc_device->getDeviceInfoStr().prepend("Disconnect from:"));
        emit deleteDevice(DeviceIndex());
        if(!threadQuiting(tqdc_thread, 2000)){
            qWarning() << "Can't stop device thread";
        }
        tqdc_device = nullptr;
    }

    ui->tableWidgetDevice->item(0, COL_DEV_SERIAL)->setText("");
    ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setText("No device selected");
    updateTableSize(ui->tableWidgetDevice);
    updateControlState();

    calibrationDialog->init(DeviceCalibration());
}

void MainWindow::connectToDev(const DeviceDescription &dd)
{
    if(tqdc_device != nullptr)
        disconnectFromDev();

    config.curDev = dd.getIndex();
    calibrationDialog->init(config.getCurrentCalibration());
    ui->widgetHistMngr->updateInlCorr(config.curDev, config.getCurrentCalibration().inl);

    ui->tableWidgetDevice->item(0, COL_DEV_SERIAL)->setText(dd.getIndex().getSerialStr());
    ui->tableWidgetDevice->item(0, COL_DEV_STATUS)->setText("connecting...");
    ui->tableWidgetDevice->item(0, COL_DEV_STREAM)->setText("");
    devStatusLabel->setText("connecting...");
    devStatusLabel->setStyleSheet("QLabel {color : grey; }");
    updateTableSize(ui->tableWidgetDevice);

    updateGuiVal();

    try{
        tqdc_device = new TQDCDevice(dd.device_id, dd.serial_id);
        connect(this, &MainWindow::deleteDevice, tqdc_device, &QtMregDevice::deleteHardware);
        connect(this, &MainWindow::setupChanged, tqdc_device, &TQDCDevice::setSetup);
        connect(this, &MainWindow::setupChangedForce, tqdc_device, &TQDCDevice::writeSetupForce);
        connect(this, &MainWindow::softReset, tqdc_device, &TQDCDevice::softReset);
        connect(this, &MainWindow::blockDeviceRun, tqdc_device, &TQDCDevice::blockDeviceRun);
        connect(this, &MainWindow::startCalibration, tqdc_device, &TQDCDevice::startCalibration);
        connect(tqdc_device, &TQDCDevice::calibrateFinished, this, &MainWindow::calibrateFinished);
        connect(this, &MainWindow::setDevAddr, tqdc_device, &QtMregDevice::setHostAddress);
        connect(discoverDialog, &DiscoverDialog::deviceDiscovered, tqdc_device, &QtMregDevice::deviceUpdated);
        connect(tqdc_device, &TQDCDevice::setStatus, this, &MainWindow::setDeviceStatus);
        connect(tqdc_device, &TQDCDevice::deviceInfoUpdated, this, &MainWindow::deviceInfoUpdated);
        connect(tqdc_device, &TQDCDevice::statusUpdated, this, &MainWindow::deviceStatusUpdated);
        connect(tqdc_device, &TQDCDevice::deviceConnected, this, &MainWindow::deviceConnected);
        connect(tqdc_device, &TQDCDevice::deviceDisconnected, this, &MainWindow::deviceDisconnected);
        tqdc_device->moveToThread(tqdc_thread);
        connect(this, &MainWindow::closeAll, tqdc_device, &TQDCDevice::deleteLater);
        connect(tqdc_device, &TQDCDevice::destroyed, tqdc_thread, &QThread::quit);
        tqdc_thread->start();

        // Connecting
        QHostAddress hostAddr;
        if(dd.ip_addr)
            hostAddr.setAddress(dd.ip_addr);
        if(hostAddr.isNull()){
            if(knownDevDescr.contains(dd.getIndex()) && knownDevDescr[dd.getIndex()].ip_addr)
                hostAddr.setAddress(knownDevDescr[dd.getIndex()].ip_addr);
        }
        if(!hostAddr.isNull())
            emit setDevAddr(hostAddr);

        sendLog(QString("Selected device %1")
               .arg(config.curDev.getSerialStr()));
        emit setupChanged(config.getCurrentConfig(),
                          config.getCurrentCalibration());
    } catch (std::exception &e){
        sendLog(QString("Connect to device %1").arg(e.what()));
    }
}

void MainWindow::sendLog(QString l)
{
    emit addLog(l);
}

QString MainWindow::updateFileName(bool setTime)
{
    QString timeSubStr = config.fileTime?setTime?QDateTime::currentDateTime().toString("yyyyMMdd-HHmmss"):"<yyyyMMdd-HHmmss>":"";
    QString fileName = QString("%1%2%3")
            .arg(config.filePrefix)
            .arg(timeSubStr)
            .arg(config.fileSuffix);
    if(fileName.isEmpty())
        fileName = "tqdc";
    fileName.append('.').append(config.dataFileIsRaw?"dat":"txt");

    if(!config.fileDir.isEmpty())
        fileName.prepend(QString("%1%2")
                         .arg(config.fileDir)
                         .arg(QString(QDir::separator())));
    ui->lineEditFilePath->setText(fileName);
    return fileName;
}

void MainWindow::addMStreamDump(const DeviceIndex &index)
{
    MStreamDumpParameters mstream_parameters;
    mstream_parameters.deviceIndex = index;
    const auto dd = discoverDialog->getDescriptionByIndex(index);
    if(dd.getIndex() == index && dd.ip_addr){
        mstream_parameters.deviceAddress = QHostAddress(dd.ip_addr);
    }
    Q_ASSERT(dd.ip_addr);
    mstream_parameters.bigFragId = true;
    mstream_parameters.hwBufSize = tqdc_status.mstreamHwBufSize;
    mstream_parameters.ackSizeLimit = tqdc_status.mstreamMultiAck ? tqdc_status.mstreamHwBufSize : 1;
    mstream_parameters.preRunInit();
    mstream = new MStreamDump(mstream_parameters, false);
    connect(mstream, &MStreamDump::closing, pnpServer, &PNPServer::closeProgram);
    connect(mstream, &MStreamDump::updateProgramDescription, pnpServer, &PNPServer::updateProgramDescription);
    connect(mstream, &MStreamDump::updateProgramDescription, this, &MainWindow::programDiscovered);

    connect(mstream, &MStreamDump::stat_updated, ui->statistic_output, &StatisticOutput::stat_updated);
    connect(mstream, &MStreamDump::stat_from_receiver_updated, ui->statistic_output, &StatisticOutput::stat_from_receiver_updated);
    connect(this, &MainWindow::mstream_reset, mstream, &MStreamDump::msdump_reset);
    connect(discoverDialog, &DiscoverDialog::deviceDiscovered, mstream, &MStreamDump::receivedMldpDescr);

    tqdc_thread = new QThread(this);
    tqdc_thread->setObjectName("MStream_"+index.getSerialStr());
    connect(mstream, &MStreamDump::destroyed, tqdc_thread, &QThread::quit);
    connect(this, &MainWindow::deleteDevice, mstream, &MStreamDump::delete_dump);

    mstream->moveToThread(tqdc_thread);
    tqdc_thread->start();

    ui->statistic_output->add_device(index);
}

void MainWindow::updateTableSize(QTableWidget * const table, bool widthOnly)
{
    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

