#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QAbstractButton>
#include <QHostAddress>
#include <QProcess>
#include <QMessageBox>
#include <QMetaType>
#include <QTime>

#include "Config.h"
#include "TQDCDevice.h"
#include "TQDC2StratupOptions.h"
#include "configuration-manager/ConfigurationManager.h"
#include "mstream-lib/MStreamDump.h"

namespace Ui {
class MainWindow;
}

class QSemaphore;
class QTableWidget;
class QSemaphore;
class QTimer;
class AnalyzerStats;
class CalibrationDialog;
class DeviceData;
class DeviceDescription;
class DiscoverDialog;
class TqdcFileReader;
class TqdcFileWriter;
class LogDialog;
class TQDCAnalyzer;
class TqdcStatus;
class TcpClient;
class PNPServer;
class ProgramDescription;
class WriterStatistic;

class TqdcZmqPublisher;
class ZmqDialog;

Q_DECLARE_METATYPE(QHostAddress)

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(const TQDC2StartupOptions &startup_options, QWidget *parent = 0);
    ~MainWindow();

signals:
    void setFileNam(QString);
    void setupChanged(const DeviceSetup &, const DeviceCalibration &);
    void setupChangedForce();
    void startCalibration();
    void deleteDevice(const DeviceIndex &);
    void setDevAddr(const QHostAddress &);
    void setTcpAddress(const QHostAddress &, int port);
    void closeAll();
    void addLog(QString);
    void resetAnalyzerStats();
    void startWrite(const QString &fileName, bool rawDataWriter);
    void stopWrite();
    void matchWinUpdated(int matchWin);
    void requestRawData(uint events);
    void softReset();
    void blockDeviceRun(bool isBlock);
    void requestInlCorr(const DeviceIndex index);
    void prepareForStop(quint32 stopEvNum);
    void mstream_reset(DeviceIndex index, bool bigFragId,
                      quint16 hwBufSize, quint16 ackSize,
                      QString fn=QString(), bool doStatClear = false);
public slots:
    void onCellThresholdChange(int thr);

private slots:
    void on_pushButtonChangeDevice_clicked();
    void on_pushButtonStart_clicked();
    void on_pushButtonStop_clicked();
    void on_pushButtonBlockRun_clicked();

    void on_groupBoxFileWrite_toggled(bool en);
    void on_radioButtonRawFile_toggled(bool checked);
    void on_pushButtonChangeDir_clicked();
    void on_lineEditFilePrefix_textChanged(const QString &prefix);
    void on_checkBoxAddTime_toggled(bool checked);
    void on_lineEditFileSuffix_textChanged(const QString &suffix);

    void on_radioButtonGain1_toggled(bool checked);
    void on_radioButtonGain4_toggled(bool checked);
    void on_checkBoxAdcZs_toggled(bool checked);
    void on_checkBoxTrigTimer_toggled(bool checked);
    void on_spinBoxTimeTrig_valueChanged(int rateHz);
    void on_checkBoxTrigTtl_toggled(bool checked);
    void on_checkBoxTrigTtc_toggled(bool checked);
    void on_checkBoxSelfTrig_toggled(bool checked);
    void on_spinBoxSelfTrigDelay_valueChanged(int delay);

    void on_spinBoxThreshold_valueChanged(int thr);
    void on_spinBoxLatency_valueChanged(int lat);
    void on_spinBoxMatchWin_valueChanged(int win);
    void on_spinBoxAdcWin_valueChanged(int val);
    void on_spinBoxAdcPreWin_valueChanged(int preWin);

    void on_tabWidget_currentChanged(int index);

    void on_tableWidgetChCfg_cellChanged(int row, int column);

    void temperatureSet(QMap<int, double> temps);
    void setDeviceStatus(const QString &status);
    void deviceStatusUpdated(const TqdcStatus &status);
    void tcpClientStatusUpdated(bool connOk, double dataRate);
    void deviceConnected();
    void deviceDisconnected();
    void deviceUpdated(const DeviceDescription &dd);
    void programDiscovered(const ProgramDescription &pd);
    void showWaveform(const DeviceData &data, bool);
    void updateHitStats(const AnalyzerStats &stats);
    void rawDataUpdated(const QStringList &data);
    void deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCount);
    void writerStatisticUpdated(const WriterStatistic &s);
    void analyzerReadyForStop(quint32 evNum);
    void fileWriterReadyForStop(quint32 evNum);

    void on_actionExit_triggered();
    void on_actionCalibrationData_triggered();
    void on_actionSeparate_channel_enable_toggled(bool checked);
    void on_actionReset_MStream_triggered();
    void on_actionSend_soft_reset_triggered();
    void on_action0MQ_publisher_triggered();
    void on_actionSave_INL_corrections_triggered();
    void on_actionTxt_data_format_triggered();
    void on_actionRaw_data_format_triggered();
    void on_actionAbout_triggered();

    void calibrateFinished(const DeviceIndex &index,
                           const QMap<int, double> &adcVals, const QMap<int, double> &rms,
                           const QMap<int, double> &adcVals4, const QMap<int, double> &rms4,
                           const QMap<int, int> &dacOffset);
    void calibrationRequest();
    void clearCalibration();
    void inlCorrUpdated(const DeviceIndex &index,
                        const QMap<int,QVector<double> > &inlCorr);

    void on_pushButtonRequestRaw_clicked();
    void on_spinBoxRawEvents_valueChanged(int limit);

    void on_actionConfiguration_manger_triggered();
    void switch_to_config(QString program_index, QString configuration_name);
    void zmq_cfg_change(const ZmqConfig &cfg);
    void write_config_timer();

private:
    void read_config();
    void write_config(bool force=false);
    void initTable();
    void setTableRowsCount(int nch);
    void setupDevice(bool force=false);
    void updateGuiVal();
    void updateControlState();
    void disconnectFromDev();
    void connectToDev(const DeviceDescription &dd);
    void sendLog(QString l);
    QString updateFileName(bool setTime=false);
    void addMStreamDump(const DeviceIndex &index);

    ConfigurationManager *configuration_manager;

public:
    static void updateTableSize(QTableWidget * const table, bool widthOnly=false);

    Ui::MainWindow *ui;
    QSemaphore * const analyzerSem;// free space in analyzer queue
    QSemaphore * const writerSem;// free space in writer queue (raw or decoded)
    QSemaphore * const histSem;// free space in histogram queue
    TqdcFileReader * const reader;
    TQDCAnalyzer * const analyzer;
    QThread * const analyzerThread;
    DiscoverDialog * const discoverDialog;
    PNPServer * const pnpServer;
    Config config;
    Config bakConfig;
    QTimer *config_timer;
    bool write_request;
    TQDCDevice * tqdc_device;
    QThread * tqdc_thread;
    TqdcStatus tqdc_status;
    MStreamDump * mstream;
    QThread * mstream_thread;
    int blockControl;
    TqdcFileWriter * const fileWriter;
    QThread * const fileWriterThread;
    QString mStreamPath;
    TcpClient * const tcpClient;
    QThread * const tcpClientThread;
    QLabel * const devStatusLabel;
    QLabel * const busyStatusLabel;
    QLabel * const analyzerStatusLabel;
    QLabel * const writerStatusLabel;
    QLabel * const semStatusLabel;
    LogDialog *const logDialog;
    QMap<DeviceIndex, DeviceDescription> knownDevDescr;
    QTime lastWaveformTime;
    CalibrationDialog * const calibrationDialog;
    bool isDevBlocked;
    bool analIsReadyToStop;
    bool writerIsReadyToStop;
    bool lookForStop;
    quint32 stopEvNum;
    TqdcZmqPublisher * const tqdcZmqPublisher;
    ZmqDialog * const zmqDialog;
};

#endif // MAINWINDOW_H
