#include "TQDCAnalyzer.h"
#include <QByteArray>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QSemaphore>
#include <QTimer>
#include <iostream>
#include "mregdev/device_id.h"


namespace {
const QString HPTDC_ERR_STR[15] = {
    "Hit lost in group 0 from read-out fifo overflow",
            "Hit lost in group 0 from L1 buffer overflow",
            "Hit error have been detected in group 0",
            "Hit lost in group 1 from read-out fifo overflow",
            "Hit lost in group 1 from L1 buffer overflow",
            "Hit error have been detected in group 1",
            "Hit lost in group 2 from read-out fifo overflow",
            "Hit lost in group 2 from L1 buffer overflow",
            "Hit error have been detected in group 2",
            "Hit lost in group 3 from read-out fifo overflow",
             "Hit lost in group 3 from L1 buffer overflow",
             "Hit error have been detected in group 3",
             "Hits rejected because of programmed event size limit",
             "Event lost (trigger FIFO ovefrlow)",
             "Internal fatal chip error has been detected"
};
}
TQDCAnalyzer::TQDCAnalyzer(QSemaphore *_analyzerSem, QSemaphore *_histSem, QSemaphore *_writerSem, QObject *parent) :
    QObject(parent),
    analyzerSem(_analyzerSem),
    histSem(_histSem),
    writerSem(_writerSem),
    statTimer(new QTimer(this)),
    readTimer(new QTimer(this)),
    prevEventNum(0),
    enabled(true),
    signalInverted(false),
    writeDecodedEnabled(false),
    rawDataEvents(0),
    printBadEvent(true),
    lookForStop(false),
    stopEvNum(0)
{
    connect(statTimer, &QTimer::timeout, this, &TQDCAnalyzer::sendStat);
    connect(readTimer, &QTimer::timeout, this, &TQDCAnalyzer::read);
    statTimer->start(500);
}

void TQDCAnalyzer::dataAccuired(QByteArray newData, bool)
{
    if(data.size())
        data.append(newData);
    else
        data = newData;
    if(!skipData()){
        previousProceed=true;
        if(enabled){
            read();
        }
    }

    analyzerSem->release();
}


void TQDCAnalyzer::read()
{
    curPos = 0;
    curEventStartPos = 0;
//    qInfo()<<"read begin:"<<rdata.size();

    QElapsedTimer decodeTime;
    decodeTime.start();
    int eventDecoded=0;
    while(true) {
        try{
            if(!readEvent()) {
//                qInfo()<<"End of data";
                break;
            }
            ++eventDecoded;
            if(rawDataEvents) {
                if(--rawDataEvents == 0) {
                    emit rawDataUpdated(rawDataTxt);
                }
            }
//            if(deviceData.events.size() > 1000)
//                flushData();
        } catch(const char* message) {
            ++analyzerStats.formatErrorCnt;

            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(mess);
            } else{
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        } catch(const QString message) {
            ++analyzerStats.formatErrorCnt;
            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(message);
            } else {
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        }
    }
    analyzerStats.decodeTimeNs += decodeTime.nsecsElapsed();
    analyzerStats.eventDecoded += eventDecoded;
    analyzerStats.sizeDecoded += curEventStartPos;

    if(curEventStartPos) {
        if(curEventStartPos==data.size())
            data.clear();
        else
            data = data.right(data.size()-curEventStartPos);
    }

    flushData();
}

void TQDCAnalyzer::resetStats()
{
    analyzerStats.clear();
    lastHptdcErr.clear();
}

void TQDCAnalyzer::startWrite(QString, bool rawData)
{
    writeDecodedEnabled = !rawData;
}

void TQDCAnalyzer::stopWrite()
{
    writeDecodedEnabled = false;
}

void TQDCAnalyzer::sendStat()
{
    printBadEvent = true;
    emit deviceStatsUpdated(analyzerStats);
}

void TQDCAnalyzer::requestRawData(uint events)
{
    rawDataTxt.clear();
    rawDataEvents = events;
}

void TQDCAnalyzer::prepareForStop(quint32 evNum)
{
    if(analyzerStats.lastEvNum == evNum){
        emit gotStopEvernt(evNum);
    } else {
        lookForStop = true;
        stopEvNum = evNum;
    }
}

bool TQDCAnalyzer::findNextEventHeader()
{
    qint64 shift = 0;
    const uint dataSize = data.size();
    const char * dataStart = data.data();
    uint pos;
    while(true) {
        shift += 2;
        pos = curEventStartPos+shift;
        if(pos+sizeof(MpdEventHdr) > dataSize) {
            break;
        }
        mpdEventHdr = (MpdEventHdr *) (dataStart+pos);
        if(mpdEventHdr->sync == MPD_SYNC_NORM || mpdEventHdr->sync == MPD_SYNC_EOB) {
            curPos = curEventStartPos + shift;

            QString mess = QString("Syncword was found at 0x%1; Shift:0x%2")
                    .arg(curEventStartPos,0,16).arg(shift,0,16).toStdString().c_str();
            qInfo()<<mess;

            return true;
        }
    }
    curEventStartPos = dataSize;
    return false;
}

bool TQDCAnalyzer::readEvent()
{
    curEventStartPos = curPos;
    int bRead;
    int leftSize;
    if(data.size()<curPos+sizeof(MpdEventHdr))
        return false;

    mpdEventHdr = (MpdEventHdr *)(data.data()+curPos);

    if(mpdEventHdr->sync != MPD_SYNC_NORM) {
        if(mpdEventHdr->sync == MPD_SYNC_EOB){
            curPos += mpdEventHdr->length + sizeof(MpdEventHdr);
            return true;
        }
        QString mess = QString("Format error: wrong MPD_SYNC_WORD: %1; at file pos:%2")
                .arg(mpdEventHdr->sync, 0, 16).arg(curPos);
        throw(mess.toStdString().c_str());
    }
    if(data.size() < curPos+mpdEventHdr->length+sizeof(MpdEventHdr))
        return false;

    if(rawDataEvents) {
//        printRawTxt(sizeof(MpdEventHdr), "MpdEventHdr");
        rawDataTxt << QString("\t%1:").arg("MpdEventHdr");
        rawDataTxt << QString("%1:\t\tSyncWord").arg(*(quint32*)(data.data()+curPos), 8, 16, QChar('0'));
        rawDataTxt << QString("%1:\t\tlen:%2").arg(*(quint32*)(data.data()+curPos+4), 8, 16, QChar('0'))
                      .arg(mpdEventHdr->length);
        rawDataTxt << QString("%1:\t\tevNum:%2").arg(*(quint32*)(data.data()+curPos+8), 8, 16, QChar('0'))
                      .arg(mpdEventHdr->evNum);
    }

    curPos += sizeof(MpdEventHdr);
    leftSize = mpdEventHdr->length;

//    qInfo()<<"Event number is "<<mpdEventHdr->evNum;
    while(leftSize) {
        bRead = readDevice(leftSize);
        leftSize -= bRead;
        if(leftSize<0)
            throw("Format error: Sum of MpdDevice blocks is lager then declared in MpdEvent header");
    }
    prevEventNum = mpdEventHdr->evNum;
    lastEventNumAtPrevFrame = prevEventNum;
    analyzerStats.lastEvNum = prevEventNum;
    if(lookForStop &&  stopEvNum == prevEventNum) {
        emit gotStopEvernt(stopEvNum);
        lookForStop = false;
        stopEvNum = 0;
    }

    return true;
}

void TQDCAnalyzer::flushData()
{
    if(deviceData.events.empty())
        return;

    int retryCnt=0;
    while (!histSem->tryAcquire(1, 100)) {
        if((retryCnt++)%10==0)
            qInfo()<<"There is no free space for histogram queue";
    }
    if(writeDecodedEnabled) {
        while (!writerSem->tryAcquire(1, 100)) {
            if((retryCnt++)%10==0) {
                qInfo()<<"There is no free space for writer queue";
            }
        }
    }
    emit deviceDataUpdated(deviceData, writeDecodedEnabled);
    deviceData.clear();
}

void TQDCAnalyzer::decodeTdcBlock(const quint32 *payloadPtr)
{
    int tdcWord = 0;
    int tdcWordCnt=0;
    quint16 tdcTS;
    QString wordStr;
    QMap<int, QVector<TqdcEdge> > chEdgeMap;
    QMap<int, QString> hptdcErrors;
    while(dataBlockHdr->dataLength>tdcWord*sizeof(quint32)){
        const quint32 &word = *(payloadPtr+tdcWord);
        ++tdcWordCnt;
        const int dt = word >> 28;
//        qInfo() << tdcWord<< ":  " << hex << word << dec;
        wordStr = QString("%1:\t\t").arg(word, 8, 16, QChar('0'));
        switch (dt) {
        case 2: // tdc ev hdr
        {
            tdcTS = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF;
            if(rawDataEvents)
                wordStr.append(QString("ev hdr evNum:%1").arg(tdcEv));
//            qInfo() << "tdc ev hdr: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
        }
            break;
        case 3: // tdc ev trl
        {
            quint16 wordCnt = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF;
            if(rawDataEvents)
                wordStr.append(QString("ev trl evNum:%1; words:%2").arg(tdcEv).arg(wordCnt));
//            qInfo() << "tdc ev trl: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
            Q_ASSERT(tdcWordCnt == wordCnt);
            tdcWordCnt=0;
        }
            break;
        case 4: // tdc leading edge
//            qInfo() << tdcWord<< ":  " << hex << word << dec;
        case 5: // tdc trl edge
        {
            const int ch = (word>>21)&0xF;
            const quint32 time = word&0xFFFFF;
            const bool leading = (dt==5)==signalInverted;
            if(rawDataEvents)
                wordStr.append(QString("\t%1 ch:%2; t:%3")
                               .arg(leading?"leading":"faling").arg(ch).arg(time));
            chEdgeMap[ch].push_back(TqdcEdge(tdcTS,time, leading));
            break;
        }
        case 6:
        {
            const int tdcId = (word>>24)&0xF;
            QStringList errList;
            for(int errBit=0; errBit< 15; ++errBit){
                if((word & 0x7FFF) & 1<<errBit)
                    errList << QString("0x%1: %2").arg(1<<errBit, 0, 16).arg(HPTDC_ERR_STR[errBit]);
            }
            QString errStr = QString("Got HPTDC error (tdcId=%1) word 0x%2:\n%3")
                    .arg(tdcId).arg(word, 8, 16)
                    .arg(errList.join("\n"));
            hptdcErrors[tdcId] = errStr;
            break;
        }
        default:
            break;
        }
        if(rawDataEvents)
            rawDataTxt << wordStr;
        ++tdcWord;
    }

    // Handle HPTDC error mask
    {
        DeviceIndex index = DeviceIndex(mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);
        auto &lastErrMap = lastHptdcErr[index];
        if(lastErrMap != hptdcErrors) {
            for(int tdcId : lastErrMap.keys()) {
                if(!hptdcErrors.contains(tdcId)) {
                    qInfo() << QString("%1 has no longer error at HPTDC #%2; prev errStr:%3")
                               .arg(index.getIdent()).arg(tdcId).arg(lastErrMap[tdcId]);
                } else {
                    qWarning() << QString("%1 HPTDC #%2 change err from '%3' to '%4'")
                                  .arg(index.getIdent()).arg(tdcId)
                                  .arg(lastErrMap[tdcId]).arg(hptdcErrors[tdcId]);
                }
            }
        }
        for(int tdcId : hptdcErrors.keys()) {
            if(!lastErrMap.contains(tdcId)) {
                qWarning() << QString("%1 HPTDC #%2 got new error: '%3'")
                              .arg(index.getIdent()).arg(tdcId)
                              .arg(hptdcErrors[tdcId]);
            }
        }
        lastErrMap = hptdcErrors;
    }


//    bool signal=false;
    curTqdcEv.ch = -1;
    for (int ch : chEdgeMap.keys()) {
        for (const TqdcEdge &edge : chEdgeMap[ch]) {
            if(edge.leading) {
                if(curTqdcEv.ch != (quint8)-1){
                    curCompleteEvent.tqdcEvents[ch].push_back(curTqdcEv);
                    ++analyzerStats.tdcHits[ch];
                    curTqdcEv.clear();
                    curTqdcEv.ch = -1;
                }
                curTqdcEv.ch = ch;
                curTqdcEv.leadingTime = edge.time;
                curTqdcEv.tdcTs = edge.tdcTs;
            } else {
                if(curTqdcEv.ch == ch){
                    curTqdcEv.trailingTime = edge.time;
                    curCompleteEvent.tqdcEvents[ch].push_back(curTqdcEv);
                    ++analyzerStats.tdcHits[ch];
                    curTqdcEv.clear();
                    curTqdcEv.ch = -1;
                }
            }
        }
        if(curTqdcEv.ch != (quint8)-1) {
            curCompleteEvent.tqdcEvents[curTqdcEv.ch].push_back(curTqdcEv);
            ++analyzerStats.tdcHits[curTqdcEv.ch];
            curTqdcEv.clear();
            curTqdcEv.ch = -1;
        }
    }
}

void TQDCAnalyzer::decodeAdcBlock(const quint32 *payloadPtr)
{
    int blockWord = 0;
    int totHitLen = -1;
    int curHitLen =0;
    qint16 sample1, sample2;
    std::vector<double> &waveForm = curAdcEv.wf;
    while(dataBlockHdr->dataLength > blockWord*sizeof(quint32)) {
        const quint32 &word = *(payloadPtr+blockWord);
        ++blockWord;
        if(totHitLen==-1) {
            curAdcEv.ch = dataBlockHdr->ch;
            curAdcEv.adcTs = word & 0xFFFF;
            totHitLen = word >> 16;
            if(totHitLen%sizeof(quint16))
                throw(QString("Format error: bad size of adc data len: %1").arg(totHitLen));
            if(dataBlockHdr->dataLength < blockWord*sizeof(quint32)+totHitLen) {
                throw(QString("Format error:there is no enough data "
                              "in adc data block for hit with length %1")
                        .arg(totHitLen));
            }
            totHitLen /= sizeof(qint16);

            waveForm.reserve(totHitLen);
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\tlen:%2; ts:%3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(totHitLen).arg(curAdcEv.adcTs);
            }
            if(totHitLen==0) {
                totHitLen = -1;
            }
        } else {
//            sample1 = (qint16)(word&0xFFFF);
//            sample2 = (qint16)((word>>16)&0xFFFF);
            waveForm.push_back((qint16)(word&0xFFFF));
            ++curHitLen;
            // skip sample2 in case totHitLent is odd
            if(curHitLen != totHitLen) {
                waveForm.push_back((qint16)((word>>16)&0xFFFF));
                ++curHitLen;
            }
            if(curHitLen == totHitLen) {
                curCompleteEvent.adcEvents[curAdcEv.ch].push_back(curAdcEv);
                ++analyzerStats.adcHits[curAdcEv.ch];
                curAdcEv.clear();
                totHitLen = -1;
                curHitLen=0;
            }
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\t%2 %3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(sample2, 10).arg(sample1, 10);
            }
        }
    }
}

void TQDCAnalyzer::printRawTxt(uint size, QString header, int shift)
{
    if(!rawDataEvents) return;
    if(!header.isEmpty())
        rawDataTxt << QString("\t%1:").arg(header);
    quint32 *word = (quint32*)(data.data()+curPos+shift);
    for(uint w=0; w<(size/sizeof(quint32)); ++w){
        rawDataTxt << QString("%1").arg(*word, 8, 16, QChar('0'));
        ++word;
    }
}

int TQDCAnalyzer::readDevice(int upperLimitToDevLength)
{
    int byteRead = 0;

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr))
        throw("Format error: There is no data for MpdDeviceHdr");

    mpdDeviceHdr = (MpdDeviceHdr *) (data.data()+curPos);
    if(rawDataEvents) {
        printRawTxt(sizeof(MpdDeviceHdr), "MpdDeviceHdr");
    }

//    qInfo()<<" Device:"<<hex<<mpdDeviceHdr->deviceSerial<<dec;
    curPos += sizeof(MpdDeviceHdr);
    byteRead += sizeof(MpdDeviceHdr);

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length)
        throw("Format error: There is no data for MpdDevice payload");

    if(mpdDeviceHdr->deviceId != DEVICE_ID_TQDC16VS_E) {
        curPos += mpdDeviceHdr->length;
        byteRead += mpdDeviceHdr->length;
        return byteRead;
    }
//    qInfo()<<"mpdDevice: length:"<<mpdDeviceHdr->length<<" serial:"<<hex<<mpdDeviceHdr->deviceSerial;

    quint32 devPayloadLen=0;
    curCompleteEvent.clear();
    curCompleteEvent.evNum = mpdEventHdr->evNum;
    while(mpdDeviceHdr->length > devPayloadLen) {
        mstreamHdr = (MpdMStreamHdr *) (data.data()+curPos);
//        qInfo()<<"mstreamHdr: length:"<<mstreamHdr->words32b*sizeof(quint32);
        if(rawDataEvents) {
            printRawTxt(sizeof(MpdMStreamHdr), "MStreamHdr");
        }

        curPos += sizeof(MpdMStreamHdr);
        devPayloadLen += sizeof(MpdMStreamHdr);
        const quint32 mStreamWholeSize = mstreamHdr->words32b*sizeof(quint32);
        switch (mstreamHdr->subtype) {
        case 0:
        {
//            if(mstreamHdr->words32b*sizeof(quint32) != sizeof(AdcDataType1))
//                throw( QString("Format error: Invalid size(%1 Bytes) of MStream subtype=0 block; should be %2 Bytes.")
//                    .arg(mstreamHdr->words32b*sizeof(quint32)).arg(sizeof(AdcDataType0)).toStdString().c_str());

            tsTAI = (AdcDataType1 *) (data.data()+curPos);
            curCompleteEvent.taiFlags = tsTAI->taiFlags;
            curCompleteEvent.taiSec = tsTAI->wfTsHi;
            curCompleteEvent.taiNSec = tsTAI->wfTsLow;
            quint32 blockLen = sizeof(AdcDataType1);
            if(rawDataEvents) {
//                printRawTxt(sizeof(AdcDataType1), "TAI");
                rawDataTxt << QString("\t%1:").arg("TAI");
                rawDataTxt << QString("%1:\t\t%2").arg(*(quint32*)(data.data()+curPos), 8, 16, QChar('0'))
                              .arg(QDateTime::fromTime_t(tsTAI->wfTsHi).toString());
                rawDataTxt << QString("%1:\t\tflag:%2; ns:%3").arg(*(quint32*)(data.data()+curPos+4), 8, 16, QChar('0'))
                              .arg(tsTAI->taiFlags).arg(tsTAI->wfTsLow);
            }

            while(mStreamWholeSize>blockLen) {
                dataBlockHdr = (DataBlockHeader *) (data.data()+curPos+blockLen);
                if(blockLen+sizeof(DataBlockHeader)+dataBlockHdr->dataLength > mStreamWholeSize)
                    throw(QString("Format error: TQDC Data Block length (%1) is greater than left size:"
                                  "MStream-parsed-DataBlockHeader=%2-%3-%4")
                        .arg(dataBlockHdr->dataLength).arg(mStreamWholeSize).arg(blockLen).arg(sizeof(DataBlockHeader)));
                blockLen += sizeof(DataBlockHeader);

                const quint32* payloadPtr = (const quint32*) (data.data()+curPos+blockLen);
                switch (dataBlockHdr->type) {
                case 0:
                    if(rawDataEvents) {
//                        rawDataTxt << QString("\t%1:").arg("TDC Block");
                        rawDataTxt << QString("%1:\t    TDC Block len:%2")
                                      .arg(*(quint32*)(data.data()+curPos+blockLen-sizeof(DataBlockHeader)), 8, 16, QChar('0'))
                                      .arg(dataBlockHdr->dataLength);
                    }
                    decodeTdcBlock(payloadPtr);
                    break;
                case 1:
                    if(rawDataEvents) {
//                        printRawTxt(dataBlockHdr->dataLength, "Data type: ADC", blockLen);
//                        rawDataTxt << QString("\t%1:").arg("ADC Block");
                        rawDataTxt << QString("%1:\t    ADC Block len:%2; ch:%3")
                                      .arg(*(quint32*)(data.data()+curPos+blockLen-sizeof(DataBlockHeader)), 8, 16, QChar('0'))
                                      .arg(dataBlockHdr->dataLength).arg(dataBlockHdr->ch);
                    }
                    decodeAdcBlock(payloadPtr);
                    break;
                default:
                    if(rawDataEvents) {
                        printRawTxt(dataBlockHdr->dataLength+sizeof(DataBlockHeader),
                                    QString("Unknown data block:%1").arg(dataBlockHdr->type),
                                    blockLen-sizeof(DataBlockHeader));
                    }

                    throw(QString("Format error: unknown TQQDC Data type: %1").arg(dataBlockHdr->type));
                    break;
                }
                blockLen += dataBlockHdr->dataLength;
            }

//            qInfo()<<"Sub0: chMask:"<<hex<<adcChMask;
            break;
        }
        default:
            if(rawDataEvents) {
                printRawTxt(mstreamHdr->words32b*sizeof(quint32), "MStreamPayload");
            }
            throw( QString("Format error: Invalid subtype number in MStream header: %1.")
                .arg(mstreamHdr->subtype).toStdString().c_str());
            break;
        }

        curPos += mstreamHdr->words32b*sizeof(quint32);
//        qInfo()<<"MS block:ev="<<mpdEventHdr->evNum<<"; subtype="<<mstreamHdr->subtype<<"; curpos="<<curpos;
        devPayloadLen += mstreamHdr->words32b*sizeof(quint32);
    }
    if(mpdDeviceHdr->length != devPayloadLen)
        throw(QString("Format error: Sum of sizes of MStream blocks(%1) isn't equal to one declared in MpdDevice header(%2)")
            .arg(devPayloadLen).arg(mpdDeviceHdr->length).toStdString().c_str());
    deviceData.events.push_back(curCompleteEvent);
    deviceData.index = DeviceIndex(mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);

    byteRead += devPayloadLen;
    return byteRead;

    /*

    vmeData = (VMEDataTypes *) (rdata.data()+curpos);

   // qInfo()<<"VME Word = "<<(vmeData->vmeVol)::hex;


    unsigned int vol = vmeData->vmeVol;
//    qInfo()<<"VME Word = "<<hex<<vol;
    unsigned int vmeType = (vol & 0xF0000000)>>28;
//    qInfo()<<"VME type = "<<vmeType;

    switch (vmeType) {
    case 13: { // Spill trailer
        curpos += 4;
        breaded += 4;
        step += 4;
        break;}
    case 12: { // Spill header
        curpos += 4;
        breaded += 4;
        step += 4;
        break;}
    case 11: { // Event trailer
        curpos += 4;
        breaded += 4;
        step += 4;
        evHdr = false;
        evTrl = true;

        break;}
    case 10: { // Event header
        unsigned int evNum = (vol & 0xFFFFF)>>0; //Mask for event number
//        qInfo()<<"Event Number is "<<evNum;
        curpos += 4;
        breaded += 4;
        step += 4;

        evHdr = true;
        evTrl = false;

        break;}
    case 8: { // Module header
        devID = (vol & 0x7F0000)>>16;
//        qInfo()<<"Device ID is "<<devID;
        curpos += 4;
        breaded += 4;
        step += 4;
        modHdr = true;

        break;}
    case 15: { // Padding
        curpos += 4;
        breaded += 4;
        step += 4;
        break;}
    case 14: { // Status
        curpos += 4;
        breaded += 4;
        step += 4;
        break;}
    default: { // ...

        curpos += 4;
        breaded += 4;
        step += 4;
        break;}
    }

    if(devID == 9) {

    while(modHdr) {
        vmeData = (VMEDataTypes *) (rdata.data()+curpos);
        vol = vmeData->vmeVol;
        vmeType = (vol>>28) & 0xF;


//        qInfo()<<"VME Word = "<<hex<<vol;
//        qInfo()<<"VME type = "<<vmeType;
//        qInfo()<<"VME Word = "<<hex<<vol<<dec<<"\tVME type = "<<vmeType;

        if(adcTsWordCnt==2 && vmeType != 5){
            // flush ADC data
            curAdcEv.evNum = mpdEventHdr->evNum;
            curAdcEv.adcTs = adcTmstmp-trigTmstmp;
            adcTsWordCnt = 0;
            deviceData.adcEvents.push_back(curAdcEv);
            curAdcEv.clear();
        }
        if(vmeType)
        switch (vmeType) {

        case 9: { // Module trailer

            curpos += 4;
            breaded += 4;
            step += 4;
            modHdr = false;
            devID = -1;

            break;}
        case 7: { // ?!?
            curpos += 4;
            breaded += 4;
            step += 4;
            break;}
        case 6: { // TDC error
            u_int16_t err = vol & 0x7FFF;
            if(err){
                deviceData.errors.push_back(err);
            }
            curpos += 4;
            breaded += 4;
            step += 4;
            break;}
        case 2: { // TDC event header
            TDCevHdr = true;
//            quint16 tdcTime = vol & 0xFFF;
            curpos += 4;
            breaded += 4;
            step += 4;

            break;}
        case 3: { // TDC event trailer
            Q_ASSERT(TDCevHdr && (adcTsWordCnt==0));
            TDCevHdr = false;
            adcTsWordCnt = 0;

            curpos += 4;
            breaded += 4;
            step += 4;

            break;}
        case 4: { // ADC timestamp and TDC data
            int mode = (vol>>26) & 0x3;
//            qInfo()<<"mode = "<<mode4;

            chanNum = (vol>>19) & 0x1F;

            if(TDCevHdr) {
                Q_ASSERT(mode==0);
//                qInfo()<<"TDC Channel";
//                qInfo()<<"Channel number = "<<chanNum;
                // leading or combined mode? TODO
                tdcData = vol & 0x7FFFF;// mode == 0 TDC data
                curTqdcEv.evNum = mpdEventHdr->evNum;
                curTqdcEv.ch = chanNum;
                curTqdcEv.leadingTime = tdcData;
                deviceData.tqdcEvents.push_back(curTqdcEv);
            } else {
                // ADC event timestamp
                if((vol>>16) & 0x1) {
                    curAdcEv.mode = (vol>>26) & 0x3;
                    adcTmstmp = vol & 0xFFFF; // ADC timestamp
                } else {
                    trigTmstmp = vol & 0xFFFF; // Trigger timestamp
                }
                //                 qInfo()<<"TDC word = "<<tdcWrd+1;
                adcTsWordCnt +=1; //For Type 5
            }


            curpos += 4;
            breaded += 4;
            step += 4;

            break;}
        case 5: { // TDC and ADC data
            int mode = (vol >> 26) & 0x3;
            chanNum = (vol >> 19) & 0x1F;
            wordData = vol & 0x7FFFF; // TDC or ADC data

            if(mode)
                Q_ASSERT(adcTsWordCnt == 2);
//                qInfo()<<"ADC Channel";
//                qInfo()<<"Channel number = "<<chanNum;
            switch (mode) {
            case 0: {
                //TDC trailing time measurement
                curTqdcEv.ch = chanNum;
                curTqdcEv.trailingTime = wordData;
                //and rcdata must be here!
//                qInfo()<<"TDC time measurement";
                break;}
            case 1: {
                //adc sample in calibration mode
                curAdcEv.wf.push_back(wordData & 0xFFFF);
//                qInfo()<<"calibration";
                break;}
            case 2: {
                //adc sample
                curAdcEv.wf.push_back(wordData & 0xFFFF);
                //                    qInfo()<<"sampling";
                break;}
            case 3: {
                // adc integration
                curAdcEv.sumQ = wordData;
//                qInfo()<<"Sum of adc samples = "<<wordData;
                break;}
            default:
                break;
            }

            curpos += 4;
            breaded += 4;
            step += 4;

            break;}

        default: {
            curpos += 4;
            breaded += 4;
            step += 4;
            break;}
        }
    }
    }
}
    return breaded;
*/
}

//quint32 TQDCAnalyzer::VmeChnlMap()
//{
//    QMap<QPair<quint32, quint8>, TQDCData> allData;
//}

void TQDCAnalyzer::sendLog(QString s) const
{
    emit addLog(s);
}

bool TQDCAnalyzer::skipData()
{
    return false;
    int freeSpace=analyzerSem->available();
    if(freeSpace>30)
        return false;
    int s;
    if(freeSpace<5)
        s=99;
    else
        s=100-freeSpace*3;

    const int rand=qrand()%100;
    bool res=rand<s;
    if(previousProceed)
        qInfo()<<"freeSpace="<<freeSpace<<"  RandNumber="<<rand<<" threshold="<<s<<(res?" - skipp":" - proceed");

    return res;

}
