#include "TQDCDevice.h"
#include "mregdev/qtmregdevice.h"

#include <QDateTime>
#include <QFile>
#include <QTextStream>
#include <QVector>
#include <math.h>

#include "mregdev/mregdevice.h"
#include "regio/regop.h"
#include <TqdcRegs.h>


namespace {
const int tqdc16vs_adc_scale = 3;
const int tqdc_hptdc_scale = 1;
const int tqdc16vs_adc_lat_offset = 1; // in 8ns steps
const int tqdc16vs_hptdc_lat_offset = 5; // in 24ns steps
int hptdcSetupIndex = 0;
const int CALIBRATE_LIMIT = 10e3;

const bool ADC_12_BIT_MODE = false;
const double DAC_V_REF = 3.3 * 82.5 / (82.5 + 130); // in V (1.2812)
const quint16 DAC_RANGE = (1<<12) - 1;
const int DAC_DEFAULT_ZERO = 1920;
constexpr fw_version_t REV_MSTREAM_CFG = fw_version_t(2, 0, 25502);
constexpr fw_version_t REV_MAX_SAMPLES_CFG =fw_version_t(2, 0, 24223);

const int TQDC_CH_TO_DAC_CH_MAP[16] = {
    3,2,1,0, 4,5,6,7, 0,1,2,3, 7,6,5,4
};
const int tqdc_map_chtdc2inp[2][32] = {
    {15,-1,-1,-1,14,-1,-1,-1,13,-1,-1,-1,12,-1,-1,-1,11,-1,-1,-1,10,-1,-1,-1, 9,-1,-1,-1, 8,-1,-1,-1 },
    { 7,-1,-1,-1, 6,-1,-1,-1, 5,-1,-1,-1, 4,-1,-1,-1, 3,-1,-1,-1, 2,-1,-1,-1, 1,-1,-1,-1, 0,-1,-1,-1 }
};

const QString HPTDC_ST_STR[16] = {
    "Vernier Error",
    "Coarse Error",
    "Channel Select Error",
    "L1 Buffer Parity Error",
    "Trigger FIFO Parity Error",
    "Trigger Matching State Error",
    "Readout FIFO Parity Error",
    "Readout State Error",
    "Setup Parity Error",
    "Control Parity Error",
    "JTAG Instruction Error",
    "TDC Have Readout Token",
    "Readout FIFO Full",
    "Readout FIFO Empty",
    "Trigger FIFO Full",
    "Trigger FIFO Empty"
};
const quint16 HPTDC_ERR_MASK = 0x7FF;

// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}}

using mlink::RegOpVector;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;


TQDCDevice::TQDCDevice(quint16 devId, quint64 devSerial):
    QtMregDevice(devId, devSerial),
    setupCacheValid(false),
    hitCntTime(0),
    isBlock(false)
{
    qRegisterMetaType<DeviceSetup>();
}

void TQDCDevice::writeSetup()
{
    status.index = setup.index;
    if (!isConnected()) return;
    try {
        StopRun();
        if(!waitEndOfRun()) {
            setupCacheValid = false;
            if(isConnected()) {
                qWarning("Failed wait device pause (run bit disabled; ev queue is empty). Writing setup is failed.");
            } else {
                return;
            }
        }


        RegOpVector regOp;

        if (!setupCacheValid || setupCache.trigsetup.all != setup.trigsetup.all)
            regOp << WriteTrigSetup(setup.trigsetup);

        if (!setupCacheValid || setupCache.trigPeriod != setup.trigPeriod) {
            regOp << RegWrite32(REG32_TRIG_TIMER_PERIOD, setup.trigPeriod);
        }
        if (!setupCacheValid || setupCache.selfTrigDelay != setup.selfTrigDelay) {
            regOp << RegWrite(REG_TRIG_SELFTRIG_DELAY, setup.selfTrigDelay);
        }

        quint16 chEnMaskAdc = 0;
        quint16 chTrigMask = 0;
        quint32 chEnMaskTdc1 = 0, chEnMaskTdc2 = 0;
        quint16 gainCtrl = 0;
        quint16 chInvert = 0;
        for(int ch=0; ch<setup.getNCh(); ch++) {
            if(setup.chTdcEn.value(ch, false)) {
                if(ch/8) {
                    chEnMaskTdc1 |= 0xF << ((7-ch%8)*4);
                } else {
                    chEnMaskTdc2 |= 0xF << ((7-ch%8)*4);
                }
            }
            if(setup.chAdcEn.value(ch, false))
                chEnMaskAdc |= (quint16)1 << ch;
            if(setup.chTrigEn.value(ch, false)) {
                chTrigMask |= (quint16)1 << ch;
            }
            if(ch%2==0){
                if(setup.chGain.value(ch, false))
                    gainCtrl |= (quint16)1 << (ch/2);

            } else {
                if(setup.chGain.value(ch, false) != setup.chGain.value(ch-1, false))
                    qWarning() << "Chennal gain for ch " << ch-1
                               << " and for ch " << ch << " are differ";
            }
            if(setup.chInv.value(ch, false))
                chInvert |= (quint16)1 << ch;
        }

        regOp << RegWrite(REG_SELF_TRIG_MASK, chTrigMask);
        regOp << RegWrite(REG_ADC_CH_EN, chEnMaskAdc);
        regOp << RegWrite(REG_ADC_GAIN_CRTL, gainCtrl);
        regOp << RegWrite(REG_ADC_MODE, setup.zsEn ? REG_ADC_MODE_NORM : REG_ADC_MODE_BASELINE);

        regOp << RegWrite32(REG32_TDC_GEN_CTRL, REG32_TDC_GEN_CTRL_BIT_STANDBY);
        regOp << RegWrite32(REG32_TDC_CH_EN_1, chEnMaskTdc1);
        regOp << RegWrite32(REG32_TDC_CH_EN_2, chEnMaskTdc2);
        if (!setupCacheValid || setupCache.latency != setup.latency
                || setupCache.matchWin != setup.matchWin) {
            quint32 mW = tqdc_hptdc_scale*setup.matchWin;
            quint32 lat = tqdc16vs_hptdc_lat_offset + tqdc_hptdc_scale*setup.latency;
            regOp << RegWrite32(REG32_TDC_TRIG_WIN_SETUP,
                                (mW&0xFFF)<<12 | (lat & 0xFFF));
        }

        regOp << RegWrite32(REG32_TDC_GEN_CTRL, 0);

        if (!setupCacheValid || setupCache.chThr != setup.chThr
                 || calibrationCache.thrOffset != calibration.thrOffset
                 || setupCache.chInv != setup.chInv) {
            regOp << setChThreshold(setup.chThr, calibration.thrOffset, setup.chInv);
        }
        if (!setupCacheValid || setupCache.chInv != setup.chInv) {
            regOp << RegWrite(REG_ADC_DAC_INVERT, chInvert);
        }


        if (!setupCacheValid || setupCache.latency != setup.latency)
            regOp << setAdcLatency(setup.latency);

        if (!setupCacheValid || setupCache.matchWin != setup.matchWin)
            regOp << RegWrite(REG_ADC_MATCH_WIN, tqdc16vs_adc_scale*setup.matchWin);
        if (!setupCacheValid || setupCache.adcPreWin != setup.adcPreWin)
            regOp << RegWrite(REG_ADC_PREWIN, setup.adcPreWin);
        if (!setupCacheValid || setupCache.readCellNumber != setup.readCellNumber)
            regOp << RegWrite(REG_ADC_WIN, setup.readCellNumber);

        regOpExec(regOp);

        if(setup.run) {
            if(!setupCacheValid || !setupCache.run)
                resetEvNumAndStats();
            StartRun();
        }

//        readoutHptdcSetup();

        writeLiveMagic();
        setupCache = setup;
        calibrationCache = calibration;
        setupCacheValid = true;
    }
    catch (std::runtime_error &e) {
        setupCacheValid = false;
        qCritical() << QString("Communication error: %1").arg(e.what());
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }

    writeLiveMagic();
    clearStatistics();
}

void TQDCDevice::setSetup(const DeviceSetup &s, const DeviceCalibration &cal)
{
    setup = s;
    calibration = cal;
    writeSetup();
}

void TQDCDevice::writeSetupForce()
{
    StopRun();
    setupCacheValid = false;
    on_deviceConnected();
//    writeSetup();
}

void TQDCDevice::startCalibration()
{
    QElapsedTimer t;
    t.start();
    QMap<int, double> mean, mean4, rms, rms4;

    regWrite(REG_ADC_GAIN_CRTL, 0); // all ch gain x1
    calibrateBaseline(mean, rms);
    regWrite(REG_ADC_GAIN_CRTL, 0xFF); // all ch gain x4
    calibrateBaseline(mean4, rms4);

    qInfo() << "calibrateBaseline:"<<t.restart()<<"ms";
//    qInfo() << "Calibration result is:" << calReult.join(' ');
    QMap<int, int> dacOffset = calibrateDac();
    qInfo() << "calibrateDac:"<<t.elapsed()<<"ms";
    setupCacheValid = false;
    emit calibrateFinished(getDeviceIndex(),
                           mean, rms,
                           mean4, rms4,
                           dacOffset);
    writeSetup();
}

void TQDCDevice::softReset()
{
    if(!isConnected())
        return;
    setupCacheValid = false;
    RegOpVector op;
    op << RegWrite(REG_DEVICE_CTRL, 1);
    op << RegWrite(REG_DEVICE_CTRL, 0);
    regOpExec(op);
}

void TQDCDevice::blockDeviceRun(bool _isBlock)
{
    isBlock = _isBlock;
    if(isBlock) {
        if(isConnected())
            StopRun();
    } else
        writeSetup();
}

void TQDCDevice::statusPoll()
{
    QtMregDevice::statusPoll();
    if(isConnected() && !setupCacheValid)
        writeSetup();

    status.temp.clear();
    status.temp[0] = getTemperature();
    status.connected = isConnected();

    RegOpVector v;
    int regPos=0;
    v << RegWrite(REG_TRIG_CSR, REG_TRIG_CSR_BIT_CNT_LOCK); regPos+=1;
    int trigQueuePos  = regPos;
    v << RegRead(REG_TRIG_QUEUE); regPos+=1;
    int runStatusPos = regPos;
    v << RegRead(REG_DEVICE_CTRL); regPos+=1;
    const int busyStatusPos = regPos;
    v << RegRead(REG_TRIG_BUSY_STATUS); regPos+=1;
    const int evAllPos = regPos;
    v << RegRead64(REG64_TRIG_CNT_ALL); regPos+=4;
    const int evCandPos = regPos;
    v << RegRead64(REG64_TRIG_CNT_CANDID); regPos+=4;
    const int evAcceptedPos = regPos;
    v << RegRead64(REG64_TRIG_CNT_ACCEPTED); regPos+=4;
    v << RegWrite(REG_TRIG_CSR, 0); regPos+=1;
    const int pllUnlockPos = regPos;
    v << RegRead(REG_PLL_UNLOCK_CNT); regPos+=1;
    const int pllStatusPos = regPos;
    v << RegRead(REG_PLL_STATUS); regPos+=1;
    const int taiPos = regPos;
    v << RegRead32(REG_TAI_0); regPos+=2;
    const int desStatusPos = regPos;
    v << RegRead(REG_DES_STATUS); regPos+=1;

    const int bmcVerPos = regPos;
    v << RegRead(REG_BMC_VERSION); regPos+=1;
    v << RegRead(REG_BMC_REVERSION); regPos+=1;
    const int bmcTempPos = regPos;
    v << RegRead(REG_BMC_TEMP_1); regPos+=1;
    v << RegRead(REG_BMC_TEMP_2); regPos+=1;
    v << RegRead(REG_BMC_TEMP_3); regPos+=1;
    v << RegRead(REG_BMC_TEMP_4); regPos+=1;

    RegOpVector regRes = regOpExec(v);
    if(regRes.size() != v.size())
        return;

    v.clear();
    regPos = 0;
    v << RegWrite(REG_HIT_CNT_CTRL, 1); regPos+=1;
    const int hitCntPos = regPos;
    for(int ch=0; ch<getNumberOfChannels(); ++ch) {
        v << RegRead32(REG32_HIT_CNT_CH1+ch*2); regPos+=2;
    }
    const int hitTimePos = regPos;
    v << RegRead64(REG64_HIT_CNT_TIME); regPos+=4;
    v << RegWrite(REG_HIT_CNT_CTRL, 0); regPos+=1;
    RegOpVector regResHit = regOpExec(v);
    if(regResHit.size() != v.size())
        return;

    status.runFinished = (regRes.at(runStatusPos).data & REG_DEVICE_CTRL_BIT_RUN) == 0;
    status.runFinished &= (regRes.at(trigQueuePos).data&0xf) == 0;

    status.busyStatus = regRes.at(busyStatusPos).data;
    status.statistic.evNumAll = get64(regRes, evAllPos);
    status.statistic.evNumCandidateRun = get64(regRes, evCandPos);
    status.statistic.evNumAccepted = get64(regRes, evAcceptedPos);

    status.statistic.pllUnlockCnt = regRes[pllUnlockPos].data & 0xFFF;
    status.pllStatus = regRes[pllStatusPos].data & 0xFFF;

    status.adcStatus = regRes[desStatusPos].data;
    status.adcSelfTestOk = isDesStatusOk(status.adcStatus);
    status.wrTime = get32(regRes, taiPos);

    status.mcuVer = regRes[bmcVerPos].data;
    status.mcuRev = regRes[bmcVerPos+1].data;
    if(status.mcuVer) {
        for(int i=0; i<4; ++i) {
            quint16 &rawTemp = regRes[bmcTempPos+i].data;
            if(rawTemp != 0xFFFF)
                status.temp[1+i] = decodeTemp_v1(rawTemp);
        }
    }

//    status.statistic.timeValidLoss = regRead32(REG_WR_SYNC_LOST_COUNTER);
//    status.statistic.rxCodeError = regRead32(REG_WR_LINK_ERROR_COUNTER);
//    status.statistic.missExtTrig = regRead32(REG_TRIG_ON_XOFF_ERROR_COUNTER);

    const quint64 hitTime = get64(regResHit, hitTimePos);
    const quint64 hitTimeShift = hitTime - hitCntTime;
    const double CLOCK_RATE = 125e6;
    const double HIT_RATE_UPDATE_PERIOD = CLOCK_RATE; // 1 sec

    if(hitTimeShift>HIT_RATE_UPDATE_PERIOD) {
        hitCntTime = hitTime;
        for(int ch=0; ch<getNumberOfChannels(); ++ch) {
            const quint32 chCnt = get32(regResHit, hitCntPos+ch*2);
            const quint32 chCntShift = chCnt-hitCnt[ch];
            hitCnt[ch] = chCnt;
            status.hitRate[ch] = CLOCK_RATE*chCntShift/hitTimeShift;
        }
    } else if(hitTimeShift==0) {
        status.hitRate.clear();
    }

    RegOpVector htpdcRegOp;
    for (int hptdcId = 0; hptdcId < 2; ++hptdcId) {
        for(int reg=0; reg<4; ++reg) {
            htpdcRegOp << RegRead(REG_HPTDC1_STATUS_1+(hptdcId*4)+reg);
        }
    }

    RegOpVector hptdcRes = regOpExec(htpdcRegOp);
    for (int hptdcId = 0; hptdcId < 2; ++hptdcId) {
        for(int reg=0; reg<4; ++reg) {
            status.hptdcStatus[hptdcId].regValues[reg] = hptdcRes[hptdcId*4+reg].data;
        }
    }

    emit statusUpdated(status);
}

void TQDCDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
    status.fwStr = getFirmwareVersionStr();
    // Before this rev. there was no default value for MLink CSR registers
    if(getDeviceFwVersion() > REV_MSTREAM_CFG) {
        status.mstreamMultiAck = hasMultiAckSupport();
        status.mstreamHwBufSize = getHwBufSize();
    }
    detectChannelMaxSampleCount();

    setupCacheValid = false;
    writeClockDataDelay();
    setupAdc();
    resetDac();
    writeTime();

    regWrite32(REG32_TDC_GEN_CTRL, REG32_TDC_GEN_CTRL_BIT_STANDBY);
    quint32 measureCtrl = regRead32(REG32_TDC_MEASURE_CTRL);

    {
        measureCtrl &= ~((quint32)0xF<<3);
        measureCtrl |= 0xA<<3; // width_select
    }
    measureCtrl |= 3<<8; // mode_rc mode_rc_compression
    measureCtrl &= ~((quint32)1<<13); // disable trailing
//    measureCtrl |= 1<<9; // mode_rc
    regWrite32(REG32_TDC_MEASURE_CTRL, measureCtrl);
    regWrite32(REG32_TDC_GEN_CTRL, 0);

    writeSetup();

    emit deviceInfoUpdated(getDeviceIndex(), getNumberOfChannels(), maxSampleCount);
}

void TQDCDevice::on_deviceDisconnected()
{
    QtMregDevice::on_deviceDisconnected();
    setupCacheValid = false;
}

RegOpVector TQDCDevice::WriteTrigSetup(reg_trig_ctrl trigsetup)
{
    RegOpVector regOp;
    regOp << RegWrite(REG_TRIG_SRC, trigsetup.all);
    return regOp;
}

mlink::RegOpVector TQDCDevice::setChThreshold(QMap<int, int> chThr_mV, QMap<int, int> thrOffset, QMap<int, bool> chInv)
{
    QMap<int, int> chThr;
//    qInfo() << "Set threshold:";
    for(int ch : chThr_mV.keys()) {
        int dacOffset = thrOffset.value(ch, DAC_DEFAULT_ZERO);
        int mV = chThr_mV[ch];
        if(chInv.value(ch, false))
            mV *= -1;
        chThr[ch] = dacValue(mV, dacOffset);
//        qInfo() << QString("%1: %2 (%3 mv)")
//                   .arg(ch).arg(chThr[ch]).arg(mV);
    }
    return setChThreshold(chThr);
}

RegOpVector TQDCDevice::setChThreshold(QMap<int, int> chThr)
{
    RegOpVector regOp;

    for(auto ch : chThr.keys()) {
        if(ch > static_cast<int>(sizeof(TQDC_CH_TO_DAC_CH_MAP)/sizeof(int))) {
            qWarning() << "ch out of range";
            continue;
        }
        quint16 dacData = chThr[ch] & 0xFFF;
        dacData |= TQDC_CH_TO_DAC_CH_MAP[ch]<<12;
        regOp << RegWrite(ch/8 ? REG_DAC2_CTRL : REG_DAC1_CTRL,
                                 dacData);
//        qInfo() << hex << (ch/8 ? REG_DAC2_CTRL : REG_DAC1_CTRL) << dacData << dec;
    }
    regOpExec(regOp);
    regOp.clear();
    regOp << RegWrite(REG_DAC1_CTRL, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC2_CTRL, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC1_CTRL, DAC_CMD_SINGLE_UPDATE);
    regOp << RegWrite(REG_DAC2_CTRL, DAC_CMD_SINGLE_UPDATE);

    return regOp;
}

RegOpVector TQDCDevice::setAdcLatency(int latency)
{
    RegOpVector regOp;
    regOp << RegWrite(REG_ADC_LAT, tqdc16vs_adc_lat_offset + tqdc16vs_adc_scale*latency);  //Set number of points saved after incoming trigger
//    RegOpVector result = regOpExec(regOp);
//    if(result.size() != regOp.size())
//        qWarning() << "regOpExec error in SetTrigDelay";
    return regOp;
}

void TQDCDevice::StartRun()
{
    if(isBlock)
        return;
    RegOpVector v;
//    v << RegWrite(REG_DEVICE_CTRL, 0); // clear RUN
    v << RegWrite(REG_DEVICE_CTRL, REG_DEVICE_CTRL_BIT_RUN); // set RUN
    regOpExec(v);
}

void TQDCDevice::StopRun()
{
    regWrite(REG_DEVICE_CTRL, 0); // clear RUN
}

bool TQDCDevice::isDesStatusOk(const quint16 &adc_status)
{
    bool lock_ok = true;
    for(int n=0; n<getDeserializeNum(); ++n)
        if(((adc_status>>(2*n)) & 3) != 1)
            lock_ok = false;

    return lock_ok;
}

int TQDCDevice::getNumberOfChannels() const
{
    int nch;
    switch (getDeviceId()) {
    case DEVICE_ID_TQDC16VS_E: nch = 16; break;
    case DEVICE_ID_ADC64: nch =  64; break;
    case DEVICE_ID_ADC32P: nch =  32; break;
    case DEVICE_ID_ADC64S2: nch =  64; break;
    case DEVICE_ID_ADC64WR: nch =  64; break;
    case DEVICE_ID_ADC64VE: nch =  64; break;
    case DEVICE_ID_ADC8E: nch = 8; break;
    case DEVICE_ID_8CADC: nch = 8; break;
    case DEVICE_ID_ADCM16: nch = 16; break;
    case DEVICE_ID_ADC16LTCM: nch = 16; break;
//    case DEVICE_ID_ADC16LTCS: nch = 16; break;
    case DEVICE_ID_ADCM163: nch = 16; break;
    case DEVICE_ID_BPM4100: nch = 4; break;
    case DEVICE_ID_EVADC: nch = 8; break;
    default: nch = 1; break;
    }
    return nch;
}

void TQDCDevice::detectChannelMaxSampleCount()
{
    maxSampleCount = 0;
    if(isConnected() && getDeviceFwVersion() >= REV_MAX_SAMPLES_CFG) {
        quint32 s = regRead(REG_CH_DPM_KS);
        maxSampleCount = s<<10;
    }
    if(maxSampleCount==0) {
        switch (getDeviceId()) {
        case DEVICE_ID_TQDC16VS_E:
            maxSampleCount = 10000/8; // 10us
            break;
        case DEVICE_ID_ADC64:
            maxSampleCount = 0x0800; // 2k
            break;
        case DEVICE_ID_ADC64S2:
            maxSampleCount = 0x2000; // 8k
            break;
        case DEVICE_ID_ADC64WR:
            maxSampleCount = 0x2000; // 8k
            break;
        case DEVICE_ID_ADC64VE:
            maxSampleCount = 0x1000; // 4k
            break;
        case DEVICE_ID_ADC8E:
            maxSampleCount = 0x2000; // 8k
            break;
        case DEVICE_ID_ADCM163:
            maxSampleCount = 0x1000; // 4k
            break;
        default:
            maxSampleCount = 0;
            break;
        }
    }
}

void TQDCDevice::setupAdc()
{
    // set deserializers to reset
    regWrite(REG_DES_CTRL, ADC_CLK_CTRL_BIT_ADC_RESET);

    // Double loop for SPI1 and SPI2
    for(int i=0;i<2;i++){
        const int writeReg = REG_ADC0_SPI_WRITE + i*8;
        const int adrReg = REG_ADC0_SPI_ADDR + i*8;

        regWrite(writeReg, 1<<7); // reset bit
        regWrite(adrReg, LTM9011_ADC_SPI_A0_REG);

//        regWrite(writeReg, 0x40); // randomizer
        regWrite(writeReg, 0x20); // 2's complement
        regWrite(adrReg, LTM9011_ADC_SPI_A1_REG);

        if(ADC_12_BIT_MODE) {
            regWrite(writeReg, 0x2); // 2-lane 12-bit ser
            regWrite(adrReg, LTM9011_ADC_SPI_A2_REG);
        } else {
            regWrite(writeReg, 0x0); // 2-lane 16-bit ser
            regWrite(adrReg, LTM9011_ADC_SPI_A2_REG);
        }

//        regWrite(writeReg, 0x00);
//        regWrite(adrReg, LTM9011_ADC_SPI_A4_REG);
//        regWrite(writeReg, 0xbf);
//        regWrite(adrReg, LTM9011_ADC_SPI_A3_REG);
    }
//    setCustomPattern(0x01ff);

    // set deserializers to normal work
    regWrite(REG_DES_CTRL, 0);

//    ResetAdc();
}

void TQDCDevice::resetDac()
{
    RegOpVector regOp;
    regOp << RegWrite(REG_DAC1_CTRL, DAC_CMD_RESET); // DAC Data and control reset
    regOp << RegWrite(REG_DAC2_CTRL, DAC_CMD_RESET); // DAC Data and control reset
    regOpExec(regOp);

//    regWrite(REG_DAC1_CTRL, 0x800C); // No gain, buf input, V Ref
//    regWrite(REG_DAC2_CTRL, 0x800C); // 0x800C
}

void TQDCDevice::writeTime()
{
    uint sec = QDateTime::currentDateTime().toTime_t();
    RegOpVector regOp;
    regOp << RegWrite32(REG_TAI_0, sec);
    regOp << RegWrite(REG_TAI_UPDATE, 0);
    regOpExec(regOp);
}

void TQDCDevice::writeClockDataDelay()
{
    quint16 clockDelay=0;
    quint16 dataDelay=0;
    switch (getDeviceSerialNumber()) {
    case 0x076D2B58:
        clockDelay=25; dataDelay=5; break;
    case 0x076D2B5F:
    case 0x076D3C47:
    case 0x076D3D38:
    case 0x076D3D37:
        clockDelay=29; dataDelay=5; break;
    case 0x076D3C21:
    case 0x076D45E9:
    case 0x080C635F:
        clockDelay=28; dataDelay=5; break;
    case 0x076D3C3C:
        clockDelay=28; dataDelay=4; break;
    case 0x06E9A6A9:
        clockDelay=23; dataDelay=2; break;
    case 0x080C6CB2:
        clockDelay=25; dataDelay=2; break;
    case 0x080C635B:
        clockDelay=24; dataDelay=2; break;
    case 0x0A79F593:
        clockDelay=10; dataDelay=14; break;
    case 0x0A7A3132:
    case 0x0A7A3133:
        clockDelay=26; dataDelay=3; break;
    case 0x0A7A328D:
        clockDelay=10; dataDelay=13; break;

    default:
        return;
    }
    RegOpVector regOp;
    regOp << RegWrite(REG_DES_IDELAY_TAP_VAL, clockDelay);
    regOp << RegWrite(REG_DES_IDELAY_LOAD_MASK, 0x5555);
    regOp << RegWrite(REG_DES_IDELAY_TAP_VAL, dataDelay);
    regOp << RegWrite(REG_DES_IDELAY_LOAD_MASK, 0xAAAA);
    regOpExec(regOp);
}

void TQDCDevice::ResetAdc()
{
//    regWrite(REG_DES_CTRL, ADC_CLK_CTRL_BIT_ADC_RESET);
//    regWrite(REG_DES_CTRL, 0);  //work on edge
}

//QString TQDCDevice::adcStatusStr(int status)
//{
//    quint8 adc_locked = status & 0xFF;
//    quint8 adc_ramp_error = (status >> 8) & 0xFF;
//    return QString("ADC status: error=%1, lock=%2").arg(adc_ramp_error, 8, 2, QChar('0')).arg(adc_locked, 8, 2, QChar('0'));
//}

void TQDCDevice::readoutHptdcSetup()
{
    const int HPTDC_SETUP_LEN = 41;
    RegOpVector regOp;
    for(int i=0; i<HPTDC_SETUP_LEN*2; ++i)
        regOp << RegRead(0x1000 + i);
    RegOpVector regRes = regOpExec(regOp);

    for(int tdcId=0; tdcId<2; ++tdcId){
        QFile file(QString("hptdc_eth_%1_%2.dat").arg(tdcId?"B":"A").arg(hptdcSetupIndex));
//        QFile file(QString("hptdc_%1_%2.dat").arg("eth").arg(hptdcSetupIndex));
        file.open(QIODevice::WriteOnly);
        QTextStream out(&file);
        for(int i=0; i<HPTDC_SETUP_LEN; ++i) {
            mlink::RegOp op = regRes[HPTDC_SETUP_LEN*tdcId+i];
            out << QString("%1 %2: ")
                   .arg(i*2, 3, 10, QChar('0'))
                   .arg(i*2*8, 3, 10, QChar('0'));
            for(int p=7; p>=0; --p)
                out << ((op.data & (1<<p)) ? "1":"0");
            out << QString(" %1\n").arg(op.data&0xFF, 0, 16);

            out << QString("%1 %2: ")
                   .arg(i*2+1, 3, 10, QChar('0'))
                   .arg(i*2*8+8, 3, 10, QChar('0'));
            for(int p=15; p>=8; --p)
                out << ((op.data & (1<<p)) ? "1":"0");
            out << QString(" %1\n").arg(op.data>>8, 0, 16);
            out.flush();
        }

        file.close();
    }
    ++hptdcSetupIndex;
}

bool TQDCDevice::waitEndOfRun()
{
    QElapsedTimer timer;
    timer.start();
    while(regRead(REG_TRIG_QUEUE) & 0xF){
        if(timer.elapsed() > 2000){
            qWarning() << "regRead(REG_TRIG_QUEUE):"<< hex << regRead(REG_TRIG_QUEUE)<<dec;
            return false;
        }
    }
    return true;
}

void TQDCDevice::resetEvNumAndStats()
{
    RegOpVector regOp;
    regOp << RegWrite32(REG32_TRIG_EVENT_NUM_LOAD, 0);
    regOp << RegWrite(REG_TRIG_CSR, REG_TRIG_CSR_BIT_CNT_RESET); // reset stats
    regOp << RegWrite(REG_TRIG_CSR, 0);
    regOpExec(regOp);
}

void TQDCDevice::calibrateBaseline(QMap<int, double> &mean, QMap<int, double> &rms)
{
    QMap<int, QVector<qint16>> adcVals;

    RegOpVector regOp;
    for(int ch=0; ch < getNumberOfChannels(); ++ch) {
        regOp << RegRead(REG_ADC_CH_DATA_BASE+ch);
        adcVals[ch].reserve(CALIBRATE_LIMIT);
        mean[ch] = 0;
        rms[ch] = 0;
    }
    if(!isConnected())
        return;

    int loop;
    for(loop = 0; loop<CALIBRATE_LIMIT; ++loop) {
        RegOpVector regRes = regOpExec(regOp);
        if(regRes.size() != regOp.size()) {
            qWarning() << "error in getting current adc value while calibration";
            break;
        }
        for(int ch=0; ch < getNumberOfChannels(); ++ch) {
            const qint16 &data = regRes[ch].data;
            adcVals[ch].push_back((qint16)data);
            mean[ch] += data;
        }
    }
    QStringList calReult;
    for(int ch=0; ch < getNumberOfChannels(); ++ch) {
        double &chMean = mean[ch];
        if(loop)
            chMean /= loop;

        double &chRms=rms[ch];
        for (qint16 val : adcVals[ch]) {
            chRms += (chMean-val)*(chMean-val);
        }
        if(loop)
            chRms /= loop;
        chRms = sqrt(chRms);

        calReult << QString::number(chMean, 'f', 2);
    }
}

QMap<int, int> TQDCDevice::calibrateDac()
{
    QMap<int, int> chThr;
    if(!isConnected())
        return chThr;

    regWrite(REG_ADC_DAC_INVERT, 0);
    const int cal_dac_delta_mv = 100;
    const int num_samples = 20;
    const quint16 dacMaxVal = dacValue(cal_dac_delta_mv);
    quint16 dacVal = dacValue(-cal_dac_delta_mv);

    QMap<int, QMap<quint16, int> > chCompState; // <ch, <dacVal, cmpCnt> >
    RegOpVector getCmpSt;
    for(int i=0; i<num_samples; ++i)
        getCmpSt << RegRead(REG_ADC_CMP_STATE);

    for(;dacVal<dacMaxVal;++dacVal) {
        for(int ch=0; ch<16; ++ch)
            chThr[ch] = dacVal;
        regOpExec(setChThreshold(chThr));

        if(!isConnected())
            break;
        RegOpVector regRes = regOpExec(getCmpSt);
        for(int ch=0; ch<16; ++ch)
            chCompState[ch][dacVal]=0;
        for(int i=0; i<num_samples; ++i) {
            const quint16 &dacSt = regRes[i].data;
            for(int ch=0; ch<16; ++ch){
                if(dacSt & (1<<ch))
                    ++chCompState[ch][dacVal];
            }
        }
    }
    chThr.clear();
    for (int ch : chCompState.keys()) {
        for (int dacVal : chCompState[ch].keys()) {
            int v = chCompState[ch][dacVal];
            if(v < (num_samples/2)){
                chThr[ch] = dacVal;
//                qInfo() << QString("%1: %2 (%3mV)")
//                           .arg(ch).arg(dacVal)
//                           .arg(dac_mV(dacVal));
                break;
            }
        }
    }
    return chThr;
}

void TQDCDevice::clearStatistics()
{
    status.statistic.clear();
}

bool TQDCDevice::checkLiveMagic()
{
    bool liveMagicCorrect = QtMregDevice::checkLiveMagic();
    if(!liveMagicCorrect) {
        setupCacheValid = false;
    }
    return liveMagicCorrect;
}

quint16 TQDCDevice::dacValue(int mV)
{
    return dacValue(mV, DAC_DEFAULT_ZERO);
}

quint16 TQDCDevice::dacValue(int mV, int dacShift)
{
    int dacData = 1e-3*mV/DAC_V_REF * DAC_RANGE+dacShift;
    if(dacData < 0)
        return 0;
    else if(dacData > DAC_RANGE)
        return DAC_RANGE;
    else
        return dacData;
}

double TQDCDevice::dac_mV(quint16 dacData)
{
    Q_ASSERT(dacData<DAC_RANGE);
    return 1e3*DAC_V_REF/DAC_RANGE*((int)dacData-DAC_DEFAULT_ZERO);
}

bool HptdcStatus::isOk() const
{
    return (regValues[0] & HPTDC_ERR_MASK) && isDllLock();
}

bool HptdcStatus::isDllLock() const
{
    return regValues[1] & (1<<12);
}

bool HptdcStatus::hasError() const
{

    return !isDllLock() || (regValues[0] & 0x7F);
}

QString HptdcStatus::getSummary() const
{
    QStringList sList;
    for(int i=0; i<16; ++i)
        if(regValues[0] & (1<<i) & HPTDC_ERR_MASK)
            sList << HPTDC_ST_STR[i];

//    quint8 readOutFifoOccup = regValues[1] & 0xFF;
//    quint8 trigFifoOccup = (regValues[1]>>12) & 0xFF;

//    quint8 occupL1Buffer[4];
//    occupL1Buffer[0] = regValues[2] & 0xFF;
//    occupL1Buffer[1] = (regValues[2]>>12) & 0xFF;
//    occupL1Buffer[2] = regValues[3] & 0xFF;
//    occupL1Buffer[3] = (regValues[4]>>12) & 0xFF;
    if(!isDllLock())
        sList << QString("Dll %1lock").arg(isDllLock()?"":"un");

    return sList.join(' ');
}
