﻿#include <QDialog>
#include <QByteArray>

#include "Config.h"
#include "mongo/BaseConfig.h"
#include "mongo/StartupOptions.h"

namespace  {
const QString CLIENT_SECTION = "Clients";
const QString DEVICE_PREFIX = "DEV_";
const QString DEFAULT_DEV = "DEFAULT_DEV";
}

Config::Config(const StartupOptions &startupOptions) :
    BaseConfig()
    , program_type(startupOptions.program_type)
    , program_index(startupOptions.program_index)
    , configuration_name(startupOptions.configuration_name)
    , debug_mode(startupOptions.debug_mode)
    , runEnable(false)
    , allowRemoteControl(true)
{
    update_root();
    default_root = root;
    default_root.set_data("configuration_name", QString("default"));
}

void Config::read_config()
{
    devicesCfg.clear();
    root = BaseConfig::read_config(program_type, program_index, configuration_name);
    if (!root.isEmpty()) {
        program_index = root.get_data<QString>("program_index", program_index);

//        runEnable = root.get_data<bool>("runEnable", runEnable);
        runId.runIndex = root.get_data<QString>("run_index", runId.runIndex);
        runId.runNumber = root.get_data<int>("run_number", runId.runNumber);
        localRunNumber = root.get_data<int>("localRunNumber", localRunNumber);
        allowRemoteControl = root.get_data<bool>("allowRemoteControl", allowRemoteControl);

        for(QString child_name : root.children().keys()) {
            if(child_name.startsWith(DEVICE_PREFIX)) {
                TtvxsConfig devCfg = get_device_config(root.child(child_name));
                if(!devCfg.device_index.isNull())
                    devicesCfg[devCfg.device_index] = devCfg;
            } else if(child_name == CLIENT_SECTION){
                RootConfig root_child = root.child(child_name);
                for(auto client_root : root_child.children_const().values()) {
                    ClientConfig cc;
                    cc.id = client_root.get_data<int>("id");
                    cc.program_type = client_root.get_data<QString>("program_type");
                    cc.program_index = client_root.get_data<QString>("program_index");
                    cc.readout = client_root.get_data<bool>("readout", true);
                    cc.evNumCheck = client_root.get_data<bool>("evNumCheck", true);
                    if(!cc.program_type.isEmpty() && !cc.program_index.isEmpty()) {
                        clients[cc.id] = cc;
                    }
                }
            }
        }

        default_cfg = get_device_config(root.child(DEFAULT_DEV));
    }
}

void Config::write_config()
{
    update_root();
    BaseConfig::write_config(root);
}

void Config::unify_config()
{
    for(auto i : devicesCfg.keys()) {
        if(devicesCfg[i].enabled)
            devicesCfg[i].unify_config(default_cfg);
    }
}

RootConfig Config::get_root_config(const TtvxsConfig &devCfg)
{
    RootConfig child;
    if(devCfg.device_index.isNull())
        child.item_name = DEFAULT_DEV;
    else {
        child.item_name = QString("%1%2").arg(DEVICE_PREFIX)
                .arg(devCfg.device_index.getIdent());

        child.set_data("devId", devCfg.device_index.getDevId());
        child.set_data("devSer", devCfg.device_index.getSerial());
    }

    child.set_data("enabled", devCfg.enabled);
    child.set_data("run", devCfg.run);
    child.set_data("trig_open", devCfg.trig_open);

    child.set_data("trigSrc", devCfg.trigSrc.all);
    child.set_data("vxsChMask", devCfg.vxsChMask);
    child.set_data("trigTimerFreq", devCfg.trigTimerFreq);
    child.set_data("trigRandomFreq", devCfg.trigRandomFreq);
    child.set_data("trigRandomDeadTimeUs", devCfg.trigRandomDeadTimeUs);

    child.set_data("aux_cnt_for_trig", devCfg.aux_cnt_for_trig);
    child.set_data("stat_readout", devCfg.stat_readout);
    child.set_data("stat_period", devCfg.stat_period);
    child.set_data("custom_name", devCfg.custom_name);

    child.append_child(devCfg.stat_preset.variables_to_root());

    return child;
}

TtvxsConfig Config::get_device_config(const RootConfig &child)
{
    TtvxsConfig devCfg;
    if(child.item_name.startsWith(DEVICE_PREFIX)) {
        DeviceIndex index = DeviceIndex(child.get_data<quint16>("devId"),
                                        child.get_data<quint64>("devSer"));
        if(index.isNull())
            return devCfg;

        devCfg.device_index = index;
    } else if(child.item_name != DEFAULT_DEV) {
        return devCfg;
    }

    devCfg.enabled = child.get_data<bool>("enabled", devCfg.enabled);
    devCfg.run = child.get_data<bool>("run", devCfg.run);
//            devCfg.trig_open = child.get_data<bool>("trig_open");

    devCfg.trigSrc.all = child.get_data<quint16>("trigSrc", devCfg.trigSrc.all);
    devCfg.vxsChMask = child.get_data<quint32>("vxsChMask", devCfg.vxsChMask);
    devCfg.trigTimerFreq = child.get_data<int>("trigTimerFreq", devCfg.trigTimerFreq);
    devCfg.trigRandomFreq = child.get_data<int>("trigRandomFreq", devCfg.trigRandomFreq);
    devCfg.trigRandomDeadTimeUs = child.get_data<double>("trigRandomDeadTimeUs", devCfg.trigRandomDeadTimeUs);

    devCfg.aux_cnt_for_trig = child.get_data<bool>("aux_cnt_for_trig", devCfg.aux_cnt_for_trig);
    devCfg.stat_readout = child.get_data<bool>("stat_readout", devCfg.stat_readout);
    devCfg.stat_period = child.get_data<quint16>("stat_period", devCfg.stat_period);

    devCfg.stat_preset.root_to_variables(child.child(StatPreset::DEFAULT_ITEM_NAME));
    devCfg.custom_name = child.get_data<QString>("custom_name", devCfg.custom_name);
    return devCfg;
}

void Config::update_root()
{
    root.remove_all_data();
    root.set_data("program_type", program_type);
    root.set_data("program_index", program_index);
    root.set_data("configuration_name", configuration_name);

    root.set_data("runEnable", runEnable);
    root.set_data("run_index", runId.runIndex);
    root.set_data("run_number", runId.runNumber);
    root.set_data("debug_mode", debug_mode);
    root.set_data("localRunNumber", localRunNumber);
    root.set_data("allowRemoteControl", allowRemoteControl);

    root.append_child(get_root_config(default_cfg));
    for(auto devCfg: devicesCfg.values())
        root.append_child(get_root_config(devCfg));


    if(!clients.isEmpty()) {
        RootConfig root_clients = RootConfig(CLIENT_SECTION);
        int id=0;
        for (const auto &cc : clients.values()) {
            RootConfig root_child_client = RootConfig(QString("%1_%2").arg(cc.program_type).arg(cc.program_index));
            root_child_client.set_data("id", ++id);
            root_child_client.set_data("program_type", cc.program_type);
            root_child_client.set_data("program_index", cc.program_index);
            root_child_client.set_data("readout", cc.readout);
            root_child_client.set_data("evNumCheck", cc.evNumCheck);
            root_clients.append_child(root_child_client);
        }
        root.append_child(root_clients);
    }
}
