#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QElapsedTimer>
#include <QPair>
#include <QMap>
#include <QVector>
#include <QDateTime>
#include <QHostAddress>

#include <mldiscover/DeviceIndex.h>
#include <mldiscover/DeviceDescription.h>

class DiscoverDialog;
class DeviceDescription;
class QLabel;

namespace Ui {
class MainWindow;
}

struct IpChange {
    IpChange() : ip(0) {}
    IpChange(quint32 i, QDateTime t) :
        ip(i), time(t) {}

    QString getIpStr() const { return QHostAddress(ip).toString(); }
    quint32 ip;
    QDateTime time;
};

struct LoopInfo {
    LoopInfo(uint loopId) :
        id(loopId),
        powerOnTime(QDateTime::currentDateTime())
    {
    }
//    LoopInfo(uint loopId, QDateTime startTime) :
//        id(loopId),
//        start(startTime)
//    {}
    LoopInfo() : id(0) {}
    QString toString() const {
        const QString format = "dd/MM/yyyy hh:mm:ss";
        QString msg = QString("\t%1: %2+%3 %4+%5").arg(id)
                .arg(powerOnTime.isNull() ? "--" : powerOnTime.toString(format))
                .arg((powerOnTime.isNull() || start.isNull()) ? "--" : QString("%1").arg(powerOnTime.secsTo(start), 2))
                .arg(powerOffTime.isNull() ? "--" : powerOffTime.toString(format))
                .arg((powerOffTime.isNull() || end.isNull()) ? "--" : QString("%1").arg(powerOffTime.secsTo(end), 2));
        return msg;
    }

    uint id;
    QDateTime powerOnTime;
    QDateTime start;
    QDateTime powerOffTime;
    QDateTime end;
};

struct DevStat {
    DevStat() {
        clear();
    }
    DevStat(const DeviceIndex &i) :
        index(i),
        slot(0)
    {
        clear();
    }
    void clear() {
        attemptsCnt = 0;
        onlineCnt = 0;
        online = false;
        list.clear();
    }
    bool operator< (const DevStat& v) const
    {
        return slot < v.slot || (slot == v.slot && index.getSerial() < v.index.getSerial());
    }

    DeviceIndex index;

    quint32 attemptsCnt;
    quint32 onlineCnt;
    bool online;
    QMap<uint, LoopInfo> list;
    QList<IpChange> ipList;
    DeviceDescription lastDescr;
    uint slot;
};

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent = nullptr);
    ~MainWindow();

private slots:
    void on_pushButtonAdd_clicked();
    void on_pushButtonRemove_clicked();
    void on_pushButtonPower_clicked();
    void on_spinBoxTimeout_valueChanged(int);
    void on_radioButtonRun_toggled(bool checked);
    void on_pushButtonReset_clicked();
    void on_pushButtonInfo_clicked();

    void on_tableWidget_cellChanged(int row, int column);
    void on_tableWidget_currentCellChanged(int currentRow, int currentColumn, int previousRow, int previousColumn);

    void deviceUpdated(const DeviceDescription &dd);
    void deviceExpired(const DeviceIndex &index);

    void switchPower(bool force=false);
    void updateStatus();
    bool getCrateState();

private:
    void initTable();
    void readCfg();
    void saveCfg();
    void addDev(const DeviceIndex &index, uint slot=0);
    DeviceIndex getComboBoxDevIndex(int index) const;
    DeviceIndex getDevIndex(int row) const;
    int getDevRow(const DeviceIndex &index) const;
    bool isAllDevInState();
    void updateDeviceInfo(const DeviceIndex &index);
    void updateStatusBar();
    void addLog(QString log);
    bool switchPowerScript(bool on);

    Ui::MainWindow *ui;

    DiscoverDialog * const discoverDialog;
    QMap<DeviceIndex, DevStat> devStats;
    QTimer * powerTimer;
    QTimer * statusTimer;
    QTimer * crateStatusTimer;
    bool powerOn;
    uint loopId;
    uint loopGoodCnt;
    uint loopBadCnt;
    bool run;
    QElapsedTimer switchTime;
    QElapsedTimer pingTime;
    QLabel * const mainStatus;
    QLabel * const crateStatus;
};

#endif // MAINWINDOW_H
