#include "DiscoverDialog.h"
#include "ui_DiscoverDialog.h"

#include "mldiscover/MldpListener.h"
#include "ManualAddDialog.h"

#include <QtNetwork>
#include <QCheckBox>
#include <QDebug>
#include <QElapsedTimer>
#include <QHostInfo>
#include <QLabel>
#include <QPushButton>
#include <QRadioButton>
#include <QButtonGroup>

namespace  {
const bool HIDING_SELECTED_ENABLED = false;
enum TableColumns{
    COL_ENABLE,
    COL_STATE,
    COL_DEV_TYPE_TXT,
    COL_SERIAL_ID_TXT,
    COL_FIRM_WARE,
    COL_SLOT,
    COL_IP,
    COL_MAC,
    COL_MASTER,
    COL_M_STREAM,
    COL_TYPE_NUM,
    COL_SERIAL_NUM,
    COL_TOTAL
};
}

DiscoverDialog::DiscoverDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::DiscoverDialog),
    mldpListener(new MldpListener(this)),
    count(0),
    outOfDateTimer(new QTimer(this)),
    resizeTimer(new QTimer(this)),
    useCheckBoxes(true),
    printInfo(false),
    printInfoFull(false),
    buttonGroupEnabled(new QButtonGroup(this)),
    show_mac_adress(false)
{
    ui->setupUi(this);
    setWindowTitle("Device Discovery Dialog");
    setTableHeaders(true);
    setTableHeaders(false);
    //    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
    ui->pushButtonTest->hide();
    ui->checkBoxShowMAC->setVisible(false);

    QMap<QString, QString> map;
    // TODO: make configurable
    map["159.93.126.0"] = "Nuclotron GPN";
    map["192.168.126.0"] = "Nuclotron Intranet";
    map["10.18.99.0"] = "White Rabbit Network";
    map["10.18.69.0"] = "AFI-DAQ Network";
    map["10.18.97.0"] = "BMN-DAQ Network";
    map["10.18.100.0"]= "MPD-TEST-DAQ Network";

    for (const auto &key : map.keys()) {
        QHostAddress ip(key);
        addSubnet(ip.toIPv4Address(), map[key]);
    }

    connect(mldpListener, &MldpListener::received, this, &DiscoverDialog::descriptionUpdated);
    connect(mldpListener, &MldpListener::received, this, &DiscoverDialog::deviceDiscoveredAll);

    outOfDateTimer->setInterval(1000);
    connect(outOfDateTimer, &QTimer::timeout, this, &DiscoverDialog::checkOutOfDate);
    outOfDateTimer->start();
    resizeTimer->setInterval(100);
    resizeTimer->setSingleShot(true);
    connect(resizeTimer, &QTimer::timeout, this, &DiscoverDialog::resizeTables);
    updateLocalIp();
    auto t = new QTimer(this);
    connect(t, &QTimer::timeout,
            this, &DiscoverDialog::checkHostInfoUpdates);
    t->start(1000);
}

DiscoverDialog::~DiscoverDialog()
{
    delete ui;
}

void DiscoverDialog::descriptionUpdated(const DeviceDescription &dd)
{
    if (printInfo && printInfoFull) {
        qDebug()<<QTime::currentTime().toString()<<" Discovered device:";
        qDebug()<<QString("[Chassis ID]: MAC=%1").arg(dd.haddr, 0, 16);
        qDebug()<<QString("[Port ID]: port=%1").arg(dd.portID);
        qDebug()<<QString("[TTL]: %1").arg(dd.ttl);
//        qDebug()<<QString("[System Name]: %1").arg(sys_name);
//        qDebug()<<QString("[System Description]: %1").arg(sys_desc);

//        qDebug()<<QString("[System Capabilities]: %1").arg(sysCap, 2, 16);
//        qDebug()<<QString("[Enabled Capabilities]: %1").arg(enCap, 2, 16);

//        qDebug()<<QString("[Management Address]: %1, intNumSubtype=%2, intNum=%3, OID=%4")
//                  .arg(mngAddrStr)
//                  .arg(intNumSubtype)
//                  .arg(intNum)
//                  .arg(oid);
        qDebug()<<QString("[Organizationally Specific]: device_id=0x%1 serial_id=%2")
                  .arg(dd.device_id, 2, 16, QChar('0'))
                  .arg(dd.serial_id, 10, 16, QChar('0'));

        if (dd.master_locked){
            qDebug()<<QString("[Organizationally Specific]: Master: ip=%1(%2) port=%3")
                      .arg(QHostAddress(dd.master_ip).toString())
                      .arg(dd.master_mac, 12, 16, QChar('0'))
                      .arg(dd.master_udp_port);
        } else {
            qDebug()<<QString("[Organizationally Specific]: Master is free");
        }

        if (dd.mstream_locked){
            qDebug()<<QString("[Organizationally Specific]: MStream: ip=%1(%2) port=%3")
                      .arg(QHostAddress(dd.mstream_ip).toString())
                      .arg(dd.mstream_mac, 12, 16, QChar('0'))
                      .arg(dd.mstream_udp_port);
        } else {
            qDebug()<<QString("[Organizationally Specific]: MStream is free");
        }
    }

    if (!dd.isNull() && checkDeviceIdFilter(dd.device_id)) {
        insertDev(dd);
        emit deviceDiscovered(dd);
    }
}

bool DiscoverDialog::checkDeviceIdFilter(quint16 deviceId) const
{
    return deviceIdFilter.empty() || deviceId==0 || deviceIdFilter.contains(deviceId) || deviceIdFilter.contains(deviceIdAnyDevice);
}

void DiscoverDialog::updateShownDevices()
{
    if (printInfo && printInfoFull)
        qDebug()<<"deviceIdShown:"<<hex<<deviceIdShown;
    //Remove all rows
//    ui->tableWidgetAvailable->clear();
    ui->tableWidgetAvailable->setRowCount(0);

    for (const auto &dd : availableDevices.values()) {
        if(checkDeviceIdFilter(dd.device_id))
            insertRow(true, dd);
        else
            availableDevices.remove(dd.getIndex());
    }

    if(HIDING_SELECTED_ENABLED) {
//        ui->tableWidgetSelected->clear();
        for(int row=0; row<ui->tableWidgetSelected->rowCount(); ++row)
            delete ui->tableWidgetSelected->cellWidget(row, COL_ENABLE);
        ui->tableWidgetSelected->setRowCount(0);

        for (const auto &dd : selectedDevices.values()) {
            insertRow(false, dd);
        }
    }
}

void DiscoverDialog::updateDeviceTypeInfo(const DeviceDescription &dd)
{
    if(deviceTypeMap.contains(dd.device_id)) {
        QList<QString> curTypes = deviceTypeMap.values(dd.device_id);
        if(curTypes.contains(dd.getModelStr())) {
            return;
        }

        bool curIsHex, newIsHex;
        quint16 id1 = 0, id2;
        if(curTypes.size() == 1) {
            id1 = deviceTypeMap.values(dd.device_id).first().toUShort(&curIsHex, 16);
        } else {
            curIsHex = false;
        }

        id2 = dd.getModelStr().toUShort(&newIsHex, 16);
        if(curIsHex) {
            // Cur. type is hex number
            if(newIsHex) {
                // Both has hex deviceId
                Q_ASSERT(id1==id2);
                return;
            } else {
                // dd has model string; need to update
                deviceTypeMap.remove(dd.device_id);
                deviceTypeMap.insert(dd.device_id, dd.getModelStr());
                deviceTypeMapUpdated(dd.device_id);
            }
        } else {
            // Cur. type is String model or list of models
            if(!newIsHex) {
                deviceTypeMap.insert(dd.device_id, dd.getModelStr());
                deviceTypeMapUpdated(dd.device_id);
            }
            return;
        }
    } else {
        deviceTypeMap.insert(dd.device_id, dd.getModelStr());
        QListWidgetItem *i = new QListWidgetItem (dd.getModelStr());
        i->setData(Qt::UserRole, dd.device_id);
        ui->listWidgetTypes->addItem(i);
        ui->listWidgetTypes->setItemSelected(i, true);
    }
}

void DiscoverDialog::updateHostInfo(QTableWidget *table)
{
    bool upper = table == ui->tableWidgetAvailable;
    const int rows = table->rowCount();
    bool isChanged = false;
    for(int row=0; row<rows; ++row) {
        const DeviceIndex i = getDeviceIndexOfRow(upper, row);
        Q_ASSERT((upper?availableDevices:selectedDevices).contains(i));
        const DeviceDescription &dd = (upper?availableDevices:selectedDevices).value(i);
        if(dd.isNull())
            continue;

        QString itemText;
        QTableWidgetItem *item = table->item(row, COL_IP);
        if(getHostMap().contains(dd.ip_addr)) {
            itemText = getHostMap()[dd.ip_addr];
            if(item->text() != itemText) {
                isChanged = true;
                item->setText(itemText);
                item->setToolTip(QHostAddress(dd.ip_addr).toString());
            }
        }

        item = table->item(row, COL_MASTER);
        if(dd.master_locked) {
            itemText = getHostMap().value(dd.master_ip, QHostAddress(dd.master_ip).toString());
            itemText.append(":").append(QString::number(dd.master_udp_port));
        } else
            itemText = "free";
        if(item->text() != itemText) {
            isChanged = true;
            item->setText(itemText);
            item->setToolTip(dd.master_locked ? QHostAddress(dd.master_ip).toString() : "");
        }

        item = table->item(row, COL_M_STREAM);
        if(dd.mstream_locked) {
            itemText = getHostMap().value(dd.mstream_ip, QHostAddress(dd.mstream_ip).toString());
            itemText.append(":").append(QString::number(dd.mstream_udp_port));
        } else
            itemText = "free";
        if(item->text() != itemText) {
            isChanged = true;
            item->setText(itemText);
            item->setToolTip(QHostAddress(dd.mstream_ip).toString());
        }
    }

    if(!isChanged) {
        isChanged = true;
        table->setUpdatesEnabled(false);
    }

    if(isChanged)
        table->setUpdatesEnabled(true);
}

void DiscoverDialog::updateSubnetList(const DeviceDescription &dd)
{
    if(dd.ip_addr == 0) {
        addSubnet(0, "No ip");
    } else {
        const quint32 subnet = dd.ip_addr&0xFFFFFF00;
        addSubnet(subnet);
    }
}

bool DiscoverDialog::isInShownSubnet(const DeviceDescription &dd) const
{
    return subnetShown.isEmpty() || subnetShown.contains(dd.ip_addr&0xFFFFFF00);
}

void DiscoverDialog::addSubnet(quint32 subnet, QString name)
{
    if(knownSubnets.contains(subnet))
        return;
    QListWidget *list = ui->listWidgetSubnet;
    knownSubnets.insert(subnet);
    int row = list->count();
    QString toolTip = QString("%1.%2.%3.%4")
            .arg(subnet>>24).arg((subnet>>16)&0xFF)
            .arg((subnet>>8)&0xFF).arg(subnet&0xFF);
    if(name.isNull())
        name = toolTip;
    QListWidgetItem *item = new QListWidgetItem(name);
    item->setData(Qt::UserRole, subnet);
    item->setToolTip(toolTip);
    list->insertItem(row, item);
    list->setItemSelected(item, true);
}

void DiscoverDialog::deviceTypeMapUpdated(quint16 id)
{
    QListWidget *list = ui->listWidgetTypes;
    for(int row=0; row<list->count(); ++row) {
        QListWidgetItem *item = list->item(row);
        if(item->data(Qt::UserRole)==id) {
            item->setText(deviceTypeMap.values(id).join(","));
            return;
        }
    }
    Q_ASSERT_X(false, "updateDeviceTypeInfo", "can't find list item with needed devId");
}

bool DiscoverDialog::isDevShown(bool upper, const DeviceDescription &dd) const
{
    if(!upper && !HIDING_SELECTED_ENABLED)
        return true;
    if(!deviceIdShown.isEmpty() && !deviceIdShown.contains(dd.device_id))
        return false;
    return isInShownSubnet(dd);
}

void DiscoverDialog::updateLocalIp()
{
    localIp.clear();
    for (const auto &address : QNetworkInterface::allAddresses()) {
        if (address.protocol() == QAbstractSocket::IPv4Protocol && address != QHostAddress(QHostAddress::LocalHost))
             localIp << address.toIPv4Address();
    }
}

void DiscoverDialog::insertDev(DeviceDescription dd)
{
//    if (deviceIdShown)
//        return;
    const DeviceIndex &index = dd.getIndex();
    if(availableDevices.contains(index)){
        DeviceDescription &ddOld = availableDevices[index];
        if(dd.deviceInfoAreSame(ddOld)) {
            ddOld.discoverTime = dd.discoverTime;
            ddOld.discoverTimer = dd.discoverTimer;
            return;
        }
    } else if(selectedDevices.contains(index)) {
        DeviceDescription &ddOld = selectedDevices[index];
        if(dd.deviceInfoAreSame(ddOld) && devicePreviousState.value(index, false)) {
            ddOld.discoverTime = dd.discoverTime;
            ddOld.discoverTimer = dd.discoverTimer;
            return;
        }
    }
    if (!checkDeviceIdFilter(dd.device_id))
        return;
    updateDeviceTypeInfo(dd);
    if(dd.ip_addr && !getHostMap().contains(dd.ip_addr))
        dnsLookup.reverseLookup(QHostAddress(dd.ip_addr).toString());
    if(dd.master_ip && !getHostMap().contains(dd.master_ip))
        dnsLookup.reverseLookup(QHostAddress(dd.master_ip).toString());
    if(dd.mstream_ip && !getHostMap().contains(dd.mstream_ip))
        dnsLookup.reverseLookup(QHostAddress(dd.mstream_ip).toString());


    if(availableDevices.contains(index)){
        dd.enabled = availableDevices[index].enabled;
        availableDevices[index] = dd;
        updateTableRow(true, dd);
        return; // Already exist
    }
    if(selectedDevices.contains(index)){
        dd.enabled = selectedDevices[index].enabled;
        selectedDevices[index] = dd;
        updateTableRow(false, dd);
        return; // Already exist
    }

    availableDevices[index] = dd;
    insertRow(true, dd);
    updateSubnetList(dd);
}

void DiscoverDialog::insertRow(bool isFree, const DeviceDescription dd)
{
    if(!isDevShown(isFree, dd))
        return;

    QTableWidget *table = isFree ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    table->setSortingEnabled(false);

    int newRow = table->rowCount();
    table->setRowCount(newRow+1);

    if (!isFree) {
        // Add CheckBox or RadioButton
        QWidget *cellWidget = new QWidget(table);
        table->setCellWidget(newRow, COL_ENABLE, cellWidget);

        QAbstractButton *boolWidget;
        if (useCheckBoxes) {
            boolWidget = new QCheckBox(cellWidget);
        } else {
            boolWidget = new QRadioButton(cellWidget);
            buttonGroupEnabled->addButton(boolWidget);
        }
        connect(boolWidget, &QAbstractButton::toggled,
                this, &DiscoverDialog::setDeviceEnabled);

        QHBoxLayout *layout = new QHBoxLayout(cellWidget);
        layout->setAlignment(Qt::AlignCenter);
        cellWidget->setLayout(layout);
        layout->addWidget(boolWidget);

        // boolWidget->updateGeometry();
        // Add Status Icon
        cellWidget = new QWidget(table);
        table->setCellWidget(newRow, COL_STATE, cellWidget);

        QLabel *label = new QLabel(cellWidget);

        layout = new QHBoxLayout(cellWidget);
        layout->setAlignment(Qt::AlignCenter);
        cellWidget->setLayout(layout);
        layout->addWidget(label);

        setOnlineState(newRow, !dd.isOutOfDate(), dd.getIndex());
    }
    table->setItem(newRow, COL_DEV_TYPE_TXT, new QTableWidgetItem());
    table->setItem(newRow, COL_SERIAL_ID_TXT, new QTableWidgetItem());
    table->setItem(newRow, COL_FIRM_WARE, new QTableWidgetItem());
    table->setItem(newRow, COL_SLOT, new QTableWidgetItem());
    table->setItem(newRow, COL_IP, new QTableWidgetItem());
    table->setItem(newRow, COL_MAC, new QTableWidgetItem());
    table->setItem(newRow, COL_MASTER, new QTableWidgetItem("free"));
    table->setItem(newRow, COL_M_STREAM, new QTableWidgetItem("free"));
    table->setItem(newRow, COL_TYPE_NUM, new QTableWidgetItem(QString::number(dd.device_id)));
    table->setItem(newRow, COL_SERIAL_NUM, new QTableWidgetItem(QString::number(dd.serial_id)));
    table->item(newRow, COL_DEV_TYPE_TXT)->setToolTip(QString("0x%1").arg(dd.device_id, 2, 16, QChar('0')));
    if (printInfo && printInfoFull)
        qDebug()<<"Device row inserted row="<<newRow<<hex<<dd.getModelStr()<<dd.getSerialIdStr();

    updateTableRow(isFree, dd, newRow);
    table->setSortingEnabled(true);
}


void DiscoverDialog::removeRow(bool free, DeviceIndex index)
{
    if (printInfo && printInfoFull)
        qDebug()<<"Remove from"<<(free?"upper":"bottom")<<" table row with index:"<<index;
    QTableWidget *table = free ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    for (int row=0; row<table->rowCount(); ++row){
        DeviceIndex dIndex = getDeviceIndexOfRow(free, row);
//        if (printInfo && printInfoFull)
//            qDebug() << free << row << hex << dIndex;
        if (dIndex == index) {
            table->removeRow(row);
            if (printInfo && printInfoFull)
                qDebug()<<"remove row="<<row<<"in "<<(free?"available":"selected")<<" table;"<<hex<<dIndex.second;
            return;
        }
    }
}

void DiscoverDialog::moveRows(bool downwards)
{
    QVector<int> rows = getSelectedRows(downwards);
    QSet<DeviceIndex> indexes;
    for (int i=rows.size()-1; i>=0; --i) {
        int row = rows[i];
        DeviceIndex index = getDeviceIndexOfRow(downwards, row);
        indexes.insert(index);
    }

    for (const auto &index : indexes)
        moveRow(downwards, index);
    checkOutOfDate();
}

void DiscoverDialog::moveRow(bool downwards, DeviceIndex index)
{
    QMap<DeviceIndex, DeviceDescription> &fromMap = downwards ? availableDevices : selectedDevices;
    QMap<DeviceIndex, DeviceDescription> &toMap = downwards ? selectedDevices : availableDevices;
    Q_ASSERT(fromMap.contains(index));
    Q_ASSERT(!toMap.contains(index));
    DeviceDescription dd = fromMap.take(index);
    bool expired = dd.isOutOfDate();
    if(expired && (downwards || !devicePreviousState.contains(index) || devicePreviousState[index]))
        emit deviceExpired(index);

    removeRow(downwards, index);
    if(downwards || !expired) {
        toMap[index] = dd;
        insertRow(!downwards, dd);
    }
}

void DiscoverDialog::setOnlineState(int row, bool isOnline, DeviceIndex index)
{
    QTableWidget *table = ui->tableWidgetSelected;
    QLabel *label = qobject_cast<QLabel*>(table->cellWidget(row, COL_STATE)->layout()->itemAt(0)->widget());
    QPixmap p;
    if (isOnline){
        label->setToolTip("Device sends discover packets");
        p = QApplication::style()->standardPixmap(QStyle::SP_DialogApplyButton);
    } else {
        label->setToolTip("Device doesn't send discover packets");
//        p = QApplication::style()->standardPixmap(QStyle::SP_DialogCancelButton);
        p = QApplication::style()->standardPixmap(QStyle::SP_DialogCloseButton);
    }
    label->setPixmap(p);
    devicePreviousState[index] = isOnline;
}


void DiscoverDialog::updateTableRow(bool free, const DeviceDescription dd, int row)
{
    updateSubnetList(dd);
    if(!isDevShown(free, dd))
        return;

    QTableWidget *table = free ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    bool isChanged = false;
    bool sorting = table->isSortingEnabled();
    table->setSortingEnabled(false);
    DeviceIndex index = dd.getIndex();
    if(row==-1 || index != getDeviceIndexOfRow(free, row)){
        row = getRowByIndex(free, index);
    }
    Q_ASSERT(row!=-1);

    if (printInfo)
        qDebug()<<QString("Device updated s/n:%1 (type:%2) (%3ms)")
                  .arg(dd.getSerialIdStr())
                  .arg(dd.getModelStr())
                  .arg(dd.discoverTimer.elapsed());

    if (!free){
        // Update enable state
        QAbstractButton *button = qobject_cast<QAbstractButton*>(
                    table->cellWidget(row, COL_ENABLE)->layout()->itemAt(0)->widget()
                    );
        Q_ASSERT(button);
        if(button->isChecked() != dd.enabled) {
            isChanged = true;
            button->setChecked(dd.enabled);
        }

        // Update icon
        DeviceIndex index = dd.getIndex();
        bool isOnline = !dd.isOutOfDate();
        if (!devicePreviousState.contains(index) || devicePreviousState[index] != isOnline){
            isChanged = true;
            setOnlineState(row, isOnline, index);
        }
    }

    QString modelStr = dd.getModelVerStr();
    if(dd.modelname_str.isEmpty()){
        modelStr = deviceTypeMap.values(dd.device_id).join("/");
    }
    if(modelStr != table->item(row, COL_DEV_TYPE_TXT)->text()) {
        isChanged = true;
        table->item(row, COL_DEV_TYPE_TXT)->setText(modelStr);
    }
    if(dd.getSerialIdStr() != table->item(row, COL_SERIAL_ID_TXT)->text()) {
        isChanged = true;
        table->item(row, COL_SERIAL_ID_TXT)->setText(dd.getSerialIdStr());
    }
    if(dd.getFirmWareStr() != table->item(row, COL_FIRM_WARE)->text()) {
        isChanged = true;
        table->item(row, COL_FIRM_WARE)->setText(dd.getFirmWareStr());
    }
    QString slotNumberStr = QString("%1").arg(dd.chassis_slot,2,10,QChar('0'));
    if (slotNumberStr != table->item(row, COL_SLOT)->text()) {
        isChanged = true;
        table->item(row, COL_SLOT)->setText(slotNumberStr);
    }

    // COL_IP
    bool ipResolved = getHostMap().contains(dd.ip_addr);
    QString ipStr = ipResolved ? getHostMap()[dd.ip_addr] : QHostAddress(dd.ip_addr).toString();
    if(ipStr != table->item(row, COL_IP)->text()) {
        isChanged = true;
        table->item(row, COL_IP)->setText(ipStr);
        if(ipResolved)
            table->item(row, COL_IP)->setToolTip(QHostAddress(dd.ip_addr).toString());
    }
    // COL_MAC
    QString mac_adress = QString("%1").arg(dd.haddr, 12, 16, QChar('0'));
    mac_adress = mac_adress.toUpper();
    const int step = 2;
    for(int i=step; i< mac_adress.size(); i+= step+1){
        mac_adress.insert(i, ":");
    }

    table->item(row, COL_MAC)->setText(mac_adress);

    // COL_MASTER
    QString masterStr;
    if (dd.master_locked) {
        ipResolved = getHostMap().contains(dd.master_ip);
        ipStr = ipResolved ? getHostMap()[dd.master_ip] : QHostAddress(dd.master_ip).toString();
        masterStr = QString("%1:%2").arg(ipStr).arg(dd.master_udp_port);
    } else {
        masterStr = "free";
    }
    QTableWidgetItem * const masterItem = table->item(row, COL_MASTER);
    if(masterStr != masterItem->text()) {
        isChanged = true;
        masterItem->setText(masterStr);
        masterItem->setToolTip((ipResolved && dd.master_locked) ? QHostAddress(dd.master_ip).toString() : "");
    }
    QColor masterItemCol = (!dd.master_locked || localIp.contains(dd.master_ip)) ? Qt::black : (free ? QColor(0xf48000):Qt::red);
    if(masterItemCol != masterItem->textColor()) {
        isChanged = true;
        masterItem->setForeground(masterItemCol);
    }

    // COL_M_STREAM
    QString mStreamStr;
    if (dd.mstream_locked) {
        ipResolved = getHostMap().contains(dd.mstream_ip);
        ipStr = ipResolved ? getHostMap()[dd.mstream_ip] : QHostAddress(dd.mstream_ip).toString();
        mStreamStr = QString("%1:%2").arg(ipStr).arg(dd.mstream_udp_port);
    } else {
        mStreamStr = "free";
    }
    QTableWidgetItem * const mstreamItem = table->item(row, COL_M_STREAM);
    if(mStreamStr != mstreamItem->text()) {
        isChanged = true;
        mstreamItem->setText(mStreamStr);
        mstreamItem->setToolTip(ipResolved && dd.mstream_locked ? QHostAddress(dd.mstream_ip).toString() : "");
    }
    QColor mstreamItemCol = (!dd.mstream_locked || localIp.contains(dd.mstream_ip)) ? Qt::black : (free ? QColor(0xf48000):Qt::red);
    if(mstreamItemCol != mstreamItem->textColor()) {
        isChanged = true;
        mstreamItem->setForeground(mstreamItemCol);
    }

    if(isChanged && !resizeTimer->isActive())
        resizeTimer->start();

#ifdef QT_DEBUG
    checkRows();
#endif
    table->setSortingEnabled(sorting);
}


void DiscoverDialog::setTableHeaders(bool upper)
{
    QTableWidget *table = upper ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents); //resize
    table->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents); //resize
    table->verticalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
    connect(table, &QTableWidget::cellDoubleClicked,
            this, &DiscoverDialog::tableDoubleClick);

    table->setColumnCount(COL_TOTAL);
    table->setHorizontalHeaderItem(COL_ENABLE, new QTableWidgetItem("En"));
    table->setHorizontalHeaderItem(COL_STATE, new QTableWidgetItem("St."));
    table->setHorizontalHeaderItem(COL_DEV_TYPE_TXT, new QTableWidgetItem("Type"));
    table->setHorizontalHeaderItem(COL_SERIAL_ID_TXT, new QTableWidgetItem("Serial ID"));
    table->setHorizontalHeaderItem(COL_FIRM_WARE, new QTableWidgetItem("Firmware"));
    table->setHorizontalHeaderItem(COL_SLOT, new QTableWidgetItem("Slot"));
    table->setHorizontalHeaderItem(COL_IP, new QTableWidgetItem("IP Address"));
    table->setHorizontalHeaderItem(COL_MAC, new QTableWidgetItem("MAC Address"));
    table->setHorizontalHeaderItem(COL_MASTER, new QTableWidgetItem("Master"));
    table->setHorizontalHeaderItem(COL_M_STREAM, new QTableWidgetItem("MStream"));
    table->setHorizontalHeaderItem(COL_TYPE_NUM, new QTableWidgetItem("type num"));
    table->setHorizontalHeaderItem(COL_SERIAL_NUM, new QTableWidgetItem("S/N"));

    table->sortByColumn(COL_DEV_TYPE_TXT, Qt::AscendingOrder);

    table->hideColumn(COL_TYPE_NUM);
    table->hideColumn(COL_SERIAL_NUM);
    if (upper){
        table->hideColumn(COL_ENABLE);
        table->hideColumn(COL_STATE);
    }
    if(!show_mac_adress){
        table->hideColumn(COL_MAC);
    }
}

DeviceIndex DiscoverDialog::getDeviceIndexOfRow(bool upper, int row) const
{
    QTableWidget *table = upper ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    DeviceIndex index;
    Q_ASSERT(row<table->rowCount());
    index.first = table->item(row, COL_TYPE_NUM)->text().toInt();
    index.second = table->item(row, COL_SERIAL_NUM)->text().toLongLong();
    return index;
}

int DiscoverDialog::getRowByIndex(bool upper, DeviceIndex index) const
{
    QTableWidget *table = upper ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    for(int row=0;row<table->rowCount(); ++row){
        if(getDeviceIndexOfRow(upper, row) == index){
            return row;
        }
    }

    return -1;
}

void DiscoverDialog::checkRows() const
{
    return;
    for(int k=0;k<2;++k){
        bool upper = k;
        QTableWidget *table = upper ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
        const QMap<DeviceIndex, DeviceDescription> &list = upper ? availableDevices : selectedDevices;
        int rows = table->rowCount();
        for(int row=0; row<rows; ++row){
            DeviceIndex i = getDeviceIndexOfRow(upper, row);
            bool found = false;
            for (const auto &dd : list) {
                if(dd.getIndex()==i){
                    found = true;
                    break;
                }
            }
            Q_ASSERT(found);
        }
        Q_ASSERT(rows==list.size());
    }
}


QVector<int> DiscoverDialog::getSelectedRows(bool free) const
{
    QVector<int> rows;
    QTableWidget *table = free ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    for (const auto &range : table->selectedRanges()){
        for (int row=range.topRow(); row<=range.bottomRow(); ++row)
            rows.push_back(row);
    }
    std::sort(rows.begin(), rows.end());
    return rows;
}

void DiscoverDialog::setSelected(const DeviceDescription &dd)
{
    QVector<DeviceDescription> list;
    list.push_back(dd);
    setSelected(list);
}

void DiscoverDialog::setSelected(QVector<DeviceDescription> list)
{
    // All selected devices should be in list
    QList<DeviceIndex> toRemove = selectedDevices.keys();
    for (const auto &dd : list) {
        DeviceIndex index = dd.getIndex();
        if(selectedDevices.contains(index)) {
            DeviceDescription &dd_old = selectedDevices[index];
            if(dd_old.enabled != dd.enabled){
                dd_old.enabled = dd.enabled;
                updateTableRow(false, dd_old);
            }
            toRemove.removeAll(index);
        } else {
//            qCritical()<<"In Discovery dialog there are selected devices that wasn't be selected.";
        }
    }
    for (const auto &i : toRemove) {
        moveRow(false, i);
    }

    // Add missing devices
    for (const auto &dd : list) {
        addSelected(dd);
    }
    selectedDevices_backup = selectedDevices;
}

void DiscoverDialog::addSelected(const DeviceDescription &dd)
{
    DeviceIndex index = dd.getIndex();
    if(selectedDevices.contains(index))
        return; // Already in selected list

    updateDeviceTypeInfo(dd);
    if(availableDevices.contains(index)){
        // move from devices to selected
        moveRow(true, index);
        Q_ASSERT(selectedDevices.contains(index));
        if(selectedDevices[index].enabled != dd.enabled) {
            selectedDevices[index].enabled = dd.enabled;
            updateTableRow(false, selectedDevices[index]);
        }
        return;
    }

    // Unknown
    selectedDevices[index] = dd;
    insertRow(false, dd);
    selectedDevices_backup = selectedDevices;
}

void DiscoverDialog::show_mac_adresses()
{
    ui->checkBoxShowMAC->setVisible(true);
}

DeviceDescription DiscoverDialog::getDescriptionByIndex(DeviceIndex index) const
{
    if(selectedDevices.contains(index))
        return selectedDevices[index];
    if(availableDevices.contains(index))
        return availableDevices[index];
    return DeviceDescription();
}

void DiscoverDialog::on_checkBoxShowMAC_clicked(bool checked)
{
    show_mac_adress = checked;

    if(!checked){
        ui->tableWidgetAvailable->hideColumn(COL_MAC);
        ui->tableWidgetSelected->hideColumn(COL_MAC);
    }
    else{
        ui->tableWidgetAvailable->showColumn(COL_MAC);
        ui->tableWidgetSelected->showColumn(COL_MAC);
    }
}

void DiscoverDialog::accept()
{
    QDialog::accept();
    selectedDevices_backup = selectedDevices;
}

void DiscoverDialog::reject()
{
    QDialog::reject();
    setSelected(selectedDevices_backup.values().toVector());
}


void DiscoverDialog::checkOutOfDate()
{
    bool isChanged=false;
    for (const auto &dd : availableDevices.values()) {
        if (dd.isOutOfDate()){
            const DeviceIndex &index = dd.getIndex();
            emit deviceExpired(index);
            if (printInfo)
                qDebug() << QString("Device s/n:%1 (type:%2) is out of date (%3ms)")
                            .arg(dd.getSerialIdStr())
                            .arg(dd.getModelStr())
                            .arg(dd.discoverTimer.elapsed());
            removeRow(true, index);
            availableDevices.remove(index);
            isChanged = true;
        }
    }

    for (const auto &dd : selectedDevices.values()) {
        if (dd.isOutOfDate()) {
            if(!isDevShown(false, dd))
                continue;

            const DeviceIndex &index = dd.getIndex();
            if (!devicePreviousState.contains(index) || devicePreviousState[index] != false){
                emit deviceExpired(index);
                const int row = getRowByIndex(false, index);
                Q_ASSERT(row != -1);
                if (printInfo)
                    qDebug() << QString("Device s/n:%1 (type:%2) is out of date (%3ms)")
                                .arg(dd.getSerialIdStr())
                                .arg(dd.getModelStr())
                                .arg(dd.discoverTimer.elapsed());
                setOnlineState(row, false, index);
            }
            isChanged = true;
        }
    }

    if(isChanged && !resizeTimer->isActive())
        resizeTimer->start();
}

void DiscoverDialog::setDeviceEnabled(bool st)
{
    QWidget *button = (QWidget *)sender();
    for (int row=0;row<ui->tableWidgetSelected->rowCount(); ++row){
        QWidget *widget = ui->tableWidgetSelected->cellWidget(row, COL_ENABLE);

        if (widget->isAncestorOf(button) ){
            DeviceIndex index = getDeviceIndexOfRow(false, row);
            Q_ASSERT(selectedDevices.contains(index));
            selectedDevices[index].enabled = st;
            return;
        }
    }
    // happens when removing row but CheckBox/RadioButton still exists
    qDebug()<<"Can't find device for toggled CheckBox/RadioButton";
    Q_ASSERT(false);
}

void DiscoverDialog::on_pushButtonTest_clicked()
{
//    ui->tableWidgetSelected->updateGeometry(); //resize
    ui->tableWidgetSelected->resizeRowsToContents();
    return;
    int h = ui->tableWidgetSelected->rowHeight(0);
    qDebug()<<"rowHeight:"<<h;
    qDebug()<<ui->tableWidgetSelected->cellWidget(0,0)->height();
    qDebug()<<ui->tableWidgetSelected->cellWidget(0,0)->layout()->itemAt(0)->widget()->height();
    ui->tableWidgetSelected->cellWidget(0,0)->layout()->itemAt(0)->widget()->showMaximized();
    qDebug()<<ui->tableWidgetSelected->cellWidget(0,0)->layout()->itemAt(0)->widget()->height();
}


void DiscoverDialog::on_pushButtonAddDiscoverDev_clicked()
{
    moveRows(true);
}

void DiscoverDialog::on_pushButtonRemoveDiscoverDev_clicked()
{
    moveRows(false);
}

void DiscoverDialog::printCols(bool upper, int col) const
{
    if (!printInfo)
        return;
    QTableWidget *table = upper ? ui->tableWidgetAvailable : ui->tableWidgetSelected;
    for (int row=0; row<table->rowCount(); ++row){
        QTableWidgetItem *item = table->item(row, col);
        if (item){
            QString s= item->text();
            qDebug()<<row<<col<<":"<<s;
        }
    }
}

void DiscoverDialog::on_listWidgetTypes_itemSelectionChanged()
{
    QSet<quint16> shownDevices;
    for (auto item : ui->listWidgetTypes->selectedItems()) {
        int id = item->data(Qt::UserRole).toUInt();
        shownDevices.insert(id);
    }
//    qDebug()<<hex<<shownDeices;
    if(deviceIdShown != shownDevices){
        deviceIdShown = shownDevices;
        updateShownDevices();
    }
}

void DiscoverDialog::on_listWidgetSubnet_itemSelectionChanged()
{
    QSet<quint32> subnetSet;
    for (auto item : ui->listWidgetSubnet->selectedItems()) {
        const quint32 &subnet = item->data(Qt::UserRole).toUInt();
        subnetSet.insert(subnet);
    }
    if(subnetShown != subnetSet) {
        subnetShown = subnetSet;
        updateShownDevices();
    }
}

void DiscoverDialog::on_pushButtonManualAdd_clicked()
{
    ManualAddDialog addDialog(this);
    addDialog.setDeviceIdMap(deviceTypeMap);
    if(addDialog.exec() == QDialog::Accepted){
        DeviceDescription dd(addDialog.getDeviceId(), addDialog.getSerial());
        addSelected(dd);
    }
}

void DiscoverDialog::tableDoubleClick(int row, int col)
{
    Q_UNUSED(col);
    bool upper = sender() == ui->tableWidgetAvailable;
    Q_ASSERT(upper || sender() == ui->tableWidgetSelected);
    DeviceIndex index = getDeviceIndexOfRow(upper, row);
    moveRow(upper, index);
}

void DiscoverDialog::resizeTables()
{
    ui->tableWidgetSelected->resizeRowsToContents();
    ui->tableWidgetAvailable->resizeRowsToContents();
    ui->tableWidgetSelected->resizeColumnsToContents();
    ui->tableWidgetAvailable->resizeColumnsToContents();
}

void DiscoverDialog::checkHostInfoUpdates()
{
    static DnsReverseMap hostMapCache;
    const DnsReverseMap &hostMap = dnsLookup.getHostMap();
    if (hostMap == hostMapCache) return;
    hostMapCache = hostMap;
    updateHostInfo(ui->tableWidgetAvailable);
    updateHostInfo(ui->tableWidgetSelected);
}
