#include "FlashDev.h"

#include <chrono>
#include <thread>
#include <iostream>
#include <cstdio>
#include <cstring>

#include "FlashDevTypes.h"

using namespace std::chrono;

namespace {
std::string formatTime(int64_t ms)
{
    std::string strStr;

    int64_t s = ms/1000;
    if(ms==0 || (ms%=1000)!=0) {
        strStr += std::to_string(ms) + "ms";
    }

    int64_t min = s/60;
    s %=60;
    if(s)
        strStr = std::to_string(s) +"sec "+strStr;
    if(min)
        strStr = std::to_string(min) +"min "+strStr;

    return strStr;
}

int64_t elapsedMs(steady_clock::time_point tStart) {
    auto tNow = steady_clock::now();
    int64_t int_ms = duration_cast<milliseconds>(tNow - tStart).count();
    return int_ms;
}
}

int FlashDev::findBitFileHeader(const char *fileData, int dataLen)
{
    const uint32_t *valSync;
    const uint32_t BIT_FILE_SYNC_WORD = 0x665599AA;
    const int headerSizeMax = std::min(dataLen-static_cast<int>(sizeof(uint32_t)), 256);
    for(int headerSize = 0; headerSize<headerSizeMax; ++headerSize) {
        valSync = reinterpret_cast<const uint32_t*>(fileData+headerSize);
        if(*valSync == BIT_FILE_SYNC_WORD) {
            std::printf("findBitFileHeader: %#x(%d)\n", headerSize, headerSize);
            int printPos=std::max(headerSize-8, 0);
            int printPosMax=std::min(headerSize+4*10,
                                     dataLen-static_cast<int>(sizeof(uint32_t)));
            for(;printPos<printPosMax; printPos+=4) {
                std::printf("%#x: %08x\n", printPos,
                            *reinterpret_cast<const uint32_t*>(fileData+printPos));
            }
            return headerSize;
        }
    }
    return -1;
}

bool FlashDev::startWrite(uint32_t id, const char *fileData, uint32_t dataLen,
                          uint32_t headerSize, bool isGoldenImage, bool verify)
{
    if(!enabled)
        return false;

    std::printf("%s started to write id:%d\n", getPrefix().c_str(), id);

    cmdPerformed = 0;
    fileHeaderSize = headerSize;
    fullSize = dataLen;

    if(!checkFileSize(isGoldenImage)) {
        std::fprintf(stderr, "file is too big\n");
        return false;
    }
    if((fullSize-fileHeaderSize)<MAX_BLOCK_SIZE) {
        std::fprintf(stderr, "file is too small\n");
        return false;
    }

    writeOk=true;

    // Erasing ...
    writeOk &= sectorErase(isGoldenImage, fullSize-fileHeaderSize);

    // Writing ...
    if(writeOk && !writeFlash(fileData, isGoldenImage)) {
        writeOk &= false;
        std::fprintf(stderr, "Writing flash was failed\n");
    }

    // Readback ...
    if(verify && writeOk && !verifyFlash(isGoldenImage, fullSize, fileData)) {
        writeOk &= false;
        std::fprintf(stderr, "Readback flash was failed\n");
    }

    return writeOk;
}

void FlashDev::patchGoldenImage(void *data) const
{
    const uint32_t word8_wd_timer  = 0x00000140;
    const uint32_t word10_wbstar = 0x00008000;
    const uint32_t word12_iprog = 0x0f000000;
    uint32_t *val;
    const uint32_t dummy_word_shift = 0;

    val = reinterpret_cast<uint32_t*>(data)+8+dummy_word_shift;
    std::printf("Change 8th word from '%8x' to '%8x'\n", *val, word8_wd_timer);
    *val = word8_wd_timer;

    val = reinterpret_cast<uint32_t*>(data)+10+dummy_word_shift;
    std::printf("Change 10th word from '%8x' to '%8x'\n", *val, word10_wbstar);
    *val = word10_wbstar;

    val = reinterpret_cast<uint32_t*>(data)+12+dummy_word_shift;
    std::printf("Change 12th word from '%8x' to '%8x'\n", *val, word12_iprog);
    *val = word12_iprog;
}

void FlashDev::reverseByteOrder(std::vector<uint16_t> &dataVec)
{
    uint16_t vTmp;
    for(uint32_t i=0; i<dataVec.size(); i += 2) {
        uint16_t &v1 = dataVec[i];
        uint16_t &v2 = dataVec[i+1];

        vTmp = ((v2&0xFF)<<8) | (v2>>8);
        v2 = ((v1&0xFF)<<8) | (v1>>8);
        v1 = vTmp;
    }
}

bool FlashDev::waitCmdExecution(int timeoutMs, bool extraWaitCheck)
{
    if(timeoutMs==0)
        timeoutMs = WRITE_TIMEOUT_MS;
    auto commandTimer = steady_clock::now();
    uint16_t readBackVal;
    int64_t elapsedTime;
    while(true) {
        readBackVal = devReadReg(getRegBase()+REG_READBACK);
        if((readBackVal >> 14) == 0 && (!extraWaitCheck || (readBackVal&0xFF)==0x80))
            break;
        elapsedTime = elapsedMs(commandTimer);
        if(elapsedTime > timeoutMs) {
            std::fprintf(stderr, "Command execution timeouted(%d ms): readBackVal=%#x\n",
                         timeoutMs, readBackVal);
            return false;
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }
    return true;
}

bool FlashDev::configureFlash()
{
    flashConfigured = false;
    // Configure flash
    // SPI single, default drive strength (see flash manual)
    devWriteReg(getRegBase()+REG_FLASH_ADDR, 0xef);
    devWriteReg(getRegBase()+REG_CMD, CMD_WRITE_EN);
    if(!waitCmdExecution()) {
        std::fprintf(stderr, "[Configure flash] CMD_WRITE_EN failed\n");
        return false;
    }
    ++cmdPerformed;
    if(!waitCmdExecution()) {
        std::fprintf(stderr, "[Configure flash] Write enable hasn't passed\n");
        return false;
    }

    devWriteReg(getRegBase()+REG_CMD, CMD_WRVECR);
    ++cmdPerformed;
    if(!waitCmdExecution()) {
        std::fprintf(stderr, "[Configure flash] Failed to write preparing command\n");
        return false;
    }

    ++cmdPerformed;
    devWriteReg(getRegBase()+REG_CMD, CMD_RDID);
    if(!waitCmdExecution()) {
        std::fprintf(stderr, "[Configure flash] Read JEDEC_ID failed\n");
        return false;
    }
    if(canReadFlashId()) {
        updateFlashIdData();
        if(!flashIdData.isValid()) {
            std::fprintf(stderr, "[Configure flash] Invalid flash id data\n");
            return false;
        }
    } else {
        uint16_t val = devReadReg(getRegBase()+REG_READBACK);
        if(val != JEDEC_ID_OK) {
            std::fprintf(stderr, "[Configure flash] Unsupported JEDEC_ID (%#x), should be JEDEC_ID_OK(%#x)\n",
                         val, JEDEC_ID_OK);
            return false;
        }
    }

    flashConfigured = true;
    return true;
}

bool FlashDev::sectorErase(bool isGoldenImage, uint32_t sizeToErase)
{
    if(!flashConfigured && !configureFlash()){
        std::fprintf(stderr, "Failed to write preparing command\n");
        return false;
    }

    std::printf("Start sector erasing...\n");
    bool allOk = true;
    auto eraseTime = steady_clock::now();

    sendProgress(0);
    const uint32_t posOffset = getImageOffset(isGoldenImage ? 0 : 1);
    uint32_t erasePos = 0;
    while(erasePos < sizeToErase) {
        if(!isConnected()) {
            allOk = false;
            std::fprintf(stderr, "Device has gone offline on sector erasing. Aborting...\n");
            break;
        }
        if(!enabled) {
            allOk = false;
            std::fprintf(stderr, "Device was disabled on sector erasing. Aborting...\n");
            break;
        }

        devWriteReg32(getRegBase()+REG_FLASH_ADDR, erasePos+posOffset);
        devWriteReg(getRegBase()+REG_CMD, CMD_WRITE_EN);
        ++cmdPerformed;
        if(!waitCmdExecution()) {
            std::cerr<<"err in sectorErase::CMD_WRITE_EN";
            allOk = false;
            break;
        }
        devWriteReg(getRegBase()+REG_CMD, CMD_SE);
        ++cmdPerformed;
        if(!waitCmdExecution(SECTOR_ERASE_TIMEOUT, true)) {
            std::cerr<<"err in sectorErase::CMD_SE";
            allOk = false;
            break;
        }
        read_rdsr(); // delay after sector erase
        erasePos += SECTOR_SIZE;

        sendProgress(static_cast<int>(100*erasePos/sizeToErase));
        if(!allOk)
            break;
    }
    std::printf("Sector erase %s in %s\n",
                (allOk ? "finished" : "failed"),
                formatTime(elapsedMs(eraseTime)).c_str());
    return allOk;
}

bool FlashDev::bulkErase()
{
    std::printf("Start bulk erasing...\n");
    auto eraseTime = steady_clock::now();
    bool allOk=true;

    devWriteReg(getRegBase()+REG_CMD, CMD_WRITE_EN);
    ++cmdPerformed;
    if(!waitCmdExecution()) {
        allOk = false;
        std::cerr <<"err in bulkErase::CMD_WRITE_EN";
    }

    if(allOk) {
        devWriteReg(getRegBase()+REG_CMD, CMD_BE);
        ++cmdPerformed;
        if(!waitCmdExecution(BULK_ERASE_TIMEOUT)){
            allOk = false;
            std::cerr <<"Failed wait end of bulk erase.";
        }
    }

    std::printf("Bulk erase %s in %s\n",
                allOk ? "finished" : "failed",
                formatTime(elapsedMs(eraseTime)).c_str());
    return allOk;
}

bool FlashDev::writeFlash(const char *fileData, bool isGoldenImage)
{
    if(!flashConfigured && !configureFlash()){
        std::fprintf(stderr, "Failed to write preparing command\n");
        return false;
    }
    std::printf("Start flash writting...\n");
    bool allOk=true;
    auto writeTime = steady_clock::now();

    sendProgress(0);
    uint32_t wroteSize = 0; // bytes
    const uint32_t addrBase = getImageOffset(isGoldenImage ? 0 : 1);

    const size_t count=MAX_BLOCK_SIZE/sizeof(uint16_t); // in 16-bit words
    std::vector<uint16_t> blockData;
    blockData.resize(count);
    bool atEnd=false;
    uint32_t blockSize=0; // bytes
    while(fileHeaderSize+wroteSize < fullSize) {
        if(!isConnected()) {
            allOk = false;
            std::fprintf(stderr, "Device has gone offline on writing bit-file. Aborting...\n");
            return false;
        }
        if(!enabled) {
            allOk = false;
            std::fprintf(stderr, "Device was disabled on writing bit-file. Aborting...\n");
            break;
        }

        if(atEnd) {
            blockSize = MAX_BLOCK_SIZE;
            memcpy(&blockData[0], fileData+fileHeaderSize, blockSize);
            if(isGoldenImage)
                patchGoldenImage(blockData.data());
        } else {
            blockSize = std::min(MAX_BLOCK_SIZE, fullSize-fileHeaderSize-MAX_BLOCK_SIZE-wroteSize);
            if(blockSize != MAX_BLOCK_SIZE)
                memset(&blockData[0], 0, MAX_BLOCK_SIZE);
            memcpy(&blockData[0], fileData+fileHeaderSize+MAX_BLOCK_SIZE+wroteSize, blockSize);
            if(blockSize+sizeof(uint32_t)<= MAX_BLOCK_SIZE) {
                blockData[blockSize/sizeof(uint16_t)] = 0x2a50;
                blockData[blockSize/sizeof(uint16_t)+1] = 0x4a90;
            }
        }
        reverseByteOrder(blockData);

        uint32_t addr[count];
        for(uint32_t w=0; w<count; ++w)
            addr[w] = getRegBase()+REG_DATA+w;
        bool blkWriteOk = devBlkWriteReg(count, addr, blockData.data());
        if(!blkWriteOk) {
            std::fprintf(stderr, "Error while perform block write of data\n");
            return false;
        }
//        memWriteBlk(MEM_DATA, dataVec);

        devWriteReg32(getRegBase()+REG32_DATA_ADDR, atEnd ? addrBase : (addrBase + MAX_BLOCK_SIZE + wroteSize));

        for(int i=0; i<1 && allOk; ++i) {
            devWriteReg(getRegBase()+REG_CMD, CMD_WRITE_EN);
            ++cmdPerformed;
            if(!waitCmdExecution()) {
                std::fprintf(stderr, "err in writeFlash::CMD_WRITE_EN\n");
                allOk = false;
                break;
            }
            devWriteReg(getRegBase()+REG_CMD, CMD_WRITE);
            ++cmdPerformed;
            if(!waitCmdExecution(0, true)) {
                allOk = false;
                std::fprintf(stderr, "err in writeFlash::CMD_WRITE\n");
                break;
            }
            read_rdsr(); // delay after write
        }
        wroteSize += blockSize;
        sendProgress(100*wroteSize/(fullSize-fileHeaderSize));
        if(fullSize- (fileHeaderSize+wroteSize) <= MAX_BLOCK_SIZE) {
            atEnd = true;
        }

        if(!allOk)
            break;
    }

    std::printf("Wrote %d of %d bytes (+%d bytes in header) in %s.\n",
               wroteSize, fullSize-fileHeaderSize, fileHeaderSize,
               formatTime(elapsedMs(writeTime)).c_str());
    return allOk;
}

bool FlashDev::verifyFlash(bool isGoldenImage, uint32_t sizeToRead, const char *data)
{
    if(!flashConfigured && !configureFlash()){
        std::fprintf(stderr, "Failed to write preparing command\n");
        return false;
    }

    std::printf("Start verify...\n");
    bool allOk = true;
    auto readTime = steady_clock::now();

    uint32_t startAddr = getImageOffset(isGoldenImage ? 0 : 1);

    const size_t maxBlockCount=MAX_BLOCK_SIZE/sizeof(uint16_t); // in 16-bit words
    std::vector<uint16_t> readbackData;
    readbackData.resize(maxBlockCount);
    std::vector<uint16_t> originalData;
    originalData.resize(maxBlockCount);

    uint32_t addr[maxBlockCount];
    for(uint32_t w=0; w<maxBlockCount; ++w)
        addr[w] = getRegBase()+REG_DATA+w;

    sendProgress(0);
    uint32_t readSize=0, curBlockSize=0;
    while(fileHeaderSize+readSize < sizeToRead) {
        if(!isConnected()) {
            allOk = false;
            std::fprintf(stderr, "Device has gone offline on verifyFlash. Aborting...\n");
            break;
        }
        if(!enabled) {
            allOk = false;
            std::fprintf(stderr, "Device was disabled on verifyFlash. Aborting...\n");
            break;
        }
        curBlockSize = std::min(MAX_BLOCK_SIZE, sizeToRead-fileHeaderSize-readSize);

        devWriteReg32(getRegBase()+REG_FLASH_ADDR, startAddr + readSize);
        devWriteReg(getRegBase()+REG_CMD, CMD_READ);
        if(!waitCmdExecution(READ_TIMEOUT_MS)) {
            std::fprintf(stderr, "err in readbackFlash::CMD_READ\n");
            allOk = false;
            break;
        }
        ++cmdPerformed;

        bool blkReadOk = devBlkReadReg(MAX_BLOCK_SIZE/sizeof(uint16_t), addr, readbackData.data());
        if(!blkReadOk) {
            std::fprintf(stderr, "Error while perform block read of data in verifyFlash\n");
            return false;
        }

        if(curBlockSize != MAX_BLOCK_SIZE) {
            memset(&originalData[0], 0, MAX_BLOCK_SIZE);
            originalData[curBlockSize/sizeof (uint16_t)] = 0x2a50;
            originalData[curBlockSize/sizeof (uint16_t) + 1] = 0x4a90;
        }
        memcpy(&originalData[0], data+fileHeaderSize+readSize, curBlockSize);
        if(isGoldenImage && readSize==0) {
            patchGoldenImage(originalData.data());
        }

        allOk &= memcmp(&originalData[0], readbackData.data(), curBlockSize)==0;
        if(!allOk) {
            if(!allOk) printf("\nError block:\n");
            if(readSize==0) printf("First block:\n");
            if(readSize+curBlockSize+fileHeaderSize==sizeToRead)
                printf("Last block:\n");
            for(uint32_t w=0; w<MAX_BLOCK_SIZE/sizeof(uint16_t); ++w) {
                const uint16_t &v1 = originalData[w];
                const uint16_t &v2 = readbackData[w];
                printf("%#4x+%#4x: %4x %s %4x\n", readSize,
                       addr[w]-getRegBase()-REG_DATA, v1, v1==v2?"==":"!=", v2);
            }
        }
        readSize += curBlockSize;

        sendProgress(100*readSize/(sizeToRead-fileHeaderSize));
        if(!allOk)
            break;
    }
    std::printf("Verify flash %s in %s\n", (allOk ? "finished" : "failed"),
                formatTime(elapsedMs(readTime)).c_str());
    return allOk;
}

void FlashDev::readStat()
{
    if(!isConnected() )
        return;
    uint16_t statVal1 = devReadReg(getRegBase()+REG_STAT_1);
    uint16_t statVal2 = devReadReg(getRegBase()+REG_STAT_2);

    devWriteReg(getRegBase()+REG_CMD, CMD_RDSR);
    uint16_t val = devReadReg(getRegBase()+REG_READBACK);

    std::printf("Statistic: cmdPerformed=%d reg(%#x)=%d; reg(%#x)=%d; RDSR=%#x\n",
                cmdPerformed,
                getRegBase()+REG_STAT_1, statVal1,
                getRegBase()+REG_STAT_2, statVal2,
                val);
}

uint16_t FlashDev::read_rdsr()
{
    devWriteReg(getRegBase()+REG_CMD, CMD_RDSR);
    uint16_t val = devReadReg(getRegBase()+REG_READBACK);
//    std::printf("rdsr = %x\n", val);
    return val;
}

bool FlashDev::checkFileSize(bool isGoldenImage) {
    const uint img_size = fullSize-fileHeaderSize;
    return img_size < (isGoldenImage ? getFlashSize() : getImageOffset(1));
}

void FlashDev::updateFlashIdData()
{
    const int data_size = sizeof(FlashIdData)/sizeof(uint16_t);
    uint32_t addr[data_size];
    const uint32_t data_offset = getRegBase()+REG_DATA+MAX_BLOCK_SIZE/sizeof(uint16_t)-data_size;
    for(uint32_t w=0; w<data_size; ++w)
        addr[w] = data_offset+w;

    bool blkReadOk = devBlkReadReg(data_size, addr, reinterpret_cast<uint16_t*>(&flashIdData));
    if(!blkReadOk) {
        std::fprintf(stderr, "Error while perform block read of data in verifyFlash\n");
        return;
    }

    const auto capacity = flashIdData.getFlashCapacity_Mib();
    std::fprintf(stderr, "%s Flash capacity: %dMib (%dMiB)\n",
                 getPrefix().c_str(), capacity, capacity/8);
}

uint32_t FlashDev::getFlashSize() const
{
    if(!flashIdData.isValid())
        return DEV_FULL_MEM;

    //                              from Mi-bits to bytes
    return flashIdData.getFlashCapacity_Mib()/8*0x100000;
}

uint32_t FlashDev::getImageOffset(int image_index) const
{
    return image_index ? getFlashSize()/2 : 0;
}
