#ifndef FLASHDEV_H
#define FLASHDEV_H

#include <vector>
#include <stdint.h>
#include <string>

struct FlashIdData
{
    std::string getMemoryType() const
    {
        switch (memoryType) {
        case 0xBA: return "3V";
        case 0xBB: return "1.8V";
        default: return "unknown";
        }
    }

    uint32_t getFlashCapacity_Mib() const
    {
        if(isValid()) {
            if(memoryCapacity>=0x17 && memoryCapacity<=0x19)
                return 1 << (memoryCapacity-0x11);
            if(memoryCapacity>=0x20 && memoryCapacity<=0x22)
                return 1 << (memoryCapacity-0x17);
        }
        return 0;
    } // in Mb

    bool isValid() const
    {
        return  manufacturerID==0x20 &&
                remaingDataSize==0x10 &&
                deviceConfigInfo==0x0;
    }

    uint8_t manufacturerID = 0;
    uint8_t memoryType = 0;
    uint8_t memoryCapacity = 0;
    uint8_t remaingDataSize = 0;
    uint8_t extendedId = 0;
    uint8_t deviceConfigInfo = 0;
    uint8_t customizedFactoryData[14] = {0};
};

class FlashDev
{
public:
    virtual uint32_t getRegBase() const = 0;
    static int findBitFileHeader(const char *fileData, int dataLen);

protected:
    bool startWrite(uint32_t id, const char *fileData, uint32_t dataLen,
                    uint32_t headerSize, bool isGoldenImage=false, bool verify=true);
    void patchGoldenImage(void *data) const;
    void reverseByteOrder(std::vector<uint16_t> &dataVec);
    bool waitCmdExecution(int timeoutMs=0, bool extraWaitCheck = false);
    bool configureFlash();
    virtual bool sectorErase(bool isGoldenImage, uint32_t sizeToErase);
    bool bulkErase();

    virtual bool writeFlash(const char *fileData, bool isGoldenImage=false);
    virtual bool verifyFlash(bool isGoldenImage, uint32_t sizeToRead, const char *data);
    void readStat();
    uint16_t read_rdsr();
    virtual bool isConnected() const {return true;}
    virtual uint16_t devReadReg(int regAddr) = 0;
    virtual void devWriteReg(int regAddr, uint16_t data) = 0;
    virtual void devWriteReg32(int regAddr, uint32_t data) = 0;
    virtual bool devBlkWriteReg(uint32_t count, const uint32_t *addr, uint16_t *data) = 0;
    virtual bool devBlkReadReg(uint32_t count, const uint32_t *addr, uint16_t *data) = 0;

    virtual void sendProgress(int p) = 0;
    virtual std::string getPrefix() const = 0;
    virtual bool checkFileSize(bool isGoldenImage);
    virtual bool canReadFlashId() const {return false;}
    void updateFlashIdData();
    uint32_t getFlashSize() const;
    uint32_t getImageOffset(int image_index) const;

    bool flashConfigured = false;
    int cmdPerformed = 0;
    uint32_t fileHeaderSize = 0;
    uint32_t fullSize = 0;
    bool enabled = true;
    bool writeOk = true;
    FlashIdData flashIdData;
};

#endif // FLASHDEV_H
