#include "HistWorker.h"
#include <QSemaphore>
#include <QCoreApplication>

#include <DeviceData.h>
#include "DeviceHelper.h"

HistWorker::HistWorker(QSemaphore *_histSem) :
    QObject(nullptr),
    histSem(_histSem),
    taiReperSec(0),
    taiReperNs(0),
    data_processing(false),
    destroy_flg(false),
    curEv(nullptr)
{
    qRegisterMetaType<HistDevType>();
}

HistWorker::~HistWorker()
{
#ifdef TEST_PLL_DOUBLE_PULSE
    flushTestDp();
#endif
    destroy_flg = true;
    trashAllData();
}

void HistWorker::requestInlResults(bool allRes) const
{
    if(allRes) {
        emit inlResultsUpdated(inlResults, allRes);
        return;
    }
    InlResults inlRes;
    if(HistId::inl == -1) return;

    QVector<double> codeDensData(1024);
    for(DeviceIndex index: histContainer.inlHist.keys()){
        DeviceCalibration devCal = inlResults.value(index, DeviceCalibration(index));
        for(int ch: histContainer.inlHist[index].keys()){
            const Histogramm &cutHist = histContainer.inlHist[index][ch];
            const int nBins = cutHist.nBins;
            if(nBins != 1024) continue;

            for (int bin = 0; bin < nBins; ++bin) {
                codeDensData[bin] = cutHist.bins[bin];
            }
//            inlRes[index][ch] = makeDNLCorrectionTable(codeDensData);
            devCal.inl[ch] = codeDensData;
            codeDensData.clear();
            codeDensData.resize(1024);
        }
        if(!devCal.inl.isEmpty()) {
            if(devErrors.contains(index))
                devCal.errMask = devErrors[index];
            inlRes[index] = devCal;
        }
    }
    emit inlResultsUpdated(inlRes);
}

void HistWorker::tdcTtvxMapUpdated(TdcTtvxsMapType ttMap)
{
    tdcTtvxsMap = ttMap;
}

void HistWorker::hist_cfg_updated(const HistMngrCfg &newCfg)
{
    cfg = newCfg;
    for(const HistCfg &hc : newCfg.histCfg) {
        if(histContainer.baseHist.contains(hc.groupId)) {
            Histogramm &hist = histContainer.baseHist[hc.groupId];
            if(hist.nBins != hc.nBins ||
                    !qFuzzyCompare(hist.min, hc.min) ||
                    !qFuzzyCompare(hist.max, hc.max)) {
                histContainer.baseHist[hc.groupId].resize(hc.nBins, hc.min, hc.max);
            }
        } else {
            histContainer.baseHist.insert(hc.groupId, Histogramm(hc.title, hc.nBins, hc.min, hc.max));
            qInfo() << hc.groupId << "Create hist:" << hc.title << hc.nBins;
        }
    }
}

void HistWorker::inlConfigUpdated(InlResults inl)
{
    curDevCorrMap = inl;
}

void HistWorker::init(const HistMngrCfg &cfg)
{
    hist_cfg_updated(cfg);
}

void HistWorker::reset()
{
    knownDevices.clear();
    cfg.devChOffset.clear();
    histContainer.inlHist.clear();
    histContainer.inlCheckHist.clear();
    clearAll();
    emit resetHistMngr();
}

void HistWorker::clearAll()
{
    for (auto &hId: histContainer.baseHist.keys())
        histContainer.baseHist[hId].reset();
    for (auto &index: histContainer.inlHist.keys()) {
        auto &inlHists = histContainer.inlHist[index];
        for (auto &ch: inlHists.keys())
            inlHists[ch].reset();
    }
    for (auto &index: histContainer.inlCheckHist.keys()) {
        auto &inlCheckHists = histContainer.inlCheckHist[index];
        for (auto &ch: inlCheckHists.keys())
            inlCheckHists[ch].reset();
    }
    recreateTaiHist(0);
    inlResults.clear();
}

void HistWorker::clearHist(int id, DeviceIndex index, int ch)
{
    // TODO: find-out for what purpuse of full cleaning
    for (auto &hId: histContainer.baseHist.keys())
        histContainer.baseHist[hId].reset();
    for (auto &index: histContainer.inlHist.keys()) {
        auto &inlHists = histContainer.inlHist[index];
        for (auto &ch: inlHists.keys())
            inlHists[ch].reset();
    }
    for (auto &index: histContainer.inlCheckHist.keys()) {
        auto &inlCheckHists = histContainer.inlCheckHist[index];
        for (auto &ch: inlCheckHists.keys())
            inlCheckHists[ch].reset();
    }
    if(id==-1 || id == HistId::inl) {
        if(index.isNull())
            histContainer.inlHist.clear();
        else if(histContainer.inlHist.contains(index)){
            auto &devHists = histContainer.inlHist[index];

        }
    }
    if(id==-1 || id == HistId::inlCheck) {

    }
    recreateTaiHist(0);
    refreshCurrentPlot(true);
    inlResults.clear();
}

void HistWorker::handle_data()
{
    data_processing = true;
    while(!data_list.isEmpty()) {
        if(destroy_flg || !cfg.globalEnabled) {
            trashAllData();
            break;
        }
        DecodedData data = data_list.takeFirst();
        histSem->release();

        int tdcHits;
        int ev_cnt=0;
        auto evIt = data.events.begin();
        while(evIt != data.events.end()) {
            CompleteEvent &ev = evIt.value();
            if(++ev_cnt>1000) {
                ev_cnt = 0;
                // for quick handle clean & destroy signals
                QCoreApplication::processEvents(QEventLoop::AllEvents, 10);
            }
            if(destroy_flg)
                break;
            tdcHits=0;
            dataProcessingSingleEv(ev, tdcHits);

            if(histContainer.baseHist.contains(HistId::tdcHitEv) && cfg.histCfg.value(HistId::tdcHitEv).histGroupEn) {
                auto &hTdcHitEv = histContainer.baseHist[HistId::tdcHitEv];
                hTdcHitEv.fill(tdcHits);
            }
            ++evIt;
        }
        if(!destroy_flg)
            refreshCurrentPlot();
    }
    data_processing = false;
}

void HistWorker::dataUpdated(const DecodedData &data)
{
    if(destroy_flg) {
        histSem->release();
        return;
    }

    data_list << data;
    if(data_processing)
        return;
    handle_data();
}

void HistWorker::refreshCurrentPlot(bool force)
{
    emit hist_updated(histContainer, force);
}

void HistWorker::addDevice(const DeviceIndex &index)
{
    if(knownDevices.contains(index)) return;

    auto devChCnt = getDevChCntMap();

    const int nch = getDeviceChNum(index);
    // Add INL/INL-check hist
    for(int ch=0; ch<nch; ++ch) {
        if(HistId::inl!=-1 && !histContainer.inlHist[index].contains(ch)) {
            const auto &hCfg = cfg.histCfg[HistId::inl];
            QString title = QString("%1 %2 ch:%3").arg(hCfg.title).arg(index.getIdent()).arg(ch);
            histContainer.inlHist[index].insert(ch, Histogramm(title, hCfg.nBins, hCfg.min, hCfg.max));
        }
        if(HistId::inlCheck!=-1 && !histContainer.inlCheckHist[index].contains(ch)) {
            const auto &hCfg = cfg.histCfg[HistId::inlCheck];
            QString title = QString("%1 %2 ch:%3").arg(hCfg.title).arg(index.getIdent()).arg(ch);
            histContainer.inlCheckHist[index].insert(ch, Histogramm(title, hCfg.nBins, hCfg.min, hCfg.max));
        }
    }

    ///TODO:: find-out this asserrtt
//    Q_ASSERT(!cfg.devChOffset.contains(index));
    knownDevices << index;
    int devChOffset = 0;
    int maxChNum = cfg.getMaxChNum();

    auto it = cfg.devChOffset.find(index);
    if(it == cfg.devChOffset.end()) {
        devChOffset = maxChNum;
        maxChNum += getDeviceChNum(index);
        cfg.devChOffset[index] = devChOffset;
    } else {
        devChOffset = it.value();
    }

    Histogramm *chHist = getHistById(HistId::tdcCh);
    if(chHist) {
        chHist->resize(maxChNum, 0, maxChNum);
        fillDevChHist(devChCnt);
        QString &tool_tip = cfg.toolTipSuffix[HistId::tdcCh];
        if(!tool_tip.isEmpty())
            tool_tip.append('\n');
        tool_tip.append(QString("%1-%2: %3").arg(devChOffset).arg(maxChNum).arg(index.getIdent()));
    }
    Histogramm *hitHist = getHistById(HistId::tdcHitEv);
    if(hitHist) {
        auto &bins = hitHist->bins;
        auto old_bins = bins;
        int maxHit = maxChNum ? maxChNum*10 : 100;
        hitHist->resize(maxHit, 0, maxHit);
        old_bins.resize(bins.size());
        bins = old_bins;
    }

    emit devListUpdated(cfg.devChOffset, histContainer);
}

void HistWorker::dataProcessingSingleEv(CompleteEvent &ev, int &tdcHitsPerEv)
{
    curEv = &ev;
    TdcEvent tdcStopEvent;
    if(!cfg.stopDevIndex.isNull() && ev.tdcEv.contains(cfg.stopDevIndex)) {
        TdcEvent tdcEvent = ev.tdcEv[cfg.stopDevIndex];
        if(tdcEvent.tdcHits.contains(cfg.stopCh) &&
                !tdcEvent.tdcHits[cfg.stopCh].isEmpty()) {
            tdcStopEvent = tdcEvent;
        }
    }
//    {
//        // Logic to get first
//        DeviceIndex customStopDev(0xd0,0x0aa13347);
//        auto it=ev.tdcEv.find(customStopDev);
//        if(it != ev.tdcEv.constEnd()){
//            auto &customStopEv = it.value();
//            auto itCh=customStopEv.tdcHits.constBegin();
//            while(itCh!=customStopEv.tdcHits.constEnd()) {
//                if(!itCh.value().isEmpty()){
//                    tdcStopEvent = customStopEv;
//                    auto ch = static_cast<quint16>(itCh.key());
//                    if(ch != cfg.stopCh) {
//                        qInfo().noquote() << QString("cfg.stopCh: %1 -> %2")
//                                             .arg(cfg.stopCh,2).arg(ch,2);
//                        cfg.stopCh = ch;
//                    }
//                    break;
//                }
//                ++itCh;
//            }
//        }
//    }

    auto &stopHits = tdcStopEvent.tdcHits[cfg.stopCh];
    auto tdcEvIt = ev.tdcEv.begin();
    while(tdcEvIt != ev.tdcEv.end()) {
        TdcEvent &tdcEv = tdcEvIt.value();
        if(!knownDevices.contains(tdcEv.index)) {
            addDevice(tdcEv.index);
        }
        dataProcessingTdc(tdcEv, tdcHitsPerEv, tdcStopEvent, stopHits);
        ++tdcEvIt;
    }
    curEv = nullptr;
}

void HistWorker::dataProcessingTdc(TdcEvent &ev, int &tdcHitsPerEv, TdcEvent &tdcStopEvent, QVector<TdcChHit> &stopHits)
{
    const int &chOffset = cfg.devChOffset[ev.index];
    // TAI processing
    if(ev.tai.flags == 2 &&
            histContainer.baseHist.contains(HistId::tdcTaiTime) &&
            cfg.histCfg.value(HistId::tdcTaiTime).histGroupEn) {
        Histogramm &hist = histContainer.baseHist[HistId::tdcTaiTime];

        if(hist.nBins == 0) {
            taiReperSec = ev.tai.sec;
            taiReperNs = 0;
            hist = recreateTaiHist(cfg.taiTsWindow);
        }
        double dt = ev.tai.sec-taiReperSec;
        dt += (ev.tai.nSec - taiReperNs)*1e-9;
        hist.fill(dt);
    }

    // TDC processing
    auto tdcChIt = ev.tdcHits.begin();
    while(tdcChIt != ev.tdcHits.end()) {
        const auto &ch = tdcChIt.key();
        QVector<TdcChHit> &tdcHits = tdcChIt.value();
        tdcHitsPerEv += tdcHits.size();

        auto tdcChHitIt = tdcHits.begin();
        while (tdcChHitIt != tdcHits.end()) {
            TdcChHit &tdcHit = *tdcChHitIt;
            Histogramm *hist = getHistById(HistId::tdcCh);
            if(hist && cfg.histCfg.value(HistId::tdcCh).histGroupEn)
                hist->fill(ch+chOffset);
            hist = getHistById(HistId::tdcTime);
            if(hist && cfg.histCfg.value(HistId::tdcTime).histGroupEn) {
                if(tdcHit.hasLeadTimeInl)
                    hist->fill(tdcHit.leadTimeInl);
                else if(tdcHit.hasTrailTimeInl)
                    hist->fill(tdcHit.trailTimeInl);
                else
                    hist->fill(tdcHit.getHitTime(getInl(ev, tdcHit)));
            }

            hist = getHistById(HistId::tdcWidth);
            if(hist && cfg.histCfg.value(HistId::tdcWidth).histGroupEn &&
                    tdcHit.leadingTime && tdcHit.trailingTime)
                hist->fill(tdcHit.hasWidthTimeInl ? tdcHit.widthTimeInl : tdcHit.getWidthNs(getInl(ev, tdcHit)));
            if(HistId::inl != -1) {
                hist = getHistByIdIndexCh(HistId::inl, ev.index, ch);
                if(hist && cfg.histCfg.value(HistId::inl).histGroupEn)
                    hist->fill(tdcHit.leadingTime%1024);
//                histTRoot[HistId::chInl+ch]->Fill(tdcHit.leadingTime%1024);
            }

            if(cfg.histCfg.value(HistId::tdcDoublePulseTime).histGroupEn)
                hist = getHistById(HistId::tdcDoublePulseTime);
            else
                hist = nullptr;

            if((hist!=nullptr || (cfg.histCfg.value(HistId::inlCheck).histGroupEn))
                    && (ev.index != tdcStopEvent.index || ch != cfg.stopCh))
                fillDoublePulse(hist, tdcStopEvent, stopHits, ev, tdcHits);
            ++tdcChHitIt;
        }
        ++tdcChIt;
    }

    // Grab stat info for INL results
    if(HistId::inl != -1) {
        for (auto errWord: ev.errors) {
            auto &hptdcMap = inlResults[ev.index].errMask;
            const uint hptdcId = (errWord>>24) & 0xF;
            const quint16 errMask = errWord & 0x3FFF;
            hptdcMap[hptdcId] |= errMask;
        }
        if(!ev.fw_rev.isNull()) {
            inlResults[ev.index].fw_versions.insert(ev.fw_rev.toUInt());
        }
        if(!ev.bmc_rev.isNull()) {
            inlResults[ev.index].bmc_versions.insert(ev.bmc_rev.toUInt());
        }
        auto &minTemp = inlResults[ev.index].minTemp;
        auto &maxTemp = inlResults[ev.index].maxTemp;
        auto &tempStat = inlResults[ev.index].tempStat;
        for (auto temp_id: ev.temp.keys()) {
            const auto &temp = ev.temp[temp_id];
            auto &tempStatOne = tempStat[temp_id];
            tempStatOne.addTemp(temp);

            if(!tempStatOne.initialized)
                continue;
            if(!tempStatOne.temp_valid(temp)) {
                qWarning().noquote() << QString("[%1] Bad temp: temp_id=%2 mean=%3 temp=%4")
                                        .arg(ev.index.getSerialStr())
                                        .arg(temp_id)
                                        .arg(tempStatOne.mean, 0, 'f', 2)
                                        .arg(temp, 0, 'f', 2);
                continue;
            }
            if(minTemp.contains(temp_id))
                minTemp[temp_id] = std::min(minTemp[temp_id], temp);
            else
                minTemp[temp_id] = temp;
            if(maxTemp.contains(temp_id))
                maxTemp[temp_id] = std::max(maxTemp[temp_id], temp);
            else
                maxTemp[temp_id] = temp;
        }
    }

#ifdef TEST_PLL_DOUBLE_PULSE
    for (auto temp_id: ev.temp.keys()) {
        const auto &temp = ev.temp[temp_id];
        auto &v = testDp.temp[ev.index][temp_id];
        v.first += temp;
        ++v.second;
    }
#endif

    // Errors processing
    if(hasHistById(HistId::errors) && cfg.histCfg.value(HistId::errors).histGroupEn){
        auto histErr = getHistById(HistId::errors);
        std::vector<int> errorsBits;
        errorsBits.resize(16, 0);
        for (auto errWord: ev.errors) {
            errWord &= 0x3FFF;
            for(ulong bit=0; errWord; ++bit){
                if(errWord&1){
                    ++errorsBits[bit];
                }
                errWord = (errWord>>1);
            }
        }
        for(ulong bit=0; bit<16; ++bit){
            if(errorsBits[bit])
                histErr->fill(bit, errorsBits[bit]);
        }
    }
#ifdef TEST_PLL_DOUBLE_PULSE
    if(testDp.started && testDp.expired(tdcStopEvent.tai)) {
        flushTestDp();
    } else {
        testDp.started = true;
        testDp.startTai = tdcStopEvent.tai;
    }
#endif
}

void HistWorker::fillDoublePulse(Histogramm *histDp, TdcEvent &tdcStopEvent, QVector<TdcChHit> &stopHits, TdcEvent &ev, QVector<TdcChHit> &tdcEvents)
{
    const TAI *stopTai = &tdcStopEvent.tai;
    const TAI *evTai = &ev.tai;
    TdcTtvxsMapType::const_iterator it;
    // TODO: Move look for stopTai to upper method
    QMap<DeviceIndex, TtvxsEvent>::const_iterator it1;
    if((it=tdcTtvxsMap.find(tdcStopEvent.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()){
            stopTai = &it1.value().tai;
        }
    }
    if((it=tdcTtvxsMap.find(ev.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()) {
            evTai = &it1.value().tai;
        }
    }

    if(evTai->flags != 2 || stopTai->flags!=2)
        return;

    QMap<int, Histogramm> *inlCheckMap = (HistId::inlCheck == -1) ?
                nullptr : &histContainer.inlCheckHist[ev.index];

    auto stopHitIt = stopHits.begin();
    while(stopHitIt != stopHits.end()) {
        TdcChHit &stopHit = *stopHitIt;
        const double stopTimeNs = stopHit.hasLeadTimeInl ? stopHit.leadTimeInl
                                                         : stopHit.getLeadTimeNs(getInl(tdcStopEvent, stopHit));

        double dtTai = 0; // in ns
        dtTai += (evTai->sec - stopTai->sec)*1e9;
        dtTai += evTai->nSec - stopTai->nSec;
        auto tdcHitIt = tdcEvents.begin();
        while(tdcHitIt != tdcEvents.end()) {
            TdcChHit &tdcHit = *tdcHitIt;
            const double hitTimeNs = tdcHit.hasLeadTimeInl ? tdcHit.leadTimeInl
                                                           : tdcHit.getLeadTimeNs(getInl(ev, tdcHit));
            double dt = dtTai+hitTimeNs-stopTimeNs;
            if(histDp)
                histDp->fill(dt);
#ifdef TEST_PLL_DOUBLE_PULSE
            testDp.dt += dt;
            ++testDp.cnt;
#endif
            if(inlCheckMap) {
                inlCheckMap->operator[](tdcHit.ch).fill(dt);
            }
            ++tdcHitIt;
        }
        ++stopHitIt;
    }
}

QVector<double> HistWorker::getInl(const TdcEvent &ev, const TdcChHit &tdcEv) const
{
    return getInl(ev.index, tdcEv.ch);
}

QVector<double> HistWorker::getInl(const DeviceIndex &index, uint ch) const
{
    return curDevCorrMap.value(index).inl.value(ch, QVector<double>());
}

Histogramm &HistWorker::recreateTaiHist(int winSec)
{
    Histogramm &hist = histContainer.baseHist[HistId::tdcTaiTime];
    const QString title = cfg.histCfg[HistId::tdcTaiTime].title;
    if(winSec)
        cfg.toolTipSuffix[HistId::tdcTaiTime] = QString(" (reper: %1)")
                .arg(QDateTime::fromTime_t(taiReperSec).toString("dd/MM/yy hh:mm:ss"));
    else
        cfg.toolTipSuffix.remove(HistId::tdcTaiTime);
    hist = Histogramm(title, (winSec ? winSec*10 : 0), 0, winSec);
    refreshCurrentPlot(true);
    return hist;
}

bool HistWorker::hasHistById(int id) const
{
    return histContainer.baseHist.contains(id);
}

Histogramm *HistWorker::getHistById(int id)
{
    return hasHistById(id) ? &histContainer.baseHist[id] : nullptr;
}

Histogramm *HistWorker::getHistByIdIndexCh(int id, const DeviceIndex &index, uint ch)
{
    if(id == HistId::inl &&
            histContainer.inlHist.contains(index) &&
            histContainer.inlHist[index].contains(ch)) {
        return &histContainer.inlHist[index][ch];
    }
    if(id == HistId::inlCheck &&
            histContainer.inlCheckHist.contains(index) &&
            histContainer.inlCheckHist[index].contains(ch)) {
        return &histContainer.inlCheckHist[index][ch];
    }
    return nullptr;
}

QMap<DeviceIndex, QMap<uint, uint> > HistWorker::getDevChCntMap()
{
    QMap<DeviceIndex, QMap<uint, uint> > devChCnt;
    Histogramm *chHist = getHistById(HistId::tdcCh);
    if(!chHist)
        return  devChCnt;

    for (const DeviceIndex &i: cfg.devChOffset.keys()) {
        const uint &chOffset = cfg.devChOffset[i];
        const uint nch = getDeviceChNum(i);
        for(uint ch=0; ch<nch; ++ch) {
            const int globCh = static_cast<int>(chOffset+ch);
            const uint chCnt = static_cast<uint>(chHist->bins.value(globCh));
            if(chCnt)
                devChCnt[i][ch] = chCnt;
        }
    }
    return devChCnt;
}

void HistWorker::fillDevChHist(QMap<DeviceIndex, QMap<uint, uint> > map)
{
    if(!histContainer.baseHist.contains(HistId::tdcCh) || HistId::tdcCh == -1)
        return;

    Histogramm &chHist = histContainer.baseHist[HistId::tdcCh];
    for (const DeviceIndex &i: map.keys()) {
        Q_ASSERT(cfg.devChOffset.contains(i));
        const uint chOffset = cfg.devChOffset[i];
        const auto &devChCnt = map[i];
        for(uint ch : devChCnt.keys()) {
            chHist.fill(chOffset+ch, devChCnt[ch]);
        }
    }
}

void HistWorker::trashAllData()
{
    histSem->release(data_list.size());
    data_list.clear();
}

QStringList aaa(TestDp testDp, DeviceIndex i, bool full) {
    QStringList str;
    str << QString("0x%1").arg(i.getSerial(),8,16,QChar('0'));
    for(int tempId=0; tempId<5; ++tempId) {
        const auto &val = testDp.temp[i][tempId];
        if(full)
            str << QString("%1").arg(val.first,5,'f',1)
                << QString("%1").arg(val.second,5);
        else
            str << (val.second ? QString("%1").arg(val.first/val.second,5,'f',1) : "-");
    }
    return str;
}

void HistWorker::flushTestDp()
{
    if(!testDp.started)
        return;
    QStringList str;
    str << "xxx"
            << QString("%1").arg(testDp.dt, 6, 'f', 3)
            << QString("%1").arg(testDp.cnt, 5)
               ;
    QStringList str2;
    str2 << "yyy"
            << (testDp.cnt ? QString("%1").arg(testDp.dt/testDp.cnt, 6, 'f', 3) : "-");
    DeviceIndex i1(0xd0, 0x0a9fa26a);
    DeviceIndex i2(0xd0, 0x0aa148f3);
    str << aaa(testDp, i1, 1);
    str2<< aaa(testDp, i1, 0);
    str << aaa(testDp, i2, 1);
    str2<< aaa(testDp, i2, 0);
//    qInfo().noquote() << str.join('\t');
    qInfo().noquote() << str2.join('\t');
    testDp.clear();
}






