#include "Histogramm.h"

#include <numeric>
#include <math.h>

Histogramm::Histogramm():
    Histogramm("", 0, 0, 0)
{}

Histogramm::Histogramm(QString _title, int _nBins, double _min, double _max):
    title(_title),
    nBins(_nBins),
    min(_min),
    max(_max),
    nEntries(0)
{
    resize(nBins, min, max);
}

bool Histogramm::isEmpty()
{
    if(title.isEmpty() && nBins == 0 && min == 0 && max == 0)
        return true;
    else
        return false;
}

void Histogramm::reset()
{
    if(nBins)
        memset(&bins[0], 0, static_cast<size_t>(nBins)*sizeof(double));
    under_flow = 0;
    over_flow = 0;
    nEntries = 0;
    min_filled_val = 0;
    max_filled_val = 0;
}

void Histogramm::resize(int _nBins, double _min, double _max)
{
    nBins = _nBins;
    min = _min;
    max = _max;
    bins.resize(nBins);
    reset();
    if(nBins)
        binStep = (max-min)/nBins;
    else
        binStep = 0;
}

void Histogramm::fill(double v, int w)
{
    Q_ASSERT(w>0);
    if(v < min) {
        ++under_flow;
    } else if(v >= max) {
        ++over_flow;
    } else {
        int binNum = findBin(v);
        Q_ASSERT(binNum>=0 && binNum<nBins);
        bins[binNum] += w;
    }
    if(nEntries) {
        min_filled_val = qMin(min_filled_val, v);
        max_filled_val = qMax(max_filled_val, v);
    } else {
        min_filled_val = v;
        max_filled_val = v;
    }
    nEntries += static_cast<uint>(w);
}

int Histogramm::findBin(double v) const
{
    if(nBins && v >= min && v < max)
        return static_cast<int>(floor((v-min)/binStep));
    else
        return -1;
}

uint Histogramm::getEntries() const
{
    return nEntries;
}

double Histogramm::getBinLower(int bin) const
{
    return min+binStep*bin;
}

double Histogramm::getBinCenter(int bin) const
{
    return min+binStep*(bin+0.5);
}

double Histogramm::getBinUpper(int bin) const
{
    return min+binStep*(bin+1);
}

DistributionParams Histogramm::getDistribution() const
{
    DistributionParams fit;
    fit.cnt = nEntries-getUnderFlow()-getOverFlow();
    int minBin=-1;
    int maxBin=-1;
    for(int bin=0; bin<nBins; ++bin) {
        const auto &binVal = bins[bin];
        if(!qFuzzyIsNull(binVal)) {
            fit.mean += getBinCenter(bin)*binVal;
            if(minBin==-1)
                minBin = bin;
            maxBin = bin;
        }
    }
    if(fit.cnt) {
        fit.mean /= fit.cnt;

        for(int bin=0; bin<nBins; ++bin) {
            const auto &binVal = bins[bin];
            if(!qFuzzyIsNull(binVal)) {
                double dx = getBinCenter(bin)-fit.mean;
                fit.rms += dx*dx*binVal;
            }
        }
        fit.rms = sqrt(fit.rms/fit.cnt);
        if(minBin!=-1) {
            fit.min = getBinLower(minBin);
            fit.max = getBinUpper(maxBin);
        }
    }
    return fit;
}
