//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "ClientManagerWidget.h"
#include "ui_ClientManagerWidget.h"
#include "Client.h"
#include "add-pnp-dialog/AddPnpDialog.h"
#include "daq-config/DaqConfig.h"
#include "pnp-server/PNPMessage.h"
#include "remote-control-server/RunControlCommands.h"
#include "base/ThreadController.h"

#include <QCheckBox>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QJsonObject>
#include <QJsonDocument>
#include <QTimer>

namespace  {
enum COLUMNS{ CLIENTS_COL_ID, CLIENTS_COL_READOUT,
              CLIENTS_COL_PROTOCOL, CLIENTS_COL_IS_VME,
              CLIENTS_COL_DESCR, CLIENTS_COL_TYPE, CLIENTS_COL_INDEX,
              CLIENTS_COL_STATUS, CLIENTS_COL_TOTAL};

const QStringList EV_NUM_SUPPORT_LIST = QStringList()
        << DaqConfig::getTypeName(DaqConfigAdc64System)
        << DaqConfig::getTypeName(DaqConfigHrbSystem)
        << DaqConfig::getTypeName(DaqConfigLand)
        << DaqConfig::getTypeName(DaqConfigTdcControl)
        << DaqConfig::getTypeName(DaqConfigTtvxs)
        << DaqConfig::getTypeName(DaqConfigRunControl);

void setItemTextColor(QTableWidgetItem *item, QColor c) {
    auto foreground = item->foreground();
    foreground.setColor(c);
    item->setForeground(foreground);
}

const int CLIENT_REMOVE_TIMEOUT = 1000;
}

ClientManagerWidget::ClientManagerWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ClientManagerWidget),
    m(new RcClientManager(this)),
    controlEnabled(true),
    statusBackupTimer(new QTimer(this)),
    curSpillNum(0),
    run_enable(false)
{
    qRegisterMetaType<CLIENT_PROTO_ID>();
    ui->setupUi(this);
    setupTable();
    connect(statusBackupTimer, &QTimer::timeout,
            this, &ClientManagerWidget::checkBackupStatus);
    statusBackupTimer->setInterval(500);
    statusBackupTimer->setSingleShot(true);

    connect(this, &ClientManagerWidget::programDiscovered,
            this, &ClientManagerWidget::gotProgramDescription);
    connect(this, &ClientManagerWidget::programClosed,
            this, &ClientManagerWidget::gotClosedProgram);
}

ClientManagerWidget::~ClientManagerWidget()
{
    removeAllClients();
    delete ui;
}

void ClientManagerWidget::init(QMap<int, ClientConfig> client_config_)
{
    removeAllClients();

    for(ClientConfig cc: client_config_.values()) {
        int i = addClient(cc);
        if(i==-1) {
            qWarning()<<QString("Failed to add client from saved configuration: %1 %2 id=%4")
                        .arg(cc.program_type).arg(cc.program_index).arg(cc.id);
        } else {
            highlightClient(cc.id);
        }
    }
    updateVisibilityColumn();
    emit updateTabState(getClientsStates());
}

void ClientManagerWidget::set_run_enable(bool run_enable_)
{
    run_enable = run_enable_;
    updateButtonsState();
}

void ClientManagerWidget::onRestartAll(const RcRunId &runId)
{
    backPressureMap.clear();
    m->ri.set(runId, RS_STARTING);

    bool hasAnyEnabledClient=false;
    for(auto clInfo : clientsInfo) {
        if(clInfo.cfg.readout) {
            hasAnyEnabledClient=true;
            break;
        }
    }
    if(!hasAnyEnabledClient) {
        m->ri.setState(RS_STARTED);
        sendClientMap();
        return;
    }

    for(auto &clInfo : clientsInfo) {
        if(clInfo.cfg.readout)
            clInfo.state = RS_STARTING;
        else
            clInfo.state = RS_STARTED;
    }

    emit sendRunState(RS_STARTING, m->ri.id);
    QElapsedTimer timer;
    timer.start();

    bool looping=true;
    // wait all ended
    while(looping && !timer.hasExpired(5000)){
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        looping=false;
        for(auto clInfo : clientsInfo) {
            if(!clInfo.cfg.readout)
                continue;
            if(clInfo.state == RS_STARTING){
                looping=true;
                break;
            }
        }
    }

    bool isOk=true;
    for(auto clInfo : clientsInfo) {
        if(!clInfo.cfg.readout)
            continue;
        if(clInfo.state != RS_STARTED) {
            qCritical() << QString("Client #%1 '%2' not started: remote state=%3")
                           .arg(clInfo.getId()).arg(getClientDescr(clInfo.getId()))
                           .arg(clInfo.state==RS_ERROR?"error":"starting");
            isOk=false;
            break;
        } else {
            qDebug()<<"Client #"<<clInfo.getId()<<" started";
        }
    }

    sendClientMap();
//    (isOk?qInfo():qWarning())<<"Clients starting finished; "<< (isOk?"success":"error");
    m->ri.setState(isOk ? RS_STARTED : RS_ERROR);
}

void ClientManagerWidget::onStopAll()
{
    backPressureMap.clear();
    m->ri.setState(RS_STOPPING);

    bool hasAnyEnabledClient=false;
    for(auto clInfo : clientsInfo) {
        if(clInfo.cfg.readout){
            hasAnyEnabledClient=true;
            break;
        }
    }
    if(!hasAnyEnabledClient){
        m->ri.setState(RS_STOPPED);
        return;
    }

    for(auto &clientInfo : clientsInfo) {
        if(clientInfo.cfg.readout)
            clientInfo.state = RS_STOPPING;
        else
            clientInfo.state = RS_STOPPED;
    }

    emit sendRunState(RS_STOPPING, m->ri.id);
    QElapsedTimer timer;
    timer.start();

    bool looping=true;
    // wait all ended
    while(looping && !timer.hasExpired(5000)){
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        looping=false;
        for(auto clientInfo : clientsInfo) {
            if(!clientInfo.cfg.readout)
                continue;
            if(clientInfo.state == RS_STOPPING){
                looping=true;
                break;
            }
        }
    }

    bool isOk=true;
    for(auto clientInfo : clientsInfo) {
        if(!clientInfo.cfg.readout)
            continue;
        if(clientInfo.state != RS_STOPPED){
            qCritical() << QString("Client #%1 '%2' not stoped: remote state=%3")
                           .arg(clientInfo.getId()).arg(getClientDescr(clientInfo.getId()))
                           .arg(clientInfo.state==RS_ERROR?"error":"stoping");
            isOk=false;
        } else {
            qDebug()<<"Client #"<<clientInfo.getId()<<" stopped";
        }
    }

//    (isOk?qInfo():qWarning())<<"Clients stopping finished; "<< (isOk?"success":"error");
    m->ri.setState(isOk ? RS_STOPPED : RS_ERROR);
}

void ClientManagerWidget::setupTable()
{
    QTableWidget *table = ui->tableWidgetClients;
    table->setColumnCount(CLIENTS_COL_TOTAL);
#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    table->setHorizontalHeaderItem(CLIENTS_COL_ID, new QTableWidgetItem("id"));
    table->setHorizontalHeaderItem(CLIENTS_COL_READOUT, new QTableWidgetItem("enable"));
    table->setHorizontalHeaderItem(CLIENTS_COL_IS_VME, new QTableWidgetItem("VME"));
    table->setHorizontalHeaderItem(CLIENTS_COL_DESCR, new QTableWidgetItem("descr"));
    table->setHorizontalHeaderItem(CLIENTS_COL_PROTOCOL, new QTableWidgetItem("Proto"));
    table->setHorizontalHeaderItem(CLIENTS_COL_TYPE, new QTableWidgetItem("type"));
    table->setHorizontalHeaderItem(CLIENTS_COL_INDEX, new QTableWidgetItem("index"));
    table->setHorizontalHeaderItem(CLIENTS_COL_STATUS, new QTableWidgetItem("status"));
    table->hideColumn(CLIENTS_COL_TYPE);
    table->hideColumn(CLIENTS_COL_INDEX);
//    table->hideColumn(COL_ID);
}

void ClientManagerWidget::on_pushButtonAdd_clicked()
{
    QVector<ProgramDescription> freeProgs = getFreeProgs();

    if(freeProgs.isEmpty()) {
        ui->pushButtonAdd->setEnabled(false);
        QMessageBox::information(this, "Add client...", "There is no free program");
        return;
    }

    AddPnpDialog dialog(freeProgs, true, this);
    connect(this, &ClientManagerWidget::programDiscovered,
            &dialog, &AddPnpDialog::gotProgramDescription);
    connect(this, &ClientManagerWidget::programClosed,
            &dialog, &AddPnpDialog::gotClosedProgram);
    dialog.set_function_filter(std::bind(&ClientManagerWidget::isFreeProgs, this, std::placeholders::_1));
    if(dialog.exec() == QDialog::Accepted) {
        for(auto &descr : dialog.getProgDescr()) {
            ClientConfig cc;
            cc.program_type = descr.type;
            cc.program_index = descr.index;
            cc.readout = true;
            cc.evNumCheck = isSupportEvNumCheck(descr.type);

            int id = addClient(cc);
            if(id != -1) {
                bool ok;
                const ProgramInterface &i = descr.getProgramInterface(ProgramInterfaceRemoteControl, &ok);
                if(ok && !i.host.isNull() && i.port){
                    emit setClientAddress(id, i.host, i.port);
                    Q_ASSERT(clientsInfo.contains(id));
                    clientsInfo[id].prog_uuid = descr.uuid;
                }

                highlightClient(id);
            }
        }
    }
    updateVisibilityColumn();
    sendUpdatedClientConfig();
}

void ClientManagerWidget::on_pushButtonRemove_clicked()
{
    int rowToRemove = getSelectedRow();
    if(rowToRemove == -1){
        ui->pushButtonRemove->setEnabled(false);
        return;
    }

    int clientId = getClientId(rowToRemove);
    Q_ASSERT(clientId > 0);
    removeClient(clientId);
    sendUpdatedClientConfig();
}

void ClientManagerWidget::on_tableWidgetClients_itemSelectionChanged()
{
    bool enRemove = controlEnabled && (getSelectedRow() != -1) && !run_enable;
    ui->pushButtonRemove->setEnabled(enRemove);
}

void ClientManagerWidget::clientRunError(const RcRunId &runId, QString errStr)
{
    const int id = getClientId(sender());
    if(id <= 0)
        return;
    const int clientRow = getClientRow(id);
    bool ok;
    int clientId = ui->tableWidgetClients->item(clientRow, CLIENTS_COL_ID)->text().toInt(&ok);
    Q_ASSERT(ok);

    for(auto clInfo : clientsInfo.values()) {
        if(clInfo.getId() == clientId){
            QString clErrStr = QString("Error in client #%1 '%2'' (runNumber=%3)")
                    .arg(clientId).arg(getClientDescr(clientId))
                    .arg(runId.runNumber);
            if(!errStr.isEmpty()){
                clErrStr.append(":").append(errStr);
            }
            qCritical() << clErrStr;
            emit runError(runId, clErrStr);
            return;
        }
    }
}

void ClientManagerWidget::clientRunWarn(const RcRunId &runId, QString warnStr)
{
    if (runId != m->ri.id || warnStr.trimmed().isEmpty())
        return;
    const int row = getClientRow(sender());
    QTableWidgetItem *item = ui->tableWidgetClients->item(row, CLIENTS_COL_STATUS);
    const int clientId = getClientId(row);
    if(statusBackup.contains(clientId)) {
        statusBackup[clientId].spillNum = curSpillNum;
    } else {
        StatusBackup &backup = statusBackup[clientId];
        backup.prevStatus = item->text();
        backup.spillNum = curSpillNum;
    }
    if(!statusBackupTimer->isActive())
        statusBackupTimer->start();

    item->setText(warnStr);
    setItemTextColor(item, QColor("#FF9500"));

    updateWarnString();
}

void ClientManagerWidget::clientBackPressure(const RcRunId &runId, bool bp)
{
    int clientId = getClientId(sender());
    if(clientId == -1)
        return;

    bool oldBP = isBackPressureOn();
    clientsInfo[clientId].backPressure = bp;
    bool newBP = isBackPressureOn();

    if(oldBP != newBP)
        emit backPressure(runId, newBP);
}

void ClientManagerWidget::clientStatusUpdated(int id, QString str)
{
    if(statusBackup.contains(id)) {
        statusBackup[id].prevStatus = str;
    } else {
        const int row = getClientRow(sender());
        if(row!=-1)
            ui->tableWidgetClients->item(row, CLIENTS_COL_STATUS)
                    ->setText(str.trimmed().isEmpty()?"unknown state":str);
        else
            qDebug()<<"recive state string from unknown client:"<<str;
    }
}

void ClientManagerWidget::clientDescrUpdated(int id, QString str)
{
    Q_ASSERT(clientsInfo.contains(id));
    const auto &clInfo = clientsInfo[id];
    Q_ASSERT(clInfo.client==sender());

    if(str.isEmpty()) {
        qWarning() << QString("Got empty description for client #%1 '%2'")
                      .arg(id).arg(getClientDescr(id));
    } else {
        const int row =  getClientRow(id);
        Q_ASSERT(row != -1);
        if(row == -1)
            qDebug()<<"recive info from unknown client:"<<str;
        ui->tableWidgetClients->item(row, CLIENTS_COL_DESCR)->setData(Qt::DisplayRole, str);
    }
}

void ClientManagerWidget::clientRunStateUpdated(int clientId, RUN_STATES state, const RcRunId &runId)
{
    if (runId.runIndex != m->ri.id.runIndex){
        qDebug()<<QString("Got wrong response(runIndex=%1; curRunIndex=%2) from client #%3")
                  .arg(runId.runIndex)
                  .arg(m->ri.id.runIndex)
                  .arg(clientId);
        return;
    }
    if (runId.runNumber < m->ri.id.runNumber){
        qDebug()<<QString("Got old response(run=%1; curRun=%2) from client #%3")
                  .arg(runId.runNumber)
                  .arg(m->ri.id.runNumber)
                  .arg(clientId);
        return;
    }
    Q_ASSERT(m->ri.id.runNumber==runId.runNumber);
    if (clientsInfo.contains(clientId)) {
        if(state==RS_ERROR)
            qDebug()<<"client returned error; id="<<clientId;
        clientsInfo[clientId].state = state;
    } else {
        qDebug()<<"Got clientRunStateUpdated from unknown id="<<clientId;
    }
}

void ClientManagerWidget::clientReadoutToggled(bool readout)
{
    QTableWidget *table = ui->tableWidgetClients;
    bool isClientVme = false;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->cellWidget(row, CLIENTS_COL_READOUT)!=sender())
            continue;

        bool prevOk = getClientsStates();
        int clientId = getClientId(row);
        Q_ASSERT(clientId != -1);
        if(clientsInfo.contains(clientId)){
            // ClientConfig can be absent in config if we adding a client
            auto &cfg = clientsInfo[clientId].cfg;
            cfg.readout = readout;
            sendUpdatedClientConfig();
            highlightClient(clientId);
            isClientVme = cfg.hasVme();
            updateVisibilityColumn();
        } else {
            qInfo() << QString("Client #%1 '%2' was %3abled");
        }
        qInfo() << QString("Client #%1 '%2' was %3abled")
                   .arg(clientId).arg(getClientDescr(clientId))
                   .arg(readout?"en":"dis");
        bool curOk = getClientsStates();
        if(prevOk != curOk)
            emit updateTabState(curOk);
        if(readout)
            emit startListen(clientId);
        else
            emit stopListen(clientId);
        // emit updateClientReadout(clientId, readout);
        break;
    }
    if(isClientVme)
        emit vmeFlagUpdated();
}

void ClientManagerWidget::gotProgramDescription(const ProgramDescription &newDescr)
{
    bool missing = true;
    for (auto &descr : progDescrs) {
        if(newDescr.sameProgram(descr)) {
            missing = false;
            descr = newDescr;
            break;
        }
    }
    if(missing)
        progDescrs.push_back(newDescr);

    bool ok;
    for(auto &clInfo : clientsInfo) {
        if(clInfo.cfg.program_type == newDescr.type && clInfo.cfg.program_index == newDescr.index) {
            const ProgramInterface &i = newDescr.getProgramInterface(ProgramInterfaceRemoteControl, &ok);
            if(ok && !i.host.isNull() && i.port) {
                emit setClientAddress(clInfo.getId(), i.host, i.port);
                clInfo.prog_uuid = newDescr.uuid;
            }
            break;
        }
    }

    bool enAdd = !getFreeProgs().isEmpty() && !run_enable;
    if(ui->pushButtonAdd->isEnabled() != enAdd)
        ui->pushButtonAdd->setEnabled(enAdd);
}

void ClientManagerWidget::gotClosedProgram(const ProgramDescription &descr)
{
    for(int i=0; i< progDescrs.size(); ++i) {
        if(progDescrs[i].uuid == descr.uuid){
            progDescrs.remove(i);
            break;
        }
    }
    bool enAdd = !getFreeProgs().isEmpty() && !run_enable;
    if(ui->pushButtonAdd->isEnabled() != enAdd)
        ui->pushButtonAdd->setEnabled(enAdd);

    for(auto clInfo : clientsInfo) {
        if(clInfo.prog_uuid == descr.uuid) {
            emit setClientEmptyAddress(clInfo.getId());
        }
    }
}

void ClientManagerWidget::clientReady4Run(int id, bool ready)
{
    if(!clientsInfo.contains(id))
        return;
    auto &clInfo = clientsInfo[id];
    if(clInfo.ready4Run == ready)
        return;

    bool prevOk = getClientsStates();
    clInfo.ready4Run = ready;
    highlightClient(id);
    bool ok = getClientsStates();
    if(prevOk != ok)
        emit updateTabState(ok);
}

void ClientManagerWidget::clientVmeFlag(int id, bool hasVme)
{
    if(!clientsInfo.contains(id))
        return;
    clientsInfo[id].cfg.has_vme = hasVme;
    int row = getClientRow(id);
    if(row==-1)
        return;
    auto item = ui->tableWidgetClients->item(row, CLIENTS_COL_IS_VME);
    item->setText(hasVme ? QChar(0x2713) : QString());
    item->setTextAlignment(Qt::AlignCenter);
    sendUpdatedClientConfig();
    emit vmeFlagUpdated();
    updateVisibilityColumn(CLIENTS_COL_IS_VME);
}

void ClientManagerWidget::clientProtocolChanged(int id, CLIENT_PROTO_ID protoId)
{
    if(clientsInfo.contains(id)) {
        clientsInfo[id].cfg.protoId = protoId;
    }

    int row = getClientRow(id);
    if(row==-1)
        return;

    auto item = ui->tableWidgetClients->item(row, CLIENTS_COL_PROTOCOL);
    item->setText(Client::getProtocolIdStr(protoId));
    item->setToolTip(Client::getProtocolIdStr(protoId, false));

    QColor col;
    switch (protoId) {
    case CLIENT_PROTO_V3:
        col = Qt::black; break;
    case CLIENT_PROTO_V2:
        col = Qt::darkYellow; break;
    default:
        col = Qt::red;
    }

    auto fg = item->foreground();
    fg.setColor(col);
    item->setForeground(fg);
//    updateVisibilityColumnProtocol();
    updateVisibilityColumn(CLIENTS_COL_PROTOCOL);
}

void ClientManagerWidget::setControlEnable(bool enabled)
{
    controlEnabled = enabled;
    setEnabled(controlEnabled);
    updateButtonsState();
}

void ClientManagerWidget::updateButtonsState()
{
    bool enAdd = !getFreeProgs().isEmpty() && !run_enable;
    if(ui->pushButtonAdd->isEnabled() != enAdd)
        ui->pushButtonAdd->setEnabled(enAdd);
    bool enRemove = controlEnabled && (getSelectedRow() != -1) && !run_enable;
    if(ui->pushButtonRemove->isEnabled() != enRemove)
        ui->pushButtonRemove->setEnabled(enRemove);
    for(int row=0; row<ui->tableWidgetClients->rowCount(); ++row)
        ui->tableWidgetClients->cellWidget(row, CLIENTS_COL_READOUT)->setEnabled(controlEnabled && !run_enable);
}

void ClientManagerWidget::checkBackupStatus()
{
    for(int id: statusBackup.keys()) {
        const StatusBackup &backup = statusBackup[id];
        if(backup.isActive(curSpillNum))
            continue;
        int row = getClientRow(id);
        if(row!=-1) {
            auto item = ui->tableWidgetClients->item(row, CLIENTS_COL_STATUS);
            item->setText(backup.prevStatus);
            setItemTextColor(item, QColor());
        }
        statusBackup.remove(id);
    }
    if(!statusBackup.isEmpty())
        statusBackupTimer->start();
    updateWarnString();
}

int ClientManagerWidget::addClient(ClientConfig cc)
{
    if(cc.program_type.isEmpty() || cc.program_index.isEmpty())
        return -1;
    int maxClientId=0;
    auto &clientId = cc.id;
    if(clientId==0) {
        qWarning() << "attempt to add client with id==0; clientId=0 is prohibited";
        clientId = -1;
    }
    auto descr = cc.program_type + " #" + cc.program_index;
    for(auto cl_info : clientsInfo) {
        maxClientId = qMax(maxClientId, cl_info.getId());

        if(cc.program_type == cl_info.cfg.program_type
                && cc.program_index == cl_info.cfg.program_index) {
            qWarning() << "Can't add client with dublicated " << descr;
            return -1;
        }
        if(clientId == cl_info.getId()) {
            qWarning() << QString("Desiread clientId (%1) is already exist").arg(clientId);
            clientId = -1;
        }
    }

    if(clientId == -1)
        clientId = maxClientId+1;

    ClientInfo &clInfo = clientsInfo[clientId];
    clInfo.cfg = cc;

    Client *c = clInfo.client = new Client(clientId, cc.program_type, cc.program_index);

    QTableWidget *table = ui->tableWidgetClients;
    int row = table->rowCount();
    table->setRowCount(row+1);

    QTableWidgetItem *item;
    table->setItem(row, CLIENTS_COL_ID, item=new QTableWidgetItem(QString::number(clientId)));
    item->setData(Qt::UserRole, clientId);
    QCheckBox *cbReadout = new QCheckBox(table);
    cbReadout->setChecked(false);
    table->setCellWidget(row, CLIENTS_COL_READOUT, cbReadout);
    table->setItem(row, CLIENTS_COL_IS_VME, item = new QTableWidgetItem());
    const bool isVme = cc.program_type==DaqConfig::getTypeName(DaqConfigVmeDaq);
    if(isVme)
        item->setText(QChar(0x2713));
    table->setItem(row, CLIENTS_COL_DESCR, new QTableWidgetItem(descr));
    table->setItem(row, CLIENTS_COL_PROTOCOL, new QTableWidgetItem());
    table->setItem(row, CLIENTS_COL_TYPE, new QTableWidgetItem(cc.program_type));
    table->setItem(row, CLIENTS_COL_INDEX, new QTableWidgetItem(cc.program_index));
    table->setItem(row, CLIENTS_COL_STATUS, item=new QTableWidgetItem("creating..."));
    setItemTextColor(item, Qt::red);

    c->setObjectName(QString("Client for %1 thread").arg(descr));
    new ThreadController(c, this);
    connect(this, &ClientManagerWidget::deleteClient, c, &Client::deleteClient);
    connect(c, &Client::runStateUpdated, this, &ClientManagerWidget::clientRunStateUpdated);
    connect(c, &Client::runError, this, &ClientManagerWidget::clientRunError);
    connect(c, &Client::runWarning, this, &ClientManagerWidget::clientRunWarn);
    connect(c, &Client::backPressure, this, &ClientManagerWidget::clientBackPressure);
    connect(c, &Client::statusUpdated, this, &ClientManagerWidget::clientStatusUpdated);
    connect(c, &Client::descrUpdated, this, &ClientManagerWidget::clientDescrUpdated);
    connect(c, &Client::ready4Run, this, &ClientManagerWidget::clientReady4Run);
    connect(c, &Client::vmeFlag, this, &ClientManagerWidget::clientVmeFlag);
    connect(c, &Client::protocolChanged, this, &ClientManagerWidget::clientProtocolChanged);
    connect(this, &ClientManagerWidget::setClientAddress, c, &Client::setAddress);
    connect(this, &ClientManagerWidget::setClientEmptyAddress, c, &Client::setEmptyAddress);
    connect(this, &ClientManagerWidget::startListen, c, &Client::startReadOut);
    connect(this, &ClientManagerWidget::stopListen, c, &Client::stopReadOut);
    connect(this, &ClientManagerWidget::sendRunState, c, &Client::sendRunState);
//    connect(this, &ClientManagerWidget::updateClientReadout,
//            c, &Client::updateClientReadout);

    connect(c, &Client::config_transfer, this, &ClientManagerWidget::config_transfer);
    connect(cbReadout, &QCheckBox::toggled, this, &ClientManagerWidget::clientReadoutToggled);

    if(cc.evNumCheck) {
        connect(this, &ClientManagerWidget::evNumRequest,
                c, &Client::evNumRequest);
        connect(c, &Client::evNumResponse,
                this, &ClientManagerWidget::evNumResponse);
    } else {
        qInfo() << QString("Remove client #%1 (%2 %3) from event number check")
                   .arg(clientId).arg(cc.program_type).arg(cc.program_index);
    }
    clientProtocolChanged(clientId, CLIENT_PROTO_UNKNOWN);

    cbReadout->setChecked(cc.readout);
    return clientId;
}

int ClientManagerWidget::getClientRow(QObject *sender) const
{
    int id = getClientId(sender);
    if(id != -1)
        return getClientRow(id);
    return -1;
}

int ClientManagerWidget::getClientRow(int clientId) const
{
    for(int row=0;row<ui->tableWidgetClients->rowCount(); ++row){
        const int id = getClientId(row);
        if(id == clientId)
            return row;
    }
    return -1;
}

int ClientManagerWidget::getClientId(int row) const
{
    bool ok;
    const int id = ui->tableWidgetClients->item(row, CLIENTS_COL_ID)->
            data(Qt::DisplayRole).toInt(&ok);

    return  ok ? id : -1;
}

int ClientManagerWidget::getClientId(QObject *sender) const
{
    for(auto clInfo : clientsInfo) {
        if(sender==clInfo.client)
            return clInfo.getId();
    }
    return -1;
}

QString ClientManagerWidget::getClientDescr(int clientId) const
{
    const int row = getClientRow(clientId);
    return ui->tableWidgetClients->item(row, CLIENTS_COL_DESCR)->text();
}

QVector<ProgramDescription> ClientManagerWidget::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;

    for(const ProgramDescription &descr: progDescrs) {
        if(isFreeProgs(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool ClientManagerWidget::isFreeProgs(const ProgramDescription &descr) const
{
    bool ok;
    if(descr.isEmpty())
        return false;
    const ProgramInterface &interface = descr.getProgramInterface(ProgramInterfaceRemoteControl, &ok);
    if(!ok || !interface.enabled || !interface.isFree)
        return false;

    for(auto clInfo : clientsInfo.values()) {
        if(clInfo.cfg.program_type == descr.type
                && clInfo.cfg.program_index == descr.index)
            return false;
    }
    return true;
}

int ClientManagerWidget::getSelectedRow() const
{
    int row=-1;
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidgetClients->selectedRanges();
    if(!ranges.isEmpty()){
        QTableWidgetSelectionRange firstRange = ranges[0];
        row = firstRange.topRow();
    }
    return row;
}

bool ClientManagerWidget::getClientsStates() const
{
    for(auto clInfo : clientsInfo) {
        if(clInfo.cfg.readout && !clInfo.ready4Run)
            return false;
    }
    return true;
}

void ClientManagerWidget::highlightClient(int id)
{
    Q_ASSERT(clientsInfo.contains(id));
    const auto &clInfo = clientsInfo[id];
    bool isOk = (!clInfo.cfg.readout) || clInfo.ready4Run;
    auto item = ui->tableWidgetClients->item(getClientRow(id), CLIENTS_COL_STATUS);
    setItemTextColor(item, isOk ? QColor() : Qt::red);
}

void ClientManagerWidget::sendClientMap() const
{
    QMap<int, QString> map;

    for(auto clInfo : clientsInfo) {
        if(!clInfo.cfg.readout || !clInfo.cfg.evNumCheck) continue;
        if(!isSupportEvNumCheck(clInfo.cfg.program_type))
            continue;
        const int id = clInfo.getId();
        const int row = getClientRow(id);
        if(row == -1) {
            qWarning() << QString("Can't find row for client with id:%1 %2 %3")
                          .arg(clInfo.getId())
                          .arg(clInfo.cfg.program_type)
                          .arg(clInfo.cfg.program_type);
            continue;
        }

        map[id] = getClientDescr(id);
    }
    emit  clientMapUpdated(map);
}

void ClientManagerWidget::updateWarnString() const
{
    QStringList clWarnStr;
    for(int clientId: statusBackup.keys()) {
        int row = getClientRow(clientId);
        if(row==-1)
            continue;

        clWarnStr << QString("%1:%2")
                     .arg(ui->tableWidgetClients->item(row, CLIENTS_COL_DESCR)->text())
                     .arg(ui->tableWidgetClients->item(row, CLIENTS_COL_STATUS)->text());
    }
    emit runWarn(m->ri.id, clWarnStr.join("; "));
}

bool ClientManagerWidget::isSupportEvNumCheck(QString type)
{
    return EV_NUM_SUPPORT_LIST.contains(type);
}

void ClientManagerWidget::updateVisibilityColumn(int col)
{
    for(int c=(col==-1?0:col); c<CLIENTS_COL_TOTAL; ++c) {
        std::function<bool(ClientConfig)> isHidden = nullptr;
        switch (c) {
        case CLIENTS_COL_PROTOCOL:
            isHidden = [](ClientConfig cfg) { return !cfg.readout || cfg.protoId == CLIENT_PROTO_V3; };
            break;
        case CLIENTS_COL_IS_VME:
            isHidden = [](ClientConfig cfg) { return !cfg.readout || !cfg.hasVme(); };
            break;
        default:
            continue;
        }

        bool hide=true;
        for(auto clInfo : clientsInfo) {
            hide = isHidden(clInfo.cfg);
            if(!hide) break;
        }
        ui->tableWidgetClients->setColumnHidden(c, hide);
        if(col != -1)
            break;
    }
}

void ClientManagerWidget::removeAllClients()
{
    QElapsedTimer timer;
    timer.start();
    emit deleteClient(0);
    while(!clientsInfo.isEmpty())
        removeClient(clientsInfo.first().getId(), -1, &timer);
}

void ClientManagerWidget::removeClient(int clientId, int rowToRemove, QElapsedTimer *timer)
{
    if(rowToRemove==-1) {
        rowToRemove = getClientRow(clientId);
        Q_ASSERT(rowToRemove >= 0);
    }
    if(rowToRemove != -1)
        ui->tableWidgetClients->removeRow(rowToRemove);

    auto clInfoIt = clientsInfo.find(clientId);
    Q_ASSERT(clInfoIt != clientsInfo.end());
    if(clInfoIt == clientsInfo.end())
        return;

    qDebug()<<"Going to delete client #"<<clientId;
    emit deleteClient(clientId);

    clientsInfo.remove(clientId);
}

void ClientManagerWidget::sendUpdatedClientConfig()
{
    QMap<int, ClientConfig> client_config;
    for(auto clInfo : clientsInfo)
        client_config[clInfo.getId()] = clInfo.cfg;
    emit client_config_changed(client_config);
}

bool ClientManagerWidget::isBackPressureOn() const
{
    for(const auto &clInfo: clientsInfo)
        if(clInfo.backPressure && clInfo.cfg.readout)
            return true;
    return false;

}
