//
//    Copyright 2016 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
#include "MetricSenderHelper.h"

#include <QProcessEnvironment>
#include <QUrl>
#include <QHostInfo>
#include <QCoreApplication>
#include <QRegularExpression>
#include <QDateTime>

namespace {
// Set environment variable in form hostname:port
static const char * METRICS_ENV_VAR = "METRICS_HOST";
static const int METRICS_URL_DEFAULT_PORT = 2003;
static const char *METRICS_URL_DEFAULT_HOST = "localhost";
static const char *METRICS_URL_DEFAULT_SCHEME = "udp";
}

MetricSenderHelper::MetricSenderHelper(QObject *parent)
    : QObject(parent),
      metricNamespace(MetricSenderHelper::defaultNamespace()),
      enable(false)
{
    // parse environment variable
    const QString &envStr = QProcessEnvironment::systemEnvironment().value(METRICS_ENV_VAR);
    if (envStr.isEmpty()) {
        qInfo() << QString("Metrics disabled: %1 not defined")
                   .arg(METRICS_ENV_VAR);
        return;
    }
    url = QUrl::fromUserInput(envStr);
    if (!url.isValid()) {
        qWarning() << QString("Metrics disabled: '%1' is not a valid URL")
                      .arg(envStr);
        return;
    }
    if (url.scheme().compare("http") == 0)
        url.setScheme(METRICS_URL_DEFAULT_SCHEME);
    if (url.scheme().compare("udp") != 0) {
        qWarning() << QString("Metrics disabled: %1 is not a supported scheme")
                      .arg(url.scheme());
        return;
    }
    if (url.port() <= 0)
        url.setPort(METRICS_URL_DEFAULT_PORT);

    if (url.host().isEmpty()) {
        url.setHost(METRICS_URL_DEFAULT_HOST);
    }
    // resolve to IP address
    const QHostInfo &hostInfo = QHostInfo::fromName(url.host());
    const QList<QHostAddress> &addresses = hostInfo.addresses();
    if (addresses.isEmpty()) {
        qWarning() << QString("Metrics disabled: no IP address for %1")
                      .arg(url.toDisplayString());
        return;
    }
    hostAddrCached = addresses.first(); // TODO: asynchronous resolver

    enable = true;
    qInfo().noquote() << QString("[MetricSender]: Metrics destination %1")
               .arg(url.toDisplayString());
}

QString MetricSenderHelper::defaultNamespace()
{
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    auto name = QCoreApplication::applicationName() + "." + shortHostName;
    name.replace(QRegularExpression("[^A-Za-z0-9.]"), QChar('_'));
    return name;
}

QString MetricSenderHelper::formatMetricLine(const QString &name, const QVariant &value) const
{
    QString s = name;
    s.replace(QRegularExpression("[^A-Za-z0-9.]"), QChar('_'));
    s = metricNamespace + "." + s;
    s += " ";
    s += QString("%2").arg(value.toString());
    s += " ";
    s += QString::number(QDateTime::currentDateTime().toTime_t());
    s += "\n";
    return s;
}

QHostAddress MetricSenderHelper::hostAddr() const
{
    return hostAddrCached;
}

quint16 MetricSenderHelper::hostPort() const
{
    return url.port(METRICS_URL_DEFAULT_PORT);
}

QUrl MetricSenderHelper::getUrl() const
{
    return url;
}
