/*
**    Copyright 2006-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MLINKDEVICEINTERFACEBASE_H
#define MLINKDEVICEINTERFACEBASE_H

#include <cstddef>
#include <cstdint>
#include <array>
#include "mlink.h"
#include "mlink/mlink_frame_header.h"

class QString;

namespace mlink {

typedef int mlink_event_t;
class RegIoPacket;

class MlinkDeviceInterfaceBase
{
public:
    MlinkDeviceInterfaceBase &operator=(const MlinkDeviceInterfaceBase &) = delete;
    MlinkDeviceInterfaceBase &operator=(const MlinkDeviceInterfaceBase &&) = delete;

    virtual ~MlinkDeviceInterfaceBase() = default;
    const char *ml_strerror(int errnum) const;
    virtual void print_link_error(int ret, const QString &text) const = 0;
    virtual void print_frame_error(int ret, const QString &text) const = 0;
    virtual void ml_print_raw_buffer(const void *ptr, size_t size) const = 0;

    virtual int ml_open (const char *devname);
    int ml_reset ();
    bool isOpen() const;
    bool is_link_up() const;
    int ml_get_driver_version(unsigned long *n);
    [[deprecated]] int ml_get_board_serial_id(uint64_t *n);
    virtual int ml_close ();

    // *data - pointer to contents
    // len - size of contents in 32-bit words
    // type - frame type
    // src, dst - address
    // seq - sequence number
    // timeout - milliseconds
    int ml_send_packet(const RegIoPacket &packet, int timeout);

    // *flen - on input, user shall specify the allocated data buffer size, words
    // *flen - on output, it contains the actual number of words received
    // others same as above
    int ml_recv_packet(RegIoPacket *packet, int timeout);
    int ml_recv_data_frame_multi(uint32_t fdata[], size_t fsize, struct frameset_t *frameset, int timeout);


    // rx/tx bytes counter
    int ml_get_rxb();
    int ml_get_txb();
    // rx/tx frames counter
    int ml_get_rxf();
    int ml_get_txf();
    void ml_clear_rxtx_counters();

    int ml_get_event (mlink_event_t *ev, int timeout_ms);

    int ml_read_port(int port, unsigned int offset, unsigned int *data);
    int ml_write_port(int port, unsigned int offset, unsigned int data);

    int ml_read_flasher(unsigned int offset, unsigned int *data);
    int ml_write_flasher(unsigned int offset, unsigned int data);

    int ml_read_board_info(board_info_t *data);

protected:
    enum {
        ERR_TIMERECV = 10001,
        ERR_TIMESEND = 10002,
        ERR_BADCRC   = 10003,
        ERR_REMOTEIO = 10004,
        };

    virtual int ml_dma_read (uint32_t buf[], size_t nwords, size_t * words_received, int timeout);
    virtual int ml_dma_write (const uint32_t buf[], size_t nwords, int timeout);


    int ml_fd = -1;
    unsigned int rxb = 0, txb = 0, rxf = 0, txf = 0;

private:
    int ml_get_driver_status (mlink_status_t *st) const;
    [[deprecated]] int ml_get_caps (mlink_caps_t *caps);

    std::array<uint32_t, ML_MAX_RAW_FRAME_WORDS> tx_buffer;
    std::array<uint32_t, ML_RX_BUFFER_WORDS> rx_buffer;
};

} // namespace mlink

#endif // MLINKDEVICEINTERFACEBASE_H
