//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "mregdevice.h"

#include <QDebug>

#include "regio/regop.h"
#include "regio/crc8_dallas_maxim.h"

namespace mlink {

QString DeviceAttr::getFirmwareVersionStr() const
{
    return QString("%1.%2.%3")
        .arg(fw_version.v_maj)
        .arg(fw_version.v_min)
        .arg(fw_version.revision);
}

MregDevice::MregDevice(RegIOBase &p, int address)
    : MlinkDevice(p, address)
{
}

MregDeviceBpm::MregDeviceBpm(RegIOBase &p, int address)
    : MregDevice(p, address)
{
}

MregDeviceAdcm::MregDeviceAdcm(RegIOBase &p, int address)
    : MregDevice(p, address)
{
}

quint64 MregDeviceBpm::readOneWireId(bool checkAccess)
{
    if (checkAccess && !hardwareAccessAllowed()) return 0;
    std::vector<quint16> v = regReadBlk(REG64_SERIAL_NUM, 0x4, checkAccess);
/*
    std::vector<quint16> v;
    v.push_back(regRead(REG_SERIAL_NUM));
    v.push_back(regRead(REG_SERIAL_NUM+1));
    v.push_back(regRead(REG_SERIAL_NUM+2));
    v.push_back(regRead(REG_SERIAL_NUM+3));
*/
    std::vector<quint16> c(8);
    bool all_zeroes = true;
    bool all_ones = true;
    for(int i=0; i<8; i++) {
        c[i] = (v[i/2] >> ((i%2)?8:0)) & 0xFF;
        if (c[i] != 0) all_zeroes = false;
        if (c[i] != 0xFF) all_ones = false;
    }
    if (all_zeroes || all_ones) {
        qWarning() << "No identification chip present";
        return 0;
    }

    //  for(size_t i=0; i<c.size(); i++)
    //	cout << hex << i << ": " << (unsigned int)c[i] << endl;

    const int testID = 0x28;
    unsigned char crc = 0;
    for(int i=0; i<7; i++)
        crc = crc8_dallas_maxim(c[i], crc);
    if (c[0] != testID) {
        qWarning() << QString("Unknown identification chip: read %1, expected %2").arg(c[0], 2, 16, QChar('0')).arg(testID, 2, 16, QChar('0'));
        return 0; // no-op
    }
    if (crc != c[7]) {
        qWarning() << QString("Identification chip CRC mismatch: read %1, calculated %2").arg(c[7], 2, 16, QChar('0')).arg((int)crc, 2, 16, QChar('0'));
        return 0;
    }
    quint64 id = quint64(c[1])
                 | (quint64(c[2]) << 8)
                 | (quint64(c[3]) << 16)
                 | (quint64(c[4]) << 24)
                 | (quint64(c[5]) << 32)
                 | (quint64(c[6]) << 40);
    attr.id_chip_present = true;
    attr.device_serial = id;
    return id;
}

quint64 MregDeviceAdcm::readOneWireId(bool checkAccess)
{
    if (checkAccess && !hardwareAccessAllowed()) return 0;
    quint32 serial = regRead(ADCM16_SERIAL_ID_REG, checkAccess);
    quint32 serialHi = regRead(ADCM16_SERIAL_ID_HI_REG, checkAccess);
    serial |= serialHi<<16;
    attr.id_chip_present = true;
    attr.device_serial = serial;
    return serial;
}

bool MregDeviceBpm::readFirmwareId()
{
    if (!hardwareAccessAllowed()) return false;
    RegOpVector regOp({RegRead(REG_FW_VER), RegRead(REG_FW_REV)});
    auto regRes = regOpExec(regOp);
    if(regOp.size() != regRes.size())
        return false;
    attr.fw_version = fw_version_t(regRes[0].data, regRes[1].data);
    return true;
}

bool MregDeviceBpm::readMStreamCfg()
{
    if (!hardwareAccessAllowed()) return false;
    quint16 var = regRead(REG_MSTREAM_CFG);
    attr.mstreamMultiAck = var & 0x8;
    attr.mstreamHwBufSize = var ? (1<<(var&0x7)) : 8;
    return var;
}

bool MregDeviceAdcm::readFirmwareId()
{
    if (!hardwareAccessAllowed()) return false;
    RegOpVector regOp({RegRead(ADCM16_FW_VER_REG), RegRead(ADCM16_FW_REV_REG)});
    auto regRes = regOpExec(regOp);
    if(regOp.size() != regRes.size())
        return false;
    attr.fw_version = fw_version_t(regRes[0].data, regRes[1].data);
    return true;
}

bool MregDeviceAdcm::readMStreamCfg()
{
    if (!hardwareAccessAllowed()) return false;
    quint16 var = regRead(ADCM16_MSTREAM_CFG);
    attr.mstreamMultiAck = var & 0x8;
    attr.mstreamHwBufSize = var ? (1<<(var&0x7)) : 8;
    return var;
}

void MregDevice::print() const
{
  qDebug() << QString("Device ID %1, Firmware %2, Serial Number %3, Address %4, Temp %5")
          .arg(getDeviceId(), 4, 16, QChar('0'))
          .arg(attr.getFirmwareVersionStr())
          .arg(getDeviceSerial(), 8, 16, QChar('0'))
          .arg(getAddress(), 4, 16, QChar('0'))
          .arg(attr.temp, 0, 'f', 1);
}

double MregDevice::convertOneWireTemp(quint16 rawTemp)
{
    rawTemp &= 0xFFF;
    if (rawTemp & 0x800) rawTemp = -(rawTemp&0x7FF);
    return 0.0625 * rawTemp;
}

double MregDeviceBpm::readTemp()
{
    if (!hardwareAccessAllowed()) return -255;
    attr.temp = convertOneWireTemp(regRead(REG_TEMPERATURE));
    return attr.temp;
}

double MregDeviceAdcm::readTemp()
{
    if (!hardwareAccessAllowed()) return -255;
    attr.temp = convertOneWireTemp(regRead(ADCM16_TEMPERATURE_REG));
    return attr.temp;
}

quint64 MregDeviceBpm::readTimestampCurrent()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CURRENT_TIMESTAMP);
}

quint64 MregDeviceBpm::readTimestampRun()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CNT_NCU);
}

quint64 MregDeviceBpm::readTimestampTrig()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CNT_KCU);
}

quint32 MregDeviceBpm::writeLiveMagic(quint32 value)
{
    if (!hardwareAccessAllowed()) return 0;
    regWrite(REG_LIVE_MAGIC, value);
    quint16 rd = regRead(REG_LIVE_MAGIC);
    return rd;
}

quint32 MregDeviceBpm::readLiveMagic()
{
    if (!hardwareAccessAllowed()) return 0;
    quint16 rd = regRead(REG_LIVE_MAGIC);
    return rd;
}

quint64 MregDeviceAdcm::readTimestampCurrent()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_REG64);
}

quint64 MregDeviceAdcm::readTimestampRun()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_A);
}

quint64 MregDeviceAdcm::readTimestampTrig()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_B);
}

quint32 MregDeviceAdcm::writeLiveMagic(quint32 value)
{
    if (!hardwareAccessAllowed()) return 0;
    RegOpVector op;
    op << RegWrite(ADCM16_LIVE_MAGIC, value);
    op << RegRead(ADCM16_LIVE_MAGIC);
    const RegOpVector res = regOpExec(op);

    if(res.size()==2)
        return res.at(1).data;
    return 0;
}

quint32 MregDeviceAdcm::readLiveMagic()
{
    if (!hardwareAccessAllowed()) return 0;
    quint16 rd = regRead(ADCM16_LIVE_MAGIC);
    return rd;
}

quint16 MregDevice::readDeviceId(bool checkAccess)
{
    quint16 id = regRead(REG_DEVICE_ID, checkAccess);
//    qDebug() << QString("Read reg %1: %2").arg(REG_DEVICE_ID, 4, 16, QChar('0')).arg(id, 4, 16, QChar('0'));
//    id &= 0xFF00;
    id = id>>8;
    attr.device_id = id;
    return id;
//    for (int i=0; i<0x100; i++) {
//        const quint16 rd = regRead(i);
//        qDebug() << QString("reg[%1] = %2").arg(i, 2, 16, QChar('0')).arg(rd, 4, 16, QChar('0'));
//        usleep(50000);
//    }
}

//quint16 MregDeviceBpm::readDeviceId()
//{
//    quint16 id = regRead(REG_DEVICE_ID);
//    attr.device_id = id;
//    return id;
//}

//quint16 MregDeviceAdcm::readDeviceId()
//{
//    quint16 id = regRead(ADCM16_STATUS_REG) & 0xFF00;
//    attr.device_id = id;
//    return id;
//}

quint16 MregDevice::detect()
{
    if (!enableState) return 0;
//    qDebug() << QString("MregDevice: Trying address %1").arg(getAddress(), 4, 16, QChar('0'));
    try {
//        onlineState = true;
        quint16 id = readDeviceId(false);
        bool id_present = (id != 0) && ((id&0xFFFF) != 0xFFFF);
        if (!id_present) {
            qWarning() << QString("No device ID found");
            return 0;
        }
        onlineState = id_present;
        readOneWireId();
        readFirmwareId();
        readMStreamCfg();
        readTemp();
        return id;
    }

    catch (std::exception &e) {
        if (onlineState)
            qWarning() << QString("MregDevice::detect failed: %1").arg(e.what());
        onlineState = false;
    }

//    if (isOnline())
//        print();
    return 0;
}

} // namespace mlink
