#include "BaseConfig.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QHostInfo>
#include <QElapsedTimer>

BaseConfig::BaseConfig():
    mongo_alive(false)
{
#ifdef MONGO_DRIVERS
    mongo_alive = mongo_db.try_connection();
#endif
}

void BaseConfig::init(QString program_type)
{
    local_db.init(program_type);
#ifdef MONGO_DRIVERS
    if(mongo_alive){
        mongo_db.mongo_init();
        if (!synchronization(program_type)){
            qWarning() << "Synchronize Mongo saved failed!";
        }
    }
#endif
}

RootConfig BaseConfig::read_config(QString program_type, QString program_index, QString config_name)
{
    RootConfig to_return = RootConfig();
    if(program_type == ""){
        qWarning() << "read settings failed. Can't find program type";
        return to_return;
    }

#ifdef MONGO_DRIVERS
    if(mongo_alive && (program_index != PROGRAM_INDEX_DEFAULT)){
        //---read settings from mongo
        to_return = mongo_db.read_config(program_type, program_index, config_name);
    }
#endif
    if(to_return.isEmpty())
        to_return = local_db.read_config(program_type, program_index, config_name);
    return to_return.is_phantom() ? RootConfig() : to_return;
}

bool BaseConfig::write_config(RootConfig root)
{
    bool ok = true;
    QString program_type = root.get_data<QString>("program_type");
    QString program_index = root.get_data<QString>("program_index");
    QString configuration_name = root.get_data<QString>("configuration_name");

    if(program_type.isEmpty()){
        qWarning() << "write settings failed. Can't find program type";
        return false;
    }

    if(program_index.isEmpty()){
        root.set_data("program_index", PROGRAM_INDEX_DEFAULT);
    }
    if(configuration_name.isEmpty()){
        root.set_data("configuration_name", PROGRAM_CONFIG_DEFAULT);
    }

    //---change date of creation to current date_time
    QDateTime current_dt = QDateTime::currentDateTime();
    root.set_data("date_of_modification", current_dt);

    //---set current host name
    root.set_data("current host", QHostInfo::localHostName());

#ifdef MONGO_DRIVERS
    if(mongo_alive && (program_index != PROGRAM_INDEX_DEFAULT)){
        ok &= mongo_db.write_config(root);
    }
#endif

    ok &= local_db.write_config(root);

    return ok;
}

QMap<QString, QStringList> BaseConfig::get_config_map(QString program_type)
{
    if(program_type.isEmpty()) {
        return QMap<QString, QStringList>();
    }

#ifdef MONGO_DRIVERS
    if(mongo_alive)
        return mongo_db.get_config_map(program_type);
#endif
    return local_db.get_config_map(program_type);
}

QDateTime BaseConfig::get_date_of_modification(QString program_type,
                                               QString program_index,
                                               QString config_name)
{
#ifdef MONGO_DRIVERS
    if(mongo_alive) {
        return mongo_db.get_date_of_modification(program_type,
                                                 program_index,
                                                 config_name);
    }
#endif
    return local_db.get_date_of_modification(program_type,
                                             program_index,
                                             config_name);
}

bool BaseConfig::transform_to_phantom(RootConfig root_config)
{
    QString program_type = root_config.get_data<QString>("program_type");
    QString program_index = root_config.get_data<QString>("program_index");
    QString configuration_name = root_config.get_data<QString>("configuration_name");
    QDateTime date_of_modification = root_config.get_modification_time();

    RootConfig phCfg;
    phCfg.set_data("program_type", program_type);
    phCfg.set_data("program_index", program_index);
    phCfg.set_data("configuration_name", configuration_name);
    phCfg.set_data("date_of_modification", date_of_modification);
    phCfg.set_data("delete_datetime", QDateTime::currentDateTime());
    phCfg.set_data("delete_host", QHostInfo::localHostName());

    if(program_type.isEmpty()){
        return false;
    }

    bool ok = true;
#ifdef MONGO_DRIVERS
    if(mongo_alive){
        if(!mongo_db.delete_config(program_type, program_index, configuration_name)){
            qWarning() << "Remove configuration from Mongo failed. Collection:"
                       << COLLECTION_SAVED;
        }
        ok &= mongo_db.write_config(phCfg);
    }
#endif

    ok &= local_db.delete_config(program_type, program_index, configuration_name);
    if(!mongo_alive) {
        ok &= local_db.write_config(phCfg);
    }
    return true;
}

bool BaseConfig::delete_config(QString program_type, QString program_index,
                               QString config_name)
{
    if(program_type.isEmpty()){
        return false;
    }

    if(mongo_alive){
        #ifdef MONGO_DRIVERS
        if(!mongo_db.delete_config(program_type, program_index, config_name)){
            qWarning() << "Remove configuration from Mongo failed. Collection:"
                       << COLLECTION_SAVED;
        }
        #endif
    }

    local_db.delete_config(program_type, program_index, config_name);
    return true;
}

bool BaseConfig::write_history_config(RootConfig root)
{
    QString program_type = root.get_data<QString>("program_type");
    QString run_index = root.get_data<QString>("run_index");
    int run_number = root.get_data<int>("run_number",
                                        std::numeric_limits<int>::max());

    bool check_primary_keys = false;
    if(program_type.isEmpty()){
         qWarning() << "Write history config warning: program_type is empty";
         check_primary_keys = true;
    }
    if(run_index.isEmpty()){
         qWarning() << "Write history config warning: run_index is empty";
         check_primary_keys = true;
    }

    if(run_number == std::numeric_limits<int>::max()){
         qWarning() << "Write history config warning: run_number is empty";
         check_primary_keys = true;
    }

    if(check_primary_keys){
        return false;
    }

    //---change date of creation to current date_time
    QDateTime current_dt = QDateTime::currentDateTime();
    root.set_data("date_of_modification", current_dt);

    //---set current host name
    root.set_data("current host", QHostInfo::localHostName());

    if(mongo_alive){
        #ifdef MONGO_DRIVERS
        mongo_db.write_history_config(root);
        #endif
    }


    local_db.write_history_config(root);

    return true;
}

int BaseConfig::check_run_number(QString program_type, QString run_index, int run_number)
{
    if(mongo_alive){
        #ifdef MONGO_DRIVERS
        return mongo_db.check_run_number(program_type, run_index, run_number);
        #endif
    }

    return local_db.check_run_number(program_type, run_index, run_number);
}

bool BaseConfig::write_device_config(RootConfig root)
{
    bool to_return = true;

    QString family = root.get_data<QString>("family");
    if(family.isEmpty()){
        qWarning() << "write device config failed. Can't find field family";
        return false;
    }

    quint32 serial = root.get_data<quint32>("serial", 0);
    if(serial == 0){
        qWarning() << "write device config failed. Can't find field serial";
        return false;
    }

    //---change date of creation to current date_time
    QDateTime current_dt = QDateTime::currentDateTime();
    root.set_data("date_of_modification", current_dt);

    //---set current host name
    root.set_data("current host", QHostInfo::localHostName());

    if(mongo_alive){
        #ifdef MONGO_DRIVERS
        //---change date of creation to current date_time
        to_return &= mongo_db.write_device_config(root);
        #endif
    }
    else{
        qWarning() << "write device config failed. Mongo database is not alived";
    }
    to_return = local_db.write_device_config(root);

    return  to_return;
}

//------------- LOCAL ----- MONGO ---
// case 1     [normal]    [normal]
// case 2     [normal]    [fantom]
// case 3     [normal]    [      ]
// case 4     [fantom]    [normal]
// case 5     [fantom]    [fantom]
// case 6     [fantom]    [      ]
// case 7     [      ]    [normal]
// case 8     [      ]    [fantom]

#ifdef MONGO_DRIVERS
bool BaseConfig::synchronization(QString program_type)
{
    Q_ASSERT(!program_type.isEmpty());
    qInfo() << "Mongo synchronization to local";
    if(!local_db.local_db_chekup()){
        qWarning().noquote() << QString("\033[91m") +
                                QString("Mongo synchronization failed! ")+
                                QString("\033[0m");
        return false;
    }
    auto list_local = local_db.get_rootconfig_map(program_type);
    auto list_mongo = mongo_db.get_rootconfig_map(program_type);

    for(auto pi : list_local.keys()) {
        auto &list_local_pi = list_local[pi];
        for(auto cn : list_local_pi.keys()) {
            auto &rc_local = list_local_pi[cn];
            QDateTime local_date = rc_local.get_delete_time();
            bool local_dead = !local_date.isNull();
            if(!local_dead)
                local_date = rc_local.get_modification_time();

            bool contains = list_mongo.contains(pi) && list_mongo[pi].contains(cn);
            if(!contains) {
                if(local_dead){
                    local_db.delete_config(program_type, pi, cn); // case 6
                }
                else{
                    mongo_db.write_config(rc_local); // case 3
                }
                continue;
            }

            auto &rc_mongo = list_mongo[pi][cn];
            QDateTime mongo_date = rc_mongo.get_delete_time();
            bool mongo_dead = !mongo_date.isNull();
            if(!mongo_dead)
                mongo_date = rc_mongo.get_modification_time();
            bool local_is_newer = local_date > mongo_date;
            if(local_is_newer) {
                if(local_dead) {
                    mongo_db.delete_config(program_type, pi, cn); // case 4 and 5
                    local_db.delete_config(program_type, pi, cn);
                }
                else{
                    mongo_db.write_config(rc_local); // case 1 and 2
                }

            } else {
                if(mongo_dead) {
                    local_db.delete_config(program_type, pi, cn); // case 2 and 5
                    list_local_pi.remove(cn);
                }
                else{
                    if(rc_local != rc_mongo)
                        local_db.write_config(rc_mongo, false); // case 1 and 4
                }

            }
        }
    }
    local_db.sync();

    for(auto pi : list_mongo.keys()) {
        auto &list_mongo_pi = list_mongo[pi];
        for(auto cn : list_mongo_pi.keys()) {
            auto &rc_mongo = list_mongo_pi[cn];
            bool contains = list_local.contains(pi) && list_local[pi].contains(cn);
            if(contains) continue; // case 1, 2, 4, 5
            if(!rc_mongo.get_delete_time().isNull()) continue; // case 8
            local_db.write_config(rc_mongo); // case 7
        }
    }
    return true;
}
#endif

void BaseConfig::root_debug_print(RootConfig tree_item)
{
    qInfo() << "===ROOT DEBUG PRINT===";

    qInfo() << tree_item.id.toString();
    for(QString data_field : tree_item.data().keys()){
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::Int){
            qInfo() << "MY FIELD:"<< data_field << tree_item.get_data<int>(data_field);
        }
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QString){
            qInfo() << "MY FIELD:"<< data_field << tree_item.get_data<QString>(data_field);
        }
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::Bool){
           qInfo() << "MY FIELD:" << data_field << tree_item.get_data<bool>(data_field);
        }

        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QByteArray){
            qInfo() << "MY FIELD:" << data_field << tree_item.get_data<QByteArray>(data_field);
        }

        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QDateTime){
            qInfo() << "MY FIELD:" << data_field << tree_item.get_data<QDateTime>(data_field);
        }
    }

    for(QString client : tree_item.children().keys()){
        qInfo() << "MY CHILD:" << client;
        root_debug_print(tree_item.children().value(client));
    }
}

void BaseConfig::init_tests()
{
//    RootConfig row_1;
//    row_1.set_data("program_type", QString("test"));
//    row_1.set_data("program_index", QString("1"));
//    row_1.set_data("configuration_name", QString("1"));
//    row_1.set_data("date_of_modification", QDateTime::currentDateTime());


//    RootConfig row_3;
//    row_3.set_data("program_type", QString("test"));
//    row_3.set_data("program_index", QString("1"));
//    row_3.set_data("configuration_name", QString("2"));
//    row_3.set_data("date_of_modification", QDateTime::currentDateTime());
//    qInfo() << "=== cur 3 " << QDateTime::currentDateTime();

//    sleep(3);

    RootConfig row_2;
    row_2.set_data("program_type", QString("test"));
    row_2.set_data("program_index", QString("1"));
    row_2.set_data("configuration_name", QString("2"));
    row_2.set_data("date_of_modification", QDateTime::currentDateTime());
    qInfo() << "=== cur 2 " << QDateTime::currentDateTime();

//    local_db.write_config(row_1);
//    local_db.write_config(row_3);
//    mongo_db.write_config(row_2);
}


//---- garbage
//    init_tests();

//    QElapsedTimer timer1;
//    timer1.start();

//    qInfo() << "--------- local ms= " << timer1.elapsed();
//    for(QString program_index : list_local.keys()){
//        for(QString config_name : list_local[program_index].keys()){
//            RootConfig root_local = list_local[program_index][config_name];
//            QString prog_type = root_local.get_data<QString>("program_type");
//            QString program_index = root_local.get_data<QString>("program_index");
//            QString config_name3 = root_local.get_data<QString>("configuration_name");
//            qInfo() << "=== " << prog_type << " " << program_index << " " << config_name3;
//        }
//    }

//    QElapsedTimer timer2;
//    timer2.start();

//    qInfo() << "--------- mongo  ms= " << timer2.elapsed();
//    for(QString program_index : list_mongo.keys()){
//        for(QString config_name : list_mongo[program_index].keys()){
//            RootConfig root_mongo = list_mongo[program_index][config_name];
//            QString prog_type = root_mongo.get_data<QString>("program_type");
//            QString program_index = root_mongo.get_data<QString>("program_index");
//            QString config_name3 = root_mongo.get_data<QString>("configuration_name");
//            qInfo() << "=== " << prog_type << " " << program_index << " " << config_name3;
//        }
//    }

//    for(QString pi_local : list_local.keys()){
//        for(QString cn_local : list_local[pi_local].keys()){
//            bool found = false;
//            RootConfig local = list_local[pi_local][cn_local];
//            RootConfig mongo;
//            for(QString pi_mongo : list_mongo.keys()){
//                for(QString cn_mongo : list_mongo[pi_mongo].keys()){
//                    if((pi_local == pi_mongo) && (cn_local== cn_mongo)){
//                        found = true;
//                        mongo = list_mongo[pi_mongo][cn_mongo];
//                        if(!local.is_phantom() && !local.is_phantom())
//                            synchronization_case(1, local, mongo);
//                        if(!local.is_phantom() && mongo.is_phantom())
//                            synchronization_case(2, local, mongo);
//                        if(local.is_phantom() && !mongo.is_phantom())
//                            synchronization_case(3, local, mongo);
//                        if(local.is_phantom() && mongo.is_phantom())
//                            synchronization_case(4, local, mongo);
//                    }
//                }
//            }
//            if(!found){
//                if(local.is_phantom()){
//                    synchronization_case(5, local, mongo);
//                }
//                else{
//                    synchronization_case(6, local, mongo);
//                }
//            }
//        }
//    }

//    for(QString pi_mongo : list_mongo.keys()){
//        for(QString cn_mongo : list_mongo[pi_mongo].keys()){
//            bool found = false;
//            RootConfig mongo = list_mongo[pi_mongo][cn_mongo];
//            RootConfig local;
//            for(QString pi_local : list_local.keys()){
//                for(QString cn_local : list_local[pi_mongo].keys()){
//                    if((pi_local == pi_mongo) && (cn_local== cn_mongo)){
//                        found = true;
//                    }
//                }
//            }
//            if(!found){
//                if(mongo.is_phantom()){
//                    synchronization_case(7, local, mongo);
//                }
//                else{
//                    synchronization_case(8, local, mongo);
//                }
//            }
//        }
//    }

//                    qInfo() << "=== local pr ind: " << program_index_local << " " << config_name_local << " " << program_index_mongo << " "<< config_name_mongo ;
    /*

*/

//    local_db.clear_saved_settings();
//    QStringList types_list = mongo_db.get_types_list(COLLECTION_SAVED);
//    for(int i=0; i<types_list.size(); ++i){
//        QMap<QString, QStringList> config_map = mongo_db.get_config_map(COLLECTION_SAVED, types_list[i]);
//        for(QString program_index : config_map.keys()){
//            for(int k=0; k<config_map[program_index].size(); ++k){
//                RootConfig root = mongo_db.read_config(types_list[i], program_index, config_map[program_index][k]);
//                local_db.write_config(root);
//            }
//        }
//    }

//    local_db.clear_history_settings();
//    QStringList hist_types_list = mongo_db.get_types_list(COLLECTION_HISTORY);
//    for(int i=0; i<hist_types_list.size(); ++i){
//        QMap<QString, QStringList> hist_config_map = mongo_db.get_config_map(COLLECTION_HISTORY, hist_types_list[i]);
//        for(QString run_index : hist_config_map.keys()){
//            for(int k=0; k<hist_config_map[run_index].size(); ++k){
//                RootConfig root = mongo_db.read_history_config(hist_types_list[i], run_index, hist_config_map[run_index][k].toInt());
//                local_db.write_history_config(root);
//           }
//       }
//    }
