#include "StartupOptions.h"
#include "BaseConfig.h"

#include <QCoreApplication>
#include <QDebug>

const int CELL_WIDTH = 25;

StartupOptions::StartupOptions(const QString &program_type) :
    program_type(program_type),
    program_index(PROGRAM_INDEX_DEFAULT),
    configuration_name(PROGRAM_CONFIG_DEFAULT),
    list(false),
    version(false),
    debug_mode(false)
{
    QString rev;
    #ifdef GIT_DESCR
        rev = QString("%1").arg(GIT_DESCR);
        rev = rev.remove(QRegExp("^.*\\:"));
    #endif
    program_version = rev;
    setApplicationDescription(QCoreApplication::applicationName());
    addHelpOption();

    addPositionalArgument("program_index",
    QCoreApplication::translate("main","with which index to run the program"));

    QCommandLineOption list_option( QStringList()  << "l" <<  "list", "List available configurations");

    QCommandLineOption version_option( QStringList()  << "v" <<  "version", "Current program version");

    QCommandLineOption config_option(
        QStringList() << "config",
        QCoreApplication::translate("main", "Startup with certain configuration"),
        "config"
    );

    QCommandLineOption debug_option( QStringList()  << "d" <<  "debug", "Run in debug mode");

    addOption(list_option);
    addOption(version_option);
    addOption(config_option);
    addOption(debug_option);
}

void StartupOptions::parse_options()
{
    process(QCoreApplication::arguments());

    const QStringList args = positionalArguments();
    if (!args.empty()) {
        program_index = args[0];
    }

    list = isSet("list");
    version = isSet("version");
    debug_mode = isSet("debug");

    BaseConfig base_config;
    auto config_map = base_config.get_config_map(program_type);

    if(list){
        show_config_list(config_map);
    }

    if(version){
        if(program_version.isEmpty()){
            qInfo().noquote() << QString("\033[91m") + QString("Unknown program version") + QString("\033[0m");

        }
        else{
            qInfo().noquote() << QString("\033[32m") + QString("program version: %1").arg(program_version) + QString("\033[0m");
        }
    }

    if(isSet("config")){
        configuration_name = QString::fromStdString(value("config").toStdString());
        if(!config_map[program_index].contains(configuration_name)){
            qCritical().noquote() << QString("No configuration called \"%1\"").arg(configuration_name);
            show_config_list(config_map);
            showHelp();
        }
        if(configuration_name.isEmpty()||(configuration_name.startsWith('-'))) {
            showHelp();
        }
    }
}

void StartupOptions::show_config_list(const QMap<QString, QStringList> &config_map)
{
    QString output_str;
    output_str = ":\n";
    if(config_map.isEmpty()){
        output_str.append(QString("No saved %1 configurations. \n").arg(program_type));
    }
    else{
        output_str.append(QString("List of saved %1 configurations: \n").arg(program_type));
    }

    QString  separator;
    for(int i=0; i<(2*CELL_WIDTH+1); ++i){
        separator.append("-");
    }

    separator.prepend("+");
    separator.append("+");
    separator.append("\n");

    int tail_1 = CELL_WIDTH - QString(" Program index").length();
    QString tail_1_str;
    for(int i=0; i<tail_1; ++i){
        tail_1_str.append(" ");
    }

    int tail_2 = CELL_WIDTH - QString(" Configuration name").length();
    QString tail_2_str;
    for(int i=0; i<tail_2; ++i){
        tail_2_str.append(" ");
    }

    output_str.append(separator);
    QString labels = QString("|") + "\033[91m" + " Program index" + "\033[0m" + tail_1_str +
                     "|" + "\033[91m" + " Configuration name" + "\033[0m" + tail_2_str +  "|" + "\n";
    output_str.append(labels);
    output_str.append(separator);

    for (const QString& program_index : config_map.keys()) {
        for(int k=0; k<config_map[program_index].size(); ++k){
            QString configuration_name = config_map[program_index][k];
            int tail_1 = CELL_WIDTH - program_index.length() -1;
            int tail_2 = CELL_WIDTH - configuration_name.length() -1;

            QString tail_1_str;
            for(int i=0; i<tail_1; ++i){
                tail_1_str.append(" ");
            }

            QString tail_2_str;
            for(int i=0; i<tail_2; ++i){
                tail_2_str.append(" ");
            }

            QString table_data = QString("| ") + program_index + tail_1_str + "| " +
                                 configuration_name + tail_2_str +  "|" + "\n";
            output_str.append(table_data);
            output_str.append(separator);
        }
    }

    output_str.append("\n");
    qInfo().noquote() << output_str;
}
