
#include "SingleMultiCastListener.h"

#include <QNetworkInterface>
#include <QTimer>
#include <QUdpSocket>

SingleMultiCastListener::SingleMultiCastListener(QObject *parent,
                                                 QNetworkInterface interface,
                                                 QHostAddress _groupAddress,
                                                 quint16 port) :
    QObject(parent),
    udpSocket(nullptr),
    groupAddress(_groupAddress)
{
    qRegisterMetaType<QHostAddress>();

    udpSocket = new QUdpSocket(this);
    connect(udpSocket, &QUdpSocket::readyRead, this, &SingleMultiCastListener::processPendingDatagrams);

    QHostAddress bindAddress = QHostAddress::AnyIPv4;
    QUdpSocket::BindMode bindMode = QUdpSocket::ShareAddress | QAbstractSocket::ReuseAddressHint;
    if (! udpSocket->bind(bindAddress, port, bindMode)) {
        qCritical() << interface.humanReadableName() << QString("bind(%1, %2, %3): %4")
                       .arg(bindAddress.toString()).arg(port).arg(bindMode).arg(udpSocket->errorString());
        delete udpSocket;
        return;
    }
    udpSocket->setMulticastInterface(interface);

    QTimer::singleShot(0, this, &SingleMultiCastListener::start);
}


SingleMultiCastListener::~SingleMultiCastListener()
{
    stop();
}

void SingleMultiCastListener::processPendingDatagrams()
{
    if (!udpSocket) return;
    while (udpSocket->hasPendingDatagrams()) {
        QByteArray datagram;
        QHostAddress host;
        datagram.resize(static_cast<int>(udpSocket->pendingDatagramSize()));
        udpSocket->readDatagram(datagram.data(), datagram.size(), &host);
        emit received(datagram, host);
    }
}

void SingleMultiCastListener::start()
{
    if (!udpSocket) return;
    const QNetworkInterface &interface = udpSocket->multicastInterface();
    const quint16 port = udpSocket->localPort();


    if (! udpSocket->joinMulticastGroup(groupAddress, interface)) {
        qCritical() << QString("joinMulticastGroup %1 on %2: %3")
                           .arg(groupAddress.toString())
                           .arg(interface.humanReadableName())
                           .arg(udpSocket->errorString());
        delete udpSocket;
        return;
    }
    qInfo() << QString("joined multicast group %1 port %2 on %3")
                   .arg(groupAddress.toString())
                   .arg(port)
                   .arg(interface.humanReadableName());
}

void SingleMultiCastListener::stop()
{
    if (!udpSocket) return;
    const QNetworkInterface &interface = udpSocket->multicastInterface();
    if (interface.flags() & QNetworkInterface::IsUp) {
        if (! udpSocket->leaveMulticastGroup(groupAddress, interface)) {
            qCritical() << QString("leaveMulticastGroup %1 on %2: %3")
                           .arg(groupAddress.toString())
                           .arg(interface.humanReadableName())
                           .arg(udpSocket->errorString());
        } else {
            qInfo() << QString("left multicast group %1 port %2 on %3")
                       .arg(groupAddress.toString())
                       .arg(udpSocket->localPort())
                       .arg(interface.humanReadableName());
        }
    }
    disconnect(udpSocket, &QUdpSocket::readyRead, nullptr, nullptr);
    udpSocket->close();
}
