#include "AddPnpDialog.h"
#include "ui_AddPnpDialog.h"

#include <QStringList>
#include <QHeaderView>
#include <QTableWidget>
#include <QComboBox>
#include <QPushButton>
#include "pnp-server/PNPMessage.h"

namespace {
    enum COLUMNS{COL_PROG_TYPE, COL_HOST_NAME, COL_PROG_INDEX, COL_TOTAL};
}

AddPnpDialog::AddPnpDialog(QVector<ProgramDescription> progDescrs, bool multi_select, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::AddClientDialog),
    multi_select(multi_select),
    filter_func(nullptr)
{
    ui->setupUi(this);
    ui->labelProgInfo->hide();

    for(int i=0; i<progDescrs.size(); ++i){
        progs[progDescrs[i].uuid] = progDescrs[i];
    }

    init_table(ui->leftTableWidget);

    if(multi_select) {
        init_table(ui->rightTableWidget);
    } else {
        ui->rightTableWidget->hide();
        ui->pushButtonAdd->hide();
        ui->pushButtonRemove->hide();
    }

    populate_left_table();
    populate_all_filter();
    check_ok_button_enable();
}

//---public
AddPnpDialog::~AddPnpDialog()
{
    delete ui;
}

QString AddPnpDialog::getProgType() const
{
    if(multi_select && !selected_progs.isEmpty())
        return "";
    else
        return selected_progs.first().type;
}

QString AddPnpDialog::getProgHost() const
{
    if(multi_select && !selected_progs.isEmpty())
        return "";
    else
        return selected_progs.first().hostName;
}

QString AddPnpDialog::getProgIndex() const
{
    if(multi_select && !selected_progs.isEmpty())
        return "";
    else
        return selected_progs.first().index;
}

QVector<ProgramDescription> AddPnpDialog::getProgDescr() const
{
    return selected_progs.values().toVector();
}

void AddPnpDialog::set_program_type_filter(QStringList filter)
{
    global_type_filter = filter;

    ui->rightTableWidget->setRowCount(0);
    populate_left_table();
    populate_all_filter();
}

void AddPnpDialog::set_function_filter(std::function<bool(const ProgramDescription&)> f)
{
    filter_func = f;
    if(!filter_func) return;

    for(const auto &pd : progs.values()) {
        if(!filter_func(pd)) {
            int row = find_table_row(ui->leftTableWidget, pd);
            if(row>=0)
                ui->leftTableWidget->removeRow(row);
            progs.remove(pd.uuid);
        }
    }

    populate_left_table();
    populate_all_filter();
}

//---pivate
void AddPnpDialog::populate_left_table()
{
    auto missing_progs = progs.keys();
    auto table = ui->leftTableWidget;
    for(int row=0; row< table->rowCount(); ++row){
        auto pd = row_to_progdescr(table, row);
        if(pd)
            missing_progs.removeAll(pd->uuid);
    }

    for (QUuid uuid : missing_progs) {
        const auto &pd = progs[uuid];
        if(!global_type_filter.isEmpty() && !global_type_filter.contains(pd.type)){
            continue;
        }
        if(!filter_func || filter_func(pd))
            table_append_row(ui->leftTableWidget, pd);
    }
}

void AddPnpDialog::populate_all_filter()
{
    populate_program_filter(ui->comboBoxType);
    populate_program_filter(ui->comboBoxHost);
    populate_program_filter(ui->comboBoxIndex);
}

void AddPnpDialog::populate_program_filter(QComboBox *box)
{
    QTableWidget *table = ui->leftTableWidget;
    QString *cur_filter=nullptr;
    QString type = program_type_filter, host = host_name_filter, index = program_index_filter;
    QString default_item;
    QString ProgramDescription:: * pd_field;
    if(box==ui->comboBoxType) {
        cur_filter = &program_type_filter;
        type.clear();
        QString type = program_type_filter, host = host_name_filter, index = program_index_filter;
        default_item = "All types";
        pd_field = &ProgramDescription::type;
    } else if(box==ui->comboBoxHost) {
        cur_filter = &host_name_filter;
        host.clear();
        default_item = "All hosts";
        pd_field = &ProgramDescription::hostName;
    } else if(box==ui->comboBoxIndex) {
        cur_filter = &program_index_filter;
        index.clear();
        default_item = "All indexes";
        pd_field = &ProgramDescription::index;
    }

    QStringList filter_items;
    for(int row=0; row<table->rowCount(); ++row){
        ProgramDescription *pd = row_to_progdescr(table, row);
        if(multi_select){
            if(pd && is_prog_match_filter(*pd, type, host, index)  && !selected_progs.contains(pd->uuid)) {
                const QString &val = pd->*pd_field;
                if(!filter_items.contains(val)){
                    filter_items.append(val);
                }
            }
        }
        else{
            if(pd && is_prog_match_filter(*pd, type, host, index)) {
                const QString &val = pd->*pd_field;
                if(!filter_items.contains(val)){
                    filter_items.append(val);
                }
            }
        }
    }
    std::sort(filter_items.begin(), filter_items.end());
    filter_items.prepend(default_item);

    bool old_state = box->blockSignals(true);
    for(int i=0; i<box->count(); ){
        if(filter_items.contains(box->itemText(i)))
            ++i;
        else
            box->removeItem(i);
    }
    for(int i=0; i<filter_items.size();){
        auto item = filter_items[i];
        if(box->itemText(i) != item)
            box->insertItem(i, item);
        else
            ++i;
    }

    box->blockSignals(old_state);
    set_current_index(box, *cur_filter);
}

void AddPnpDialog::filter_item_changed()
{
    QTableWidget *table = ui->leftTableWidget;
    for(int row=0; row<table->rowCount(); ++row) {
        ProgramDescription *pd = row_to_progdescr(table, row);
        if(!pd) continue;
        if(multi_select){
            if(is_prog_match_filter(*pd) && !selected_progs.contains(pd->uuid))
                table->showRow(row);
            else
                table->hideRow(row);
        }
        else{
            if(is_prog_match_filter(*pd))
                table->showRow(row);
            else
                table->hideRow(row);
        }
    }

    populate_all_filter();
    check_add_button_enable();
}

void AddPnpDialog::update_detail_label(ProgramDescription *pd)
{
    ui->labelProgInfo->show();
    ui->labelProgInfo->clear();
    ui->labelProgInfo->setText(pd ? pd->toString() : "");
    ui->labelProgInfo->adjustSize();
    detailed_program_uuid = pd ? pd->uuid : QUuid();
}

void AddPnpDialog::set_current_index(QComboBox *box, QString &item)
{
    if(item.isEmpty()){
        box->setCurrentIndex(0);
        return;
    }

    for(int i=0; i<box->count(); ++i){
        if(box->itemText(i) == item){
            box->setCurrentIndex(i);
            return;
        }
    }

    item.clear();
    box->setCurrentIndex(0);
    filter_item_changed();
}

void AddPnpDialog::init_table(QTableWidget *table)
{
    table->setColumnCount(COL_TOTAL);
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    table->setHorizontalHeaderItem(COL_PROG_TYPE, new QTableWidgetItem("Program type"));
    table->setHorizontalHeaderItem(COL_HOST_NAME, new QTableWidgetItem("Host name"));
    table->setHorizontalHeaderItem(COL_PROG_INDEX, new QTableWidgetItem("Program index"));

    QString table_style = (
    "   QTableView {                            "
    "       border: 1px solid gray;             "
    "       border-radius: 5px;                 "
    "   }                                       "
    );

    table->setStyleSheet(table_style);
    table->verticalHeader()->setVisible(false);
    if(!multi_select)
        table->setSelectionMode(QAbstractItemView::SingleSelection);
    table->horizontalHeader()->setStretchLastSection(true);
}

void AddPnpDialog::table_append_row(QTableWidget *table, const ProgramDescription &pd)
{
    bool old_state = table->blockSignals(true);

    int row = find_table_row(table, pd);
    if(row >= 0) {
        table->showRow(row);
    } else {
        int row_count = find_table_proper_row(table, pd);
        table->insertRow(row_count);
        table->setItem(row_count, COL_PROG_TYPE, new QTableWidgetItem(pd.type));
        table->setItem(row_count, COL_HOST_NAME, new QTableWidgetItem(pd.hostName));
        table->setItem(row_count, COL_PROG_INDEX, new QTableWidgetItem(pd.index));
        table->item(row_count, COL_PROG_TYPE)->setData(Qt::UserRole, pd.uuid.toRfc4122());
    }

    table->resizeColumnsToContents();
    table->blockSignals(old_state);
}

void AddPnpDialog::push_button_add_enable(bool enable)
{
    QPushButton *button = ui->pushButtonAdd;
    if(enable){
        button->setEnabled(true);
        change_button_color(button, QColor(159, 255, 73));
    }
    else{
        button->setEnabled(false);
        change_button_color(button, QColor(255, 255, 255));
    }
}

void AddPnpDialog::push_button_remove_enable(bool enable)
{
    QPushButton *button = ui->pushButtonRemove;
    if(enable){
        button->setEnabled(true);
        change_button_color(button, QColor(248, 86, 27));
    }
    else{
        button->setEnabled(false);
        change_button_color(button, QColor(255, 255, 255));
    }
}

void AddPnpDialog::change_button_color(QPushButton *push_button, QColor color)
{
    QPalette pal = push_button->palette();
    pal.setColor(QPalette::Button, color);
    push_button->setAutoFillBackground(true);
    push_button->setPalette(pal);
    push_button->update();
}

QUuid AddPnpDialog::row_to_uuid(QTableWidget *table, int row)
{
    auto item = table->item(row, COL_PROG_TYPE);
    if(!item) {
        qWarning() << "Can't define uuid in row_to_uuid method";
        return QUuid();
    }
    return QUuid::fromRfc4122(item->data(Qt::UserRole).toByteArray());
}

ProgramDescription *AddPnpDialog::row_to_progdescr(QTableWidget *table, int row)
{
    auto uuid = row_to_uuid(table, row);
    auto &prog_list = (table==ui->leftTableWidget) ? progs : selected_progs;
    auto progIt = prog_list.find(uuid);
    if(progIt == prog_list.end()){
        qWarning() << "Can't find program description in row_to_progdescr method";
        return nullptr;
    }
    return &progIt.value();
}

void AddPnpDialog::check_add_button_enable()
{
    bool found=false;
    for(auto range : ui->leftTableWidget->selectedRanges()) {
        for(int row=range.topRow(); row<=range.bottomRow(); ++row){
            if(!ui->leftTableWidget->isRowHidden(row)) {
                found = true;
                break;
            }
        }
        if(found) break;
    }
    push_button_add_enable(found);
}

void AddPnpDialog::check_ok_button_enable()
{
    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(!selected_progs.isEmpty());
}

int AddPnpDialog::find_table_proper_row(QTableWidget *table, const ProgramDescription &pd)
{
    int row =0;
    for(;row<table->rowCount(); ++row){
        ProgramDescription *prog_descr = row_to_progdescr(table, row);
        if(!prog_descr) continue;

        if(pd.type < prog_descr->type){
            return row;
        }else if(pd.type == prog_descr->type) {
            if(pd.hostName < prog_descr->hostName){
                return row;
            }else if(pd.hostName == prog_descr->hostName) {
                if(pd.index < prog_descr->index){
                    return row;
                }else if(pd.index == prog_descr->index) {
                    return row;
                }
            }
        }
    }
    return row;
}

int AddPnpDialog::find_table_row(QTableWidget *table, const ProgramDescription &pd)
{
    for(int row =0;row<table->rowCount(); ++row){
        auto uuid = row_to_uuid(table, row);
        if(uuid == pd.uuid)
            return row;
    }
    return -1;
}

bool AddPnpDialog::is_prog_match_filter(const ProgramDescription &pd) const
{
    return is_prog_match_filter(pd, program_type_filter, host_name_filter, program_index_filter);
}

bool AddPnpDialog::is_prog_match_filter(const ProgramDescription &pd, QString type, QString host, QString index) const
{
    return (((pd.type == type) || (type.isEmpty())) &&
       ((pd.hostName == host) || (host.isEmpty())) &&
            ((pd.index == index || (index.isEmpty()))));
}

void AddPnpDialog::add_prog(int row)
{
    ProgramDescription *pd = row_to_progdescr(ui->leftTableWidget, row);
    ui->leftTableWidget->hideRow(row);
    if(pd) {
        table_append_row(ui->rightTableWidget, *pd);
        selected_progs[pd->uuid] = *pd;
    }
}

void AddPnpDialog::remove_prog(int row)
{
    ProgramDescription *pd = row_to_progdescr(ui->rightTableWidget, row);
    auto item = ui->rightTableWidget->item(row, COL_PROG_TYPE);
    if(pd) {
        if(item->backgroundColor() != Qt::gray)
            table_append_row(ui->leftTableWidget, *pd);
        selected_progs.remove(pd->uuid);
    }
    ui->rightTableWidget->removeRow(row);
}

void AddPnpDialog::set_background_row_color(int row, QColor color) const
{
    for(int col=0; col< COL_TOTAL; ++col)
        ui->rightTableWidget->item(row, col)->setBackgroundColor(color);
}

//--- public slots
void AddPnpDialog::gotProgramDescription(ProgramDescription pd)
{
    if((global_type_filter.isEmpty() || global_type_filter.contains(pd.type))
            && (!filter_func || filter_func(pd))) {
        const bool add_row = !progs.contains(pd.uuid);
        progs[pd.uuid] = pd;
        if(add_row)
            table_append_row(ui->leftTableWidget, pd);
    } else {
        if(progs.contains(pd.uuid)) {
            const int row = find_table_row(ui->leftTableWidget, pd);
            if(row>=0)
                ui->leftTableWidget->removeRow(row);
            progs.remove(pd.uuid);
        }
    }

    if(selected_progs.contains(pd.uuid)) {
        selected_progs[pd.uuid] = pd;
        int row = find_table_row(ui->rightTableWidget, pd);
        if(row >= 0)
            set_background_row_color(row, Qt::transparent);
    }
    if(detailed_program_uuid == pd.uuid) {
        update_detail_label(&pd);
    }

    populate_left_table();
    populate_all_filter();
}

void AddPnpDialog::gotClosedProgram(const ProgramDescription &closedDescr)
{
    if(!progs.contains(closedDescr.uuid))
        return;

    for(int row=0; row<ui->leftTableWidget->rowCount(); ++row){
        ProgramDescription *pd = row_to_progdescr(ui->leftTableWidget, row);
        if(pd && closedDescr.uuid == pd->uuid){
            ui->leftTableWidget->removeRow(row);
            progs.remove(closedDescr.uuid);
            break;
        }
    }
    for(int row=0; row<ui->rightTableWidget->rowCount(); ++row){
        ProgramDescription *pd = row_to_progdescr(ui->rightTableWidget, row);
        if(pd && closedDescr.uuid == pd->uuid)
            set_background_row_color(row, Qt::gray);
    }
    if(detailed_program_uuid == closedDescr.uuid)
        update_detail_label(nullptr);
}

//--- private slots
void AddPnpDialog::on_comboBoxType_currentIndexChanged(const QString &type)
{
    if(type == "All types")
        program_type_filter.clear();
    else
        program_type_filter = type;

    filter_item_changed();
}

void AddPnpDialog::on_comboBoxHost_currentIndexChanged(const QString &host)
{
    if(host == "All hosts")
        host_name_filter.clear();
    else
        host_name_filter = host;

    filter_item_changed();
}

void AddPnpDialog::on_comboBoxIndex_currentIndexChanged(const QString &index)
{
    if(index == "All indexes")
        program_index_filter.clear();
    else
        program_index_filter = index;

    filter_item_changed();
}

void AddPnpDialog::on_leftTableWidget_itemDoubleClicked(QTableWidgetItem *item)
{
    if(multi_select) {
        ui->rightTableWidget->clearSelection();

        add_prog(item->row());

        populate_all_filter();
        check_ok_button_enable();
        check_add_button_enable();
    } else {
        selected_progs.clear();
        auto pd = row_to_progdescr(ui->leftTableWidget, item->row());
        Q_ASSERT(pd);
        if(pd) {
            selected_progs[pd->uuid] = *pd;
            accept();
        }
    }
}

void AddPnpDialog::on_rightTableWidget_itemDoubleClicked(QTableWidgetItem *item)
{
    ui->leftTableWidget->clearSelection();
    push_button_remove_enable(false);

    remove_prog(item->row());

    populate_all_filter();
    check_ok_button_enable();
}

void AddPnpDialog::on_leftTableWidget_itemSelectionChanged()
{
    auto select = ui->leftTableWidget->selectedItems();

    ProgramDescription *pd=nullptr;
    if(!select.isEmpty()) {
        ui->rightTableWidget->clearSelection();
        pd = row_to_progdescr(ui->leftTableWidget, select.last()->row());
    }

    update_detail_label(pd);
    if(!multi_select) {
        selected_progs.clear();
        if(pd)
            selected_progs[pd->uuid] = *pd;
    }
    check_ok_button_enable();
    check_add_button_enable();
}

void AddPnpDialog::on_rightTableWidget_itemSelectionChanged()
{
    auto select = ui->rightTableWidget->selectedItems();
    push_button_remove_enable(!select.isEmpty());

    ProgramDescription *pd = nullptr;
    if(!select.isEmpty()){
        ui->leftTableWidget->clearSelection();
        pd = row_to_progdescr(ui->rightTableWidget, select.last()->row());
    }
    update_detail_label(pd);
}

void AddPnpDialog::on_pushButtonAdd_clicked()
{
    push_button_add_enable(false);
    QTableWidget *table = ui->leftTableWidget;
    for(auto range : table->selectedRanges()) {
        for(int row=range.topRow(); row<=range.bottomRow(); ++row){
            if(table->isRowHidden(row)) continue;
            ProgramDescription *pd = row_to_progdescr(table, row);
            if(pd) {
                table_append_row(ui->rightTableWidget, *pd);
                selected_progs[pd->uuid] = *pd;
            }
            table->hideRow(row);
        }
    }

    populate_all_filter();
    check_ok_button_enable();
}

void AddPnpDialog::on_pushButtonRemove_clicked()
{
    push_button_remove_enable(false);

    QTableWidget *table = ui->rightTableWidget;
    while(!table->selectedRanges().isEmpty())
        remove_prog(table->selectedRanges().first().bottomRow());

    populate_all_filter();
    check_ok_button_enable();
}
