#ifndef EVNUMCHECKER_H
#define EVNUMCHECKER_H

/*
 *  Local mode - EvNumChecker works at top level run control program
 *               inits check by evNumCheckTimer, gets 1st trigEvNum,
 *               collects evNum from clientMap and vmeMap
 *               gets 2nd trigEvNum, checks
 *  Non-local mode - EvNumChecker works at client level run control program
 *  Client mode - EvNumChecker works at device control program (like ADC64-system)
 *
 *  <1st (initial) evNum> <= {<clients evNum>, <vme evNum>, <dev evNum>} <= <last evNum>
 *  1st and last evNum - from top trig device
 *
 * Local mode (TTVXS with 1 module, U40VE-RC or UT24):
 * 1. Starts periodicaly
 * 2. Get 1st evNum
 * 3. Request sub-systems evNum
 * 4. Get last evNum
 * 5. Check
 *
 * Non-local mode:
 * 1. Wait request from Remote control
 * 2. Request evNum from sub-systems and trig devices (TTVXS with any modules, U40VE-RC or UT24)
 * 3. Sends all collected evNUm to Remote control
 *
 * Client mode:
 * 1. Wait request from Remote control
 * 2. Request evNum from controlled devices
 * 3. Sends all collected evNUm to Remote control
*/

#include "types.h"

#include <QObject>
#include <QMap>
#include <QSet>
#include <QVector>

class QTimer;

class EvNumChecker : public QObject
{
    Q_OBJECT
public:
    explicit EvNumChecker(QObject *parent = nullptr);
    virtual ~EvNumChecker();
    void stopCheck(bool clearWarnSet=true);

signals:
    void getTrigEvNum(int checkIndex, bool local=true); // to trig device (u40ve-rc/ttvxs)
    void evNumRequest(int checkIndex); // to Clients, FWME2TMWR and regular devices
    void completedEvNumCheck(int checkIndex, EvNumSet evNums);
    void checkFail(QString mess);

public slots:
    void startCheck(); // trigged by start run or by evNumCheckTimer
    void set_dev_serials(QVector<quint64> serialList);

    void updateClientMap(QMap<int, QString> map);
    void updateVmeMap(QMap<quint64, QString> map);

    void onRemoteEvNumCheckRequest(int checkIndex);
    void gotTopLevelEvNum(int checkIndex, quint64 evNum);
    void gotClientEvNum(int checkIndex, int clientIndex, EvNumSet evNum);
    void gotVmeTrigEvNum(int checkIndex, quint64 vmeSerial, quint64 evNum);
    void gotDevEvNum(int checkIndex, quint64 serial, quint64 evNum);

private slots:
    void checkTimeout();

private:
    enum WARN_ID{WARN_DOUBLE_REQUEST, WARN_MISSING_INITIAL, WARN_NOT_IN_RANGE,
                 WARN_EV_NUM_DECR, WARN_TIMEOUT, WARN_SUMMARY_ERR};

    void gotEvNum(int checkIndex, quint64 serial, quint64 evNum);
    bool checkAllCollected(bool info_only=false);
    void performTopLevelCheck(int checkIndex, quint64 evNum);
    void sendRemoteEvNumCheckResponce(int checkIndex);
    void setWarn(WARN_ID id, quint64 subId, QString msg);
    void setWarn(WARN_ID id, QString msg);
    void clearWarn(WARN_ID id, quint64 subId=0);

    bool completed;
    bool checkEnabled; // top-level check
    int curCheckIndex;
    QTimer * const evNumCheckTimer; // starts next checking at top-level run control
    QTimer * const checkExpireTimer; // every mode has its own check timeout

    quint64 trigEvNum; // 1st (initial) evNum from main trig device for local mode
    QMap<int, EvNumSet> clientEvNum; // list acquired evNum from clients
    QMap<quint64, quint64> evNumbers; // evNum list from device in client mode, TTVXS or form VME2TMWR

    QVector<quint64> devSerials; // pending devices (for client mode)
    QMap<int, QString> clientMap; // index:name (clients name)
    QMap<quint64, QString> vmeMap; // serial:name (user name of VME2TMWR devices)

    QMap<WARN_ID, QMap<quint64, QString> > warnSet;
};



#endif // EVNUMCHECKER_H
