#include "RemoteControlServer.h"
#include "RunControlCommands.h"
#include "pnp-server/pnp-server/PNPMessage.h"
#include "base/RcRunId.h"

#include "QCoreApplication"
#include "QTcpServer"
#include "QTcpSocket"
#include "unistd.h"
#include "stdexcept"

namespace {
const qint32 MAX_BLOCK_SIZE = 1e5;

#pragma pack(push, 1)
struct TypeLengthHeader {
    quint32 type;
    quint32 len;
};
#pragma pack(pop)
}

RemoteControlServer::RemoteControlServer(QObject *parent, quint16 port, QString _descr) :
    QObject(parent),
    qTcpServer(new QTcpServer(this)),
    tcpSocket(nullptr),
    descr(_descr.isEmpty() ? QCoreApplication::applicationName() : _descr),
    status(port)
{
    qRegisterMetaType<RemoteControlServerStatus> ();
    connect(qTcpServer, &QTcpServer::newConnection,
            this, &RemoteControlServer::updateConnections);
}

RemoteControlServer::~RemoteControlServer()
{
    blockSignals(true);
    startListen(false);
    delete qTcpServer;
    blockSignals(false);
}

void RemoteControlServer::config_transfer(QByteArray data) const
{
    send_byte_array(RC_MSG_CONFIG, data);
}

void RemoteControlServer::send_command_message(QString mess) const
{
    if(!status.listening)
        return;

    QByteArray mess_ba = mess.toUtf8();
    send_byte_array(RC_MSG_CMD, mess_ba);
}

void RemoteControlServer::send_byte_array(quint32 type, QByteArray payload) const
{
    if (!isConnectionAlive()) {
        return;
    }
    QByteArray data = QByteArray(sizeof(TypeLengthHeader), Qt::Uninitialized);
    data.append(payload);
    quint16 check = 0;

    TypeLengthHeader *hdr = reinterpret_cast<TypeLengthHeader *>(data.data());
    hdr->type = type;
    hdr->len = static_cast<quint32>(data.size()) + sizeof (check);

    check = qChecksum(data.data(), static_cast<uint>(data.size()), Qt::ChecksumIso3309);
    data.append(reinterpret_cast<char*>(&check), sizeof (check));

    int written = static_cast<int>(tcpSocket->write(data));

    if(written == data.size()) {
        if(type == RC_MSG_CONFIG || type == RC_MSG_CONFIG_OLD) {
            qInfo()<< "RemoteControlServer sent configuration";
        }
    } else {
        qWarning()<< "RemoteControlServer failed to send info string";
    }
}

void RemoteControlServer::readCommandRequest()
{
    try{
        while(tcpSocket->bytesAvailable()){
            data.append(tcpSocket->readAll());

            while(!data.isEmpty()) {
                bool type_warn=true, size_warn=true;
                TypeLengthHeader *hdr=nullptr;
                while(static_cast<uint>(data.size()) >= sizeof(TypeLengthHeader)) {
                    hdr= reinterpret_cast<TypeLengthHeader*>(data.data());
                    if((hdr->type == RC_MSG_CONFIG) || (hdr->type == RC_MSG_CONFIG_OLD)
                            || (hdr->type == RC_MSG_CMD) || (hdr->type == RC_MSG_CMD_OLD)) {
                        if(hdr->len <= MAX_BLOCK_SIZE) {
                            break;
                        } else if(size_warn) {
                            size_warn = false;
                            qWarning() << QString("RemoteControlServer message error: block size is %1. Permissible maximum is %2")
                                          .arg(hdr->len).arg(MAX_BLOCK_SIZE);
                        }
                    } else if(type_warn) {
                        qWarning() << "RemoteControlServer message error: unknown message type";
                        type_warn = false;
                    }

                    data.remove(0, 1);
                    hdr = nullptr;
                }
                if(!hdr)
                    continue;

                if(data.size() < static_cast<int>(hdr->len))
                    continue;

                quint32 tsd_size = hdr->len - sizeof(quint16);
                QByteArray tsd_data = data.left(static_cast<int>(tsd_size));
                quint16 check_calc = qChecksum(tsd_data.data(), tsd_size, Qt::ChecksumIso3309);

                quint16 *check = reinterpret_cast<quint16*>(data.data()+tsd_size);
                if(*check != check_calc){
                    qWarning() << QString("RemoteControlServer message error: checksum is incorrect.");
                    data.remove(0, 1);
                    continue;
                }
                else{
                    switch (hdr->type) {
                    case RC_MSG_CONFIG_OLD :
                    case RC_MSG_CONFIG :
                       qWarning() << "RemoteControlServer message warning: unexpected type of message";
                       break;
                    case RC_MSG_CMD_OLD :
                    case RC_MSG_CMD :
                       command_line_parse(tsd_data.mid(2*sizeof (qint32)));
                       break;
                    }
                    data.remove(0, static_cast<int>(hdr->len));
                }
            }
        }
    }
    catch(...) {
        qWarning()<<"Exception in RemoteControlServer::readCommandRequest()";
    }
}

void RemoteControlServer::command_line_parse(QByteArray data)
{
    QString mess = QString::fromUtf8(data.data(), data.size());
    for (QString command : mess.split(RC_CMD_END)) {
        command = command.trimmed();
        qDebug() << QString("Remote command:").append(command);
        QStringList commandParts = command.split(" ", QString::SkipEmptyParts);
        if(commandParts.isEmpty())
            continue;
        if(commandParts.first() == RC_CMD_START
                || commandParts.first() == RC_CMD_STOP) {
            // start <globIndexStr> <runNumber> [startEvNumber]
            // stop <globIndexStr> <runNumber> [stopEvNumber]

            qInfo() << QString("RemoteControlServer received command:").append(commandParts.first());
            bool ok = commandParts.size()>=3;

            int runNumber=0;
            if(ok)
                runNumber = commandParts[2].toInt(&ok);

            quint64 evNumber=0;
            if(ok && commandParts.size()>3)
                evNumber = commandParts[3].toULongLong(&ok);

            if(ok) {
                const QString &runIndex = commandParts[1];
                if(commandParts.first() == RC_CMD_START)
                    emit startRequest(RcRunId(runIndex, runNumber), evNumber);
                else
                    emit stopRequest(RcRunId(runIndex, runNumber), evNumber);
            }
            else{
                qWarning() << "RemoteControlServer can't define evNumber in command:"
                           << command;
            }
        } else if(commandParts.first() == RC_CMD_STATE) {
            // state
            qInfo() << QString("RemoteControlServer received command:").append(RC_CMD_STATE);
            if(commandParts[1] != ""){
                emit timeToStart(commandParts[1]);
            } else{
                emit stateRequest();
            }
        }else if(commandParts.first() == RC_CMD_EV_NUM_CHECK) {
            // EvNumCheck
//            qInfo() << QString("Received command:").append(RC_CMD_EV_NUM_CHECK);
            bool ok = commandParts.size()==2;

            int checkIndex = 0;
            if(ok)
                checkIndex = commandParts[1].toInt(&ok);
            if(ok)
                emit evNumCheckRequest(checkIndex);
            else
                qWarning() << "RemoteControlServer can't define checkIndex in command:"
                           << command;
        }
    }
}

void RemoteControlServer::sendState(QString state, bool read4Run)
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_STATE)
            .arg(state)
            .arg(RC_CMD_END);
    send_command_message(mess);

    mess = QString("%1 %2 %3")
            .arg(RC_CMD_READY4RUN)
            .arg(read4Run ? "1" : "0")
            .arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::onRunError(const RcRunId &runId, QString runError)
{
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_ERROR)
            .arg(QString(runId))
            .arg(runError.replace("\r", "").replace("\n","").trimmed())
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onRunWarning(const RcRunId &runId, QString warnStr)
{
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_WARN)
            .arg(QString(runId))
            .arg(warnStr.replace("\r", "").replace("\n","").trimmed())
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(runId.runIndex.isEmpty())
        return;
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_BACK_PRESSURE)
            .arg(QString(runId))
            .arg(backPressure ? "1" : "0")
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onVmeFlag(bool hasVme)
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_VME_FLAG)
            .arg(hasVme ? "1" : "0")
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::startListen(bool state)
{
    if(status.listening != state) {
        if(state) {
            listen();
        } else {
            close();
        }
        status.listening = state;
    }
    updateStatus();
}

void RemoteControlServer::completedCmd(bool success, QString cmd, int arg) const
{
    QString mess = QString("%1 %2 %3 %4")
            .arg(success ? RC_CMD_SUCCESS_PREFIX : RC_CMD_FAILED_PREFIX)
            .arg(cmd)
            .arg(arg)
            .arg(RC_CMD_END);
    send_command_message(mess);
    // sendLog(QString("Send to remote control:").append(mess.trimmed()));
}

void RemoteControlServer::listen()
{
    while(!qTcpServer->listen(QHostAddress::Any, status.preferedTcpPort)) {
//        QAbstractSocket::SocketError errNum = qTcpServer->serverError();
//        if(errNum==QAbstractSocket::AddressInUseError){
//            if(!portInUse){
//                QString mess = tr("Unable to start the remote control server on port %1: %2, restart in 5 s")
//                        .arg(tcpPort).arg(qTcpServer->errorString());
//                sendLog(mess, Qt::red);
//                portInUse=true;
//                qFatal(mess.toStdString().c_str());
//            }
//            exit(1);
//            sleep(5);
//            continue;
//        }
        QString s = tr("RemoteControlServer unable to listen on TCP port %1: %2")
                .arg(status.preferedTcpPort).arg(qTcpServer->errorString());
        qCritical() << s;
        throw std::runtime_error(s.toStdString());
    }
    status.tcpPort = qTcpServer->serverPort();

    qInfo() << QString("RemoteControlServer listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(status.tcpPort);
}

void RemoteControlServer::close()
{
    qTcpServer->close();
    if(tcpSocket){
        tcpSocket->abort();
        delete tcpSocket;
        tcpSocket = nullptr;
        status.portInUse=false;
    }
    qDebug() << QString("TcpServer closed on %1:%2")
                .arg(qTcpServer->serverAddress().toString())
                .arg(status.tcpPort);
}

void RemoteControlServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            status.portInUse=false;
        }
        if(tcpSocket==nullptr){
            tcpSocket = pendingConnection;
            status.portInUse=true;
            status.peerHost = tcpSocket->peerAddress();
            status.peerPort = tcpSocket->peerPort();
            connect(tcpSocket, &QTcpSocket::readyRead,
                    this, &RemoteControlServer::readCommandRequest);
            connect(tcpSocket, &QTcpSocket::disconnected,
                    this, &RemoteControlServer::onDisconnected);
            QHostInfo info = QHostInfo::fromName(status.peerHost.toString());
            qInfo() << QString("RemoteControlServer accepted connection from %1:%2")
                       .arg(info.hostName())
                       .arg(tcpSocket->peerPort());
            updateStatus();
            sendDescription();
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(tcpSocket==nullptr){
        qDebug()<<"There are no connections.";
    }
}

bool RemoteControlServer::isConnectionAlive() const
{
    return tcpSocket != nullptr && tcpSocket->isValid() && (tcpSocket->state() == QTcpSocket::ConnectedState);
}

void RemoteControlServer::completedStartStop(bool start, bool success, const RcRunId &runId) const
{
    QString mess = QString("%1 %2 %3 %4")
            .arg(success ? RC_CMD_SUCCESS_PREFIX : RC_CMD_FAILED_PREFIX)
            .arg(start ? RC_CMD_START : RC_CMD_STOP)
            .arg(QString(runId))
            .arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::updateStatus() const
{
    emit connectionStatusUpdated(status);
}

void RemoteControlServer::sendDescription() const
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_DESCR).arg(descr).arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::onDisconnected()
{
    status.portInUse=false;
    updateStatus();
    qInfo() << "RemoteControlServer disconnected";
}

void RemoteControlServer::changeProgramDescription(const ProgramDescription &pd)
{
    descr = pd.getName();
    sendDescription();
}

void RemoteControlServer::evNumCheckResponse(int checkIndex, const EvNumSet &evNums) const
{
    QString evNumsStr;
    for (quint64 evNum : evNums) {
        evNumsStr += QString::number(evNum)+" ";
    }
    QString mess = QString("%1 %2 %3 %4")
            .arg(RC_CMD_EV_NUM_CHECK)
            .arg(checkIndex)
            .arg(evNumsStr)
            .arg(RC_CMD_END);
    send_command_message(mess);
}
