#include "ZmqPublisher.h"

#include <QTimer>
#include <QDebug>

#include <zmq.hpp>

namespace  {

bool read_msg(void* s, zmq_event_t* event, char* ep)
{
    int rc ;
    zmq_msg_t msg1;  // binary part
    zmq_msg_init (&msg1);
    zmq_msg_t msg2;  //  address part
    zmq_msg_init (&msg2);
    rc = zmq_msg_recv (&msg1, s, ZMQ_DONTWAIT);
    if (rc == -1 && zmq_errno() == ETERM)
        return false;
    assert (rc != -1);
    assert (zmq_msg_more(&msg1) != 0);
    rc = zmq_msg_recv (&msg2, s, 0);
    if (rc == -1 && zmq_errno() == ETERM)
        return false;
    assert (rc != -1);
    assert (zmq_msg_more(&msg2) == 0);
    // copy binary data to event struct
    const char* data = (char*)zmq_msg_data(&msg1);
    memcpy(&(event->event), data, sizeof(event->event));
    memcpy(&(event->value), data+sizeof(event->event), sizeof(event->value));
    // copy address part
    const size_t len = zmq_msg_size(&msg2) ;
//    ep = memcpy(ep, zmq_msg_data(&msg2), len);
    memcpy(ep, zmq_msg_data(&msg2), len);
    *(ep + len) = 0 ;
    return true;
}

void close_zero_linger(void *socket)
{
    int linger = 0;
    zmq_setsockopt(socket, ZMQ_LINGER, &linger, sizeof(linger));
    zmq_close(socket);
}
}


ZmqPublisher::ZmqPublisher(QObject *parent) :
    QObject(parent),
    test_timer(new QTimer(this)),
    status_timer(new QTimer(this)),
    context(nullptr),
    socket(nullptr),
    socket_mon(nullptr)
{
    connect(test_timer, &QTimer::timeout,
            this, &ZmqPublisher::test_send);
    test_timer->setSingleShot(false);
    test_timer->setInterval(1000);

    connect(status_timer, &QTimer::timeout,
            this, &ZmqPublisher::send_status);
    status_timer->setSingleShot(false);
    status_timer->start(1000);
}

ZmqPublisher::~ZmqPublisher()
{
    closeSocket();
}

void ZmqPublisher::test_send()
{
    if(!socket || !cfg.test_mode)
        return;

    zmq::message_t msg(14);
    memset(msg.data(), 0, 14);
    auto buf = reinterpret_cast<char*>(msg.data());
    ++test_id;
    sprintf(buf, "%03d %10d", 500+test_id%5, test_id);
    socket->send(msg, ZMQ_DONTWAIT);
    if(cfg.print_messages)
        qInfo().noquote() << "[ZMQ::test]" << buf;
    ++status.packages;
}

void ZmqPublisher::send_status()
{
    if(socket_mon) {
        zmq_event_t ev;
        char address[1024];
        while(read_msg(socket_mon, &ev, address)) {
            qInfo() << "get_monitor_event" << eventStr(ev.event) << address << dec << ev.value;
        }
    }
    emit statusUpdated(status);
}

void ZmqPublisher::config_changed(const ZmqConfig &config)
{
    bool recreate = (cfg.port != config.port) || (cfg.enabled != config.enabled);
    cfg = config;

    if(recreate) {
        closeSocket();
        if(cfg.enabled)
            createSocket();
    }

    if(cfg.enabled && cfg.test_mode) {
        if(!test_timer->isActive())
            test_timer->start();
    } else {
        test_timer->stop();
    }
    send_status();
}

bool ZmqPublisher::createSocket()
{
    status.working = false;

    if(!context)
        context = new zmq::context_t(1);
    Q_ASSERT(!socket);
    socket = new zmq::socket_t(*context, ZMQ_PUB);

    try {
        QString addr = QString("tcp://*:%1").arg(cfg.port ? QString::number(cfg.port) : "*");

        socket->bind(addr.toStdString().c_str());
        if(cfg.port) {
            status.port = cfg.port;
        } else {
            char buf[60];
            size_t len=50;
            memset(buf, 0, 60);
            socket->getsockopt(ZMQ_LAST_ENDPOINT, &buf, &len);
            auto port_str = strrchr(buf, ':');
            status.port = atoi(&port_str[1]);
        }
        qInfo() << "ZMQ pulisher started at port" << status.port;
        status.working = true;
    }  catch (const zmq::error_t &err) {
        qWarning().noquote() << "zmq bind error " << err.num() << err.what();
    }

//    connectMonitor();

    if(status.working && cfg.test_mode)
        test_send();
    return status.working;
}

void ZmqPublisher::closeSocket()
{
    if(socket) {
        close_zero_linger(socket);
        delete socket;
        socket = nullptr;
    }
    if(socket_mon) {
        close_zero_linger(socket_mon);
        delete socket_mon;
        socket_mon = nullptr;
    }
    if(context) {
        zmq_ctx_term(context);
        context->close();
        delete context;
        context = nullptr;
    }
    status.working = false;
}

bool ZmqPublisher::connectMonitor()
{
    if(!socket) return false;
    if(socket_mon) return true;

    if(!socket_mon)
        socket_mon = new zmq::socket_t(*context, ZMQ_PAIR);

    int rc = zmq_socket_monitor (socket, "inproc://monitor-client", ZMQ_EVENT_ALL);
    if(rc != 0) {
        int err = zmq_errno();
        qWarning() << "zmq_socket_monitor err:"<< err << zmq_strerror(err);
        return false;
    }
    assert (rc == 0);

     // Connect these to the inproc endpoints so they'll get events
     rc = zmq_connect (socket_mon, "inproc://monitor-client");
     assert (rc == 0);
     return true;
}

int ZmqPublisher::get_monitor_event(void *monitor, int *value, char **address)
{
    // First frame in message contains event number and value
    zmq_msg_t msg;
    zmq_msg_init (&msg);
    if (zmq_msg_recv (&msg, monitor, 0) == -1)
        return -1; // Interruped, presumably
    assert (zmq_msg_more (&msg));

    uint8_t *data = reinterpret_cast<uint8_t *>(zmq_msg_data (&msg));
    uint16_t event = *reinterpret_cast<uint16_t *>(data);
    if (value)
        *value = *reinterpret_cast<int32_t *>(data + 2);

    // Second frame in message contains event address
    zmq_msg_init (&msg);
    if (zmq_msg_recv (&msg, monitor, 0) == -1)
        return -1; // Interruped, presumably
    assert (!zmq_msg_more (&msg));

    if (address) {
        uint8_t *data = reinterpret_cast<uint8_t *>(zmq_msg_data (&msg));
        size_t size = zmq_msg_size (&msg);
        *address = reinterpret_cast<char *>(malloc (size + 1));
        memcpy (*address, data, size);
        *address [size] = 0;
    }
    return event;
}

QString ZmqPublisher::eventStr(int evNum) const
{
    switch (evNum) {
#if ZMQ_VERSION > ZMQ_MAKE_VERSION(4, 6, 0)
    case ZMQ_STREAM_NOTIFY: return "ZMQ_STREAM_NOTIFY";
#endif

    case ZMQ_EVENT_CONNECTED:        return "ZMQ_EVENT_CONNECTED";
    case ZMQ_EVENT_CONNECT_DELAYED:  return "ZMQ_EVENT_CONNECT_DELAYED";
    case ZMQ_EVENT_CONNECT_RETRIED : return "ZMQ_EVENT_CONNECT_RETRIED";
    case ZMQ_EVENT_LISTENING :       return "ZMQ_EVENT_LISTENING";
    case ZMQ_EVENT_BIND_FAILED :     return "ZMQ_EVENT_BIND_FAILED";
    case ZMQ_EVENT_ACCEPTED :        return "ZMQ_EVENT_ACCEPTED";
    case ZMQ_EVENT_ACCEPT_FAILED :   return "ZMQ_EVENT_ACCEPT_FAILED";
    case ZMQ_EVENT_CLOSED :          return "ZMQ_EVENT_CLOSED";
    case ZMQ_EVENT_CLOSE_FAILED :    return "ZMQ_EVENT_CLOSE_FAILED";
    case ZMQ_EVENT_DISCONNECTED :    return "ZMQ_EVENT_DISCONNECTED";
    case ZMQ_EVENT_MONITOR_STOPPED : return "ZMQ_EVENT_MONITOR_STOPPED";
    case ZMQ_EVENT_ALL :             return "ZMQ_EVENT_ALL";
    }
    return QString("0x%1 %2").arg(evNum, 4, 16, QChar('0') ).arg(evNum, 5);
}

void ZmqPublisher::publish(const QString &mess_str)
{
    const size_t mess_size = static_cast<size_t>(mess_str.size());
    zmq::message_t msg(mess_size);
//    zmq::message_t msg(mess_size+2);
    if(mess_size)
        memcpy (msg.data (), mess_str.toStdString().c_str(), mess_size);
//    auto buf = reinterpret_cast<char*>(msg.data());
//    buf[mess_size]='\n';
//    buf[mess_size+1]=0;
    socket->send(msg, ZMQ_DONTWAIT);
    ++status.packages;
    if(cfg.print_messages)
        qInfo().noquote() << "[ZMQ]" << mess_str;
}
